/**
 *	@file we_application.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora

 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file defines the class required for creating an Application.
 *
 *	It manages the control flow of an application.
*/

#ifndef WE_APPLICATION_H_
#define WE_APPLICATION_H_

#include <memory>
#include <string>
#include "../../common/include/definitions.h"
#include "../../common/include/we_size.h"

namespace WE
{

class ToolkitWEApplication;
class WEWidget;
class WEColor;


/**
 *	@class WEApplication
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief This class is the base class that shall be created for any application using the WE Application
 *	Framework. It performs event handling, meaning that it receives events from the underlying window system
 *	and dispatches them to the relevant widgets.
 *
 *	@namespace WE
 *
 *	@warning If not inheriting WEApplication then do not instantiate WEApplication as follows
 *			WEApplication* app = new WEApplication(argc,argv,true);
 *			Here true indicates that the application is application which requires a method to be invoked
 *			during IDLE Processing, which would internally invoke mainLoop() unnecessarily
 *			An application can be created as follows:
 *			WEApplication* app = new WEApplication(argc,argv);
 *
 *
*/
class WEApplication
{

public:

	/**
	 *	@fn WEApplication(int& argc,char *argv[],bool mainLoop = false)
	 *
	 *	@brief Constructor of the WEApplication. It must be created before any other objects
	 *	(widgets) related to the user interface are created.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] argc command line argument argc
	 *	@param[in] argv command line argument argv
	 *	@param[in] mainLoop True indicates the application requires a method(mainLoop) to
	 *	be invoked during idle processing this method can be overridden in the main class of teh Application inheriting WEApplication.
	 *
	 *
	 *	A sample application can be created as follows:
	 *
	 *	WEApplication* app =  new WEApplication(argc,argv);
	 *	//Application Logic
	 *	app->exec();
	 *
	 * The Sample Application can also be created by inheriting WEApplication
	 *
	 * class SampleApplication : public WEApplication
	 * {
	 * 		SampleApplication(int& argc,char *argv[]):WEApplication(argc,argv){}
	 * }
	 *
	 *
	 *	If an application requires a method to be invoked during IDLE processing of the underlying System
	 *	. For such purposes	an application can specify true as the third parameter to WEApplication(by default False).
	 *	Specifying true will allow the application to override the mainLoop method which will be invoked during IDLE processing
	 *	This could be used in a State Driven Application. Also an application must create widgets with in the Main thread only.
	 *	For E.g.
	 * 	class SampleApplication : public WEApplication
	 * 	{
	 * 		SampleApplication(int& argc,char *argv[]):WEApplication(argc,argv){}
	 * 		void mainLoop();
	 * 	}
	 *
	 *	SampleApplication::mainLoop()
	 *	{
	 *		//Application Logic
	 *	}
	*/
	WEApplication(int& p_argc,char *p_argv[],bool p_mainLoopApplication = false);
	//don't allow copy
	WEApplication& operator=(const WEApplication & p_copy) = delete;
	WEApplication(const WEApplication & p_copy) = delete;

	/**
	 *	@fn execute()
	 *
	 *	@brief This method of is used to start event handling. The main
	 *	It receives paint event from the underlying window system and
	 *	dispatches these to the application widgets.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns Returns the value that was set to exit().
	 *
	 *	@warning It is the responsibility of the user creating the instance
	 *	to delete the instance.
	 *
	 *	This method must be called for the widgets to receive the events.
	*/
	int execute();

	/**
	 *	@fn exit(int returnCode=0)
	 *
	 *	@brief This method tells the application to exit with a return code.By Default 0.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param returnCode The execute() function returns this returnCode.
	 *
	*/
	void exit(int p_returnCode=0);

	/**
	 *	@fn mainLoop()
	 *
	 *	@brief This method can be overridden by application which require a method tobe invoked continuously during
	 *	IDLE processing of the System. Such application can be created by specifying true for the main loop in the
	 *	constructor of WEApplication.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@see WEApplication()
	*/
	virtual void mainLoop();

	/**
	 *	@fn ~WEApplication()
	 *
	 *	@brief Destrcutor
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	*/
	virtual ~WEApplication();

	/**
	 *	@fn isMainLoopApplication()
	 *
	 *	@brief Returns whether the application is a mainLoop Application.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns True if the application invokes a method during IDLE Processing.
	*/
	bool isMainLoopApplication() const;

	/**
	 *	@fn void enablePainting(bool p_paint)
	 *
	 *	@brief Enable or disable painting
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	This method calls the similar one from QTWEServer.
	 *
	 *	@warning Has no effect on a Linux PC
	 *
	 *	@param p_paint true if painting shall be enabled, false if painting shall be disabled
	 */
	void enablePainting(bool p_paint);

	/**
	 *	@fn void setCursorVisible(bool p_visible)
	 *
	 *	@brief Show or hide the mouse cursor
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_visible true if the mouse cursor shall be shown, false if the mouse cursor shall be hidden
	 *
	 *	@warning Has no effect on a windows PC
	 */
	void setCursorVisible(bool p_visible);

	/**
	 *	@fn const WESize getDisplayResolution(void)
	 *
	 *	@brief Retrieve the pixel resolution of the display on which the application runs on
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns A WESize instance that contains the width and height in pixel of the display
	 */
	static const WESize getDisplayResolution(void);

	/**
	 *	@fn const enum DisplayOrientation getDisplayOrientation(void)
	 *
	 *	@brief Retrieve the display orientation
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns The Orientation of the display
	 */
	static enum DisplayOrientation getDisplayOrientation(void);

	/**
	 *	@brief Main task of this method is to cast 'enum DisplayOrientation' to int.
	 *
	 *	@see ToolkitWEApplication::setScreenRotation(int)
	 */
	int setDisplayOrientation(enum DisplayOrientation p_orientation);

	/**
	 *	@fn int setTerminationListenerFunction(void (*functionPointer)(void*), bool p_terminateWithLastWindow = true)
	 *
	 *	@brief Set a callback function that gets called when the process gets killed from outside.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param functionPointer pointer to the function that shall be called
	 *	@param p_terminateWithLastWindow Set to "true" if "functionPointer" should be called when this' last window is closed
	 *
	 *	@returns WE_SUCCESS if the callback function was correctly set
	 *	@returns WE_GENERAL_ERROR else
	 *
	 *	The implementation of this function is OS dependent.
	 *	On Linux system signal handlers are installed which call the
	 *	function pointed to by functionPointer.
	 *
	 *	p_terminateWithLastWindow may not be implemented if there should be no support from the toolkit used.
	 *	It is supported for Qt though.
	 */
	int setTerminationListenerFunction(void (*p_functionPointer)(void*), bool p_terminateWithLastWindow = true);

	void setApplicationName(const std::string &);
	const std::string & getApplicationName( void ) const;

	void setApplicationVersion(const std::string &);
	const std::string & getApplicationVersion( void ) const;

	void setActiveWindow(WEWidget * p_widget);

	/**
	 *	@fn bool isServerAppln()
	 *
	 *	@brief This method check if the application context in which
	 *	this method is called, is a Qt Server application or Client application.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@returns TRUE if application is server
	 *
	 */
	bool isServerAppln();

	/**
	 *	@fn bool isApplicationFocussed()
	 *
	 *	@brief This method check if the application currently has focused an
	 *	it is an active application
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@returns TRUE if application is focused, FALSE otherwise
	 *
	 */
	bool isApplicationFocussed();

	/**
	 *	@fn void processEvents()
	 *
	 *	@brief This method can be called for immediate handling of events in current context.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@NOTE: This function was introduced to give Qt more chances to process input events while
	 *	       our large stage engine is running in Qt idle processing. Result is mainly the faster
	 *	       response on key, mouse and encoder events.
	 */
	void processEvents();

	/**
	 *	@fn void setTouchInput(bool p_setTouch)
	 *
	 *	@brief Enable or disable touch input events
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@warning Has no effect on a Linux PC
	 *
	 *	@param p_setTouch true if touch input events shall be enabled, false if touch input events shall be disabled
	 */
	void setTouchInput(bool p_setTouch);

	/**
	 * Return the flag indicating if touch shall be enabled or not (set via setTouchInput())
	 */
	bool isTouchEnabled(void);

	/**
	 *	Reloads mouse driver on Qt QWS systems. Can be used after touch calibration to use the new calibration values.
	 *	Currently only works on QWS systems.
	 */
	void reloadMouse(void);

	void setCursorBlank();

	/**
	 *	@fn void setBackgroundColor(WEColor &p_color)
	 *
	 *	@brief Sets the background color and transparency for the application.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_color The color (red, green, blue and transparency) to apply to for the background
	 */
	void setBackgroundColor(WEColor &p_color);


	/**
	 *	@fn void tryEnableSquish(int p_port);
	 *
	 *	@brief 	Try to enable Squish (load Squish Plugin) for Test-Automation.
	 *
	 *	@param p_port Port for Squish connection
	 *
	 *	NOTE:
	 *	Currently this method is only available for Topcon OS  (OE Angstroem).
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 */
	void tryEnableSquish(int p_port);

	/**
	 *	@fn static void sigHandler (int sig)
	 *
	 *	@brief Signal handler for all "quit" signals
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param sig The number of the received signal
	 *
	 *	@returns absolutely nothing
	 *
	 *	The handler functions just calls the callback function set with
	 *	int WEApplication::setTerminationListenerFunction(void (*functionPointer)(void*))
	 */
	static void sigHandler (int p_sig);
private:
	/**
	 *	@fn int computeDisplayOrientation(void)
	 *
	 *	@brief Computes the display orientation and stores it in member variable
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS if orientation was successfully read.
	 *	@returns WE_READ_ERROR if orientation was not successfully read and set to the default value (no rotation).
	 */
	int computeDisplayOrientation(void);

	/**
	 *	@fn int computeDisplayResolution(void)
	 *
	 *	@brief Computes the display resolution and stores it in member variable
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS
	 */
	int computeDisplayResolution(void);


	std::unique_ptr<ToolkitWEApplication> toolkitWEApplication;		//Reference to the ToolkitApplication
	bool mainLoopApplication;				//Whether the application is a mainLoop Application
	static WESize displayResolution;			//!< display resolution, (-1,-1) when resolution is unknown
	static enum DisplayOrientation displayOrientation;	//!< display orientation, ORIENTATION_UNKNOWN when orientation is not available

	static WEApplication * applicationInstance;	//!< The instance of this application (no real singleton handling needed since there can be only one application)
};
}
#endif /* WE_APPLICATION_H_ */
