/**
 *	@file qt_we_server.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Definition of QTWEServer class
 *
 *	@see qt_we_server_x11.cpp for an attempt to explain the purpose of this class.
 */

#ifndef QT_WE_SERVER_H_
#define QT_WE_SERVER_H_

#include <memory> //unique_ptr
#include <iostream>
#include <vector>

#include <QKeyEvent>

#include "../../../../win_mgr/include/we_window_info.h"

class QAbstractNativeEventFilter;


namespace WE
{
class WEColor;
class QTWEApplication;
class QtWEKeyboardFilter;

/**
 *	@class QTWEServer
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Class responsible for calling QWSServer related functions
 *
 *	@namespace WE
 *
 *	@warning To use the window managing functionality, this class is to be used in WEWinMgr class.
 *
 *	@see WEWinMgr
 */
class QTWEServer
{
public:
	/**
	 *	@fn ~QTWEServer()
	 *
	 *	@brief Destructor of QTWEServer
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	virtual ~QTWEServer() = default;

	/**
	 *	@fn static QTWEServer* createInstance(void)
	 *
	 *	@brief Static method that creates and returns a pointer to the instance of this class
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_wmApp The application providing the window manager process (nullptr for QWS based devices)
	 *
	 *	@returns Pointer to the instance of this class
	 *
	 *	If the class has not been instantiated, yet, this function creates
	 *	the object and returns the address of it. When calling multiple times, this function
	 *	always returns a pointer to the same instance (singleton).
	 */
	static QTWEServer* createInstance(QTWEApplication * p_wmApp = nullptr);

	/**
	 *	@fn static QTWEServer* getInstance()
	 *
	 *	@brief static method which returns the pointer to the current instance of this class
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Pointer to the existing instance, otherwise nullptr.
	 *
	 *	This function returns a pointer to the current available instance (singleton) of this class. It does not
	 *	create a new instance when no instance is available. When using carefully (i.e. being sure that
	 *	an instance exists, this function can be used for doing single line function calls to the instance.
	 */
	static QTWEServer* getInstance()
	{
		//If no instance of the QTWEServer is existing, abort process
		if (nullptr == theInstance.get())
		{
			std::cerr << "QTWEServer::getInstance() called without creating an QTWEServer instance before." << std::endl;
			std::cerr << "Use QTWEServer::createInstance() in advance to create one!" << std::endl;
			std::cerr << "Aborting..." << std::endl;
			abort();
		}

		//return a pointer to the instance of the QTWEServer
		return theInstance.get();
	}

	/**
	 *	@fn static void enablePainting(bool p_paint)
	 *
	 *	@brief Enable or disable painting
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@warning Has no effect on a Linux PC
	 *
	 *	@param p_paint true if painting shall be enabled, false if painting shall be disabled
	 */
	void enablePainting(bool p_paint);

	/**
	 *	@fn static void setCursorVisible(bool p_visible)
	 *
	 *	@brief Show or hide the mouse cursor
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_visible true if the mouse cursor shall be shown, false if the mouse cursor shall be hidden
	 *
	 *	@warning Has no effect on a windows PC
	 */
	void setCursorVisible(bool p_visible);

	/**
	 *	@fn void setBackgroundColor(WEColor &color, bool p_useBrush = true)
	 *
	 *	@brief Sets the background color, transparency and brush usage for the server.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param color The color (red, green, blue and transparency) to apply to for the background
	 *	@param p_useBrush Allows enabling and disabling of using a brush
	 *		when painting the background. Default value: Brush enabled.
	 */
	void setBackgroundColor(WEColor &p_color, bool p_useBrush = true);

	/**
	 *	@fn std::vector<WEWindowInfo> clientWindows()
	 *
	 *	@brief This method provides the list of the Client windows available.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@returns returns the vector list of the client windows
	 *
	 */
	virtual std::vector<WEWindowInfo> clientWindows()
	{
		std::vector<WEWindowInfo> runningApps;

		return runningApps;
	}

	/**
	 *	@fn void activateWindow(WEWindowInfo p_winInfo)
	 *
	 *	@brief This method provides the functionality to activate the
	 *	application window passed as an input parameter.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@param[in] p_winInfo context of the current application window
	 *
	 *	@returns at present always return WE_SUCCESS
	 *
	 *	This method uses the Qt's QWS functionality to raise selected
	 *	window.
	 *	It traversed through the complete list of windows, raise/show
	 * 	the next application window and lower/hide all others except
	 *	the window managing pclient.
	 *	By using the hide/show techniques we can easily route the events
	 *	from the Qt to the focused application. Qt Event mechanism only
	 *	transfer the events to the active applications and by lowering the
	 *	application, the application become inactive so it does not receives
	 *	the events.
	 *	So at a given instance, only two applications are active, first one
	 *	is the window managing server pclient application and second one
	 *	is the user application which is currently focused.
	 */
	virtual std::int32_t activateWindow(WEWindowInfo /*p_winInfo*/)
	{
		return 0;
	}

	/**
	 *	@fn WEWindowInfo currentActiveWindow()
	 *
	 *	@brief This method returns the current focussed window that is visible to the user.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@returns context of the current application window
	 *
	 *	This method uses the Qt's QWSServer::windowAt function to retrieve
	 *	the current window. In our panels we always draw the application
	 *	window in full screen mode so if we can get the window at point 1,1 location
	 *	then we could easily get the current focused window that is visible to the user.
	 *
	 *	NOTE: Have to delete the memory used by the return pointer when it is used.
	 *
	 */
	WEWindowInfo currentActiveWindow();

	/**
	 *	@fn static void setTouchInput(bool p_setTouch)
	 *
	 *	@brief Enable or disable touch input events
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@warning Has no effect on a Linux PC
	 *
	 *	@param p_setTouch true if touch input events shall be enabled, false if touch input events shall be disabled
	 */
	void setTouchInput(bool p_setTouch);


	/**
	 *	Documentation from 'we_application.h':
	 *	Reloads mouse driver on Qt QWS systems. Can be used after touch calibration to use the new calibration values.
	 *	Currently only works on QWS systems.
	 */
	void reloadMouse(void);


	/**
	 *	@fn int setAlarmPriority(unsigned int p_winID, int p_priority)
	 *
	 *	@brief Set alarm priority for a certain window
	 *
	 *	@author mlenhart@topcon.com
	 *	@author Michael Lenhart
	 *
	 *	@param p_winID ID of the window whose alarm priority changed
	 *	@param p_priority priority of the alarm, -1 for no alarm
	 *
	 *	@return An error code: WE_SUCCESS in case of success.
	 *	@return WE_SOCKET_ERROR if there is no connection to window manager.
	 *	@return WE_SEND_ERROR if no message could be sent.
	 *	@return WE_GENERAL_ERROR for other errors.
	 *
	 *	On QWS based systems this will contact the window manager/server via QCopChannel.
	 *	On X11 based systems a message will be sent via X server connection.
	 */
	static int setAlarmPriority(unsigned int p_winID, int p_priority);


	/**
	 *	@fn int sendToggleMassage(void)
	 *
	 *	@brief Send message to window manager in order to switch to next window
	 *
	 *	@author mlenhart@topcon.com
	 *	@author Michael Lenhart
	 *
	 *	@return An error code: WE_SUCCESS in case of success.
	 *	@return WE_SOCKET_ERROR if there is no connection to window manager.
	 *	@return WE_SEND_ERROR if no message could be sent.
	 *	@return WE_GENERAL_ERROR for other errors.
	 *
	 *	On QWS based systems this will contact the window manager/server via QCopChannel.
	 *	On X11 based systems a message will be sent via X server connection.
	 */
	static int sendToggleMassage(void);


	/**
	 * Depending on the used window system (X11 or QWS), the screen rotation is either retrieved
	 * via xcb-randr on X11 systems or by reading the rotation from the QWS_DISPLAY environment variable on QWS systems.
	 * @return 0, 90, 180, 270
	 */
	static int getScreenRotation(void);


	/**
	 *	@brief Introduced to rotate display at runtime.
	 *
	 *	@param p_screenRotation Set to 0, 90, 180 or 270. All other values may be considered invalid.
	 *
	 *	@returns WE_SUCCESS in case of success.
	 *	@returns WE_NOT_SUPPORTED_ERROR if parameter is not set to one of the supported values.
	 *	@returns WE_ALREADY_EXISTS_ERROR if rotation is already set to p_rotation.
	 *
	 *	For QWS and X11 this method rotates the display immediately at runtime.
	 *	For Wayland runtime rotation has to be supported by the compositor - as
	 *	long as Weston is used the display rotation is considered a configuration
	 *	to be done in the 'weston.ini' file before Weston gets started.
	 */
	static int setScreenRotation(int p_screenRotation);


protected:
	explicit QTWEServer(QTWEApplication * p_wmApp);
	QTWEServer() = default;

	static inline std::unique_ptr<QTWEServer> theInstance{};
	static int screenRotation;		//!< For PC (== egspro): Just return in 'get' method what was set before (i.e. don't rotate anything but remember the setting)

	QTWEApplication * wmApp{nullptr};		//!< At least for Qt5 + X11 we need the application to communicate with the X server.
	//QtWEKeyboardFilter * keyboardFilter;	//!< We need to store the keyboard filter because it has to be called inside the sendKey() method.
	QAbstractNativeEventFilter * nativeEventFilter{nullptr};	//!< For X11 we need to get XCB events
	QObject * keyFilter{nullptr};			//!< QtWEKeyboardFilter is implemented for QWS plus we prefer to have the filter where the WM code is.
};
} // namespace WE

#endif /* QT_WE_SERVER_H_ */
