/**
 *	@file we_win_mgr.h
 *
 *	@author herpreet.sura@aricent.com
 *	@author Herpreet Sura
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Provides the class declaration of the interfaces that are
 *	required for the Window manager to operate.
 *
 */

#ifndef WE_WIN_MGR_H_
#define WE_WIN_MGR_H_

#include <vector>
#include <map>

#include "we_window_info.h"


namespace WE
{
/**
 *	@class WEWinMgr

 *	@brief Common methods managing windows in combination with QTWEServer.
 *
 *	@author herpreet.sura@aricent.com
 *	@author Herpreet Sura
 *
 *	This class provides the interfaces to get the list top
 *	level windows of the currently running applications. After the user
 *	selection of the desired application to launch, this also raise
 *	the window to make it in front.
 *
 *	@see WE::QTWEServer
 */
class WEWinMgr
{
public:
	/**
	 *	@fn WEWinMgr()
	 *
	 *	@brief Deleted constructor for the window manager class.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	Constructor is deleted since window manager only contains static methods and members and we
	 *	expect that we never will have an instance of this class.
	 */
	WEWinMgr() = delete;
	~WEWinMgr() = delete;

	/**
	 *	@fn static std::vector<WEWindowInfo> getClientWindows()
	 *
	 *	@brief This method provides the list of the running client windows
	 *
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@returns returns the vector list of the client applications
	 *
	 *	This method uses the QTWEServer's getClientWindows() interface method
	 *	to provide this functionality.
	 *
	 *	@see QTWEServer
	 */
	static std::vector<WEWindowInfo> getClientWindows();


	/**
	 *	@fn static std::int32_t activateWindow(WEWindowInfo p_appInfo)
	 *
	 *	@brief This method provides the functionality to activate the
	 *	application window passed as an input parameter.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@param[in] p_appInfo context of the current application window
	 *
	 *	@returns at present always return WE_SUCCESS
	 *
	 *	This method uses the QTWEServer's activateWindow() interface method
	 *	to provide this functionality.
	 *
	 *	@see QTWEServer
	 */
	static std::int32_t activateWindow(WEWindowInfo p_appInfo);


	/**
	 *	@fn static WEWindowInfo* currentActiveWindow()
	 *
	 *	@brief This method returns the current focussed window that is visible to the user.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@returns context of the current application window
	 *
	 *	This method uses the QTWEServer's currentActiveWindow() interface method
	 *	to provide this functionality.
	 *
	 *	@see QTWEServer
	 */
	static WEWindowInfo currentActiveWindow();


	/*
	 *	@fn static bool windowIsAlive(WE::WEWindowInfo p_winInfo)
	 *
	 *	@brief Figure out if p_winInfo still describes an existing window.
	 *
	 *	@param p_winInfo Description of the window to find.
	 *
	 */
	static bool windowIsAlive(WE::WEWindowInfo p_winInfo);


	/*
	 *	@fn static static WE::WEWindowInfo doWindowSwitch(void)
	 *
	 *	@brief Switch to next window
	 *
	 *	@param none
	 *
	 *	This is called when a "toggle" message is received indicating that the next window should be shown.
	 *
	 *	It is also used if a window is destroyed which was currently shown on screen.
	 */
	static WE::WEWindowInfo doWindowSwitch(void);



	/*
	 *	@fn static WE::WEWindowInfo findLivingWindowToActivate(int p_winID)
	 *
	 *	@brief Set alarm priority of a window and get application which is currently active resulting from that modification.
	 *
	 *	@param p_winID	ID of the window that should be shown. "-1" to only update current state.
	 *
	 *	@return The window that should come to front
	 *
	 *	This is called when either an alarm priority changed or an event suggests that a window
	 *	tries to come to front.
	 *
	 *	The window description returned is either (ordered by the algorithm's priority)
	 *	- the one for the alarm with highest priority
	 *	- the one belonging to p_winID (if it points to a valid window)
	 *	- the currently active window
	 *	- an alternative existing window
	 *	- a bogus description if there is no active window
	 */
	static WE::WEWindowInfo findLivingWindowToActivate(int p_winID = -1);


	/*
	 *	@fn static WE::WEWindowInfo newAlarmFromWindow(int winID, int priority)
	 *
	 *	@brief Set alarm priority of a window and get application which is currently active resulting from that modification.
	 *
	 *	@param p_winID	ID of the window that should be modified
	 *	@param p_priority alarm priority, "-1" for "no alarm"
	 *
	 *	This is called when an application sends a message indicating that the alarm priority of
	 *	(one of) its windows changed.
	 *
	 *	It is also used if a window is destroyed which showed an alarm when it was destroyed.
	 */
	static WE::WEWindowInfo newAlarmFromWindow(int p_winID, int p_priority);

	static WE::WEWindowInfo getActiveAlarm(void);

private:
	// Everything else uses "WE::WEWindowInfo" which contains a win ID - shouldn't we use this everywhere? - undecided since 2013
	static std::uint32_t lastActiveApplicationWinID;	// this now keeps the window ID of the active application

	static std::map<int, std::pair<int, unsigned int> > priorityWindows;	//!< first: winID, second: <priority, timestamp>
	static WE::WEWindowInfo alarmToBeShown;	//!< Used as fallback if an normal window should manage to appear above an alarm.
};
}

#endif /* WE_WIN_MGR_H_ */
