/**
 *	@file definitions.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Definitions to be used for target building.
 *
 *	@see build_options_PC.h build_options_A3.h
 *
 *	This file contains global defines (e.g. build target,
 *	CAN hardware, etc.). These definitions shall be used
 *	in the build_options_XX.h files.
 */

#ifndef DEFINITIONS_H_
#define DEFINITIONS_H_

namespace WE
{

/**
 *	@enum targets
 *
 *	@brief Contains the targets for which WEAPI can be built
 */
enum targets
{
	TARGET_PC_LINUX,	//!< target is a Linux PC
	TARGET_PC_WINDOWS,	//!< target is a Windows PC
	TARGET_EGS,		//!< target is EGSPRO or EGSCORE platform (e.g. A8, A6G2, B3, B4, B6)
	TARGET_B2		//!< target is OPUS B2
};

/**
 *	@enum CANHardware
 *
 *	@brief Contains supported CAN hardware
 */
enum CANHardware
{
	CAN_PEAK_LINUX,	//!< use the peak Linux hardware driver (PC or A5)
	CAN_FLEXCAN		//!< use the flexCAN Linux hardware driver (A3)
};

/**
 *	@enum CANSoftwareInterface
 *
 *	@brief Contains supported CAN interfaces
 */
enum CANSoftwareInterface
{
	CANIF_SOCKETCAN,	//!< IOCTL socket CAN can be used
	CANIF_PEAK_CHAR_DRIVER	//!< No socketCAN available, use character driver
};

/**
 *	@enum KeyLayout
 *
 *	@brief Contains the key layout of the build target
 *
 *	Use the numbers directly in keyboard layout configuration files.
 *	A combination with OPUS 44-1 is also possible, e.g. KEYS_A3_NO_KEYS | KEYS_OPUS_44_1
 */
enum KeyLayout
{
	KEYS_PC_KEYBOARD 	= 0x01,	//!< A normal PC keyboard shall be used
	KEYS_DEPRECATED		= 0x02,	//!< The keys of OPUS A5 shall be used
	KEYS_A3_NO_KEYS		= 0x03,	//!< The keys of OPUS A3 without any keys shall be used
	KEYS_A3_STANDARD	= 0x04,	//!< The standard keys of an OPUS A3 shall be used
	KEYS_A3_KEYPAD		= 0x05,	//!< The keys of OPUS A3 with a numeric keypad shall be used
	KEYS_A6_NO_KEYS		= 0x06,	//!< The keys of OPUS A6 without any keys shall be used
	KEYS_A6_STANDARD	= 0x07,	//!< The standard keys of an OPUS A6 shall be used
	KEYS_A6_KEYPAD		= 0x08,	//!< The keys of OPUS A6 with a numeric keypad shall be used
	KEYS_A6_STANDARD_PORTRAIT	= 0x09,	//!< The standard keys of an OPUS A6 Portrait shall be used
	KEYS_RVC		= 0x0A,	//!< The keys of OPUS RVC shall be used

	KEYS_OPUS_44_1		= 0x10  //!< The OPUS 44-1 can be used as a remote control via CAN bus
};

/**
 *	@enum DisplayOrientation
 *
 *	@brief Contains the display orientation
 */
enum DisplayOrientation
{
	DISPLAY_ORIENTATION_0 = 0,		// normal
	DISPLAY_ORIENTATION_90 = 90,		// 90 degree clock wise
	DISPLAY_ORIENTATION_180 = 180,		// top down
	DISPLAY_ORIENTATION_270 = 270,		// 90 degree counter clock wise
	DISPLAY_ORIENTATION_UNKNOWN = -1
};
}//end of namespace WE

#endif /* DEFINITIONS_H_ */
