/**
 *	@file we_color.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file defines the Colors based on RGB.
 *
 *	This file defines the colors based on RGB(Red, Green and Blue Components).
 */


#ifndef WE_COLOR_H_
#define WE_COLOR_H_

#include <QtGui/QColor>
#include <boost/serialization/export.hpp>
// AXIVION NextLine Generic-LocalInclude : For some reason Axivion thinks that we don't need the following include => but we do!
#include <boost/serialization/access.hpp>

namespace WE
{

/**
 *	@enum Color
 *
 *	@brief This enum defines the set of pre-defined colors.
 */
typedef enum
{
	//Taken From ISO 11783-6
	black 	= 0x0,		//!< Black Color with RGB value (0,0,0)
	white 	= 0x1,		//!< White Color with RGB value (255,255,255)
	green 	= 0x2,		//!< Green Color with RGB value (0,153,0)
	teal 	= 0x3,		//!< Teal Color with RGB value (0,153,153)
	maroon 	= 0x4,		//!< Maroon Color with RGB value (153,0,0)
	purple 	= 0x5,		//!< Purple Color with RGB value (153,0,153)
	olive 	= 0x6,		//!< Olive Color with RGB value	(153,153,0)
	silver 	= 0x7,		//!< Silver Color with RGB value (204,204,204)
	grey 	= 0x8,		//!< Grey Color with RGB value (153,153,153)
	blue 	= 0x9,		//!< Blue Color with RGB value (0,0,255)
	lime 	= 0xa,		//!< Lime Color with RGB value (0,255,0)
	cyan 	= 0xb,		//!< Cyan Color with RGB value (0,255,255)
	red 	= 0xc,		//!< Red Color with RGB value (255,0,0)
	magenta = 0xd,		//!< Magenta Color with RGB value (255,0,255)
	yellow 	= 0xe,		//!< Yellow Color with RGB value (255,255,0)
	navy 	= 0xf,		//!< Navy Color with RGB value (0,0,153)
	we_blue = 0x10,		//!< Wachendorff Blue with RGB value (0, 107, 179)
	nocolor = 0xffff	//!< Transparent (0,0,0) alpha = 0
} Color;

/**
 *	@enum Opaqueness
 *
 *	@brief This enum defines the set of pre-defined states of "opaque".
 */
enum Opaqueness
{
	transparent	= 0x00,	//!<completely see-through, invisible>
	opaque		= 0xff	//!<opaque - hiding everything behind/underneath>
};



/**
 *	@class WEColor
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief Class that contains color information.
 *
 *	This class contains color information i.e. Red, Green, Blue and Alpha
 *	components of a color.
 */
class WEColor
{
public:

	/**
	 *	@fn WEColor()
	 *
	 *	@brief Default Constructor
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 */
	WEColor() = default;


	/**
	 *	@fn WEColor::WEColor(unsigned int p_rgb, unsigned int p_alpha = WE::opaque)
	 *
	 *	@brief Constructor creating a color object from 24 bit RGB definition and optional alpha value.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_rgb Color definition in format 0x00RRGGBB
	 *	@param p_alpha Opaqueness value [0: transparent to 0xff: opaque]
	 *
	 *
	 */
	explicit WEColor(unsigned int p_rgb, unsigned int p_alpha = WE::opaque);

	/**
	 *	@fn WEColor(uint8 red, uint8 green, uint8 blue, uint8 alpha = WE::opaque)
	 *
	 *	@brief To construct a Color with the specified RGB Value.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_red Red Component of the Color
	 *	@param[in] p_green Green Component of the Color
	 *	@param[in] p_blue Blue Component of the Color
	 */
	explicit WEColor(std::uint8_t p_red, std::uint8_t p_green, std::uint8_t p_blue, std::uint8_t p_alpha = WE::opaque);


	/**
	 *	@fn WEColor(const WEColor &color)
	 *
	 *	@brief Copy Constructor.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] color Color of which the copy is to be created.
	 */
	WEColor(const WEColor &p_color) = default;

	/**
	 *	@fn WEColor(Color color)
	 *
	 *	@brief Convenience Constructor to initialise the color from
	 *	the pre-defined set of colors.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] color Color with which the WEColor is to be initialised.
	 */
	explicit WEColor(Color p_color);

	virtual ~WEColor() = default;


	/**
	 *	@fn getRed()
	 *
	 *	@brief Returns the red component of the color.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns Red component of the Color.
	 */
	std::uint8_t getRed() const;


	/**
	 *	@fn getGreen()
	 *
	 *	@brief Returns the green component of the color.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns Green component of the Color.
	 */
	std::uint8_t getGreen() const;


	/**
	 *	@fn getBlue()
	 *
	 *	@brief Returns the blue component of the color.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns Blue component of the Color.
	 */
	std::uint8_t getBlue() const;


	/**
	 *	@fn getAlpha()
	 *
	 *	@brief Returns the Transparency level of the color.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns Alpha component of the Color.
	 */
	std::uint8_t getAlpha() const;


	/**
	 *	@fn setRed(uint8 red)
	 *
	 *	@brief Sets the Red component of the color.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] red Red component of the color to be set.
	 */
	void setRed(std::uint8_t p_red);


	/**
	 *	@fn setGreen(uint8 green)
	 *
	 *	@brief Sets the Green component of the color.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] green Green component of the color to be set.
	 */
	void setGreen(std::uint8_t p_green);


	/**
	 *	@fn setBlue(uint8 blue)
	 *
	 *	@brief Sets the Blue component of the color.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] blue Blue component of the color to be set.
	 */
	void setBlue(std::uint8_t p_blue);


	/**
	 *	@fn setAlpha(uint8 alpha)
	 *
	 *	@brief Sets the Alpha component of the color.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] alpha Alpha component of the color to be set. 0 indicates fully transparent;
	 *				any other value shall be set to 255 to indicate Non Transparent.
	 */
	void setAlpha(std::uint8_t p_alpha);


	/**
	 *	@fn setColor(Color p_color)
	 *
	 *	@brief Convenience method to set the color from the
	 *	using the pre-defined set of colors.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_color Color to be set.
	 *
	 *	@see Color
	 *
	 *	Please see Color to see the set fo pre-defined colors and their RGB
	 *	components.
	*/
	void setColor(Color p_color);


	/**
	 *	@fn operator=(const WEColor& color)
	 *
	 *	@brief Assignment operator to allow the assigning of LHS WEColor
	 *	of "=" to RHS WEColor.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] color WEColor to be assigned to LHS of "=".
	 *
	 */
	WEColor& operator=(const WEColor& p_color) = default;

	WEColor& operator=(const Color& p_color);


	/**
	 *	@fn bool operator==(const WEColor& colorRHS) const
	 *
	 *	@brief Comparison operator
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] colorRHS RHS value of the operator
	 *
	 */
	bool operator==(const WEColor& p_colorRHS) const;
	bool operator!=(const WEColor& p_colorRHS) const;
	bool operator<(const WEColor & p_colorRHS) const;


	/**
	 *	@fn int to24Bit(void)
	 *
	 *	@brief Convert a WEColor object into a 24 bit number (format is 0xRRGGBB)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns integer containing 24bit RGB value (0xRRGGBB)
	 */
	int to24Bit(void) const;


	/**
	 *	@fn int to32Bit(void)
	 *
	 *	@brief Convert a WEColor object into a 32 bit number (format is 0xRRGGBBAA)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns integer containing 32bit RGBA value (0xRRGGBBAA)
	 */
	int to32Bit(void) const;


	/**
	 *	@fn QColor toQColor(void)
	 *
	 *	@brief Convert a WEColor object into a QColor object
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@returns the QColor object
	 */
	QColor toQColor(void) const;


	/**
	 *	@fn bool WEColor::isGrey(void)
	 *
	 *	@brief Figure out if this color is a shade of grey.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns true if this->red == this->green == this->blue
	 *	@returns false otherwise
	 *
	 *	This method also considers white and black as a shade of grey.
	 *
	 */
	bool isGrey(void) const;


	/**
	 *	@fn bool isOpaque(void) const
	 *
	 *	@brief Figure out if this color is opaque
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@returns true if this->alpha == 255
	 *	@returns false otherwise
	 *
	 */
	bool isOpaque(void) const;


	/**
	 *	@fn bool hasAlpha(void) const
	 *
	 *	@brief Figure out if this color is not opaque
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@returns true if this->alpha != 255
	 *	@returns false otherwise
	 *
	 */
	bool hasAlpha(void) const;


	/**
	 *	@fn bool isTransparent(void)
	 *
	 *	@brief Figure out if this color is completely transparent
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@returns true if this->alpha == 0
	 *	@returns false otherwise
	 *
	 */
	bool isTransparent(void) const;


	/**
	 *	@fn unsigned char getGrey(void)
	 *
	 *	@brief Calculate grey value for this color.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns Intensity for grey value
	 *
	 *
	 *	Grey value is calculated as (red+green+blue) / 3
	 */
	unsigned char getGrey(void) const;


	/**
	 *	@fn unsigned char getGrey(void)
	 *
	 *	@brief Set this color to its complimentary by inverting the red green and blue value.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns nothing
	 *
	 *
	 *	Complimentary RGB value is calculated as (red XOR 0xff), (green XOR 0xff), (blue XOR 0xff).
	 *	Result is that red is converted to green, blue to yellow etc. and vice versa.
	 */
	void setComplimentaryColor( void );
	WEColor getComplimentaryColor(void) const;

	/**
	 * Function to dimm an RGB value to a certain percent.
	 * This can be used e.g. to dimm a multi color LED
	 * @param p_valueToDimm The initial RGB color value
	 * @param p_percent Dimm percentage
	 * @return the dimmed RGB value
	 */
	static unsigned int dimmRGBValue(unsigned int p_valueToDimm, unsigned char p_percent);


	static const WEColor transparentColor;	//!< Static instance of a transparent WEColor


private:
	std::uint8_t red{0};		//!< red component
	std::uint8_t green{0};	//!< green component
	std::uint8_t blue{0};		//!< blue component
	std::uint8_t alpha{WE::opaque};	//!< transparency information (0 fully transparent, any other value Non transparent)

	//SERIALIZATION
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int);
};	//End of WEColor class
}	//end of namespace WE

BOOST_CLASS_EXPORT_KEY(WE::WEColor)

#endif /* WE_COLOR_H_ */
