/**
 *	@file we_rotation.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller

 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2013
 *
 *	@brief Implements a type (as class) to handle rotation values.
 *
 *	@see we_rotation.cpp
 *
 *	This file contains a class used to handle rotation values.
 *	Rotation values are measured in degrees. The rotation value
 *	is in range 0 to 359. This is ensured by using modulo 360
 *	operation.
 *
 *	NOTE about performance of this class:
 *	There is no CPP file for this class, because of performance optimization.
 *	I tried to make this class as performant as a primitive integer data type.
 *	So I compared the assembler code of the following instructions:
 *		int i = 9;
 *		i += 5;
 *	and
 *		WERotation r(9);
 *		r += 5;
 *	=> But I disables the normalization in WERotation, to make both code
 *	   snippets doing the same stuff.
 *	I did that using -O1 as compiler option to gcc.
 *	And I found out, that the WERotation header only inline implementation
 *	of WERotation produces the same assembler instructions as the similar
 *	integer implementation.
 *		1 for the constructor
 *		1 for the += operation
 *	With compiler option -O0 or the original CPP implementation of this class
 *	we already get multiple assembler instructions for the constructor. Same
 *	happens with the +0 operation. But this wasn't my intention when designing
 *	this class. Especial temporary creations of a WERotation would get far
 *	too expensive when WERotation will once become used more frequently ;)
 */

#ifndef WE_ROTATION_H_
#define WE_ROTATION_H_

#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>

namespace WE
{

class WERotation
{
public:
	/**
	 *	@fn WERotation
	 *
	 *	@brief Constructor for creating rotation object with zero degrees rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	inline WERotation() = default;

	/**
	 *	@fn WERotation(int p_degree)
	 *
	 *	@brief Constructor for creating rotation object with p_degree degrees rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_degree Rotation value used for initializing the WERotation object.
	 */
	inline explicit WERotation(int p_degree)
		: rotation(p_degree)
	{
		normalize();
	}

	/**
	 * Default destructor. No dynamic memory.
	 */
	~WERotation() = default;

	/**
	 *	@fn WERotation(const WERotation& p_rotation)
	 *
	 *	@brief Constructor for creating rotation object and initialize with p_rotation object.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_rotation WERotation object used to initializing the called WERotation object.
	 */
	inline WERotation(const WERotation& p_rotation)
		: rotation(p_rotation.rotation)
	{
		normalize();
	}

	/**
	 *	@fn WERotation& operator= (const int& p_degree)
	 *
	 *	@brief Assignment operator to set a new rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_degree Rotation value to be set.
	 *
	 *	@return Returns the called object itself.
	 */
	inline WERotation& operator= (const int& p_degree)
	{
		setRotation(p_degree);
		return *this;
	}

	/**
	 *	@fn WERotation& operator= (const WERotation& p_rotation)
	 *
	 *	@brief Assignment operator toset a new rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_rotation WERotation object used to be set.
	 *
	 *	@return Returns the called object itself.
	 */
	inline WERotation& operator= (const WERotation& p_rotation)
	{
		if(this != &p_rotation)
		{
			setRotation(p_rotation);
		}
		return *this;
	}

	/**
	 *	@fn WERotation& operator+= (const int& p_degree)
	 *
	 *	@brief Assignment operator to add a rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_degree Rotation value to be added.
	 *
	 *	@return Returns the called object itself.
	 */
	inline WERotation& operator+= (const int& p_degree)
	{
		addRotation(p_degree);
		return *this;
	}

	/**
	 *	@fn WERotation& operator+= (const WERotation& p_rotation)
	 *
	 *	@brief Assignment operator to add a rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_rotation Rotation object to be added.
	 *
	 *	@return Returns the called object itself.
	 */
	inline WERotation& operator+= (const WERotation& p_rotation)
	{
		addRotation(p_rotation);
		return *this;
	}


	/**
	 *	@fn WERotation& operator-= (const int& p_degree)
	 *
	 *	@brief Assignment operator to subtract a rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_degree Rotation value to be subtracted.
	 *
	 *	@return Returns the called object itself.
	 */
	inline WERotation& operator-= (const int& p_degree)
	{
		substractRotation(p_degree);
		return *this;
	}

	/**
	 *	@fn WERotation& operator-= (const WERotation& p_rotation)
	 *
	 *	@brief Assignment operator to subtract a rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_rotation Rotation object to be subtracted.
	 *
	 *	@return Returns the called object itself.
	 */
	inline WERotation& operator-= (const WERotation& p_rotation)
	{
		substractRotation(p_rotation);
		return *this;
	}

	/**
	 *	@fn inline void setRotation(const int& p_degree)
	 *
	 *	@brief Sets a new rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_degree Rotation value to be set.
	 */
	inline void setRotation(const int& p_degree)
	{
		this->rotation = p_degree;
		normalize();
	}

	/**
	 *	@fn inline void setRotation(const WERotation& p_rotation)
	 *
	 *	@brief Sets a new rotation object.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_rotation Rotation object to be set.
	 */
	inline void setRotation(const WERotation& p_rotation)
	{
		setRotation(p_rotation.toInt());
	}

	/**
	 *	@fn inline void addRotation(const int& p_degree)
	 *
	 *	@brief Add a rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_degree Rotation value to be added.
	 */
	inline void addRotation(const int& p_degree)
	{
		this->rotation += p_degree;
		normalize();
	}

	/**
	 *	@fn inline void addRotation(const WERotation& p_rotation)
	 *
	 *	@brief Add a rotation object.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_rotation Rotation object to be added.
	 */
	inline void addRotation(const WERotation& p_rotation)
	{
		addRotation(p_rotation.toInt());
	}

	/**
	 *	@fn inline void substractRotation(const int& p_degree)
	 *
	 *	@brief Subtract a rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_degree Rotation value to be subtracted.
	 */
	inline void substractRotation(const int& p_degree)
	{
		this->rotation -= p_degree;
		normalize();
	}

	/**
	 *	@fn inline void substractRotation(const WERotation& p_rotation)
	 *
	 *	@brief Subtract a rotation object.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_rotation Rotation object to be subtracted.
	 */
	inline void substractRotation(const WERotation& p_rotation)
	{
		substractRotation(p_rotation.toInt());
	}

	/**
	 *	@fn inline int toInt() const
	 *
	 *	@brief Delivers the rotation value as integer.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return The rotation value.
	 */
	inline int toInt() const
	{
		return this->rotation;
	}

	/**
	 *	@fn inline static int normalize(const int p_degree)
	 *
	 *	@brief This function normalizes the the given value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_degree The value to be normalized.
	 *
	 *	This function is used to normalize p_degree value in
	 *	the range from 0 to 359 degrees.
	 */
	inline static int normalize(const int p_degree)
	{
		int returnValue = p_degree % moduloValue;

		/* Special handling for negativ values.
		 * Example: value -1 should be corrected to 359.
		 */
		if(returnValue < 0)
		{
			returnValue += moduloValue;
		}

		return returnValue;
	}

private:
	/**
	 *	@fn inline void normalize()
	 *
	 *	@brief This function normalizes the objects rotation value.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	This function is used to normalize the objects rotation value in
	 *	the range from 0 to 359 degrees.
	 */
	inline void normalize()
	{
		this->rotation = normalize(this->rotation);
	}

	static const int moduloValue = 360;	//! Defines the value range of a rotation.
	int rotation{0};	//! Member storing the rotation value

	//SERIALIZATION
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int );

	/*
	 * These friend declarations are necessary to avoid extra function calls
	 * when comparing objects of this class. They are allowing functions that
	 * are highly related to this object a direct access to the private rotation
	 * value. So these functions do not need to use the public getter and
	 * setter methods.
	 */
	friend WERotation operator+ (const WERotation& p_lhs, const int& p_rhs);
	friend WERotation operator+ (const int& p_lhs, const WERotation& p_rhs);
	friend WERotation operator+ (const WERotation& p_lhs, const WERotation& p_rhs);
	friend WERotation operator- (const WERotation& p_lhs, const int& p_rhs);
	friend WERotation operator- (const int& p_lhs, const WERotation& p_rhs);
	friend WERotation operator- (const WERotation& p_lhs, const WERotation& p_rhs);
	friend bool operator< (const WERotation& p_lhs, const int& p_rhs);
	friend bool operator< (const int& p_lhs, const WERotation& p_rhs);
	friend bool operator< (const WERotation& p_lhs, const WERotation& p_rhs);
	friend bool operator> (const WERotation& p_lhs, const int& p_rhs);
	friend bool operator> (const int& p_lhs, const WERotation& p_rhs);
	friend bool operator> (const WERotation& p_lhs, const WERotation& p_rhs);
	friend bool operator<= (const WERotation& p_lhs, const int& p_rhs);
	friend bool operator<= (const int& p_lhs, const WERotation& p_rhs);
	friend bool operator<= (const WERotation& p_lhs, const WERotation& p_rhs);
	friend bool operator>= (const WERotation& p_lhs, const int& p_rhs);
	friend bool operator>= (const int& p_lhs, const WERotation& p_rhs);
	friend bool operator>= (const WERotation& p_lhs, const WERotation& p_rhs);
	friend bool operator== (const WERotation& p_lhs, const int& p_rhs);
	friend bool operator== (const int& p_lhs, const WERotation& p_rhs);
	friend bool operator== (const WERotation& p_lhs, const WERotation& p_rhs);
	friend bool operator!= (const WERotation& p_lhs, const int& p_rhs);
	friend bool operator!= (const int& p_lhs, const WERotation& p_rhs);
	friend bool operator!= (const WERotation& p_lhs, const WERotation& p_rhs);
};

/**
 *	@fn inline WERotation operator+ (const WERotation& p_lhs, const int& p_rhs)
 *
 *	@brief Adds a Rotation Object and an Integer
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation object to be added.
 *	@param[in] p_rhs Rotation value to be added.
 *
 *	@return Returns a WERotation object containing the result of the add operation.
 */
inline WERotation operator+ (const WERotation& p_lhs, const int& p_rhs)
{
	return WERotation(p_lhs.rotation + p_rhs);
}

/**
 *	@fn inline WERotation operator+ (const int& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Adds an Integer and a Rotation Object
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation value to be added.
 *	@param[in] p_rhs Rotation object to be added.
 *
 *	@return Returns a WERotation object containing the result of the add operation.
 */
inline WERotation operator+ (const int& p_lhs, const WERotation& p_rhs)
{
	return WERotation(p_lhs + p_rhs.rotation);
}

/**
 *	@fn inline WERotation operator+ (const WERotation& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Adds two Rotation Objects
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs First rotation object to be added.
 *	@param[in] p_rhs Second rotation object to be added.
 *
 *	@return Returns a WERotation object containing the result of the add operation.
 */
inline WERotation operator+ (const WERotation& p_lhs, const WERotation& p_rhs)
{
	return WERotation(p_lhs.rotation + p_rhs.rotation);
}

/**
 *	@fn inline WERotation operator- (const WERotation& p_lhs, const int& p_rhs)
 *
 *	@brief Subtract an Integer value from a Rotation Object
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation object to be subtracted from.
 *	@param[in] p_rhs Rotation value to be subtracted.
 *
 *	@return Returns a WERotation object containing the result of the subtract operation.
 */
inline WERotation operator- (const WERotation& p_lhs, const int& p_rhs)
{
	return WERotation(p_lhs.rotation - p_rhs);
}

/**
 *	@fn inline WERotation operator- (const int& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Subtract an Rotation Object from an Integer value.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation value to be subtracted from.
 *	@param[in] p_rhs Rotation object to be subtracted.
 *
 *	@return Returns a WERotation object containing the result of the subtract operation.
 */
inline WERotation operator- (const int& p_lhs, const WERotation& p_rhs)
{
	return WERotation(p_lhs - p_rhs.rotation);
}

/**
 *	@fn inline WERotation operator- (const WERotation& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Subtract two Rotation Objects
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation object to be subtracted from.
 *	@param[in] p_rhs Rotation object to be subtracted.
 *
 *	@return Returns a WERotation object containing the result of the subtract operation.
 */
inline WERotation operator- (const WERotation& p_lhs, const WERotation& p_rhs)
{
	return WERotation(p_lhs.rotation - p_rhs.rotation);
}

/**
 *	@fn inline bool operator< (const WERotation& p_lhs, const int& p_rhs)
 *
 *	@brief Compares a rotation object and a rotation value.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation object to be compared.
 *	@param[in] p_rhs Rotation value to be compared.
 *
 *	@return Returns true if p_lhs is smaller than p_rhs. Otherwise false is returned.
 */
inline bool operator< (const WERotation& p_lhs, const int& p_rhs)
{
	return (p_lhs.rotation < WERotation::normalize(p_rhs));
}

/**
 *	@fn inline bool operator< (const int& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares a rotation value and a rotation object.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation value to be compared.
 *	@param[in] p_rhs Rotation object to be compared.
 *
 *	@return Returns true if p_lhs is smaller than p_rhs. Otherwise false is returned.
 */
inline bool operator< (const int& p_lhs, const WERotation& p_rhs)
{
	return (WERotation::normalize(p_lhs) < p_rhs.rotation);
}

/**
 *	@fn inline bool operator< (const WERotation& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares two rotation objects.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs First rotation object to be compared.
 *	@param[in] p_rhs Second rotation object to be compared.
 *
 *	@return Returns true if p_lhs is smaller than p_rhs. Otherwise false is returned.
 */
inline bool operator< (const WERotation& p_lhs, const WERotation& p_rhs)
{
	return (p_lhs.rotation < p_rhs.rotation);
}

/**
 *	@fn inline bool operator> (const WERotation& p_lhs, const int& p_rhs)
 *
 *	@brief Compares a rotation object and a rotation value.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation object to be compared.
 *	@param[in] p_rhs Rotation value to be compared.
 *
 *	@return Returns true if p_lhs is bigger than p_rhs. Otherwise false is returned.
 */
inline bool operator> (const WERotation& p_lhs, const int& p_rhs)
{
	return (p_lhs.rotation > WERotation::normalize(p_rhs));
}

/**
 *	@fn inline bool operator> (const int& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares a rotation value and a rotation object.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation value to be compared.
 *	@param[in] p_rhs Rotation object to be compared.
 *
 *	@return Returns true if p_lhs is bigger than p_rhs. Otherwise false is returned.
 */
inline bool operator> (const int& p_lhs, const WERotation& p_rhs)
{
	return (WERotation::normalize(p_lhs) > p_rhs.rotation);
}

/**
 *	@fn inline bool operator> (const WERotation& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares two rotation objects.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs First rotation object to be compared.
 *	@param[in] p_rhs Second rotation object to be compared.
 *
 *	@return Returns true if p_lhs is bigger than p_rhs. Otherwise false is returned.
 */
inline bool operator> (const WERotation& p_lhs, const WERotation& p_rhs)
{
	return (p_lhs.rotation > p_rhs.rotation);
}

/**
 *	@fn inline bool operator<= (const WERotation& p_lhs, const int& p_rhs)
 *
 *	@brief Compares a rotation object and a rotation value.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation object to be compared.
 *	@param[in] p_rhs Rotation value to be compared.
 *
 *	@return Returns true if p_lhs is less or equal than p_rhs. Otherwise false is returned.
 */
inline bool operator<= (const WERotation& p_lhs, const int& p_rhs)
{
	return (p_lhs.rotation <= WERotation::normalize(p_rhs));
}

/**
 *	@fn inline bool operator<= (const int& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares a rotation value and a rotation object.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation value to be compared.
 *	@param[in] p_rhs Rotation object to be compared.
 *
 *	@return Returns true if p_lhs is less or equal than p_rhs. Otherwise false is returned.
 */
inline bool operator<= (const int& p_lhs, const WERotation& p_rhs)
{
	return (WERotation::normalize(p_lhs) <= p_rhs.rotation);
}

/**
 *	@fn inline bool operator<= (const WERotation& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares two rotation objects.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs First rotation object to be compared.
 *	@param[in] p_rhs Second rotation object to be compared.
 *
 *	@return Returns true if p_lhs is less or equal than p_rhs. Otherwise false is returned.
 */
inline bool operator<= (const WERotation& p_lhs, const WERotation& p_rhs)
{
	return (p_lhs.rotation <= p_rhs.rotation);
}

/**
 *	@fn inline bool operator>= (const WERotation& p_lhs, const int& p_rhs)
 *
 *	@brief Compares a rotation object and a rotation value.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation object to be compared.
 *	@param[in] p_rhs Rotation value to be compared.
 *
 *	@return Returns true if p_lhs is bigger or equal than p_rhs. Otherwise false is returned.
 */
inline bool operator>= (const WERotation& p_lhs, const int& p_rhs)
{
	return (p_lhs.rotation >= WERotation::normalize(p_rhs));
}

/**
 *	@fn inline bool operator>= (const int& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares a rotation value and a rotation object.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation value to be compared.
 *	@param[in] p_rhs Rotation object to be compared.
 *
 *	@return Returns true if p_lhs is bigger or equal than p_rhs. Otherwise false is returned.
 */
inline bool operator>= (const int& p_lhs, const WERotation& p_rhs)
{
	return (WERotation::normalize(p_lhs) >= p_rhs.rotation);
}

/**
 *	@fn inline bool operator>= (const WERotation& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares two rotation objects.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs First rotation object to be compared.
 *	@param[in] p_rhs Second rotation object to be compared.
 *
 *	@return Returns true if p_lhs is bigger or equal than p_rhs. Otherwise false is returned.
 */
inline bool operator>= (const WERotation& p_lhs, const WERotation& p_rhs)
{
	return (p_lhs.rotation >= p_rhs.rotation);
}

/**
 *	@fn inline bool operator== (const WERotation& p_lhs, const int& p_rhs)
 *
 *	@brief Compares a rotation object and a rotation value.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation object to be compared.
 *	@param[in] p_rhs Rotation value to be compared.
 *
 *	@return Returns true if p_lhs is equal to p_rhs. Otherwise false is returned.
 */
inline bool operator== (const WERotation& p_lhs, const int& p_rhs)
{
	return (p_lhs.rotation == WERotation::normalize(p_rhs));
}

/**
 *	@fn inline bool operator== (const int& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares a rotation value and a rotation object.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation value to be compared.
 *	@param[in] p_rhs Rotation object to be compared.
 *
 *	@return Returns true if p_lhs is equal to p_rhs. Otherwise false is returned.
 */
inline bool operator== (const int& p_lhs, const WERotation& p_rhs)
{
	return (WERotation::normalize(p_lhs) == p_rhs.rotation);
}

/**
 *	@fn inline bool operator== (const WERotation& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares two rotation objects.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs First rotation object to be compared.
 *	@param[in] p_rhs Second rotation object to be compared.
 *
 *	@return Returns true if p_lhs is equal to p_rhs. Otherwise false is returned.
 */
inline bool operator== (const WERotation& p_lhs, const WERotation& p_rhs)
{
	return (p_lhs.rotation == p_rhs.rotation);
}

/**
 *	@fn inline bool operator!= (const WERotation& p_lhs, const int& p_rhs)
 *
 *	@brief Compares a rotation object and a rotation value.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation object to be compared.
 *	@param[in] p_rhs Rotation value to be compared.
 *
 *	@return Returns true if p_lhs is not equal to p_rhs. Otherwise false is returned.
 */
inline bool operator!= (const WERotation& p_lhs, const int& p_rhs)
{
	return (p_lhs.rotation != WERotation::normalize(p_rhs));
}

/**
 *	@fn inline bool operator!= (const int& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares a rotation value and a rotation object.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs Rotation value to be compared.
 *	@param[in] p_rhs Rotation object to be compared.
 *
 *	@return Returns true if p_lhs is not equal to p_rhs. Otherwise false is returned.
 */
inline bool operator!= (const int& p_lhs, const WERotation& p_rhs)
{
	return (WERotation::normalize(p_lhs) != p_rhs.rotation);
}

/**
 *	@fn inline bool operator!= (const WERotation& p_lhs, const WERotation& p_rhs)
 *
 *	@brief Compares two rotation objects.
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param[in] p_lhs First rotation object to be compared.
 *	@param[in] p_rhs Second rotation object to be compared.
 *
 *	@return Returns true if p_lhs is not equal to p_rhs. Otherwise false is returned.
 */
inline bool operator!= (const WERotation& p_lhs, const WERotation& p_rhs)
{
	return (p_lhs.rotation != p_rhs.rotation);
}
}	//End of namespace WE

template<class Archive>
void WE::WERotation::serialize(Archive & p_ar, const unsigned int )
{
	p_ar & rotation;
}

#ifndef DOXYGEN_SHOULD_SKIP_THIS
//Explicitly initialize / generate code for serialization because AppGraphicDDO is abstract
template void WE::WERotation::serialize(
		boost::archive::binary_oarchive & ar,	// OUTput archive
		const unsigned int version
);


template void WE::WERotation::serialize(
		boost::archive::binary_iarchive & ar,	// INput archive
		const unsigned int version
);
#endif

#endif /* WE_ROTATION_H_ */
