/**
 *	@file we_dei_ipc_client_handler.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief DEIIPCClientHandler class definition, handles DEI's IPC messaging.
 *
 *	@see DEIIPCServerHandler
 *	@see WEIPC
 *	@see WEEndPointHandler
 *
 *	This class implements the interface of WEEndPointHandler. It's purpose is
 *	to get informed about incoming IPC messages and the handling of them.
 */

#ifndef WE_DEI_IPC_CLIENT_HANDLER_H_
#define WE_DEI_IPC_CLIENT_HANDLER_H_

#include "we_dei_ipc_thread.h"
#include "../../os_abstraction/endpoint_manager/include/we_endpoint_handler.h"

#include "../dei_ipc_api/src/dei_ipc_data.h"

#include <unordered_map>

namespace WE
{

class WEDEI;

typedef std::unordered_multimap<unsigned int, std::string> SubscriberMap; //!< multi-hasmap <ulong, string>


/**
 *	@class DEIIPCClientHandler
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Implementing the interface of WEEndPointHandler. Handling of DEI IPC traffic.
 *
 *	@namespace WE
 *
 *	@see WEEndPointHandler
 *
 *	The onEndPointReadable method gets invoked by DEIIPCServerHandler when a new IPC messages
 *	comes in on one of the connected end points.
 *	After reception of the message it will be extracted and appropriate actions will be taken
 *	(e.g. if a INT_VARIABLE_GET_REQUEST comes in, the value of the variable gets extracted from DEI
 *	and the value will be sent back via a INT_VARIABLE_GET_RESPONSE).
 */
class DEIIPCClientHandler : public WEEndPointHandler
{
public:
	/**
	 *	@fn DEIIPCClientHandler(bool startIPCThread)
	 *
	 *	@brief Standard constructor of DEIIPCClientHandler
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	explicit DEIIPCClientHandler(WEDEI * p_dei, bool p_startIPCThread);

	/**
	 *	@fn ~DEIIPCClientHandler()
	 *
	 *	@brief Standard destructor of DEIIPCClientHandler
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~DEIIPCClientHandler();

	/**
	 *	@fn void onEndPointReadable(WEEndPoint* endPoint, WEEndPointManager* endPointMgr)
	 *
	 *	@brief Function that handles incoming IPC messages.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_endPoint Endpoint that got readable (message came in)
	 *	@param[in] p_endPointMgr The endpoint manager
	 *
	 *	@returns void
	 *
	 *	@see WEIPCEndPoint
	 *	@see WEEndPointManager
	 *	@see WEEndPointHandler
	 *
	 *	This method gets called when an endpoint is readable (a message was sent to this endpoint).
	 *	It receives the message, decomposes it and takes the appropriate action.
	 */
	void onEndPointReadable(std::shared_ptr<WEEndPoint> p_endPoint, WEEndPointManager* p_endPointMgr) override;

	/**
	 *	@fn int sendIPCMessage(WEIPCEndPoint* p_endPoint,
			WEIntVariableData * p_intVar,
			enum WEMessageType p_typeOfMessage,
			bool p_status)
	 *
	 *	@brief Send an IPC message with WEIntVariableData as data.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_endPoint The endpoint that shall send the message.
	 *	@param[in] p_data The data that shall be sent.
	 *	@param[in] p_stringToSend The value of the string for all string related stuff. For integer stuff just provide an empty string ("").
	 *	@param[in] p_typeOfMessage The type of message that shall be sent.
	 *	@param[in] p_status The status of the message (true/false, good/bad) that shall be sent.
	 *
	 *	@returns WE_SUCCESS if sending was successful.
	 *	@returns WE_TYPE_ERROR if the type of message will not work with WEIntVariableData
	 *	@returns the error code of WEIPCEndPoint's sendMessage() function
	 *
	 *	@see WEIPCEndPoint
	 *	@see WEIntVariableData
	 *	@see WEMessageType
	 *	@see WEMessage
	 *
	 *	This functions constructs a new WEMessage using the provided parameters.
	 *	The constructed message gets sent by the endpoint provided as first parameter.
	 */
	int sendIPCMessage(std::weak_ptr<WEIPCEndPoint> p_endPoint,
			IPCData * p_data,
			const std::string & p_stringToSend,
			IPCMessageType p_typeOfMessage,
			bool p_status);

	/**
	 *	@fn int sendUpdateToIPCObservers(unsigned short p_varIndex, unsigned short p_varSubIndex, int p_valueToSend, bool p_success = true)
	 *
	 *	@brief Sending updated values to interested IPC subsribers
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_varIndex Index of the updated variable
	 *	@param p_varSubIndex Subindex of the updated variable
	 *	@param p_valueToSend The updated value that will be sent over IPC
	 *	@param p_success Flag that sets the status of the IPC message that will be sent to the observers
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns anything 'sendIPCMessage()' may return
	 *
	 *	This function gets called every time a variable value is changed. All other processes interested
	 *	in the value of this variable get informed automatically.
	 *
	 *	@see sendIPCMessage(WEIPCEndPoint *, WEIntVariableData *, enum WEMessageType, bool)
	 */
	int sendUpdateToIPCObservers(unsigned short p_varIndex, unsigned short p_varSubIndex, int p_valueToSend, bool p_success = true);

	/**
	 *	@fn int sendUpdateToIPCObservers(unsigned short p_varIndex, unsigned short p_varSubIndex, std::string p_valueToSend, bool p_success = true)
	 *
	 *	@brief Sending updated values to interested IPC subsribers
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_varIndex Index of the updated variable
	 *	@param p_varSubIndex Subindex of the updated variable
	 *	@param p_valueToSend The updated value that will be sent over IPC
	 *	@param p_success Flag that sets the status of the IPC message that will be sent to the observers
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns anything 'sendIPCMessage()' may return
	 *
	 *	This function gets called every time a variable value is changed. All other processes interested
	 *	in the value of this variable get informed automatically.
	 *
	 *	@see sendIPCMessage(WEIPCEndPoint *, WEStringVariableData *, std::string, enum WEMessageType, bool)
	 */
	int sendUpdateToIPCObservers(unsigned short p_varIndex, unsigned short p_varSubIndex, const std::string & p_valueToSend, bool p_success = true);

	/**
	 *	@fn void onEndPointReconnected(WEEndPoint* p_endpoint);
	 *
	 *	@brief Call this function if a IPC connection was lost and could be re-established
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_endpoint Endpoint of the connection that was re-connected
	 *
	 *	Internally re-subscribes the variables owned by the variable owner identified by p_endpoint
	 */
	void onEndPointReconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;

	/**
	 *	@fn void onEndPointDisconnected(WEEndPoint* p_endpoint)
	 *
	 *	@brief Call this function if a IPC connection was lost
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endpoint Endpoint of the connection that was disconnected
	 *
	 */
	void onEndPointDisconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;

	void startIPCThread(void);

private:

		WEIPCThread theIPCThread;

	/**
	 *	@fn int addIPCSubscriber(unsigned short p_index, unsigned short p_subindex, std::string p_endPointPeerAddress)
	 *
	 *	@brief Adds the peer address of an endpoint who is interested in a variable's value to the multimap of IPC subsribers
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the variable to which the endpoint wants to subscribe
	 *	@param p_subindex Subindex of the variable to which the endpoint wants to subscribe
	 *	@param p_endPointPeerAddress Peer address of the endpoint that is interested in the variable's value
	 *
	 *	@returns always WE_SUCCESS
	 */
	int addIPCSubscriber(unsigned short p_index, unsigned short p_subindex, const std::string & p_endPointPeerAddress);

	/**
	 *	@fn int removeIPCSubscriber(std::string p_peerAddress)
	 *
	 *	@brief Removes all IPC subscribers of a peer from the multimap
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_peerAddress Peer address of the endpoint that was interested in the variable's value
	 *
	 *	@returns always WE_SUCCESS
	 *
	 *	@see addIPCSubscriber
	 */
	int removeIPCSubscriber(const std::string & p_peerAddress);

	/**
	 * Used by onEndPointReadable() to handle a variable get request
	 */
	int handleVariableGetRequest(IPCData & p_ipcData, std::weak_ptr<WEIPCEndPoint> p_endPoint);

	/**
	 * Used by onEndPointReadable() to handle a variable get response
	 */
	int handleVariableGetResponse(const IPCData & p_ipcData, std::weak_ptr<WEIPCEndPoint> p_endPoint, bool p_messageStatus);

	/**
	 * Used by onEndPointReadable() to handle a variable set request
	 */
	int handleVariableSetRequest(const IPCData & p_ipcData, bool p_messageStatus);

	/**
	 * Used by onEndPointReadable() to handle a variable set response
	 */
	int handleVariableSetResponse(const IPCData & p_ipcData, std::weak_ptr<WEIPCEndPoint> p_endPoint, bool p_messageStatus);

	/**
	 * Used by onEndPointReadable() to handle a variable event subscriprion request
	 */
	int handleEventSubscriptionRequest(IPCData & p_ipcData, std::weak_ptr<WEIPCEndPoint> p_endPoint);

	/**
	 * Used by onEndPointReadable() to handle a variable event subscription response
	 */
	int handleEventSubscriptionResponse(const IPCData & p_ipcData, std::weak_ptr<WEIPCEndPoint> p_endPoint, bool p_messageStatus);

	/**
	 * Used by onEndPointReadable() to handle a variable update event
	 */
	int handleVariableUpdateEvent(const IPCData & p_ipcData, std::weak_ptr<WEIPCEndPoint> p_endPoint, bool p_messageStatus);

	SubscriberMap ipcSubscribers; //!< Multimap of combined variable key <-> IPC peer addresses

	WEEndPointManager * endPointMgr{nullptr}; //!< Pointer to the manager of all IPC connections

	WEDEI * dei{nullptr};
};
} // namespace WE

#endif /* WE_DEI_IPC_CLIENT_HANDLER_H_ */
