/**
 *	@file we_dei.h
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Thread to decouple the sending of IPC messages
 *
 */

#ifndef WE_DEI_IPC_THREAD_H_
#define WE_DEI_IPC_THREAD_H_

#include <condition_variable>
//#include <mutex>
#include <list>

struct IPCMessage;

namespace WE
{

class WEIPCEndPoint;
class WEThread;

/**
 *	@class WEIPCListElement
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@brief class which represents one Element of the send list
 *
 *	@namespace WE
 */
class WEIPCListElement
{
public:
	WEIPCListElement() = default;

	virtual ~WEIPCListElement() = default;

	//public members
	IPCMessage *messageToSend{nullptr};
	std::weak_ptr<WEIPCEndPoint> endPoint{};

	//Seems like we neither use copy nor assignment
	WEIPCListElement(const WEIPCListElement & p_event) = delete;
	WEIPCListElement & operator=(const WEIPCListElement &p_event) = delete;
};

/**
 *	@class WEIPCThread
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@brief represents the thread class
 *
 *	@namespace WE
 */
class WEIPCThread
{

public:

	WEIPCThread();
	virtual ~WEIPCThread();

	/**
	 *	@fn void create()
	 *
	 *	@brief creates the thread / allocates memory
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns void
	 *
	 */
	void create();

	/**
	 *	@fn void release();
	 *
	 *	@brief releases the thread
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] args
	 *
	 *	@returns void
	 *
	 *	first it signals the thread with the condition variable to wake up
	 *	afterwards the release method waits until the thread ends.
	 */
	void release();

	/**
	 *	@fn void start();
	 *
	 *	@brief sets a flag that the thread should start
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns void
	 *
	 */
	void start();

	/**
	 *	@fn void stop();
	 *
	 *	@brief sets a flag that the thread should stop
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] args
	 *
	 *	@returns void
	 *
	 */
	void stop();

	/**
	 *	@fn static void run(void* args);
	 *
	 *	@brief run method for the thread
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] args
	 *
	 *	@returns void
	 *
	 */
	static void run(void* p_args);

	/**
	 * Wrapper for the wait_for function of our condition_variable. Like this we don't need to return a
	 * non-const reference to it...
	 */
	template< class Rep, class Period >
	std::cv_status conditionalWait(std::unique_lock<std::mutex>& p_lock,
	                         const std::chrono::duration<Rep, Period>& p_relTime)
	{
		return this->condition.wait_for(p_lock, p_relTime);
	}

	/**
	 *	@fn bool isStopped()
	 *
	 *	@brief ask the flag stopped for its state
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns true if the stopped is set or false if not
	 *
	 */
	bool isStopped() const
	{
		return stopped;
	}

	/**
	 *	@fn bool isStarted()
	 *
	 *	@brief ask the flag started for its state
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns true if the started is set or false if not
	 *
	 */
	bool isStarted() const
	{
		return started;
	}

	/**
	 *	@fn bool hasNewData()
	 *
	 *	@brief ask the flag newData for its state
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns true if the newData is set or false if not
	 *
	 */
	bool hasNewData() const
	{
		return newData;
	}

	/**
	 *	@fn void setNewData(bool p_nd)
	 *
	 *	@brief ask the flag newData for its state
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_nd true or false
	 *
	 *	@returns void
	 *
	 */
	void setNewData(bool p_nd)
	{
		newData = p_nd;
	}

	WEIPCListElement * getFirstListElement(void)
	{
		return this->theMessageList.front();
	}

	void popFirstListElement(void)
	{
		this->theMessageList.pop_front();
	}

	bool isListEmpty()
	{
		return this->theMessageList.empty();
	}

	/**
	 *	@fn void pushToStack(WEMessage *messageToSend, WEIPCEndPoint* p_endPoint)
	 *
	 *	@brief ask the flag newData for its state
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] messageToSend pointer to message
	 *	@param[in] p_endPoint pointer to end point
	 *
	 *	@returns void
	 *
	 */
	void pushToStack(IPCMessage *p_messageToSend, std::weak_ptr<WEIPCEndPoint> p_endPoint);

private:

	bool started;
	bool stopped;
	bool newData;

	void init();
	void reset();

	std::condition_variable condition;
	std::mutex the_mutex;
	WEThread myThread{"IPC Thread"};
	std::list<WEIPCListElement*> theMessageList;
};
}
#endif /* WE_DEI_IPC_THREAD_H_ */
