/**
 *	@file we_variable_event_dispatcher.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Class for distributing events coming from WEDEI
 *
 *	@see WEDEI
 */

#ifndef WE_VARIABLE_EVENT_DISPATCHER_H_
#define WE_VARIABLE_EVENT_DISPATCHER_H_

#include <set>
#include <unordered_map>
#include <mutex>
#include <memory>

#include "../../event_handling/include/we_signal.h"

namespace WE
{

class WEEventQueue;

typedef std::unordered_multimap< unsigned int, WESignalListener * > DispatchMap;

/**
 *	@class WEVariableEventDispatcher
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Distribute events from WEDEI to interested WEEventListeners
 *
 *	@namespace WE
 *
 *	@see WEDEI
 *	@see WESignalListener
 *
 *	This class distributes events received from WEDEI to any WESignalListener
 *	that wants to be informed about variable events. Event listeners can be
 *	added to this class via the addMapping() function. If an event from WEDEI
 *	arrives, the event will be distributed to event listeners that are interested
 *	in events belonging to the variable that has changed.
 *	Since WEDEI is a singleton (only one instance per process), this class is
 *	also implemented as a singleton.
 */
class WEVariableEventDispatcher
{
private:
	/**
	 *	@fn WEVariableEventDispatcher()
	 *
	 *	@brief Private constructor of WEVariableEventDispatcher (singleton).
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WEVariableEventDispatcher();

	DispatchMap variableListenerMapping{};
	std::mutex mapMutex{};	//!< variableListenerMapping can be accessed and manipulated by different threads. Access needs to be synchronized.
	WEEventQueue * deiEventQueue{nullptr}; //!< Pointer to the outgoing event queue of WEDEI
	WESignal lastEvent{}; //!< Data of the last received event is stored here

	std::set< WESignalListener * > veryCuriousListeners{};

	inline static std::unique_ptr<WEVariableEventDispatcher> eventDispatcherInstance{};

public:
	/**
	 *	@fn WEVariableEventDispatcher(const WEVariableEventDispatcher&)
	 *
	 *	@brief Deleted copy constructor of WEVariableEventDispatcher, no copies are allowed.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WEVariableEventDispatcher(const WEVariableEventDispatcher & p_disp) = delete;

	/**
	 *	@fn WEVariableEventDispatcher& operator=(WEVariableEventDispatcher const&)
	 *
	 *	@brief Overloaded assignment operator of WEVariableEventDispatcher. No assignment is allowed (private)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WEVariableEventDispatcher& operator=(WEVariableEventDispatcher const&) = delete; //no assignment allowed

	/**
	 *	@fn ~WEVariableEventDispatcher()
	 *
	 *	@brief Destructor of WEVariableEventDispatcher
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WEVariableEventDispatcher();

	/**
	 *	@fn static WEVariableEventDispatcher * instance(void)
	 *
	 *	@brief Returns pointer to the only existing instance of WEVariableEventDispatcher
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Static pointer to the instance of WEVariableEventDispatcher
	 *
	 *	@warning This function has to be called to call all other functions of WEVariableEventDispatcher
	 *
	 *	Since WEVariableEventDispatcher is a singleton, this function has to be called in order to be able
	 *	to call any other public function.
	 *	E.g. WEVariableEventDispatcher::instance()->processEvents(50);
	 */
	static WEVariableEventDispatcher * instance(void);

	/**
	 * Release the static instance of this class. Do NOT call instance() after that!
	 */
	static void release(void);

	/**
	 *	@brief Creates a mapping between variable index/subindex and the pointer to the variable listener.
	 *		This function locks the map mutex until it is finished.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_varIndex Index of the variable for which the mapping shall be created
	 *	@param p_varSubindex Subindex of the variable for which the mapping shall be created
	 *	@param p_listener Pointer to the event listener whose signalOccurred() function shall be called if a
	 *		event is received that matches index/subindex
	 *
	 *	@returns WE_SUCCESS
	 */
	int addMapping(unsigned short p_varIndex, unsigned short p_varSubindex, WESignalListener * p_listener);

	/**
	 *	@brief Creates a mapping between variable index/subindex and the pointer to the variable listener.
	 *		This function locks the map mutex until it is finished.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_varKey Key of the variable for which the mapping shall be created
	 *	@param p_listener Pointer to the event listener whose signalOccurred() function shall be called if a
	 *		event is received that matches index/subindex
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	p_varKey shall be calculated as ( index << 16 | (subindex & 0xffff) ).
	 *
	 *
	 *	@see WEDEI::getCombinedKey()
	 *	@see WEDEI::calculateCombinedKey()
	 *
	 */
	int addMapping(unsigned int p_varKey, WESignalListener * p_listener);

	/**
	 *	@brief Removes a mapping between variable index/subindex and the pointer to the variable listener.
	 *		This function locks the map mutex until it is finished.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_varIndex Index of the variable for which the mapping shall be removed
	 *	@param p_varSubindex Subindex of the variable for which the mapping shall be removed
	 *	@param p_listener Pointer to the event listener to which the mapping shall be removed
	 *
	 *	@returns WE_SUCCESS if the mapping was removed successfully
	 *	@returns WE_NOT_FOUND_ERROR if the mapping was not found
	 */
	int removeMapping(unsigned short p_varIndex, unsigned short p_varSubindex, WESignalListener * p_listener);

	/**
	 *	@fn bool isEventSupported(WEEventType p_type)
	 *
	 *	@brief Tests if a provided event type is supported by this class
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_type Type of the event that shall be tested for support
	 *
	 *	@returns true if event type is supported
	 *	@returns false if event type is not supported
	 *
	 *	@see WESignal
	 */
	static bool isEventSupported(WEEventType p_type);

	/**
	 *	@fn int processEvents(unsigned int p_timeout)
	 *
	 *	@brief Function that reads events from the event queue and distributes according to mapping settings
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timeout Read events until timeout expires or event queue is empty. p_timeout == 0 will process
	 *		events until the queue is empty
	 *
	 *	@returns WE_SUCCESS if all events from the event queue were processed and the event queue is empty
	 *	@returns WE_TIMEOUT_ERROR if not all of the events could be processed but the time ran up
	 *
	 *	This functions reads events from the event queue and distributes them according to the configured
	 *	mappings. While processing and forwarding the events, the map mutex will be locked until
	 *	the function is done.
	 */
	int processEvents(unsigned int p_timeout);

	void subscribeForAllVariables(WESignalListener * p_listener);
	void removeSubscriptionForAllVariables(WESignalListener * p_listener);

	static int debugSection;
};
}//end of namespace WE
#endif /* WE_VARIABLE_EVENT_DISPATCHER_H_ */
