/**
 *	@file we_variable_owner.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Definition of the class WEVariableOwner
 *
 *	@see WEDEI
 *
 *	This the definition of an abstract interface for an variable owner.
 *	The interface has to be implemented by every class that sees itself as an owner of a variable.
 */

#ifndef WE_VARIABLE_OWNER_H_
#define WE_VARIABLE_OWNER_H_

#include <string>

namespace WE
{

typedef enum
{
	PROTOCOL_INTERNAL,
	PROTOCOL_INTER_PROCESS,
	PROTOCOL_CAN_FREESTYLE,
	PROTOCOL_CAN_OPEN,
	PROTOCOL_CAN_J1939,
	PROTOCOL_UNDEFINED
}WEProtocolType;

/**
 *	@class WEVariableOwner
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Definition of the abstract interface WEVariableOwner
 *
 *	@namespace WE
 *
 *	This abstract interface has to be implemented by every class that "owns" variables.
 *	The two methods defined in this interface get called whenever an update request
 *	(e.g. INT_VARIABLE_SET_REQUEST) arrives on DEI's IPC interface. The owner of a
 *	variable is the only one who really can know how to handle such an request.
 *	An example implementation is WEHardwareDaemon that owns all internal hardware variables,
 *	it knows what has to be done when someone wants to write a new value to e.g. the
 *	"@DispBacklightIntensity" variable.
 */
class WEVariableOwner
{
public:
	/**
	 *	@fn virtual ~WEVariableOwner()
	 *
	 *	@brief Virtual empty destructor. Needs to be overwritten by every inheriting class.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WEVariableOwner() = default;

	/**
	 *	@fn virtual int updateRequest(unsigned short p_index, unsigned short p_subindex, int p_intValue) = 0
	 *
	 *	@brief Method to be called when integer variable shall be changed.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_index Index of the variable that was requested to be updated
	 *	@param[in] p_subindex Subindex of the variable that was requested to be updated
	 *	@param[in] p_intValue Integer value that shall be written to the variable (or the hardware or whatever)
	 *
	 *	@returns Implementation specific return codes.
	 *
	 *	This is an abstract method that needs to be overwritten by every inheriting class.
	 *	It gets called when DEI receives a "SET" message over IPC from another process.
	 */
	virtual int updateRequest(unsigned short p_index, unsigned short p_subindex, int p_intValue) = 0;

	/**
	 *	@fn virtual int updateRequest(unsigned short p_index, unsigned short p_subindex, std::string p_stringValue) = 0
	 *
	 *	@brief Method to be called when string variable shall be changed.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_index Index of the variable that was requested to be updated
	 *	@param[in] p_subindex Subindex of the variable that was requested to be updated
	 *	@param[in] p_stringValue String value that shall be written to the variable (or the hardware or whatever)
	 *
	 *	@returns Implementation specific return codes.
	 *
	 *	This is an abstract method that needs to be overwritten by every inheriting class.
	 *	It gets called when DEI receives a "SET" message over IPC from another process.
	 */
	virtual int updateRequest(unsigned short p_index, unsigned short p_subindex, const std::string & p_stringValue) = 0;

	// JU
	/**
	 *	@fn virtual int getValueRequest(unsigned short p_index, unsigned short p_subindex) = 0;
	 *
	 *	@brief Method to be called when a variable of an foreign stack shall be read.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param[in] p_index Index of the variable where the value which we want to read shall be stored
	 *	@param[in] p_subindex Subindex of the variable where the value which we want to read shall be stored
	 *
	 *	@returns Implementation specific return codes.
	 *
	 *	This is an abstract method that needs to be overwritten by every inheriting class.
	 *	It gets called when DEI receives a "GET" message over IPC from another process.
	 */
	virtual int getValueRequest(unsigned short p_index, unsigned short p_subindex) = 0;

	/**
	 *	@fn const std::string & getOwnerString(void)
	 *
	 *	@brief Returns a string representing the name or IPC address of a variable owner
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Name or IPC address of the variable owner
	 */
	const std::string & getOwnerString(void) const
	{
		return ownerName;
	}

	WEProtocolType getType(void) const
	{
		return this->protocolType;
	}

protected:
	std::string ownerName{};	//!< Name of the variable owner
	WEProtocolType protocolType{PROTOCOL_UNDEFINED};
};
} // namespace WE


#endif /* WE_VARIABLE_OWNER_H_ */
