/**
 *	@file we_variable.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Interface to a WEVariable (abstract)
 *
 *	This file describes the common interface of all WE variables.
 *	The class WEVariable is abstract.
 */

#ifndef WE_VARIABLE_H_
#define WE_VARIABLE_H_

#include <map>
#include <mutex>
#include <atomic>

#include "../../../event_handling/include/we_action.h"	//for WEActionOrigin

#include "we_variable_definitions.h"

namespace WE {

class WEVariableOwner;

/**
 *	@class WEVariable
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief WEVariable interface to be implemented by every specialized variable.
 *
 *	@namespace WE
 *
 *	@see Every class that inherits from this class
 *
 *	This class provides a common interface for all kinds of variables.
 *	WEDEI is a friend of this class and is the only class that can access the
 *	protected methods of this interface.
 *	The methods that are common for all variables are already implemented here
 *	(e.g. getters and setters or observer management), every inheriting class can
 *	use these implementations.
 */
class WEVariable : public WEActionOrigin
{
protected:
	std::atomic<int>	intValue{0}; //!< Integer value of the variable, only accessible if storageType == STORAGE_TYPE_INTEGER

	int 			defaultIntValue{-1};
	int 			maxLength{0}; //!< Max length in bit for integers, in byte for strings
	int 			maxIntValue{0};
	int 			minIntValue{0};
	int 			numberOfBits{32};
	int 			stringLength{0}; //!< Length of the string value
	unsigned int 		combinedKey{0}; //!< Combination of index and subindex, this is the unique identifier of a variable
	unsigned int 		externalIdentifier{0};	//!< Identifier of the variable on an external owner (e.g. CANopen object dictionary index and subindex of another CANopen device)

	unsigned short 		index{0}; //!< Index of the variable
	unsigned short 		subindex{0}; //!< Subindex of the variable

	WEVariableAccessType 	accessType{ACCESS_TYPE_UNDEFINED}; //!< Access type of the variable, see WEVariableAccessType
	WEVariableType 		type{INTERNAL_VARIABLE}; //!< Type of the variable, see WEVariableType
	WEVariableStorageType 	storageType{STORAGE_TYPE_UNDEFINED}; //!< Storage type of the variable, see WEVariableStorageType
	WEInternalVariableType 	internalVariableType{INTERNAL_VARIABLE_UNDEFINED_TYPE};

	WEVariableOwner * 	ownerPointer{nullptr};

	bool 			forceWriting{false};	//!< if true, then setting a value will always be  processed, even if new and current value are same
	bool 			isSigned{true};
	bool 			remanent{false};
	bool 			valueIsValid{false}; //!< Flag that indicates if the stored value of the variable is valid or not
	bool 			allowSimulationValueChangeOnly{false};

	std::string 		defaultStringValue{};
	std::string 		name{"INVALID VARIABLE"}; //!< Unique name of the variable
	std::string 		ownerString{}; //!< String that contains the owner of the variable (e.g. IPC path)
	std::string 		ownerName{}; //!< String that contains the name of the owner of the variable (e.g. PClient)
	std::string 		stringValue{}; //!< String value of the variable, only accessible if storageType == STORAGE_TYPE_STRING
	std::string 		groupName{};

	mutable std::mutex the_mutex{};
	std::map<int, std::shared_ptr<WE::WEAction>> eventActionMap{};//!< reference to the Event Action Map <Event ID, WEVariableAction>

public:
	/**
	 *	@fn virtual ~WEVariable()
	 *
	 *	@brief virtual destructor of WEVariable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@see All inheriting classes.
	 *
	 *	Since this interface also holds some data, we can clear it up in this destructor.
	 *	The derived classes will only have to care for their specialized data.
	 */
	virtual ~WEVariable();

	WEVariable() = default;

	/**
	 *	Ask Dirk if name of the method should not be sufficient as documentation.
	 *	... or if you want to know why p_maxLength is something similar as numberOfBits?
	 *
	 * @param p_index
	 * @param p_subindex
	 * @param p_name
	 * @param p_storageType
	 * @param p_maxLength: Only getters and setters here ... maybe evaluated elsewhere
	 * @param p_access
	 * @param p_minValue
	 * @param p_maxValue
	 * @param p_defIntValue
	 * @param p_defStringValue
	 * @param p_initAsValid
	 * @param p_type
	 * @param p_remanent
	 * @param numberOfBits: 0..32, mainly to create correct data type for CANopen
	 * @param isSigned
	 * @param p_forceSetting
	 */
	WEVariable(unsigned short p_index,
			unsigned short p_subindex,
			const std::string & p_name,
			WEVariableStorageType p_storageType,
			int p_maxLength,
			WEVariableAccessType p_access,
			int p_minValue,
			int p_maxValue,
			int p_defIntValue,
			const std::string & p_defStringValue,
			bool p_initAsValid,
			WEVariableType p_type,
			bool p_remanent,
			int p_numberOfBits,
			bool p_isSigned,
			bool p_forceWriting = false);

	/**
	 *	Copy constructor
	 */
	WEVariable(const WEVariable &p_var);
	
	/**
	 * Move constructor. Don't rely on p_var after the move...
	 */
	WEVariable(WEVariable &&p_var);

	/**
	 *	Something similar to a copy constructor.
	 *
	 *	@param p_var
	 */
	explicit WEVariable(WEVariable * p_var);

	/**
	 *	Assignment operator
	 */
	WEVariable & operator=(const WEVariable &p_var);
	
	/**
	 * Move assignemnt operator. Don't rely on p_var after using this.
	 */
	WEVariable& operator=(WE::WEVariable && p_var);


	/**
	 *	@fn WEVariableType getType(void) const
	 *
	 *	@brief Simple get method. Returns the type of the variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns the type of the variable, see WEVariableType
	 */
	WEVariableType getType(void) const
	{
		return this->type;
	}

	/**
	 *	@fn unsigned short getIndex(void) const
	 *
	 *	@brief Simple get method. Returns the index of the variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Index of the variable
	 */
	unsigned short getIndex(void) const
	{
		return this->index;
	}

	/**
	 *	@fn unsigned short getSubindex(void) const
	 *
	 *	@brief Simple get method. Returns the subindex of the variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Subindex of the variable
	 */
	unsigned short getSubindex(void) const
	{
		return this->subindex;
	}

	/**
	 * This function also implements the WEActionOrigin interface
	 */
	const std::string & getName(void) const override
	{
		return this->name;
	}

	/**
	 *	@fn unsigned short getStringLength(void) const
	 *
	 *	@brief Simple get method. Returns the length of the string value of the variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns the length of the string value
	 */
	int getStringLength(void) const
	{
		return this->stringLength;
	}

	/**
	 *	@fn WEVariableStorageType getStorageType(void) const
	 *
	 *	@brief Simple get method. Returns the storage type of the variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns the storage type of the variable, see WEVariableStorageType
	 */
	WEVariableStorageType getStorageType(void) const
	{
		return this->storageType;
	}

	/**
	 * Implementation of WEActionOrigin interface.
	 */
	unsigned int getID() const override
	{
		return this->combinedKey;
	}

	int getIntegerValue(void) const;

	/**
	 *	@fn int setEventAction(int p_eventType, std::shared_ptr<WE::WEAction> p_varAction)
	 *
	 *	@brief This method associates the variable Action with the event and store with the
	 *	@brief variable in a hash map.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@returns WE_SUCCESS  for successful association of action with variable event
	 *	@returns WE_EVENT_CONFIGURATION_FAILED if action passed as parameter is Null
	 *
	 */
	int setEventAction(int p_eventType, std::shared_ptr<WE::WEAction> p_varAction);

	/**
	 *	@fn std::shared_ptr<WE::WEAction> getAction(int p_eventType)
	 *
	 *	@brief This method provides the Action specified for an event associated with the variable
	 *
	 *	@returns the Action to execute
	 */
	std::shared_ptr<WE::WEAction> getAction(int p_eventType);

	/**
	 *	@fn std::map<int, std::shared_ptr<WE::WEAction>> getVariableEventActionMap(void)
	 *
	 *	@brief This method provides the Event Action map associated with the variable
	 *
	 *	@returns the Event Action Map of variable
	 */
	std::map<int, std::shared_ptr<WE::WEAction>> getVariableEventActionMap();

	/**
	 *	@fn int setVariableEventActionMap(std::map<int, std::shared_ptr<WE::WEAction>> p_map)
	 *
	 *	@brief This method store the Event Action map associated with the variable
	 *
	 *	@param[in] p_map map to be set.
	 *
	 *	@returns WE_SUCCESS
	 */
	int setVariableEventActionMap(const std::map<int, std::shared_ptr<WE::WEAction>> & p_map);


	/**
	 *	Assuming method name is enough documentation.
	 */
	int getMaxIntValue() const;


	/**
	 *	Assuming method name is enough documentation.
	 */
	int getMaxLength() const;


	/**
	 *	Assuming method name is enough documentation.
	 */
	int getMinIntValue() const;


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setDefaultIntValue(int p_defaultIntValue);


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setDefaultStringValue(const std::string & p_defaultStringValue);


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setMaxIntValue(int p_maxIntValue);


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setMaxLength(int p_maxLength);


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setMinIntValue(int p_minIntValue);


	/**
	 *	@fn unsigned int getCombinedKey(void) const
	 *
	 *	@brief Simple get method. Returns the combined key of the variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns the combined key of the variable
	 */
	unsigned int getCombinedKey(void) const
	{
		return this->combinedKey;
	}

	/**
	 *	@fn WEVariableAccessType getAccessType(void) const
	 *
	 *	@brief Simple get method. Returns the access type of the variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns the access type of the variable, see WEVariableAccessType
	 */
	WEVariableAccessType getAccessType(void) const
	{
		return this->accessType;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setVariableType(WEVariableType p_type)
	{
		this->type = p_type;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setOwner(WEVariableOwner * p_owner)
	{
		this->ownerPointer = p_owner;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setOwnerString(const std::string & p_owner)
	{
		this->ownerString = p_owner;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setOwnerName(const std::string & p_owner)
	{
		this->ownerName = p_owner;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	int getDefaultIntValue(void) const
	{
		return this->defaultIntValue;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	const std::string & getDefaultStringValue(void) const
	{
		return this->defaultStringValue;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	WEVariableOwner * getOwnerPointer(void) const
	{
		return this->ownerPointer;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setExternalIdentifier(unsigned int p_id)
	{
		this->externalIdentifier = p_id;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	unsigned int getExternalIdentifier(void) const
	{
		return this->externalIdentifier;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setValueToValid(bool p_valid);

	/**
	 *	@fn bool isValid(void) const
	 *
	 *	@brief Returns if the value of the variable is valid or not.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns true if the value of the variable is valid
	 *	@returns false if the value of the variable is not valid
	 */
	bool isValid(void) const
	{
		return valueIsValid;
	}

	/**
	 *	@fn bool getForceWriting(void) const
	 *
	 *	@brief Returns if setting of variable value is forced.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns true if setting value of the variable is forced
	 *	@returns false if setting value of the variable is not forced
	 *
	 *	When setting a variable value, the value is only set, if the new and current variable value
	 *	are not equal. That is the default behavior. In some cases, especially when setting a variable
	 *	triggers an action, it might be useful to set all values, even if new and current variable value are same.
	 *	The current behavior can be read by this function.
	 */
	bool getForceWriting(void) const;

	/**
	 *	@fn bool setForceWriting(bool p_forceWriting)
	 *
	 *	@brief Changes if setting of variable value is forced.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_forceSetting if set to true, setting variable value is always forced.
	 *
	 *	When setting a variable value, the value is only set, if the new and current variable value
	 *	are not equal. That is the default behavior. In some cases, especially when setting a variable
	 *	triggers an action, it might be useful to set all values, even if new and current variable value are same.
	 *	The current behavior can be changed by this function.
	 */
	void setForceWriting(bool p_forceWriting);


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setSigned(bool p_signed)
	{
		this->isSigned = p_signed;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	bool isVariableSigned(void) const
	{
		return this->isSigned;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	int setNumberOfBits(int p_bits);


	/**
	 *	Assuming method name is enough documentation.
	 */
	int getNumberOfBits(void) const
	{
		return this->numberOfBits;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setInternalVariableType(WEInternalVariableType p_type)
	{
		this->internalVariableType = p_type;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	WEInternalVariableType getInternalVariableType(void) const
	{
		return this->internalVariableType;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	bool isRemanent(void) const
	{
		return this->remanent;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	unsigned short getDataType(void) const;


	/**
	 *	Assuming method name is enough documentation.
	 */
	const std::string & getGroupName(void) const;


	/**
	 *	Assuming method name is enough documentation.
	 */
	void setGroupName(const std::string & p_name);


	/**
	 *	Assuming method name is enough documentation.
	 */
	int getIntValueCopy(void) const;


	/**
	 *	@fn const std::string & getOwner(void) const
	 *
	 *	@brief Simple get method. Returns the name of the owner of the variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns the name of the owner of the variable
	 */
	const std::string & getOwner(void) const
	{
		return this->ownerString;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	const std::string & getOwnerName(void) const
	{
		return this->ownerName;
	}


	/**
	 *	Assuming method name is enough documentation.
	 */
	bool isVariableChangeOnlyAllowedBySimulation(void) const
	{
		return this->allowSimulationValueChangeOnly;
	}

	void setAllowSimulationValueChangeOnly(bool p_allow)
	{
		this->allowSimulationValueChangeOnly = p_allow;
	}

	/**
	 *	@fn bool isEventExists(void)
	 *
	 *	@brief This method return true if any event is associated with the variable
	 *
	 *	@returns TRUE, if event is defined, otherwise FALSE
	 */
	bool isEventExists() const;

	/**
	 *	@fn unsigned int createCombinedKey(void)
	 *
	 *	@brief Creates the combined key of a variable using it's index and subindex.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns The created combined key
	 *
	 *	With this function the combined key of a variable is created and stored in the member
	 *	combinedKey. The created key is returned by the function.
	 */
	inline unsigned int createCombinedKey(void)
	{
		this->combinedKey = this->index;
		this->combinedKey = this->combinedKey << 16;
		this->combinedKey |= this->subindex;
		return this->combinedKey;
	}

	const std::string & getStringValue(void) const;

	/**
	 *	@fn int setValue(int  p_valueToSet)
	 *
	 *	@brief Pure virtual function to set a new integer value to a variable.
	 *		Needs to be implemented by every variable type.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_valueToSet New integer value that shall be set to the variable
	 *
	 *	@returns Dependent on the implementation of the concrete variable type
	 *
	 *	@see Every class derived from WEVariable
	 *
	 * 	@warning This functions LOCKS THE VARIABLE'S MUTEX when the value is changed. Watch out for deadlocks!
	 *
	 *	With this function a new value can be set to the variable. with the
	 *	name of the DEI it is checked if this DEI is allowed to directly
	 *	write to the variable.
	 */
	int setValue(int p_valueToSet);

	/**
	 *	@fn int setValue(std::string p_valueToSet)
	 *
	 *	@brief Pure virtual function to set a new string value to a variable.
	 *		Needs to be implemented by every variable type.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_valueToSet New string value that shall be set to the variable
	 *
	 *	@returns Dependent on the implementation of the concrete variable type
	 *
	 *	@see Every class derived from WEVariable
	 *
	 * 	@warning This functions LOCKS THE VARIABLE'S MUTEX when the value is changed. Watch out for deadlocks!
	 *
	 *	With this function a new value can be set to the variable. with the
	 *	name of the DEI it is checked if this DEI is allowed to directly
	 *	write to the variable.
	 */
	int setValue(const std::string & p_valueToSet);

	/**
	 * This function writes the int or string value of the variable into a file.
	 * @param p_path Path in which the variable file shall be stored in
	 * @return WE_SUCCESS if writing was successful
	 * 	WE_TYPE_ERROR if the variable's storage type is neither string nor int
	 * 	or return value of WEFilesAndFolders::writeStringToFile()
	 *
	 * @warning This functions LOCKS THE VARIABLE'S MUTEX when calling WEFilesAndFolders::writeStringToFile. Watch out for deadlocks!
	 *
	 * The file name is set in this function based on the index and subindex of the variable.
	 * If the variable's storage type is integer, the int value is converted to a string when writing it into the file.
	 */
	int writeRemanentValue(const std::string & p_path) const;
};
} // namespace WE

#endif /* WE_VARIABLE_H_ */
