/**
*	@file we_action.h
*
*	@author dsl@wachendorff.de
*	@author Dirk Sterzel
*
*	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
*
*	@brief Class definition of WEAction
*
*	see class description
*
*/

#ifndef WE_ACTION_H_
#define WE_ACTION_H_

#include <boost/archive/binary_iarchive.hpp>

namespace WE
{


/**
 *	@class WEAction
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief This class is used for defining an Action. Application can define the actions by inheriting from this class
 *	and implementing the executeAction interface of this class. Please see details
 *
 *	@namespace WE
 *
 *	@see PClientBaseAction
 *
 *	The DDOs provide an interface to set the Events/Actions (setEventAction).
 *	The applications can create the Actions (using WEAction) and configure them
 *	with in the DDOs for the appropriate events.

 *
*/

typedef enum
{
	EXECUTE_SCRIPT_ACTION,
	JUMP_TO_PAGE_ACTION,
	SET_INT_VALUE_ACTION,
	SET_STRING_VALUE_ACTION,
	SCALE_VALUE_ACTION,
	NEXT_SOFTKEY_LEVEL_ACTION,
	PREVIOUS_SOFTKEY_LEVEL_ACTION,
	START_APPLICATION_ACTION,
	SEND_EVENT_TO_PLUGIN_ACTION,
	SET_OBJECT_OPEN_FOR_INPUT,
	NAVIGATION_UP_ACTION,
	NAVIGATION_DOWN_ACTION,
	NAVIGATION_CONFIRM_ACTION,
	OPEN_PDF_ACTION,
	MULTIMEDIA_ACTION,

	ACTION_UNKNOWN
}WEActionType;

/**
 * Interface class to make it possible to identify
 * the origin of an action. Every class that needs to be identified as
 * an action origin needs to implement this interface.
 */
class WEActionOrigin
{
public:
	WEActionOrigin() = default;

	virtual ~WEActionOrigin() = default;

	virtual unsigned int getID() const = 0;
	virtual const std::string & getName() const = 0;

private:
};

class WEAction
{
public:
	WEAction() = default;

	/**
	 *	@fn virtual bool canExecute() = 0
	 *
	 *	@brief verify if the action can be executed or not. This method is used by WEActionQueue to
	 *	validate before executing the action.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@returns TRUE if action is valid to be executed,
	 *	@returns FALSE if action is no more valid to be executed
	 *
	 */
	virtual bool canExecute() const = 0;

	/**
	 *	@fn virtual int executeAction(void* arg) = 0
	 *
	 *	@brief perform the action.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@param arg pointer to the incoming value and also used to return the scaled value,
	 *						used mainly for ScaleValueAction and CopyValueAction in future.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	Seems parameter is highly context specific:
	 *	- VT actions need input or navigation events which are provided by WEActionQueue.
	 *	- DEI uses variable values and calls 'executeAction()' directly.
	 *	- In many cases nullptr seems to be fine.
	 */
	virtual int executeAction(void* p_arg) = 0;
	virtual ~WEAction() = default;

	virtual WEActionType getActionType(void) const
	{
		return this->type;
	}

	void setOrigin(WEActionOrigin * p_origin)
	{
		this->origin = p_origin;
	}

	void setTrigger(unsigned int p_trigger)
	{
		this->trigger = p_trigger;
	}

protected:
	WEActionType type{ACTION_UNKNOWN};
	WEActionOrigin * origin{nullptr};
	unsigned int trigger{0};	// event type which triggered this action
private:
	//SERIALIZATION
	//This struct/class is exported in we_action_queue.cpp
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int )
	{
		p_ar & type;
	}
};
} // namespace WE


#endif /* WE_ACTION_H_ */
