/**
 *	@file we_can_remote_handler.h
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd Faisal
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief handler for opus 44-1 events
 *
 *	@see WEInputEventManager
 *
 *	Opus 44-1 is a remote control type of device that is connected on the CAN bus and uses CANopen protocol.
 *	It has an encoder and 2 more keys Shift and Back.
 *	This class provides the handler for opus 44-1 events which is invoked every time an event event arrives
 *	at the can port connected to opus 44-1 (e.g; can0) and reads CAN messages from the can port and translates it
 *	into WEInputEvent and post to the event queue. It also calculates long presses for encoder and keys.
 *
 */


#ifndef WE_CAN_REMOTE_HANDLER_H_
#define WE_CAN_REMOTE_HANDLER_H_

#include "../../os_abstraction/endpoint_manager/include/we_endpoint_handler.h"

namespace WE
{

class WEInputEventManager;
class WEEndPoint;
class WEEndPointManager;

class CANRemoteHandler : public WEEndPointHandler
{
private:
	WEInputEventManager* inputEventManager;		//!< the input event manager using this handler, mainly required to post events to its queue.

public:
	/**
	 *	@fn CANRemoteHandler(WEInputEventManager* inputEventManager)
	 *
	 *	@brief constructor for the handler
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@param inputEventManager the input event manager using this handler
	 */
	explicit CANRemoteHandler(WEInputEventManager* p_inputEventManager);

	/**
	 *	@fn ~CANRemoteHandler()
	 *
	 *	@brief default destructor for the hanlder
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 */
	virtual ~CANRemoteHandler();

	/**
	 *	@fn void onEndPointReadable(WEEndPoint* p_endPoint, WEEndPointManager* p_endPointMgr)
	 *
	 *	@brief Invoked every time from the waitForEndPoints() if the corresponding endpoint
	 *		becomes readable which happens when a CAN message arrives at the can port.
	 *
	 *	@param p_endPoint The End point that is found to be readable. This is actually a WECANRaw pointer.
	 *	@param p_endPointMgr The End Point manager managing the endpoint.
	*/
	void onEndPointReadable(std::shared_ptr<WEEndPoint> p_endPoint, WEEndPointManager* p_endPointMgr) override;

	/**
	 *	@fn void onEndPointReconnected(WEEndPoint* p_endpoint)
	 *
	 *	@brief This function gets called every time a endpoint gets re-connected
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@param p_endpoint The endpoint (WECANRaw*) which was reconnected
	 */
	void onEndPointReconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;

	/**
	 *	@fn void onEndPointDisconnected(WEEndPoint* p_endpoint)
	 *
	 *	@brief Call this function if a IPC connection was lost
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endpoint Endpoint of the connection that was disconnected
	 *
	 */
	void onEndPointDisconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;
};
}
#endif /* WE_CAN_REMOTE_HANDLER_H_ */
