/**
 *	@file we_encoder_dynamic.h
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file contains the interface definitions of the encoder dynamics calculation.
 *
 *	@see WE::WEInputEventManager
 *
 *	Encoder dynamics is used to make the rotary encoder of the operator panel more useful
 *	especially if a wide range of values is to be changed by turning the encoder.
 *	The idea behind that is that the increment per encoder click increases when the encoder
 *	is turned faster i.e. the time between ticks is low.
 *	The parameters should be set from application while the actual calculation is intended
 *	to be called from WEInputEventManager.
 *	Like that every encoder move event will contain its direction and separately its increment.
 *	This means the receiver of the event can decide to apply the dynamic increment or only the
 *	single click which triggered the event.
 *
 *	The original approach was: Calculate gap between encoder events and with that go to a look-up table.
 *	The look-up table had fixed time gap definitions but adjustable increments to apply for each gap.
 *	The shorted the gap the higher the increment.
 *
 *	When encoder was modified to accumulate ticks (i.e. send the first one directly but then wait
 *	for some time and put the number of ticks for that period into the event sent) the approach changed:
 *	The number of ticks in the event was used as index into the look-up table now.
 *	This approach had the disadvantage that there were too few events to smoothen them with low pass filtering
 *	which resulted in an unusable dynamic on Opus A3 while being acceptable on Opus A6 and RVC.
 *	On RVC we anyways only get accumulated events from BSP so that we do not really have a choice.
 *
 *	Finally we combined the two approaches:
 *	Calculate dynamic on single events whenever possible but forward accumulated events to save
 *	refreshing efforts in the application.
 *
 *
 */


#ifndef WE_ENCODER_DYNAMIC_H_
#define WE_ENCODER_DYNAMIC_H_

#include <memory> //unique_ptr
#include <vector>

namespace WE
{

class WEInputEvent;

/**
 *	@struct WEEncoderDynamicPair
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@brief This struct contains a time gap (in milliseconds) and it's respective increment.
 *
 *	@namespace WE
 *
 *	@see WEEncoderDynamic::setDynamicParameters
 *	@see WEEncoderDynamic::calculateIncrement
 *
 *	A vector of this struct is used to define the increment for different encoder speeds.
 */

struct WEEncoderDynamicPair
{
	unsigned int timeGap;	//!< Apply increment if gap between events is timeGap [ms] or shorter
	unsigned int increment;	//!< increment that is applied for one click if time between clicks is timeGap or shorter
};

/**
 *	@class WEEncoderDynamic
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@brief This class is a singleton calculating encoder dynamic based on time between clicks.
 *
 *	@namespace WE
 *
 *
 *	This class evaluates the timestamps of encoder events in order to calculate increment per
 *	click where the increment increases when time between clicks becomes shorter.
 *	Algorithm to do the calculation is fixed but can be widely influenced by parameters.
 */

class WEEncoderDynamic
{
private:

	int userFactor{1000};			//!< fine tuning by user; 1000: unchanged, 100: only 10% of increment applied
	// assume a vector of 10 elements should fit in most cases - can be increased anyways
	std::vector< WEEncoderDynamicPair > dynamicParameters{10};	//!< one factor  and timeGap for each speed interval, [0]: fastest speed == smallest timeGap
	int lowPassFactor{500};			//!< timeGap is calculated as ( timeGap * lowPassFactor + newTimeGap * (1000-lowPassFactor) ) / 1000
	unsigned int timeGap{100};			//!< low pass filtered (->average) time gap between clicks
	unsigned int increment{1};		//!< low pass filtered (->average) increment of last clicks
	unsigned int initTimeGap{10};		//!< to be calculated from slowest dynamicParameter
	unsigned long long lastTimeStamp{0};	//!< remember when last event before this one was triggered

	static inline std::unique_ptr<WEEncoderDynamic> theInstance{};	//!< pointing to the singleton instance
	// see if we can keep the constants with the default values inside the class itself ...
	static const int HUNDRED_PERCENT{1000};
	static const int DEFAULT_USER_FACTOR{HUNDRED_PERCENT};
	static const int DEFAULT_LOW_PASS{900};
	static const size_t DEFAULT_DYNAMIC_SIZE{7};			//!< size is expected to be >= 1, i.e. we expect at least one element in the following arrays
	static inline const int DEFAULT_TIME_GAPS [DEFAULT_DYNAMIC_SIZE]{   2,    8,   12,   20,   35,   55,  100};	// These are borrowed from the 16 bit systems
	static inline const int DEFAULT_INCREMENTS [DEFAULT_DYNAMIC_SIZE]{4096, 1024,  256,   64,   16,    4,    1};	//
	static const unsigned int DEFAULT_INIT_TIME_GAP{500};		//!< half a second should be without increment in any case

	static int debugSection;

	/**
	 *	@fn WEEncoderDynamic()
	 *
	 *	@brief Private standard constructor - we expect this class to be a singleton.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@see instance()
	 *	@see initialize()
	 *
	 *	This constructor will not do initialization - "instance()" will create and
	 *	initialize the singleton when it is called the first time.
	 */
	WEEncoderDynamic() = default;



public:

	/**
	 *	@fn ~WEEncoderDynamic()
	 *
	 *	@brief Destructor of the encoder dynamics. Frees memory, also deletes all variables.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 */
	virtual ~WEEncoderDynamic() = default;

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEEncoderDynamic(const WEEncoderDynamic &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEEncoderDynamic & operator=(const WEEncoderDynamic &) = delete;

	/**
	 *	@fn static WEEncoderDynamic * instance()
	 *
	 *	@brief Get the one instance of encoder dynamics that exists per process (or create it).
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns pointer to the instance of the WEEncoderDynamic
	 *	@returns nullptr if the instance did not exist and could not be created and initialized.
	 *
	 *	@warning When using other methods they have to be called by calling the instance method!
	 *		e.g. WEEncoderDynamic::instance()->otherMethod();
	 *
	 *	Since WEEncoderDynamic is a singleton, there is only one instance per process allowed.
	 *	Therefore this static method has to be used to call any other methods of
	 *	WEEncoderDynamic.
	 */
	static WEEncoderDynamic * instance();



	/**
	 *	@fn initialize()
	 *
	 *	@brief Initialize all members to default settings so that WEEncoderDynamic can be used.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *
	 *	@returns WE_SUCCESS in case of success.
	 *
	 *	@see instance()
	 *
	 *	This initialization method is called whenever the constructor is called i.e. when
	 *	the instance() method is called the first time.
	 *	It may be called again later in order to reset all members to default values.
	 */
	int initialize();



	/**
	 *	@fn getLastTimeStamp() const
	 *
	 *	@brief Get timestamp of last event evaluated.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns Milliseconds as indicated by timestamp of last event evaluated
	 *
	 */
	unsigned long long getLastTimeStamp() const;



	/**
	 *	@fn getLowPassFactor() const
	 *
	 *	@brief Get factor used in calculation of average time gap between events.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns Factor used in low pass calculation of time gap between events.
	 *
	 *	timeGap is calculated as
	 *	( newTimeGap * lowPassFactor + timeGap * (1000-lowPassFactor) ) / 1000
	 *
	 */
	int getLowPassFactor() const;



	/**
	 *	@fn setLowPassFactor(int p_lowPassFactor)
	 *
	 *	@brief Set factor used in calculation of average time gap between events.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] <p_lowPassFactor> New factor to be used in low pass filtering.
	 *			Valid values are 1 to 1000
	 *
	 *	@returns WE_SUCCESS if new value was accepted
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if new value is too high or too low.
	 *
	 *	timeGap is calculated as
	 *	( newTimeGap * lowPassFactor + timeGap * (1000-lowPassFactor) ) / 1000
	 *	Note that a value of 1000 effectively deactivates the filtering.
	 *
	 */
	int setLowPassFactor(int p_lowPassFactor);



	/**
	 *	@fn getDynamicParameters() const
	 *
	 *	@brief Get vector of parameters currently used in calculation of dynamic increment.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns Vector of parameters currently used in calculation of dynamic increment.
	 *
	 */
	std::vector<WEEncoderDynamicPair> getDynamicParameters() const;



	/**
	 *	@fn int setDynamicParameters(const std::vector<WEEncoderDynamicPair>& p_dynamicParameters)
	 *
	 *	@brief Set parameters consisting of timeGap and increment to be applied for it.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] <p_dynamicParameters> Vector of timeGaps and their assigned increments
	 *
	 *	@returns WE_SUCCESS if new value was accepted
	 *
	 *	For the algorithm to work as expected the vector is expected to be sorted by timeGap
	 *	values, i.e. dynamicParameters[0] shall contain the shortest timeGap describing the
	 *	highest speed of the encoder.
	 *
	 *	@see calculateIncrement()
	 *	@see WE::WEEncoderDynamicPair
	 *
	 */
	int setDynamicParameters(const std::vector<WEEncoderDynamicPair>& p_dynamicParameters);



	/**
	 *	@fn getTimeGap() const
	 *
	 *	@brief Get current (average) gap between encoder click events.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *
	 *	@returns Low pass filtered time (i.e. average) gap between encoder events processed.
	 *
	 *	@see setLowPassFactor()
	 *
	 *	timeGap is calculated as
	 *	( newTimeGap * lowPassFactor + timeGap * (1000-lowPassFactor) ) / 1000
	 */
	unsigned int getTimeGap() const;



	/**
	 *	@fn getUserFactor() const
	 *
	 *	@brief Get level of fine-tuning user applied to encoder dynamic.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *
	 *	@returns userFactor (1000 == 100.0%: keep behavior as defined)
	 *
	 *	Resulting increment is calculated as
	 *	(value taken from dynamicParameters) * userFactor / 1000
	 */
	int getUserFactor() const;



	/**
	 *	@fn setUserFactor(int p_userFactor)
	 *
	 *	@brief Set level of fine-tuning user applied to encoder dynamic.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] <p_userFactor> New factor to be used in user fine-tuning
	 *			Valid values are > 0; 1000 keeps behavior as defined
	 *
	 *	@returns WE_SUCCESS if new value was accepted
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if new value is too low.
	 *
	 *	Resulting increment is calculated as
	 *	(value taken from dynamicParameters) * userFactor / 1000
	 *	Note that a value of 1000 == 100.0% effectively deactivates user fine tuning and
	 *	directly uses the settings from dynamicParameters.
	 *
	 */
	int setUserFactor(int p_userFactor);



	/**
	 *	@brief Calculate increment and set it as eventParameter2 into the event provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *
	 *	@param <p_newEvent> encoder event to be analyzed
	 *	@param <p_accumulatedEvent> Event collecting multiple encoder clicks
	 *
	 *	@returns WE_SUCCESS in case of success or an error code otherwise
	 *	@returns WE_GENERAL_ERROR if timeStamp decreased compared to last event
	 *	@returns WE_TYPE_ERROR if p_event is not an ENCODER_MOVE_EVENT
	 *
	 *	@see setUserFactor()
	 *	@see setDynamicParameters()
	 *
	 *	This function performs the following steps with the event provided:
	 *	1) Calculate the time gap between this event and the one before.
	 *	2) Calculate new average time gap.
	 *	3) Look up the increment in dynamicParameters fitting the current time gap.
	 *	4) Scale it by user factor.
	 *	5) Copy the resulting increment into p_accumulatedEvent.
	 */
	int calculateIncrement(WEInputEvent & p_newEvent, WEInputEvent & p_accumulatedEvent);
};	// class
}	// namespace
#endif /* WE_ENCODER_DYNAMIC_H_ */
