/**
 *	@file we_event.h
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Abstract base class to carry the event information and functions to access these elements
 *
 *	The general idea is to have two contexts dealing with events:
 *	* GUI context handling operator input and drawing - aka 'slow thread'.
 *	* Data context handling hardware, CAN busses, object pool state - aka 'fast thread'.
 *	Events are used to communicate between the two and synchronize them (for the most part without any locking).
 *
 *	Events created in one context are handled there first and then sent via queue to the other
 *	context for final handling (typically triggering an action or applying changes to widget):
 *
 *			GUI context				queue				Data context
 *
 *			     ,->	WEToolkitEvent		--,
 *	creation	------->	WEInputEvent		------->	WEIncident	------->	trigger action
 *			     '->	WENavigationEvent	--'
 *
 *	modify widget	<-------	WETrigger		<------		WESignal	<-------	creation
 *
 *
 *	@see WEToolkitEvent - to be used for events from the widget (is a WEInputEvent and hence a WEIncident)
 *	@see WEInputEvent - to be used for operator input (is a WEIncident)
 *	@see WENavigationEvent - to be used for 'navigation actions' (such as setting focus) triggered by operator input or command (is a WEInputEvent and hence a WEIncident)
 *	@see WEIncident - to communicate from operator input/drawing (slow thread) to hardware and data maintenance (fast thread)

 *	@see WESignal - to be used for hardware events and other events staying in 'fast thread'
 *	@see WETrigger - to communicate from hardware and data maintenance (fast thread) to operator input/drawing (slow thread)
 *
 *	@see WESomeEvent - to be used as a container class e.g. for queues (virtually inherits all the others, i.e. is big enough to carry any of them)
 *
 */



/*
 *	Improvements were incorporated in redesign covered with https://gei-jira1.topcon.com/browse/PRJ-147
 */



#ifndef WE_EVENT_H_
#define	WE_EVENT_H_

#include <cinttypes>
#include <string>

namespace WE
{

enum WEKeyCodes
{	// The following values are based on the WE enumeration convention, to make the use of key codes OS independent
	// the actual os code mapping can be found in WEInputEventManager which can be overwritten by an application.
	WE_KEY_ESC	= 0,
	WE_KEY_HOME	= 1,
	WE_KEY_TAB	= 2,
	WE_KEY_ENTER	= 3,
	WE_KEY_UP	= 4,
	WE_KEY_DOWN	= 5,
	WE_KEY_END	= 6,
	WE_KEY_POWER	= 7,
	WE_KEY_UNKNOWN	= 0xffff
};


/**
 *	@enum WEEventClass
 *
 *	@brief IDs for all event classes. Note that WEEvent is only an abstract base type.
 *
 *	This enum mainly exists to avoid a lot of 'dynamic_cast' calls in WESomeEvent.
 *
 *	@see WESomeEvent
 *
*/
enum WEEventClass
{
	WE_EVENT				= 0,
	WE_INCIDENT,
	WE_TOOLKIT_EVENT,
	WE_INPUT_EVENT,
	WE_NAVIGATION_EVENT,

	WE_TRIGGER,
	WE_SIGNAL,

	WE_SOME_EVENT
};


/**
 *	@enum WEEventType
 *
 *	@brief IDs for all event types. Note that WEEvent is only an abstract base type.
 *
 *	@namespace WE
 *
 *	@see WEEvent
 *
*/
enum WEEventType
{
	UNKNOWN_EVENT				=  0,

	// Events (1-26) as defined in ISO 11783-6 Section A.2
	// NOTE: Do not change the event numbers from 1-26 as they are directly
	// mapped to ISO 11783-6 Event IDs
	WORKING_SET_ACTIVE_EVENT		=  1,	// Trigger it when pool is made active by the operator using Window Management OR due to transfer/loading of a pool
	WORKING_SET_DEACTIVE_EVENT		=  2,	// Trigger it when a pool is made deactive by the operator using Window Management OR due to transfer/loading of a pool
	SHOW_EVENT				=  3,	// Trigger it for Show/Hide command
	HIDE_EVENT				=  4,	// Trigger it for Show/Hide command
	ENABLE_EVENT				=  5,	// Trigger it for Enable/Disable command
	DISABLE_EVENT				=  6,	// Trigger it for Enable/Disable command
	CHANGE_ACTIVE_DATAMASK_EVENT		=  7,	// Trigger it for change datamask command
	CHANGE_SOFTKEY_MASK_EVENT		=  8,	// Trigger it for change softkey mask command
	CHANGE_ATTRIBUTE_EVENT			=  9,	// Trigger it for change attribute command for all objects
	CHANGE_BACKGROUND_COLOR_EVENT		= 10,	// Trigger it for change background color commands
	CHANGE_FONT_ATTRIBUTES_EVENT		= 11,	// Trigger it for change font attribute command
	CHANGE_LINE_ATTRIBUTES_EVENT		= 12,	// Trigger it for change line attribute command
	CHANGE_FILL_ATTRIBUTES_EVENT		= 13,	// Trigger it for change fill attribute command
	CHANGE_CHILD_LOCATION_EVENT		= 14,	// Trigger it for change child location command
	CHANGE_SIZE_EVENT			= 15,	// Trigger it for change size command
	CHANGE_VALUE_EVENT			= 16,	// Trigger it for change numeric/string value command
	CHANGE_PRIORITY_EVENT			= 17,	// Trigger it for change priority command
	CHANGE_END_POINT_EVENT			= 18,	// Trigger it for change end point command
	ENABLED_WITH_FOCUS_EVENT		= 19,	// switch to border state ENABLED_WITH_FOCUS, ISO VT "On Input Field Selection"
	FOCUS_LOST_EVENT			= 20,	// switch to border state ENABLED_WITHOUT_FOCUS, ISO VT "On Input Field Deselection"
	ESCAPE_EVENT				= 21,	// escape means is used while object is OPEN_FOR_INPUT, , ISO VT "On ESC"
	CONFIRM_EDITED_VALUE_EVENT		= 22,	// encoder is pushed or edited object touched while object is OPEN_FOR_INPUT (On Entry of Value)
	CONFIRM_NEW_EDITED_VALUE_EVENT		= 23,	// encoder is pushed while object is OPEN_FOR_INPUT and value really changed (On Entry of New Value)
	KEY_PRESS_EVENT				= 24,	// Key Press event, ISO VT "On Key Press"
	KEY_RELEASE_EVENT			= 25,	// Key Release event, ISO VT "On Key Release"
	CHANGE_CHILD_POSITION_EVENT		= 26,	// Trigger it for change child position command

	// Events internal to ISO VT but not mapped to ISO 11783-6
	ADD_OBJECT_POOL_EVENT			= 27,
	DELETE_OBJECT_POOL_EVENT		= 28,	// Delete the pool from memory due to delete pool command (sent from fast thread to slow thread)
	SELECT_INPUT_OBJECT_EVENT		= 29,	// Select (Focus/open for input) the object using the navigation scheme (sent from fast thread to slow thread)
	ESCAPE_INPUT_EVENT			= 30,	// escape/abort the operator input using the navigation scheme (sent from fast thread to slow thread)
	SHOW_AUX_SCREEN_EVENT			= 31,	// Display the Aux Assignment Screen (slow thread)
	REFRESH_AUX_SCREEN_EVENT		= 32,	// clear the current aux screen and send a RECREATE_AUX_SCREEN_DDOS_EVENT to fast thread (slow thread)
	RECREATE_AUX_SCREEN_DDOS_EVENT		= 33,	// Recreate the DDOs on the aux screen for the specified pool (fast thread)
	// not used: NAVIGATION_EVENT			= 34,	// forward an event from fast thread to the navigation scheme in the slow thread
	SHOW_ALARM_IN_SETTINGS_PROJECT		= 35,	// Set the path of the dead Working set icon and trigger alarm in display_seetings_project via variables
	TRIGGER_ALARM				= 36,	// trigger alarm when working set or aux input is dead (from fast thread to slow thread where icons for working set or aux input are created)
	AUX_INPUT_UNIT_DEAD_EVENT		= 37,	// Remove assignments of the aux inputs from the dead pool from all other pools (slow thread to fast thread)

	// ISO VT related but internal events
	DELETE_OBJECT_POOL_CONFIRM_EVENT	= 100,	// Triggered by DELETE_OBJECT_POOL_EVENT, this event triggers final destruction of the pool data
	REMOVE_POOL_WIDGETS_EVENT		= 101,	// Triggered by pool update which requires to remove all widgets while merging
	REMOVE_POOL_WIDGETS_CONFIRM_EVENT	= 102,	// Triggered by removal of page when pool should be merged with pool update
	CLEAR_J1939_ECUS_EVENT			= 103,	// Triggered on receiving low power event so as to clear the ecus in the stack.

	// UT related event introduced with UT plugin
	NEXT_OBJECT_POOL_EVENT			= 110,	// triggered by operator input to cycle pools in DDO
	SWAP_OBJECT_POOL_EVENT			= 111,	// triggered by operator input to exchange pools between view and primary DDO

	// ******************************* QT mouse events ********************************************
	/* These events are triggered when the widget controller is notified about these events.
	 * Inside the widget controller we reroute these events to the input event manager, where
	 * they will appear as MOUSE_PRESS_EVENT, MOUSE_RELESE_EVENT, etc.
	 * Having both: QT_MOUSE_PRESS_EVENT and MOUSE_PRESS_EVENT allows
	 * to distinguish between the raw events sent by QT and the ones coming from the
	 * input event manager. Meanwhile we also have 'toolkitEventOccurred()' and 'inputEventOccurred()'.
	 * If we ever move away from QT, it should be easy to check where the mouse events are coming
	 * from and route them to the input event manager. No further changes on the event mechanism
	 * would be needed ... well ... at least not for this purpose ;-)
	 */
	TOOLKIT_MOUSE_PRESS_EVENT		= 200,	//!< QT Mouse Press event
	TOOLKIT_MOUSE_RELEASE_EVENT		= 201,	//!< QT Mouse Release event
	TOOLKIT_MOUSE_DOUBLE_PRESS_EVENT	= 202,	//!< Qt Mouse Double press event
	TOOLKIT_MOUSE_MOVE_EVENT		= 203,	//!< QT Mouse move event
	TOOLKIT_MOUSE_ENTER_EVENT		= 204,	//!< QT Mouse Enter event
	TOOLKIT_MOUSE_EXIT_EVENT		= 205,	//!< QT Mouse Exit event
	TOOLKIT_WHEEL_EVENT			= 206,	//!< Qt Wheel event
	TOOLKIT_CLICK_EVENT			= 207,	//!< Click event. Whenever a mouse event cannot be send to a widget, a click event hs to be send
	TOOLKIT_MOUSE_ESCAPE_EVENT		= 208,	//!< widget is not touched anymore while finger still on screen (e.g. we could trigger a release event on a button)

	// ******************************* QT key events ********************************************
	TOOLKIT_KEY_PRESS_EVENT			= 300,	//!< Qt Key Press event
	TOOLKIT_KEY_RELEASE_EVENT		= 301,	//!< Qt Key Release event
	TOOLKIT_KEY_AUTO_REPEAT_PRESS_EVENT	= 302,	//!< Qt Key Press event
	TOOLKIT_KEY_AUTO_REPEAT_RELEASE_EVENT	= 303,	//!< Qt Key Release event

	// ******************************* QT Window events ********************************************
	// these events are triggered by the Qt when the widget (mainly the main window) gets the de/activated
	TOOLKIT_WINDOW_ACTIVE_EVENT		= 400,	//!< widget/main window got the focus so e.g.; activate the associated working set
	TOOLKIT_WINDOW_DEACTIVE_EVENT		= 401,	//!< widget/main window lost the focus so e.g.; deactivate the associated working set

	TOOLKIT_HIDE_EVENT			= 410,	//!< widget is hidden (e.g. we could trigger a release event on a button)

	// ******************************* key events **********************************************
	KEY_LONG_PRESS_EVENT			= 500,	//!< Key Long press event
	KEY_DOUBLE_PRESS_EVENT			= 501,	//!< Key double press event
	KEY_LONG_RELEASE_EVENT			= 502,	//!< Key Long Release event

	// ******************************* mouse events ********************************************
	MOUSE_PRESS_EVENT			= 600,	//!< Mouse Press event
	MOUSE_RELEASE_EVENT			= 601,	//!< Mouse Release event
	MOUSE_MOVE_EVENT			= 602,	//!< Mouse move event
	MOUSE_ENTER_EVENT			= 603,	//!< Mouse Enter event
	MOUSE_EXIT_EVENT			= 604,	//!< Mouse Exit event
	MOUSE_LONG_PRESS_EVENT			= 605,	//!< Mouse Long press event
	MOUSE_DOUBLE_PRESS_EVENT		= 606,	//!< Mouse Double press event
	MOUSE_LONG_RELEASE_EVENT		= 607,	//!< Mouse Long release event
	MOUSE_ESCAPE_EVENT			= 608,	//!< Finger dragged away and released elsewhere

	// ******************************* encoder events ******************************************
	ENCODER_MOVE_EVENT			= 700,	//!< Encoder Moved up/down
	// obsolete: 	ENCODER_PRESS_EVENT			= 701,	//!< Encoder Push
	// obsolete:	ENCODER_RELEASE_EVENT			= 702,	//!< Encoder release after push
	// obsolete:	ENCODER_LONG_PRESS_EVENT		= 703,	//!< Encoder long push event after press event
	// obsolete:	ENCODER_LONG_RELEASE_EVENT		= 704,	//!< Encoder long release event after long push
	// obsolete:	ENCODER_DOUBLE_PRESS_EVENT		= 705,	//!< Encoder double press event

	CONFIRM_INPLACE_INPUT_EVENT		= 710,	// Value confirmd in inplace editor

	// ******************************* power management events *********************************
	POWER_MANAGEMENT_TIMEOUT_EVENT			= 800,	//!< Power management event about timeout value
	POWER_MANAGEMENT_DEVICE_STATE_EVENT		= 801,	//!< Power management event about device state
	POWER_MANAGEMENT_POWER_STATE_EVENT		= 802,	//!< Power management event about power state
	POWER_MANAGEMENT_POWER_STATE_CHANGE_EVENT	= 803,	//!< Power management event about power state change
	POWER_MANAGEMENT_PROPAGATION_TIME_CHANGE_EVENT	= 804,	//!< propagation time changed
	POWER_MANAGEMENT_PROPAGATION_EVENT		= 805,	//!< propagation event before power state change

	// ******************************* other hardware events ***********************************
	IO_TOGGLE_EVENT	= 900,		//!< The value of an IO has changed

	// ******************************* navigation events ***************************************
	NAVIGATION_INIT_EVENT			=  999,	//!< Just need an ID to trigger navigation initialization (i.e. not really an event)
	OPEN_FOR_INPUT_EVENT			= 1000,	//!< switch to border state OPEN_FOR_INPUT
	ENABLED_OUTPUT_EVENT			= 1001,	//!< switch to border state ENABLED_OUTPUT
	DISABLED_EVENT				= 1002,	//!< switch to border state DISABLED

	CHANGE_EDITED_VALUE_EVENT		= 1003,	//!< encoder is turned while object is OPEN_FOR_INPUT

	CHANGE_INPLACE_VALUE_EVENT		= 1004,	//!< Intermediate value while inplace editing is active
	CONFIRM_INPLACE_VALUE_EVENT		= 1005,	//!< Confirmed value as result of inplace editing
//	ESCAPE_INPLACE_EDIT_EVENT		= 1006,	//!< Inplace editing was canceled

	// unused: CHANGE_SOFTKEY_LEVEL_EVENT		= 1007,	//!< Notifies change of softkey level
	// not triggered because CHANGE_SOFTKEY_LEVEL_EVENT is unused: CONFIRM_CHANGE_SOFTKEY_LEVEL_EVENT	= 1008,

	JUMP_TO_PAGE_EVENT			= 1009,

	CLICK_EVENT				= 1010,	//!< Trigger beeper click event

	// ******************************* variable events *****************************************
	// All variable events have the combined key ((index<<16)|subindex) stored in eventParameter2
	VARIABLE_VALUE_CHANGED_EVENT_INT		= 1100,	//!< inform (local) subscribers
	VARIABLE_VALUE_CHANGED_EVENT_STRING		= 1101,	//!< inform (local) subscribers
	// RESERVED: VARIABLE_GET_VALUE_REQUEST_EVENT		= 1102,	//!< ask to get value without subscription
	VARIABLE_GET_VALUE_RESPONSE_EVENT_INT		= 1103,	//!< get value without subscription
	VARIABLE_GET_VALUE_RESPONSE_EVENT_STRING	= 1104,	//!< get value without subscription

	// ***************************User defined Variable Events **********************************
	VARIABLE_CHANGE_EVENT_BY_OWNER		= 1200,	//!< Event raised when variable value is changed by Owner
	VARIABLE_CHANGE_REQUEST_EVENT_INT	= 1201,	//!< Event raised when value change of Integer Variable is requested, not by Owner
	VARIABLE_CHANGE_REQUEST_EVENT_STRING	= 1202,	//!< Event raised when value change of String Variable is requested, not by Owner
	VARIABLE_VALIDITY_CHANGED		= 1203,	//!< Event raised when the validity of a variable has changed

	// ******************************* protocol events *****************************************

	PROTOCOL_TIMEOUT_EVENT			= 1300,	//!< ep3 defines engaged (true) or resolved (false); parameters 1, 2 are implementation specific
	CAN_MAPPING_TIMEOUT_EVENT		= 1301,
	CAN_MAPPING_TIMEOUT_RESOLVED_EVENT	= 1302,
	HEARTBEAT_TIMEOUT_EVENT			= 1303,
	HEARTBEAT_TIMEOUT_RESOLVED_EVENT	= 1304,
	J1939_TP_MESSAGE_SEND			= 1305,	//!< result of sending a J1939 transport protocol message
	CAN_MAPPING_RECEIVED_EVENT		= 1306,

	// ******************************* logging events ******************************************
	LOG_TIMER1_EXPIRED_EVENT		= 1400,
	LOG_TIMER2_EXPIRED_EVENT		= 1401,
	LOG_TIMER3_EXPIRED_EVENT		= 1402,

	// ******************************* application events **************************************
	// 'application' is only true because PClient still only supports one project. Should be 'pool' events.
	APPLICATION_INIT_EVENT			= 1500,
	APPLICATION_REPEAT_EVENT		= 1501,
	APPLICATION_EXIT_EVENT			= 1502,
	// unused: APPLICATION_PROGRESS_EVENT		= 1503,
	// unused: APPLICATION_RESULT_EVENT		= 1504,
	APPLICATION_INIT_FINISHED_EVENT		= 1505,

	// ******************************* screen events *******************************************
	SCREEN_INIT_EVENT			= 1600,
	SCREEN_REPEAT_EVENT			= 1601,
	SCREEN_EXIT_EVENT			= 1602,
	SCREEN_INIT_FINISHED_EVENT		= 1603,

	// ******************************* EWAIN events *******************************************
	STATE_CHANGED_EVENT			= 1700,

	// obsolete: USER_CHANGED_VALUE_EVENT		= 1799,	// just introduced to avoid abuse of PROPERTY_CHANGE_EVENT

	PROPERTY_CHANGE_EVENT			= 1800,
//	SET_PROPERTY_EVENT			= 1801,	// never used to generate an event
//	SET_OBJECT_PROPERTY_EVENT		= 1802,	// unused since 2013
	SET_PROPERTY_IN_ALL_WIDGETS_EVENT	= 1803,
//	GET_CURRENT_VARIABLE_VALUES_EVENT	= 1804,
	CLEAN_UP_EVENT				= 1805,	// used to finally delete generated DDOs from AppAuxObjectPointerDDO

//	INTERNAL_FORWARD_EVENT			= 1900,	//!< needed to internally forward events from navigation scheme to DDO via widget controller

	TIMER_TIMEOUT_EVENT			= 2000,

	MEDIA_EVENT				= 2100,

	WE_MAX_EVENT_TYPE_NUMBER		= 9999
};


enum WEEventOrigin
{
    EVENT_ORIGIN_OPERATOR,
    EVENT_ORIGIN_SYSTEM,
    EVENT_ORIGIN_COMMAND
};





// back to the roots: Doesn't matter where I have to search for some time to figure out how an event works ...
// ... so instead of spending time here I can similarly well just search for the event type and see how it's used.




//VARIABLE_CHANGE_REQUEST_EVENT_STRING
//VARIABLE_CHANGE_REQUEST_EVENT_INT
//--> only an ID to find and trigger action in 'WEDEI::setVariableValue()'
// i.e. they don't use any event type





/**
 *	@struct WEEvent
 *
 *	@brief Event base structure to carry all kinds of events in the system.
 *
 */
struct WEEvent
{
	WEEventClass	eventClass{};	//!< Most sophisticated class of this event (i.e. the one whose 'virtual' methods are called)
	WEEventType	eventType{};	//!< Type of event e.g; keyPressEvent, mousePressEvent etc.
	std::uint32_t	timestamp{0};
	void *		widget{nullptr};	//!< Some arbitrary pointer with a suggested target (but usage depends on eventType)
	void *		controller{nullptr};	//!< For input and navigation events this is the widget controller (or nothing)
	void *		object{nullptr};	//!< Some arbitrary pointer with a suggested target (but usage depends on eventType)
	void *		pointer{nullptr};	//!< Some arbitrary pointer usage depends on eventType, sometimes it's a pool
	std::uint32_t	uvalue{0};	//!< Some arbitrary value of the event (usage depends on eventType)
	std::int32_t	value{0};	//!< Some arbitrary value of the event (usage depends on eventType)
	std::string strValue{""};

	/**
	 *	@brief Constructor creating an empty event of type p_type
	 *
	 *	@param p_type The type of the event
	 *
	 *	Preferring to NOT have a parameterless constructor as some members don't get
	 *	initialized if it is called in virtual inheritance used for the events.
	 *	I.e. prefer to know and prevent that at compile time.
	 */

	/**
	 *	@brief Constructor creating an empty event of type p_type
	 *
	 *	@param p_type The type of the event
	 *	@param p_class The class type this instance really is
	 *
	 */
	WEEvent(WEEventType p_type, WEEventClass p_class)
	:eventClass(p_class),
	 eventType(p_type)
	{
		//the other members are directly initialized in the member declaration, see above
	}

	virtual ~WEEvent()
	{
		// nothing to clean up - but not allowed to be abstract
	};


	// We could define an abstract 'clone()' method to copy the real events into and out of queues etc.
	// see https://stackoverflow.com/questions/31343225/create-instance-of-derived-abstract-class-in-c
	// Meanwhile decided to solve this problem with 'WESomeEvent'

//	/**
//	 *	@fn WEEvent(enum WEEventType p_type, std::uint32_t p_timestamp)
//	 *
//	 *	@brief Constructor creating an empty event of type p_type
//	 *
//	 *	@param p_type The type of the event
//	 *	@param p_timestamp Timestamp in milliseconds when the event was created
//	 */
//	WEEvent(WEEventType p_type, std::uint32_t p_timestamp)
//	{
//		recycle(p_type, p_timestamp);
//	}
//
//
//	/**
//	 *	@fn WEEvent(WEEvent &p_event)
//	 *
//	 *	@brief Copy constructor for non constant events
//	 *
//	 *	@param p_event The event to copy from
//	 */
//	WEEvent(WEEvent &p_event)
//	:eventType(p_event.eventType),
//	 timestamp(p_event.timestamp),
//	 widget(p_event.widget),
//	 controller(p_event.controller),
//	 object(p_event.object),
//	 pointer(p_event.pointer),
//	 uvalue(p_event.uvalue),
//	 value(p_event.value)
//	{
//	}
//
//
//	/**
//	 *	@fn WEEvent(const WEEvent &p_event)
//	 *
//	 *	@brief Copy constructor for constant events
//	 *
//	 *	@param p_event The event to copy from
//	 */
//	WEEvent(const WEEvent &p_event)
//	:eventType(p_event.eventType),
//	 timestamp(p_event.timestamp),
//	 widget(p_event.widget),
//	 controller(p_event.controller),
//	 object(p_event.object),
//	 pointer(p_event.pointer),
//	 uvalue(p_event.uvalue),
//	 value(p_event.value)
//	{
//	}
//
//// Seems to work well without overwriting operators e.g. for proper queue handling
//// I.e. structs are simple enough to live with the default operators.
//	/**
//	 *	@fn WEEvent & operator=(const WEEvent &p_event)
//	 *
//	 *	@brief Assignment operator
//	 *
//	 *	@param p_event The event to copy from
//	 */
//	WEEvent & operator=(const WEEvent &p_event)
//	{
//		if(this != &p_event)
//		{
//			this->eventType		= p_event.eventType;
//			this->timestamp		= p_event.timestamp;
//			this->widget		= p_event.widget;
//			this->controller	= p_event.controller;
//			this->object		= p_event.object;
//			this->pointer		= p_event.pointer;
//			this->uvalue		= p_event.uvalue;
//			this->value		= p_event.value;
//		}
//		return *this;
//	}
//
//
//	/**
//	 *	@brief Recycling/cleaning up the event and setting the event type
//	 *
//	 *	@param p_type The event type
//	 *
//	 *	Call this function whenever you want to reuse an existing event.
//	 *	NOTE: same e.g. event = WEEvent(EVENT_TYPE) but faster, because no no temporary object is needed
//	 */
//	inline void recycle(WEEventType p_type)
//	{
//		recycle(p_type, 0);
//	}
//
//
//	/**
//	 *	@brief Recycling/cleaning up the event and setting the event type and timestamp
//	 *
//	 *	@param p_type The event type
//	 *	@param p_timestamp The timestamp
//	 *
//	 *	Call this function whenever you want to reuse an existing event.
//	 *	Only type and timestamp are modified.
//	 *	NOTE: same e.g. event = WEEvent(EVENT_TYPE, timestamp) but faster, because no no temporary object is needed
//	 */
//	inline void recycle(WEEventType p_type, std::uint32_t p_timestamp)
//	{
//		this->eventType = p_type;
//		this->timestamp = p_timestamp;
//	}

	//**********functions for setting event parameters **************************************************
	//* always use these functions so that compiling fails when somebody changes the event structure    *
	//***************************************************************************************************


//	inline void setInternalForwardParameter(WEEvent * const p_event)
//	{
//		eventParameter.internalForward.event = p_event;
//	}
};	// struct WEEvent

std::string translateEventTypeToString(int p_eventType);
} // namespace

#endif /* WE_EVENT_H_ */
