/*
 *	we_input_event.h
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2020
 *
 *	@brief Based on WEEvent and intended for operator input.
 */

#ifndef SRC_EVENT_HANDLING_INCLUDE_WE_INPUT_EVENT_H_
#define SRC_EVENT_HANDLING_INCLUDE_WE_INPUT_EVENT_H_

#include "we_incident.h"

namespace WE
{

// Wanted to not convert key into encoder events anymore - which is done for the encoder push only now.
// Seems we have too many places where specific handling is implemented only for encoder turn - so keep that.
// Anyways will probably have to convert them to navigation events first before they do anything (once we have the configurable navigation scheme)
// TOOLKIT_WHEEL_EVENT
// ENCODER_MOVE_EVENT
//struct encoderMoveParameter
//{
//	// use weCode: std::int32_t direction;	//<! encoder rotation direction
//	// use WEEvent::value: std::uint32_t steps;		//<! number of steps the encoder was rotated
//	// use WEEvent::uvalue: std::uint32_t dynamicSteps;	//<! number of dynamic steps corresponding to rotated steps
//};

// TOOLKIT_CLICK_EVENT
// CLICK_EVENT
//struct clickParameter
//{
//	// empty
//};

// TOOLKIT_MOUSE_PRESS_EVENT
// TOOLKIT_MOUSE_RELEASE_EVENT
// TOOLKIT_MOUSE_MOVE_EVENT
// MOUSE_PRESS_EVENT
// MOUSE_RELEASE_EVENT
// MOUSE_MOVE_EVENT
// MOUSE_ENTER_EVENT
// MOUSE_EXIT_EVENT
// MOUSE_LONG_PRESS_EVENT
// MOUSE_DOUBLE_PRESS_EVENT
// MOUSE_LONG_RELEASE_EVENT
struct MouseParameter
{
	std::int32_t posX;		//!< absolute X Position on screen
	std::int32_t posY;		//!< absolute Y Position on screen
	// use WEInputEvent::modifiers: std::uint32_t modifiers;	//!< Qt modifiers
	// only used for drag events in VT i.e. only properly set for mouse move events: bool isPressed;		//!< mouse button was pressed?
	// use WEEvent::controller: void *controller;	//!< internal usage, needed for event routing
	// use WEEvent::object: void * object;		//!< the DDO - internal usage, needed for event routing
	// use WEEvent::uvalue: std::uint32_t activeMaskID;	//!< Used by ISO-VT only to report the active mask.
	// use WEEvent::value: int buttonValueToBeSet;	//!< Button needed a means to transport button value to other thread with the same event - meanwhile only uses key events for that
	int relX;		//! position inside the widget
	int relY;		//! position inside the widget
};

// TOOLKIT_KEY_PRESS_EVENT
// TOOLKIT_KEY_RELEASE_EVENT
// KEY_PRESS_EVENT
// KEY_RELEASE_EVENT
// KEY_LONG_PRESS_EVENT
// KEY_DOUBLE_PRESS_EVENT
// KEY_LONG_RELEASE_EVENT
//struct keyParameter
//{
//	// use WEInputEvent::weCode: std::int32_t weCode;		//!< WE key code/number
//	// use WEInputEvent::osCode: std::int32_t osCode;		//!< OS Code (Qt Code)
//	// use WEInputEvent::modifiers: std::uint32_t modifiers;	//!< Qt modifiers
//	// use WEEvent::object: void *object;		//!< used by ISO-VT only
//	// use WEEvent::controller: void *controller;	//!< filled by widget controller for event routing
//	// use WEEvent::uvalue: std::uint32_t activeMaskID;	//!< Used by ISO-VT only to report the active mask.
//	// use WEEvent::value: int buttonValueToBeSet;	//!< Button needs a means to transport button value to other thread with the same event
//};


// introduced CONFIRM_INPLACE_INPUT_EVENT just to pipe it through navigation scripts
// uses the same format as navigation events CONFIRM_EDITED_VALUE_EVENT and esp. CONFIRM_INPLACE_VALUE_EVENT


union inputEventParameter
{
	struct MouseParameter			mouse;
};



// inheriting WEIncident as many user inputs are also represented in CAN messages (at least for the VT)
struct WEInputEvent:virtual WEIncident
{

	std::int32_t weCode;		//!< WE key code/number
	std::int32_t osCode;		//!< OS Code (Qt Code)
	std::uint32_t modifiers;	//!< Qt modifiers (Shift/Ctrl/Alt ...)

	union inputEventParameter inputPara;


	/**
	 *	@brief Constructor directly defining the event type - UNKNOWN_EVENT if none provided
	 *
	 *	@param p_type The type of event to be created
	 *	@param p_class Keep default - is only modified by classes inheriting this one
	 *
	 *	@warning Logically p_type should clearly define what event class to generate but nothing is enforced there.
	 *
	 *	Marked as 'explicit' i.e. not allowed for implicit casting.
	 */
	explicit WEInputEvent(WEEventType p_type = UNKNOWN_EVENT, WEEventClass p_class = WE_INPUT_EVENT)
	:WEEvent(p_type, p_class),	// otherwise the empty constructor is called
	 WEIncident(p_type, p_class),
	 weCode(-1),
	 osCode(-1),
	 modifiers(0)
	{
		memset(&(this->inputPara), 0, sizeof(this->inputPara));
		this->value = (-1);	// SOFTKEY_UNDEFINED
	}

	virtual ~WEInputEvent()
	{
		// nothing to clean up
	}


	inline void setEncoderMoveParameter(std::int32_t p_direction, std::uint32_t p_steps, std::uint32_t p_dynamicSteps)
	{
		this->weCode	= p_direction;
		this->value	= p_steps;
		this->uvalue	= p_dynamicSteps;
	}


	inline void setMouseParameter(std::int32_t p_posX, std::int32_t p_posY, std::uint32_t p_modifiers, bool p_isPressed, int p_relX, int p_relY, void * p_widget = nullptr, std::uint32_t p_activeMaskID = 0xffff /*NULL_OBJECT_ID*/)
	{
		this->modifiers = p_modifiers;
		this->widget = p_widget;
		this->uvalue = p_activeMaskID;
		this->value = p_isPressed;	// e.g. released in a terminating 'move' event
		this->inputPara.mouse.posX = p_posX;
		this->inputPara.mouse.posY = p_posY;
		this->inputPara.mouse.relX = p_relX;
		this->inputPara.mouse.relY = p_relY;
		// only used in VT: this->para.mouse.isPressed = p_isPressed;
		// this->controller is set separately
	}

	inline void setKeyParameter(std::int32_t p_osKeyCode, std::uint32_t p_modifiers, void * const p_object = nullptr, std::uint32_t p_activeMaskID = 0xffff /*NULL_OBJECT_ID*/, std::int32_t p_weKeyCode = WE_KEY_UNKNOWN)
	{
		this->weCode = p_weKeyCode;
		this->osCode = p_osKeyCode;
		this->modifiers = p_modifiers;
		this->object = p_object;	// seems nobody sets that that via this method
		this->uvalue = p_activeMaskID;
		this->value = -1;	// buttonValueToBeSet
	}

	// this only exists to send an input event of type CONFIRM_INPLACE_INPUT_EVENT to navigation scheme
	inline void setConfirmEditedValueParameter(std::int32_t p_intValue, const std::string & p_stringValue = "", void * const p_object = nullptr, bool p_onlyChanged = false)
	{
		this->pointer	= nullptr;
		this->value	= p_intValue;		//!< value as integer
		this->strValue	= p_stringValue;	//!< value as string pointer
		this->object	= p_object;		//!< trigger action of this object in fast thread
		this->uvalue	= p_onlyChanged;	//!< To be used when Confirm edited value is misused for send value directly change events
		// leave this->controller untouched here - it's filled by the widget controller
	}
};



/**
 *	@brief Event listener interface that needs to be implemented to get notified of input events.
 *
 *	Subscription to events has to be implemented separately but they are expected to be handled in
 *	'inputEventOccurred()'.
 */
class WEInputEventListener
{
public:
	/**
	 *	@brief Handle events in this method.
	 *
	 *	@param[in] p_event The event that has occurred.
	 *
	 *	@returns an error code - WE_SUCCESS in case the event was successfully handled.
	 *
	 *	p_event should potentially be completed in this call but it should be carefully considered
	 *	to e.g. alter the event type since the event may be forwarded to other entities
	 *	after this method terminated. They are probably expecting the original content then.
	 *	I.e. generally p_event should be copied before the existing content is modified.
	 *
	 *	It is suggested to use this method to handle input events in the GUI context (slow thread).
	 *	If they need (additional) handling in data context (fast thread) it should be implemented
	 *	in 'incidentOccurred(WEIncident)'
	 *
	 *	@see incidentOccurred(WEIncident)
	 */
	virtual int inputEventOccurred(WEInputEvent & p_event) = 0;
	virtual ~WEInputEventListener() = default;
};
}	// namespace WE


#endif /* SRC_EVENT_HANDLING_INCLUDE_WE_INPUT_EVENT_H_ */
