/*
 *	we_some_event.h
 *
 *	WESomeEvent inherits all event types and is intended to be used e.g. in event queues.
 */

#ifndef SRC_EVENT_HANDLING_INCLUDE_WE_SOME_EVENT_H_
#define SRC_EVENT_HANDLING_INCLUDE_WE_SOME_EVENT_H_

// WEIncident:
#include "we_navigation_event.h"
#include "we_toolkit_event.h"
// WETrigger:
#include "we_signal.h"

#include <iostream>

namespace WE
{

struct WESomeEvent:public WEToolkitEvent, public WENavigationEvent, public WESignal
{
	// just need a struct which can fit all event types - no extra members or methods
	// but just inheriting all the leafs of the WEEvent hierarchy

	WEEventClass	createdFrom;	// remember what this really contains - makes casting easier

	/**
	 *	@brief Constructor directly defining the event type - UNKNOWN_EVENT if none provided
	 *
	 *	@param p_type The type of event to be created
	 *	@param p_class Keep default - is only modified by classes inheriting this one
	 *
	 *	@warning Logically p_type should clearly define what event class to generate but nothing is enforced there.
	 *
	 *	Marked as 'explicit' i.e. not allowed for implicit casting.
	 */
	explicit WESomeEvent(WEEventType p_type = UNKNOWN_EVENT, WEEventClass p_class = WE_SOME_EVENT)
	:	// otherwise the empty constructor is called ... and the order has to be correct to avoid warning
		// and it was 'bubble-sort' with the warning to find the right order :-(
	 WEEvent(p_type, p_class),
	 WEIncident(p_type, p_class),
	 WEInputEvent(p_type, p_class),
	 WETrigger(p_type, p_class),
	 WEToolkitEvent(p_type, p_class),
	 WENavigationEvent(p_type, p_class),
	 WESignal(p_type, p_class),
	 createdFrom(WE_EVENT)			// don't know yet - but this is the common base class
	{
		// Just a bare-bones constructor to create UNKNOWN_EVENT if queue is empty
	}

	/**
	 *	@brief Copy constructor for WEEvent - just using the one from WEEvent.
	 *
	 *	@warning This only copies members of the base class
	 */
	explicit WESomeEvent(WEEvent & p_event)
	:WEEvent(p_event.eventType, WE_SOME_EVENT),
	 WEIncident(p_event.eventType, WE_SOME_EVENT),
	 WEInputEvent(p_event.eventType, WE_SOME_EVENT),
	 WETrigger(p_event.eventType, WE_SOME_EVENT),
	 WEToolkitEvent(p_event.eventType, WE_SOME_EVENT),
	 WENavigationEvent(p_event.eventType, WE_SOME_EVENT),
	 WESignal(p_event.eventType, WE_SOME_EVENT),
	 createdFrom(p_event.eventClass)
	{
		switch (this->createdFrom)
		{
		case WE_EVENT:	// should never happen as this is the abstract base class
		{
			WEEvent * myself = this;
			*myself = p_event;
		}
			break;
		case WE_INCIDENT:	// C-style and static cast doesn't work with abstract base class :-(
					// So, we rely on not triggering an exception with the dynamic cast here
		{
			WEIncident * myself = this;
			*myself = dynamic_cast<WEIncident &>(p_event);
		}
			break;
		case WE_TOOLKIT_EVENT:
		{
			WEToolkitEvent * myself = this;
			*myself = dynamic_cast<WEToolkitEvent &>(p_event);
		}
			break;
		case WE_INPUT_EVENT:
		{
			WEInputEvent * myself = this;
			*myself = dynamic_cast<WEInputEvent &>(p_event);
		}
			break;
		case WE_NAVIGATION_EVENT:
		{
			WENavigationEvent * myself = this;
			*myself = dynamic_cast<WENavigationEvent &>(p_event);
		}
			break;
		case WE_TRIGGER:
		{
			WETrigger * myself = this;
			*myself = dynamic_cast<WETrigger &>(p_event);
		}
			break;
		case WE_SIGNAL:
		{
			WESignal * myself = this;
			*myself = dynamic_cast<WESignal &>(p_event);
		}
			break;
		case WE_SOME_EVENT:
			*this = dynamic_cast<WESomeEvent &>(p_event);
			break;
		default:
			std::cerr << "WESomeEvent(WEEvent &): ERRROR - created from unknown event class " << this->createdFrom << std::endl;
			this->createdFrom = WE_EVENT;
			break;
		}
	}

	virtual ~WESomeEvent()
	{
		// nothing to clean up
	}
};
} // namespace


#endif /* SRC_EVENT_HANDLING_INCLUDE_WE_SOME_EVENT_H_ */
