/*
 *	we_toolkit_event.h
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2020
 *
 *	@brief Based on WEEvent and intended for 'raw' operator input as provided by the toolkit.
 *
 *	Main purpose is to add some pointers before converting them into our own WEInputEvent type (which they inherit anyways).
 *	This is especially helpful for mouse events which the toolkit sends to the touched widget.
 */

#ifndef SRC_EVENT_HANDLING_INCLUDE_WE_TOOLKIT_EVENT_H_
#define SRC_EVENT_HANDLING_INCLUDE_WE_TOOLKIT_EVENT_H_


#include "we_input_event.h"

namespace WE
{



//TOOLKIT_WINDOW_ACTIVE_EVENT
// set in we_widget.cpp and only evaluated in the VT it seems (PClient only has one window)
//struct windowActiveParameter
//{
//	// empty
//};

//TOOLKIT_WINDOW_DEACTIVE_EVENT
//struct windowDeactiveParameter
//{
//	// empty
//};

// TOOLKIT_HIDE_EVENT
//struct hideParameter
//{
//	// use WEEvent::object: void *object;	//!< the object referenced with the event
//};


union toolkitEventParameter
{
};



// Toolkit event inherits 'WEInputEvent' as most toolkit events get converted to input events anyways.
struct WEToolkitEvent:virtual WEInputEvent
{

	union toolkitEventParameter tkpara;	// other name, so that we don't hide the one from WEInputEvent

	/**
	 *	@brief Constructor directly defining the event type - UNKNOWN_EVENT if none provided
	 *
	 *	@param p_type The type of event to be created
	 *	@param p_class Keep default - is only modified by classes inheriting this one
	 *
	 *	@warning Logically p_type should clearly define what event class to generate but nothing is enforced there.
	 *
	 *	Marked as 'explicit' i.e. not allowed for implicit casting.
	 */
	explicit WEToolkitEvent(WEEventType p_type = UNKNOWN_EVENT, WEEventClass p_class = WE_TOOLKIT_EVENT)
	:WEEvent(p_type, p_class),	// otherwise the empty constructor is called
	 WEIncident(p_type, p_class),
	 WEInputEvent(p_type, p_class)
	{
		memset(&(this->tkpara), 0, sizeof(this->tkpara));
	}

	virtual ~WEToolkitEvent() = default;

	// currently no methods to fill certain events
};


/**
 *	@brief Event listener interface that needs to be implemented to get notified of input events.
 *
 *	Subscription to events has to be implemented separately but they are expected to be handled in
 *	'inputEventOccurred()'.
 */
class WEToolkitEventListener
{
public:
	/**
	 *	@brief Handle events in this method.
	 *
	 *	@param[in] p_event The event that has occurred.
	 *
	 *	@returns an error code - WE_SUCCESS in case the event was successfully handled.
	 *
	 *	p_event should potentially be completed in this call but it should be carefully considered
	 *	to e.g. alter the event type since the event may be forwarded to other entities
	 *	after this method terminated. They are probably expecting the original content then.
	 *	I.e. generally p_event should be copied before the existing content is modified.
	 *	Note that this method quite likely is the exception from the rule: Typically it
	 *	converts the WEToolkitEvent event type into the matching WEInputEvent type and forwards it
	 *	to some implementation of inputEventOccurred().
	 *
	 *	It is suggested to use this method to handle toolkit events in the GUI context (slow thread).
	 *	If they need (additional) handling in data context (fast thread) it should be implemented
	 *	in 'incidentOccurred(WEIncident &)'
	 *
	 *	@see incidentOccurred(WEIncident &)
	 *	@see inputEventOccurred(WEInputEvent &)
	 */
	virtual int toolkitEventOccurred(WEToolkitEvent & p_event) = 0;
	virtual ~WEToolkitEventListener() = default;
};
}	// namespace WE

#endif /* SRC_EVENT_HANDLING_INCLUDE_WE_TOOLKIT_EVENT_H_ */
