/**
 *	@file we_border_profile.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief QT Widget Class
 *
 *	@see WEWidget
 *
 *	@warning no default border anymore! If you want to have a border you have to set one in the widget!
 *
 *	usage:
 *
 * 	weltWidget->setBorderProfile(&myProfile);
 *
 * 	That sets a whole profile at once into the widget.
 *
 */
#ifndef WE_BORDER_PROFILE_H_
#define WE_BORDER_PROFILE_H_

//Included the header files for Data Members
#include <vector>
#include <array>
#include "../../../common/include/we_color.h"
#include "../../include/we_datatypes_gui.h"

namespace WE
{
/**
 *	@class WEBorderProfile
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@brief Class that represents a Widget's border profile
 *
 *	@namespace WE
 */
class WEBorderProfile
{
public:

	/**
	 *	@fn WEBorderProfile()
	 *
	 *	@brief default constructor
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 */
	WEBorderProfile();


	/**
	 *	@fn WEBorderProfile(int borderWidth,
	 * 			Color disabledColor,
	 * 			Color openForInputColor,
	 * 			Color withFocusColor,
	 * 			Color withoutFocusColor)
	 *
	 *	@brief convenience constructor to create a border profile
	 *	of the specified borderWidth with the specified colors
	 *	for each border state.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 * 	@param[in] borderWidth		Border width to be used for the border profile
	 *	@param[in] disabledColor	Color to be used for the border in Disabled state
	 *	@param[in] openForInputColor	Color to be used for the border in Open For Input state
	 *	@param[in] withFocusColor	Color to be used for the border in Enabled With Focus state
	 *	@param[in] withoutFocusColor	Color to be used for the border in Enabled Without Focus state
	 *
	 *	This internally calls the constructor using WEColor references.
	 */
	WEBorderProfile(int p_borderWidth,
			Color p_disabledColor,
			Color p_openForInputColor,
			Color p_withFocusColor,
			Color p_withoutFocusColor);


	/**
	 *	@fn WEBorderProfile(
	 *			int borderWidth,
	 * 			WE::WEColor & p_disabledColor,
	 * 			WE::WEColor & p_openForInputColor,
	 * 			WE::WEColor & p_withFocusColor,
	 * 			WE::WEColor & p_withoutFocusColor);
	 *
	 *	@brief convenience constructor to create a border profile
	 *	of the specified borderWidth with the specified colors
	 *	for each border state.
	 *
	 * 	@param[in] borderWidth		Border width to be used for the border profile
	 *	@param[in] disabledColor	Color to be used for the border in Disabled state
	 *	@param[in] openForInputColor	Color to be used for the border in Open For Input state
	 *	@param[in] withFocusColor	Color to be used for the border in Enabled With Focus state
	 *	@param[in] withoutFocusColor	Color to be used for the border in Enabled Without Focus state
	 *
	 *
	 */
	WEBorderProfile(
			int p_borderWidth,
			WE::WEColor & p_disabledColor,
			WE::WEColor & p_openForInputColor,
			WE::WEColor & p_withFocusColor,
			WE::WEColor & p_withoutFocusColor);


	/**
	 *	@fn WEBorderProfile(const WEBorderProfile& p_borderprofile)
	 *
	 *	@brief copy constructor
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 * 	@param[in] p_borderprofile other borderprofile
	 *
	 *	needed when giving the whole object
	 */
	WEBorderProfile(const WEBorderProfile& p_borderprofile);


	/*--------------------------------------------------------*/
	/**
	 *	@fn ~WEBorderProfile()
	 *
	 *	@brief Standard destructor of WEBorderProfile
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	Destroys the border profile object
	 */
	/*--------------------------------------------------------*/
	virtual ~WEBorderProfile();


	/**
	 *	@fn int setBorderProfile(WEBorderState p_state, std::vector<WEColor> p_borderProfile)
	 *
	 *	@brief Sets complete profile for a defined state and creates a internal copy
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  	@returns WEError code
	 *
	 *	will reset only the given state
	 */
	int setBorderProfile(WEBorderState p_state, const std::vector<WEColor> & p_borderProfile);


	/**
	 *	@fn getBorderProfile(WEBorderState p_state)
	 *
	 *	@brief Sets one border to a color in a state
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @returns WEError code
	 *
	 *	will not reset the whole profile. Only overwrite a available color. Will check on index
	 */
	const std::vector<WEColor> & getBorderProfile(WEBorderState p_state) const;


	/**
	 *	@fn setBorderColor(WEBorderState p_state, unsigned int p_index, WEColor p_color)
	 *
	 *	@brief Sets one border index to a color in a state only for convinience
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  	@returns WEError code
	 *
	 * 	@param[in] p_state border state
	 * 	@param[in] p_index index inside the border state
	 * 	@param[in] p_color set it to that color
	 *
	 *	will not reset the whole profile. Only overwrite a available color. Will check on index if index < border profile size
	 */
	int setBorderColor(WEBorderState p_state, unsigned int p_index, WEColor p_color);


	/**
	 *	@fn int addBorderColor(WEBorderState p_state, WEColor p_color)
	 *
	 *	@brief Add one border to a profile
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  	@returns WEError code
	 *
	 * 	@param[in] p_state state
	 * 	@param[in] p_color color
	 *
	 *	add one border color at the end of the border profile
	 */
	int addBorderColor(WEBorderState p_state, WEColor p_color);


	/**
	 *	@fn WEColor getBorderColor(WEBorderState p_state, unsigned int p_index)
	 *
	 *	@brief gets one border color
	 *
	 *	@param _state Border state from which the color should be retrieved.
	 *	@param p_index The line index whose color should be returned.
	 *
	 *  	@returns WEColor stored, white in case of error
	 *
	 */
	WEColor getBorderColor(WEBorderState p_state, unsigned int p_index) const;


	/**
	 *	@fn int clearBorderProfile(WEBorderState p_state)
	 *
	 *	@brief gets one border color
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  	@returns WEError code
	 *
	 * 	@param[in] p_state state to be cleared
	 *
	 *	clear a border for a state.
	 */
	int clearBorderProfile(WEBorderState p_state);


	/**
	 *	@fn int clearAllBorderProfiles()
	 *
	 *	@brief gets one border color
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  	@returns WEError code
	 *
	 *	clear all borders
	 */
	int clearAllBorderProfiles();


	/**
	 *	@fn int getBorderSize(WEBorderState p_state)
	 *
	 *	@brief gets border count
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 * 	@param[in] p_state state of border to return the size
	 *
	 *  	@returns border count or -1 if error
	 */
	int getBorderSize(WEBorderState p_state) const;


	/**
	 *	@fn int release()
	 *
	 *	@brief release all
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	clear all borders, release members, nulls members
	 */
	void release();


	/**
	 *	@fn operator=(const WEBorderProfile& p_borderprofile)
	 *
	 *	@brief assignment operator
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 * 	@param[in] p_borderprofile const ref to other borderprofile
	 */
	WEBorderProfile& operator=(const WEBorderProfile& p_borderprofile);


	/**
	 *	@fn bool operator==(const WEBorderProfile &o) const
	 *
	 *	@brief comparison operator
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 * 	@param[in] p_borderprofile const ref to other borderprofile
	 */
	bool operator==(const WEBorderProfile &p_o) const;


	/**
	 *	@fn bool operator!=(const WEBorderProfile &o) const
	 *
	 *	@brief comparison operator
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 * 	@param[in] p_borderprofile const ref to other borderprofile
	 */
	bool operator!=(const WEBorderProfile &p_o) const;


	/**
	 *	@fn bool operator<(const WEBorderProfile &o) const
	 *
	 *	@brief comparison operator
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 * 	@param[in] p_borderprofile const ref to other borderprofile
	 */
	bool operator<(const WEBorderProfile &p_o) const;


	bool areColorsEqual(int p_state) const;

private:

	/**
	 *	@fn init()
	 *
	 *	@brief initialize the members.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 */
	void init();


	/**
	 *	@fn int release(WEBorderState p_state)
	 *
	 *	@brief release only one state
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 * 	@param[in] p_state state to be released
	 */
	void release(WEBorderState p_state);


	/**
	 *	@fn int create(WEBorderState p_state)
	 *
	 *	@brief create only one state, to have a container for a copy
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 * 	@param[in] p_state state to be created
	 */
	void create(WEBorderState p_state);

	/**
	 *	@fn int setBorderProfile (int borderWidth,
			WE::WEColor & p_disabledColor,
			WE::WEColor & p_openForInputColor,
			WE::WEColor & p_withFocusColor,
			WE::WEColor & p_withoutFocusColor)
	 *
	 *	@brief convenience menthod to create a border profile
	 *	of the specified borderWidth with the specified colors
	 *	for each border state.
	 *
	 * 	@param[in] borderWidth		Border width to be used for the border profile
	 *	@param[in] disabledColor	Color to be used for the border in Disabled state
	 *	@param[in] openForInputColor	Color to be used for the border in Open For Input state
	 *	@param[in] withFocusColor	Color to be used for the border in Enabled With Focus state
	 *	@param[in] withoutFocusColor	Color to be used for the border in Enabled Without Focus state
	 *
	 *
	 */
	int setBorderProfile (int p_borderWidth,
			WE::WEColor & p_disabledColor,
			WE::WEColor & p_openForInputColor,
			WE::WEColor & p_withFocusColor,
			WE::WEColor & p_withoutFocusColor);

	void updateEqualColors(int p_state);

	//!<Array for the Border Profile Colors
	std::array<std::vector<WEColor>, BORDERPROFILE_STATE_COUNT> borderProfiles{}; //!<internal border storages
	std::array<bool, BORDERPROFILE_STATE_COUNT> stateCreated{false, false, false, false, false, false}; //!<stores if a state was set or created
	std::array<bool, BORDERPROFILE_STATE_COUNT> equalColors{false, false, false, false, false, false};

	//SERIALIZATION
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int );
};
} // end of namespace WE

BOOST_CLASS_EXPORT_KEY(WE::WEBorderProfile)

#endif /* WE_BORDER_PROFILE_H_ */
