/**
 *	@file we_gui_utils.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief  This file contains common utility function that that can be used
 *	through out the gui.
 *
 */


#ifndef WE_GUI_UTILS_H_
#define WE_GUI_UTILS_H_

#include <map>
#include <string>
//Included here to avoid inclusion of WESize everywhere WEGUIUtils is used.
#include "../../include/we_datatypes_gui.h"
#include "../../../common/include/we_size.h"


namespace WE
{

/**
 *	@class WEGUIUtils
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief This class contains the static utility methods to
 *	be used with in an application.
 *
 *	@namespace WE
 *
 */
class WEGUIUtils
{
public:
	WEGUIUtils() = delete;
	~WEGUIUtils() = delete;

	/**
	 *	@fn getScreenDimensions()
	 *
	 *	@brief This method retrieves the dimension of the Screen
	 *	which includes the x,y width and height of the screen.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns WERect The dimensions of the screen.
	 *	@returns nullptr if WEApplication is not initialised.
	 *
	 *	@see WEApplication.
	 *
	 *	@warning A WEApplication must be instantiated before using
	 *	this method to retrieve the dimension of the screen.
	 *	this Please see WEApplication to create/initialise the
	 *	application.
	 *
	 *	@warning Do not delete the returned size pointer! Destructor of
	 *	WEGUIUtils will delete the size!
	 */
	static const WESize & getScreenDimensions();

	/**
	 *	@fn fromUnicode(WE::WEEncodeType p_encodeType,std::wstring p_str)
	 *
	 *	@brief This method converts the given wide string p_str(in UTF-8) to the encoding scheme specified p_encodeType.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns string The string encoded in the specified format.
	 *
	 *	@Warning: The current implementation requires Qt and should hence reside in a "qt" folder.
	 */
	static std::string fromUnicode(WE::WEEncodeType p_encodeType, const std::wstring & p_str);

	/**
	 *	@fn static std::wstring toUnicode(WE::WEEncodeType encodeType,std::string p_str)
	 *
	 *	@brief This method decodes the given encoded (p_encodeType) string p_str in to unicode.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns string The string encoded in the specified format.
	 *
	 *	@Warning: The current implementation requires Qt and should hence reside in a "qt" folder.
	 */
	static std::wstring toUnicode(WE::WEEncodeType p_encodeType, const std::string & p_str);


	/**
	 *	@fn static std::string convertEncoding(WE::WEEncodeType p_encodeTypeIn, WE::WEEncodeType p_encodeTypeOut, std::string p_str)
	 *
	 *	@brief Convert p_str from one char set encoding to the other.
	 *
	 *	@param[in] p_encodeTypeIn Character set p_str is encoded in.
	 *	@param[in] p_encodeTypeOut Character set returned string should be encoded in.
	 *	@param[in] p_str String that should be recoded.
	 *
	 *	@returns Converted string
	 *
	 *	This method can e.g. be used to convert a ISO 8859-1 encoded string to UTF-8 or vice versa.
	 *
	 *	@Warning: The current implementation requires Qt and should hence reside in a "qt" folder.
	 *
	 *	@see fromUnicode()
	 *	@see toUnicode()
	 */
	static std::string convertEncoding(WE::WEEncodeType p_encodeTypeIn, WE::WEEncodeType p_encodeTypeOut, const std::string & p_str);



	static bool isCharacterPrintable(wchar_t p_ch);

	/**
	 *	@fn getTextEncodeName(WEEncodeType p_EncodeType)
	 *
	 *	@brief get the Qt Text Codec name for the specified encode type
	 *
	 *	@author satheeshkumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 * 	@param[in] p_EncodeType Enocode Type
	 *
	 * 	@returns Qt Text Codec Name
	 *
	 * 	Intentionally non-const!
	 */
	static std::string getTextEncodeName(WEEncodeType p_EncodeType);

	/**
	 *	@fn static std::map<WEEncodeType,std::string> getTextEncodingNameMap()
	 *
	 *	@brief Retrieve the Text Encoding Scheme enum to name map.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 * 	@param none
	 *
	 * 	@returns Text Encoding Scheme enum to name map.
	 */
	static const std::map<WEEncodeType,std::string> & getTextEncodingNameMap();
private:
	static WESize screenSize;
	static std::map<WEEncodeType,std::string> encodeEnumToNameMapping;	//Text Encoding Scheme enum to name map.

	//Initialise the Text Encoding Scheme enum to name map.
	static void initEnumToNameMapping();
};
}

#endif /* WE_GUI_UTILS_H_ */
