/**
 *	@file we_rect.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file contains the definition of the WERect data Type.
 *
 */


#ifndef WE_RECT_H_
#define WE_RECT_H_

#include <QRect>

namespace WE
{
/**
 *	@class WERect
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief Defines a Rectangle
 *
 *	@namespace WE
 *
 *	@see we_datatypes_gui.cpp
 *
 *	This class defines a rectangle which is internally represented by a x
 *	coordinate, y coordinate, width and height.
 */

class WERect
{
public:

	/**
	 *	@fn WERect
	 *
	 *	@brief Default Constructor of the Rectangle
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 */
	WERect() = default;

	/**
	 *	@fn WERect(int x,int y,int width,int height)
	 *
	 *	@brief Constructs a Rectangle with the specified location and size.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_x X Coordinate of the Rectangle.
	 *	@param[in] p_y Y Coordinate of the Rectangle.
	 *	@param[in] p_width Width of the Rectangle.
	 *	@param[in] p_height Height of the Rectangle.
	 *
	 */
	WERect(int p_x, int p_y, int p_width, int p_height);

	/**
	 *	@fn ~WERect()
	 *
	 *	@brief Destructor of the WERect Class.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	Destructor of the WERect Class.
	 */
	virtual ~WERect() = default;

	/**
	 *	@fn setX(int x)
	 *
	 *	@brief Sets the X Coordinate of the Rectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_x X Coordinate of the Rectangle.
	 *
	 *	Sets the X Coordinate of the Rectangle.
	 */
	void setX(int p_x);

	/**
	 *	@fn getX(void)
	 *
	 *	@brief Gets the X Coordinate of the Rectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns 	X Coordinate of the Rectangle.
	 *
	 *	Gets the X Coordinate of the Rectangle.
	 */
	int getX(void) const;

	/**
	 *	@fn setY(int y)
	 *
	 *	@brief Sets the Y Coordinate of the Rectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_y Y Coordinate of the Rectangle.
	 *
	 *	Sets the Y Coordinate of the Rectangle.
	 */
	void setY(int p_y);

	/**
	 *	@fn getY(void)
	 *
	 *	@brief Gets the Y Coordinate of the Rectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns 	Y Coordinate of the Rectangle.
	 *
	 *	Gets the Y Coordinate of the Rectangle.
	 */
	int getY(void) const;

	/**
	 *	@fn setWidth(int width)
	 *
	 *	@brief Sets the Width of the Rectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_width Width of the Rectangle.
	 *
	 *	Sets the width of the Rectangle.
	 */
	void setWidth(int p_width);

	/**
	 *	@fn getWidth(void)
	 *
	 *	@brief Gets the Width of the Rectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns 	Width of the Rectangle.
	 *
	 *	Gets the Width of the Rectangle.
	 */
	int getWidth(void) const;

	/**
	 *	@fn setHeight(int height)
	 *
	 *	@brief Sets the Height of the Rectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_height Height of the Rectangle.
	 *
	 *	Sets the height of the Rectangle.
	 */
	void setHeight(int p_height);

	/**
	 *	@fn getHeight(void)
	 *
	 *	@brief Gets the Height of the Rectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns 	Height of the Rectangle.
	 *
	 *	Gets the Height Coordinate of the Rectangle.
	 */
	int getHeight(void) const;

	/**
	 *	@fn setSize(int width,int height)
	 *
	 *	@brief Sets the Size of the Rectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_width Width of the Rectangle.
	 *	@param[in] p_height Height of the Rectangle.
	 *
	 *	Sets the size of the Rectangle.
	 */
	void setSize(int p_width,int p_height);

	void setGeometry(int p_x, int p_y, int p_width, int p_height);


	/**
	 *	@fn bool enclosing(WERect * p_other, WERect * p_result);
	 *
	 *	@brief Provide rectangle just covering this and the other one.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_other Rectangle with which the enclosing rect will be created.
	 *	@param p_result Rectangle which should contain the enclosing area. nullptr if result should only be the return value.
	 *
	 *	@returns TRUE if this enclosing rectangle has width and height unequal to zero
	 *	@returns FALSE Otherwise
	 *
	 *	@see contains(int x,int y)
	 *	@see bool intersects(WERect &)
	 *	@see bool intersection(WERect *, WERect *)
	 *
	 *
	 *
	 */
	bool enclosing(WERect * p_other, WERect * p_result = nullptr) const;


	/**
	 *	@fn bool intersection(WERect * p_other, WERect * p_result);
	 *
	 *	@brief Returns TRUE if this rectangle intersects with rectangle p_other.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_other Rectangle with which the intersection is to be checked.
	 *	@param p_result Rectangle which should contain the intersection area. nullptr if result should only be the return value.
	 *
	 *	@returns TRUE if this rectangle intersects with rectangle p_other
	 *	@returns FALSE Otherwise
	 *
	 *	@see contains(int x,int y)
	 *	@see bool intersects(WERect &)
	 *	@see bool enclosing(WERect *, WERect *)
	 *
	 *
	 *	Returns TRUE if this rectangle intersects with rectangle rect (there is
	 *	at least one pixel that is within both rectangles).
	 *
	 */
	bool intersection(WERect * p_other, WERect * p_result = nullptr) const;


	/**
	 *	@fn intersects(WERect &rect)
	 *
	 *	@brief Returns TRUE if this rectangle intersects with rectangle rect.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] rect Rectangle with which the intersection is to be checked.
	 *
	 *	@returns TRUE if this rectangle intersects with rectangle rect ;Otherwise
	 *	@returns FALSE
	 *
	 *	@see contains(int x,int y)
	 *
	 *	Returns TRUE if this rectangle intersects with rectangle rect (there is
	 *	at least one pixel that is within both rectangles).
	 *
	 */
	bool intersects(const WERect &p_rect) const;


	/**
	 *	@fn contains(int x,int y)
	 *
	 *	@brief Returns TRUE if the point(x,y) lies within the rectangle or on
	 *	its boundary.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] x Coordinate of the point to be checked for containment.
	 *	@param[in] y Coordinate of the point to be checked for containment.
	 *
	 *	@returns TRUE if this point(x,y) lies inside the rectangle; Otherwise
	 *	@returns FALSE
	 *
	 *	@see intersects(WERect &rect)
	 *
	 */
	bool contains(int p_x, int p_y) const;

	/**
	 *	@fn WERect operator-(const WERect &rect)
	 *
	 *	@brief  minus operator for the WERectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] rect Constant reference of the RHS of the -operator. The value
	 *	that is to be subtracted.
	 *
	 *	@returns WERect The result of the subtraction.
	 *
	 */
	WERect operator-(const WERect &p_rect);

	/**
	 *	@fn WERect& operator=(const WERect &rect)
	 *
	 *	@brief  Assignment operator for the WERectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] rect Constant reference of the RHS of the = operator. The value
	 *	which is to be copied to the LHS of the operator.
	 *
	 *	@returns WERect LHS of the assignment operator.
	 *
	 */
	WERect& operator=(const WERect &p_rect) = default;

	/**
	 *	@fn WERect& operator=(const QRect &rect)
	 *
	 *	@brief  Assignment operator for QRect.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] rect Constant reference of the RHS of the = operator. The value
	 *	which is to be copied to the LHS of the operator.
	 *
	 *	@returns WERect LHS of the assignment operator.
	 *
	 */
	WERect& operator=(const QRect &p_rect);

	WERect(const WERect & p_rect) = default;

	explicit WERect(const QRect & p_rect);

	/**
	 *	@fn bool operator==(const WERect &rect) const
	 *
	 *	@brief  Comparison operator for WERect
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@returns The created QRect
	 *
	 */
	bool operator==(const WERect &p_rect) const;

	/**
	 *	@fn bool operator!=(const WERect &rect) const
	 *
	 *	@brief  Comparison operator for WERect
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@returns The created QRect
	 *
	 */
	bool operator!=(const WERect &p_rect) const;

	/**
	 *	@fn QRect toQRect() const
	 *
	 *	@brief  Converts this WERect to a QRect
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@returns The created QRect
	 *
	 */
	QRect toQRect() const;

private:
	int x{0};	//!< X Coordinate of the Rectangle
	int y{0};	//!< Y Coordinate of the Rectangle
	int width{0};	//!< Width of the Rectangle
	int height{0};	//!< Height of the Rectangle
};
}
#endif /* WE_RECT_H_ */
