/**
 *	@file we_region.h
 *
 *	@author TANMOY.DEB AT ARICENT.COM
 *	@author TANMOY DEB
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief The implementation of Region in the WEApi Layer.
 *
 */
#ifndef WE_REGION_H_
#define WE_REGION_H_

#include <memory>
#include "we_rect.h"

namespace WE
{

class ToolkitWERegion;

/**
 *	@enum WERegionType
 *
 *	@author TANMOY.DEB AT ARICENT.COM
 *	@author TANMOY DEB
 *
 *	@brief Specifies the shape of the region to be created.
 *
 *	@namespace WE
 */
enum WERegionType
{
	REGION_ELLIPSE,		//the region is an ellipse inside the rectangle
	REGION_RECTANGLE	//the region covers the entire rectangle
};
/**
 *	@class WERegion
 *
 *	@author Tanmoy Deb
 *	@author Tanmoy.deb AT aricent.com
 *
 *	@brief The WERegion class specifies a clip region for a painter
 *
 * WERegion is used with WEPainter::setClipRegion() to limit the paint area to what needs
 * to be painted.WERegion is the best tool for minimizing the amount of screen area to be
 * updated by a repaint.
 *
 *
 */


class WERegion
{



public:
	/**
	 *	@fn WERegion(WERect &p_rect, WE::WERegionType p_regionType)
	 *
	 *	@brief The definition of WERegion.
	 *
	 *	@author Tanmoy Deb
	 *	@author tanmoy.deb AT aricent.com
	 *
	 *	@param p_rect The rectangular region.
	 *	@param[in] p_regionType This determines the region type. Either Rectangular or Elliptical.
	 *
	 *	@see WEWidget.cpp WEWidget.h WERect.cpp
	 *
	 */
	WERegion(WERect &p_rect, WE::WERegionType p_regionType);

	WERegion();


	/**
	 *	@fn ~WERegion()
	 *
	 *	@brief That deletes the WERegion.
	 *
	 *	@author Tanmoy Deb
	 *	@author tanmoy.deb AT aricent.com
	 *
	 *
	 */
	virtual ~WERegion();


	/**
	 *	@fn getToolkitWERegion()
	 *
	 *	@author Tanmoy Deb
	 *	@author tanmoy.deb AT aricent.com
	 *
	 *
	 *	@returns The reference to the ToolkitWERegion. That will be used to access the
	 *	QRegion from the WERegion.
	 *
	 *	Get the toolkitWERegion.
	 */
	ToolkitWERegion* getToolkitWERegion() const;

	/**
	 *	@fn void setToolkitWERegion(ToolkitWERegion *p_toolkitWERegion)
	 *
	 *	@brief Set the toolkit Region
	 *
	 *	@author tanmoy.deb AT aricent.com
	 *	@author Tanmoy Deb
	 *
	 *	@param[in] p_toolkitRegion The toolkit Region
	 *
	 *	Set the ToolkitWERegion.
	 */
	void setToolkitWERegion(ToolkitWERegion *p_toolkitWERegion);

	/**
	 *	@fn void init()
	 *
	 *	@brief initialize the Region
	 *
	 *	@author tanmoy.deb AT aricent.com
	 *	@author Tanmoy Deb
	 *
	 *	Set the ToolkitWERegion.
	 */
	void init();

	/**
	 *	@fn getBoundingRect()
	 *
	 *	@brief Returns the bounding rectangle of this region.
	 *
	 *	@author Tanmoy Deb
	 *	@author tanmoy.deb AT aricent.com
	 */
	WERect getBoundingRect() const;

	/**
	 *	@fn operator+ (WERegion &p_weRegion)
	 *
	 *	@brief Returns a region which is p_weRegion added with this region.
	 *
	 *	@author Tanmoy Deb
	 *	@author tanmoy.deb AT aricent.com
	 *
	 *
	 *	@param[in] WERegion
	 *
	 *	@returns WERegion
	 *
	 *
	 */
	WERegion operator+ (const WERegion &p_weRegion);

	/**
	 *	@fn operator- (WERegion &p_weRegion)
	 *
	 *	@brief Returns a region which is r subtracted from this region.
	 *
	 *	@author tanmoy.deb AT aricent.com
	 *	@author Tanmoy Deb
	 *
	 *	@param[in] WERegion The region to be subtracted from the another region.
	 *
	 *	@returns WERegion the result region
	 *
	 */
	WERegion operator- (const WERegion &p_weRegion);

	// The copy Constructor

	WERegion(const WERegion& p_WERegion);

	WERegion& operator=(const WERegion& p_other); // Assignment operator

private:
	// The region Type
	WE::WERegionType regionType{REGION_RECTANGLE};
	// The rectangle
	WERect rect{};
	// The toolkit WERegion
	std::unique_ptr<ToolkitWERegion> toolkitWERegion;
};
}
#endif /* WE_REGION_H_ */
