/**
 *	@file qt_we_region.h
 *
 *	@author tanmoy deb AT aricent.com
 *	@author Tanmoy Deb
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file contains the definition of the QT specific definition of the WERegion
 *
 */

#ifndef QT_WE_PAINTER_PATH_H_
#define QT_WE_PAINTER_PATH_H_

#include <QtGui/QPainterPath>
#include "../include/toolkit_we_painter_path.h"
#include "../../common/include/we_region.h"

namespace WE
{
class WEFont;

/*
 *	This class inherit from the two basic classes:
 *	ToolkitWEPainterPath - interface class for the PainterPaths
 *	QPainterPath - graphic Path class
 */
class QTWEPainterPath : public ToolkitWEPainterPath, public QPainterPath
{
public:
	/**
	 *	@fn QTWEPainterPath(ToolkitWEPainterPath *p_toolkitWEPainterPath)
	 *
	 *	@brief Standard constructor of QTWEPainterPath class
	 *
	 *	@author TANMOY.DEB AT ARICENT.COM
	 *	@author TANMOY DEB
	 *
	 *	@param p_toolkitWEPainterPath Reference to the ToolkitWEPainterPath for which this QTWEPainterPath is created.
	 *
	 *	The constructor create the qt region interface.By default
	 *	the parent is nullptr indicating a window will be created.
	 */
	/*--------------------------------------------------------*/
	explicit QTWEPainterPath(ToolkitWEPainterPath *p_toolkitWEPainterPath);


	/**
	 *	@fn  QTWEPainterPath(WERect &p_rect, WE::WERegionType regionType, bool antialiasing, int startAngle=0, int spanAngle=0)
	 *
	 *	@brief The standard constructor for the WERegion.
	 *
	 *	@author TANMOY.DEB AT ARICENT.COM
	 *	@author TANMOY DEB
	 *
	 *	@param[in] p_rect The rectangle of the painter path.
	 *	@param[in] p_regionType The type of region (elipse or rectangle)
	 *	@param[in] p_antialiasing Flag if anti aliasing shall be used
	 *	@param[in] p_startAngle The Starting Angle
	 *	@param[in] p_spanAngle The extent angle
	 *
	 *	The base Constructor of the QTWEPainterPath.
	 */
	QTWEPainterPath(const WERect &p_rect, WE::WERegionType p_regionType, bool p_antialiasing, int p_startAngle=0, int p_spanAngle=0);


	/*--------------------------------------------------------*/
	/**
	 *	@fn ~QTWEPainterPath()
	 *
	 *	@brief Standard destructor of QTWEPainterPath class
	 *
	 *	@author TANMOY.DEB AT ARICENT.COM
	 *	@author TANMOY DEB
	 *
	 *	The destructor deletes the qt Region interface.
	 */
	/*--------------------------------------------------------*/
	virtual ~QTWEPainterPath() = default;


	/**
	 *	@fn ToolkitWEPainterPath* subtracted ( ToolkitWEPainterPath  *p_toolkitWEPainterPath )
	 *
	 *	@brief Returns a region which is subtracted from this region.
	 *
	 *	@author tanmoy.deb AT aricent.com
	 *	@author Tanmoy Deb
	 *
	 *	@param[in] p_toolkitWEPainterPath The region to be substracted from the another region.
	 *
	 *	@returns The result region
	 *
	 */
	ToolkitWEPainterPath* subtracted ( ToolkitWEPainterPath  *p_toolkitWEPainterPath ) override;


	/**
	 *	@fn ToolkitWEPainterPath* united ( ToolkitWEPainterPath *p_toolkitWEPainterPath )
	 *
	 *	@brief Returns a region which is added with this region.
	 *
	 *	@author Tanmoy Deb
	 *	@author tanmoy.deb AT aricent.com
	 *
	 *
	 *	@param[in] p_toolkitWEPainterPath
	 *
	 *	@returns ToolkitWEPainterPath The result region.
	 */
	ToolkitWEPainterPath* united ( ToolkitWEPainterPath *p_toolkitWEPainterPath ) override;


	int getStartAngle() const override;


	int getSpanAngle() const override;


	/**
	 *	@fn WERect boundingRect()
	 *
	 *	@brief Returns the bounding rectangle of this region.
	 *
	 *	@author TANMOY.DEB AT ARICENT.COM
	 *	@author TANMOY DEB
	 *
	 *	@returns The bounding rectangle of this region.
	 *
	 *	@see WERect.cpp
	 *
	 *	Returns the bounding rectangle of this region.
	 */
	WERect boundingRect() override;


	void addText(WEFont &p_font, const std::string & p_text, int p_flag) override;

	void addEllipse (WERect & p_rect) override;
	void addPolygon (WEPoint * p_points, unsigned int p_numberOfPoints) override;

	void setFillRule(int p_rule) override;

private:
	// The private constructor of the QTWEPainterPath
	explicit QTWEPainterPath (const QPainterPath &p_painterPath);
};
}


#endif /* QT_we_painter_path_H_ */
