/**
 *	@file toolkit_we_font.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief The Toolkit Interface  for the Font class
 *
 */

#ifndef TOOLKIT_WE_FONT_H_
#define TOOLKIT_WE_FONT_H_

#include "../../include/we_datatypes_gui.h"	// WEAlignment

class QFont;

namespace WE
{
class WEPainter;
class WERect;

/**
 *	@class WEFont
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@brief This class specifies a font to be used for drawing text.
 *
 *	@namespace WE
 *
 *	This is the abstract class, used as an interface to the toolkit.
 *	Should not have Constructor, only virtual Destructor function.
 *	All the rest methods should be override in the tollkit class
 */
class ToolkitWEFont
{
public:
	/**
	 *	@fn ~ToolkitWEFont()
	 *
	 *	@brief Default Destructor
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 */
	virtual ~ToolkitWEFont() = default;

//-----------------------------------------------------------------------------
	/**
	 *	@fn virtual void setFontFamily(std::string p_familyName) = 0
	 *
	 *	@brief Sets the family of the font
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_familyName Name of the Font family.
	 */
	virtual void setFontFamily(const std::string & p_familyName) = 0;

	/**
	 *	@fn virtual const std::string & getFontFamily() = 0
	 *
	 *	@brief Returns Name of the font.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns Name of the Font.
	 *
	 *	Name of the font is nullptr terminated
	 */
	virtual std::string getFontFamily() const = 0;

	/**
	 *	@fn virtual void setFontStyle(int fontStyle) = 0
	 *
	 *	@brief Sets the style of the font to Normal(if enable is FALSE) or Italic(if enable is TRUE)
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_fontStyle the font style that shall be set
	 */
	virtual void setFontStyle(int p_fontStyle) = 0;

	/**
	 *	@fn virtual int getFontStyle() = 0
	 *
	 *	@brief Gets the style of the font: Normal - FALSE, Italic - TRUE
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@return true if italic, otherwise returnes false
	 */
	virtual int getFontStyle() const = 0;

	/**
	 *	@fn virtual void setFontSize(int p_fontSize, int p_fontWidth = 0) = 0
	 *
	 *	@brief Sets the size of the font.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_fontSize Sets the size of the font.
	 *	@param[in] p_fontWidth Width of a single character if a specific one should be required
	 */
	virtual void setFontSize(int p_fontSize, int p_fontWidth = 0) = 0;

	/**
	 *	@fn virtual int getFontSize() = 0
	 *
	 *	@brief Returns the size of the font.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns size of the Font.
	 */
	virtual int getFontSize() const = 0;
//-----------------------------------------------------------------------------
	/**
	 *	@fn virtual int getStringWidth(WEPainter* p_painter, std::string p_text) = 0
	 *
	 *	@brief Returns the size in pixel of the string.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns Size in pixel of the string.
	 */
	virtual int getStringWidth(WEPainter* p_painter, const std::string & p_text) const = 0;
	virtual int getStringWidth(WEPainter* p_painter, const std::wstring & p_text) const = 0;


	/**
	 *	@fn int getStringWidthWithoutPainter(std::string p_text)
	 *
	 *	@brief Get width of the string provided using latin 1 encoding.
	 *
	 *	@param p_text Text to be evaluated
	 *
	 *	@returns The width in pixels of p_text rendered for the screen with this font.
	 *
	 */
	virtual int getStringWidthWithoutPainter(const std::string & p_text) const = 0;
	virtual int getStringWidthWithoutPainter(const std::wstring & p_text) const = 0;


	/**
	 *	@fn virtual WERect getBoundingRect(WERect p_rect, WEAlignment p_align, bool p_autowrap, std::string p_text, bool p_wrapOnHyphen = false) = 0
	 *
	 *	@brief Get bounding rectangle of the string provided using latin 1 encoding.
	 *
	 *	@param p_rect	Rectangle where p_text should be shown, e.g. widget boundaries.
	 *	@param p_align	Flags defining how p_text should be aligned.
	 *	@param p_autowrap True if p_text should be automatically wrapped
	 *	@param p_text	Text to be evaluated.
	 *	@param p_wrapOnHyphen	False: hyphen character (0x2D) is not considered line breaking
	 *
	 *	@returns The bounding rect of p_text rendered for the screen with this font and the parameters given.
	 *
	 */
	virtual WERect getBoundingRect(const WERect & p_rect, WEAlignment p_align, bool p_autowrap, const std::string & p_text, bool p_wrapOnHyphen = false) const = 0;


	/**
	 * 	@fn WERect getBoundingRect(WERect p_rect, WEAlignment p_align, bool p_autowrap, std::wstring p_text, bool p_wrapOnHyphen = false)
	 *
	 *	@brief Get bounding rectangle of the string provided using unicode encoding.
	 *
	 *	@param p_rect	Rectangle where p_text should be shown, e.g. widget boundaries.
	 *	@param p_align	Flags defining how p_text should be aligned.
	 *	@param p_autowrap True if p_text should be automatically wrapped
	 *	@param p_text	Text to be evaluated.
	 *	@param p_wrapOnHyphen	False: hyphen character (0x2D) is not considered line breaking
	 *
	 *	@returns The bounding rect of p_text rendered for the screen with this font and the parameters given.
	 *
	 *	Qt consider the hyphen character (0x2D) also a line separator character for the text wrapping , so wrapping
	 *	occurs at the hyphen character by default. So when the wrapOnHyphen option is disabled then wrap should not
	 *	happen at the hyphen and this is achieved by replacing the hyphen character as non breaking hyphen character.
	 *	This option is basically required by ISO-VT
	 */
	virtual WERect getBoundingRect(const WERect & p_rect, WEAlignment p_align, bool p_autowrap, const std::wstring & p_text, bool p_wrapOnHyphen = false) const = 0;


	/**
	 *	@fn virtual WERect getBoundingRectUTF8(WERect p_rect, WEAlignment p_align, bool p_autowrap, std::string p_text, bool p_wrapOnHyphen = false) = 0
	 *
	 *	@brief Get bounding rectangle of the string provided using unicode encoding.
	 *
	 *	@param p_rect	Rectangle where p_text should be shown, e.g. widget boundaries.
	 *	@param p_align	Flags defining how p_text should be aligned.
	 *	@param p_autowrap True if p_text should be automatically wrapped
	 *	@param p_text	Text to be evaluated.
	 *	@param p_wrapOnHyphen	False: hyphen character (0x2D) is not considered line breaking
	 *
	 *	@returns The bounding rect of p_text rendered for the screen with this font and the parameters given.
	 *
	 */
	virtual WERect getBoundingRectUTF8(const WERect & p_rect, WEAlignment p_align, bool p_autowrap, const std::string & p_text, bool p_wrapOnHyphen = false) const = 0;

	/**
	 *	@fn virtual const QFont * getQFont() const = 0
	 *
	 *	@brief Returns a pointer to the QFont object of this ToolkitWEFont,
	 *	       if the toolkits implementation is based on it
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@returns pointer to a QFont object
	 */
	virtual const QFont * getQFont() const = 0;

	/**
	 * Method to add a font to the application that is not installed on the OS
	 * @param p_fontPath Path to the font (including font file)
	 */
	static void addFontToApplication(const std::string & p_fontPath);

	static void printAvailableFonts();
};
}
#endif /* TOOLKIT_WE_FONT_H_ */
