/**
 *	@file qt_we_widget.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief QT Widget Class
 *
 *	@see WEWidget, WEToolkitFont, qt_we_font.cpp
 *
 *	In this file the font for the QT library is defined
 */
#ifndef QT_WE_FONT_H_
#define QT_WE_FONT_H_

//Included the header file for Inherited Classes
#include <QtGui/QFont>
#include "../include/toolkit_we_font.h"

namespace WE
{
/**
 *	@class QTWEFont
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@brief The QTWEFont class secifies a font used for drawing text
 *	This class is derived from the two basic classes:
 *	WEFont - interface class for the fonts
 *	QFont - graphic library class
 *
 *	@namespace WE
 *
 */
class QTWEFont : public ToolkitWEFont, public QFont
{
public:
	/**
	 *	@fn QTWEFont()
	 *
	 *	@brief Standard constructor of QTWEFont.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	Contructs a font object that uses the application's default font
	 */
	QTWEFont();

	/**
	 *	@fn QTWEFont(std::string & p_family, int p_size, int p_style)
	 *
	 *	@brief Constructor of QTWEFont that also initializes the interface.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_family
	 *	@param[in] p_size
	 *	@param[in] p_style
	 *
	 *	Constructs a font object with the specified family, size and style settings
	 *	If size is <=0, it is set to DEF_FONT_SIZE = 12
	 */
	QTWEFont(const std::string & p_family, int p_size, int p_style);

	/**
	 *	@fn ~QTWEFont()
	 *
	 *	@brief Standard destructor of QTWEFont
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	Destroys the font object and frees all allocated resources
	 */
	virtual ~QTWEFont() = default;


	/**
	 *	@fn void setFontFamily(std::string p_family)
	 *
	 *	@brief Function to set the font family
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_family for the font family name (for Example "Times")
	 *
	 *	Overloaded function from QT
	 *	Set the family name of the font.
	 */
	void setFontFamily(const std::string & p_fontFamily) override;


	/**
	 *	@fn std::string getFontFamily() const
	 *
	 *	@brief Function to get the font family
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@return the font family name
	 *
	 *	Overloaded function from QT
	 *	Returns the requested font family name,
	 *	i.e. the name set in the constructor or the last setFontName() call
	 */
	std::string getFontFamily() const override;


	/**
	 *	@fn void setFontSize(int p_fontSize, int p_fontWidth = 0)
	 *
	 *	@brief Function to set the sizeh of the font
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_fontSize for the font size
	 *	@param[in] p_fontWidth Width of a single character if a specific one should be required
	 *
	 *	Set the font size to p_size pixel
	 */
	void setFontSize(int p_fontSize, int p_fontWidth = 0) override;


	/**
	 *	@fn int getFontSize()
	 *
	 *	@brief Function to get the size of the font
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@return the pixel size of the font
	 *
	 *	Overloaded function from QT: QFontMetrics
	 */
	int getFontSize() const override;

	/**
	 *	@fn void setFontStyle(int p_fontStyle)
	 *
	 *	@brief Function to set the font style
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_fontStyle for the font style from the type WEFontStyle
	 *
	 *	Set the style of the font to the mask.
	 *	The mask should be build as following: WE::STYLE_BOLD | WE::STYLE_ITALIC
	 *	Each of the values in the mask should be of the enum WEFontStyle
	 */
	void setFontStyle(int p_fontStyle) override;

	/**
	 *	@fn int getFontStyle()
	 *
	 *	@brief Function to get the style mask of the font
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@return the mask with the styles of the font
	 *
	 *	The return value is integer and is a bitmask with the current font style
	 *	The mask Example:WE::STYLE_BOLD | WE::STYLE_ITALIC | WE::STYLE_UNDERLINED
	 *	Each of the values in the mask is of the enum WEFontStyle
	 */
	int getFontStyle() const override;

	/**
	 *	@fn int getStringWidth(WEPainter* p_painter, std::string p_text)
	 *
	 *	@brief Function to meagure the width of the string
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param p_painter Pointer to painter
	 *	@param[in] p_text for the text, which width should be calculated
	 *
	 *	@return the width of a string in pixel
	 *
	 *	Overloaded function from QT: QFontMetrics
	 */
	int getStringWidth(WEPainter* p_painter, const std::string & p_text) const override;
	int getStringWidth(WEPainter* p_painter, const std::wstring & p_text) const override;


	/**
	 *	@fn int getStringWidthWithoutPainter(std::string p_text)
	 *
	 *	@brief Get width of the string provided using latin 1 encoding.
	 *
	 *	@param p_text Text to be evaluated
	 *
	 *	@returns The width in pixels of p_text rendered for the screen with this font.
	 *
	 */
	int getStringWidthWithoutPainter(const std::string & p_text) const override;
	int getStringWidthWithoutPainter(const std::wstring & p_text) const override;

	/**
	 *	@fn WERect getBoundingRect(WERect p_rect, WEAlignment p_align, bool p_autowrap, std::string p_text, bool p_wrapOnHyphen = false)
	 *
	 *	@brief Get bounding rectangle of the string provided using latin 1 encoding.
	 *
	 *	@param p_rect	Rectangle where p_text should be shown, e.g. widget boundaries.
	 *	@param p_align	Flags defining how p_text should be aligned.
	 *	@param p_autowrap True if p_text should be automatically wrapped
	 *	@param p_text	Text to be evaluated.
	 *	@param p_wrapOnHyphen	False: hyphen character (0x2D) is not considered line breaking
	 *
	 *	@returns The bounding rect of p_text rendered for the screen with this font.
	 *
	 */
	WERect getBoundingRect(const WERect & p_rect, WEAlignment p_align, bool p_autowrap, const std::string & p_text, bool p_wrapOnHyphen = false) const override;

	/**
	 *	@fn WERect getBoundingRect(WERect p_rect, WEAlignment p_align, bool p_autowrap, std::wstring p_text, bool p_wrapOnHyphen = false)
	 *
	 *	@brief Get bounding rectangle of the string provided using unicode encoding.
	 *
	 *	@param p_rect	Rectangle where p_text should be shown, e.g. widget boundaries.
	 *	@param p_align	Flags defining how p_text should be aligned.
	 *	@param p_autowrap True if p_text should be automatically wrapped
	 *	@param p_text	Text to be evaluated.
	 *	@param p_wrapOnHyphen	False: hyphen character (0x2D) is not considered line breaking
	 *
	 *	@returns The bounding rect of p_text rendered for the screen with this font and the parameters given.
	 *
	 *	Qt consider the hyphen character (0x2D) also a line separator character for the text wrapping , so wrapping
	 *	occurs at the hyphen character by default. So when the wrapOnHyphen option is disabled then wrap should not
	 *	happen at the hyphen and this is achieved by replacing the hyphen character as non breaking hyphen character.
	 *	This option is basically required by ISO-VT
	 */
	WERect getBoundingRect(const WERect & p_rect, WEAlignment p_align, bool p_autowrap, const std::wstring & p_text, bool p_wrapOnHyphen = false) const override;

	/**
	 *	@fn WERect getBoundingRectUTF8(WERect p_rect, WEAlignment p_align, bool p_autowrap, std::string p_text, bool p_wrapOnHyphen = false)
	 *
	 *	@brief Get bounding rectangle of the string provided using unicode encoding.
	 *
	 *	@param p_rect	Rectangle where p_text should be shown, e.g. widget boundaries.
	 *	@param p_align	Flags defining how p_text should be aligned.
	 *	@param p_autowrap True if p_text should be automatically wrapped
	 *	@param p_text	Text to be evaluated.
	 *	@param p_wrapOnHyphen	False: hyphen character (0x2D) is not considered line breaking
	 *
	 *	@returns The bounding rect of p_text rendered for the screen with this font and the parameters given.
	 *
	 */
	WERect getBoundingRectUTF8(const WERect & p_rect, WEAlignment p_align, bool p_autowrap, const std::string & p_text, bool p_wrapOnHyphen = false) const override;

	/**
	 *	@fn const QFont * getQFont() const
	 *
	 *	@brief Returns a pointer to the underlying QFont object of this QTWEFont
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@returns pointer to a QFont object
	 */
	const QFont * getQFont() const override;

	static void printAvailableFonts();

private:
	/**
	 *	@fn WERect getBoundingRect(WERect p_rect, WEAlignment p_align, bool p_autowrap, QString p_text, bool p_wrapOnHyphen = false)
	 *
	 *	@brief Get bounding rectangle of the string provided.
	 *
	 *	@param p_rect	Rectangle where p_text should be shown, e.g. widget boundaries.
	 *	@param p_align	Flags defining how p_text should be aligned.
	 *	@param p_autowrap True if p_text should be automatically wrapped
	 *	@param p_text	Text to be evaluated.
	 *	@param p_wrapOnHyphen	False: hyphen character (0x2D) is not considered line breaking
	 *
	 *	@returns The bounding rect of p_text rendered for the screen with this font.
	 *
	 */
	WERect getBoundingRect(const WERect & p_rect, WEAlignment p_align, bool p_autowrap, QString p_text, bool p_wrapOnHyphen = false) const;

	/**
	 *	@fn int setISOVTFontParameters(int p_height, int p_width)
	 *
	 *	@brief Do the fine tuning required for ISO-VT while avoiding memory leak cause by Qt.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_height Character == line height
	 *	@param[in] p_width Width of a single character, 0 for proportional font
	 *
	 *	@return An error code - WE::WE_SUCCESS in case of success
	 *
	 *	Fine tune parameters so that height and width of the characters match requirements
	 *	of ISO 11783-6.
	 *
	 *	@warning Current implementation assumes that modified fonts "DejaVu Sans Low" and "DejaVu Sans Mono Low" are applied.
	 *
	 */
	int setISOVTFontParameters(int p_height, int p_width);

	unsigned int fontStyle{0};

	QFont * qFont{nullptr};
};
} // end of namespace WE
#endif /* QT_WE_FONT_H_ */
