/**
 *	@file we_image.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Image Class
 *
 *	@see WEToolkitImage
 */
#ifndef WE_IMAGE_H_
#define WE_IMAGE_H_

#include <memory>
#include <vector>
//Included files Data Members
#include "../../../common/include/we_color.h"
// AXIVION NextLine Generic-LocalInclude : Removing the next include and replacing it with forward decalarations has A LOT of side effects => keep it
#include "../../texture_cache/include/we_texture_cache.h"

namespace WE
{

class ToolkitWEImage;
class WEWidget;

typedef enum
{
	BMP,	//!< Windows Bitmap
	GIF,	//!< Graphic Interchange Format
	JPG,	//!< Joint Photographic Experts Group
	PNG,	//!< Portable Network Graphics
	TIFF	//!< Tagged Image File Format
}WEImageFormat; //<MKA 04.08.2009>: changed WEFormat to WEImageFormat


class WEImage
{
public:

	/**
	 *	@fn WEImage()
	 *
	 *	@brief Standard constructor of WEImage
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 */
	WEImage();

	/**
	 *	@fn ~WEImage()
	 *
	 *	@brief Standard destructor of WEImage
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 */
	virtual ~WEImage();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEImage(const WEImage &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEImage & operator=(const WEImage &) = delete;

	/**
	 *	@fn void release(bool p_forceRelease = false)
	 *
	 *	@brief releases memory
	 *
	 *	@param[in] p_forceRelease true : release the image forcibly.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 */
	void release(bool p_forceRelease = false);

	/**
	 *	@fn void getToolkitInterface()
	 *
	 *	@brief gets pointer of the toolkit interface
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns pointer to WEToolkitImage
	 *
	 * returns a pointer to the WEToolkitImage Interface
	 *
	 */
	std::shared_ptr<ToolkitWEImage> getToolkitInterface() const
	{
		return this->toolkitImage;
	}

	/**
	 *	@fn getWidth()
	 *
	 *	@brief get image width
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns width of image
	 *
	 */
	int getWidth() const;

	/**
	 *	@fn getHeight()
	 *
	 *	@brief get image height
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns height of image
	 *
	 */
	int getHeight() const;

	/**
	 *	@fn createTransparent(bool p_bTransparent)
	 *
	 *	@brief creates a transparent image by color set with setTransparentColor
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_bTransparent create = true or false for reset.
	 *
	 *	makes image transparent with the set RGB values by setTransparentColor. Transparency
	 *	by color only works with no alpha channel available. Alpha channels are not
	 *	overwritten.
	 */
	int createTransparent(bool p_bTransparent);

	/**
	 *	@fn setTransparent(bool p_bTransparent, int p_colorIndex)
	 *
	 *	@brief set the color index for transparency
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_bTransparent create = true or false for reset.
	 *	@param[in] p_colorIndex color index works only for images max indexed8
	 *
	 */
	int setTransparent(bool p_bTransparent, int p_colorIndex);


	/**
	 *	@fn int createImage(const std::string & p_strFile)
	 *
	 *	@brief creates image and stores pointer internally
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_strFile File + full path
	 *
	 *	@returns WE_error code
	 *
	 * creates an Image that can be stored as a pointer. In later implementation it
	 * looks up a list of images and only creates a new image if the image is not
	 * present in memory.
	 *
	 *
	 *	@see createImage(std::string p_strFile, std::uint8_t p_R, std::uint8_t p_G, std::uint8_t p_B)
	 *	@see createImage(std::string p_strFile, WEColor& p_transparency)
	 */
	int createImage(const std::string & p_strFile);


	/**
	 *	@fn int createRawImage(std::string p_strFile)
	 *
	 *	@brief Creates a basic raw image
	 *
	 *	@author satheeshkumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *	@param[in] p_strFile image file path
	 *	@param[in] p_strFile filename
	 *	@param[in] p_R transparency
	 *	@param[in] p_G transparency
	 *	@param[in] p_B transparency
	 *
	 *	@returns WE_error code
	 *
	 *	Creates a raw image which does not depend on the image list or any widget.
	 *	Main purpose is to use this in fill patterns where transparency has to be OFF
	 *	even if the same image is used on screen (in an image DDO) with transparency ON.
	 *
	 *	The image list currently only looks for the image path - so rather than modifying the
	 *	whole image list for this we keep the image within this WEImage.
	 *
	 *	If somebody understanding all the image container and image list magic should have some spare time we could integrate this in image list.
	 */
	int createRawImage(const std::string & p_strFile);


	/**
	 *	@fn int createRawImage(QPixmap *p_pixmap);
	 *
	 *	@brief creates a WEImage
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] p_pixmap the QPixmap
	 *
	 *	@returns WE_error code
	 */
	int createRawImage(QPixmap *p_pixmap);


	/**
	 *	@fn int createRawImage(QSGTexture *p_texture, WETextureCache::Data *p_cacheData=nullptr)
	 *
	 *	@brief creates a WEImage from the p_texture
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] p_texture the QSGTexture
	 *	@param[in] p_cacheData optional pointer to the WETextureCache::Data object of the texture
	 *
	 *	@returns WE_error code
	 */
	int createRawImage(QSGTexture *p_texture, WETextureCache::Data *p_cacheData=nullptr);


	/**
	 *	@fn int deleteRawImage()
	 *
	 *	@brief Deletes the created raw image
	 *
	 *	@author satheeshkumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *
	 */
	void deleteRawImage();


	/**
	 * 	@brief This is a static method used to store images e.g. from an ISO pool to file system.
	 *
	 *	This mainly exists to store images coming from an ISO-VT object pool with the original
	 *	references to the color table.
	 *	Refer to ISO-VT pool parser to see how the different formats defined in ISO 11783-6 are
	 *	converted into the common format expected here.
	 *
	 *	@param indexed8ImageData pointer to the raw image data - one byte per pixel, new lines start at next 32 bit boundary
	 *	@param imageWidth	width in pixels
	 *	@param imageHeight	height in pixels
	 *	@param colorTable	e.g. obtained from WEApp::ColorTable::getAllColors()
	 *	@param imageFileFormat	image file format (e,g; PNG, JPG etc)
	 *	@param imageFilePath	absolute path of the image file which shall be created
	 *	@returns		an error code - WE_SUCCESS in case of success
	 *
	 *	@see WEApp::PoolParser::saveRawDataAsImageFile()
	 *	@see WEApp::ColorTable::getAllColors()
	 */
	static int saveRawDataAsImageFile(
			unsigned char * p_indexed8ImageData,
			unsigned int p_imageWidth,
			unsigned int p_imageHeight,
			std::vector<unsigned int> & p_colorTable,
			const std::string & p_imageFileFormat,
			const std::string & p_imageFilePath);


	/**
	 *	@fn setTransparentColor(std::uint8_t p_R, std::uint8_t p_G, std::uint8_t p_B)
	 *
	 *	@brief sets the RGB values for the transparent color
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_R transparency
	 *	@param[in] p_G transparency
	 *	@param[in] p_B transparency
	 *
	 *	sets the transparent color later used by createTransparent
	 */
	int setTransparentColor(std::uint8_t p_R, std::uint8_t p_G, std::uint8_t p_B);


	/**
	 *	@fn setTransparentColor(WEColor& p_transparency)
	 *
	 *	@brief sets the RGB values for the transparent color
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_transparency WEColor& p_transparency
	 *
	 *	sets the transparent color later used by createTransparent
	 */
	int setTransparentColor(const WEColor& p_transparency);


	/**
	 *	@fn hasAlphaChannel()
	 *
	 *	@brief returns if an Alpha channel is available or not
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns true or false
	 *
	 */
	bool hasAlphaChannel() const;


	/**
	 *	@fn isTransparent()
	 *
	 *	@brief Returns true if the image has transparency added by transparency color.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns true or false
	 *
	 */
	bool isTransparent() const;


	int setSize(int p_width, int p_height, bool p_resize = true);


	void setWidget(WEWidget * p_widget);


	bool isSVG(void) const;
	bool isGIF(void) const;

	const std::string & getFilePath(void) const;


	/**
	 *	@fn bool isTextureCacheable(void) const
	 *
	 *	@brief Returns true if the texture of this image can be added to the texture cache
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@see void setIsTextureCacheabl(bool cacheable)
	 *
	 *	@returns true if the texture of this image can be added to the texture cache
	 */
	bool isTextureCacheable(void) const;


	/**
	 *	@fn void setIsTextureCacheable(bool cacheable)
	 *
	 *	@brief Set if the texture of this image can be added to the texture cache
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] cacheable true if the image is cacheable
	 *
	 *	@see bool isTextureCacheable(void) const
	 */
	void setIsTextureCacheable(bool p_cacheable);


	/**
	 *	@fn static bool getUseImageList( void )
	 *
	 *	@brief Check if every WEImage will maintain its own ToolkitWEImage or not
	 *
	 *	@param None
	 *
	 *	@returns true if image list should be used and one ToolkitWEImage instance should be shared for all images with the same name
	 *	@returns false if image list should not be used and each instance of WEImage maintains its own ToolkitWEImage
	 *
	 *	This method returns the value of the respective private static member.
	 *	If only some images should be handled without the image list, use
	 *	createRawImage(std::string) instead.
	 *
	 *	@see int createRawImage( std::string )
	 *	@see static int setUseImageList( bool )
	 */
	static bool getUseImageList( void );


	/**
	 *	@fn static int setUseImageList( bool p_useIt )
	 *
	 *	@brief Define if every WEImage will maintain its own ToolkitWEImage or not.
	 *
	 *	@param[in] p_useIt - true if image list should be used, false otherwise
	 *
	 *	@returns WE::WE_SUCCESS
	 *
	 *	This method sets the value of the respective private static member.
	 *	If only some images should be handled without the image list, use
	 *	createRawImage(std::string) instead.
	 *
	 *	This is intended to stay unchanged during the lifetime of an application
	 *	i.e. the idea is to define the policy at application initialization.
	 *
	 *	@warning createImage(const std::string&) will use the current setting whenever it is called.
	 *
	 *	@see int createRawImage( std::string )
	 *	@see static bool getUseImageList( void )
	 */
	static int setUseImageList( bool p_useIt );




	static int debugSection;

private:

	/**
	 *	@fn init()
	 *
	 *	@brief Standard init for class
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	private at first.
	 */
	void init();


	std::shared_ptr<ToolkitWEImage> toolkitImage;		//!< Pointer to toolkit image
	bool imageCreatedWithoutTheList;	//!< Remember if we have to delete our image or tell the list to reduce usage count
	bool textureIsCacheable;		//!< True if the texture of this image can be added to the texture cache
	WEColor transparentColor;		//!< Color to be used for transparency mask

	std::string fileName;			//!< Image path

	WEWidget * widget;

	static bool useWEImageList;		// True (default): Use the same image ToolkitWEImage whenever the same filename is used, False: Every WEImage has its own ToolkitWEImage.
};
} // end of namespace WE
#endif /* WE_IMAGE_H_ */
