/**
 *	@file qt_we_image.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief QT Image Class
 *
 *	@see WEToolkitImage
 *
 *	In this file the image for the QT library is defined
 */
#ifndef QT_WE_IMAGE_H_
#define QT_WE_IMAGE_H_

#include <QQuickItem>
#include <QtSvg/QSvgRenderer>

#include "../include/toolkit_we_image.h"

namespace WE
{
/**
 *	@class QTWEImage
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@brief The QTWEImage class specifies an image to be drawn with Qt.
 *	This class is derived from the two basic classes:
 *	ToolkitWEImage - interface class for the image
 *	QPixmap - Qt lib class used for showing images ... which meanwhile became QObject
 *
 *	@namespace WE
 *
*/
class QTWEImage : public ToolkitWEImage, public QObject
{
public:
	/*--------------------------------------------------------*/
	/**
	 *	@fn QTWEImage()
	 *
	 *	@brief Standard constructor of QTWEImage.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *
	 */
	/*--------------------------------------------------------*/
	QTWEImage();


	/*--------------------------------------------------------*/
	/**
	 *	@fn ~QTWEImage()
	 *
	 *	@brief Standard destructor of QTWEImage
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 */
	/*--------------------------------------------------------*/
	virtual ~QTWEImage();


	/**
	 *	@fn createImage(std::string p_strFile)
	 *
	 *	@brief creates a QPixmap object
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_strFile filename
	 *
	 *	@returns WE_SUCCES or WE_FILE_ERROR
	 *
	 *	if that function fails the filename is invalid
	 *
	 *	@Warning: This follows Markus' special "init create release reset" design pattern which is described in a WE wiki article.
	 *	@Warning: This seems to include that it is forbidden to call this method twice without calling another in between.
	 *	@Warning: Tried to fix that - which may have side effects.
	 */
	int createImage(const std::string & p_strFile) override;


	/**
	 *	@fn createImage(std::string p_strFile, std::uint8_t p_R, std::uint8_t p_G, std::uint8_t p_B)
	 *
	 *	@brief creates a QPixmap object
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_strFile file
	 *	@param[in] p_R transparency
	 *	@param[in] p_G transparency
	 *	@param[in] p_B transparency
	 *
	 *	@returns WE_SUCCES or WE_FILE_ERROR
	 *
	 *	if that function fails the filename is invalid
	 *
	 *	creates image and makes it transparent with the set RGB values. Transparency
	 *	by color only works with no alpha channel available. Alpha channels are not
	 *	overwritten.
	 *
	 *	@Warning: This follows Markus' special "init create release reset" design pattern which is described in a WE wiki article.
	 *	@Warning: This seems to include that it is forbidden to call this method twice without calling another in between.
	 *	@Warning: Tried to fix that - which may have side effects.
	 *	@Warning: This is costly if setting the size later as it will create a copy of the image and transparency mask that later on would have to be replaced.
	 */
	int createImage(const std::string & p_strFile, std::uint8_t p_R, std::uint8_t p_G, std::uint8_t p_B);


	/**
	 *	@fn createImage(const QPixmap *p_pixmap)
	 *
	 *	@brief creates an image with the given pixmap
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] p_pixmap the QPixmap
	 *
	 *	@returns WE_error
	 */
	int createImage(QPixmap *p_pixmap) override;


	/**
	 *	@fn int createImage(QSGTexture *p_texture, WETextureCache::Data *p_cacheData=nullptr)
	 *
	 *	@brief creates an image with the given texture
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] p_texture the QSGTexture
	 *	@param[in] p_cacheData optional pointer to the WETextureCache::Data object of the texture
	 *
	 *	@returns WE_error
	 */
	int createImage(QSGTexture *p_texture, WETextureCache::Data *p_cacheData=nullptr) override;


	/**
	 *	@fn QSGTexture * getTexture()
	 *
	 *	@brief Gets the texture of this image or creates a new texture if it needs an update and createIfNeeded is true.
	 *
	 *	@warning: This method should only be called from the Qt quick scene graph rendering thread.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param createIfNeeded create a texture of this image if there is no texture or the texture is invalid
	 *
	 *	@see void releaseTexture()
	 *
	 *	@returns the texture
	 */
	QSGTexture * getTexture(bool p_createIfNeeded=true) override;


	/**
	 *	@fn QSvgRenderer * getSVGRenderer();
	 *
	 *	@brief Gets the QSvgRenderer which is needed to render the image if it is a SVG image.
	 *	       Call isSVG to check if the image is a SVG before calling this method.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@see bool isSVG(void) const
	 *
	 *	@returns the QSvgRenderer instance
	 */
	QSvgRenderer * getSVGRenderer();


	/**
	 *	@fn void releaseTexture()
	 *
	 *	@brief Releases the texture created by createTexture()
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@see QSGTexture * getTexture(bool createIfNeeded=true)
	 */
	void releaseTexture() override;


	/**
	 *	@fn void invalidateTexture()
	 *
	 *	@brief Invalidates the texture of this image
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@see bool textureNeedsUpdate()
	 */
	void invalidateTexture() override;


	/**
	 *	@fn bool textureNeedsUpdate()
	 *
	 *	@brief Returns true if the texture of this image needs an update.
	 *	       This is the case if the texture was never created or invalidateTexture was called before.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@see void invalidateTexture()
	 *
	 *	@returns true or false
	 */
	bool textureNeedsUpdate() const override;


	/**
	 *	@fn void setCacheData(WETextureCache::Data *p_cacheData)
	 *
	 *	@brief Setter for the WETextureCache::Data object of this image
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] p_cacheData pointer to the WETextureCache::Data object of the texture of this image
	 */
	void setCacheData(WETextureCache::Data *p_cacheData) override;


	/**
	 * 	@brief This should be a static method used to store images e.g. from an ISO pool to file system.
	 *
	 *	This mainly exists to store images coming from an ISO-VT object pool with the original
	 *	references to the color table.
	 *	Refer to ISO-VT pool parser to see how the different formats defined in ISO 11783-6 are
	 *	converted into the common format expected here.
	 *	It is not really static because we cannot define it as 'static' and 'virtual' in the interface.
	 *
	 *	@param indexed8ImageData pointer to the raw image data - one byte per pixel, new lines start at next 32 bit boundary
	 *	@param imageWidth	width in pixels
	 *	@param imageHeight	height in pixels
	 *	@param colorTable	e.g. obtained from WEApp::ColorTable::getAllColors()
	 *	@param imageFileFormat	image file format (e,g; PNG, JPG etc)
	 *	@param imageFilePath	absolute path of the image file which shall be created
	 *	@returns		an error code - currently always WE_SUCCESS
	 *
	 *	@see WEApp::PoolParser::saveRawDataAsImageFile()
	 *	@see WEApp::ColorTable::getAllColors()
	 */
	int saveRawDataAsImageFile(
			unsigned char* p_indexed8ImageData,
			unsigned int p_imageWidth,
			unsigned int p_imageHeight,
			std::vector<unsigned int> & p_colorTable,
			const std::string & p_imageFileFormat,
			const std::string & p_imageFilePath) const override;


	/**
	 *	@fn setSize(int p_iWidth, int p_iHeight, bool p_resize = true)
	 *
	 *	@brief set the size of the image
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_iWidth width of the image
	 *	@param[in] p_iHeight height of the image
	 *	@param[in] p_resize resize the image if width and height are not the same
	 *
	 *	Nobody knows what the "p_resize" parameter should be good for - throughout all of our
	 *	code there isn't a single call for "setSize()" with more than two parameters.
	 *
	 */
	void setSize(int p_iWidth, int p_iHeight, bool p_resize = true) override;


	/**
	 *	@fn release()
	 *
	 *	@brief releases memory
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 */
	void release() override;


	/**
	 *	@fn getWidth()
	 *
	 *	@brief get image width
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns width of image
	 *
	 */
	int getWidth() const override;


	/**
	 *	@fn getHeight()
	 *
	 *	@brief get image height
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns height of image
	 *
	 */
	int getHeight() const override;


	/**
	 *	@fn getRootObject()
	 *
	 *	@brief gets a pointer to the object of the underlying toolkit
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns pointer to the QPixmap, because of Qt
	 *
	 */
	void* getRootObject() const override;


	/**
	 *	@fn int createTransparent(bool p_bTransparent)
	 *
	 *	@brief Creates a transparent image by color set with alternative createTransparent(), or switch it off
	 *
	 *	@param[in] p_bTransparent create = true or false for reset.
	 *
	 *	@returns WE_SUCCESS or failure
	 *
	 *	Makes image transparent at all pixels of transparency color.
	 *	Main intention here is to switch transparency off or to re-apply the same
	 *	transparency color as before.
	 *
	 *	@see int createTransparent(bool p_bTransparent, WEColor & p_transparencyColor)
	 */
	// this is an additional interface which isn't needed at the moment it seems
	//	int createTransparent(bool p_bTransparent);


	/**
	 *	@fn int createTransparent(bool p_bTransparent, WEColor & p_transparencyColor)
	 *
	 *	@brief Makes image transparent at all pixels of color p_transparencyColor.
	 *
	 *	@param[in] p_bTransparent create = true or false for reset.
	 *	@param[in] p_transparencyColor Color to create mask if p_bTransparent is true.
	 *
	 *	@returns WE_SUCCESS or failure
	 *
	 *	Makes image transparent at all pixels of color p_transparencyColor. Alpha channels of the image are not
	 *	overwritten. Alpha value of p_transparencyColor is ignored.
	 *
	 *	Only one of the transparency methods can be applied at any point of time
	 *	(i.e. calling one reverts effect of the one called before).
	 *
	 *	@see int setTransparent(bool, int)
 	 */
	int createTransparent(bool p_bTransparent, const WEColor & p_transparencyColor) override;


private:
	/**
	 *	@fn int createTransparent(bool p_bTransparent, WEColor & p_transparencyColor)
	 *
	 *	@brief Makes image transparent at all pixels of color p_transparencyColor.
	 *
	 *	@param[in] p_bTransparent create = true or false for reset.
	 *	@param[in] p_transparencyColor Color to create mask if p_bTransparent is true.
	 *	@param[in] p_width < 0: keep size, else: Create image of new size.
	 *	@param[in] p_height < 0: keep size, else: Create image of new size.
	 *
	 *	@returns WE_SUCCESS or failure
	 *
	 *	Makes image transparent at all pixels of color p_transparencyColor. Alpha channels of the image are not
	 *	overwritten. Alpha value of p_transparencyColor is ignored.
	 *
	 *	Only one of the transparency methods can be applied at any point of time
	 *	(i.e. calling one reverts effect of the one called before).
	 *
	 *	@see int setTransparent(bool, int)
 	 */
	int createTransparent(bool p_bTransparent, const WEColor & p_transparencyColor, int p_width, int p_height);


	/**
	 *	@fn int createTexture()
	 *
	 *	@brief Creates a texture of the internal pixmap to draw the image with the Qt quick scene graph.
	 *
	 *	@Warning: This method should only be called from the Qt quick scene graph rendering thread.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@returns WE_SUCCESS or failure
	 */
	int createTexture();


public:
	/**
	 *	@fn int setTransparent(bool p_bTransparent, int p_colorIndex)
	 *
	 *	@brief Makes image transparent at all pixels of color index.
	 *
	 *	@param[in] p_bTransparent create = true or false for reset.
	 *	@param[in] p_colorIndex Color index.
	 *
	 *	@returns WE_SUCCESS or failure
	 *
	 *	This method is required for ISO-VT, only one of the transparency
	 *	methods can be applied at any point of time (i.e. calling one
	 *	reverts effect of the one called before).
	 *
	 *	@see int createTransparent(bool, WEColor &)
	 */
	int setTransparent(bool p_bTransparent, int p_colorIndex) override;


private:
	/**
	 *	@fn int setTransparent(bool p_bTransparent, int p_colorIndex)
	 *
	 *	@brief Makes image transparent at all pixels of color index.
	 *
	 *	@param[in] p_bTransparent create = true or false for reset.
	 *	@param[in] p_colorIndex Color index.
	 *	@param[in] p_width < 0: keep size, else: Create image of new size.
	 *	@param[in] p_height < 0: keep size, else: Create image of new size.
	 *
	 *	@returns WE_SUCCESS or failure
	 *
	 *	This method is required for ISO-VT, only one of the transparency
	 *	methods can be applied at any point of time (i.e. calling one
	 *	reverts effect of the one called before).
	 *
	 *	@see int createTransparent(bool, WEColor &)
	 */
	int setTransparent(bool p_bTransparent, int p_colorIndex, int p_width, int p_height);


public:
	/**
	 *	@fn hasAlphaChannel()
	 *
	 *	@brief returns if an Alpha channel is available or not
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns true or false
	 *
	 */
	bool hasAlphaChannel() const override;


	/**
	 *	@fn isTransparent()
	 *
	 *	@brief Returns true if the image has transparency added by transparency color.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns true or false
	 *
	 */
	bool isTransparent() const override;


	// Seems this came in with SVG support ... a little documentation would be nice anyways
	// Has to be public because WEImageList calls it
	void setWidget(WEWidget * p_widget) override;	//!< Seems this is needed to scale the SVG to the widget's size and rotate them around their center


	// Seems this came in with SVG support ... a little documentation would be nice anyways
	// public because WEImage calls it
	bool isSVG(void) const override;
	bool isGIF(void) const override;


private:

	void init();
	QImage loadQImageFromFile() const;

	QBitmap * bitmapMask;	//!< Pointer to QBitmap for color transparency
	QPixmap * pixmap;		//!< Pointer to (potentially modified) image to display
	QPixmap * orgPixmap;	//!< Pointer to original image
	QImage originalImage;		//!< QImage holding the original image data including a valid color map for indexed images
	QSvgRenderer * svgRenderer;	//!< Pointer to QSvgRenderer which is needed to render the image if it is a svg image (might be nullptr, use getSVGRenderer)
	QSGTexture * texture;		//!< Pointer to the texture of this image (should hold the same image content as pixmap)
	QSGTexture * originalUserTexture;//!<Pointer to the original texture set by createImage(QSGTexture*)
	bool textureIsValid;		//!< true if the texture is valid

	WEColor transparentColor;	//!< Color used to create transparency mask in "createTransparent"
	int transparentColorIndex;	//!< < 0: WEColor used for transparency, else: color index used for transparency
	bool transparentFlag;		//!< true if we added transparency to the image - independent of alpha channel
	QString filePath;		//!< The file path to this image as QString
	bool userPixmap;		//!< true if the pixmap was provided by the user via createImage(QPixmap*)
	WETextureCache::Data *cacheData;//!< Pointer to the cache data of this image in the texture cache

	bool gif;			//!< true if this image is a (animated?) GIF image
	bool svg;			//!< true if this image is a SVG image

	// Seems this came in with SVG support ... a little documentation would be nice anyways
	QQuickItem * widget;	//!< Seems this is needed to scale the SVG to the widget's size and rotate them around their center
};
} // end of namespace WE
#endif /* QT_WE_IMAGE_H_ */
