/**
 *	@file we_datatypes_gui.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Definition of the general data types that can be used with WE Window and widgets.
 *
 *	@see we_datatypes_gui.cpp
 *
 *	This file contains the definition of the data types used with the WE Windows and Widgets.
 */


#ifndef WE_GUI_DATATYPES_H_
#define WE_GUI_DATATYPES_H_

#include <boost/serialization/export.hpp>
#include <boost/serialization/access.hpp>

namespace WE
{
/**
 *	@struct WEPoint
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief Represents a point.
 *
 *	@namespace WE
 *
 *	It represents a point specified by an X Coordinate and a Y Coordinate.
 */
struct WEPoint
{
	int x{0};	//!< X Coordinate of the point.
	int y{0};	//!< Y Coordinate of the point.
	WEPoint () = default;
	WEPoint (int p_x, int p_y)
	:x{p_x},
	 y{p_y}
	{
		//nothing more to do
	}
	virtual ~WEPoint() = default;
private:
	//SERIALIZATION
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int );
};

// Enums
// ------------------------------------------------
// ------------------------------------------------
enum WEDisplayVideo
{
	VIDEO_NONE = 0,
	VIDEO_UNMIRRORED = 1,
	VIDEO_HORIZONTALLY_MIRRORED = 2,
	VIDEO_VERTICALLY_MIRRORED = 3,
	VIDEO_BOTH = 4
};


enum WEScalingFactor
{
	SCALING_FACTOR_WHOLE_WIDGET = 0,
	SCALING_FACTOR_WHOLE_IMAGE
};

enum WEBorderState
{
	BORDER_STATE_UNDEFINED = 0,
	OPEN_FOR_INPUT,			//!< open for input
	ENABLED_WITH_FOCUS,		//!< enable with focus, encoder push would open it for input
	ENABLED_WITHOUT_FOCUS,		//!< the frame could be selected, but is not selected
	ENABLED_OUTPUT,			//!< output object showing valid value
	DISABLED,			//!< Passive/showing invalid value
	BORDERPROFILE_STATE_COUNT
};


// WARNING: Current implementation relies on these constants being identical with those of the underlying graphical toolkit
typedef enum
{
	ALIGN_UNDEFINED		= 0x0000,

	ALIGN_LEFT		= 0x0001,
	ALIGN_RIGHT		= 0x0002,
	ALIGN_H_CENTER		= 0x0004,
	ALIGN_H_MASK		= 0x000f,

	ALIGN_TOP		= 0x0020,
	ALIGN_BOTTOM		= 0x0040,
	ALIGN_V_CENTER		= 0x0080,
	ALIGN_V_MASK 		= 0x00f0,

	ALIGN_CENTER		= ALIGN_V_CENTER | ALIGN_H_CENTER,
	ALIGN_TOP_CENTER	= ALIGN_TOP | ALIGN_H_CENTER,
	ALIGN_BOTTOM_CENTER	= ALIGN_BOTTOM | ALIGN_H_CENTER,

	ALIGN_TOP_LEFT		= ALIGN_LEFT | ALIGN_TOP,
	ALIGN_CENTER_LEFT	= ALIGN_LEFT | ALIGN_V_CENTER,
	ALIGN_BOTTOM_LEFT	= ALIGN_LEFT | ALIGN_BOTTOM,

	ALIGN_TOP_RIGHT		= ALIGN_RIGHT | ALIGN_TOP,
	ALIGN_CENTER_RIGHT	= ALIGN_RIGHT | ALIGN_V_CENTER,
	ALIGN_BOTTOM_RIGHT	= ALIGN_RIGHT | ALIGN_BOTTOM,

	ALIGN_DEFAULT		= ALIGN_TOP_LEFT
}WEAlignment;

typedef enum
{
	NO_DEFLECTION = 0,
	CLOCKWISE,
	ANTICLOCKWISE
}WEDeflectionDirectionType;

typedef enum
{
	NO_FORMAT = 0,
	ROUND,
	SQUARE
}WETickPointFormatType;

typedef enum
{
	ELLIPSE_CLOSED		= 0x00,
	ELLIPSE_OPEN		= 0x01,
	ELLIPSE_CLOSED_SEGMENT	= 0x02,
	ELLIPSE_CLOSED_SECTION	= 0x03
}WEEllipseType;

typedef enum
{
	ISO_8859_1 = 0,
	ISO_8859_15 = 1,
	ISO_8859_2 = 2,
	RESERVED_3 = 3,
	ISO_8859_4 = 4,
	ISO_8859_5 = 5,
	RESERVED_6 = 6,
	ISO_8859_7 = 7,
	UTF_8 = 8,
	RESERVED,
	PROPRIETARTY
}WEEncodeType;
//While adding any value here please also update the WEGUIUtils::initEnumToNameMapping()

typedef enum
{
	LINE_DIRECTION_TOP_LEFT_TO_BOTTOM_RIGHT	= 0,
	LINE_DIRECTION_BOTTOM_LEFT_TO_TOP_RIGHT	= 1
}LineDirection;

typedef enum
{
	NO_AXIS = 0,
	AXIS_HORIZONTAL,	//!< horizontal mode
	AXIS_VERTICAL	//!< vertical mode
}WEAxisDirectionType;

typedef enum
{
	NO_BARGRAPH_DIRECTION = 0,
	BARGRAPH_POSITIVE,
	BARGRAPH_NEGATIVE
}WEDirectionType;

typedef enum
{
	NO_MODE = 0,
	NORMAL,	//!< normal mode
	EDIT	//!< edit mode
}WEListModeType;

enum WENumberDisplayFormat
{
	NUMBER_FORMAT_DECIMAL,
	NUMBER_FORMAT_EXPONENTIAL
};

typedef enum
{
	POLYGON_TYPE_CONVEX 	= 0x0,
	POLYGON_TYPE_NON_CONVEX	= 0x1,
	POLYGON_TYPE_COMPLEX	= 0x2,
	POLYGON_TYPE_OPEN	= 0x3
}WEPolygonType;
}	//End of namespace WE

BOOST_CLASS_EXPORT_KEY(WE::WEPoint)

#endif /* WE_GUI_DATATYPES_H_ */
