/**
 *	@file qt_we_multimedia.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Class definitions of MyVideoWidget and QtWEMultimediaSource
 */


#ifndef GSTREAMER_WE_MULTIMEDIA_H_
#define GSTREAMER_WE_MULTIMEDIA_H_

//#include <QKeyEvent>
#include <QQuickWidget>
#include <QObject>
#include <QQuickItem>
#include <tr1/unordered_map>
#include <gst/gst.h>

namespace WE {

class WEWidget;
class WEImageWidget;
class WEContainer;
class WEEvent;
class WEEventListener;
class WEMultimediaPlayer;
class WEEventQueue;
class WESignalListener;

class GStreamerMultimediaSource : public QObject
{
	Q_OBJECT
public:
	/**
	 * Constructor. Initialize members with (hopefully) sane default values
	 */
	GStreamerMultimediaSource();

	/**
	 * Destructor. Frees memory and disconnects video and audio pathes.
	 * Does not delete the widgets! (WEWidget is deleted by standard WE ways and the phonon video
	 * widget SHOULD be deleted by Qt's automatic freeing because it is a child of another
	 * QWidget that gets destroyed)
	 */
	virtual ~GStreamerMultimediaSource();

	void sendEvent(void);

	void addListener(WESignalListener * p_listener);
	void removeListener(WESignalListener * p_listener);

	int createAlbumArt(void);

	void createQuickItems(void);

	int state;		//!< current state (see enum MediaStates)
	int oldState;		//!< old state that was active before current state
	int loopCount;		//!< How often this file has repeated already
	int maxLoops;		//!< How often shall this media file be repeated (-1 for infinite loop)
	int category;		//!< Either multimedia, alarm, event, etc. see enum MediaCategory in we_multimedia_player.h
	unsigned int latestPosition;	//!< latest position in the stream in seconds
	unsigned int positionQueryTime;	//!< a tick in the future at which we shall query the current position again
	unsigned int playbackGap;	//!< playback gap in milliseconds between each loop
	unsigned int nextPlaybackLoopTime;	//!< tick in the future when playback shall be started again after a loop with playback gap
	gint64 duration;	//!< Duration of the whole stream in 1/1000000000s
	bool isVideo;		//!< flag to tell if source is a video file
	bool isMpegAudio;	//!< we only want to extract album art from mpeg audio files
	bool doRepeat;		//!< flag that is set after a loop with playback gap
	bool metadataChanged;	//!< Flag that is set when the title was extracted from a tag message
	bool manualStopOrPause;	//!< Flag to indicate that playback was stopped manually so that we don't repeat after EOS
	QQuickItem * videoWidget;	//!< The QQuickItem which is set as "widget" property to the qmlglsink
	QQuickItem * waitItem;
	QQuickWidget * videoParent;
	GstElement * play;	//!< The gstreamer pipeline (uridecodebin => (automatically chosen decoders) => glupload => glcolorconvert => qmlglsink)
	GstElement * sink;	//!< The video sink
	GstElement * glupload;
	GstElement * glcolorconvert;
	GstElement * src;	//!< uridecodebin element to automatically choose correct decoders for the given uri (e.g. file path)
	GstElement * audio;	//!< audio sink
	GstElement * volumeElement;	//!< volume element to control volume of the pipeline
	GstBus * bus;	//!< gstreamer message bus
	WEImageWidget * albumArt;	//!< If this media is audio, it may have album/cover art stored in it's meta data. If that's the case we store the imgae in here
	unsigned char volume;	//!< The intended volume. What is really set is based on volume variables!
	std::string path;	//!< The identifier of the media file (it's file path)
	std::string title;	//!< If source contains a title in it's meta data, it will be stored here after receiving the appropriate "tag" message
	std::string errorMessage;	//!< Error message as reported by gstreamer
	std::string debugInformation;	//!< Debug information as reported by gstreamer

	static WEEventQueue * eventQueue;

protected:
	std::set<WESignalListener *> eventListeners;	//!< A set of event listeners that want to get events by this source
private slots:
	void doShowVideoItem(bool p_show);
	void videoWidgetDestroyed(QObject * p_obj);
signals:
	void showVideoItem(bool p_show);
};

typedef std::unordered_map<std::string, GStreamerMultimediaSource*> sourceMap;	//!< Typedef of the hash map that contains all source files identified by it's file path
}  // namespace WE


#endif /* GSTREAMER_WE_MULTIMEDIA_H_ */
