/**
 *	@file qt_we_multimedia.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Class definitions of MyVideoWidget and QtWEMultimediaSource
 */


#ifndef QT_WE_MULTIMEDIA_H_
#define QT_WE_MULTIMEDIA_H_

#include <QtGlobal>

#include <QObject>
#include <QImage>
#include <QMouseEvent>
#include <QLabel>
#include <QUrl>
#include <unordered_map>

#include <QAudioOutput>
#include <QMediaPlayer>
#include <QVideoWidget>

namespace WE {

class WEWidget;
class WEMultimediaPlayer;
class WEEventQueue;
class WESignalListener;

/**
 * MyVideoWidget subclasses QVideoWidget because we want to be able to switch to
 * fullscreen when someone touches the widget.
 * Maybe we should make this configurable in future ;)
 */
class MyVideoWidget : public QVideoWidget
{
	Q_OBJECT
public:
	/**
	 * Constructor, calls Phonon::VideoWidget constructor and initializes members
	 * @param parent QWidget parent (in our case it is WEWidget's toolkit widget)
	 */
	explicit MyVideoWidget( QWidget * parent = 0 );

	/**
	 * Nothing to do in the destructor
	 */
	virtual ~MyVideoWidget();

	void enterFullscreen(bool enter);

signals:
	void fullscreenChanged(bool);

protected:
	/**
	 * Overwritten mouseReleaseEvent function. We switch to or back from fullscreen mode
	 * when receiving a mouse release event. When entering fullscreen mode, we store width, height and parent
	 * in our members and then set our own parent to nullptr and resize ourself to display width and height.
	 * When leaving fullscreen mode, we set our original size again and also set our parent again.
	 * Event will also be forwarded to  Phonon::VideoWidget::mouseReleaseEvent()
	 * @param event The mouse release event. We do not care for it's values.
	 */
	void mouseReleaseEvent ( QMouseEvent * event ) override;
	void keyReleaseEvent(QKeyEvent* p_qtKeyEvent) override;

	bool fullscreenMode;	//!< Flag that shows if this widget is in fullscreen mode or not
	int widthBeforeFullscreen;	//!< Store the original width of the widget in here before switching to fullscreen
	int heightBeforeFullscreen;	//!< Store the original height of the widget in here before switching to fullscreen
	QWidget * parentBeforeFullscreen;	//!< The parent of this widget before switching to fullscreen
};


class AudioOutputAbstraction : public QAudioOutput
{

};

/**
 * QtWEMultimediaSource represents one media file with all information needed by the multimedia player
 * who uses it. It's a QObject subclass because it receives signals from the phonon media object
 * about state and playback time changes (these are then forwarded as WESignal to a listener) and also a
 * signal when the phonon video widget is destroyed so that it can clean up.
 */
class QtWEMultimediaSource : public QObject
{
	Q_OBJECT
public:
	/**
	 * Constructor. Initialize members with (hopefully) sane default values
	 * @param parent The QObject parent, can be omitted. Not really interesting for our application
	 */
	explicit QtWEMultimediaSource(WEMultimediaPlayer * p_player, QObject * parent = nullptr);

	/**
	 * Destructor. Frees memory and disconnects video and audio pathes.
	 * Does not delete the widgets! (WEWidget is deleted by standard WE ways and the phonon video
	 * widget SHOULD be deleted by Qt's automatic freeing because it is a child of another
	 * QWidget that gets destroyed)
	 */
	virtual ~QtWEMultimediaSource();

	/**
	 * If the media file is an audio file, this function tries to find the album/cover art of the
	 * audio file in it's meta data. It stores the image as a QImage.
	 * @return WE_SUCCESS if album art was successfully found and QImage was created.
	 * 	WE_GENERAL_ERROR if creating the image failed or no album art was found in meta data
	 */
	int createAlbumArt();

	/**
	 *	Send current state?!
	 */
	void fireEvent(void);

	/**
	 *	Send new state?!
	 */
	void fireEvent(qint64 p_duration, qint64 p_position);

	static WEEventQueue * eventQueue;

public slots:
	/**
	 * Slot for handling the finished() signal of the media object.
	 * If looping is configured, it will start to playback the media again.
	 */
	void finished();

	/**
	 * Slot that is called by the media object's tick() signal while playback.
	 * This is configured to happen every 1000ms.
	 * WESignal for listener is sent out when this slot is called.
	 * @param time Current time of the media object in ms.
	 */
	void currentPositionChanged(qint64 time);

	/**
	 * Slot that is called by the media object if it's total time changes.
	 * Should actually not happen since we do not stream anything and if we set
	 * another source file, we will create a new media object.
	 * WESignal for listener is sent out when this slot is called.
	 * @param time Total time of the media object in ms.
	 */
	void totalTimeChanged(qint64 time);

	/**
	 * This slot is called when the satte of the media object changes. States can be stop,
	 * play, pause, loading, buffering, etc. See Phonon::State documentation for a complete list.
	 * WESignal for listener is sent out when this slot is called.
	 * @param newstate The new state
	 */
	void stateChanged(QMediaPlayer::State newstate);

	void mediaStatusChanged(QMediaPlayer::MediaStatus status);


	/**
	 * This slot is called when the Phonon video widget is being deleted (right before it is really gone).
	 * Gives us a chance to disconnect pathes and set our widget pointers to nullptr.
	 * @param obj Pointer to the object that is destroyed.
	 */
	void widgetDestroyed(QObject * obj);

	/**
	 * Use this function to try to extract interesting meta data and store them as member so that they can be quickly accessed.
	 * Currently only extracting the title is supported. If extracting the title failed, the title member has an empty
	 * string after calling this method.
	 */
	void updateMetadataFromFile(void);

	/**
	 * This slot will be called by a timer when looping is configured for this file and a playback gap was configured.
	 */
	void playAndIncrement(void);

	/**
	 * Same as playAndIncrement() but without incrementing the loop counter (when configured for endless looping)
	 */
	void justPlay(void);

	void enterFullscreen(bool enter);

	void fullscreenChanged(bool enabled);

	void addListener(WESignalListener * p_listener);
	void removeListener(WESignalListener * p_listener);

	void displayErrorMessage(QMediaPlayer::Error err);

	int createVideoSettings(bool videoAvailable = true);

private:
	friend class WEMultimediaPlayer;	//!< The multimedia player may access private members
	std::string path;	//!< The identifier of the media file (it's file path)
	unsigned char volume;	//!< The intended volume. What is really set is based on volume variables!
	int loopCount;	//!< How often this file has repeated already
	int maxLoops;	//!< How often shall this media file be repeated (-1 for infinite loop)
	unsigned int playbackGap;	//!< playback gap in milliseconds between each loop
	int state;	//!< The state of this media file as reported by phonon
	int oldState;
	int category;	//!< Either multimedia, alarm, event, etc. see enum MediaCategory in we_multimedia_player.h
	WEWidget * videoWidget;	//!< The WEWidget which's QWidget acts as a parent for our own video widget

	bool isVideo;

	QMediaPlayer * mediaObject;	//!< The phonon media object that is used for controlling playback
	AudioOutputAbstraction * audioOutput;	//!< The phonon audio output that is connected to a media object for audio output

	//Phonon::Path videoPath;	//!< Phonon creates a path between media object and video widget. We save this so that we can disconnect the path again when done
	//Phonon::Path audioPath;	//!< Phonon creates a path between media object and audio output. We save this so that we can disconnect the path again when done
	MyVideoWidget * phononVideoWidget;	//!< Our subclass of a phonon VideoWidget used for video rendering
	std::set< WESignalListener * > eventListeners;	//!< A set of event listeners that want to get events by this source
	QImage * albumArt;	//!< If this media is audio, it may have album/cover art stored in it's meta data. If that's the case we store the imgae in here
	std::string title;	//!< If title could be extracted by updateMetadataFromFile() this contains the title. If it failed: empty string.
	QLabel * albumArtLabel;	//!< Label that renders the album art image if available
	WEMultimediaPlayer * thePlayer;
};

typedef std::unordered_map<std::string, QtWEMultimediaSource*> sourceMap;	//!< Typedef of the hash map that contains all source files identified by it's file path

}  // namespace WE


#endif /* QT_WE_MULTIMEDIA_H_ */
