/**
 *	@file we_geometry_node.h
 *
 *	@author preths@topcon.com
 *	@author Patrick Reths
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2020
 *
 */

#ifndef WE_GEOMETRY_NODE_H
#define WE_GEOMETRY_NODE_H

#include <iostream>
#include <QSGTextureMaterial>

#include "we_node_image_support.h"

namespace WE
{

class WEFlatColorMaterial;
class WEVertexColorMaterial;

/**
 *
 * This class can be used to draw geometry with the Qt quick scene graph.
 *
 */
class WEGeometryNode : virtual public WEBasicGeometryNode
{

public:

	enum MaterialType {
		FlatColorMaterial,
		VertexColorMaterial,
		TextureMaterial
	};

	explicit WEGeometryNode(unsigned int p_drawingMode, int p_vertexCount, int p_indexCount=0, MaterialType p_materialType=WEGeometryNode::defaultColorMaterialType);

	virtual ~WEGeometryNode();

	void setDrawingMode(unsigned int p_drawingMode);
	void setSize(int p_vertexCount, int p_indexCount=0);
	void setVertex(int p_index, float p_x, float p_y);
	void setIndex(int p_index, quint16 p_vertexIndex);
	void setColoredVertex(int p_index, float p_x, float p_y, WEColor p_color);
	void useInternalMaterial();
	void setColor(const WEColor& p_color, bool p_applyToAll=true);
	void setImage(WEImage *p_image, const QSize& p_geometrySize);
	void setTexture(QSGTexture *p_texture);
	void apply();

	unsigned int getDrawingMode() const
	{
		return this->geometry->drawingMode();
	}

	int getVertexCount() const
	{
		return this->geometry->vertexCount();
	}

	int getIndexCount() const
	{
		return this->geometry->indexCount();
	}

	MaterialType getMaterialType() const
	{
		return this->materialType;
	}

	QSGGeometry::Point2D getVertex(int p_index) const
	{
		QSGGeometry::Point2D ret{};

		switch (this->materialType)
		{
		case FlatColorMaterial:
		{
			QSGGeometry::Point2D *p = this->geometry->vertexDataAsPoint2D();
			if (nullptr != p)
			{
				ret = p[p_index];
			}
			else
			{
				ret.set(0, 0);
			}
			break;
		}
		case VertexColorMaterial:
		{
			QSGGeometry::ColoredPoint2D *p = this->geometry->vertexDataAsColoredPoint2D();
			if (nullptr != p)
			{
				ret.set(p[p_index].x, p[p_index].y);
			}
			else
			{
				ret.set(0, 0);
			}
			break;
		}
		case TextureMaterial:
		{
			QSGGeometry::TexturedPoint2D *p = this->geometry->vertexDataAsTexturedPoint2D();
			if (nullptr != p)
			{
				ret.set(p[p_index].x, p[p_index].y);
			}
			else
			{
				ret.set(0, 0);
			}
			break;
		}
		default:
			std::cerr << "WEGeometryNode::getVertex: MaterialType is invalid" << std::endl;
			break;
		}

		applyLinePointAlignment(ret.x, ret.y, false);

		return ret;
	}

	float getVertexX(int p_index) const
	{
		return getVertex(p_index).x;
	}

	float getVertexY(int p_index) const
	{
		return getVertex(p_index).y;
	}

	quint16 getIndex(int p_index) const
	{
		quint16 vertexIndex = 0;
		quint16 *i = this->geometry->indexDataAsUShort();

		if (nullptr != i)
		{
			vertexIndex = i[p_index];
		}

		return vertexIndex;
	}

	WEColor getColor() const
	{
		return this->color;
	}

	QSGTexture * getTexture() const
	{
		return this->imageSupport.getTexture();
	}

	WEImage * getImage() const
	{
		return this->imageSupport.getImage();
	}


protected:

	typedef struct
	{
		unsigned char r, g, b, a;
	}Color;

	// Vertex colored geometry holds its color in each vertex instead in the material.
	// This allows the renderer to batch more nodes together, because they can be drawn with the same material.
	// Use QSG_RENDERER_DEBUG=render to debug node batching and
	// GALLIUM_HUD="fps,draw-calls" to show the number of OpenGL draw calls. (GALLIUM_HUD requires a gallium based graphics driver like etnaviv.)
	static const MaterialType defaultColorMaterialType;

	QSGGeometry * geometry;
	const MaterialType materialType;
	WEFlatColorMaterial * flatColorMaterial;
	WEVertexColorMaterial * vertexColorMaterial;
	WEColor color;
	WEGeometryNode::Color premultipliedColor;
	QSize geometrySize; // Used to calculate texture coordinates
	WENodeImageSupport imageSupport;

	void createInternalMaterial();
	void setColoredVertex(int p_index, float p_x, float p_y, const WEGeometryNode::Color& p_premultipliedColor);
	void applyLinePointAlignment(float& p_x, float& p_y, bool p_add) const;
	WEGeometryNode::Color premultiplyColor(const WEColor& p_color) const;
};
} //End of namespace WE
#endif // WE_GEOMETRY_NODE_H
