/**
 *	@file we_node_utils.h
 *
 *	@author preths@topcon.com
 *	@author Patrick Reths
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2020
 *
 */

#ifndef WE_NODE_UTILS_H
#define WE_NODE_UTILS_H

#include <vector>
//#include <QSGNode>
//#include <QSGTexture>
#include <QRunnable>
#include <QQuickWindow>
#include <QGuiApplication>

//#include "we_controller_node.h"
//#include "we_geometry_node.h"
//#include "../../widgets/include/we_widget.h"

namespace WE
{

class WEBaseNode;
class WEControllerNode;
class WEBasicGeometryNode;
class WEGeometryNode;

/**
 *
 * This namespace contains some utility functions which are used by the nodes.
 *
 */
namespace WENodeUtils
{


QWindow * getWindow();


QSGTexture * createTextureFromImage(QImage * p_image, bool p_isOpaque=false);


void releaseQSGObject(QObject * p_object, QQuickWindow * p_window=nullptr);
void releaseQSGObjects(const std::vector<QObject*>& p_objects, QQuickWindow * p_window=nullptr);


void drawNode(WEBaseNode *p_node, WEControllerNode *p_controller);
void drawNode(WEGeometryNode *p_node, WEControllerNode *p_controller);


void markDirtyMaterial(WEBasicGeometryNode *p_node);


unsigned int getNumberOfEllipseSegments(unsigned int p_radiusX, unsigned int p_radiusY);


template <class T>
void setMaterial(T *p_node, QSGMaterial *p_material, bool p_ownsMaterial)
{
	if ((nullptr != p_node) && (nullptr != p_material))
	{
		if (p_material != p_node->material())
		{
			p_node->T::setMaterial(p_material);

			if (p_ownsMaterial)
			{
				p_node->setFlag(QSGNode::OwnsMaterial);
			}
		}
	}
}


template <class T>
void releaseQSGObject(T *p_object, QQuickWindow *p_window)
{
	if (nullptr != p_object)
	{
		std::vector<T*> objects;
		objects.push_back(p_object);
		releaseQSGObjects(objects, p_window);
	}
}


template <class T>
void releaseQSGObjects(const std::vector<T*>& p_objects, QQuickWindow *p_window)
{
	class ReleaseObjectRunnable : public QRunnable
	{
		std::vector<T*> objects;

	public:

		explicit ReleaseObjectRunnable(const std::vector<T*>& p_objects) : QRunnable(),
		objects(p_objects)
		{
			setAutoDelete(true);
		}

		~ReleaseObjectRunnable() = default;

		void run() override
		{
			for (T *t: this->objects)
			{
				if (nullptr != t)
				{
					delete t;
					//std::cout << "*** WENodeUtils::releaseQSGObjects (runnable)" << std::endl;
				}
			}
		}
	};


	if (nullptr == p_window)
	{
		p_window = static_cast<QQuickWindow*>(getWindow());
	}

	if (nullptr != p_window && p_window->isVisible())
	{
		p_window->scheduleRenderJob(new ReleaseObjectRunnable(p_objects), QQuickWindow::AfterSynchronizingStage);
	}
	else
	{
		for (T *t: p_objects)
		{
			if (nullptr != t)
			{
				delete t;
				//std::cout << "*** WENodeUtils::releaseQSGObjects (direct)" << std::endl;
			}
		}
	}
}
} //End of namespace WENodeUtils
} //End of namespace WE
#endif // WE_NODE_UTILS_H
