/**
 *	@file we_painter.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file defines the WEPainter to be used for painting the widgets.
 *
 *	@see we_painter_factory.h, we_widget.h
 *
 *	This file defines the WEPainter which provides the various drawing primitives
 *	and other interfaces to configure the Painter.
 */


#ifndef WE_PAINTER_H_
#define WE_PAINTER_H_

//Included headers are used in the implementation,
// and in widgets which override the paint method.
#include <vector>
#include <QPolygon>
#include "../../include/we_datatypes_gui.h"
#include "../../common/include/we_rect.h"
#include "../../common/include/we_region.h"
#include "../../common/include/we_painter_path.h"
#include "../../font/include/we_font.h"


namespace WE
{
/**
 *	@enum WERenderHints
 *
 *	@brief RenderingHints to provide the flags to the Painter.
 */
typedef enum
{
	ANTIALIAS_NONE = 0x00,
	ANTIALIASING = 0x01,
	TXT_ANTIALIASING = 0x02,
	SMOOTH_PIXMAP_TRANSFORM = 0x04,
	HIGH_QUALITY_ANTIALIASING = 0x08,
	NON_COSMETIC_DEFAULT_PEN = 0x10
}WERenderHints;

class WEImage; //forward declaration so the class compiles, will be replaced
struct WEBrush;
struct WEPen;

/**
 *	@class WEPainter
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief The WEPainter class performs painting on widgets and windows. It provide various method to draw the primitives
 *	and configure the painter. It can be used to paint any object that uses WEWidget.
 *
 *
 *	@namespace WE
 *
 *	@see we_widget.h
 *	@see we_datatypes_gui.h
 *
 *	This is a pure abstract class which can be instantiated with WEPainterFactory.
 *	The WEPainter class performs painting on widgets and windows. It provides
 *	various method to draw the primitives and configure the painter. It can be
 *	used to paint any object that uses WEWidget. The core functionality
 *	of  WEPainter is drawing, but the class also provide several functions that
 *	allows you to customise WEPainter's settings and its rendering quality.
 *	The common use of WEPainter is inside a widget's paint event: Construct and
 *	then customise (e.g. set the pen or the brush) the painter. Then draw.
 *	For example:
 *	void SampleWidget::paint()
 *	{
 *		//Create a Painter using the WEPainterFactory
 *		WEPainter* WEPainterFactory::getInstance()->createWEPainter(this->weWidget);
 *		painter->setPen(somePen);
 *		painter->drawText(x,y,"WE");
 *	}
 *
 *	@warning Only one painter instance can be created for a widget.
 *	@warning Painter initialised with a null widget may cause problems.
 */
class WEPainter
{

public:
	/**
	 *	@fn ~WEPainter()
	 *
	 *	@brief Destructor for WEPainter.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 */
	virtual ~WEPainter() = default;

	/**
	 *	@fn setRenderHint(WERenderHints renderHint)
	 *
	 *	@brief Sets the specified render hint on the painter.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] renderHint WERenderHint to be set on the painter.
	 *
	 *	@see WERenderHints
	 *	@see void setRenderHints(WERenderHints renderHints)
	 *
	 *	Sets the specified rendering hint on the painter.
	 */
	virtual void setRenderHint(WERenderHints p_renderHint) =0;

	/**
	 *	@fn setRenderHints(WERenderHints renderHints)
	 *
	 *	@brief Sets the render hints off the painter to the specified render hints.
	 *	       This sets all specified render hints and clears all render hints which are not specified.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] renderHints WERenderHints to be set on the painter.
	 *
	 *	@see WERenderHints
	 *	@see void setRenderHint(WERenderHints renderHint)
	 */
	virtual void setRenderHints(WERenderHints p_renderHints) =0;

	/**
	 *	@fn getRenderHints()
	 *
	 *	@brief Returns the current render hint of the painter.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@return WERenderHint currently used by the painter.
	 *
	 *	@see WERenderHints
	 */
	virtual int getRenderHints() const =0;

	/**
	 *	@fn isRenderHintActive(WERenderHints renderHint)
	 *
	 *	@brief Checks if RenderHint is set.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@return true 	= render hint is set
	 *		false	= render hint is not set
	 *
	 *	@see WERenderHints
	 */
	virtual bool isRenderHintActive(WERenderHints p_renderHint) const =0;

	/**
	 *	@fn setClip(const WERect & rect)
	 *
	 *	@brief It is used to replace the current clip rectangle and sets
	 *	the clip region to the given rectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *  @param[in] rect WERect Rectangle Area to be clipped.
	 *
	 */
	virtual void setClip(const WERect & p_rect)=0;

	/**
	 *	@fn setClipRegion(const WERegion &p_region)
	 *
	 *	@brief It is used to replace the current clip region and sets
	 *	the clip region to the given rectangle.
	 *
	 *	@author tanmoy.deb@aricent.com
	 *	@author Tanmoy Deb
	 *
	 *	@param[in] p_region The region.area to be clipped.
	 *
	 */
	virtual void setClipRegion(const WERegion & p_region)=0;


	virtual void setClipPath(const WEPainterPath & p_painterPath) = 0;


	/**
	 *	@fn void drawPath(const WEPainterPath & p_path)
	 *
	 *	@brief Draw p_path using the current pen and brush.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_path Path to be drawn
	 *
	 *	@returns nothing
	 *
	 *
	 *	This method can be used to e.g. draw an arched bargraph or some other more
	 *	complex shapes.
	 *
	 */
	virtual void drawPath(const WEPainterPath & p_path)  = 0;


	/**
	 *	@fn save(void)
	 *
	 *	@brief Saves the current painter state. A save() must be followed by a
	 *	corresponding restore().
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@see restore(void)
	 *
	 *	Saves the current painter state. A save() must be followed by a
	 *	corresponding restore().
	 */
	virtual void save(void) =0;

	/**
	 *	@fn restore(void)
	 *
	 *	@brief Restores the current painter state (pops a saved state off the stack).
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@see save(void)
	 *
	 *	@warning A corresponding save() must have been performed before the restore.
	 *
	 *	Restores the current painter state (pops a saved state off the stack).
	 */
	virtual void restore(void) =0;

	/**
	 *	@fn drawPoint(int x,int y)
	 *
	 *	@brief Draws a single point at the given position using the current pen's color.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] x XCoordinate at which to draw the point.
	 *	@param[in] y YCoordinate at which to draw the point.
	 *
	 */
	virtual void drawPoint(int p_x, int p_y) =0;

	/**
	 *	@fn drawLine(int x1,int y1,int x2,int y2)
	 *
	 *	@brief Draws a line from (x1, y1) to (x2, y2) and sets the current pen
	 *	position to (x2, y2).
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] x1 X1 Coordinate of the line.
	 *	@param[in] y1 Y1 Coordinate of the line.
	 *	@param[in] x2 X2 Coordinate of the line.
	 *	@param[in] y2 Y2 Coordinate of the line.
	 *
	 */
	virtual void drawLine(int p_x1, int p_y1, int p_x2, int p_y2) =0;


	/**
	 * @fn void drawLine(float x1,float y1,float x2,float y2);
	 *
	 * @brief Use floating point coordinates to draw a line from (x1, y1) to (x2, y2) and sets the current pen
	 *	position to (x2, y2).
	 *
	 * @author ajay.bisht@aricent.com
	 * @author Ajay Bisht
	 *
	 * @param[in] x1 X1 Coordinate of the line.
	 * @param[in] y1 Y1 Coordinate of the line.
	 * @param[in] x2 X2 Coordinate of the line.
	 * @param[in] y2 Y2 Coordinate of the line.
	 *
	 */
	virtual void drawLine(float p_x1, float p_y1, float p_x2, float p_y2)  = 0;

	/**
	 *	@fn drawRect(int x1,int y1,int p_width,int p_height, float p_xRadius, float p_yRadius)
	 *
	 *	@brief Draws a Rectangle from upper left corner(p_x, p_y) with width = p_width and height = p_height
	 *	and sets the current pen position to (x2, y2).
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_x X Coordinate of the Rectangle.
	 *	@param[in] p_y Y Coordinate of the Rectangle.
	 *	@param[in] p_width Width of the Rectangle.
	 *	@param[in] p_height Height of the Rectangle.
	 *	@param[in] p_xRadius specify the radii of the ellipses defining the corners of the rounded rectangle.
	 *	@param[in] p_yRadius specify the radii of the ellipses defining the corners of the rounded rectangle.
	 *
	 *	Draws a Rectangle from upper left corner(p_x, p_y) with width = p_width and height = p_height
	 *	and sets the current pen position to (x2, y2).
	 *	Rectangle is drawn with the current pen and brush.
	 *
	 *	@warning This method is a direct forward to the underlying toolkit.
	 *		Currently this means that the coordinates given are used by the CENTER of the WEPen.
	 */
	virtual void drawRect(int p_x, int p_y, int p_width, int p_height, float p_xRadius = 0.0, float p_yRadius = 0.0) = 0;

	/**
	 *	@fn drawRect(const WERect &rect)
	 *
	 *	@brief Convenience method to draw a rectangle with the coordinates as specfied by rect.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] rect WERect specifies the coordinates of the rectangle.
	 *
	 *	This is an overloaded member function.
	 */
	virtual void drawRect(const WERect & p_rect)       =0;

	/**
	 *	@fn fillRect(int p_x, int p_y, int p_width, int p_height, const WEColor& color)
	 *
	 *	@brief Convenience method to set the brush to the specified color and draw a
	 *	rectangle with the coordinates as specified.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_x X Coordinate of the Rectangle.
	 *	@param[in] p_y Y Coordinate of the Rectangle.
	 *	@param[in] p_width Width of the Rectangle.
	 *	@param[in] p_height Height of the Rectangle.
	 *	@param[in] p_color WEColor to be set to the brush for filling the rectangle.
	 *
	 */
	virtual void fillRect(int p_x, int p_y, int p_width, int p_height, const WEColor& p_color) =0;

	/**
	 *	@fn drawText(int x,int y,const char *str)
	 *
	 *	@brief Draws the given string str beginning at the given position.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] x X Coordinate of the position at which to draw the text.
	 *	@param[in] y Y Coordinate of the position at which to draw the text.
	 *	@param[in] str Text to be drawn at the specified position.
	 *	@param[in] auto wrap enable/disable
	 *	@param[in] Text Encode Type
	 *	@param[in] widget transparent enable/disable option
	 *	@param[in] wrapOnHyphen flag to set the text wrap on hyphen
	 *	@param[in] p_trim flag to trim whitespace based on alignment
	 *
	 *	This function does not handle the newline character (\n), as it cannot break text into multiple lines, and it cannot
	 *	display the newline character.
	 *
	 */
	virtual void drawText(int p_x,int p_y,const char *p_str) =0;

	virtual void drawText(const WERect & p_rect, int p_flag, const std::string &p_str, bool p_autowrap = false, WEEncodeType p_EncodeType = UTF_8, bool p_transparent = false, bool p_wrapOnHyphen = false, bool p_trim = true) = 0;
	virtual void drawText(const WERect & p_rect, int p_flag, const std::wstring &p_str, bool p_autowrap = false, bool p_transparent = false, bool p_wrapOnHyphen = false, bool p_trim = true) = 0;

	/**
	 *	@fn virtual void drawArc(int x,int y,int width, int height,int startAngle,int endAngle, bool chord = false) = 0
	 *
	 *	@brief Draws the Arc defined by the rectangle beginning at (x, y) with the specified width and height,
	 *	and the given startAngle and endAngle. Not finally tested
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] x X Coordinate of the position at which to draw the ellipse.
	 *	@param[in] y Y Coordinate of the position at which to draw the ellipse.
	 *	@param[in] width Width of the rectangle enclosing the ellipse.
	 *	@param[in] height Height of the rectangle enclosing the ellipse.
	 *	@param[in] startAngle Start angle(in degrees) of the ellipse.
	 *	@param[in] endAngle End angle(in degrees) of the ellipse.
	 *	@param[in] chord If set to true, a chord is drawn instead of an arch
	 *
	 */
	virtual void drawArc(int p_x,int p_y,int p_width, int p_height, int p_startAngle,int p_endAngle, bool p_chord = false)  = 0;

	/**
	 *	@fn virtual void drawArc(const WERect &rect,int startAngle,int endAngle, bool chord = false) = 0
	 *
	 *	@brief Convenience method to draw the Arc defined by the rectangle rect and the given startAngle and endAngle.Not finally tested
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] rect WERect coordinates of the rectangle enclosing the ellipse.
	 *	@param[in] startAngle Start angle(in degrees) of the ellipse.
	 *	@param[in] endAngle End angle(in degrees) of the ellipse.
	 *	@param[in] chord If set to true, a chord is drawn instead of an arch
	 *
	 */
	virtual void drawArc(const WERect & p_rect, int p_startAngle, int p_endAngle, bool p_chord = false)  = 0;

	/**
	 *	@fn drawEllipse(int x,int y,int width, int height,int startAngle,int endAngle)
	 *
	 *	@brief Draws the ellipse defined by the rectangle beginning at (x, y) with the specified width and height,
	 *	and the given startAngle and endAngle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] x X Coordinate of the position at which to draw the ellipse.
	 *	@param[in] y Y Coordinate of the position at which to draw the ellipse.
	 *	@param[in] width Width of the rectangle enclosing the ellipse.
	 *	@param[in] height Height of the rectangle enclosing the ellipse.
	 *	@param[in] startAngle Start angle(in degrees) of the ellipse.
	 *	@param[in] endAngle End angle(in degrees) of the ellipse.
	 *
	 */
	virtual void drawEllipse(int p_x, int p_y, int p_width, int p_height,int p_startAngle = 0, int p_endAngle = 360) =0;

	//virtual void drawArchedbargraph(int x, int y, int width, int height, int startAngle, int endAngle, int graphWidth) = 0;

	/**
	 *	@fn drawEllipse(const WERect &rect,int startAngle,int endAngle)
	 *
	 *	@brief Convenience method to draw the ellipse defined by the rectangle rect and the given startAngle and endAngle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] rect WERect coordinates of the rectangle enclosing the ellipse.
	 *	@param[in] startAngle Start angle(in degrees) of the ellipse.
	 *	@param[in] endAngle End angle(in degrees) of the ellipse.
	 *
	 */
	virtual void drawEllipse(const WERect & p_rect, int p_startAngle, int p_endAngle) =0;

	/**
	 *	@fn drawPolygon(WEPoint *points,int pointCount)
	 *
	 *	@brief Draws the polygon defined by the first pointCount points in the array points using the current pen and brush.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] points WEPoint Array of points of the polygon.
	 *	@param[in] pointCount Count of the number of points in the polygon.
	 *
	 *	The first point is implicitly connected to the last point, and the polygon is filled with the current brush().
	 */
	virtual void drawPolygon(WEPoint *p_points, int p_pointCount) =0;
	virtual void drawPolygon(std::vector<WEPoint> & p_points, bool p_connect = false)  = 0;
	virtual void drawPolygon(QPolygon & p_polygon, bool p_connect = false)  = 0;

	/**
	 *	@fn setBrush(const WEBrush &brush)
	 *
	 *	@brief Sets the painter's brush to the given brush.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] brush WEBrush to be set for the painter.
	*/
	virtual void setBrush(const WEBrush & p_brush) =0;

	/**
	 *	@fn getBrush()
	 *
	 *	@brief Gets the painter's current brush.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns Current brush set for the painter.
	 *
	 */
	virtual const WEBrush& getBrush() const =0;

	/**
	 *	@fn setPen(const WEPen &pen)
	 *
	 *	@brief Sets the painter's pen to be the given pen.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] pen WEPen to be set for the painter.
	 */
	virtual void setPen(const WEPen & p_pen) =0;

	/**
	 *	@fn getPen()
	 *
	 *	@brief Gets the painter's current pen.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns Current pen set for the painter.
	 */
	virtual const WEPen& getPen() const =0;

	/**
	 *	@fn setFont(WEFont* font)
	 *
	 *	@brief Sets the painter's font to the given font.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] font WEFont to be set for the painter for
	 *	drawing the text.
	 *
	 *	This method copies all the properties from the font provided.
	 *	The pointer provided only has to be valid until the method returns.
	 *
	 *
	 */
	virtual void setFont(WEFont* p_font) =0;

	/**
	 *	@fn virtual void drawImage(int p_x, int p_y, WEImage* p_image, int width = 0, int height = 0) = 0
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns void
	 *
	 *	@param[in] x x location of image.
	 *	@param[in] y y location of image.
	 *	@param[in] pointer to image.
	 */
	virtual void drawImage(int p_x, int p_y, WEImage* p_image, int p_width = 0, int p_height = 0) = 0;


	/**
	 *	@fn drawImage(int p_x, int p_y, int p_width, int p_height, WEImage* p_image)
	 *
	 *	@brief paints WEImage
	 *
	 *	@author Tanmoy Deb
	 *	@author tanmoy.deb@aricent.com
	 *
	 *	@returns void
	 *
	 *	@param[in] p_x x location of image.
	 *	@param[in] p_y y location of image.
	 *	@param[in] p_width The width of the image
	 *	@param[in] p_height The height of the image
	 *	@param[in] p_image pointer to image.
	 *
	 * drawImage implementation. needs a pointer to a WEImage object. Added for comfortable useage.
	 *
	 */
	virtual void drawImage(int p_x, int p_y, int p_width, int p_height, WEImage* p_image) = 0;


/**
	 *	@fn virtual void drawImage(int p_x, int p_y, WEImage* p_image, double p_degree, int width = 0, int height = 0) = 0
	 *
	 *	@brief paints WEImage
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns void
	 *
	 *	@param[in] p_x x location of image.
	 *	@param[in] p_y y location of image.
	 *	@param[in] p_image pointer to image.
	 *	@param[in] p_degree angle to rotate the image in the clockwise direction(Negative degrees means Anti-Clockwise).
	 *
	 * drawImage implementation. needs a pointer to a WEImage object. Added for comfortable useage.
	 *
	 */
	virtual void drawImage(int p_x, int p_y, WEImage* p_image, double p_degree, int p_width = 0, int p_height = 0) = 0;

	virtual void drawRotatedSVGImage(int p_deltaX, int p_deltaY, int p_x, int p_y, int p_width, int p_height, WEImage* p_image, double p_angle) = 0;

	/**
	 *	@fn virtual void setRotate(double p_degree) = 0
	 *
	 *	@brief set the rotate angle of an image
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns void
	 *
	 *	@param[in] p_degree angle for rotating the image
	 *
	 *	This is directly forwarded to Qt which defines
	 *	'Rotates the coordinate system clockwise. The given angle parameter is in degrees.'
	 *	Note that rotating the coordinate system clockwise probably means to
	 *	rotate the object counter-clockwise (i.e. in mathematically positive angle)
	 */
	virtual void setRotate(double p_degree) = 0;

	/**
	 *	@fn virtual void setScale(float p_x, float p_y) = 0
	 *
	 *	@brief scale the coordinate system by (p_x, p_y)
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns void
	 *
	 *	@param[in] p_x scaling offset for x-achis
	 *	@param[in] p_y scaling offset for y-achis
	 *
	 */
	virtual void setScale(float p_x, float p_y) = 0;

	/**
	 *	@fn virtual void setTranslate(float p_x, float p_y) = 0
	 *
	 *	@brief translate the coordinate system by (p_x, p_y)
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns void
	 *
	 *	@param[in] p_x translation offset for x-achis
	 *	@param[in] p_y translation offset for y-achis
	 */
	virtual void setTranslate(float p_x, float p_y) = 0;

	virtual void drawDebugPixelGrid(int p_gridDistance) = 0;


	virtual void rotate(double p_angle) = 0;
	virtual void translate(int p_x, int p_y) = 0;

	//virtual void drawOutlinePath(WEPainterPath * path, int p_outlineWidth) = 0;

	virtual void rotateMiddle(double p_angle, int p_width, int p_height) = 0;

	virtual void drawISOEllipse(const WERect & p_rect, int p_startAngle, int p_endAngle, WEEllipseType p_type) = 0;

	virtual void storeCurrentRenderHints(void) = 0;

	virtual void restoreRenderHints(void) = 0;
};
}	//end of namespace WE
#endif /* WE_PAINTER_H_ */
