/**
 *	@file we_pen.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file defines the pen which is used by the WEPainter to draw
 *	lines and outlines of shapes.
 *
 *	@see we_painter.h
 *	@see we_brush.h
 *
 *	This file defines the pen which is used by the WEPainter to draw
 *	lines and outlines of shapes.
 */


#ifndef WE_PEN_H_
#define WE_PEN_H_

#include "../../../common/include/we_color.h"

namespace WE
{

class WEColor;

/**
 *	@enum WEPenStyle
 *
 *	@brief The pen style defines the line type.
 */
typedef enum
{
	SOLID_LINE,				//!<A Plain Line
	NO_PEN,
	USER_DEFINED_LINE_PATTERN
}WEPenStyle;


/**
 *	@enum WECapStyle
 *
 *	@brief The cap style defines how the end points of lines are drawn.
 */
typedef enum
{
	SQUARE_CAP,				//!< Line is half the line width longer than the end point specified
	FLAT_CAP,				//!< A Flat end of the line - ending exactly at the end point
	ROUND_CAP
}WECapStyle;


/**
 *	@enum WEJoinStyle
 *
 *	@brief The join style defines how the joins between two lines are drawn
 *	when multiple connected lines are drawn.
 */
typedef enum
{
	MITER_JOIN,	//!< The outer edges of the lines are extended to meet at an angle, and this area is filled.
	ROUND_JOIN	//!< A circular arc between the two lines is filled.
}WEJoinStyle;


/**
 *	@struct WEPen
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief The WEPen class defines how a WEPainter should draw lines and
 *	outlines of shapes.
 *
 *	@namespace WE
 *
 *	A pen has a style, width, color, cap style and join style. The pen color
 *	defines the color of lines and text. The default line color is black.
 *
 */
struct WEPen
{
	int width{0};				//!<Width of the line drawn by the pen
	WEColor color{WE::black};			//!<Color of the line drawn by the pen
	WEPenStyle penStyle{SOLID_LINE};	//!<Pen Style to draw the line type(solid,dash) of the pen
	WEJoinStyle joinStyle{ROUND_JOIN};	//!<Join Style to draw how the end points between
	WECapStyle capStyle{FLAT_CAP};	//!<Cap Style to draw the end points of the line drawn with the pen
	unsigned short linePattern{0xffff};
	double widthFactor{1.0};	//!< If line is wider due to diagonal angle line pattern still has to fit original width according to AEF conformance test for ISO 11783-6

	WEPen() = default;

	explicit WEPen(WEPenStyle p_style)
	:penStyle(p_style)
	{
		//other members are initialized at declaration
	}
};
}	//End of Namespace WE

#endif /* WEPEN_H_ */
