/**
 * 	@file qt_we_painter.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora

 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file defines the Qt specific definition of the painter
 *	to be used for painting on the Widgets and Windows.
 *
 *	@see we_widget.h,qt_we_painter.cpp
 *
 *	This file defines the Qt specific painter to be used for painting on the Widgets and Windows. It provides
 *	the interfaces to draw the drawing primitives (point, line, rectangle, text, ellipse and poylygon) and
 *	configure the painter.
 *
 */
#ifndef QT_WE_PAINTER_H_
#define QT_WE_PAINTER_H_

#include <map>
//Included for Inherited classes.
#include <QtGui/QPainter>
#include <QWidget>
#include <QTextDecoder>
#include <QFlags>

#include "../include/we_pen.h"
#include "../include/we_brush.h"
#include "../include/we_painter.h"
#include "../../widgets/include/we_widget.h"


namespace WE
{
/**
 *	@class QTWEPainter
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief The QTWEPainter class provides the QT specific definition of the
 *	WEPainter for the painting on widgets and windows.
 *
 *	@namespace WE
 *
 *	@see we_widget.h
 *	@see we_datatypes_gui.h
 *
 *	@warning Only one painter instance can be created for a widget.
 *	@warning Painter initialised with a null widget may cause problems.
 *
 *	This class inherits from QPainter (for Qt) and WEPainter (which is used by WEWidgets to paint).
 *	All the overridden methods of WEPainter in this class are used to provide wrappers for their
 *	QT specific implementation.
 */
class WEImage; //forward declaration so the class compiles, will be replaced
class WEWidget;
class QTWEWidget;

class QTWEGIFWrapper;

class QTWEPainter : public QPainter,public WEPainter
{
public:

	/**
	 *	@fn QTWEPainter(QPaintDevice *device)
	 *
	 *	@brief Constructs a painter that begins painting the paint device immediately.
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@param[in] device Object that can be painted.
	 *	@param[in] widget Corresponding WEWidget of the paint device.
	 */
	explicit QTWEPainter(QPaintDevice *p_device, WEWidget *p_widget=nullptr);

	/**
	 *	@fn ~QTWEPainter()
	 *
	 *	@brief Standard destructor for QTWEPainter.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 */
	virtual ~QTWEPainter() = default;

	static void initDecoders();
	static void deleteDecoders();

	/**
	 *	@fn setRenderHint(WERenderHints renderHint)
	 *
	 *	@brief Sets the specified render hint on the painter.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] renderHint WERenderHint to be set on the painter.
	 *
	 *	@see WERenderHints
	 *	@see void setRenderHints(WERenderHints renderHints)
	 *
	 *	Sets the specified rendering hint on the painter.
	 */
	void setRenderHint(WERenderHints p_renderHint) override;

	/**
	 *	@fn setRenderHints(WERenderHints renderHints)
	 *
	 *	@brief Sets the render hints off the painter to the specified render hints.
	 *	       This sets all specified render hints and clears all render hints which are not specified.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] renderHints WERenderHints to be set on the painter.
	 *
	 *	@see WERenderHints
	 *	@see void setRenderHint(WERenderHints renderHint)
	 */
	void setRenderHints(WERenderHints p_renderHints) override;

	/**
	 *	@fn getRenderHints()
	 *
	 *	@brief Returns the current render hint of the painter.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@return WERenderHint currently used by the painter.
	 *
	 *	@see WERenderHints
	 */
	int getRenderHints() const override;


	/**
	 *	@fn isRenderHintActive(WERenderHints renderHint)
	 *
	 *	@brief Returns true if the given renderHint is active.
	 *	       Returns false if the given rendeHint is inactive.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@return  true 	if the given renderHint is active.
	 *	         false 	if the given rendeHint is inactive.
	 *
	 *	@see WERenderHints
	 */
	bool isRenderHintActive(WERenderHints p_renderHint) const override;

	/**
	 *	@fn setClip(WERect &rect)
	 *
	 *	@brief It is used to replace the current clip rectangle and sets
	 *	the clip region to the given rectangle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *  @param[in] rect WERect Rectangle Area to be clipped.
	 *
	 */
	void setClip(const WERect & p_rect) override;

	/**
	 *	@fn setClipRegion(WERegion &region)
	 *
	 *	@brief It is used to replace the current clip Region and sets the
	 *	clip region to the given region.
	 *
	 *	@author TANMOY.DEB AT ARICENT.COM
	 *	@author TANMOY DEB
	 *
	 *	@param[in] region WERegion Region Area to be clipped.
	 */
	void setClipRegion(const WERegion & p_region) override;



	void setClipPath(const WEPainterPath & p_painterPath) override;


	/**
	 *	@fn void drawPath(WEPainterPath & p_path)
	 *
	 *	@brief Draw p_path using the current pen and brush.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_path Path to be drawn
	 *
	 *	@returns nothing
	 *
	 *
	 *	This method can be used to e.g. draw an arched bargraph or some other more
	 *	complex shapes.
	 *
	 */
	void drawPath(const WEPainterPath & p_path) override;


	/**
	 *	@fn save(void)
	 *
	 *	@brief Saves the current painter state. A save() must be followed by a
	 *	corresponding restore().
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@see restore(void)
	 *
	 *	Saves the current painter state. A save() must be followed by a
	 *	corresponding restore().
	 */
	void save(void) override;

	/**
	 *	@fn restore(void)
	 *
	 *	@brief Restores the current painter state (pops a saved state off the stack).
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@see save(void)
	 *
	 *	@warning A corresponding save() must have been performed before the restore.
	 *
	 *	Restores the current painter state (pops a saved state off the stack).
	 */
	void restore(void) override;

	/**
	 *	@fn drawPoint(int x,int y)
	 *
	 *	@brief Draws a single point at the given position using the current pen's color.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] x XCoordinate at which to draw the point.
	 *	@param[in] y YCoordinate at which to draw the point.
	 *
	 */
	void drawPoint(int p_x, int p_y) override;

	/**
	 *	@fn drawLine(int x1,int y1,int x2,int y2)
	 *
	 *	@brief Draws a line from (x1, y1) to (x2, y2) and sets the current pen
	 *	position to (x2, y2).
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] x1 X1 Coordinate of the line.
	 *	@param[in] y1 Y1 Coordinate of the line.
	 *	@param[in] x2 X2 Coordinate of the line.
	 *	@param[in] y2 Y2 Coordinate of the line.
	 *
	 */
	void drawLine(int p_x1, int p_y1, int p_x2, int p_y2) override;


	/**
	 * @fn void drawLine(float x1,float y1,float x2,float y2);
	 *
	 * @brief Use floating point coordinates to draw a line from (x1, y1) to (x2, y2) and sets the current pen
	 *	position to (x2, y2).
	 *
	 * @author ajay.bisht@aricent.com
	 * @author Ajay Bisht
	 *
	 * @param[in] x1 X1 Coordinate of the line.
	 * @param[in] y1 Y1 Coordinate of the line.
	 * @param[in] x2 X2 Coordinate of the line.
	 * @param[in] y2 Y2 Coordinate of the line.
	 *
	 */
	void drawLine(float p_x1, float p_y1, float p_x2, float p_y2) override;

	/**
	 *	@fn drawRect(int x1,int y1,int x2,int y2, float p_xRadius, float p_yRadius)
	 *
	 *	@brief Draws a Rectangle from upper left corner(x1, y1) to bottom right corner(x2, y2)
	 *	and sets the current pen position to (x2, y2).
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_x X Coordinate of the Rectangle.
	 *	@param[in] p_y Y Coordinate of the Rectangle.
	 *	@param[in] p_width Width of the Rectangle.
	 *	@param[in] y_height Height of the Rectangle.
	 *  @param[in] p_xRadius specify the radii of the ellipses defining the corners of the rounded rectangle.
	 *	@param[in] p_yRadius specify the radii of the ellipses defining the corners of the rounded rectangle.
	 *
	 *	Draws a Rectangle from upper left corner(x1, y1) with p_width width and p_height height.
	 *	Rectangle is drawn with the currrent pen and brush.
	 *
	 *	@warning This method is a direct forward to the underlying toolkit.
	 *		Currently this means that the coordinates given are used by the CENTER of the WEPen.
	 *
	 */
	void drawRect(int p_x, int p_y, int p_width, int p_height, float p_xRadius = 0.0, float p_yRadius = 0.0) override;

	/**
	 *	@fn drawRect(WERect &rect)
	 *
	 *	@brief Convenience method to draw a rectangle with the coordinates as specfied by rect.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] rect WERect specifies the coordinates of the rectangle.
	 *
	 *	This is an overloaded member function.
	 *
	 *	@warning This method is a direct forward to the underlying toolkit.
	 *		Currently this means that the coordinates given are used by the CENTER of the WEPen.
	 *
	 */
	void drawRect(const WERect & p_rect) override;

	/**
	 *	@fn fillRect(int p_x, int p_y, int p_width, int p_height, const WEColor& p_color)
	 *
	 *	@brief Convenience method to set the brush to the specified color and draw a
	 *	filled rectangle with the coordinates as specified.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_x X Coordinate of the Rectangle.
	 *	@param[in] p_y Y Coordinate of the Rectangle.
	 *	@param[in] p_width Width of the Rectangle.
	 *	@param[in] p_height Height of the Rectangle.
	 *  @param[in] p_color color
	 *
	 */
	void fillRect(int p_x, int p_y, int p_width, int p_height, const WEColor& p_color) override;

	/**
	 *	@fn drawText(int x,int y,const char *str)
	 *
	 *	@brief Draws the given string str beginning at the given position.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] x X Coordinate of the position at which to draw the text.
	 *	@param[in] y Y Coordinate of the position at which to draw the text.
	 *	@param[in] str Text to be drawn at the specified position.
	 *
	 *	This function does not handle the newline character (\n), as it cannot break text into multiple lines, and it cannot
	 *	display the newline character.
	 */
	void drawText(int p_x, int p_y, const char *p_str) override;

	/**
	 *	@fn void drawText(WERect &p_rect, int p_flag, const std::string &p_str, bool autowrap, WEEncodeType p_EncodeType = UTF_8, bool transparent = false, bool wrapOnHyphen = false, bool p_trim = true)
	 *
	 *	@brief Draw the given string str inside of the rectangle
	 *	@brief with the start position (x,y) and size (width, height).
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_rect Coordinates of the position at which to start drawing the text.
	 *	@param[in] p_flag Alignment flag (see WE::WEAlignment)
	 *	@param[in] p_str Text to be drawn at the specified position.
	 *	@param[in] autowrap Whether text is to be drawn wrapped or not.
	 *	@param[in] p_EncodeType text encoding type
	 *	@param[in] transparent flag to set the background as transparent
	 *	@param[in] wrapOnHyphen flag to set the text wrap on hyphen
	 *	@param[in] p_trim flag to trim whitespace based on alignment
	 *
	 *	This function does not handle the newline character (\n),
	 *	This as it cannot break text into multiple lines, and it cannot
	 *	display the newline character.
	 */
	void drawText(const WERect & p_rect, int p_flag, const std::string &p_str, bool p_autowrap, WEEncodeType p_encodeType = UTF_8, bool p_transparent = false, bool p_wrapOnHyphen = false, bool p_trim = true) override;

	/**
	 *	@fn void drawText(WERect &rect, int p_flag, const std::wstring &p_str, bool autowrap, bool transparent = false, bool wrapOnHyphen = false, bool p_trim = true)
	 *
	 *	@brief Draw the given wide string str inside of the rectangle
	 *	@brief with the start position (x,y) and size (width, height).
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] rect Coordinates of the position at which to start drawing the text.
	 *	@param[in] p_flag Alignment flag (see WE::WEAlignment)
	 *	@param[in] p_str Text to be drawn at the specified position.
	 *	@param[in] autowrap Whether text is to be drawn wrapped or not.
	 *	@param[in] transparent flag to set the background as transparent
	 *	@param[in] wrapOnHyphen flag to set the text wrap on hyphen
	 *	@param[in] p_trim flag to trim whitespace based on alignment
	 *
	 *	This function does not handle the newline character (\n),
	 *	This as it cannot break text into multiple lines, and it cannot
	 *	display the newline character.
	 *
	 *	This function does not specifically convert to any of the encoding schemes assuming the
	 *	characters in the wide string are in unicode format as is expected by Qt.
	 */
	void drawText(const WERect & p_rect, int p_flag, const std::wstring &p_str, bool p_autowrap, bool p_transparent = false, bool p_wrapOnHyphen = false, bool p_trim = true) override;

	/**
	 *	@fn drawEllipse(int x,int y,int width, int height,int startAngle,int endAngle)
	 *
	 *	@brief Draws the filled ellipse defined by the rectangle beginning at (x, y) with the specified width and height,
	 *	and the given startAngle and endAngle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] x X Coordinate of the position at which to draw the ellipse.
	 *	@param[in] y Y Coordinate of the position at which to draw the ellipse.
	 *	@param[in] width Width of the rectangle enclosing the ellipse.
	 *	@param[in] height Height of the rectangle enclosing the ellipse.
	 *	@param[in] startAngle Start angle(in degrees) of the ellipse.
	 *	@param[in] endAngle End angle(in degrees) of the ellipse.
	 *
	 *  <MKA 27.08.2009> changed the function so it do now what the name promises
	 *	<MKA 18.05.2010> as default it draws a 360° circle
	 */
	void drawEllipse(int p_x, int p_y, int p_width, int p_height, int p_startAngle = 0,int p_endAngle = 360) override;

	/**
	 *	@fn drawEllipse(WERect &rect,int startAngle,int endAngle)
	 *
	 *	@brief Convenience method to draw the ellipse defined by the rectangle rect and the given startAngle and endAngle.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] rect WERect coordinates of the rectangle enclosing the ellipse.
	 *	@param[in] startAngle Start angle(in degrees) of the ellipse.
	 *	@param[in] endAngle End angle(in degrees) of the ellipse.
	 *
	 *  <MKA 27.08.2009> changed the function so it do now what the name promises
	 *
	 */
	void drawEllipse(const WERect & p_rect, int p_startAngle, int p_endAngle) override;

	// void drawArchedbargraph(int x, int y, int width, int height, int startAngle, int endAngle, int graphWidth);

	/**
	 *	@fn void drawArc(int x,int y,int width, int height,int startAngle,int endAngle, bool chord = false)
	 *
	 *	@brief Draws the Arc defined by the rectangle beginning at (x, y) with the specified width and height,
	 *	and the given startAngle and endAngle. Not finally tested
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] x X Coordinate of the position at which to draw the ellipse.
	 *	@param[in] y Y Coordinate of the position at which to draw the ellipse.
	 *	@param[in] width Width of the rectangle enclosing the ellipse.
	 *	@param[in] height Height of the rectangle enclosing the ellipse.
	 *	@param[in] startAngle Start angle(in degrees) of the ellipse.
	 *	@param[in] endAngle End angle(in degrees) of the ellipse.
	 *	@param[in] chord If set to true, a chord is drawn instead of an arc
	 *
	 */
	void drawArc(int p_x ,int p_y, int p_width, int p_height,int p_startAngle, int p_endAngle, bool p_chord = false) override;

	/**
	 *	@fn void drawArc(WERect &rect,int startAngle,int endAngle, bool chord = false)
	 *
	 *	@brief Convenience method to draw the Arc defined by the rectangle rect and the given startAngle and endAngle.Not finally tested
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] rect WERect coordinates of the rectangle enclosing the ellipse.
	 *	@param[in] startAngle Start angle(in degrees) of the ellipse.
	 *	@param[in] endAngle End angle(in degrees) of the ellipse.
	 *	@param[in] chord If set to true, a chord is drawn instead of an arc
	 *
	 */
	void drawArc(const WERect & p_rect, int p_startAngle, int p_endAngle, bool p_chord = false) override;

	/**
	 *	@fn drawPolygon(WEPoint *points,int pointCount)
	 *
	 *	@brief Draws the polygon defined by the first pointCount points in the array points using the current pen and brush.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] points WEPoint Array of points of the polygon.
	 *	@param[in] pointCount Count of the number of points in the polygon.
	 *
	 *	The first point is implicitly connected to the last point, and the polygon is filled with the current brush().
	 */
	void drawPolygon(WEPoint *p_points, int p_pointCount) override;

	void drawPolygon(std::vector<WEPoint> & p_points, bool p_connect = false) override;
	void drawPolygon(QPolygon & p_polygon, bool p_connect = false) override;

	/**
	 *	@fn setBrush(WEBrush &brush)
	 *
	 *	@brief Sets the painter's brush to the given brush.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] brush WEBrush to be set for the painter.
	 */
	void setBrush(const WEBrush & p_brush) override;

	/**
	 *	@fn getBrush()
	 *
	 *	@brief Gets the painter's current brush.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns Current brush set for the painter.
	 *
	 */
	const WEBrush& getBrush() const override;

	/**
	 *	@fn setPen(WEPen &pen)
	 *
	 *	@brief Sets the painter's pen to be the given pen.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] pen WEPen to be set for the painter.
	 */
	void setPen(const WEPen & p_pen) override;

	/**
	 *	@fn getPen()
	 *
	 *	@brief Gets the painter's current pen.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns Current pen set for the painter.
	 */
	const WEPen& getPen() const override;

	/**
	 *	@fn setFont(WEFont* font)
	 *
	 *	@brief Sets the painter's font to the given font.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_font WEFont to be set for the painter for
	 *	drawing the text.
	 */
	void setFont(WEFont* p_font) override;

	/**
	 *	@fn setRotate(double p_degree)
	 *
	 *	@brief sets the rotate angle for images.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 * 	@param[in] p_degree rotate angle in degrees for images
	 *
	 */
	void setRotate(double p_degree) override;

	void rotate(double p_angle) override;

	void drawDebugPixelGrid(int p_gridDistance) override;

	/**
	 *	@fn void drawImage(int p_x, int p_y, WEImage* p_image, int width = 0, int height = 0)
	 *
	 *	@brief paints WEImage
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns void
	 *
	 *	@param[in] p_x x location of image.
	 *	@param[in] p_y y location of image.
	 *	@param[in] p_image pointer to image.
	 *	@param[in] width width of the image
	 *	@param[in] height height of the image
	 *
	 * drawImage implementation. needs a pointer to a WEImage object.
	 */
	void drawImage(int p_x, int p_y, WEImage* p_image, int p_width = 0, int p_height = 0) override;

	/**
	 *	@fn drawImage(int p_x, int p_y, int p_width, int p_height, WEImage* p_image)
	 *
	 *	@brief paints WEImage
	 *
	 *	@author Tanmoy Deb
	 *	@author tanmoy.deb@aricent.com
	 *
	 *	@returns void
	 *
	 *	@param[in] p_x x location of image.
	 *	@param[in] p_y y location of image.
	 *	@param[in] p_width The width of the image
	 *	@param[in] p_height The height of the image
	 *	@param[in] p_image pointer to image.
	 *
	 *	 drawImage implementation. needs a pointer to a WEImage object. Added for comfortable usage.
	 *
	 */
	void drawImage(int p_x, int p_y, int p_width, int p_height, WEImage* p_image) override;

	/**
	 *	@fn void drawImage(int p_x, int p_y, WEImage* p_image, double p_degree, int width = 0, int height = 0)
	 *
	 *	@brief paints WEImage
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns void
	 *
	 *	@param[in] p_x x location of image.
	 *	@param[in] p_y y location of image.
	 *	@param[in] p_image pointer to image.
	 *	@param[in] p_degree angle to rotate the image.
	 *	@param[in] width
	 *	@param[in] height
	 *
	 *	drawImage implementation. needs a pointer to a WEImage object. Added for comfortable usage
	 *	This method is used by meter widget - even for SVGs.
	 */
	void drawImage(int p_x, int p_y, WEImage* p_image, double p_degree, int p_width = 0, int p_height = 0) override;


	// this is explicitly called by WEImageWidget::paint() for SVG images:
	void drawRotatedSVGImage(int p_deltaX, int p_deltaY, int p_x, int p_y, int p_width, int p_height, WEImage* p_image, double p_angle) override;


	/**
	 *	@fn void setScale(float p_x, float p_y)
	 *
	 *	@brief scale the coordinate system by (p_x, p_y)
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns void
	 *
	 *	@param[in] p_x scaling offset for x-achis
	 *	@param[in] p_y scaling offset for y-achis
	 */
	void setScale(float p_x, float p_y) override;

	/**
	 *	@fn void setTranslate(float p_x, float p_y)
	 *
	 *	@brief translate the coordinate system by (p_x, p_y)
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns void
	 *
	 *	@param[in] p_x translation offset for x-achis
	 *	@param[in] p_y translation offset for y-achis
	 */
	void setTranslate(float p_x, float p_y) override;
	void translate(int p_x, int p_y) override;

	/**
	 *	@fn void setStringCodec(std::string p_codec)
	 *
	 *	@brief set codec used by drawText
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns void
	 *
	 *	@param[in] p_codec codec name used in drawText() default UTF-8
	 *
	 *	@see void drawText(WERect &rect, int p_flag, const std::string &p_str, bool autowrap = false)
	 */
	void setStringCodec(const std::string & p_codec)
	{
		stringCodecName = p_codec;
	}

	/**
	 *	@fn const std::string & getCurrentStringCodec()
	 *
	 *	@brief set codec used by drawText
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns std::string codec name
	 */
	const std::string & getCurrentStringCodec() const
	{
		return stringCodecName;
	}


	//void drawOutlinePath(WEPainterPath * path, int p_outlineWidth);

	void rotateMiddle(double p_angle, int p_width, int p_height) override;

	void drawISOEllipse(const WERect & p_rect, int p_startAngle, int p_endAngle, WEEllipseType p_type) override;

	void begin(QPaintDevice * p_widget);
	void end(void);

	static double gradToRad(double p_grad);

	void storeCurrentRenderHints(void) override;

	void restoreRenderHints(void) override;


private:

	/**
	 *	@fn void init()
	 *
	 *	@brief Initializes the class instance. This method is called from the constructors
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 */
	void init();

	WEPen wePen;			//!< Current pen of the painter
	WEBrush weBrush;		//!< Current brush of the painter

	//MKA 26.01.2010 to get rid of magic strings.
	std::string stringCodecName;

	static std::map<std::string,QTextDecoder*> decoders;	//Map of Encoding Scheme name to corresponding QTextDecoder's to be used for converting to unicode(UTF-8) format

	// WEFont contains color information, including "STYLE_INVERTED" which we need to set in "drawText()"

	int fontStyle;	// only really used to correctly draw inverted characters
	int charWidth;	// introduced for workaround regarding wordwrap with Qt5.

	Qt::PenStyle qtPenStyle;
	Qt::PenCapStyle qtCapStyle;
	Qt::PenJoinStyle qtJoinStyle;
	Qt::BrushStyle qtBrushStyle;

	WEWidget * toolkitWidget;

	QFlags<QPainter::RenderHint> storedRenderHints;


	/**
	 *	@fn setDefaults
	 *
	 *	@brief Set the default value for the configuration of the painter.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 */
	void setDefaults();

	/**
	 *	@fn setQtBrush()
	 *
	 *	@brief Set the QBrush according to the weBrush.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 */
	void setQtBrush();

	/**
	 *	@fn setQtPen()
	 *
	 *	@brief Set the QPen according to the wePen.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 */
	void setQtPen();


	double degree;

	/**
	 *	@fn calcSpanAngle(int p_startAngle, int p_endAngle)
	 *
	 *	@brief calculates the spanAngle
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 * 	@param[in] p_startAngle start angle
	 * 	@param[in] p_endAngle end angle
	 *
	 * 	@returns span angle
	 */
	int calcSpanAngle(int p_startAngle, int p_endAngle) const;

	/**
	 *	@fn void drawText(WERect &p_rect, int p_flag, QString wString, bool autowrap, bool transparent, bool wrapOnHyphen, bool p_trim = true)
	 *
	 *	@brief Draws the text as specified by QString
	 *
	 *	@author maneesh.arora@wachendorff.de
	 *	@author Maneesh Arora
	 *
	 *	It will forever be Maneesh's or Sathees' secret why we need this private version of "drawText()"
	 */
	void drawText(const WERect & p_rect, int p_flag, QString p_wString, bool p_autowrap, bool p_transparent, bool p_wrapOnHyphen, bool p_trim = true);
};
}	//End of namespace WE
#endif /* QT_WE_PAINTER_H_ */
