/**
 *	@file toolkit_we_factory.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Factory class that makes objects of the
 *
 *	@see we_widget.h, we_widget.cpp, qt_we_widget.h, qt_we_widget.cpp
 *
 *	Thie class provides a function that makes an object
 *	for each interface from the family of widget-classes
 */

#ifndef WE_WIDGET_FACTORY_H_
#define WE_WIDGET_FACTORY_H_

#include <memory>
#include "../../common/include/we_region.h"
#include "../../common/include/we_painter_path.h"

namespace WE
{

//class WEApplication;
class WEWidget;
class ToolkitWEImage;
class ToolkitWEFont;
class WEPainter;
//class ToolkitWEApplication;
class ToolkitWERegion;
class WERect;
class WERegion;
class WEPainterPath;



class ToolkitWEFactory
{
public:
	/*--------------------------------------------------------*/
	/**
	 *	@fn ~ToolkitWEFactory()
	 *
	 *	@brief Standard destructor of WEWidgetFactory class
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	The destructor deletes the object of the factory class.
	 */
	/*--------------------------------------------------------*/
	virtual ~ToolkitWEFactory();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	ToolkitWEFactory(const ToolkitWEFactory &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	ToolkitWEFactory & operator=(const ToolkitWEFactory &) = delete;

	/**
	 *	@fn static ToolkitWEFactory * create()
	 *
	 *	@brief The create() method creates and returns the instance of the ToolkitWEFactory of this process.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns pointer to the created instance of the ToolkitWEFactory
	 *
	 *	@warning When using other methods they have to be called by calling the instance method! (see below)
	 *		e.g. ToolkitWEFactory::instance()->otherMethod();
	 *
	 *	Since ToolkitWEFactory is a singleton, there is only one instance per process allowed.
	 *	Therefore this function shall only be called once within the process.
	 */
	//static ToolkitWEFactory * create();

	/**
	 *	@fn void release()
	 *
	 *	@brief deallocates the memory allocated by an ToolkitWEFactory instance.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	void release();

	/*--------------------------------------------------------*/
	/**
	 *	@fn static ToolkitWEFactory * getInstance()
	 *
	 *	@brief The instance() method returns the instance of the ToolkitWEFactory of this process.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns pointer to the existing instance of the ToolkitWEFactory. If no ToolkitWEFactory
	 *		 instance exists, it gives some error message and abort the process.
	 *
	 *	@warning When using other methods they have to be called by calling the instance method!
	 *		e.g. ToolkitWEFactory::instance()->otherMethod();
	 *
	 *	Since ToolkitWEFactory is a singleton, there is only one instance per process allowed.
	 *	Therefore this static method has to be used to call any other methods of
	 *	the ToolkitWEFactory.
	 */
	static ToolkitWEFactory * getInstance();

	/*--------------------------------------------------------*/
	/**
	 *	@fn createWidget(WEWidget* p_weWidget,ToolkitWEWidget *p_parentToolkitWidget)
	 *
	 *	@brief Create a concrete widget class
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_weWidget WEWidget Reference to the WEWidget for which this
	 *	toolkit widget is being created.
	 *	@param[in]  p_parentToolkitWidget ToolkitWEWidget Reference to the
	 *	parent toolkit widget of this ToolkitWEWidget.
	 *	This function create the instance of the concrete widget class
	 *	depending on the Toolkit-Settings
	 */
	/*--------------------------------------------------------*/
	//ToolkitWEWidget * createWidget(WEWidget* p_weWidget,ToolkitWEWidget *p_parentToolkitWidget);
	std::unique_ptr<ToolkitWEFont> createFont();


	/**
	 *	@fn static void releasePainterResources( void )
	 *
	 *	@brief Releases static resources otherwise used by every painter.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	Deletes static resources used by all painters.
	 *	E.g. for Qt this deletes QTextDecoder instances shared among all painter instances.
	 */
	static void releasePainterResources( void );


	/**
	 *	@fn WEToolkitImage* createImage()
	 *
	 *	@brief creates instance of an image
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @returns pointer to WEToolkitImage interface
	 *
	 * creates an Image that can be stored as a pointer.
	 */
	std::shared_ptr<ToolkitWEImage> createImage();

	/**
	 *	@fn ToolkitWERegion* createRegion(WERect &rect, WE::WERegionType regionType)
	 *
	 *	@brief Factory method for creating an instance of WERegion
	 *	for a specific Widget
	 *
	 *	@author TANMOY.DEB AT ARICENT.COM
	 *	@author TANMOY DEB
	 *
	 *	@returns ToolkitWERegion
	 *
	 */
	std::unique_ptr<ToolkitWERegion> createRegion(WERect &p_rect, WE::WERegionType p_regionType);

	/**
	 *	@fn ToolkitWEPainterPath* createPainterPath(WERect &rect, WE::WERegionType regionType, int startAngle, int spanAngle, bool antialiasing)
	 *
	 *	@brief Factory method for creating an instance of WEPainterPath
	 *	for a specific Widget
	 *
	 *	@author TANMOY.DEB AT ARICENT.COM
	 *	@author TANMOY DEB
	 *
	 *	@returns ToolkitWEPainterPath
	 *
	 */
	std::unique_ptr<ToolkitWEPainterPath> createPainterPath(WERect &p_rect, WE::WERegionType p_regionType, int p_startAngle, int p_spanAngle, bool p_antialiasing);
	std::unique_ptr<ToolkitWEPainterPath> createCopyPainterPath(ToolkitWEPainterPath * p_copy);

private:
	/*--------------------------------------------------------*/
	/**
	 *	@fn ToolkitWEFactory()
	 *
	 *	@brief Private constructor of ToolkitWEFactory class
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	This constructor should be private,
	 *	This  while it's instance should be created as singleton with the getInstance function
	 *	This constructor have no functionalities
	 */
	/*--------------------------------------------------------*/
	ToolkitWEFactory();
}; // end of WEWidgetFactory class
} // end of namespace WE
#endif /* WE_WIDGET_FACTORY_H_ */
