/**
 *	@file we_button_widget.h
 *
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This files defines a button widget
 *
 *	Button widget is a container which has a pressed and released state.
 *	If different children should be shown for both states currently one has to make
 *	them visible or invisible from the controlling level.
 *	In PClient this is done by AppButtonDDO.
 *
 *	For ISO-VT all the children have to be shown at any time but optionally a shadow border
 *	is shown to visualize the state of the button.
 *	If this shadow border isn't disabled completely the child coordinates are with respect to
 *	the area inside this border i.e. child positions would have to be modified to address this.
 *	Additionally clipping has to assure that the shadow border area only shows the border or is
 *	fully transparent i.e. the button's children may never draw there.
 *
 *	To overcome all these problems we decided to add a container widget as the actual parent of
 *	all the button's children.
 *	This provides the following advantages:
 *	- children are clipped correctly
 *	- child position can be the same as in the pool (independent of activated shadow border)
 *	- switching between shadow border on and completely off can be achieved by just moving and
 *	  resizing the container.
 *
 *	Other considered approaches were:
 *	- modify clip region for children (can be done with Qt's "setMask()") --> would have to calculate
 *	  specific mask for each child and recalculate child positions if shadow border is switched.
 *	- draw border widget on top of all children --> can't make border transparent, would again have to
 *	  recalculate child positions.
 *
 *	The approach with the container is widely copied from the list widget - we decided against a common
 *	base widget for this because both do totally different things with the children in the container.
 *
 *	To allow synchronization between these two widgets methods of both were reordered in the same logical sequence:
 *	- constructor, destructor, initialization
 *	- child addition, removal, get, set
 *	- paint
 *	- value related
 *	Header and source file are kept in the same order (with the exception that protected and private methods are
 *	moved to the bottom in the header file).
 *
 *	@see WEWidget
 *	@see WEListWidget
 *
 */

#ifndef WE_BUTTON_WIDGET_H_
#define WE_BUTTON_WIDGET_H_

//Included for data Members and Inherited classes
#include "we_int_widget.h"

namespace WE
{

class WEContainer;
class WEGeometryNode;
class WEControllerNode;
class WEImageNode;
class WEImage;
class WEPainter;
class WEWidget;

class WEButtonWidget : public WEIntWidget
{
	Q_OBJECT
	Q_PROPERTY(bool drawShadow READ isDrawShadow)
	Q_PROPERTY(unsigned int shadowWidth READ getShadowWidth)
public:
	/**
	 *	@fn WEButtonWidget(WEWidget *p_parent, void * p_controller = nullptr, int p_typeID = WIDGET_TYPE_ID_BUTTON)
	 *
	 *	@brief Constructs a WEButtonWidget.This widget can be created as
	 *	a separate window by specifying the parent as nullptr or it can be included
	 *	as a child widget of the parent.
	 *
	 *	@param[in] p_parent If parent IS nullptr then this widget becomes a window with the same properties.
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 */
	explicit WEButtonWidget(WEWidget *p_parent, void * p_controller = nullptr, int p_typeID = WIDGET_TYPE_ID_BUTTON);


	/**
	 *	@fn ~WEButtonWidget()
	 *
	 *	@brief Destructor of the widget.
	 *
	 */
	virtual ~WEButtonWidget();


	/**
	 *	@fn init()
	 *
	 *	@brief the method to initialize the member variable of the class
	 *
	 */
	void init();


	/**
	 *	@fn add(WEWidget * p_widget)
	 *
	 *	@brief add a Item to the List and as a child to the widget
	 *
	 *	@see WEWidget
	 *
	 *	Overwrite method from WEWIdget.
	 */
	int add(WEWidget * p_widget) override;


	/**
	 *	@fn void removeChild(WEWidget * p_child)
	 *
	 *	@brief remove child from widget and delete child from button. Overwritten from WEWidget
	 *
	 *	@see WEWidget
	 *
	 *	Overwrite method from WEWIdget.
	 */
	void removeChild(WEWidget * p_child) override;


	/**
	 *	@fn bool hasChildren()
	 *
	 *	@brief indicates if the widget has children
	 *
	 *	Overwrite method from WEWidget. Yes, a button object has children, but no children for editing.
	 *	So this method always returns FALSE!
	 */
	bool hasChildren() const override;


	/**
	 *	@fn WEWidget * getFirstChild()
	 *
	 *	@brief Skip the container we use here and return its getFirstChild()
	 *
	 *	Overwrite method from WEWidget.
	 */
	WEWidget * getFirstChild() const override;


	/**
	 *	@fn WEWidget * getLastChild()
	 *
	 *	@brief Skip the container we use here and return its getLastChild()
	 *
	 *	Overwrite method from WEWidget.
	 */
	WEWidget * getLastChild() const override;


	/**
	 *	@fn WEWidget * getFirstInputChild()
	 *
	 *	@brief Skip the container we use here and return its getFirstInputChild()
	 *
	 *	Overwrite method from WEWidget.
	 */
	WEWidget * getFirstInputChild() const override;


	/**
	 *	@fn WEWidget * getLastInputChild()
	 *
	 *	@brief Skip the container we use here and return its getLastInputChild()
	 *
	 *	Overwrite method from WEWidget.
	 */
	WEWidget * getLastInputChild() const override;


	/**
	 *	@fn WEWidget * getChildren(void)
	 *
	 *	@brief Skip the container we use here and return its getChildren(void)
	 *
	 *	Overwrite method from WEWidget.
	 */
	childrenBiMap * getChildren(void) const override;


	/**
	 *	@fn void setSoftkeyAsInput(bool p_input)
	 *
	 *	@brief Should this softkey be included in navigation scheme?
	 *
	 *	@oaram p_input true: Softkey should be included in navigation, false: No navigation to this softkey.
	 *
	 *	@returns nothing
	 *
	 *	If p_input is set to true and this widget is a softkey then isInputWidget() will return true and
	 *	therefore this widget will be included in navigation scheme.
	 *	The default is to exclude softkeys from navigation.
	 *
	 */
	void setSoftkeyAsInput(bool p_input);


	/**
	 *	@fn isInputWidget()
	 *
	 *	@brief Check whether the widget is an input widget or not.
	 *
	 *	@returns TRUE if the widget is an input widget, FALSE otherwise.
	 *
	 *	Only an input widget can get the focus during navigation.
	 *	The button is also used for softkeys which are only included in navigation on demand.
	 *	So to cover this specific behaviour represented by "this->softkeyIsInput" we had to
	 *	reimplement this method here.
	 *
	 *	Overwrite method from WEWidget.
	 */
	bool isInputTypeWidget() const override;


	/**
	 *	@fn setSize(int p_width, int p_height)
	 *
	 *	@brief Sets the size of the Widget.
	 *
	 *	@param[in] p_width Width to be set for the widget.
	 *	@param[in] p_height Height to be set for the widget.
	 *
	 *	We have to set the size of the items container in addition to what WEWidget would do.
	 *
	 *	Overwrite method from WEWidget.
	 */
	int setSize(int p_width, int p_height) override;

	/**
	 *	@fn setBackground(WEColor & p_bgColor)
	 *
	 *	@brief Sets the background color of the widget to p_bgColor.
	 *
	 *	@param[in] p_bgColor Background color to be set for the widget.
	 *
	 *	The method is overwritten here because it has to forward the setting
	 *	to the "items" container. The button widget itself never draws any background.
	 *
	 *	Overwrite method from WEWidget.
	 */
	int setBackground(const WEColor & p_bgColor) override;


	/**
	 *	@fn createBackgroundImage(std::string p_strFile)
	 *
	 *	@brief creates the background image
	 *
	 *	@param p_strFile path and name of the image file to be used as background
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_GENERAL_ERROR if
	 *
	 *	creates background image and stores pointer
	 *	The method is overwritten here because it has to forward the setting
	 *	to the "items" container. The button widget itself never draws any background.
	 *
	 *	If the container does not yet exist this will always return WE_SUCCESS even
	 *	though image creation may fail when container is created.
	 *
	 *	Overwrite method from WEWidget.
	 */
	int createBackgroundImage(const std::string & p_strFile) override;


	/**
	 *	@fn int setTransparent(bool p_isTransparent)
	 *
	 *	@brief Sets the transparency of this widget by forwarding it to the items container.
	 *
	 *	@param[in] p_width Width to be set for the widget.
	 *	@param[in] p_height Height to be set for the widget.
	 *
	 *	We have to set the size of the items container in addition to what WEWidget would do.
	 *
	 *	Overwrite method from WEWidget.
	 */
	int setTransparent(bool p_isTransparent) override;


	/**
	 *	@fn bool getBackgroundIsWholeButton( void )
	 *
	 *	@brief Figure out if button draws background behind shadow or not.
	 *
	 *	@param none
	 *
	 *	@returns true: Background includes shadow, false: Shadow background is transparent
	 *
	 */
	bool getBackgroundIsWholeButton( void ) const;


	/**
	 *	@fn int setBackgroundIsWholeButton( bool p_wholeButton );
	 *
	 *	@brief Define if button draws background behind shadow or not.
	 *
	 *	@param p_wholeButton true: Background includes shadow, false: Shadow background is transparent
	 *
	 *	@returns WE::WE_SUCCESS
	 *
	 */
	int setBackgroundIsWholeButton( bool p_wholeButton );


	/**
	 *	@fn int setDrawShadow(bool p_draw)
	 *
	 *	@brief Actually draw the shadow if space is reserved using setShadowOn()?
	 *
	 *	@param[in] p_draw should be true to draw the shadow or false to keep the area transparent
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	The shadow will only be drawn if space was reserved by calling setShadowOn(true).
	 *
	 *	@see setShadowOn(bool p_on)
	 */
	int setDrawShadow(bool p_draw);

	/**
	 *	@fn int setShadowWidth(unsigned int p_width)
	 *
	 *	@brief set the width of the button shadow to the parameter value
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_width the width for the button shadow
	 *
	 *	@returns WE_SUCCESS
	 */
	int setShadowWidth(unsigned int p_width);

	/**
	 *	@fn unsigned int getShadowWidth()
	 *
	 *	@brief get the current width of the button shadow
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns unsigned int for the shadow width
	 */
	unsigned int getShadowWidth() const;


	/**
	 *	@fn int setShadowColor(WEColor & p_color)
	 *
	 *	@brief Set the color to be used for the shadow area.
	 *
	 *	@param[in] p_color Color to be used for the shadow area.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	The shadow will only be drawn if space was reserved by calling setShadowOn(true)
	 *	and activated by setDrawShadow(true).
	 *
	 *	@see setShadowOn(bool p_on)
	 *	@see setDrawShadow(bool p_draw)
	 */
	int setShadowColor(const WEColor & p_color);


	/**
	 *	@fn paint(WEPainter* p_painter)
	 *
	 *	@brief paints the list
	 *
	 *	@param[in] p_painter Pointer to painter the object
	 *
	 *	@returns void
	 */
	void paint(WEPainter* p_painter) override;


	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData *p_data) override;


	/**
	 *	@fn int setValue(int p_value)
	 *
	 *	@brief Set value after limiting it to boolean range.
	 *
	 *	@param p_value Value to be set.
	 *	@param[in] p_updateScrollbar unused for buttons
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	Ensure that the button value always is "true" or "false".
	 *
	 *	Overwrite method from WEIntWidget.
	 *
	 */
	int setValue(int p_value, bool p_updateScrollbar = false) override;


	/**
	 *	@fn int setVisible(bool p_visible)
	 *
	 *	@brief Set the visibility status of a widget
	 *
	 *	@param[in] p_visible Boolean value that represents the visibility status of the widget
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This method overwrites (and internally calls WE::WEWidget::setVisible(bool)) to cover this'
	 *	widget's internal "items" container.
	 *
	 *	@see show()
	 *	@see WE::WEWidget::setVisible(bool)
	 */
	void setVisible(bool p_visible) override;


	/**
	 *	@fn void show()
	 *
	 *	@brief Show the widget
	 *
	 *	@param[in] none
	 *
	 *	@returns nothing
	 *
	 *	This method overwrites (and internally calls WE::WEWidget::show()) to cover this'
	 *	widget's internal "items" container.
	 *
	 *	@see WE::WEWidget::show()
	 *	@see setVisible(bool)
	 */
	void show();

	bool isDrawShadow(void) const;

	/**
	 *	@fn createPressedImage(std::string p_strFile)
	 *
	 *	@brief creates the pressed image
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@param p_strFile path and name of the image file to be used as image
	 *
	 *	@returns WE_SUCCESS or failure
	 *
	 *	creates pressed image and stores pointer
	 *
	 *	@see createImage(std::string p_strFile)
	 */
	int createPressedImage(const std::string & p_strFile);

	/**
	 *	@fn createPressedImage(std::string p_strFile)
	 *
	 *	@brief creates the released image
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@param p_strFile path and name of the image file to be used as image
	 *
	 *	@returns WE_SUCCESS or failure
	 *
	 *	creates released image and stores pointer
	 *
	 *	@see createImage(std::string p_strFile)
	 */
	int createReleasedImage(const std::string & p_strFile);

protected:
	/**
	 *	@fn void setContainerPositionSize(void)
	 *
	 *	@brief Sets the size of the "items" container to follow the settings for the button widget.
	 *
	 *	@param none
	 *
	 *	@returns nothing
	 *
	 *	Container size and position change with the button size and the shadow properties.
	 *
	 */
	void setContainerPositionSize(void);

	/**
	 *	@fn drawPressedImage(WEPainter * p_painter)
	 *
	 *	@brief Draws the pressed image.
	 *
	 *	@param[in] p_painter
	 *
	 *	@returns WE_SUCCESS
	 */
	void drawPressedImage(WEPainter * p_painter);

	/**
	 *	@fn drawReleasedImage(WEPainter * p_painter)
	 *
	 *	@brief Draws the released image.
	 *
	 *	@param[in] p_painter
	 *
	 *	@returns WE_SUCCESS
	 */
	void drawReleasedImage(WEPainter * p_painter);

	void updateImageNode(WEControllerNode *p_controller, WEImage * p_image);

	unsigned int	shadowWidth;	//!< The width of the shadow for the button to make 3d effect
	bool		softkeyIsInput;	//!< Include this softkey in navigation scheme? (only valid if this really is a softkey)
	bool		backgroundIsWholeButton;	//!< For PClient we want to apply background even behind shadow, for ISO-VT shadow has to be transparent
	bool		drawShadow;	//!< actually draw the shadow in the space reserved? (only valid if "shadowOn" is true)
	bool		hasContainer;	// container is already created?

	WEContainer * 	items;		// container providing correct position and clipping of children (kept its name from the list widget)
					// this can not be a member because we need a parent when calling its constructor (so we can not call it in the button's constructor)
	WEColor		shadowColor;	//!< Color to be used for the shadow drawn

	WEImage * pressedImage;
	WEImage * releasedImage;

	WEGeometryNode * shadowGeometryNode1;	//!<Node of the Widget which is used to draw shadow
	WEGeometryNode * shadowGeometryNode2;	//!<Node of the Widget which is used to draw shadow
	WEImageNode * imageNode;		//!<Node of the Widget which is used to draw the pressed/released image
};
}
#endif /* WE_BUTTON_WIDGET_H_ */
