/**
 *	@file we_frametime_item.h
 *
 *	@author preths@topcon.com
 *	@author Patrick Reths
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2020
 *
 */

#ifndef WE_FRAMETIME_ITEM_H
#define WE_FRAMETIME_ITEM_H

#include <string>

#include <QTime>
#include <QtQuick/QQuickItem>

namespace WE
{

/**
 *
 * This class can be used to print the current frame time and frame rate to stdout
 *
 */
class WEFrametimeItem: public QQuickItem
{
	Q_OBJECT

public:

	/**
	 *	@fn WEFrametimeItem(QQuickItem *parent = nullptr, int printAllAbove=20)
	 *
	 *	@brief Constructor for the WEFrametimeItem.
	 *	This class can be used to print the current frame time and frame rate to stdout. Frame times are printed every second.
	 *	Use the printAllAbove parameter to print all frame times which are above the given value.
	 *
	 *	The printed information:
	 *	- "frametime": the frame time
	 *	- "frametime fps": the theoretical frame rate of the current frame (1000 / frame time)
	 *	- "real fps": the frame rate based on the counted frames in the last second
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] parent - pointer to parent item
	 *	@param[in] printAllAbove - print all frame times above this value
	 */
	explicit WEFrametimeItem(QQuickItem *p_parent = nullptr, int p_printAllAbove=20);
	~WEFrametimeItem() = default;

	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode - the previously returned value of this function
	 *	@param data - contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *, UpdatePaintNodeData *) override;

private:
	void print(QTime p_frameTime, const std::string & p_prefix, const std::string & p_postfix);
	void updateAndPrint(QTime p_currentTime, const std::string & p_prefix, const std::string & p_postfix);
	void updateAndPrint();

	const int printAllAbove;
	int frameCounter;
	QTime oldTime;
	QTime printTime;
};
} //End of namespace WE
#endif // WE_FRAMETIME_ITEM_H

