/**
 *	@file we_graph_widget.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of WEGraphWidget class. See class description.
 */


#ifndef WE_GRAPH_WIDGET_H_
#define WE_GRAPH_WIDGET_H_

#include <deque>
#include <map>
#include <array>

#include "../../../event_handling/include/we_signal.h"
#include "we_text_widget.h"
#include "we_widget.h"
#include "we_graph_widget_types.h"

namespace WE
{

class WELineNode;
class WETextNode;
class WEControllerNode;
class WEPainter;

/**
 *	@class WEGraphWidget
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Definition of WEGraphWidget. Seed description for details.
 *
 *	@namespace WE
 *
 *	This class represents a "2D Graph". It's used to visualize the history of one or several
 *	variables. The value history of a variable is drawn as line in the graph.
 *	Optionally a cursor can be shown in the graph. This cursor shows the value of all variables
 *	at the selected line (setCursorPosition()).
 *	Various options can be set to configure the look of that line (see WEVariableGraphOptions).
 *	To set the options for all variables use setVariableOptionsPointer().
 *	To set the value history use setValueHistoryPointer().
 *	The graph itself can be configured like any other widget.
 *
 *	WEGraphWidget inherits WEWidget
 *	and WESignalListener (listen to timeout events for periodic repaint if variable values did not change)
 */
class WEGraphWidget : public WEWidget, public WETextWidget, public WESignalListener
{
	Q_OBJECT
	Q_PROPERTY(unsigned int xAxisTime READ getXAxistime)
	Q_PROPERTY(unsigned int xAxisOffset READ getXAxisOffset)
	Q_PROPERTY(int axisLineWidth READ getAxisLineWidth)
	Q_PROPERTY(int gridLineWidth READ getGridLineWidth)
	Q_PROPERTY(int cursorPosition READ getCursorPosition)
	Q_PROPERTY(unsigned short axisLinePattern READ getAxisLinePattern)
	Q_PROPERTY(unsigned short gridLinePattern READ getGridLinePattern)
	Q_PROPERTY(int yAxisPosition READ getYAxisPosition)
	Q_PROPERTY(int xAxisLabels READ getXAxisLabels)
	Q_PROPERTY(int numberOfGridLines READ getNumberOfGridLines)
	Q_PROPERTY(bool drawGridLines READ getDrawGridLines)
	Q_PROPERTY(unsigned long long xOffsetTime READ getXOffsetTime)
	Q_PROPERTY(int refreshCycleTimeMS READ getRefreshCycleTimeMS)

public:
	/**
	 * Constructor. Initializes values and the parent widgets.
	 * @param p_parent Parent widget
	 * @param p_controller widget controller
	 */
	WEGraphWidget(WEWidget * p_parent, void * p_controller);

	/**
	 * Destructor.
	 * Free memory and remove timer subscriptions.
	 */
	~WEGraphWidget();

	/**
	 * The usual paint method that draws the contents of the graph.
	 * @param painter
	 */
	void paint(WEPainter *painter) override;

	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data) override;

	/**
	 * This method is used to set the scale of X axis in seconds.
	 * @param p_seconds scale of the X axis in seconds
	 */
	void setXAchsisTime(unsigned int p_seconds);

	/**
	 * Set the offset of the X axis in seconds. With this method it can be scrolled into the past.
	 * @param p_offset offset from now in seconds
	 */
	void setXAchsisOffset(unsigned int p_offset);

	/**
	 * Set the pointer to the value history. This pointer needs to point to a map in which
	 * each variable is mapped to a deque that contains all time stamps with the
	 * matching value at that time
	 * @param p_history pointer to the value history
	 */
	void setValueHistoryPointer(std::map<unsigned int, std::deque<WEGraphData> > * p_history);

	/**
	 * Set the pointer to the map that contains all options for the variables
	 * that shall be drawn.
	 * @param p_options
	 */
	void setVariableOptionsPointer(std::map<unsigned int, WEVariableGraphOptions> * p_options);

	/**
	 * Set the drawing option for the X axis labels (full time stamps, relative time stamps or none)
	 * @param p_labels the option to set
	 */
	void setXAxisLabels(XAxisLabelOptions p_labels)
	{
		this->xAxisLabels = p_labels;
	}

	/**
	 * Set the options for the look of the axis lines
	 * @param p_width line width of the axis
	 * @param p_pattern line pattern for the axis
	 * @param p_color color of the axis lines
	 */
	void setAxisLineAttributeValues(int p_width, unsigned short p_pattern, WEColor p_color)
	{
		this->axisLineColor = p_color;
		this->axisLinePattern = p_pattern;
		this->axisLineWidth = p_width;
	}

	/**
	 * Set the options for the look of the grid lines
	 * @param p_width line width of the grid lines
	 * @param p_pattern line pattern for the grid lines
	 * @param p_color color of the grid lines
	 */
	void setGridLineAttributeValues(int p_width, unsigned short p_pattern, WEColor p_color)
	{
		this->gridLineColor = p_color;
		this->gridLinePattern = p_pattern;
		this->gridLineWidth = p_width;
	}

	/**
	 * Y axis can be position left, right or not at all
	 * @param p_pos one of the YAxisPosition enum values
	 */
	void setYAxisPosition(YAxisPosition p_pos)
	{
		this->yAxisPosition = p_pos;
	}

	/**
	 * Set the flag to draw the grid lines or not
	 * @param p_drawLines true: draw grid lines, false: don't draw grid lines
	 */
	void setDrawGridLines(bool p_drawLines)
	{
		this->drawGridLines = p_drawLines;
	}

	/**
	 * Set the number of grid lines. 1 or three lines can be drawn. To draw no grid use setDrawGridLines(false).
	 * @param p_lines either ONE_GRID_LINE or THREE_GRID_LINES
	 */
	void setNumberOfGridLines(GridLineCount p_lines)
	{
		this->numberOfGridLines = p_lines;
	}

	/**
	 * signalOccurred is used to receive a timeout event that will cause a repaint.
	 * This is needed because we also have to do repaints even when no variable changes happened
	 * (because time still progresses ;) )
	 * @param event the timeout event
	 * @return 0
	 */
	int signalOccurred(WESignal & event) override;

	/**
	 * Set the cursor position in seconds. This shows and moves the cursor.
	 * @param p_position value < 0: hide the cursor. value >= 0 move the cursor
	 */
	void setCursorPosition(int p_position)
	{
		this->cursorPosition = p_position;
	}

	unsigned int getXAxistime(void) const
	{
		return this->xAchsisTime;
	}

	unsigned int getXAxisOffset(void) const
	{
		return this->xAchsisOffset;
	}

	int getAxisLineWidth(void) const
	{
		return this->axisLineWidth;
	}

	int getGridLineWidth(void) const
	{
		return this->gridLineWidth;
	}

	int getCursorPosition(void) const
	{
		return this->cursorPosition;
	}

	unsigned short getAxisLinePattern(void) const
	{
		return this->axisLinePattern;
	}

	const WE::WEColor & getAxisLineColor(void) const
	{
		return this->axisLineColor;
	}

	unsigned short getGridLinePattern(void) const
	{
		return this->gridLinePattern;
	}

	const WE::WEColor & getGridLineColor(void) const
	{
		return this->gridLineColor;
	}

	int getYAxisPosition(void) const
	{
		return this->yAxisPosition;
	}

	int getXAxisLabels(void) const
	{
		return this->xAxisLabels;
	}

	int getNumberOfGridLines(void) const
	{
		return this->numberOfGridLines;
	}

	bool getDrawGridLines(void) const
	{
		return this->drawGridLines;
	}

	unsigned long long getXOffsetTime(void) const
	{
		return this->xOffsetTime;
	}

	void setRefreshCycleTimeMS(int p_time);

	inline int getRefreshCycleTimeMS(void) const
	{
		return this->refreshCycleTimeMS;
	}

private:
	/**
	 * The inner bounding box is a rectangle in which the grid, cursor and value lines are drawn.
	 * This function calculates the position and size of the rectangle based on the options set for the widget.
	 * The size of the inner bounding box can also change at runtime (e.g. if width of labels on Y axis change)
	 * @param p_painter the painter used for drawing this widget
	 */
	void prepareInnerBoundingBox(WEPainter * p_painter=nullptr);

	/**
	 * Function to draw the X and Y axis
	 * @param p_painter the painter used for drawing this widget
	 */
	void drawAxis(WEPainter * p_painter);

	void updateAxisNodes(WEControllerNode * controller);

	/**
	 * Function to draw the labels below the X axis.
	 * Either full time stamps, relative time stamps or none are drawn.
	 * @param painter the painter used for drawing this widget
	 */
	void drawXLabels(WEPainter * painter);

	void updateXLabelNodes(WEControllerNode * controller);

	/**
	 * This function draws the Y labels (if configured) for each variable and
	 * calls drawValueLineForVariable() to draw each value line
	 * @param p_painter the painter used for drawing this widget
	 */
	void drawValueLinesAndYLabels(WEPainter * p_painter);

	void updateValueLinesAndYLabelNodes(WEControllerNode * controller);

	void drawLabels(const std::array<std::string, LABEL_POSITION_INVALID> & labels, const WEVariableGraphOptions & item,  WEPainter * p_painter, int numberOfLabelsDrawn);

	void updateLabelNodes(unsigned int index, WEControllerNode * controller, const std::array<std::string, LABEL_POSITION_INVALID> & labels, const WEVariableGraphOptions & item, int numberOfLabelsDrawn);

	/**
	 * Draw the value line for the specified variable
	 * @param p_key unique key of the variable which's line that shall be drawn
	 * @param p_painter the painter used for drawing this widget
	 */
	void drawValueLineForVariable(unsigned int p_key, WEPainter * p_painter);

	void updateValueLineForVariableNodes(unsigned int index, unsigned int p_key, WEControllerNode * controller);

	/**
	 * Draws the cursor
	 * @param painter the painter used for drawing this widget
	 */
	void drawCursor(WEPainter *painter);

	void updateCursorNodes(WEControllerNode * controller);

	/**
	 * Function to calculate the min value, max value and value range of a given variable
	 * at a defined time interval (from current "now" to "now - this->xAxisTime)
	 * @param minValue Reference to min value variable, result will be stored in it
	 * @param maxValue Reference to max value variable, result will be stored in it
	 * @param valueRange Reference to value range variable, result will be stored in it
	 * @param now the time point for the current "now" (now is based on xAxisOffset)
	 * @param p_history pointer to the value history deque of a variable
	 */
	void calculateMinMaxAndRangeForVariable(int &minValue,
			int &maxValue,
			int &valueRange,
			const unsigned long long &now,
			std::deque<WEGraphData> * p_history) const;

	/**
	 * Calculate the min and max value of a variable in the current time frame.
	 * Basically the same as calculateMinMaxAndRangeForVariable() but without range and
	 * without having the correct variable iterator available.
	 * @param minValue Reference to min value variable, result will be stored in it
	 * @param maxValue Reference to max value variable, result will be stored in it
	 * @param key the unique key of the variable
	 * @return returns WE_SUCCESS if min and max could be calculated for the currently displayed values
	 * 	returns WE_OUT_OF_BOUNDS_ERROR if min or max could not be determined
	 * 	returns WE_NOT_FOUND_ERROR if there are no history entries for variable specified by key
	 */
	int calculateMinMax(int &minValue, int &maxValue, unsigned int key) const;

	/**
	 * Small helper function to calculate the range between two values.
	 * Calculates the correct range also for negative values.
	 * @param min minimum value of the range
	 * @param max maximum value of the range
	 * @return the calculated range
	 */
	static int calculateRange(int min, int max);

	void drawSimplifiedPolygon(WEPainter * p_painter, std::vector<int> & p_poly, simplifyAlgorithmType p_simplifyAlgorithm);

	void updateSimplifiedPolygonNodes(unsigned int index, WEControllerNode * controller, std::vector<int> & p_poly, int lineWidth, WEColor color, simplifyAlgorithmType p_simplifyAlgorithm);

	static void addPointToPolygon(std::vector<int> & p_poly, int x, int y, unsigned int & p_key, WEVariableGraphOptions & options);

	static unsigned int findBestStartInHistory(std::deque<WEGraphData> & deque, const unsigned long long & start);

	unsigned long long xOffsetTime;	//!< The X axis offset time as a timestamp in milliseconds from 1970-01-01

	unsigned int xAchsisTime;	//!< The time range of the X axis in seconds
	unsigned int xAchsisOffset;	//!< The offset time of the X axis in seconds

	int axisLineWidth;		//!< The width of the axis lines
	int gridLineWidth;		//!< The width of the grid lines
	int cursorPosition;		//!< The cursor position in pixels from the left side of the graph
	int mySpecialY;			//!< A internal helper variable for calculating the value lines

	int refreshCycleTimeMS{942};	//!< time in milliseconds when the widget shall refresh its content

	std::map<unsigned int, std::deque<WEGraphData> > * valueHistory;	//!< Pointer to the history of all variables drawn in this widget
	std::map<unsigned int, WEVariableGraphOptions> * variableOptions;	//!< Pointer to the settings for each variable (see WEVariableGraphOptions above)

	YAxisPosition yAxisPosition;	//!< The position of the Y axis (left, right, none)
	XAxisLabelOptions xAxisLabels;	//!< The X axis label options (full time stamps, relative time stamps, none)
	GridLineCount numberOfGridLines;	//!< The number of grid lines that shall be drawn (1 or 3)

	bool drawGridLines;		//!< Flag for drawing the grid lines

	WERect innerBoundingBox;	//!< Rectangle defining the inner bounding box in which the grid, cursor and value lines are drawn

	WEColor axisLineColor;		//!< The color of the axis lines
	WEColor gridLineColor;		//!< The color of the grid lines

	unsigned short axisLinePattern;	//!< The line pattern for the axis
	unsigned short gridLinePattern;	//!< The line pattern for the grid lines

	//the following polygons are 2 dimensional vectors (e.g. poly[0] = x1, poly[1] = y1, poly[2] = x2, poly[3] = y2, etc.)
	std::vector<int> simplyficationResult;	//!< The simplyfied polygon that shall be drawn (keep it as member so we don't have to do memory allocation at each paint event)
	std::vector<int> poly;	//!< The original polygon (keep it as member so we don't have to do memory allocation at each paint event)

	WELineNode * axisLineNodes[2] = {nullptr};		//!<Nodes of the Widget which are used to draw the axis
	WELineNode * gridLineNodes[9] = {nullptr};		//!<Nodes of the Widget which are used to draw the grid
	std::vector<WELineNode *> polygonLineNodes;		//!<Nodes of the Widget which are used to draw the polygon
	std::vector<std::vector<WETextNode *>> labelTextNodes;	//!<Nodes of the Widget which are used to draw the labels
	WETextNode * xLabelTextNodes[5] = {nullptr};		//!<Nodes of the Widget which are used to draw the x labels
	WELineNode * cursorLineNode;				//!<Node of the Widget which is used to draw the cursor
	WERectangleNode * cursorRectangleNode;			//!<Node of the Widget which is used to draw the cursor
	std::vector<WETextNode *> cursorTextNodes;		//!<Nodes of the Widget which are used to draw the cursor
};
}


#endif /* WE_GRAPH_WIDGET_H_ */
