/**
 *	@file we_image_widget.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This files defines a image widget
 *
 *	@see WEWidget
 *
 */

#ifndef IMAGE_WIDGET_H_
#define IMAGE_WIDGET_H_

//Included for data Members and Inherited classes
#include "we_widget.h"
#include "../../painter/qt/qt_we_painter.h"

namespace WE
{

class WEImageNode;

enum gifStates
{
	WE_GIF_STATE_STOP,
	WE_GIF_STATE_START,
	WE_GIF_STATE_PAUSE
};

// ==============================================================================================
/**
 *	@class WEImageWidget
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@brief This widget provides image display.
 *
 *	@namespace WE
 *
 *	This class includes the functions, which allows
 *	the user to display the image widgets at the given location. It inherits
 *	from WEWidget, which inherits from WEPaintListener to draw its line in the paint() event.
*/
class WEImage;

class WEImageWidget : public WEWidget
{
	Q_OBJECT
	//Q_PROPERTY(QString imagePath READ getImagePathAsQstring)
	Q_PROPERTY(bool useLocalImage READ getUseLocalImage)
	Q_PROPERTY(int gifState READ getGIFState)
	Q_PROPERTY(int gifFrameNo READ getGIFFrameNo)
	Q_PROPERTY(bool gifLoop READ getGIFLoops)
	Q_PROPERTY(int gifSpeed READ getGIFSpeed)
	Q_PROPERTY(double rotateDegree READ getRotation)

public:
	/**
	 *	@fn WEImageWidget(WEWidget *p_parent, void * p_controller = nullptr)
	 *
	 *	@brief Constructs a WEImageWidget.This widget can be created as
	 *	a separate window by specifying the parent as nullptr or it can be included
	 *	as a child widget of the parent.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_parent If parent IS nullptr then this widget becomes a window with the same properties.
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	*/
	explicit WEImageWidget(WEWidget * p_parent, void * p_controller = nullptr);


	/**
	 *	@fn ~WEImageWidget()
	 *
	 *	@brief Destructor of the widget.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	*/
	virtual ~WEImageWidget();


	//int setImagePath(std::string p_imagePath);


	void paint(WEPainter* p_painter) override;


	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData *p_data) override;


	/**
	 *	@fn int useDisplayImage(WEImage * p_image)
	 *
	 *	@brief Use image provides
	 *
	 *	@param[in] pointer to image
	 *
	 *	@returns WE_error code
	 *
	 *
	 *	This method will discard any image used before and use "p_image" instead.
	 *	"p_image" will NOT be copied or deleted - it is the responsibility of the
	 *	caller to maintain this image and to keep it valid for longer than this image
	 *	widget lives or uses this image.
	 *
	 *	@see int createDisplayImage(const std::string &)
	 */
	int useDisplayImage(WEImage * p_image);


	/**
	 *	@fn int createDisplayImage(const std::string & p_strFile)
	 *
	 *	@brief creates image
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_strFile file + fullpath to image
	 *
	 *	@returns WE_error code
	 *
	 *	This method will discard any image used before and use "p_image" instead.
	 *
	 *	@see int useDisplayImage(WEImage *)
	 */
	int createDisplayImage(const std::string & p_strFile);


	/**
	 *	@fn setSize(int p_width, int p_height)
	 *
	 *	@brief Resize the widget including the image it shows.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 * 	@param[in] p_width width
	 * 	@param[in] p_height height
	 *
	 *
	 */
	int setSize(int p_width, int p_height) override;


	/**
	 *	@fn setTransparentColorIndex(bool p_bTransparent, int p_colorIndex)
	 *
	 *	@brief set the color index for transparency
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_bTransparent create = true or false for reset.
	 *	@param[in] p_colorIndex color index works only for images max indexed8
	 *
	 */
	int setTransparentColorIndex(bool p_bTransparent, int p_colorIndex);

	/**
	 *	@fn int createTransparent(bool p_transparent, WEColor & p_color)
	 *
	 *
	 *	@brief Set the color to .
	 *
	 *	@param p_transparent "true" if pixels of p_color should be transparent, "false" if all pixels should be drawn
	 *	@param p_color Color which should be transparent if p_transparent is true
	 *
	 *	@returns WE::WE_SUCCESS in case of success
	 *	@returns WE::WE_GENERAL_ERROR otherwise
	 *
	 *
	 *	ISO 11783-6 defines that there could be a color that's transparent in an image.
	 *	This method is used to make pixels of a certain color transparent or show all pixels.
	 */
	int createTransparent(bool p_transparent, const WEColor & p_color);


	/**
	 *	@fn setRotate(double p_degree)
	 *
	 *	@brief set the rotate angle of the image
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 * 	@param[in] p_degree angle for rotating in degrees for images
	 *
	*/
	int setRotate(double p_degree);


	WEImage * getDisplayImage(void) const
	{
		return this->displayImage;
	}

	int getGIFState(void) const;
	void setGIFState(int p_state);

	int getGIFFrameNo(void) const;
	void setGIFFrameNo(int p_number);

	bool getGIFLoops(void) const;
	void setGIFLoops(bool p_loops);

	int getGIFSpeed(void) const;
	void setGIFSpeed(int p_percent);

	bool getUseLocalImage(void) const
	{
		return this->useLocalImage;
	}

	double getRotation(void) const
	{
		return this->rotateDegree;
	}

private:
	/**
	 *	@fn init()
	 *
	 *	@brief Initialize the member variables of the Widget
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	*/
	void init();

	void release();


	double rotateDegree;
	WEImage * displayImage; 	//!< pointer to WEImage

	int gifState;
	int gifFrameNo;
	int gifSpeed;

	bool gifLoop;
	bool useLocalImage;		//!< true: displayImage created here, false: displayImage provided externally

	WEColor transparentColor;	//!< Pixels of this color will be transparent if this->m_isTransparent is true

	WEImageNode * imageNode;	//!<Node of the Widget which is used to draw the image
};
}
#endif /* IMAGE_WIDGET_H_ */
