/**
 *	@file we_int_widget.h
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file defines an integer widget to be used as base class for widgets displaying an int value.
 *
 *	@see WEWidget
 *
 */
#ifndef WE_INT_WIDGET_H_
#define WE_INT_WIDGET_H_


#include "we_widget.h"

namespace WE
{


class WEIntWidget : public WEWidget
{
	Q_OBJECT
	Q_PROPERTY(int value READ getValue WRITE setValue)
	Q_PROPERTY(int absMinValue READ getAbsMin WRITE setAbsMin)
	Q_PROPERTY(int absMaxValue READ getAbsMax WRITE setAbsMax)
	Q_PROPERTY(int minValue READ getMin WRITE setMin)
	Q_PROPERTY(int maxValue READ getMax WRITE setMax)
	Q_PROPERTY(int targetValue READ getTargetValue WRITE setTargetValue)
	Q_PROPERTY(bool drawTargetLine READ getDrawTargetLine WRITE setDrawTargetLine)
	Q_PROPERTY(double offset READ getOffset)
	Q_PROPERTY(double offset2 READ getOffset2)
	Q_PROPERTY(double scale READ getScale)
	Q_PROPERTY(int numberOfDecimals READ getNumberOfDecimals)

protected:
	/**
	 *	@fn WEIntWidget(int p_typeID, WEWidget * p_parent, void * p_controller = nullptr)
	 *
	 *	@brief constructor of WEIntWidget
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_typeID: type ID for the widget to create as defined in WE::widgetTypeID
	 *	@param[in] p_parent: pointer to parent object, if nullptr then top widget
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 *
	 *	@returns nothing
	 */
	WEIntWidget(int p_typeID, WEWidget * p_parent, void * p_controller = nullptr);

public:

	/**
	 *	@fn ~WEIntWidget()
	 *
	 *	@brief deconstructor
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns nothing
	 */
	virtual ~WEIntWidget();

	/**
	 *	@fn int getValue()
	 *
	 *	@brief This method is used to get the current value of the widget
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns int value
	 */
	int getValue()
	{
		return this->value;
	};


	/**
	 *	@fn int isValueValid(int p_value)
	 *
	 *	@brief This method is used to check if a value is in the valid range.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns WE_ERROR code:
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if value too high (> absMax) or too low (< absMin)
	 *
	 *	This method checks whether p_value is in the valid range for this->value, i.e. if
	 *	absMinValue <= p_value <= absMaxValue.
	 *	This method can be used e.g. to check if a target value is inside the valid range
	 *	for value.
	 *
	 */
	virtual int isValueValid(int p_value) const;


	/**
	 *	@fn int setValue(int p_value)
	 *
	 *	@brief This method is used set the value of the IntWidget
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_value int value to be set
	 *	@param[in] p_updateScrollbar optional parameter that indicates if a possible existing scrollbar (list, table) shall be updated or not (value change could have been triggered by scrollbar itself)
	 *
	 *	@returns WE_ERROR code:
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if value too high (> absMax) or too low (< absMin)
	 *
	 *	The implementation of this method provided by WEIntWidget will check the value
	 *	against absolute min and absolute max values and reject it if the value does not
	 *	fit into the interval defined by these two values.
	 *	This behavior is appropriate for most widgets as they shall only accept values that
	 *	they can display.
	 *	For e.g. numeric widget this method has to be overwritten because a numeric widget
	 *	can generally display every value and absMin and absMax values are only relevant if
	 *	the value is edited. But the behavior for this case has to be covered by the widget's
	 *	owner (for PClient application the owner of a widget is the DDO).
	 *
	 */
	virtual int setValue(int p_value, bool p_updateScrollbar = true);


	/**
	 *	@fn int setAbsMinMax(int p_min, int p_max)
	 *	@fn int setAbsMin(int p_min)
	 *	@fn 	int setAbsMax(int p_max)
	 *
	 *	@brief These methods set the absolute min and max values
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_min absolute min value
	 *	@param[in] p_max absolute max value
	 *
	 *	@returns WE_ERROR code:
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if parameters would set absMinValue > absMaxValue
	 *
	 *	Absolute minimum and maximum value define the range of values an int widget can display.
	 *	If one tries to set its value out of this range an error will be returned.
	 *
	 *	@warning If the current value is outside of the new range it will be set to the closest limit.
	 *
	 */
	virtual int setAbsMinMax(int p_min, int p_max);
	virtual int setAbsMin(int p_min);
	virtual int setAbsMax(int p_max);
	int getAbsMin(void)
	{
		return this->absMinValue;
	}
	int getAbsMax(void)
	{
		return this->absMaxValue;
	}

	/**
	 *	@fn int setMinMax(int p_min, int p_max)
	 *	@fn int setMin(int p_min)
	 *	@fn int setMax(int p_max)
	 *
	 *	@brief Set minimum and maximum values for color change.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_min minimum value
	 *	@param[in] p_max maximum value
	 *
	 *	@returns WE_ERROR code:
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if parameters would set minValue > maxValue
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if parameters would set minValue or maxValue outside of valid range
	 *
	 *	Minimum and maximum values are used to change the color (or general appearance) of the
	 *	widget in order to show the operator that a critical value is reached.
	 *
	 *
	 *	@see setAbsMinMax(int p_min, int p_max)
	 *	@see getCurrentValueRange()
	 *	@see getCurrentValueColor()
	 */
	virtual int setMinMax(int p_min, int p_max);
	virtual int setMin(int p_min);
	virtual int setMax(int p_max);
	virtual int getMin();
	virtual int getMax();

	/**
	 *	@fn setTargetValue(int p_targetValue)
	 *
	 *	@brief This method is used set the value for the target line of the graph
	 *
	 *	@param[in] p_targetValue int value to be set
	 *
	 *	@returns WE_ERROR code
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if target value too high (> absMax) or too low (< absMin)
	 *
	 *	@warning check the parameter against the this->absoluteMin and this->absoluteMax
	 *	@warning this member variables should be set before
	 *
	 *
	 */
	virtual int setTargetValue(int p_targetValue);
	virtual int getTargetValue();

	/**
	 *	@fn int setDrawTargetLine(bool p_drawTargetLine)
	 *
	 *	@brief set drawing of target line
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] p_drawTargetLine sets if the target line is drawn or not
	 *
	 *  @returns WE_ERROR code
	 */
	virtual int setDrawTargetLine(bool p_drawTargetLine);
	virtual bool getDrawTargetLine();

	/**
	 *	@fn int setColorMinMax(WEColor & p_minColor, WEColor & p_normalColor, WEColor & p_maxColor)
	 *	@fn int setColorMin(WEColor & p_minColor)
	 *	@fn int setColorNormal(WEColor & p_normalColor)
	 *	@fn int setColorMax(WEColor & p_maxColor)
	 *
	 *	@brief set the color for below min, normal and above max value
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_maxColor color used for filling when value is above max
	 *	@param[in] p_normalColor color used for filling when value is between min and max
	 *	@param[in] p_minColor color used for filling when value is below min
	 *
	 *	@warning It could be required that a widget provides additional methods to configure
	 *	@warning what is supposed to change if value is below min or above max value.
	 *
	 *	@returns WE_ERROR code
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if one of the parameters is nullptr (nothing is changed then)
	 *
	 *	@see getCurrentValueRange()
	 *
	 */
	virtual int setColorMinMax(const WEColor & p_minColor, const WEColor & p_normalColor, const WEColor & p_maxColor);
	virtual int setColorMin(const WEColor &p_minColor);
	virtual int setColorNormal(const WEColor &p_normalColor);
	virtual int setColorMax(const WEColor &p_maxColor);


	/**
	 *	@fn int getCurrentValueColor(WEColor & p_Color)
	 *
	 *	@brief get the color to use depending on current value
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_color reference of color whose RGB values should be set
	 *
	 *
	 *	@returns WE_ERROR code
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	@see getCurrentValueRange()
	 *
	 */
	int getCurrentValueColor(WEColor & p_color);


	/**
	 *	@fn int getCurrentValueRange()
	 *
	 *	@brief get an index to use depending on current value
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] none
	 *
	 *
	 *	@returns 0 if absMinValue <= value < minValue
	 *	@returns 1 if minValue <= value <= maxValue
	 *	@returns 2 if maxValue < value <= absMaxValue
	 *
	 *	This method can be used in order to select a color, fontAttribute or other property
	 *	to be changed with the value of the widget from an array.
	 *
	 *
	 *	@see getCurrentValueColor()
	 *
	 */
	int getCurrentValueRange();

	virtual int setOffset(double p_offset);
	virtual int setOffset2(double p_offset2);
	virtual int setScale(double p_scale);
	virtual int setNumberOfDecimals(int p_noOfDecimals);

	double getOffset(void);
	double getOffset2(void);
	int getNumberOfDecimals(void);
	double getScale(void);

protected:
	/**
	 *	@fn int init()
	 *
	 *	@brief initialize the members
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns WE_ERROR code
	 */
	int init();

	/**
	 *	@fn void release()
	 *
	 *	@brief release the members
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns WE_ERROR code
	 */
	int release();
	// ======== member variables ===============================================================

	int value;		//!< value to be displayed

	int absMinValue;	//!< absolute min allowed for value
	int absMaxValue;	//!< absolute max allowed for value
	int minValue;		//!< change color if value is below this
	int maxValue;		//!< change color if value is above this
	int targetValue;	//!< place of the target if it should be drawn (applicable only for Linear and Arched bargraph)
	bool drawTargetLine; 	//!< Draw target line true/false

	WEColor * myColors[3];	//!< keep the three colors [&colorBelowMin , &colorNormal, &colorAboveMax]

	WEColor colorBelowMin;	//!< color to show if value is below min value
	WEColor colorNormal;	//!< color to show if value is above min and below max value
	WEColor colorAboveMax;	//!< color to show if value is above max value

	double offset;
	double offset2;
	double scale;
	int numberOfDecimals;		//!<Number of the decimal places

private:
	// Since this is a base class there are no private definitions
};	// class
}	// namespace

#endif /* WE_INT_WIDGET_H_ */
