/**
 *	@file we_linearbargraph_widget.h
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This files defines a linear bargraph widget
 *
 *	@see WEWidget
 *
 *
 */
#ifndef WE_LINEARBARGRAPH_WIDGET_H_
#define WE_LINEARBARGRAPH_WIDGET_H_


//Please do not review as this is only a sample/dummy widget
#include "../../include/we_datatypes_gui.h"
#include "we_int_widget.h"

namespace WE
{

class WEControllerNode;
class WEGeometryNode;
class WERectangleNode;
class WEImage;
class WEImageNode;
class WEPainter;
class WEWidget;

class WELinearBargraphWidget : public WEIntWidget
{
	Q_OBJECT
	Q_PROPERTY(unsigned short numberOfTicks READ getNumberOfTicks)
	Q_PROPERTY(bool drawTicks READ getDrawTicks)
	Q_PROPERTY(int bargraphDirection READ getBargraphDirection)
	Q_PROPERTY(bool fillMode READ getFillMode)
	Q_PROPERTY(int direction READ getDirection)
	Q_PROPERTY(bool needleImageSet READ isNeedleImageSet)
	Q_PROPERTY(bool drawEnclosingBorder READ getDrawEnclosingBorder)
public:

	/**
	 *	@fn WELinearBargraphWidget(WEWidget *parent, void * p_controller = nullptr)
	 *
	 *	@brief constructor of the meter
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] parent pointer to parent object, if nullptr then top widget
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 *
	 *	@returns nothing
	 */
	explicit WELinearBargraphWidget(WEWidget *p_parent, void * p_controller = nullptr);


	/**
	 *	@fn ~WELinearBargraphWidget()
	 *
	 *	@brief deconstructor
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns nothing
	 */
	virtual ~WELinearBargraphWidget();


	/**
	 *	@fn void paint(WEPainter* painter)
	 *
	 *	@brief draws the meter
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] painter pointer to painter object
	 *
	 *	@returns nothing
	 */
	void paint(WEPainter* p_painter) override;


	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData *p_data) override;


	/**
	 *	@fn int drawTicks(bool p_drawTicks)
	 *
	 *	@brief set drawing of ticks to true or false
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_drawTicks sets if the ticks are drawn or not
	 *
	 *	@returns WE_ERROR code
	 */
	int setDrawTicks(bool p_drawTicks);


	/**
	 *	@fn int setNumberOfTicks(uint16 p_tickCount)
	 *
	 *	@brief set the number of ticks to be drawn
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_tickCount number of ticks to be drawn
	 *
	 *	@returns WE_ERROR code
	 */
	int setNumberOfTicks(std::uint16_t p_tickCount);


	/**
	 *	@fn int setTickColor(WEColor p_tickColor)
	 *
	 *	@brief set the tick color
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_tickColor color used for drawing the ticks
	 *
	 *	@returns WE_ERROR code
	 */
	int setTickColor(WEColor p_tickColor);


	/**
	 *	@fn int setTargetLineColor(WEColor p_targetLineColor)
	 *
	 *	@brief set the color of the target line
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_targetLineColor color used for drawing the target line
	 *
	 *	@returns WE_ERROR code
	 */
	int setTargetLineColor(WEColor p_targetLineColor);


	/**
	 *	@fn int setAxisDirection(WEAxisDirectionType p_direction)
	 *
	 *	@brief set axis direction
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_direction true = horizontal / false = negative
	 *
	 *	@returns WE_ERROR code
	 *
	 *	@see WEAxisDirectionType
	 */
	int setAxisDirection(WEAxisDirectionType p_direction);

	/**
	 *	@fn int setDirection(WEDirectionType p_direction)
	 *
	 *	@brief set positive or negative direction of the bargraph
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_direction
	 *
	 *	@returns WE_ERROR code
	 *
	 *	@see WEDirectionType
	 */
	int setDirection(WEDirectionType p_direction);

	/**
	 *	@fn int setFillMode(bool p_fillMode)
	 *
	 *	@brief set fill mode on or off
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_fillMode true / false = only thin line
	 *
	 *	@returns WE_ERROR code
	 */
	int setFillMode(bool p_fillMode);

	/**
	 *	@fn setEnclosingBorderColor(WEColor p_enclosingBorderColor)
	 *
	 *	@brief set the Enclosing Border color
	 *
	 *	@author satheeshkumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *	@param[in] p_enclosingBorderColor color used for drawing the Enclosing Border
	 *
	 *	@returns WE_ERROR code
	 */
	int setEnclosingBorderColor(WEColor p_enclosingBorderColor);

	/**
	 *	@fn int setDrawEnclosingBorder(bool p_drawEnclosingBorder)
	 *
	 *	@brief set drawing of Enclosing Border to true or false
	 *
	 *	@author satheeshkumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *	@param[in] p_drawEnclosingBorder sets if the Enclosing Border is drawn or not
	 *
	 *	@returns WE_ERROR code
	 */
	int setDrawEnclosingBorder(bool p_drawEnclosingBorder);


	int createNeedleImage(const std::string & p_strFile);

	int setNeedleImage(WEImage* p_image);



	std::uint16_t getNumberOfTicks(void) const
	{
		return this->numberOfTicks;
	}

	bool getDrawTicks(void) const
	{
		return this->drawTicks;
	}

	int getBargraphDirection(void) const
	{
		return this->bargraphDirection;
	}

	bool getFillMode(void) const
	{
		return this->fillMode;
	}

	int getDirection(void) const
	{
		return this->direction;
	}

	bool isNeedleImageSet(void) const
	{
		return this->needleImageSet;
	}

	bool getDrawEnclosingBorder(void) const
	{
		return this->drawEnclosingBorder;
	}

	const WEColor& getEnclosingBorderColor() const;
	const WEColor& getTargetLineColor() const;
	const WEColor& getTickColor() const;

protected:
	virtual void paintBackgroundGraphics(WEPainter* p_painter) override;


private:

	/**
	 *	@fn void init()
	 *
	 *	@brief initialize the members
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns void
	 */
	void init();


	/**
	 *	@fn void drawValueAndTicks(WEPainter * p_painter)
	 *
	 *
	 *	@brief Draw value, target line and ticks for this bargraph.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_painter Painter to be used for drawing of this bargraph.
	 *
	 *	@returns nothing
	 *
	 *	This draws
	 *	- the value as line of filling,
	 *	- the target line (if desired) and
	 *	- ticks (if desired).
	 *	Color of target line and ticks is inverted if they would become invisible on top
	 *	of a value filling.
	 *
	 */
	void drawValueAndTicks(WEPainter * p_painter);

	void updateValueAndTicksNodes(WEControllerNode *p_controller);


	/**
	 *	@fn void drawLinearEnclosingBorder(WEPainter* painter)
	 *
	 *	@brief Draw the Enclosing Border over the Linear Widget
	 *
	 *	@author satheeshkumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *
	 *	@param[in] p_painter The painter
	 *
	 *	@see REFERENCE WE::WEArchedBarGrapgWidget::drawBorder()
	 *
	 */
	void drawLinearEnclosingBorder(WEPainter* p_painter);

	void updateLinearEnclosingBorderNodes(WEControllerNode *p_controller);


	void drawNeedleImage(WEPainter* p_painter);

	void updateNeedleImageNodes(WEControllerNode *p_controller);


	WEImage* needleImage;		//!< stores pointer to needle image.
	bool fillMode;
	bool needleImageSet;		//!< Please comment why (nullptr != this->needleImage) isn't sufficient.
	bool drawEnclosingBorder;	//!< Enable/Disable Enclosing Border
	bool drawTicks; //!< Draw ticks true/false
	WEDirectionType bargraphDirection;
	WEAxisDirectionType direction;
	WEColor targetLineColor; //!< color of ticks
	WEColor tickColor; //!< color of ticks
	WEColor enclosingBorderColor; //!< Enclosing Border color
	std::uint16_t numberOfTicks; //!< Number of ticks to be drawn


	WEImageNode * backgroundPixmapImageNode;
	WEImageNode * needleImageNode;
	WERectangleNode * linearEnclosingBorderRectangleNode;
	WERectangleNode * valueRectangleNode;
	WEGeometryNode * valueGeometryNode;
	WEGeometryNode * targetLineGeometryNode;
	WEGeometryNode * tickGeometryNode;
};
}
#endif /*WE_LINEARBARGRAPH_WIDGET_H_*/
