/**
 *	@file we_list_widget.h
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This files defines a list widget
 *
 *	@see WEWidget
 *
 *
 *	The approach with the items container is widely shared with the button widget - we decided against a common
 *	base widget for this because both do totally different things with the children in the container.
 *
 *	To allow synchronization between these two widgets methods of both were reordered in the same logical sequence:
 *	- constructor, destructor, initialization
 *	- child addition, removal, get, set
 *	- paint
 *	- value related
 *	Header and source file are kept in the same order (with the exception that protected and private methods are
 *	moved to the bottom in the header file).
 *
 *	If only one item is shown the list widget shows
 *	* the item belonging to the value in case the item is valid (not a NULL-ID and not rendered invisible via visibility)
 *	* nothing if value is below max value and the item is invalid
 *	* the last item if the value is equal to or above max value (this is a limitation made by AppIntDDO already)
 *		* if this actually shows something depends on the validity of the last item
 *
 *	If more than one item is shown the list widget
 *	* shows as many items as configured if there are enough valid ones
 *	* shows all valid items in case there are less than items configured for the list
 *		* note that this will result in reduced height of the widget
 *	* current value/selected item is identified by a focus frame using the 'ENABLED_WITH_FOCUS' state of the item
 *		* if the item should not have such property (e.g. because it's a container) the respective setting of the list is used
 *	* if an invalid item is selected, no focus frame is shown
 *	* the last item is shown if the value is equal to or above max value (this is a limitation made by AppIntDDO already)
 *		* if this actually shows something depends on the validity of the last item
 *		* add an invalid item at the end of the list if nothing should be shown for values above the list's range
 *
 *	@see WEButtonWidget
 *	@see AppIntDDO::limitToValidRange(int)
 *
 */

#ifndef LIST_WIDGET_H_
#define LIST_WIDGET_H_

//Included for data Members and Inherited classes
#include <list>
#include "../../include/we_datatypes_gui.h"
//#include "../../common/include/we_list_container.h"

#include "we_scrollbar_widget.h"

namespace WE
{

class WEContainer;


class WEListWidget : public WEScrollbarWidget
{
	Q_OBJECT
	Q_PROPERTY(unsigned int itemCount READ getItemCount)
	Q_PROPERTY(unsigned int itemsVisNormal READ getItemsVisNormal)
	Q_PROPERTY(unsigned int itemsVisEdit READ getItemsVisEdit)
	Q_PROPERTY(unsigned int itemHeight READ getItemHeight)
	Q_PROPERTY(int listMode READ getListMode)
	Q_PROPERTY(unsigned  startIndex READ getStartIndex)
	Q_PROPERTY(unsigned  drawItems READ getDrawItems)
	Q_PROPERTY(unsigned  itemDrawIndex READ getItemDrawIndex)
	Q_PROPERTY(int oldValue READ getOldValue)
	Q_PROPERTY(unsigned int focusFrameBorderWidth READ getFocusFrameBorderWidth)
	Q_PROPERTY(int highlightedPosition READ getHighlightedPosition)

public:
	/**
	 *	@fn WEListWidget(WEWidget * p_parent, void * p_controller = nullptr, int p_typeID = WIDGET_TYPE_ID_LIST)
	 *
	 *	@brief Constructs a WEListWidget.This widget can be created as
	 *	a separate window by specifying the parent as nullptr or it can be included
	 *	as a child widget of the parent.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_parent If parent IS nullptr then this widget becomes a window with the same properties.
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 */
	explicit WEListWidget(WEWidget * p_parent, void * p_controller = nullptr, int p_typeID = WIDGET_TYPE_ID_LIST);

	/**
	 *	@fn ~WEListWidget()
	 *
	 *	@brief Destructor of the widget.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 */
	virtual ~WEListWidget();


	/**
	 *	@brief Allow switching from input to output list and vice versa.
	 *
	 *	@returns WE_SUCCESS if p_typeID is accepted, WE_GENERAL_ERROR otherwise.
	 */
	int setTypeID(int p_typeID);


	/**
	 *	@fn add(WEWidget * p_widget)
	 *
	 *	@brief add a Item to the List and as a child to the widget
	 *
	 *	@see WEWidget
	 *
	 *	Overwrite method from WEWIdget.
	 */
	int add(WEWidget * p_widget) override;


	/**
	 *	@fn void removeChild(WEWidget * p_child)
	 *
	 *	@brief remove child from widget and delete child from button. Overwritten from WEWidget
	 *
	 *	@see WEWidget
	 *
	 *	Overwrite method from WEWIdget.
	 */
	void removeChild(WEWidget * p_child) override;


	/**
	 *	@fn bool hasChildren()
	 *
	 *	@brief indicates if the widget has children
	 *
	 *	@return Whether there are items or not
	 *
	 *	This method is overwritten to cover the internal "items" container.
	 *	Originally it always returned "false" because it's only called from "isInputWidget()"
	 *	and an input list never has input children.
	 *	Now to make output lists work we anyways had to overwrite isInputWidget() itself here.
	 *
	 *	Overwrite method from WEWidget.
	 */
	bool hasChildren() const override;


	/**
	 *	@fn bool isInputWidget( void )
	 *
	 *	@brief Return if navigation should take this list into account or not.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none.
	 *
	 *	@returns true if this is an output list which currently shows an input object.
	 *	@returns Returns false otherwise
	 *
	 *	If this is an input list it is probably an input itself.
	 *	An output list could only be considered an input if the child currently shown is one.
	 *
	 *	Overwrite method from WEWidget.
	 */
	bool isInputWidget( void ) const override;


	/**
	 *	@fn WEWidget * getFirstChild()
	 *
	 *	@brief Skip the container we use here and return its getFirstChild()
	 *
	 *	Overwrite method from WEWidget.
	 */
	WEWidget * getFirstChild() const override;


	/**
	 *	@fn WEWidget * getLastChild()
	 *
	 *	@brief Skip the container we use here and return its getLastChild()
	 *
	 *	Overwrite method from WEWidget.
	 */
	WEWidget * getLastChild() const override;


	/**
	 *	@fn WEWidget * getFirstInputChild()
	 *
	 *	@brief Get the focussed child if this is an output list and the child is an input. Otherwise return nullptr.
	 *
	 *	Overwrite method from WEWidget.
	 */
	WEWidget * getFirstInputChild() const override;


	/**
	 *	@fn WEWidget * getLastInputChild()
	 *
	 *	@brief Get the focussed child if this is an output list and the child is an input. Otherwise return nullptr.
	 *
	 *	Overwrite method from WEWidget.
	 */
	WEWidget * getLastInputChild() const override;


	/**
	 *	@fn WEWidget * getChildren(void)
	 *
	 *	@brief Skip the container we use here and return its getChildren(void)
	 *
	 *	Overwrite method from WEWidget.
	 */
	childrenBiMap * getChildren(void) const override;


	/**
	 *	@fn setSize(int p_width, int p_height)
	 *
	 *	@brief Sets the size of the Widget.
	 *
	 *	@param[in] p_width Width to be set for the widget.
	 *	@param[in] p_height Height to be set for the widget.
	 *
	 *	We have to set the size of the background image and the items container in addition to what WEWidget would do.
	 *
	 *	Overwrite method from WEWidget.
	 */
	int setSize(int p_width, int p_height) override;


	/**
	 *	@fn setItemHeight(int p_height)
	 *
	 *	@brief sets the height in Pixel for each and every item in the list.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_height sets height in pixel for each Item
	 *
	 *	@returns WE_ERROR code
	 */
	int setItemHeight(int p_height);


	/**
	 *	@fn setItemsVisible(int p_normal, int p_edit)
	 *
	 *	@brief sets the Item visible count
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_normal How many items shall be visible normally?
	 *	@param[in] p_edit How many items shall be visible in edit mode?
	 *
	 *	@returns WE_ERROR code.
	 *
	 *	If you set a param to -1 then the param is not overwritten.
	 *
	 *	Make sure "setItemHeight()" is called BEFORE because this resizes the list accordingly.
	 *
	 *	@see int setItemHeight(int)
	 */
	int setItemsVisible(int p_normal, int p_edit);


	/**
	 *	@fn setNumberOfListItems(int p_itemCount)
	 *
	 *	@brief sets the number of possible list items
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_itemCount item count
	 *
	 *	@returns WE_ERROR code.
	 */
	int setNumberOfListItems(int p_itemCount);


	/**
	 *	@fn paint(WEPainter* p_painter)
	 *
	 *	@brief paints the list
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_painter Pointer to painter obj
	 *
	 *	@returns void
	 */
	void paint(WEPainter* p_painter) override;


	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData *p_data) override;


	/**
	 *	@fn void repaint()
	 *
	 *	@brief repaint the list and all of its items.
	 *
	 *	@param none
	 *
	 *	We have to set the repaint flag for all the items in addition to what WEWidget would do.
	 *
	 *	Overwrite method from WEWidget.
	 */
	void repaint() override;


	/**
	 *	@fn setMode(WEListModeType p_mode)
	 *
	 *	@brief Switch between editing and normal display mode.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_mode mode for list
	 *
	 *	@returns WE_ERROR code.
	 *
	 *	Switch between editing and normal display mode.
	 *
	 *	Actually we just make sure to resize this widget so that the scollbar gets clipped or gets visible.
	 *	The first time this also triggers creation of the scrollbar.
	 *
	 *	@Warning: Problem here is that additionally this gets another parent for editing (to make sure scrollbar really is visible)
	 *	and this is maintained in DDO - so make sure that this method is only one step in a longer sequence handling mode change.
	 *	Or alternatively skip the reparenting done by DDO.
	 *
	 *	@see WEListModeType
	 */
	int setMode(WEListModeType p_mode);

	/**
	 *	@fn virtual int setBorderState(int p_state)
	 *
	 *	@brief This sets all the children to disabled also so that the status really shows on screen.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns nothing
	 *
	 *	@param p_valid Set to false if disabled pattern should be shown.
	 *
	 *	This method changes the state of the list widget.
	 *	It is overwritten here because the list widget itself is typically hidden behind (one of) its children.
	 *	Therefore the overwritten method calls setBorderState() for all the children.
	 *
	 */
	virtual int setBorderState(int p_state) override;


	/**
	 *	@fn int setMinMax(int p_min, int p_max)
	 *	@fn int setMin(int p_min)
	 *	@fn int setMax(int p_max)
	 *
	 *	@brief Set minimum and maximum values for color change.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_min minimum value
	 *	@param[in] p_max maximum value
	 *
	 *	@returns WE_ERROR code:
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if parameters would set minValue > maxValue
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if parameters would set minValue or maxValue outside of valid range
	 *
	 *	Minimum and maximum values are used to change the color (or general appearance) of the
	 *	widget in order to show the operator that a critical value is reached.
	 *	Since this is not really applicable in a list we are less strict here.
	 *
	 *
	 *	@see setAbsMinMax(int p_min, int p_max)
	 *	@see getCurrentValueRange()
	 *	@see getCurrentValueColor()
	 */
	virtual int setMinMax(int p_min, int p_max) override;
	virtual int setMin(int p_min) override;
	virtual int setMax(int p_max) override;


	/**
	 *	@fn void setValueToValid(bool p_valid)
	 *
	 *	@brief This triggers to show the disabled pattern or not.
	 *
	 *	@returns nothing
	 *
	 *	@param p_valid Set to false if disabled pattern should be shown.
	 *
	 *	This method makes sure the disabled pattern is shown on top of this widget.
	 *	This widget overwrites this method to make focus frame show 'invalid' state if required.
	 *
	 */
	virtual void setValueToValid(bool p_valid) override;


	/**
	 *	@fn int isValueValid(int p_value)
	 *
	 *	@brief This method is used to check if a value is in the valid range.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns WE_ERROR code:
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if value too high (>= no. of items) or too low (< 0)
	 *
	 *	This method checks whether p_value is in the valid range for this->value, i.e. if
	 *	it is a valid index into the list, pointing to one of the list items.
	 *	This method can be used e.g. to check if a target value is inside the valid range
	 *	for value.
	 *
	 */
	virtual int isValueValid(int p_value) const override;


	/**
	 *	@fn int setValue(int p_value)
	 *
	 *	@brief Set value and recalculate dependent members.
	 *
	 *	@param p_value Value to be set.
	 *
	 *	@returns WE_SUCCESS If value was accepted
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if value is not accepted by isValueValid().
	 *	@returns WE_SELECT_ERROR if value points to an invisible element while list is in edit mode
	 *	@returns WE_GENERAL_ERROR If p_value points to an invisible value.
	 *
	 *	Changing the value has to recalculate which items should be shown on screen.
	 *
	 *	Overwrite method from WEIntWidget.
	 *
	 *	@see isValueValid()
	 */
	int setValue(int p_value, bool p_updateScrollbar = true) override;


	/**
	 *	@fn int setWidthEditMode(int p_widthEditMode)
	 *
	 *	@brief Set width for the list to used in edit mode.
	 *
	 *	@param p_value width to be set in editing mode.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	Useful for lists which have to expand in width while editing
	 *	e.g; List used to show multiple aux assignments with "++++" DDO.
	 *
	 *	@WARNING: With introduction of WE::WEScrollbarWidget this is NOT set automatically anymore. \
	 *	If not set explicitly width will only be adjusted automatically if list is edited. \
	 *	I.e. this HAS to be set if scrollbar should be shown while list is NOT edited or is an output list.
	 */
	int setWidthEditMode(int p_widthEditMode);

	/**
	 *	@fn int getHighlightedPosition(void)
	 *
	 *	@brief Get line on screen for current value.
	 *
	 *	@param[in] none
	 *
	 *	@returns (this->value - this->startIndex) ==> this->highlightedPosition
	 *
	 *	If more than one item is shown on screen i.e. setItemsVisible() was called with a value
	 *	greater than 1 this is helpful to figure our which line on screen shows the current value.
	 *	This can be used to modify the appearance of the specific child by adding something on this screen position.
	 *
	 */
	int getHighlightedPosition(void) const;

	/**
	 *	@fn int getIndexUnderPosition(int p_x, int p_y)
	 *
	 *	@brief Figure out the list value belonging to the touched item.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_x Global x coordinate of mouse event
	 *	@param p_y Global y coordinate of mouse event
	 *
	 *	@returns The index to be applied to this->visibleItems list to show the touched item.
	 *
	 *	If only one item is shown or coordinate is outside the children area this->itemDrawIndex is returned.
	 *	Otherwise the child drawn at the indicated position is found and the respective value is returned.
	 *
	 */
	virtual int getIndexUnderPosition(int p_x, int p_y) const override;


	/**
	 *	@fn int setVisible(bool p_visible)
	 *
	 *	@brief Set the visibility status of a widget
	 *
	 *	@param[in] p_visible Boolean value that represents the visibility status of the widget
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This method overwrites (and internally calls WE::WEWidget::setVisible(bool)) to cover this'
	 *	widget's internal "items" container.
	 *
	 *	@see show()
	 *	@see WE::WEWidget::setVisible(bool)
	 */
	void setVisible(bool p_visible) override;


	/**
	 *	@brief Forward information about visibility change to (grand*) parent and re-arrange the list
	 *
	 *	This method was introduced for the list to dynamically change if
	 *	visibility of children changes.
	 *
	 *	Currently every call triggers re-arranging the list.
	 *	Maybe we should figure out if p_child really is a list item.
	 */
	virtual void childVisibilityChanged(WEWidget * p_child, bool p_visible) override;


	/**
	 *	@fn void show()
	 *
	 *	@brief Show the widget
	 *
	 *	@param[in] none
	 *
	 *	@returns nothing
	 *
	 *	This method overwrites (and internally calls WE::WEWidget::show()) to cover this'
	 *	widget's internal "items" container.
	 *
	 *	@see WE::WEWidget::show()
	 *	@see setVisible(bool)
	 */
	void show();

	unsigned int getItemCount(void) const
	{
		return this->itemCount;
	}

	unsigned int getItemsVisNormal(void) const
	{
		return this->itemsVisNormal;
	}

	unsigned int getItemsVisEdit(void) const
	{
		return this->itemsVisEdit;
	}

	unsigned int getItemHeight(void) const
	{
		return this->itemHeight;
	}

	int getListMode(void) const
	{
		return this->listMode;
	}

	unsigned getStartIndex(void) const
	{
		return this->startIndex;
	}

	unsigned getDrawItems(void) const
	{
		return this->drawItems;
	}

	unsigned getItemDrawIndex(void) const
	{
		return this->itemDrawIndex;
	}

	int getOldValue(void) const
	{
		return this->oldValue;
	}

	unsigned int getFocusFrameBorderWidth(void) const
	{
		return this->focusFrameBorderWidth;
	}


	void updateSizeWithWidth(int p_width) override;


	unsigned int getNumberOfVisibleItems(void) const override;


	unsigned int getItemAtIndex(int p_index) const override;


	/**
	 *	@brief If name of the list changes, also set name of 'items' container and 'focusFrame'.
	 */
	virtual int setName(const std::string & p_name) override;

private:

	/**
	 *	@fn int createItemsContainer()
	 *
	 *	@brief Create the Items container and focus frame for the list if not already created
	 *
	 *	@param none.
	 *
	 *	@returns WE_SUCCESS
	 *
	 */
	int createItemsContainer();

	/**
	 *	@fn void calculateStartAndEndIndex(void)
	 *
	 *	@brief Calculate which items should be shown on screen.
	 *
	 *	@param none.
	 *
	 *	@returns nothing
	 *
	 */
	void calculateStartAndEndIndex(void);


	/**
	 *	@fn init()
	 *
	 *	@brief Initialize the member variables of the Widget
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 */
	void init();


	/**
	 *	@fn release()
	 *
	 *	@brief Deletes the container we used for the items
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns WE_SUCCESS
	 */
	int release();


	/**
	 *	@fn void resizeItemsContainer( void )
	 *
	 *	@brief Resize items container based on members of the list object.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param None.
	 *
	 *	@returns Nothing
	 *
	 *	This method was created to reduce the amount of copied code.
	 *
	 */
	void resizeItemsContainer( void );


	/**
	 *	@fn unsigned int getIndexForValue(unsigned int p_value)
	 *
	 *	@brief Retrieve index into this->visibleItems that would be shown if this->value would be p_value.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_value Value whose index should be returned.
	 *
	 *	@returns Index into this->visibleItems matching p_value i.e. this->visibleItems.at(p_value) == index
	 *
	 *	This method gets all the visible items from the this->visibleItems vector and returns the index found for "p_value".
	 *	If the value can not be found it points to an invisible or not existing item.
	 *	in this case this->visibleItems.size() is returned.
	 *
	 */
	virtual unsigned int getIndexForValue(unsigned int p_value) const override;


	/**
	 *	@fn WEWidget * getItemForValue(unsigned int p_value)
	 *
	 *	@brief Retrieve widget that would be shown if this->value would be p_value.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_value Value whose widget should be returned.
	 *
	 *	@returns Pointer to the WEWidget belonging to p_value.
	 *	@returns nullptr if widget could not be obtained (doesn't exist yet, beyond the list )
	 *
	 *	This method gets all the children from the "items" container and returns the one found at position "p_value".
	 *
	 */
	WEWidget * getItemForValue(unsigned int p_value) const;


	/**
	 *	@fn void reArrangeChildren(void)
	 *
	 *	@brief Close all gaps where invisible items would show.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none.
	 *
	 *	@returns nothing
	 *
	 *	This method moves up visible children of the "items" container so that gaps are closed.
	 *
	 *	The original version of this approach comes from the LMH branch.
	 *
	 */
	void reArrangeChildren(void);


	/**
	 *	@brief Move all access to focusFrame to this one method.
	 *
	 *	@param p_show false: Hide focusFrame, otherwise: Set parameters and show it.
	 *	@param p_borderFrameSource Take border profile from this, if available (and only if 'p_show')
	 */
	void drawFocusFrame(bool p_show, WEWidget * p_borderFrameSource = nullptr);


	WEContainer *	items;		// container whose position is changed when scrolling through the list
	WEContainer *	focusFrame;	// container to indicate the child with focus if more than one child is shown and the child does not show its border

	unsigned int 	focusFrameBorderWidth;// minimum border width to be used for the focus frame

	unsigned int	itemCount;	// list index runs from 0 to (itemCount - 1)
	unsigned int	itemsVisNormal;
	unsigned int	itemsVisEdit;
	unsigned int	itemHeight;
	unsigned int	startIndex;	//!< index of first item to be shown on screen i.e. an index into this->visibleItems
	unsigned int	drawItems;	// the number of items to be shown on screen in parallel
	unsigned int	itemDrawIndex;	// the item currently drawn i.e. an index into this->visibleItems matching this->value

	int		highlightedPosition;

	int		oldValue;

	WEListModeType	listMode;

	bool hasContainer;	// container is already created?



	std::vector < unsigned int > visibleItems;	//!< list of items to choose from (position in vector reflects position in list, value is the value this list has to get to show the respective item)
							// is set in 'reArrangeChildren()' which skips all the invisible/NULL-ID children
};
}
#endif /* LIST_WIDGET_H_ */
