/**
 *	@file we_meter_widget.h
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This files defines a meter widget
 *
 *	@see WEWidget
 *
 *`
 */
#ifndef WE_METER_WIDGET_H_
#define WE_METER_WIDGET_H_

#include "we_int_widget.h"

class QPainterPath;
class QRect;
class QMouseEvent;

namespace WE
{

class WEImage;

class WEMeterWidget : public WEIntWidget
{
	Q_OBJECT
	Q_PROPERTY(unsigned short numberOfTicks READ getNumberOfTicks)
	Q_PROPERTY(bool drawTicks READ getDrawTicks)
	Q_PROPERTY(bool drawArc READ getDrawArc)
	Q_PROPERTY(double arcWidthFactor READ getArcWidthFactor)
	Q_PROPERTY(bool drawEnclosingBorder READ getDrawEnclosingBorder)
	Q_PROPERTY(double tickAngleStep READ getTickAngleStep)
	Q_PROPERTY(double internalAngle READ getInternalAngle)
	Q_PROPERTY(double needleThicknessFactor READ getNeedleThicknessFactor)
	Q_PROPERTY(unsigned int middleDotThickness READ getMiddleDotThickness)
	Q_PROPERTY(double tickHeightFactor READ getTickHeightFactor)
	Q_PROPERTY(double tickWidthFactor READ getTickWidthFactor)
	Q_PROPERTY(bool drawTickPoint READ getDrawTickPoint)
	Q_PROPERTY(double tickPointSizeFactor READ getTickPointSizeFactor)
	Q_PROPERTY(int direction READ getDirection)
	Q_PROPERTY(bool needleImageSet READ getNeedleImageSet)
	Q_PROPERTY(bool isDrawNeedleImage READ getIsDrawNeedleImage)

public:

	/**
	 *	@fn WEMeterWidget(WEWidget *parent, void * p_controller = nullptr)
	 *
	 *	@brief constructor of the meter
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] parent pointer to parent object, if nullptr then top widget
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 *
	 *	@returns nothing
	 */
	explicit WEMeterWidget(WEWidget *parent, void * p_controller = nullptr);

	/**
	 *	@fn WEMeterWidget(WEWidget *parent, void * p_controller = nullptr)
	 *
	 *	@brief constructor of the meter
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] parent pointer to parent object, if nullptr then top widget
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 *
	 *	@returns nothing
	 */
	WEMeterWidget(int typeID, WEWidget *parent, void * p_controller = nullptr);

	/**
	 *	@fn ~WEMeterWidget()
	 *
	 *	@brief deconstructor
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @returns nothing
	 */
	virtual ~WEMeterWidget();


	/**
	 *	@fn void paint(WEPainter* painter)
	 *
	 *	@brief draws the meter
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] painter pointer to painter object
	 *
	 *  @returns nothing
	 */
	void paint(WEPainter* painter) override;

	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data) override;

	/**
	 *	@fn int createNeedleImage(std::string p_strFile)
	 *
	 *	@brief creates a needle image instead of using a drawing needle
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_strFile fullpath + filename to file
	 *
	 * 	@returns WE_SUCCESS if ok, and WE_GENERAL_ERROR in other cases
	 *
	 *  	@see WEImage
	 *
	 *  A meter shall have a nice and beautiful needle.
	 *
	 */
	int createNeedleImage(const std::string & p_strFile);

	/**
	 *	@fn void release()
	 *
	 *	@brief release the members
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  	@returns void
	 */
	void release();

	/**
	 *	@fn void drawBorder(WE::WEPainter* p_painter, bool p_ellipse,int START_X, int START_Y )
	 *
	 *	@brief Draw the Border over the MweterWidget
	 *
	 *	@author Tanmoy Deb
	 *	@author tanmoy.deb@aricent.com
	 *
	 *
	 *	@param[in] p_painter The painter
	 *	@param[in] START_X  The top left co-ordinate of the outer region.
	 *	@param[in] START_Y The top left co-ordinate of the outer region
	 *
	 *	@see REFERENCE WE::WEArchedBarGrapgWidget::drawBorder()
	 *
	 */
	void drawBorder(WE::WEPainter* p_painter, bool p_ellipse,int START_X, int START_Y );

	/**
	 *	@fn void void drawMeterEnclosingBorder(WEPainter* painter)
	 *
	 *	@brief Draw the Enclosing Border over the Meter Widget
	 *
	 *	@author satheeshkumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *
	 *	@param[in] p_painter The painter
	 *
	 *	@see REFERENCE WE::WEArchedBarGrapgWidget::drawBorder()
	 *
	 */
	void drawMeterEnclosingBorder(WEPainter* p_painter);

	/**
	 *	@fn void drawNeedle(WEPainter* painter)
	 *
	 *	@brief calcs tick steps
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] painter pointer to painter
	 *
	 *  @returns void
	 */
	void drawNeedle(WEPainter* p_painter);

	void updateNeedleNodes(WEControllerNode *controller);

	/**
	 *	@fn void drawTicks(WEPainter* painter)
	 *
	 *	@brief draw the ticks into the meter
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] painter pointer to painter
	 *
	 *  @returns void
	 */
	void drawMeterTicks(WEPainter* painter);

	/**
	 *	@fn void drawMeterArc(WEPainter* painter)
	 *
	 *	@brief draw the arc into the meter
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] painter pointer to painter
	 *
	 *  @returns void
	 */
	void drawMeterArc(WEPainter* painter);

	/**
	 *	@fn int setStartAngle(std::uint16_t p_startAngle)
	 *
	 *	@brief set the start Angle
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] p_startAngle start angle of the meter
	 *
	 *  @returns WE_ERROR code
	 */
	int setStartAngle(std::uint16_t p_startAngle);


	/**
	 *	@fn int setEndAngle(std::uint16_t p_endAngle)
	 *
	 *	@brief set the end Angle
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] p_endAngle end angle of the meter
	 *
	 *  @returns WE_ERROR code
	 */
	int setEndAngle(std::uint16_t p_endAngle);

	int setNeedleThicknessFactor(double needleThicknessFactor);

	/**
	 *	@fn int setTickColor(WEColor p_tickColor)
	 *
	 *	@brief set the tick color
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] p_tickColor color used for drawing the ticks
	 *
	 *  @returns WE_ERROR code
	 */
	int setTickColor(WEColor p_tickColor);


	/**
	 *	@fn int setArcColor(WEColor p_arcColor)
	 *
	 *	@brief set the arc color
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] p_arcColor color used for drawing the arc
	 *
	 *  @returns WE_ERROR code
	 */
	int setArcColor(WEColor p_arcColor);


	/**
	 *	@fn int drawTicks(bool p_drawTicks)
	 *
	 *	@brief set drawing of ticks to true or false
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] p_drawTicks sets if the ticks are drawn or not
	 *
	 *  @returns WE_ERROR code
	 */
	int setDrawTicks(bool p_drawTicks);


	/**
	 *	@fn int setNumberOfTicks(std::uint16_t p_tickCount)
	 *
	 *	@brief set the number of ticks to be drawn
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] p_tickCount number of ticks to be drawn
	 *
	 *  @returns WE_ERROR code
	 */
	int setNumberOfTicks(std::uint16_t p_tickCount);


	/**
	 *	@fn int setDeflectionDirection(WEDeflectionDirectionType p_dir)
	 *
	 *	@brief This method sets the absolute min and max value
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_dir direction
	 *
	 *  @see WEDeflectionDirectionType
	 *
	 */
	int setDeflectionDirection(WEDeflectionDirectionType p_dir);


	/**
	 *	@fn int setSize(int p_width, int p_height)
	 *
	 *	@brief This method is overwritten from the base class we_widget
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_width width of the meter
	 *	@param[in] p_height height of the meter
	 *
	 *  @returns WE_SUCCESS if ok, and WE_GENERAL_ERROR in other cases
	 *
	 *  @see WEWidget
	 *
	 *  A meter shall have the same width and height
	 *
	 */
	int setSize(int p_width, int p_height) override;

	/**
	 *	@fn int setDrawArc(bool p_drawArc)
	 *
	 *	@brief set drawing of the arc to true or false
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  	@param[in] p_drawTicks sets if the arc is drawn or not
	 *
	 *  	@returns WE_ERROR code
	 */
	int setDrawArc(bool p_drawArc)
	{
		this->drawArc = p_drawArc;
		return 0;
	}

	/**
	 *	@fn int setArcWidthFactor(double arcWidthFactor)
	 *
	 *	@brief set the width of the drawn arc
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  	@param[in] arcWidthFactor factor for the arc width
	 *
	 *  	@returns WE_ERROR code
	 */
	int setArcWidthFactor(double arcWidthFactor);

	/**
	 *	@fn int setImagesActive(bool p_needle, bool p_border, bool p_background)
	 *
	 *	@brief set the color for below min and above max
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  	@param[in] p_needle sets if the needle image is drawn or not
	 *  	@param[in] p_border sets if the border image is drawn or not
	 *  	@param[in] p_background sets if the background image is drawn or not
	 *
	 *  	@returns WE_ERROR code
	 *
	 * 	Can be used for toggleing the diffrent images.
	 */
	int setImagesActive(bool p_needle, bool p_border, bool p_background);

	/**
	 *	@fn int setTickHeightFactor(double tickHeightFactor)
	 *
	 *	@brief set the tick height factor
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  	@param[in] tickHeightFactor factor for the tick height
	 *
	 *  	@returns WE_ERROR code
	 */
	int setTickHeightFactor(double tickHeightFactor);

	/**
	 *	@fn int setTickWidthFactor(double tickWidthFactor)
	 *
	 *	@brief set the tick width factor
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  	@param[in] tickWidthFactor factor for the tick width
	 *
	 *  	@returns WE_ERROR code
	 */
	int setTickWidthFactor(double tickWidthFactor);

	/**
	 *	@fn int setDrawTickPoint(bool drawTickPoint)
	 *
	 *	@brief set the drawing of the tick points to true of false
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  	@param[in] drawTickPoint sets if the tick points are drawn or not
	 *
	 *  	@returns WE_ERROR code
	 */
	int setDrawTickPoint(bool drawTickPoint);

	/**
	 *	@fn int setTickPointSizeFactor(double tickPointSizeFactor)
	 *
	 *	@brief set the tick point size factor
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  	@param[in] tickPointSizeFactor factor for the tick point size
	 *
	 *  	@returns WE_ERROR code
	 */
	int setTickPointSizeFactor(double tickPointSizeFactor);

	/**
	 *	@fn int setTickPointColor(WEColor tickPointColor)
	 *
	 *	@brief set the tick point color
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  	@param[in] tickPointColor color of the tick point
	 *
	 *  	@returns WE_ERROR code
	 */
	int setTickPointColor(WEColor tickPointColor);

	/**
	 *	@fn int setTickPointFormat(WETickPointFormatType tickPointFormat)
	 *
	 *	@brief set the tick point geometry format (round / square / ...)
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  	@param[in] tickPointFormat format of the tick points
	 *
	 *  	@returns WE_ERROR code
	 */
	int setTickPointFormat(WETickPointFormatType tickPointFormat);

	/**
	 *	@fn setEnclosingBorderColor(WEColor p_enclosingBorderColor)
	 *
	 *	@brief set the Enclosing Border color
	 *
	 *	@author satheeshkumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *  @param[in] p_enclosingBorderColor color used for drawing the Enclosing Border
	 *
	 *  @returns WE_ERROR code
	 */
	int setEnclosingBorderColor(WEColor p_enclosingBorderColor);

	/**
	 *	@fn int setDrawEnclosingBorder(bool p_drawEnclosingBorder)
	 *
	 *	@brief set drawing of Enclosing Border to true or false
	 *
	 *	@author satheeshkumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *  @param[in] p_drawEnclosingBorder sets if the Enclosing Border is drawn or not
	 *
	 *  @returns WE_ERROR code
	 */
	int setDrawEnclosingBorder(bool p_drawEnclosingBorder);

	/**
	 * @fn int setColorMinMax(WEColor & p_minColor,WEColor & p_normalColor,
	 * WEColor & p_maxColor)
	 *
	 * @brief sets color values for p_minColor,P_normalColor and p_maxColor
	 *
	 * @author Ajay.Bisht@aricent.com
	 * @author Ajay Bisht
	 *
	 * @param[in] p_minColor
	 * @param[in] p_normalColor
	 * @param[in] p_maxColor
	 *
	 * @returns WE::WE_SUCCESS
	 */
	int setColorMinMax(const WEColor & p_minColor,
			   const WEColor  & p_normalColor,
			   const WEColor  & p_maxColor) override;


	/**
	 *	@fn int WEMeterWidget::setColorNormal(WEColor & p_normalColor)
	 *
	 *	@brief sets value for normalColor
	 *
	 *	@author Ajay.Bisht@aricent.com
	 *	@author Ajay Bisht
	 *
	 *  @param[in] p_normalColor
	 *
	 *   @returns WE::WE_SUCCESS
	 */
	int setColorNormal(const WEColor & p_normalColor) override;

	// Overwritten setter from WEWidget to calling the method "repaintBackgroundPixmap()" after setting the value.
	// This is a workarround until the background pixmap mechanism is moved up to WEWidget.
	int setStartAngle(int p_startAngle) override;
	int setEndAngle(int p_endAngle) override;

	unsigned short getNumberOfTicks(void) const
	{
		return this->numberOfTicks;
	}

	bool getDrawTicks(void) const
	{
		return this->drawTicks;
	}

	bool getDrawArc(void) const
	{
		return this->drawArc;
	}

	double getArcWidthFactor(void) const
	{
		return this->arcWidthFactor;
	}

	bool getDrawEnclosingBorder(void) const
	{
		return this->drawEnclosingBorder;
	}

	double getTickAngleStep(void) const
	{
		return this->tickAngleStep;
	}

	double getInternalAngle(void) const
	{
		return this->internalAngle;
	}

	double getNeedleThicknessFactor(void) const
	{
		return this->needleThicknessFactor;
	}

	unsigned int getMiddleDotThickness(void) const
	{
		return this->middleDotThickness;
	}

	double getTickHeightFactor(void) const
	{
		return this->tickHeightFactor;
	}

	double getTickWidthFactor(void) const
	{
		return this->tickWidthFactor;
	}

	bool getDrawTickPoint(void) const
	{
		return this->drawTickPoint;
	}

	double getTickPointSizeFactor(void) const
	{
		return this->tickPointSizeFactor;
	}

	WEColor getTickPointColor(void) const
	{
		return this->tickPointColor;
	}

	WETickPointFormatType getTickPointFormat(void) const
	{
		return this->tickPointFormat;
	}

	int getDirection(void) const
	{
		return this->direction;
	}

	bool getNeedleImageSet(void) const
	{
		return this->needleImageSet;
	}

	bool getIsDrawNeedleImage(void) const
	{
		return this->isDrawNeedleImage;
	}

	virtual void setValueToValid(bool p_valid) override;

	bool isVtMeter;		//!< Use correct tick count even for full circles

	virtual bool amIHit(int p_globalX, int p_globalY) const override;

protected:

	/**
	 *	@fn void paintEvent(QPaintEvent* paintEvent)
	 *
	 *	@brief this method is overwritten from the class WEWidget and calls additionally the method "createBackgroundPixmap()"
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  @param[in] paintEvent event for painting which is trigger from qt
	 *
	 *  @returns nothing
	 */
	void paintEvent(QPaintEvent* paintEvent);

	/**
	 *	@fn paintBackgroundGraphics(WEPainter* painter)
	 *
	 *	@brief draws static background graphics on the widget (background, ticks, tick values, ...)
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  @param[in] painter pointer to painter object
	 *
	 *  @returns nothing
	 */
	virtual void paintBackgroundGraphics(WEPainter* painter) override;

	/**
	 *	@fn void paintForegroundGraphics(WEPainter* painter)
	 *
	 *	@brief draws high frequently updated foreground graphics on the widget (e.g. needle, digital value, ...)
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  @param[in] painter pointer to painter object
	 *
	 *  @returns nothing
	 */
	virtual void paintForegroundGraphics(WEPainter* painter);

	virtual void updateForegroundGraphicsNodes(WEControllerNode *controller);

	virtual void createBackgroundPixmap() override;

	WEMeterTextureCacheKey getBackgroundTextureCacheKey() const;

	bool isInputRoot() const override;

	/**
	 * Create the enclosingPath QPainerPath that represents the actual painting are of this meter.
	 * Used for optimized painting.
	 */
	void createEnclosingPath(void);
	QPainterPath * enclosingPath;
	QRegion paintingRegion;
private:

	/**
	 *	@fn void init()
	 *
	 *	@brief initialize the members
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  	@returns void
	 */
	void init();

	/**
	 *	@fn void calcTicks()
	 *
	 *	@brief calcs tick steps
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @returns void
	 */
	void calcTicks();

	void createAutoSVGNeedles(void);
	void createAutoSVGNeedle(WEImage * image, const WEPoint (&triangle)[3], const WEColor & needleColor);

	/**
	 *	@fn bool needleImageFileUsedByOtherImage(const WEImage * p_needleImage, const WEImage * p_otherNeedleImage)
	 *
	 *	@brief Checks if the given images use the same image file
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param p_needleImage pointer to the first image
	 *	@param p_otherNeedleImage pointer to the second image
	 *
	 *	@returns true if the images use the same image file
	 */
	bool needleImageFileUsedByOtherImage(const WEImage * p_needleImage, const WEImage * p_otherNeedleImage);

	/**
	 *	@fn bool needleImageFileUsedByOtherImages(const WEImage * p_needleImage)
	 *
	 *	@brief Checks if the given image uses the same image file as the other needle images
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param p_needleImage pointer to the image
	 *
	 *	@returns true if the image file is used by an other needle image
	 */
	bool needleImageFileUsedByOtherImages(const WEImage * p_needleImage);

	/**
	 *	@brief Do not only re-initialize or delete the object but also delete the file created for it.
	 *
	 *	@param p_needleImage Reference to the needle image member to destroy
	 *	@param p_delete true: delete the object and set p_needleImage to nullptr, otherwise only delete file and call p_needleImage->release()
	 *	@return An error code, WE_SUCCESS in case of success
	 *	@return WE_INVALID_PARAMETER_ERROR if p_needleImage is nullptr
	 *	@return Any error code WEFilesAndFolders::eraseFileOrFolder() may return.
	 *
	 *	@see WEFilesAndFolders::eraseFileOrFolder()
	 */
	int destroyNeedleImage(WEImage * & p_needleImage, bool p_delete);


	double tickAngleStep;	//!< saves the angle per tick
	double internalAngle; //!
	double needleThicknessFactor;

	WEImage* needleImage; //!< stores pointer to needle image.
	WEImage* needleImageBelowMin; //!< stores pointer to needle image.
	WEImage* needleImageAboveMax; //!< stores pointer to needle image.

	unsigned int middleDotThickness;

	WEDeflectionDirectionType direction; //!< stores current direction of the needle.

	bool drawTicks; //!< Draw ticks true/false
	bool drawArc; //!< Draw ticks true/false
	double arcWidthFactor;
	bool drawEnclosingBorder; //!< Draw Enclosing Border true/false
	bool needleImageSet;	//!< true indicates the needle image was set otherwise use the default one.
	bool isDrawNeedleImage;

	WEColor tickColor; //!< color of ticks
	WEColor arcColor; //!< color of the arc
	WEColor enclosingBorderColor; //!< color of the Enclosing Border

	std::uint16_t numberOfTicks; //!< Number of ticks to be drawn

	double tickHeightFactor;
	double tickWidthFactor;
	bool drawTickPoint;
	double tickPointSizeFactor;
	WE::WEColor tickPointColor;
	WE::WETickPointFormatType tickPointFormat;

	WEImageNode * backgroundPixmapImageNode;
	WEImageNode * needleImageNode;
};
}
#endif /* WE_METER_WIDGET_H_ */
