/**
 *	@file we_numeric_widget.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This files defines a widget for the drawing a numeric widget at the given location.
 *	@brief It inherits from WETextWidget.
 *
 *	@see WEWidget
 *
 */

#ifndef WE_NUMERIC_WIDGET_H_
#define WE_NUMERIC_WIDGET_H_

#include "../../include/we_datatypes_gui.h"
#include "we_text_widget.h"
#include "we_int_widget.h"

namespace WE
{

class WETextNode;

class WENumericWidget : public WEIntWidget, public WETextWidget
{
	Q_OBJECT
	Q_PROPERTY(int format READ getFormat)
	Q_PROPERTY(bool displayZeroAsBlank READ getDisplayZeroAsBlank)
	Q_PROPERTY(bool leadingZeros READ getLeadingZeros)
	Q_PROPERTY(bool truncate READ getTruncate)
	Q_PROPERTY(bool isSigned READ getIsSigned)
	Q_PROPERTY(QString displayedValue READ getDisplayedValueAsQString)
	Q_PROPERTY(char decimalChar READ getDecimalChar)
	Q_PROPERTY(QString zeroString READ getZeroStringAsQString)

public:
	/**
	 *	@fn WENumericWidget(WEWidget *p_parent, void * p_controller = nullptr)
	 *
	 *	@brief Constructs a WENumericWidget.This widget can be created as
	 *	a separate window by specifying the parent as nullptr or it can be included
	 *	as a child widget of the parent.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] parent If parent IS nullptr then this widget becomes a window with the same properties.
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	*/
	explicit WENumericWidget(WEWidget *p_parent, void * p_controller = nullptr);


	/**
	 *	@fn ~WENumericWidget()
	 *
	 *	@brief Destructor of the widget.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	*/
	virtual ~WENumericWidget();

	/**
	 *	@fn paint(WEPainter* painter)
	 *
	 *	@brief The drawing for this widget is performed in this method.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@see WEPainter
	 *	@see repaint()
	 *
	 *	@warning Only one instance of the WEPainter can be created for a Widget.
	 *	@warning Do not invoke this method directly, call repaint() to redraw
	 *	the contents of the widget.
	 *
	 *	@param[in] painter current WEPainter
	*/
	void paint(WEPainter* p_painter) override;


	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData *p_data) override;


	/**
	 *	@fn int setValue(int p_value)
	 *
	 *	@brief This method is used set the value of the IntWidget
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_value int value to be set
	 *	@param[in] p_updateScrollbar unused for numerics
	 *
	 *	@returns WE_ERROR code:
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	Numeric widget can generally display every value and absMin and absMax values are
	 *	only relevant if the value is edited. But the behavior for this case has to be
	 *	covered by the widget's owner (for PClient application the owner of a widget is the
	 *	DDO).
	 *	Consequently the numeric widget's implementation of setValue accepts every value
	 *	provided.
	 *
	 */
	int setValue(int p_value, bool p_updateScrollbar = false) override;


	/**
	 *	@fn int setLeadingZeros(bool p_flag)
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_flag is a flag which show if the leadings zeros should be drawn
	 *
	 *	@returns WE_SUCCESS
	 */
	int setLeadingZeros(bool p_flag);
	int setFormat(WENumberDisplayFormat p_format);
	int setDisplayZeroAsBlank(bool p_displayZerosAsBlank);

	int setTruncate(bool p_flag);
	//int setFontOutlined(bool outlined);

	static void setDecimalChar(char p_decimalChar);

	int setSign(bool p_isSigned);

	static char getDecimalChar(void)
	{
		return decimalChar;
	}

	const std::string & getDisplayedValue(void) const override
	{
		return this->displayedValue;
	}

	bool getIsSigned(void) const
	{
		return this->isSigned;
	}

	int getFormat(void) const
	{
		return this->format;
	}

	bool getDisplayZeroAsBlank(void) const
	{
		return this->displayZeroAsBlank;
	}

	bool getLeadingZeros(void) const
	{
		return this->leadingZeros;
	}

	bool getTruncate(void) const
	{
		return this->truncate;
	}

	QString getDisplayedValueAsQString(void) const
	{
		return QString::fromStdString(this->displayedValue);
	}

	static QString getZeroStringAsQString(void)
	{
		return QString::fromStdString(zeroString);
	}


private:
	/**
	 *	@fn init()
	 *
	 *	@brief the method to initialize the member variable of the class
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	*/
	void init();

	WENumberDisplayFormat format;

	bool displayZeroAsBlank;
	bool leadingZeros;		//!<The flag for to show the leading zeros
	bool truncate;				//!<Truncate the no of decimals
	bool isSigned;				// Specifies whether value is signed or not

	std::string displayedValue;

	WETextNode *textNode;

	static char decimalChar;
	static const std::string zeroString;
};
}
#endif /* WE_NUMERIC_WIDGET_H_ */
