/**
 *	@file we_rect_widget.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This files defines a widget for the drawing a rectangle widget
 *	and provides interfaces to configure the WERectWidget.
 *
 *	@see we_rect_widget.cpp
 *
 */

#ifndef WE_RECT_WIDGET_H_
#define WE_RECT_WIDGET_H_

#include "we_widget.h"

namespace WE
{

enum WERectLineSuppression
{
	SUPPRESS_NOTHING	= 0x00,
	SUPPRESS_TOP_LINE	= 0x01,
	SUPPRESS_RIGHT_LINE	= 0x02,
	SUPPRESS_BOTTOM_LINE	= 0x04,
	SUPPRESS_LEFT_LINE	= 0x08
};

/**
 *	@class WERectWidget
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@brief This widget provides rectangle display.
 *
 *	@namespace WE
 *
 *	This class includes the functions, which allows
 *	the user to display the specified rectangles at the given location. It inherits
 *	from WEWidget, which inherits from WEPaintListener to draw its line in the paint() event.
 */
class WERectWidget : public WEWidget
{
	Q_OBJECT
	Q_PROPERTY(unsigned char lineSuppression READ getLineSuppression)
public:
	/**
	 *	@fn WERectWidget(WEWidget *p_parent, void * p_controller = nullptr)
	 *
	 *	@brief Constructs a WERectWidget Widget.This widget can be created as
	 *	a separate window by specifying the parent as nullptr or it can be included
	 *	as a child widget of the parent.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] *p_parent If parent IS nullptr then this widget becomes a window with the same properties.
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 */
	explicit WERectWidget(WEWidget *p_parent, void * p_controller = nullptr);

	/**
	 *	@fn ~WERectWidget()
	 *
	 *	@brief Destructor of the widget.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 */
	virtual ~WERectWidget();

	/**
	 *	@fn init()
	 *
	 *	@brief the method to initialize the member variable of the class
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 */
	void init();

	/**
	 *	@fn void paint(WEPainter* painter)
	 *
	 *	@brief The drawing for this widget is performed in this method.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@see WEPainter
	 *	@see repaint()
	 *
	 *	@warning Only one instance of the WEPainter can be created for a Widget.
	 *	@warning Do not invoke this method directly, call repaint() to redraw
	 *	the contents of the widget.
	 *
	 */
	void paint(WEPainter* p_painter) override;

	/**
	 *	@fn setRect(int p_x1,int p_y1,int p_x2, int p_y2)
	 *
	 *	@brief Set the member variables of the rectangle widget
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_x1 X-location of the left upper corner of the rectangle
	 *	@param[in] p_y1 Y-location of the left upper corner of the rectangle
	 *	@param[in] p_x2 X-location of the right bottom corner of the rectangle
	 *	@param[in] p_y2 Y-location of the right bottom corner of the rectangle
	 */
	void setRect(int p_x1,int p_y1,int p_x2, int p_y2);

	/**
	 *	@fn setRect(WERect &p_rect)
	 *
	 *	@brief Define the rectangle, set its member variables
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] &p_rect reference to the existing rectangle
	 */
	void setRect(const WERect &p_rect);

	void setLineSuppression(unsigned char p_position);

	unsigned char getLineSuppression(void) const
	{
		return this->lineSuppression;
	}

private:
	unsigned char lineSuppression;
};
}
#endif /* WE_RECT_WIDGET_H_ */
