/**
 *	@file we_string_widget.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This files defines a widget for the drawing a string widget.
 *	@brief The class WEStringWidget inherit from the class WETextWidget, which contains the functions,
 *	@brief which are similar by string and numeric widgets.
 *
 *	@see WEWidget, WETextWidget
 *
 */

#ifndef STRING_WIDGET_H_
#define STRING_WIDGET_H_


#include "we_widget.h"
#include "we_text_widget.h"
#include "../../nodes/include/we_node_utils.h"

namespace WE
{

class WETextNode;

static const int DEFAULT_MAX_STRING_LENGTH = 255;

class WEStringWidget : public WEWidget, public WETextWidget
{
Q_OBJECT
Q_PROPERTY(QString value READ getValueAsQString WRITE setValueAsQString)
Q_PROPERTY(int maxStringLength READ getMaxStringLength WRITE setMaxStringLength)
Q_PROPERTY(bool autoWrap READ getAutoWrap WRITE setAutoWrap)
Q_PROPERTY(bool wrapOnHyphen READ getWrapOnHyphen WRITE setWrapOnHyphen)
Q_PROPERTY(bool showCursor READ getShowCursor WRITE setShowCursor)
Q_PROPERTY(unsigned int cursorPosition READ getCursorPosition WRITE setCursorPosition)
Q_PROPERTY(QString wrapedValue READ getWrappedValueAsQString)


public:
	/**
	 *	@fn WEStringWidget(WEWidget *p_parent, void * p_controller = nullptr)
	 *
	 *	@brief Constructs a WEStringWidget.This widget can be created as
	 *	a separate window by specifying the parent as nullptr or it can be included
	 *	as a child widget of the parent.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_parent If parent IS nullptr then this widget becomes a window with the same properties.
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 */
	explicit WEStringWidget(WEWidget *p_parent, void * p_controller = nullptr);

	/**
	 *	@fn ~WEStringWidget()
	 *
	 *	@brief Destructor of the widget.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 */
	virtual ~WEStringWidget();

	/**
	 *	@fn init()
	 *
	 *	@brief the method to initialize the member variable of the class
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 */
	void init();

	/**
	 *	@fn setString(int p_x, int p_y, std::string p_value)
	 *
	 *	@brief This method is used set the given string at the given location.
	 *
	 *	@param none
	 *
	 *	@return value displayed
	 *
	 *	This implements a virtual method from WEWidget.
	 */
	const std::string & getDisplayedValue(void) const override;

	/**
	 *	@fn setValue(std::string p_value)
	 *
	 *	@brief This method is used set the value of the numeric
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_value String value to be set
	 *
	 */
	virtual int setValue(const std::string & p_value);
	int setValueAsQString(QString p_value);

	/**
	 *	@fn paint(WEPainter* painter)
	 *
	 *	@brief The drawing for this widget is performed in this method.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@see WEPainter
	 *	@see repaint()
	 *
	 *	@warning Only one instance of the WEPainter can be created for a Widget.
	 *	@warning Do not invoke this method directly, call repaint() to redraw
	 *	the contents of the widget.
	 *
	 */
	void paint(WEPainter* p_painter) override;

	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData */*p_data___unused___*/) override;

	/**
	 *	@fn setAutoWrap(bool p_wrap)
	 *
	 *	@brief Set the flag, if the string should be wraps by the application or not
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@see calculateAutoWrap(WEPainter* painter)
	 *
	 *	@param[in] p_wrap boolean value to be set
	 */
	int setAutoWrap(bool p_wrap);
	bool getAutoWrap() const;

	/**
	 *	@fn setMaxStringLength(int p_maxStringLength)
	 *
	 *	@brief Sets the Maximum Length of the String that can be set to the Widget.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@warning If the current length of the String exceeds the Maximum String Length being set, then
	 *	the part of current string exceeding the Maximum String Length will be deleted.
	 *
	 *	@param[in] p_maxStringLength int Maximum String Length to be allowed.
	 */
	int setMaxStringLength(int p_maxStringLength);
	int getMaxStringLength() const;

	//int setFontOutlined(bool outlined);

	int setWrapOnHyphen(bool p_flag);
	bool getWrapOnHyphen() const;


	/**
	 *	@fn int setShowCursor(bool p_show)
	 *
	 *	@brief Figure out if this string should show a cursor.
	 *
	 *	@param none
	 *
	 *	@returns A bool indicating if a cursor should be shown.
	 *
	 *	@see int setShowCursor(bool)
	 *	@see int setCursorPosition(unsigned int)
	 */
	bool getShowCursor(void) const;


	/**
	 *	@fn int setShowCursor(bool p_show)
	 *
	 *	@brief Define if this string should show a cursor.
	 *
	 *	@param[in] p_show Set to true if a cursor should be shown within the string.
	 *
	 *	@returns an error code - WE_SUCCESS in case of success.
	 *
	 *	@see bool getShowCursor(void)
	 *	@see int setCursorPosition(unsigned int)
	 */
	int setShowCursor(bool p_show);


	/**
	 *	@fn unsigned int getCursorPosition(void)
	 *
	 *	@brief Figure out which character of the string should be highlighted as a cursor.
	 *
	 *	@param none
	 *
	 *	@returns The character position where the cursor should be shown.
	 *
	 *	@see bool getShowCursor(void)
	 *	@see int setShowCursor(bool)
	 *	@see int setCursorPosition(unsigned int)
	 */
	unsigned int getCursorPosition(void) const;


	/**
	 *	@fn int setShowCursor(bool p_show)
	 *
	 *	@brief Define which character of the string should be highlighted as a cursor.
	 *
	 *	@param[in] p_position Set to [(0) to (string length -1)] to point to a character where the cursor should be shown.
	 *
	 *	@returns an error code - WE_SUCCESS in case of success.
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if value is too high - cursor will be set to last character then.
	 *
	 *
	 *	@see unsigned int getCursorPosition(void)
	 *	@see int setShowCursor(bool)
	 *	@see bool getShowCursor(void)
	 */
	int setCursorPosition(unsigned int p_position);


	const std::string & getValue(void) const;
	QString getValueAsQString() const;
	QString getWrappedValueAsQString() const;

	void setFontAlpha(std::uint8_t p_alpha);

protected:
	int maxStringLength;		//!<Max Length of the String
private:
	bool autoWrap;		//!<Property, which is true, if the calculation for the newline is needed
	bool wrapOnHyphen;		//!<Wrap text on hyphen char
	// properties for cursor support (introduced to support VKB widget)
	bool showCursor;
	unsigned int cursorPosition;

	WEFont * invertedFont;
	WEFont * invertedResizedFont;

	WETextNode *textNodeLeft;	// text left of cursor (or without cursor)
	WETextNode *textNodeRight;	// text right of cursor (e.g. while editing)
	WETextNode *textNodeCursor;	// edited text (character) under cursor

protected:
	std::string valueToDisplay;		//!<String to be displayed
private:
	std::string wrapedValue;	//!<String to be displayed with calculated wraps
};
}
#endif /* STRING_WIDGET_H_ */
