/**
 *	@file we_text_widget.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief WETextWidget is one parent class of the WEStringWidget and WENummericWidget
 *	@brief and contains the common properties of these two widgets
 *
 *
 *	Text widget contains the font related properties which could be needed in a widget.
 *	In order to avoid inheritance conflicts it is not a WEWidget itself.
 *	So if somebody wants to implement a widget with text it should inherit from this class
 *	and WEWidget, WEIntWidget or another class based on WEWidget.
 *
 *
 *
 *	@see WEWidget, WEIntWidget, WEStringWidget, WENummericWidget
 *
 */

#ifndef TEXT_WIDGET_H_
#define TEXT_WIDGET_H_

#include "../../include/we_datatypes_gui.h"
#include "../../font/include/we_font.h"

namespace WE
{

class WEPainter;
class WEWidget;

class WETextWidget
{
public:
	/**
	 *	@fn WETextWidget(WEWidget * p_widget)
	 *
	 *	@brief Constructs a WETextWidget.This is NOT a WEWidget!
	 *
	 *	@param p_widget - pointer to the widget that inherits from "this"
	 *
	 */
	explicit WETextWidget(WEWidget * p_widget);


	/**
	 *	@fn ~WETextWidget()
	 *
	 *	@brief Destructor of the widget.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 */
	virtual ~WETextWidget();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WETextWidget(const WETextWidget &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WETextWidget & operator=(const WETextWidget &) = delete;

	/**
	 *	@fn int setWEFont(WEFont * p_font)
	 *
	 *	@brief set the font of the text
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] &p_font the address of the assigned font
	 *
	 *	@returns WE_ERROR code
	 */
	int setWEFont(const WEFont * p_font);
	//WEFont * getFont() const;

	void setFontStyle(int p_fontStyle);
	void setFontColor(const WE::WEColor & p_color);

	inline int getFontStyle(void)
	{
		return this->weFont.getFontStyle();
	}


	/**
	 *	@fn int setFontSize(int p_size)
	 *
	 *	@brief set size of the font in pixel
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_size size in pixel
	 *
	 *	@returns WE_ERROR code
	 */
	int setFontSize(int p_size);


	/**
	 *	@fn int setAlignmentFlag(WE::WEAlignment p_align)
	 *
	 *	@brief set the flag for the alignment
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_align is a flag for the alignment (both vertical and horizontal)
	 *
	 *	@returns WE_ERROR code
	 */
	int setAlignmentFlag(WE::WEAlignment p_align);


	/**
	 *	@fn WEFont * getDisabledFont()
	 *
	 *	@brief get the disabled font
	 *	(Create it if it doesn't exist.)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@see std::string getDisabledString(int p_width)
	 *
	 *	@returns the font
	 */
	WEFont * getDisabledFont();


	/**
	 *	@fn std::string getDisabledString(int p_width)
	 *
	 *	@brief Returns a string in the following form: "- - - - -"
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] p_width the maximum width of the string
	 *
	 *	@see WEFont * getDisabledFont()
	 *
	 *	@returns the string
	 */
	std::string getDisabledString(int p_width);


protected:
	WEWidget * myWidget{nullptr};	//!< Widget inheriting "this"
	WEFont weFont{"DejaVu Sans", 10, WE::STYLE_NORMAL, WE::WEColor(WE::black)};	//!<The pointer of the font, which should be used for the text display
	WEAlignment align{WE::ALIGN_CENTER};	//!<Flag for the current alignment

private:
	WEFont * disabledFont{nullptr};	//!<The pointer of the disabled font, which should be used to draw the disabled string in disabled WEWidgets
};	// class
}	// namespace
#endif /* TEXT_WIDGET_H_ */
