/**
 *	@file we_window_widget.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file defines the class for the Window widget.
 *
 *	The window can be used by other Widgets to set it as their parent.
 */


#ifndef WE_WINDOW_H_
#define WE_WINDOW_H_

#include "we_widget.h"
#include "../qt/qt_quick_we_scrollbar.h"

class QQuickCloseEvent;

namespace WE
{

class WELinearBargraphWidget;
class WEImageWidget;
class WEPainter;

/**
 *	@class WEWindowWidget
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief Class for the Window Widget.
 *
 *	@namespace WE
 *
 *	This is basically a WEWidget with some overwritten and additional
 *	functions (e.g. loading screen functionality, show modal warning message, etc.)
 *
 */
class WEWindowWidget : public WEWidget
{
	Q_OBJECT
	Q_PROPERTY(bool betaRelease READ isBetaRelease)
	Q_PROPERTY(QString buildTimeAndDate READ getBuildTimeAndDate)
	Q_PROPERTY(bool loadingScreenVisible READ isLoadingScreenVisible)
public:
	/**
	 *	@fn WEWindowWidget(WEWidget * p_parent = nullptr, void * p_controller = nullptr)
	 *
	 *	@brief Constructor for the window.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 */
	WEWindowWidget(WEWidget * p_parent = nullptr, void * p_controller = nullptr);

	/**
	 *	@fn ~WEWindowWidget()
	 *
	 *	@brief Destructor for the Window.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 */
	virtual ~WEWindowWidget();

	/**
	 *	@fn paint(WEPainter* painter)
	 *
	 *	@brief This method is the event handler of the WEWidget's paintEvent
	 *	which is automatically invoked whenever the widget needs to be repainted.
	 *	All the drawing for this Window is performed in this method. It
	 *	automatically paints all the child widgets also.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param painter The painter to be used for painting in the paint().
	 *
	 *	@see WEPainter
	 *	@see repaint()
	 *
	 *	@warning Only one instance of the WEPainter can be created for a Window.
	 *	@warning Do not invoke this method directly, call repaint() to redraw
	 *	the contents of the Window.
	 */
	void paint(WEPainter* p_painter) override;

	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData *p_data) override;

	/**
	 *	@brief Set the parent widget
	 *
	 *	@param p_newParentWidget The new parent widget
	 */
	void setParent(WEWidget * p_newParentWidget);

	/**
	 *	@brief Show or hide the window
	 *
	 *	@param p_visible Show the window if true
	 */
	void setVisible(bool p_visible) override;

	/**
	 *	@brief Set the position of the window
	 *
	 *	@param x New horizontal position
	 *	@param y New vertical position
	 */
	void setPosition(float p_x, float p_y) override;

	/**
	 *	@brief Resize the window
	 *
	 *	@param w New width
	 *	@param h New height
	 */
	void resize(int p_w, int p_h) override;

	/*
	 * Function to include a widget inside a scroll area.
	 * @param p_include: true create a scroll area and include the widget into it, false: not
	 */
	void includeInScrollArea(bool p_include);

	/**
	 * Overwritten virtual function from WEWidget.
	 * A window widget can not ahve input children and therefore will always return false.
	 * @return false. always. and forever. (maybe)
	 */
	bool hasInputChildren() const override;

	/**
	 * Overwritten function.
	 * This function calls WEWidget::setSize() and additionally automatically adjusts the
	 * size of the loading screen image and linear bargraph
	 * @param p_width	New width of the wondow
	 * @param p_height	New height of the window
	 * @return return value of WEWidget::setSize()
	 */
	int setSize(int p_width, int p_height) override;

	/**
	 * Calling this function will create in image from image path p_path and a
	 * linear bragraph that is BEHIND the image. Make parts of the image transparent
	 * to see the linear bargraph.
	 * The linear bragraph has the full height of the window and 80% of the width and is positioned
	 * at x = 10% of width and y = 0.
	 * If the image specified by p_path could not be used, the loading screen is automatically destroyed and memory
	 * for the image and the linear bargraph is freed.
	 * @param p_path Path to the image that shall be shown as loading screen image
	 * @param loadingScreenColor Color of the linear bargraph
	 * @return WE_SUCCESS if creating the loading screen image was successfull
	 * 	or the error return of WEImageWidget::createDisplayImage()
	 */
	int createLoadingScreen(const std::string & p_path, const WEColor & p_loadingScreenColor);

	/**
	 * Change the value of the loading screen bargraph
	 * @param p_value Value between 0 and 100
	 * @return WE_SUCCESS if settign the value was successfull,
	 * 	WE_NOT_SUPPORTED_ERROR if loading screen was not successfully configured with createLoadingScreen() before,
	 * 	WE_OUT_OF_BOUNDS_ERROR for every p_value < 0 or > 100
	 */
	int setLoadingPercent(int p_value);

	/**
	 * Destroy and free memory of laoding screen image and bargraph created with createLoadingScreen()
	 */
	void destroyLoadingScreen(void);

	/**
	 * Try to set a window icon. If this widget has no parent it will go up in the widget
	 * tree until a parent widget without parent is found. The icon will then be set to the top-level parent.
	 * @param p_pathToImage File path of the iamge that shall be set as window icon
	 */
	void setWEWindowIcon(const std::string & p_pathToImage);

	/**
	 * Show a modal warning message
	 * @warning: The application will block here until the warning dialog is closed!
	 * @param p_title Title of the warning message
	 * @param p_message Warning text to display
	 * @param p_autoCloseTimer Optional timer (in milliseconds). Message dialog will close after timer elapsed (only started if it is set > 0)
	 */
	void showModalWarningMessage(const std::string & p_title, const std::string & p_message, int p_autoCloseTimer = 0);

	/**
	 * Call this function to close the modal warning dialog created with showModalWarningMessage()
	 */
	void hideModalWarningMessage(void);

	/**
	 * Check if this window is marked to be beta software and not a "real" customer ready release.
	 * @return true if beta release only
	 * 	false if final customer release
	 */
	bool isBetaRelease(void) const;

	/**
	 * Mark this window to be a beta or final release
	 * @param p_beta true to mark this window as beta
	 * 	false to mark it as final customer release
	 */
	void setBetaRelease(bool p_beta);


	/**
	 * 	@brief Test wants to make sure they have the correct version
	 *	@return string listing compile time and date
	 */
	const QString getBuildTimeAndDate(void);

	/**
	 * 	@brief Test wants to make sure they have the correct version
	 *	@param p_timeAndDate - string listing compile time and date
	 */
	void setBuildTimeAndDate(const std::string & p_timeAndDate);


	/**
	 * Is loading screen currently active or not
	 */
	bool isLoadingScreenVisible(void);


private slots:

	/**
	 *	@fn void quickWindowResized()
	 *
	 *	@brief Slot which is called when the size of the Qt Quick window changes
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 */
	void quickWindowResized();

	/**
	 * We connect to the QQuickWindow close(QQuickCloseEvent *close) signal
	 * so we can emit our own windowClosing() signal
	 */
	void quickWindowClosing(QQuickCloseEvent *p_close);

	/**
	 *	@fn void verticalScrollBarValueChanged(int value)
	 *
	 *	@brief Slot which is called when the vertical scrollbar value changes
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param value the new scrollbar value
	 */
	void verticalScrollBarValueChanged(int p_value);

	/**
	 *	@fn void horizontalScrollBarValueChanged(int value)
	 *
	 *	@brief Slot which is called when the horizontal scrollbar value changes
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param value the new scrollbar value
	 */
	void horizontalScrollBarValueChanged(int p_value);




private:

	/**
	 *	@brief Set the parent widget
	 *
	 *	@param p_newParentWidget The new parent widget
	 *	@param p_force Allows setting the same parent again if true (for initialization)
	 */
	void setParent(WEWidget * p_newParentWidget, bool p_force);


	WELinearBargraphWidget * loadingBargraph;	//!< linear bargraph for showing loading progress
	WEImageWidget * loadingScreen;			//!< image for showing a loading screen
	QQuickWindow * qquickWindow;			//!< the qt quick window which is needed to show this window
	QtQuickWEScrollBar * scrollBarVertical;		//!< vertical scrollbar, needed for the includeInScrollArea functionality
	QtQuickWEScrollBar * scrollBarHorizontal;	//!< horizontal scrollbar, needed for the includeInScrollArea functionality
	int contentWidth;				//!< the width set by resize (if the includeInScrollArea functionality is used)
	int contentHeight;				//!< the height set by resize (if the includeInScrollArea functionality is used)

	bool betaRelease;				//!< Flag to indicate if this window is part of a beta release
	QString buildTimeAndDate;			//!< Allow Squish to fetch build time and date of the program shown
	bool loadingScreenVisible; //!< Flag to indicate that a loading screen is currently visible

	static const bool resizeWindowContent{false};
signals:
	/**
	 * Signal that connects to the modal warning dialog's accept() slot
	 */
	void closeWarningDialog(void);

	/**
	 * Signal that is emitted when our QQuickWindow is closed
	 */
	void windowClosing(void);
};
}
#endif /* WE_WINDOW_H_ */
