/**
 *	@file qt_quick_we_scrollbar.h
 *
 *	@author preths@topcon.com
 *	@author Patrick Reths
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2020
 *
 */

#ifndef QT_QT_QUICK_WE_SCROLLBAR_WIDGET_H
#define QT_QT_QUICK_WE_SCROLLBAR_WIDGET_H

#include <QTimer>
#include <QQuickItem>



namespace WE
{

class WEControllerNode;
class WERectangleNode;
class WEGeometryNode;

/**
 *
 * This class provides a scrollbar widget/item in the Qt quick scene
 *
 */
class QtQuickWEScrollBar : public QQuickItem
{
	Q_OBJECT

	Q_PROPERTY(int value READ getValue WRITE setValue NOTIFY valueChanged)

public:

	explicit QtQuickWEScrollBar(QQuickItem *p_parent = nullptr);
	virtual ~QtQuickWEScrollBar();

	void setValue(int p_value);
	void setRange(int p_min, int p_max);
	void setSingleStep(int p_step);
	void setPageStep(int p_step);
	void setWheelStep(int p_step);

	void mouseMoveEvent(QMouseEvent *p_event) override;
	void mousePressEvent(QMouseEvent *p_event) override;
	void mouseReleaseEvent(QMouseEvent *p_event) override;
	void wheelEvent(QWheelEvent *p_event) override;

	int getValue() const
	{
		return this->currentValue;
	}

	int getMin() const
	{
		return this->min;
	}

	int getMax() const
	{
		return this->max;
	}

	int getSingleStep() const
	{
		return this->singleStep;
	}

	int getPageStep() const
	{
		return this->pageStep;
	}

	int getWheelStep() const
	{
		return this->wheelStep;
	}

signals:
	void valueChanged(int);

protected:

	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData *p_data) override;

	int positionToValue(int p_yPos);

	void jumpToPosition(int p_yPos);

	void longPressMove();

	int currentValue;
	int singleStep;
	int pageStep;
	int wheelStep;
	int min;
	int max;

private:

	static constexpr int LONG_PRESS_TIME = 500; // in ms
	static constexpr int LONG_PRESS_ACTION_REPEAT_INTERVAL = 50; // in ms

	bool sliderPressed;
	bool buttonPressed1;
	bool buttonPressed2;
	QSize buttonSize;
	int backgroundHeight;
	int sliderHeight;
	int sliderPosition;
	int sliderPressedOffset;
	QTimer longPressTimer;
	int longPressStep;
	int longPressTargetValue;
	bool longPressDetected;
	bool longPressActionUp;

	WEControllerNode *controllerNode;
	WERectangleNode *backgroundRectangleNode;
	WERectangleNode *buttonRectangleNode1;
	WERectangleNode *buttonRectangleNode2;
	WERectangleNode *buttonBorderRectangleNode1;
	WERectangleNode *buttonBorderRectangleNode2;
	WEGeometryNode *buttonSymbolsGeometryNode;
	WERectangleNode *sliderRectangleNode;
	WERectangleNode *sliderBorderRectangleNode;

private slots:
	void recalculate();
	void longPressTimerTimeout();
};
} //End of namespace WE
#endif // QT_QT_QUICK_WE_SCROLLBAR_WIDGET_H
