/**
 *	@file can_windows_peak.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Interface functions for accessing peak CAN hardware.
 *
 *	This file provides functions to set options to peak
 *	CAN hardware using the Windows peak CAN driver.
 *	The class defined in this file inherits from the abstract
 *	class WE_CAN, a factory pattern is realized here.
 */

#ifndef CAN_WINDOWS_PEAK_H_
#define CAN_WINDOWS_PEAK_H_

#include <CanApi2.h>
#include <vector>
#include <map>

#include "../include/wecan.h"

namespace WE
{

struct PeakNet
{
	HCANHW hardwareHandle;
	HCANNET net;
	unsigned int baudrate;
	unsigned int peakBaudrate;
	std::string netName;
	unsigned int port;
};

class WECANWindowsPeak : public WECAN
{
public:
	/**
	 *	@fn WECANWindowsPeak()
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Standard constructor of WECANWindowsPeak class
	 */
	WECANWindowsPeak();

	/**
	 *	@fn virtual ~WECANWindowsPeak()
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Destructor of WECANWindowsPeak class
	 */
	virtual ~WECANWindowsPeak();

	/**
	 *	@fn setBaudrate(unsigned int port, unsigned int selectedBaudrate)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Sets the baudrate of a CAN port
	 *
	 *	@param[in] port	The port number starting by 1
	 *	@param[in] selectedBaudrate	Baudrate to set the
	 *			selected port to.
	 *
	 *	@returns WE_SUCCESS if setting the baudrate was successful.
	 *		WE_INIT_ERROR if initializing the CAN port with selected
	 *			baudrate was not successful.
	 *		WE_DEVICE_ERROR if opening the device was not successful
	 *			or the selected bitrate could not be retrieved.
	 *		WE_GENERAL_ERROR if something went completely wrong.
	 *			Should never happen.
	 *
	 *	@see libpcan.h
	 *
	 *	This function uses the Linux peak CAN driver to set
	 *	the baudrate of a CAN port. Overwritten virtual function
	 *	of WE_CAN
	 */
	int setBaudrate(unsigned int port, unsigned int selectedBaudrate) override;

	//not supported:
	virtual int getBaudrate(unsigned int port) const override;


	static std::vector<struct PeakNet> availableNets;	//!< Vector containing all available Peak CAN nets
	static std::map<unsigned int, HCANHW> portMap;		//!< Map containing the hardware handles sorted by port
	static std::map<unsigned int, HCANNET> activeNets;	//!< the baudrates of these nets were set so we consider them as active
private:
	int setBaudrateInternal(unsigned int port, unsigned int selectedBaudrate);
	static unsigned int getUserBaudrateFromPeakBitRate(int peakBitRate); //!< Provides the face value of baud rate for the given peak bit rate

	unsigned int baudrate;	//!< The baudrate that shall be set to this CAN port



};

}//end of namespace WE

#endif /* CAN_WINDOWS_PEAK_H_ */
