/**
 *	@file we_can_raw.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Definition of the WECANRaw class that acts as a raw
 *		"CAN stack" for WEAPI.
 *
 *	In this file the interface for using a raw CAN stack is defined.
 *	See the function descriptions for more information.
 */

#ifndef WE_CAN_RAW_H_
#define WE_CAN_RAW_H_

#include <memory> //unique_ptr

#include "we_can_message.h"
#include "../../../os_abstraction/endpoint_manager/include/we_endpoint.h"

namespace WE
{

/**
 *	@class WECANRaw
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief WEAPI's raw CAN stack interface
 *
 *	@namespace WE
 *
 *	@see socket_can_interface.h for definitions of the Linux PC CAN interface.
 *
 *	This class can act as a raw CAN stack. It provides
 *	methods to initialize the CAN interface, send and receive
 *	CAN methods.
 *	This class is abstract, so the create function has to be called
 *	to generate an instance of a platform specific CAN interface.
 *	Where available it uses the socketCAN interface for
 *	CAN communication.
 *	This class inherits WEEndpoint so that waiting for received messages can be
 *	handled by WEEndpointManager. Be sure to implement the WEEndpoint functions in
 *	the implementing class of WECANRaw.
 */
class WECANRaw : public WEEndPoint
{
public:

	/**
	 *	@fn WECANRaw()
	 *
	 *	@brief End point constructor.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	This constructor will open the connection to the controlled endpoint.
	 *
	 */
	WECANRaw()
	: WEEndPoint(false)
	{
		//portNo is initialized at declaration
	}

	/**
	 *	@fn ~WECANRaw()
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Standard destructor of the WECANRaw class
	 *
	 *	Standard destructor deletes the instance of CAN interface.
	 */
	virtual ~WECANRaw() = default;

	/**
	 *	@fn virtual int initialize(unsigned int portNo, unsigned int baudrateToSet, const char * interfaceName) = 0;
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Initializes the WECANRaw stack
	 *
	 *	Mainly calls the interfaces interfaceInitialize function
	 *	to initialize the chosen CAN interface. The name of the interfaces
	 *	has to be provided as parameter
	 *
	 *	@param[in] portNo Number of the CAN port
	 *	@param[in] baudrateToSet The baudrate to be set in KBit/s
	 *	@param[in] interfaceName Name of the interface
	 *		that shall be initialized.
	 *
	 *	@returns 	WE_SUCCESS if message was sent out successfully
	 *			WE_INTERFACE_ERROR if instance of interface is faulty
	 *			else: return code of interface function
	 *
	 *	@see can_interface.h socket_can_interface.h
	 */
	virtual int initialize(unsigned int p_portNo, unsigned int p_baudrateToSet, const char * p_interfaceName) = 0;

	/**
	 *	@fn virtual int initialize(unsigned int p_portNo) = 0
	 *
	 *	@brief Initialize a CAN port without setting the baudrate
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_portNo Number of the CAN port that shall be used
	 *
	 *	@returns See return codes of the functions that implement this interface
	 */
	virtual int initialize(unsigned int p_portNo) = 0;

	/**
	 *	@fn int send(WECANMessage * messageToSend)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Sends the provided CAN message.
	 *
	 *	Passes the provided CAN message to the interface's
	 *	send function.
	 *
	 *	@param[in] messageToSend struct wit CAN message data
	 *
	 *	@returns 	WE_SUCCESS if message was sent out successfully
	 *			WE_INTERFACE_ERROR if instance of interface is faulty
	 *			else: return code of interface function
	 *
	 *	@see can_interface.h socket_can_interface.h
	 */
	virtual int send(WECANMessage * p_messageToSend) = 0;

	/**
	 *	@fn virtual int receive(WECANMessage * p_messageToStoreReceivedDataTo) = 0
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Receives a CAN message and stores it provided
	 *		WECANMessage for further processing. This call is the
	 *		prefered one (using endPoints) instead of the above blockign one.
	 *
	 *	@param p_messageToStoreReceivedDataTo A pointer to a struct
	 *		in which the received CAN message gets stored.
	 *
	 *	@returns WE_SUCCESS if a new CAN message was successfully received
	 *	@returns WE_TIMEOUT_ERROR if select timed out before a message was received
	 *	@returns WE_INTERFACE_ERROR if select returned an error (other than timeout)
	 *
	 *	@see int interfaceReceive(void)
	 *
	 *	@note This is the prefered function for reading messages. Used within onEndPointReadble()
	 *	      it will never block.
	 *
	 *	The function makes reads a socket. When beeing called from onEndPointReadable()
	 *	function, this call will never block. Otherwise, if no data are readable, this
	 *	function will block until data become available.
	 */
	virtual int receive(WECANMessage * p_messageToStoreReceivedDataTo) = 0;

	/**
	 *	@fn void setPolling(bool polling, unsigned int timeout)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Sets the reading of CAN messages to blocked or polling
	 *
	 *	The reading behavior gets set to polling or blocking.
	 *	Timeout can be set for polling.
	 *
	 *	@param[in] polling Enable or disables polling read
	 *	@param[in] timeout polling timeout in milliseconds
	 *
	 *	@returns void
	 *
	 *	@see can_interface.h
	 */
	virtual void setPolling(bool p_polling, unsigned int p_timeout) = 0;

	virtual int addIDFilter(unsigned int p_ID, unsigned int p_mask) = 0;

	static int getRxPackages(unsigned int p_port, int &p_packages);
	static int getTxPackages(unsigned int p_port, int &p_packages);

	/**
	 *	@fn static WECANRaw * create(int numberOfRetries = 0, int retryTimeout = 0, int p_hardwareType = 0)
	 *
	 *	@brief Create an instance of a CAN RAW interface
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param numberOfRetries How often the send operation will be repeated when send error occurs
	 *	@param retryTimeout If send error occurs, we wait for retryTimeout microseconds  before
	 *		we try again to send (only if numberOfRetries > 0)
	 *
	 *	@returns Pointer to an unitialized CAN RAW stack
	 *
	 *	@warning Call initialize function after creating an instance
	 *
	 *	This function creates an instance of a WE CAN RAW stack.
	 *	To initialize the stack call the initialize function afterwards.
	 */
	static std::unique_ptr<WECANRaw> create(int p_numberOfRetries = 0, int p_retryTimeout = 0, int p_hardwareType = 0);

	/**
	 *	@fn static WECANRaw * create(int portNo, unsigned int baudrateToSet, const char * interfaceName, int numberOfRetries = 0, int retryTimeout = 0, int p_hardwareType = 0)
	 *
	 *	@brief Create and initialize a CAN RAW stack
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] portNo Number of the CAN port
	 *	@param[in] baudrateToSet The baudrate to be set in KBit/s
	 *	@param[in] interfaceName Name of the interface that shall be initialized.
	 *	@param numberOfRetries How often the send operation will be repeated when send error occurs
	 *	@param retryTimeout If send error occurs, we wait for retryTimeout microseconds  before
	 *		we try again to send (only if numberOfRetries > 0)
	 *
	 *	This function creates an instance of a CAN RAW stack
	 *	and initializes the stack with the given parameters.
	 */
	static std::unique_ptr<WECANRaw> create(int p_portNo, unsigned int p_baudrateToSet, const char * p_interfaceName, int p_numberOfRetries = 0, int p_retryTimeout = 0, int p_hardwareType = 0);

	unsigned int getPortNumber(void) const
	{
		return this->portNo;
	}

	static int hardwareInterfaceType;

	static int debugSection;

protected:
	unsigned int portNo{0};
};
}//end of namespace WE


#endif /* WE_CAN_RAW_H_ */
