/**
 *	@file we_feature_detection.h
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2012
 *
 *	@brief Figure out if the Opus panel is equipped with a certain feature or not.
 *
 *	This file defines class WEFeatureDetection which only contains static methods to figure out
 *	if this Opus panel is equipped with certain features.
 *
 *	Typically the methods make use of the underlying BSP's detection methods which requires them to
 *	have a device specific implementation.
 *
*/

#ifndef WE_FEATURE_DETECTION_H_
#define WE_FEATURE_DETECTION_H_



namespace WE
{


class WEFeatureDetection
{
public:

	/**
	 *	@fn static bool hasEncoder( void )
	 *
	 *	@brief Figure out whether this panel has an encoder or not.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns true if this Opus panel has an encoder
	 *	@returns false otherwise
	 *
	 */
	static bool hasEncoder( void );


	/**
	 *	@fn static bool hasTouchscreen( void )
	 *
	 *	@brief Figure out whether this panel has a touch screen or not.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns true if this Opus panel has an touch screen
	 *	@returns false otherwise
	 *
	 */
	static bool hasTouchscreen( void );


	/**
	 *	@fn static bool hasKeyboard( void )
	 *
	 *	@brief Figure out whether this panel has keys or not.
	 *
	 *	@author mlenhart@topcon.com
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns true if this Opus panel has an keyboard
	 *	@returns false otherwise
	 *
	 *	This additional method was introduced since on EGSPRO platform's
	 *	first OS distributions there is/was no proper version detection but
	 *	at least for A6G2 the keyboard could be used as indicator for ECO
	 *	versus standard devices nevertheless.
	 *
	 */
	static bool hasKeyboard( void );

	/**
	 * Runtime check if the device that we are running on is using a single core CPU
	 *
	 * @returns true if a single core CPU was detected
	 * @returns false if a multi core CPU was detected
	 */
	static bool isRunningOnSingleCore(void);

	/**
	 *	@fn WEFeatureDetection()
	 *
	 *	@brief Contructor is deleted since this class only contains static methods.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 */
	WEFeatureDetection() = delete;
	~WEFeatureDetection() = delete;
};	// class WEFeatureDetection
}	// namespace


#endif /* WE_FEATURE_DETECTION_H_ */
