/**
 *	@file we_io.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief WEAPI interface functions for accessing I/Os on an OPUS platform
 *
 *	@warning I/Os are not supported on a regular PC. Therefore, when compiling for a PC,
 *		all functions will return a WE_NOT_SUPPORTED_ERROR
 *	@warning Before including this header, we_datatypes_hardware has to be included!
 *
 *	The interface defined in this file provides functions for accessing any
 *	I/Os on an OPUS platform.
 */

#ifndef WE_IO_H_
#define WE_IO_H_

#include <limits>

namespace WE
{

class WEEventQueue;

/**
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Enum that lists types of I/Os
 *
 *	weIOType lists types of IOs that can be configured
 */
typedef enum
{
	IO_DIGITAL_INPUT,
	IO_DIGITAL_OUTPUT,
	IO_ANALOG_INPUT,
	IO_ANALOG_OUTPUT,
	IO_POWER_SENSOR,
	IO_LIGHT_SENSOR,
	IO_TEMPERATURE_SENSOR,
	IO_RADAR_SENSOR,
	IO_FREQUENCY_INPUT,
	IO_IGNITION_INPUT,
	IO_NO_TYPE
}weIOType;

/**
 *	@struct WEAnalogPortOptions
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Defines some options for events received for analog inputs
 *
 *	@namespace WE
 */
struct WEAnalogPortOptions
{
	unsigned short changeHysteresis; //!< Hysteresis for a value change
	unsigned short limitLow;	//!< Lower border of event signaling
	unsigned short limitHigh;	//!< Upper border of event signaling
};

/**
 *	@class WEInputOutput
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Interface for accessing I/O functionality
 *
 *	@namespace WE
 *
 *	This class provides an platform independent interface for accessing
 *	the I/Os of an OPUS platform.
 *	Use the create method to create an instance of the hardware interface.
 */
class WEInputOutput
{
public:
	WEInputOutput() = default;

	WEInputOutput(weIOType p_iotype, unsigned short p_portNumber);

	/**
	 *	@fn ~WEInputOutput()
	 *
	 *	@brief Standard destructor
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	Frees memory
	 */
	virtual ~WEInputOutput();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEInputOutput(const WEInputOutput &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEInputOutput & operator=(const WEInputOutput &) = delete;

	/**
	 *	@fn int initialize(weIOType ioType, unsigned short portNumber)
	 *
	 *	@brief Initialization function for a digital input
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] ioType type of the I/O to use
	 *	@param[in] portNumber Number of the port to use (e.g. 1 if you want to use digital input 1)
	 *
	 *	@returns WE_SUCCESS if initialization was successful
	 *	@returns WE_NOT_SUPPORTED_ERROR on platforms that do not support I/Os (like a PC)
	 *	@returns WE_DEVICE_ERROR On errors on the device file
	 *	@returns WE_INIT_ERROR If initialization did not work as intended (e.g. when passing wrong parameters)
	 *
	 *	@see WEInputOutput()
	 *
	 *	@warning Call this function after using the standard constructor!
	 *
	 *	This function chooses the correct device file by examining the parameters.
	 *	The device file gets opened and is ready for further use.
	 */
	int initialize(weIOType p_ioType, unsigned short p_portNumber);

	/**
	 *	@fn weIOType getType(void)
	 *
	 *	@brief Returns type of I/O
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Configured type of I/O
	 *
	 *	@see weIOType
	 *
	 *	This function just returns the type of I/O that was configured for this instance.
	 */
	weIOType getType(void) const
	{
		return type;
	}

	/**
	 *	@fn unsigned short getPortNumber(void)
	 *
	 *	@brief Returns number of the I/O
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Number of the I/O
	 *
	 *	This function returns the number of the I/O, e.g. 1 when using digitalInput1
	 */
	unsigned short getPortNumber(void) const
	{
		return portNumber;
	}

	/**
	 *	@fn int getState(bool * storeValue)
	 *
	 *	@brief Returns the state of a digital in- or output
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] storeValue The state of the digital I/O is stored to this address
	 *
	 *	@returns WE_SUCCESS if reading the state was successfull
	 *	@returns WE_READ_ERROR if the state could not be read
	 *	@returns WE_NOT_SUPPORTED_ERROR on a platform that does not support I/Os
	 *
	 *	This function returns the state of a digital in- or output.
	 *	The state is true if the in- or output level is set to on/high
	 *	and false if it is off/low.
	 *
	 *	Can not be made a const function since it remembers the last value in a member
	 */
	int getState(bool * p_storeValue);

	/**
	 *	@fn int getValue(int * storeValue)
	 *
	 *	@brief Retrieves the value of an I/O
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] storeValue The value of the I/O is stored to this address
	 *
	 *	@returns WE_SUCCESS if reading the value was successfull
	 *	@returns WE_READ_ERROR if the value could not be read
	 *	@returns WE_NOT_SUPPORTED_ERROR on a platform that does not support I/Os
	 *
	 *	This function retrieves the value of an I/O and stores it to the provided address.
	 *	The retrieved value is in mV.
	 *
	 *	Can not be made a const function since it remembers the last value in a member
	 */
	int getValue(int * p_storeValue);

	/**
	 *	@fn int on(bool state)
	 *
	 *	@brief Switches the digital output to on or off
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] state true for on, false for off
	 *
	 *	@returns WE_SUCCESS if setting the state was successfull
	 *	@returns WE_WRITE_ERROR if the state could not be written
	 *	@returns WE_NOT_SUPPORTED_ERROR on a platform that does not support I/Os
	 *
	 *	With this function a digital output can be set to high/on (true) or low/off (false).
	 */
	int on(bool p_state);

	int configureInput(int p_port, int p_value, int p_hyteresis);

	static const int debugSection;

protected:
	unsigned int lastValue{std::numeric_limits<unsigned int>::max()}; //!< Last retrieved value of the I/O
	weIOType type{IO_NO_TYPE};	//!< Type of the I/O
	unsigned int portNumber{0xFFFF};	//!< Port number of the I/O


	std::string deviceFile{"/dev/null"};	//!< File name of the device file
	int deviceFD{-1};	//!< File descriptor for the device file

	int hysteresis{-1};
	int high{-1};
	int low{-1};
	int measureMode{0};

	static inline std::string ignitionInputPath{""};
};
}//end of namespace WE

#endif /* WE_IO_H_ */
