/**
 *	@file we_netlink.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of WENetlink class
 *
 *	@see Class description below
 */


#ifndef WE_NETLINK_H_
#define WE_NETLINK_H_

#include <memory>
#include <map>
#include <set>

#include "../../../os_abstraction/endpoint_manager/include/we_endpoint.h"
#include "../../../os_abstraction/endpoint_manager/include/we_endpoint_mgr.h"
#include "../../../os_abstraction/endpoint_manager/include/we_endpoint_handler.h"

namespace WE {

class WEEvent;
class WEEndPointManager;
class WEEventQueue;

enum IOFamilies
{
	FAMILY_ANALOG,
	FAMILY_DI,
	FAMILY_DO
};

/**
 *	@class WENetlinkEndPoint
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief The WENetlinkEndPoint is a simple WEEndpoint implementation that uses netlink sockets for communication
 *
 *	@namespace WE
 */
class WENetlinkEndPoint : public WEEndPoint
{
public:
	/**
	 * Provide the generic netlink family name in the constructor.
	 * The endpoint will directly subscribe to that family.
	 * @param p_familyName The generic netlink family name to subscribe to.
	 */
	explicit WENetlinkEndPoint(const std::string & p_familyName);
	~WENetlinkEndPoint() = default;

	/**
	 * Just returns the netlink socked descriptor
	 * @return fd
	 */
	virtual int getReadEndPointID() const override;

	/**
	 * Just returns the netlink socked descriptor
	 * @return fd
	 */
	virtual int getWriteEndPointID() const override;

	/**
	 * Returns the connection status of the socket
	 * @return true if subscription and connection to the netlink family was successful, else: false
	 */
	virtual bool isConnected() const override;

	/**
	 * reconnect() is currently not implemented.
	 * @return WE_NOT_SUPPORTED_ERROR
	 */
	virtual int reconnect() override;
private:
	friend class WENetlink;	//!< Allow the Netlink class to access private properties
	std::string familyName;	//!< The generic netlink family name to which this endpoint is subscribed to
	int familyID;	//!< The generic netlink family ID to which this endpoint is subscribed to
	int fd;	//!< The socket descripto
	bool connected;	//!< Flag that tells if the socket is connected or not
};

/**
 *	@class WENetlink
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief WENetlink provides an interface to the WE generic netlink event interface
 *
 *	@namespace WE
 *
 *	The WENetlink class provides a way to subscribe to WE's generic Netlink interface.
 *	The supported hardware is: digital and analog inputs (including sensors) and digital outputs.
 *	How to use the class:
 *	- Get or create an instence with the instance() function
 *	- Add family names to which the class should subscribe with addFamily()
 *	- Add the event queue to which the WEEvents received converted from netlink events shall be send to with addEventQueue()
 *	- Add the endpoint manager that shall manage all enpoints with settEndPointManager()
 *	- call addAllEndpointsToEndpointManager() so that all created WENetlinkEndPoints will be added to the endpoint manager
 *	When all this is done, the event queue will be filled with an event as soon as one of the endpoints is readable
 *
 *	WENetlink is a singleton. Use the instance() function!
 */
class WENetlink : public WEEndPointHandler
{
public:
	/**
	 * Get the instance of this singleton calss
	 * @return The one and only instance of this class.
	 */
	static WENetlink * instance();

	/**
	 * Destructor. Frees memory.
	 */
	virtual ~WENetlink() = default;

	/**
	 * Function that is called each time one of the WENetlinkEndPoints is readable.
	 * The read netlink event is converted to an apropriate WEEvent (IOToggleEvent) and sent to the
	 * event queue.
	 * @param p_endPoint	Pointer to the endpoint that is readable
	 * @param p_endPointMgr	Pointer to the endpoint manager. Not used.
	 */
	virtual void onEndPointReadable(std::shared_ptr<WEEndPoint> p_endPoint, WEEndPointManager* p_endPointMgr) override;

	/**
	 * Reconnect is currently not supported.
	 * @param p_endpoint Not used
	 */
	virtual void onEndPointReconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;

	/**
	 * Nothing is done if the connection was lost => Not supported
	 * @param p_endpoint Not used
	 */
	virtual void onEndPointDisconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;

	/**
	 * Add a generic netlink family name. This will create a new WENetlinkEndPoint that will subscribe
	 * itself to the WE general netlink interface.
	 * @param p_familyName The family name to which a subscription shall be made.
	 * @return
	 */
	int addFamily(const std::string & p_familyName);

	/**
	 * Use this function to add all endpoints that were created with addFamily to the
	 * endpoint manager that was set with setEndPointManager().
	 */
	void addAllEndpointsToEndpointManager();

	/**
	 * Add the event queue to which all events shall be sent
	 * @param p_queue Pointer to the event queue
	 */
	void addEventQueue(WEEventQueue * p_queue);

	/**
	 * Set the endpoint manager that shall be used.
	 * @param p_manager Pointer to the endpoint manager
	 */
	void setEndPointManager(WEEndPointManager * p_manager);

	static const char * getFamilyName(int p_familyID);

private:
	/**
	 * Private constructor. Initializes members.
	 * Private because of singleton design pattern.
	 */
	WENetlink() = default;

	void handleDigitalInputEndpoint(std::shared_ptr<WENetlinkEndPoint> p_endpoint);
	void handleDigitalOutputEndpoint(std::shared_ptr<WENetlinkEndPoint> p_endpoint);
	void handleAnalogInputEndpoint(std::shared_ptr<WENetlinkEndPoint> p_endpoint);

	std::map<std::string, std::shared_ptr<WENetlinkEndPoint>> familyEndpoints{};	//!< All endpoints are stored in this map sorted by theyr family names
	std::set<std::shared_ptr<WENetlinkEndPoint>> endpoints{};	//!< A vector containing pointer to all existing WENetlinEndpoints
	WEEventQueue * eventQueue{nullptr};	//!< Pointer to the event queue to which all events will eb sent
	WEEndPointManager * endpointManager{nullptr};	//!< Pointer to the endpoint manager to which all netlink endpoints belong to
};
}  // namespace WE


#endif /* WE_NETLINK_H_ */
