#ifndef THARDWAREINTERFACE_H
#define THARDWAREINTERFACE_H

#include <QObject>

class THardwareInterface : public QObject
{
  Q_OBJECT

public:
  /**
   * \brief Get the consoles serial number.
   * \return The serial number is a string of 20 characters
   */
  virtual QString GetSerialNumber() = 0;

  // Ambient Light
  /**
   * \brief Get the ambient light value.
   * \return The ambient light as a percentage between 0 and 100.
   */
  virtual double GetAmbientLight() = 0;
  /**
   * \brief Set the hysterisis of the ambient light input.
   * \param[in] hysterisis The hysterisis value.
   * \return True if the set succeeded, false if there was an error.
   */
  virtual bool SetAmbientLightHysterisis(double hysterisis) = 0;
  /**
   * \brief Set the minimum threshold the ambient light can be before emitting AmbientLightMinThresholdReached().
   * \param[in] threshold The ambient light minimum threshold.
   * \return True if the set succeeded, false if there was an error.
   */
  virtual bool SetAmbientLightMinThreshold(double threshold) = 0;
  /**
   * \brief Set the maximum threshold the ambient light can be before emitting AmbientLightMaxThresholdReached().
   * \param[in] threshold The ambient light maximum threshold.
   * \return True if the set succeeded, false if there was an error.
   */
  virtual bool SetAmbientLightMaxThreshold(double threshold) = 0;

  // Brightness
  /**
   * \brief Get the screen brightness.
   * \return The screen brightness as a percentage between 0 and 100.
   */
  virtual double GetBrightness() = 0;
  /**
   * \brief Set the screen brightness.
   * \param[in] brightness The screen brightness as a percentage between 0 and 100.
   * \return True if the set succeeded, false if there was an error.
   */
  virtual bool SetBrightness(double brightness) = 0;

  // Power
  /**
   * \brief Get the state of the screen backlight power
   * \return True if the power is suspended, false if it isn't.
   */
  virtual bool IsPowerSuspended() = 0;
  /**
   * \brief Resume the backlight power.
   * \return True if the resume succeeded, false if there was an error.
   */
  virtual bool ResumePower() = 0;
  /**
   * \brief Suspend the backlight power.
   * \return True if the resume succeeded, false if there was an error.
   */
  virtual bool SuspendPower() = 0;

  // Digital I/O
  /**
   * \brief Disable digital output for a pin.
   * \param[in] pin The pin to disable digital output on.
   * \return True if the disable succeeded, false if there was an error.
   */
  virtual bool DisableDigitalOutput(int pin) = 0;
  /**
   * \brief Enable digital output for a pin.
   * \param[in] pin The pin to enable digital output on.
   * \return True if the enable succeeded, false if there was an error.
   */
  virtual bool EnableDigitalOutput(int pin) = 0;
  /**
   * \brief Get the status of a digital input pin.
   * \param[in] pin The pin to get the status of.
   * \return True if the pin is powered/on, false if it is unpowered/off.
   */
  virtual bool IsDigitalInputOn(int pin) = 0;

  // Radar
  /**
   * \brief Set the frequency of the radar output.
   * \param[in] frequency The frequency in Hertz.
   * \return True if the set succeeded, false if there was an error.
   */
  virtual bool SetRadarFrequency(uint frequency) = 0;

  // Temperature
  /**
   * \brief Get the console temperature.
   * \return The console temperature in degrees celcius.
   */
  virtual double GetTemperature() = 0;

  // Voltage
  /**
   * \brief Get the supply voltage of the console.
   * \return The console supply voltage in Volts.
   */
  virtual double GetSupplyVoltage() = 0;
  /**
   * \brief Get the battery voltage of the console.
   * \return The console battery voltage in Volts.
   */
  virtual double GetBatteryVoltage() = 0;

  /**
   * \brief Get the presence of physical controls for the Virtual Terminal.
   * \return True if the console has hardware buttons to control the Virtual Terminal.
   */
  virtual bool HasPhysicalVTControls() = 0;

Q_SIGNALS:
  // Ambient Light
  /**
   * \brief Emitted when the ambient light value drops below the set minimum threshold.
   * \param[out] actualValue The actual ambient light value.
   */
  void AmbientLightMinThresholdReached(double actualValue);
  /**
   * \brief Emitted when the ambient light value rises above the set maximum threshold.
   * \param[out] actualValue The actual ambient light value.
   */
  void AmbientLightMaxThresholdReached(double actualValue);

  // Digital I/O
  /**
   * \brief Emitted when the status of a digital input pin changes.
   * \param[out] pin The pin that changed.
   * \param[out] status The pin's current status, true (on) or false (off).
   */
  void DigitalInputStatusChanged(int pin, bool status);

  // Power Button
  /**
   * \brief Emitted when the power button has had a short press.
   */
  void PowerButtonShortPress();

  // Temperature
  /**
   * \brief Emitted when the console temperature changes.
   * \param[out] temperature The console's current temperature in degrees celcius.
   */
  void TemperatureChanged(double temperature);

  // Voltage
  /**
   * \brief Emitted when the console supply voltage changes.
   * \param[out] voltage The console's current supply voltage in Volts.
   */
  void SupplyVoltageChanged(double voltage);
  /**
   * \brief Emitted when the console battery voltage changes.
   * \param[out] voltage The console's current battery voltage in Volts.
   */
  void BatteryVoltageChanged(double voltage);

  /**
   * \brief Emitted when the system resumes after being put to sleep/suspended to RAM.
   * Should we implement that somehow?! Just figured out that it exists when adding serial number support.
   */
  void SystemResumed();
};

#endif
