/*
 * we_touch_calibration.h
 *
 *  Created on: Sep 22, 2016
 *      Author: dsl
 */

#ifndef SRC_HARDWARE_TOUCH_CALIBRATION_INCLUDE_WE_TOUCH_CALIBRATION_H_
#define SRC_HARDWARE_TOUCH_CALIBRATION_INCLUDE_WE_TOUCH_CALIBRATION_H_

#include <QObject>

#include <QThread> //include QThread instead of forward declaration so that the application that includes this header file does not have to include QThread itself

namespace WE {

/**
 * Utility class for starting touchscreen calibration.
 * Blocking usage:
 * - Create an instance
 * - call startTouchCalibration()
 * NonBlockingUsage:
 * - Create an instance
 * - connect your slot to signal touchCalibrationFinished
 * - call startTouchCalibration()
 * - after calibration has finished the slot that was connected to touchCalibrationFinished is called
 */
class WETouchCalibration : public QObject
{
	Q_OBJECT
public:
	WETouchCalibration();
	virtual ~WETouchCalibration();

	/**
	 * Check if touchscreen calibration is supported on the currently running device
	 */
	bool isTouchCalibrationSupported(void) const;

	/**
	 * Start touchscreen calibration. Internally calls isTouchCalibrationSupported().
	 *
	 * @param p_nonBlocking if p_nonBlocking is set to true, the call to this function will return directly
	 * 	and the calibration process is started in the background (drawn in foreground). Connect to signal
	 * 	touchCalibrationFinished to be signaled when calibration has finished.
	 * 	If p_nonBlocking is set to false, this function will only return after the calibration process has been finished.
	 *
	 * @return WE_SUCCESS if touchscreen calibration was started.
	 * 	WE_NOT_SUPPORTED_ERROR if touchscreen calibration is not supported
	 * 	WE_GENERAL_ERROR if starting touchscreen calibration failed
	 */
	int startTouchCalibration(bool p_nonBlocking);

signals:
	/**
	 * Signal set is sent when touch calibration has finished and startTouchCalibration(true) was called.
	 */
	void touchCalibrationFinished(void);

private slots:
	/**
	 * Private slot that is internally called by CalibrationThread after touch calibration finished and startTouchCalibration(true) was called.
	 */
	void nonBlockingCalibrationFinished(void);

private:
	QThread * calibrationThread;
};

/**
 * Thread that is startef when startTouchCalibration(true) was called. Will start the calibration process and runs until the calibration is done.
 */
class CalibrationThread : public QThread
{
	Q_OBJECT
public:
	void run() override;
	CalibrationThread() = default;
	virtual ~CalibrationThread() = default;
};
}  // namespace WE


#endif /* SRC_HARDWARE_TOUCH_CALIBRATION_INCLUDE_WE_TOUCH_CALIBRATION_H_ */
