

#ifndef SRC_HARDWARE_VIDEO_INPUT_INCLUDE_WE_VIDEO_INPUT_INTERFACE_H_
#define SRC_HARDWARE_VIDEO_INPUT_INCLUDE_WE_VIDEO_INPUT_INTERFACE_H_

#include <mutex>
#include <list>
#include "../../../common/include/we_rotation.h"

class QQuickItem;
class QTimer;

namespace WE {

typedef enum
{
	MIRRORED_NONE,	//!< Video is not mirrored
	MIRRORED_HORIZONTAL,	//!< Video is horizontally mirrored
	MIRRORED_VERTICAL,	//!< Video is vertically mirrored
	MIRRORED_BOTH,		//!< Video is horizontally and vertically mirrored
	MIRROR_NOT_SUPPORTED	//!< Mirroring is not supported
}WEVideoMirrorSetting;

typedef enum
{
	VIDEO_TYPE_GSTREAMER
}WEVideoInputType;

typedef enum StreamType
{
	STREAM_TYPE_UDP_RTP_H264	= 0,
	STREAM_TYPE_UDP_RTP_MJPEG	= 1,
	STREAM_TYPE_RTSP_H264		= 2,
	STREAM_TYPE_ANALOG		= 3,
	STREAM_TYPE_NOT_DEFINED		= 4
}StreamType;

enum VideoCommands
{
	VIDEO_START,
	VIDEO_STOP,
	VIDEO_CHANGE_INTERFACE_OR_PORT
};

class WEVideoInputInterface
{
public:
	static WEVideoInputInterface * getInstance(WEVideoInputType p_videoType, const std::string & p_identifier = "ID_UNUSED");
	virtual ~WEVideoInputInterface() = default;

	virtual int init(void) = 0;

	virtual WEVideoMirrorSetting getMirrorSetting(unsigned short p_cameraNumber) = 0;

	virtual WERotation getRotationSetting(unsigned short p_cameraNumber) = 0;


	virtual int getHeight(unsigned short p_cameraNumber) = 0;

	virtual int getWidth(unsigned short p_cameraNumber) = 0;

	virtual int getXPosition(unsigned short p_cameraNumber) = 0;

	virtual int getYPosition(unsigned short p_cameraNumber) = 0;

	virtual int setBrightness(unsigned short p_cameraNumber, int p_brightness) = 0;

	virtual int setContrast(unsigned short p_cameraNumber, int p_contrast) = 0;

	virtual int setHue(unsigned short p_cameraNumber, int p_hue) = 0;

	virtual int setMirror(unsigned short p_cameraNumber, WEVideoMirrorSetting p_mirrorSetting) = 0;

	virtual int setRotation(unsigned short p_cameraNumber, WERotation p_rotation) = 0;

	virtual int setPosition(unsigned short p_cameraNumber, int p_xPosition, int p_yPosition) = 0;

	virtual int setSaturation(unsigned short p_cameraNumber, int p_saturation) = 0;

	virtual int setSize(unsigned short p_cameraNumber, int p_width, int p_height) = 0;

	/**
	 * Issue command to start the video with the current settings.
	 * @warning The start command is not executed directly. Call handleQueuedCommands()
	 * from the same thread that created an instance of this class afterwards.
	 */
	virtual int start(unsigned short p_cameraNumber) = 0;

	/**
	 * Issue command to stop the video.
	 * @warning The stop command is not executed directly. Call handleQueuedCommands()
	 * from the same thread that created an instance of this class afterwards.
	 */
	virtual int stop(unsigned short p_cameraNumber) = 0;

	virtual bool isRunning(unsigned short p_cameraNumber) = 0;

	/**
	 * Issue command to change the video input / UDP port of this interface.
	 * @warning The change command is not executed directly. Call handleQueuedCommands()
	 * from the same thread that created an instance of this class afterwards.
	 */
	virtual int setCurrentInterface(int p_cameraNumber) = 0;
	virtual int getCurrentInterface(void) const = 0;

	virtual bool isInitialized() = 0;

	/**
	 * When calling start(), stop() or setCurrentInterface(), the actions are not performed directly
	 * but are stored in a queue. The reason for this is to make the video handling more thread safe.
	 * To make sure that these commands are executed, call this function regularly from the thread that
	 * created the instance of this class.
	 */
	virtual void handleQueuedCommands(void) = 0;

	bool isStopRequested(void);

	static void doQuit(void);

	/**
	 * WE need to know the main window because we have to re-parent our
	 * video item to the main window when the original parent
	 * of the video widget gets destroyed (e.g. page switch)
	 */
	static void setMainWindow(QQuickItem * p_window);

protected:
	static inline QQuickItem * mainWindow{nullptr};
	std::list<int> queuedCommands;
	std::mutex queueMutex;

	void pushCommandToQueue(int p_command);
};
}  // namespace WE



#endif /* SRC_HARDWARE_VIDEO_INPUT_INCLUDE_WE_VIDEO_INPUT_INTERFACE_H_ */
