/**
 *	@file we_endpoint_handler.h
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd Faisal
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief The interface for defining the handler that is invoked everytime an endpoint is
 *			found readable or reconnected.
 *
 *	@see WEEndPointManager, WEEndPoint
 *
 */

#ifndef WE_ENDPOINT_HANDLER_H_
#define WE_ENDPOINT_HANDLER_H_

#include <memory>

namespace WE
{

class WEEndPointManager;
class WEEndPoint;

/**
 *	@class WEEndPointHandler
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd. Faisal
 *
 *	@brief The interface for defining the handler that is invoked everytime an endpoint is
 *		found readable or reconnected.
 *
 *	@namespace WE
 *
 */
class WEEndPointHandler
{
public:
	/**
	 *	@fn onEndPointReadable(WEEndPoint* p_endPoint, WEEndPointManager* p_endPointMgr)
	 *
	 *	@brief Invoked everytime from the waitForEndPoints() if the corresponding endpoint
	 *		becomes readable.
	 *
	 *	@param p_endPoint The End point that is found to be readable.
	 *	@param p_endPointMgr The End Point manager managing the endpoint.
	 *
	 *	Generally it is invoked when a connection request is received on a server endpoint
	 *	or when a message is received on a client endpoint. For a server endpoint
	 *	WEIPCServer::acceptConnection() inside this function should return successfully.
	 *	And for a client endpoint WEIPCEndPoint::receiveMessage() inside this function
	 *	should return successfully.
	 *	It might also be invoked when connection is broken but only once. This is becuase
	 *	the connection status is not the immediate status but based on the last access results.
	 *	In such a case WEIPCEndPoint::receiveMessage() will return WE_RECEIVE_ERROR, therefore
	 *	for a failing WEIPCEndPoint::receiveMessage() it should be checked if the endpoint is
	 *	still connected using endPoint->isConnected(). And if it is not connected the endPoint
	 *	could be removed from the endPointManager using WEEndPointManager::removeEndPoint().
	 */
	virtual void onEndPointReadable(std::shared_ptr<WEEndPoint> p_endPoint, WEEndPointManager* p_endPointMgr) = 0;

	/**
	 *	@fn virtual void onEndPointReconnected(WEEndPoint* p_endpoint) = 0
	 *
	 *	@brief This function gets called every time a endpoint gets re-connected
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_endpoint The endpoint which was reconnected
	 */
	virtual void onEndPointReconnected(std::shared_ptr<WEEndPoint> p_endpoint) = 0;

	/**
	 *	@fn virtual void onEndPointDisconnected(WEEndPoint* p_endpoint) = 0
	 *
	 *	@brief This function gets called every time a endpoint gets disconnected
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endpoint The endpoint which was disconnected
	 */
	virtual void onEndPointDisconnected(std::shared_ptr<WEEndPoint> p_endpoint) = 0;

	/**
	 *	@fn ~WEEndPointHandler()
	 *
	 *	@brief Default virtual destructor.
	 *
	 *	To avoid the warning.
	 */
	virtual ~WEEndPointHandler() = default;
};
}
#endif /* WE_ENDPOINT_HANDLER_H_ */
