/**
 *	@file we_endpoint_mgr.h
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd. Faisal
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief It defines the WEEndPointManager and WEEndPointHandler classes.
 *
 *	@see we_endpoint_mgr.cpp
 *
 *	WEEndPointManager class is used to simultaneously wait for multiple end points to become
 *	readable and invoke the corresponding handler when it becomes readable.
 *	This allows the users to handle multiple endpoints like IPC socket, CAN, Pipe/EventQueue, File within a single thread.
 */


#ifndef WE_ENDPOINT_MGR_H_
#define WE_ENDPOINT_MGR_H_

#include <memory>
#include <set>
#include "we_endpoint.h"

namespace WE
{

class WEEndPointHandler;

/**
 *	@class WEEndPointManager
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd. Faisal
 *
 *	@brief This class is used to simultaneously wait for multiple end points to become
 *	readable and invoke the corresponding handler when it becomes readable. This allows the
 *	users to handle multiple endpoints like IPC socket, CAN, Pipe/EventQueue, File within a single thread.
 *
 *	@namespace WE
 */
class WEEndPointManager
{
private:
	std::set<std::pair<std::shared_ptr<WEEndPoint>, WEEndPointHandler*>> endPointsList{};	//!< Set of endpoints and handler pair pointers.
	unsigned int lastHandledTime{0};	//!< Replacing 'static' variable in 'waitForEndPoints()'
	unsigned int lastReconnectionTime{0};	//<! Remember timestamp for reconnections per instance instead of 'static' for all of them

	/**
	 *	@fn void handleDisconnectedEndPoints(void)
	 *
	 *	@brief Check if any of the endpoints registered with the manager is disconnected and trigger the according action.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	If an endpoint is disconnected, this function checks if a reconnect is allowed. If so it
	 *	will try to reconnect the endpoint. Depending on the result, it will call following endpoint
	 *	handler functions: onEndPointReconnected(), onEndPointDisconnected()
	 *
	 */
	void handleDisconnectedEndPoints(void);

public:
	/**
	 *	@fn WEEndPointManager()
	 *
	 *	@brief Default constructor. Does nothing.
	 *
	 */
	WEEndPointManager() = default;

	/**
	 *	@fn ~WEEndPointManager()
	 *
	 *	@brief Destructor for the WEEndPointManager instances.
	 *
	 *	It removes all the entries from the endPointList and frees the
	 *	memory allocated to each entry.
	 */
	virtual ~WEEndPointManager();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEEndPointManager(const WEEndPointManager &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEEndPointManager & operator=(const WEEndPointManager &) = delete;


	enum waitTypes
	{
		WAIT_PERIOD_OR_EVENT = 0,
		WAIT_NO_WAIT,
		WAIT_FOREVER_OR_EVENT
	};

	/**
	 *	@fn waitForEndPoints(unsigned int p_timeout, unsigned int p_condition)
	 *
	 *	@brief Wait for the specified time for any of the managed endpoints to become readable.
	 *
	 *	@param p_timeout Method returns if WETick::getInstance()->getTickCount() is greater than this (if p_condition == WAIT_PERIOD_OR_EVENT).
	 *	@param p_condition Covers special cases by applying values from "enum waitTypes".
	 *
	 *	@returns WE_SUCCESS If any of the end point is found readable.
	 *	@returns WE_TIMEOUT_ERROR If no end point is found readable within the specified time.
	 *	@returns WE_INTERFACE_ERROR if an error occurred while waiting.
	 *
	 *	Wait for the specified time for any of the managed endpoints to become readable and
	 *	invoke their corresponding handlers added using addEndPoint().
	 *	This method only waits for the connected end points.
	 *	Note that "p_timeout" is a point of time and NOT a period.
	 *
	 *	Based on the value of p_condition this has three different modes:
	 *	p_condition == WAIT_PERIOD_OR_EVENT: Wait until an endpoint is readable or WETick::getTickCount() >= p_timeout
	 *		Since this calculation works even for an overflow of WETick the wait period is limited to 10000 ms.
	 *		This will catch cases where this method is called although p_timeout has just passed.
	 *	p_condition == WAIT_NO_WAIT: Only call the handler if the endpoint is already readable - don't wait at all
	 *	p_condition == WAIT_FOREVER_OR_EVENT: Wait until an endpoint is readable, no timeout
	 *
	 *	However the handler might be called only once when the connection is broken because
	 *	the connection status is based on last access results.
	 *
	 *	@see addEndPoint()
	 *	@see WETick::getTickCount()
	 */
	int waitForEndPoints(unsigned int p_timeout, unsigned int p_condition = WAIT_PERIOD_OR_EVENT);


	/**
	 *	@fn addEndPoint(WEEndPoint* endPointToAdd, WEEndPointHandler* handler)
	 *
	 *	@brief Register the endPoint and the handler with the end point manager.
	 *
	 *	@param endPointToAdd The End point to be added to the manager.
	 *	@param handler The corresponding handler that is to be invoked whenever the endpoint
	 *			is found to be readable.
	 *
	 *	@returns WE_SUCCESS If the endPoint is added sucessfully.
	 *	@returns WE_ALREADY_EXISTS_ERROR If an end point with the same EndPointID already exists in the list.
	 *	@returns WE_INVALID_PARAMETER_ERROR if endPoint is nullptr or handler is nullptr.
	 *
	 *	This method registers the endPoint and the handler with the end point manager.
	 *	The next call to waitForEndPoints() will also use the added endPoint.
	 */
	int addEndPoint(std::shared_ptr<WEEndPoint> p_endPointToAdd, WEEndPointHandler* p_handler);

	/**
	 *	@fn removeEndPoint(WEEndPoint* endPointToRemove)
	 *
	 *	@brief Unregister the endPoint from the end point manager.
	 *
	 *	@param endPointToRemove The End point to be removed from the manager.
	 *
	 *	@returns WE_SUCCESS If the endPoint is removed successfully.
	 *	@returns WE_NOT_FOUND_ERROR If an end point with the same EndPointID does not exists in the list.
	 *	@returns WE_INVALID_PARAMETER_ERROR if endPoint is nullptr.
	 *
	 *	Generally end points are required to be removed when the endpoint is no more connected.
	 *	This method unregisters the endPoint from the end point manager.
	 *	The next call to waitForEndPoints() will not use the removed endPoint.
	 */
	int removeEndPoint(std::shared_ptr<WEEndPoint> p_endPointToRemove);


	/**
	 *	@fn WEEndPointHandler * getEndPointHandlerByEndPoint(WEEndPoint * p_endPoint) const
	 *
	 *	@brief Retrieve the endpoint handler of an endpoint
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_endPoint The endpoint from which the handler shall be retrieved from
	 *
	 *	@returns Pointer to the endpoint handler used by p_endPoint
	 *	@returns nullptr if no handler/endpoint was found
	 */
	WEEndPointHandler * getEndPointHandlerByEndPoint(std::shared_ptr<WEEndPoint> p_endPoint) const;
};	// class
}	// namespace
#endif /* WE_ENDPOINT_MGR_H_ */
