/**
 *	@file we_ipc.h
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd. Faisal
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Definition of an interface to access to the underlying OS
 *			suppoprted IPC mechanism.
 *
 *	@see we_ipc.cpp
 *
 *	The classes defined in this file allow access to a IPC mechanism supported
 *	by the OS.
 */

#ifndef WE_IPC_H_
#define WE_IPC_H_

#include <string>
#include <mutex>
#include "../../endpoint_manager/include/we_endpoint.h"
#include "../../thread/include/we_thread.h"
//we define the common IPC message in a header file located in the DEI IPC api directory since it
//is only used there anyway and makes it a lot easier to provide a consistent api package
#include "../../../dei/dei_ipc_api/src/dei_ipc_data.h"

//using namespace std;

namespace WE
{
class WEIPCMessageArrivalListener;
class WEIPCServer;
class WEThread;

/**
 *	@class WEIPCEndPoint
 *
 *	@brief This class represents an End point in any IPC communication.
 *
 *	@namespace WE
 *
 *	@see WEIPCServer
 *
 *	The IPC mechanism might change based on the OS. For unix
 *	domain sockets are used, for windows it could be Message Queues.
 */
class WEIPCEndPoint : public WEEndPoint
{
	// Declare WEIPCServer as a friend class so that it can use
	// the private initialize() in acceptConnection()
	friend class WEIPCServer;

private:
	int endPointID{-1};			//!< Endpoint identifier
	std::string endPointAddress{};	//!< address of the endpoint - this process' private or public socket
	std::string peerAddress{};	//!< address of the peer endpoint - other process'
	long connectTimeoutMilliSec{-1};	//!< Timeout for connectTo in milli seconds
	long receiveTimeoutMilliSec{0};	//!< Receive Timeout in milli seconds
	long sendTimeoutMilliSec{0};	//!< Send Timeout in milli seconds
	bool connectStatus{false};		//!< Connection status of the endpoint
	bool isClient{false};			//!< Only clients will try to reconnect
	WEIPCMessageArrivalListener* messageArrivalListener{nullptr}; // Listener to be invoked when message arrives
	std::unique_ptr<WEThread> messageArrivalListenerThread{};	//!< Thread to be used messageArrivalListener is set
	std::mutex messageArrivalMutex{};	//!< mutex to synschronize the message arrival from the thread and from receiveMessage()

	/**
	 *	@fn messageArrivalListenerThreadFunc(void* args)
	 *
	 *	@brief Internal function to execute in the messageArrivalListenerThread.
	 *
	 *	@param[in] args Pointer to this WEIPCEndPoint to access it within the function.
	 *
	 *	It is an infinite loop which checks the link for any received messages.
	 *	It acquires a lock on the messageArrivalMutex at the start of every iteration
	 *	and releases it at the end of each iteration. Between this lock and release
	 *	it does not block forever untill the message is received rather blocks for
	 *	very small amount of time and if it is able to receive a complete message
	 *	within this time then it passes this message to messageArrivalListener->messageArrived().
	 */
	static void messageArrivalListenerThreadFunc(void* p_args);

	/**
	 *	@fn initialize(int endPointID, const std::string& endPointAddress, const std::string& peerEndPointAddress)
	 *
	 *	@brief Private initializer to be used with WEIPCServer only.
	 *			Initializes the end point using the given endPointID.
	 *
	 *	@param[in] endPointID Already created Identifier for the EndPoint
	 *	@param[in] endPointAddress String that shall be used to identify the
	 *			end point across processes in an IPC
	 *			communication. It can be a file path or
	 *			an arbitrary string that is unique across all
	 *			processes or endpoints in the system.
	 *	@param[in] peerEndPointAddress Address of the peer endpoint to
	 *				which this endpoint is connect to.
	 *
	 *	@returns WE_SUCCESS on successfully intializing the end point.
	 *	@returns WE_INIT_ERROR when initialization fails.
	 *	@returns WE_INVALID_PARAMETER_ERROR If endPointID is negative or
	 *						if endPointAddress.length is <= 0 or
	 *						if peerEndPOintAddress.length is <= 0.
	 *
	 *	No need to call connectTo() after this if endPointID is a valid ID.
	 *	sendMessage() and receiveMessage() can be called direcly after this.
	 */
	int initialize(int p_endPointID, const std::string& p_endPointAddress, const std::string& p_peerEndPointAddress);
	/**
	 *	@fn receiveMessage(WEMessage* messageReceived, long recvTimeoutMilliSec)
	 *
	 *	@brief Private methos used to receive a message coming from the destination end point across
	 *	the link which waits for the specified timeout period. It is used in receiveMessage(WEMessage*)
	 *	and the messageArrivalListenerThreadFunc().
	 *
	 *	@param[out] message pointer to store message received from the destination.
	 *					"dataLength" in the messageReceived should be the expected
	 *					data length and the "data" should be pre-allocated.
	 *	@param[in] recvTimeoutMilliSec Receive timeout in milli seconds. Use -1 for infinite timeout period.
	 *
	 *	@returns WE_SUCCESS on successfully sending the message.
	 *	@returns WE_TIMEOUT_ERROR when recieve times out before completely receving the message.
	 *	@returns WE_RECEIVE_ERROR when endpoint is not connected or receiving fails due to any other reason.
	 *	@returns WE_INVALID_PARAMETER_ERROR if messageReceived is nullptr or
	 *						messageReceived->dataLength is 0 or
	 *						messageReceived->data is nullptr.
	 *	@returns Also returns message with actual data length and the data received.
	 *
	 *	@see WEIPCEndPoint::receiveMessage(WEMessage* message)
	 *	@see WEIPCEndPoint::messageArrivalListenerThreadFunc()
	 *
	 *	This method will try to read the complete message in two parts, first the header
	 *	bytes are read and then, depending upon the dataLength in the message header, the
	 *	data bytes are read.
	 *	If the data length received in the header is greater than the supplied value,
	 *	the original message->data is deallocated and new memory is allocated to accomodate
	 *	the incoming data.
	 *	This will return when the complete message is read or receive timeout expires.
	 */
	int receiveMessage(IPCMessage *p_message, long p_recvTimeoutMilliSec);
public:

	/**
	 *	@fn WEIPCEndPoint(bool autoReconnect)
	 *
	 *	@brief Standard constructor of WEIPCEndPoint class
	 *
	 *	The constructor initializes the data members with default values.
	 *
	 *	Call initialize(const std::string& endPointAddress) and connectTo() before calling sendMessage() and
	 *	receiveMessage() on the endpoint. OR
	 *	Call initialize(int endPointID, const std::string& endPointAddress, const std::string& peerEndPointAddress)
	 *	before calling sendMessage() and receiveMessage() on the endpoint.
	 */
	explicit WEIPCEndPoint(bool p_autoReconnect);

	/**
	 *	@fn ~WEIPCEndPoint()
	 *
	 *	@brief Destructor of WEIPCEndPoint class
	 *
	 *	The destructor calls deinitialize() which closes the endpoint.
	 */
	virtual ~WEIPCEndPoint();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEIPCEndPoint(const WEIPCEndPoint &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEIPCEndPoint & operator=(const WEIPCEndPoint &) = delete;

	/**
	 *	@fn initialize(const std::string& endPointAddress, bool isPublic = false)
	 *
	 *	@brief Initializes the end point by allocating system resources.
	 *
	 *	@param[in] endPointAddress String that shall be used to identify the
	 *		end point across processes in an IPC
	 *		communication. It can be a file path or
	 *		an arbitrary string that is unique across all
	 *		processes or endpoints in the system.
	 *	@param[in] isPublic Flag that tells if the endpoint is a server (public)
	 *		endpoint or a client (private) endpoint.
	 *
	 *	@returns WE_SUCCESS on successfully intializing the end point.
	 *	@returns WE_INIT_ERROR when initialization fails.
	 *	@returns WE_INVALID_PARAMETER_ERROR if endPointAddress.length() is <=0.
	 *
	 *	On Unix creates a socket and binds it with the specified address.
	 */
	int initialize(const std::string& p_endPointAddress, bool p_isPublic = false);

	/**
	 *	@fn deinitialize()
	 *
	 *	@brief Deinitializes the end point by deallocating system resources.
	 *
	 *	On Unix closes the socket. Frees all the allocated memory.
	 */
	void deinitialize();

	/**
	 *	@fn getEndPointAddress()
	 *
	 *	@brief Returns the address associated with this end point.
	 *
	 *	@returns End point address.
	 *
	 */
	const std::string& getEndPointAddress() const;

	/**
	 *	@fn getPeerEndPointAddress()
	 *
	 *	@brief Returns the address associated with the peer end point.
	 *
	 *	@returns Peer End point address.
	 *
	 */
	const std::string& getPeerEndPointAddress() const;

	/**
	 *	@fn getReadEndPointID()
	 *
	 *	@brief Returns the Identifier associated with the end piont.
	 *
	 *	@returns End point Identifier.
	 *
	 *	On Unix this Endpoint ID is a socket descriptor which can be used in other Unix calls
	 *	that accept a socket descriptor.
	 *	For socket end points read and write end point ids are same.
	 */
	int getReadEndPointID() const override;

	/**
	 *	@fn getWriteEndPointID()
	 *
	 *	@brief Returns the Identifier associated with the end piont.
	 *
	 *	@returns End point Identifier.
	 *
	 *	For socket end points read and write end point ids are same.
	 */
	int getWriteEndPointID() const override;

	/**
	 *	@fn connectTo(const std::string& serverAddress)
	 *
	 *	@brief connect to the specified IPC server.
	 *
	 *	Makes the OS specific calls to connect to the specified server.
	 *
	 *	@returns WE_SUCCESS If connect with server is established sucessfully.
	 *	@returns WE_TIMEOUT_ERROR If connection could not be established before connectTimeout expires.
	 *	@returns WE_INVALID_PARAMETER If serverAddress is nullptr or
	 *					serverAddressLength <= 0 or
	 *					endpoint is not initialized.
	 *	@returns WE_CONNECT_ERROR If connection could not be established for any other reason.
	 *
	 *
	 *	@param[in] serverAddress Address of the IPC server to connect to.
	 *
	 *	It returns if connection is successful, or if connection is not made
	 *	before the connect timeout, or connection fails for any other reason.
	 */
	int connectTo(const std::string& p_serverAddress);

	/**
	 *	@fn reconnect()
	 *
	 *	@brief reconnect the endpoint to the same peer with which it was connected earlier
	 *
	 *	@returns WE_SUCCESS If connect with server is re-established successfully.
	 *	@returns WE_TIMEOUT_ERROR If connection could not be established before connectTimeout expires.
	 *	@returns WE_CONNECT_ERROR If connection could not be established for any other reason.
	 *
	 *
	 *	It assumes that initialize() and connectTo() has already been called on the endpoint.
	 *	It is used to to reconnect endpoint which have been in use but got disconnected during send/receive of messages.
	 *	It initialize() the end point ONLY if the endPointID is not valid (-1) and calls connectTo() for the same peer.
	 */
	int reconnect() override;

	/**
	 *	@fn isConnected()
	 *
	 *	@brief Check whether the socket is connected to the peer IPC entity
	 *			according to the last access result not the current status.
	 *
	 *	@returns true If last access was successful.
	 *	@returns false If last access produced error.
	 */
	bool isConnected() const override;


	/**
	 *	@fn bool isClientSide()
	 *
	 *	@brief Check whether this endpoint represents the client side of a connection.
	 *
	 *	@returns true If this is the client side.
	 *	@returns false If this is the server side.
	 *
	 *
	 *	Only clients will try to reconnect to the server - never the other way around
	 */
	bool isClientSide() const
	{
		return this->isClient;
	}



	/**
	 *	@fn sendMessage(WEMessage* messageToSend)
	 *
	 *	@brief Send the specified message to the destination end point across
	 *	the link.
	 *
	 *	@param[in] message message to be sent to the destination.
	 *
	 *	@returns WE_SUCCESS on successfully sending the message.
	 *	@returns WE_TIMEOUT_ERROR when sending times out before completely sending the message.
	 *	@returns WE_SEND_ERROR when endpoint is not connected or sending fails due to any other reason.
	 *	@returns WE_INVALID_PARAMETER_ERROR if message is nullptr or
	 *						message->dataLength is 0 or
	 *						message->data is nullptr.
	 *
	 *	This method will try to send the complete message of the requested length.
	 *	This will return when the complete message is sent or send timeout expires.
	 */
	int sendMessage(IPCMessage* p_message);

	/**
	 *	@fn receiveMessage(WEMessage* messageReceived)
	 *
	 *	@brief Receive a message coming from the destination end point across
	 *	the link.
	 *
	 *	@param[out] message pointer to store message received from the destination.
	 *					"dataLength" in the messageReceived should be the expected
	 *					data length and the "data" should be pre-allocated.
	 *
	 *	@returns WE_SUCCESS on successfully sending the message.
	 *	@returns WE_TIMEOUT_ERROR when recieve times out before completely receving the message.
	 *	@returns WE_RECEIVE_ERROR when endpoint is not connected or receiving fails due to any other reason.
	 *	@returns WE_INVALID_PARAMETER_ERROR if messageReceived is nullptr or
	 *						messageReceived->dataLength is 0 or
	 *						messageReceived->data is nullptr.
	 *	@returns Also returns message with actual data length and the data received.
	 *
	 *	This method will try to read the complete message in two parts, first the header
	 *	bytes are read and then, depending upon the dataLength in the message header, the
	 *	data bytes are read.
	 *	If the data length received in the header is greater than the supplied value,
	 *	the original message->data is deallocated and new memory is allocated to accomodate
	 *	the incoming data.
	 *	This will return when the complete message is read or receive timeout expires.
	 */
	int receiveMessage(IPCMessage* p_message);

	/**
	 *	@fn setConnectTimeout(long milliSeconds)
	 *
	 *	@brief set the timeout for establishing connections.
	 *
	 *	@param[in] milliSeconds time in milli seconds to wait for.
	 *
	 *	Use 0 to return immdiately from connectTo().
	 *	Use -1 to wait forever untill the
	 *	establishment of the connection or occurrence of an error.
	 *	Default is -1.
	 */
	void setConnectTimeout(long p_milliSeconds);

	/**
	 *	@fn setReceiveTimeout(long milliSeconds)
	 *
	 *	@brief set the timeout to be used while receiving messages.
	 *
	 *	@param[in] milliSeconds time in milli seconds to wait for.
	 *
	 *	Use 0 to return immdiately from receiveMessage().
	 *	Use -1 to wait forever untill the
	 *	complete message could be received or an error occurred.
	 *	Default is -1.
	 */
	void setReceiveTimeout(long p_milliSeconds);

	/**
	 *	@fn setSendTimeout(long milliSeconds)
	 *
	 *	@brief set the timeout to be used while sending messages.
	 *
	 *	@param[in] milliSeconds time in milli seconds to wait for.
	 *
	 *	Use 0 to return immdiately from sendMessage().
	 *	Use -1 to wait forever untill the
	 *	complete message could be sent or an error occurred.
	 *	Default is -1.
	 */
	void setSendTimeout(long p_milliSeconds);


	/**
	 *	@fn setMessageArrivalListener(WEIPCMessageArrivalListener* messageListener)
	 *
	 *	@brief set the message arrival listener that is to be invoked everytime
	 *	a message arrives on the link.
	 *
	 *	@param[in] newMessageArrivalListener listener to be invoked on arrival of a message.
	 *								Use nullptr if listener is to be stopped.
	 *	@returns WE_SUCCESS on successfully intializing the thread.
	 *	@returns WE_INTERFACE_ERROR if thread creation fails.
	 *
	 *	Internally initializes a thread to listen to the link for arriving
	 *	messages. If nullptr is set thread is killed if present. If it is called twice
	 *	with non null messageListeners the first listener is replaced by the latest one
	 *	keeping the thread same.
	 *	Default is no messageListener.
	 *
	 *	On Unix if separate thread for listening on each endpoint is not desired then the
	 *	EndpointID can be used in select() where other descriptors can also be passed
	 *	and used within a single thread.
	 */
	int setMessageArrivalListener(WEIPCMessageArrivalListener* p_newMessageArrivalListener);

	/**
	 *	@fn getMessageArrivalListener()
	 *
	 *	@brief get the message arrival listener.
	 *
	 *	@returns Currently set message arrival listener or nullptr if not set.
	 */
	WEIPCMessageArrivalListener* getMessageArrivalListener();

	/**
	 *	@fn setMessageArrivalListenerActive(bool activate)
	 *
	 *	@brief Temporarily activate/deactivate the message arrival listener thread.
	 *
	 *	@param[in] activate True to activate the thread and false to deactivate.
	 *
	 *	If deactivation is required then ig gets the lock on the mutex so that
	 *	messageArrivalListenerThreadFunc goes waiting for the lock to be released.
	 *	This is required when it is not required to receive messages through the message
	 *	arrival listener and receiveMessage() is used directly.
	 *	If activating the messageArrivalListener thread is requested and thread is
	 *	already active then it does nothing. Same is true if thread is inactive and
	 *	deactivation is requested.
	 *
	 */
	void setMessageArrivalListenerActive(bool p_activate);

	static int debugSection;
};


/**
 *	@class WEIPCMessageArrivalListener
 *
 *	@brief This class is the listener interfaces for the messages received
 *	on a IPC Link.
 *	It needs to be implemented and set the WEIPCEndPoint instance inorder to
 *	to be notified of the messages received on the link.
 *
 *	@namespace WE
 *
 *	@see WEIPCEndPoint
 */
class WEIPCMessageArrivalListener
{
public:
	/**
	 *	@fn virtual void messageArrived(WEMessage* message) = 0
	 *
	 *	@brief Called everytime a message is received on the IPC link.
	 *
	 *	@param[in] message message that has arrived.
	 *
	 *	Do the processing on the message here in this function.
	 *
	 *	@warning message pointer passed here shall not be used after this
	 *		call as it will be deleted internally after this call.
	 */
	virtual void messageArrived(IPCMessage* p_message) = 0;

	/**
	 *	@fn ~WEIPCMessageArrivalListener()
	 *
	 *	@brief Dummy virtual destructor for the class.
	 *
	 *	Does nothing.
	 */
	virtual ~WEIPCMessageArrivalListener() = default;
};
}

#endif	/* WE_IPC_H_ */
