/**
 *	@file we_ipc_server.h
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd. Faisal
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief It defines the WEIPCServer class to create a server in an IPC communication.
 *
 *	@see we_ipc_server.cpp
 *
 */


#ifndef WE_IPC_SERVER_H_
#define WE_IPC_SERVER_H_

#include <string>
#include <memory>
#include "we_ipc.h"

namespace WE
{

class WEIPCEndPoint;
class WEEndPointManager;
class WEEndPointHandler;

/**
 *	@class WEIPCServer
 *
 *	@brief This class represents a server side in an IPC communication.
 *
 *	@namespace WE
 *
 *	@see WEIPCEndPoint
 *
 *	The IPC mechanism might change based on the OS. For unix
 *	domain sockets are used, for windows it could be Message Queues.
 */
class WEIPCServer
{
private:
	std::shared_ptr<WEIPCEndPoint> serverEndpoint{std::make_shared<WEIPCEndPoint>(false)};
public:
	static const int maxQueueSize = 128;	//!< Maximum connection queue size on Unix and Maximum queue size on Windows for Message Queues.
	/**
	 *	@fn WEIPCServer()
	 *
	 *	@brief Standard constructor of WEIPCServer class
	 *
	 *	The constructor initializes the data members with default values.
	 *	Call initialize(serverAddress) after this to make a usable server
	 *	end point.
	 */
	WEIPCServer() = default;

	/**
	 *	@fn ~WEIPCServer()
	 *
	 *	@brief Destructor of WEIPCServer class
	 *
	 */
	virtual ~WEIPCServer() = default;

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEIPCServer(const WEIPCServer &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEIPCServer & operator=(const WEIPCServer &) = delete;

	/**
	 *	@fn initialize(const std::string& serverAddress)
	 *
	 *	@brief Initializes the server with the specified server address.
	 *
	 *	@param[in] serverAddress String that shall be used to identify the
	 *				client across processes in an IPC
	 *				communication. It can be a file path or
	 *				an arbitrary string that is unique across all
	 *				processes or endpoints in the system.
	 *
	 *	@returns WE_SUCCESS on successfully intializing the end point.
	 *	@returns WE_INIT_ERROR when initialization fails.
	 *	@returns WE_INVALID_PARAMETER_ERROR if serverAddress.length() is <=0.
	 *
	 *	On Unix creates a socket, binds it with the specified address
	 *	and makes it a listening socket.
	 *	Call this function before calling acceptConnection().
	 */
	virtual int initialize(const std::string& p_serverAddress);

	/**
	 *	@fn acceptConnection()
	 *
	 *	@brief accept an incoming connection from an IPC client.
	 *
	 *	@returns The new endpoint established between the client and server.
	 *	@returns nullptr if connections cannot be accepted or acceptTimeout expires.
	 *
	 *	It returns if a connection is established successfully, or
	 *	if connection is not received before the accept timeout, or
	 *	if it fails for any other reason.
	 */
	std::shared_ptr<WEIPCEndPoint> acceptConnection();

	/**
	 *	@fn setAcceptTimeout(long milliSeconds)
	 *
	 *	@brief set the timeout for accepting connections.
	 *
	 *	@param[in] milliSeconds time in milli seconds to wait for.
	 *
	 *	Use 0 to return immdiately from acceptConnection().
	 *	Use -1 to wait forever untill a
	 *	connection is received or occurrence of an error.
	 *	Default is -1.
	 */
	void setAcceptTimeout(long p_milliSeconds);

	/**
	 *	@fn getServerEndPoint()
	 *
	 *	@brief Returns the server end point.
	 *
	 *	@returns The server end point.
	 *	@returns nullptr if the server is not initialized.
	 *
	 */
	const std::shared_ptr<WEIPCEndPoint> getServerEndPoint() const;

	void addToEndpointManager(WEEndPointManager * p_manager, WEEndPointHandler * p_handler);

	static int debugSection;
};
}
#endif /* WE_IPC_SERVER_H_ */
