/**
 *	@file we_power_management.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of WEAPI interface functions for accessing Power Management Daemon
 *
 *	The interface defined in this file provides functions for accessing
 *	power management API on an OPUS platform.
 */

#ifndef WE_POWER_MANAGEMENT_END_POINT_H_
#define WE_POWER_MANAGEMENT_END_POINT_H_

#include <sys/un.h>

#include "../../endpoint_manager/include/we_endpoint.h"

namespace WE
{

/**
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief Interface for power management
 *
 *	@namespace WE
 *
 *	This class provides an platform independent interface for accessing
 *	the power management of an OPUS platform.
 *	Use the init method to connect the class to the power management of the
 *	operating system.
 */
class WEPowerManagementEndPoint : public WEEndPoint
{
public:
	WEPowerManagementEndPoint();
	virtual ~WEPowerManagementEndPoint() = default;

	/**
	 *	@fn int init(void)
	 *
	 *	@brief Initialize the power management end point and connects to the socket.
	 *
	 *	@returns WE_SOCKET_ERROR when the power management socket can not be created.
	 *	@returns WE_CONNECT_ERROR when the end point can not connect to the power management socket.
	 *	@returns WE_SUCCESS when the socket connection was established successful.
	 */
	int init(void);

	/**
	 *	@fn int release(void)
	 *
	 *	@brief Release the power management end point resources.
	 *
	 *	@returns WE_FILE_ERROR when the power management socket can not be closed.
	 *	@returns WE_SUCCESS when the socket connection was closed successful.
	 */
	int release(void);

	/**
	 *	@fn int getReadEndPointID(void)
	 *
	 *	@brief Returns the Identifier used to read from the end point.
	 *
	 *	@returns End point Identifier.
	 *
	 *	This could be same as the identifier returned from the getWriteEndPointID()
	 */
	int getReadEndPointID(void) const override;

	/**
	 *	@fn int getWriteEndPointID(void)
	 *
	 *	@brief Returns the Identifier used to write to the end point.
	 *
	 *	@returns End point Identifier.
	 *
	 *	This could be same as the identifier returned from the getReadEndPointID()
	 */
	int getWriteEndPointID(void) const override;

	/**
	 *	@fn bool isConnected(void)
	 *
	 *	@brief Check whether the end point is still connected to the peer entity
	 *			according to the last access result not the current status.
	 *
	 *	@returns true If last access was successful.
	 *	@returns false If last access produced error.
	 */
	bool isConnected(void) const override;

	/**
	 *	@fn int reconnect(void)
	 *
	 *	@brief reconnect the end point to the same peer with which it was connected earlier
	 *
	 *	@returns WE_FILE_ERROR when an open power management socket can not be closed.
	 *	@returns WE_SOCKET_ERROR when the power management socket can not be created.
	 *	@returns WE_CONNECT_ERROR when the end point can not connect to the power management socket.
	 *	@returns WE_SUCCESS when the socket connection was established successful.
	 *
	 *	It depends on the type of the end point if it is really useful because there may be cases
	 *	where end point seems always connected e.g; WEFileEndpoint.
	 */
	int reconnect(void) override;

	/**
	 *	Use this function to send a message to the power management daemon.
	 *	The message pointer is platform specific. For devices it needs to be a pointer to a wepmd_msg_t struct!
	 *	The function checks if the connection has errors and tries to reconnect 10 times with a timeout of 50ms each try
	 *	=> This function may block for 500ms if the reconnect fails 10 times!
	 */
	int sendMessage(void * p_message);

private:
	int powerSock{-1};				//!< file descriptor for API socket
	struct sockaddr_un addr{};		//!< address of power management daemon API socket
};
} /* namespace WE */

#endif /* WE_POWER_MANAGEMENT_END_POINT_H_ */
