/**
 *	@file we_power_management.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of WEAPI interface functions for accessing Power Management on OPUS A3 platform.
 *
 */

#ifndef WE_POWER_MANAGEMENT_INTERFACE_A3_H_
#define WE_POWER_MANAGEMENT_INTERFACE_A3_H_

#include <memory>
#include <wepmd_api.h>

#include "../include/we_power_management.h"
#include "../include/we_power_management_end_point.h"
#include "../../../event_handling/include/we_signal.h"

namespace WE
{

/**
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief Interface for power management
 *
 *	@namespace WE
 *
 *	This class provides the interface for accessing the power management
 *	of a OPUS A3 platform.
 */
class WEPowerManagementInterfaceA3 : public WEPowerManagement
{	// Working without 'check return values of all class member functions' since 2011
public:

	/**
	 *	@brief Standard constructor
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	WEPowerManagementInterfaceA3() = default;

	/**
	 *	@brief Standard destructor
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	virtual ~WEPowerManagementInterfaceA3();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEPowerManagementInterfaceA3(const WEPowerManagementInterfaceA3 &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEPowerManagementInterfaceA3 & operator=(const WEPowerManagementInterfaceA3 &) = delete;

	/**
	 *	@brief Open connection to Power Management Daemon and reads out the available hardware devices
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS if establishing connection and reading was successful
	 *	@returns WE_SOCKET_ERROR if socket can not be created
	 *	@returns WE_CONNECT_ERROR if opening connection failed
	 */
	int init(void) override;

	/**
	 *	@brief Close connection to Power Management Daemon
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS when socket was closed successful
	 *	@returns WE_FILE_ERROR when closing socket failed
	 */
	int release(void) override;

	/**
	 *	@brief Set the event listener for power management events
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] eventListener eventListener that has to be invoked on occurrence of an event.
	 *	nullptr value has to be used to unregister the eventListener.
	 *
	 *	@returns WE_SUCCESS when the eventListener was registered successful
	 *
	 *	Data read from power management daemon are evented to the eventListener.
	 */
	int setEventListener(WESignalListener * p_eventListener) override;

	/**
	 *	@brief Set timeout limit
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_state Holds the corresponding power mode. Allowed values are
	 *	POWER_STATE_LOW_POWER and POWER_STATE_SLEEP
	 *	@param[in] p_timeout Timeout limit in seconds to set
	 *
	 *	@returns WE_SUCCESS if setting timeout was successful
	 *	@returns WE_WRITE_ERROR if setting timeout value fails
	 *	@returns WE_INVALID_PARAMETER_ERROR if timeout is not supported by given p_state
	 */
	int setTimeoutForState(wePowerStates p_state, unsigned int p_timeout) override;

	int setPropagationTimeForState(wePowerStates p_state, unsigned int p_time) override;

	/**
	 *	@brief Get timeout limit
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_state Holds the corresponding power mode. Allowed values are
	 *	POWER_STATE_LOW and POWER_STATE_SLEEP
	 *
	 *	@returns WE_SUCCESS if reading timeout values was successful
	 *	@returns WE_READ_ERROR if reading timeout value failed
	 */
	int getTimeoutForState(wePowerStates p_state) override;

	int getPropagationTimeForState(wePowerStates p_state) override;


	/**
	 *	@brief Set device state for low-power mode
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_device Holds the corresponding device id
	 *	@param[in] p_state Holds the low-power mode state (on/off)
	 *
	 *	@returns WE_SUCCES if the set request was successfully forwarded to the hardware (e.g. power daemon)
	 *	@returns Something else on error
	 */
	int setDeviceStateForLowPower(int p_device, int p_state) override;

	/**
	 *	@brief Get device state for low-power mode
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_device Holds the corresponding device id
	 *
	 *	@returns WE_SUCCESS if reading low-power state was successful
	 *	@returns WE_READ_ERROR if reading low-power state failed
	 *
	 *	@warning The answer to the get request will come via an event!
	 */
	int getDeviceStateInLowPower(int p_device) override;

	/**
	 *	@brief Set power mode
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_state The power state to set
	 *
	 *	@returns WE_SUCCESS if setting power state was successful
	 *	@returns WE_WRITE_ERROR if setting power state failed
	 *	@returns WE_INVALID_PARAMETER_ERROR if it is not possible to switch into given power state
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if power state is unknown
	 */
	int setPowerState(wePowerStates p_state) override;

	/**
	 *	@brief Get current power state
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS if reading timeout values was successful
	 *	@returns WE_READ_ERROR if reading timeout value failed
	 */
	int getPowerState(void) override;

	/**
	 *	@fn int subscribeToEndpointManager(WEEndPointManager * p_mgr)
	 *
	 *	@brief Add power management as an endpoint to the end point manager
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_mgr Pointer to the endpoint manager to which to subscribe
	 *
	 *	@returns WE_SUCCESS if adding the power management as an endpoint to the endpoint manager was successful
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_mgr is nullptr
	 *	@returns WE_ALREADY_EXISTS_ERROR if the power management was already registered.
	 *
	 *	@see int WEEndPointManager::addEndPoint(WEEndPoint* endPointToAdd, WEEndPointHandler* handler)
	 *
	 *	This function is used to add this power management  as an endpoint to the endpoint manager.
	 *	As soon as the power management is readable (a message is received) the endpoint manager calls
	 *	the "onEndPointReadable" function of the power management.
	 */
	int subscribeToEndpointManager(WEEndPointManager * p_mgr) override;

	/**
	 *	@fn onEndPointReadable(WEEndPoint* p_endPoint, WEEndPointManager* p_endPointMgr)
	 *
	 *	@brief Invoked everytime from the waitForEndPoints() if the corresponding endpoint
	 *		becomes readable.
	 *
	 **	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endPoint The End point that is found to be readable.
	 *	@param p_endPointMgr The End Point manager managing the endpoint.
	 *
	 *	Generally it is invoked when a connection request is received on a server endpoint
	 *	or when a message is received on a client endpoint. For a server endpoint
	 *	WEIPCServer::acceptConnection() inside this function should return successfully.
	 *	And for a client endpoint WEIPCEndPoint::receiveMessage() inside this function
	 *	should return successfully.
	 *	It might also be invoked when connection is broken but only once. This is becuase
	 *	the connection status is not the immediate status but based on the last access results.
	 *	In such a case WEIPCEndPoint::receiveMessage() will return WE_RECEIVE_ERROR, therefore
	 *	for a failing WEIPCEndPoint::receiveMessage() it should be checked if the endpoint is
	 *	still connected using endPoint->isConnected(). And if it is not connected the endPoint
	 *	could be removed from the endPointManager using WEEndPointManager::removeEndPoint().
	 */
	void onEndPointReadable(std::shared_ptr<WEEndPoint> p_endPoint, WEEndPointManager* p_endPointMgr) override;

	/**
	 *	@fn void onEndPointReconnected(WEEndPoint* p_endpoint)
	 *
	 *	@brief This function gets called every time a endpoint gets re-connected
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endPoint The endpoint which was reconnected
	 */
	void onEndPointReconnected(std::shared_ptr<WEEndPoint> p_endPoint) override;

	/**
	 *	@fn void onEndPointDisconnected(WEEndPoint* p_endpoint)
	 *
	 *	@brief Call this function if a IPC connection was lost
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endPoint Endpoint of the connection that was disconnected
	 *
	 */
	void onEndPointDisconnected(std::shared_ptr<WEEndPoint> p_endPoint) override;


private:
	std::shared_ptr<WEPowerManagementEndPoint> endPoint{std::make_shared<WEPowerManagementEndPoint>()};	//!< the power management end point
	WESignalListener * signalListener{nullptr};	//!< pointer to the registered event listener

	/**
	 *	@brief Function for sending events to the registered event listener
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS if no event listener is registered.
	 *	@returns forwards the result of signalOccurred function call.
	 */
	int fireSignalToSubscribers(WESignal & p_signal);

	/**
	 *	@brief Function for sending a message to the power management daemon
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS if sending message was successful.
	 *	@returns WE_WRITE_ERROR if writing message failed or was incomplete.
	 */
	int sendMessage(wepmd_msg_t &p_msg);

	/**
	 *	@brief Function for converting a power management daemon message into an WEapi event
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[out] p_signal The resulting signal.
	 *	@param[in] msg The message to convert.
	 *	@param[in] time The time stamp to assigne to the event (in milliseconds).
	 *
	 *	@returns WE_SUCCESS if converting message into event was successful.
	 *	@returns WE_NOT_SUPPORTED_ERROR if eventing a message type is not supported.
	 *	@returns WE_INVALID_MESSAGE_RECEIVED_ERROR if message contains invalid data
	 */
	static int messageToSignal(WESignal & p_signal, const wepmd_msg_t & p_msg, unsigned int p_time);

	static int timeoutMessageToSignal(WESignal & p_signal, const wepmd_msg_t & p_msg);
	static int deviceMessageToSignal(WESignal & p_signal, const wepmd_msg_t & p_msg);
	static int modeMessageToSignal(WESignal & p_signal, const wepmd_msg_t & p_msg);
	static int modeChangeMessageToSignal(WESignal & p_signal, const wepmd_msg_t & p_msg);

	virtual int disablePowerbuttonOnToLowPower(void) override;
};
} /* namespace WE */

#endif /* WE_POWER_MANAGEMENT_INTERFACE_A3_H_ */
