/**
 *	@file we_socket.h
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2012
 *
 *	@brief abstract class for the socket implementation
 */

#ifndef WE_SOCKET_H_
#define WE_SOCKET_H_

#include <deque>

namespace WE
{
#define WE_SOCKET_MAX_BUFFER_SIZE 		1500		/* in bytes => max. ethernet frame size so we should be on the safe side here when using TCP*/
#define WE_SOCKET_PORT 					15000	/* default port */
#define WE_SOCKET_IP 					"127.0.0.1" /* default IP => localhost */
#define WE_SOCKET_SERIAL_HEADER_LENGTH 	17
#define WE_SOCKET_SERIAL_HEADER_TEXT_ARCHIVE_LENGTH  16
const char WE_g_serialHeader[WE_SOCKET_SERIAL_HEADER_LENGTH] = {22, 0, 0, 0, 's', 'e', 'r', 'i', 'a', 'l', 'i', 'z', 'a', 't', 'i', 'o', 'n'};
const char WE_g_serialHeaderTextArchive[WE_SOCKET_SERIAL_HEADER_TEXT_ARCHIVE_LENGTH] = {'2', '2', ' ', 's', 'e', 'r', 'i', 'a', 'l', 'i', 'z', 'a', 't', 'i', 'o', 'n'};


//<MKA 21.02.2012> this is the abstract class. The documentation can be found @ the implementation class.
class WESocketMessage;

class WESocket
{
public:
	WESocket() = default;
	virtual ~WESocket();
	WESocket(const WESocket & p_copy) = delete;
	WESocket& operator=(const WESocket& p_copy) = delete;

	//<MKA 15.02.2012> changed the interface. It is better to use unique method names so the compiler and developer knows to 100%
	//from where this method is from.
	virtual void initSocket() = 0;
	virtual void resetSocket() = 0;

	virtual int sendSocketData(WESocketMessage *p_messageToSend) = 0;


	virtual int sendSocketData(unsigned int p_socket, int* p_sendBuffer, unsigned int p_uiLength) = 0;
	virtual int sendSocketData(int* p_sendBuffer, unsigned int p_uiLength) = 0;


	virtual int receiveSocketData(int* p_receiveBuffer, unsigned int p_uiLength) = 0;
	virtual int receiveSocketData(unsigned int p_socket, int* p_receiveBuffer, unsigned int p_uiLength) = 0;
	virtual int closeSocket(unsigned int p_socket) = 0;
	virtual int closeSocket() = 0;


	virtual int connectSocket(char* p_address, unsigned int p_port) = 0;
	virtual int connectSocket() = 0;
	virtual int bindSocket(char* p_address, unsigned int p_port) = 0;
	virtual int bindSocket() = 0;
	virtual int listenSocket() = 0;
	virtual int acceptSocket() = 0;
	virtual int exitReceiveSocketData(void) = 0;
	virtual int resetSocketPort() = 0;
	virtual void setSocketDomain(int p_domain) = 0;
	virtual int getSocketDomain() = 0;
	virtual void setSocketType(int p_type) = 0;
	virtual int getSocketType() = 0;
	virtual void setSocketProtocol(int p_protocol) = 0;
	virtual int getSocketProtocol() = 0;
	virtual void setSocketMaxQueuedConnections(int p_conn) = 0;
	virtual int getSocketMaxQueuedConnections() = 0;
	virtual void setSocketOptionsLevel(int p_level) = 0;
	virtual int getSocketOptionsLevel() = 0;
	virtual void setSocketOptionsName(int p_name) = 0;
	virtual int getSocketOptionsName() = 0;

	/**
	 * Read WE_SOCKET_MAX_BUFFER_SIZE bytes from member socket and store it into p_buffer
	 */
	virtual int receiveSocketData(char * &p_buffer) = 0;

	/**
	 * Read WE_SOCKET_MAX_BUFFER_SIZE bytes from p_socket and store it into p_buffer
	 */
	virtual int receiveSocketData(unsigned int p_socket, char * &p_receivedMessageBuffer);

	/**
	 * Send messageToSend. WESocketMessage will be serialized into a character buffer and sent over p_socket.
	 */
	virtual int sendSocketData(unsigned int p_socket, WESocketMessage *p_messageToSend);

	virtual int getSocketIdentifier(void) = 0;

	/**
	 * Returns a character buffer that contains exactly one WESocketMessage or nullptr if no messages were received.
	 * p_errorIndicator indicates errors or WE_SUCCESS (see Note below). If p_socket is not provided the member socket of this instance is used.
	 * Use the WESocketMessage constructor that takes a char * as parameter to create a message from the returned buffer.
	 * You can use WESocketMessage::getMessageType(char * p_buffer) to determine the type of message inside the buffer.
	 * Note that this function may return WE::WE_TIMEOUT_ERROR in p_errorIndicator if the non-blocking socket
	 * would normally have blocked. This is not an error but an indicator that no buffer was returned and the function can be called again.
	 */
	char * getNextSocketBuffer(int & p_errorIndicator, int p_socket = -1);

protected:
	std::deque<char *> oldBuffers{};	//!< Storing WESocketMessage buffers here when they are received via getNextSocketBuffer()
	char * restOfOldMessage{nullptr};	//!< If a buffer was received via getNextSocketBuffer() but had only a part of a message in it, we store that part here and combine it with the rest when receiving the next buffer
	int restOfOldMessageLength{0};	//!< the length in byte of restOfOldMessage

	char receiveBuffer[WE_SOCKET_MAX_BUFFER_SIZE]{};	//!< temporary buffer used internally

	static int myDebugKey;
};
}
#endif /* WE_SOCKET_H_ */
