/**
 *	@file we_socket_messages.h
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2012
 *
 *	@brief defines, structs, classes and enums for the messages
 */


/*
 * NOTE:
 *
 * This kind of class works only if we stay inside the same OS and architecture.
 * Need to check if we need these pragma for the compiler
 *
 * #pragma pack(1)   // this helps to pack the struct to 5-bytes
 * #pragma pack(0)   // turn packing of
 *
 * Why this class:
 * This class shall be used to wrap your own data. You can use this class with the serialization methods provided inside
 * the socket class implementation. You may use it by sending the class as raw data through the socket.
 *
 */


#ifndef WE_SOCKET_MESSAGES_H_
#define WE_SOCKET_MESSAGES_H_

#include <cstring>
#include <string>

namespace WE
{

// This class could benefit from some DOCUMENTATION, function headers, DOCUMENTATION, comments, DOCUMENTATION ... and documentation
// Ask the author if you should not consider method names self-explanatory.

enum WESocketMessageType
{
	SOCKET_MESSAGE_INIT 			= 0,
	SOCKET_MESSAGE_HANDSHAKE 		= 100,
	SOCKET_MESSAGE_HANDSHAKE_ACK 		= 150,
	SOCKET_MESSAGE_DATA 			= 200,
	SOCKET_MESSAGE_DATA_ACK 		= 250,
	SOCKET_MESSAGE_BYEBYE 			= 300,
	SOCKET_MESSAGE_BYEBYE_ACK 		= 350,
	SOCKET_MESSAGE_SHUTDOWN 		= 400,
	SOCKET_MESSAGE_SHUTDOWN_ACK 		= 450,
	SOCKET_MESSAGE_START_SEQ		= 500,
	SOCKET_MESSAGE_START_SEQ_ACK		= 550,
	SOCKET_MESSAGE_END_SEQ			= 600,
	SOCKET_MESSAGE_END_SEQ_ACK		= 650,
	SOCKET_MESSAGE_SEQUENCE 		= 700,
	SOCKET_MESSAGE_SEQUENCE_ACK 		= 750,
	SOCKET_MESSAGE_PING 			= 1100,
	SOCKET_MESSAGE_PONG 			= 1150
};


class WESocketMessage
{
public:
	WESocketMessage();
	explicit WESocketMessage(char * p_buffer);
	virtual ~WESocketMessage() = default;

	//Copy constructor
	WESocketMessage(const WESocketMessage & p_message);
	/**
	 *	@fn WESocketMessage & operator=(const WESocketMessage &p_message)
	 *
	 *	@brief equal operator
	 */
	virtual WESocketMessage & operator=(const WESocketMessage &p_message);

	/**
	 * Serialize this message into provided p_buffer.
	 * p_buffer needs to be allocated correctly with the correct size! (use calculateBufferSize())
	 * OVERWRITE FOR EVERY SUBCLASS!!!
	 */
	virtual int fillBuffer(char * p_buffer);

	/**
	 * Returns the number of bytes this class needs in a character buffer.
	 * OVERWRITE FOR EVERY SUBCLASS!!!
	 */
	virtual unsigned int calculateBufferSize(void);

	/**
	 * Updates the message length of p_buffer with the provided length.
	 * Also updates this->messageLength. Call this at the end of fillBuffer() in you sub class.
	 */
	void updateMessageLength(char * p_buffer, unsigned int p_newMessageLength);

	/**
	 * Returns the type of WESocketMessage inside p_buffer
	 */
	static unsigned short getMessageType(char * p_buffer);

	unsigned short messageType{SOCKET_MESSAGE_DATA};			//Needed for the state machine to know what to do next

	unsigned int messageSize{0};			//This is the full message size including header and payload, call calculateBufferSize to initialize correctly
	bool withACK{false};						//Indicates if the sender wants to have a ACK or not

	unsigned short messageSource{0};
	unsigned short messageDestination{0};
	unsigned short broadcastCounter{0};

	static inline const std::string MESSAGE_HEADER{"WESOCKETMESSAGE"};
	static inline const int MESSAGE_HEADER_SIZE{static_cast<int>(MESSAGE_HEADER.length())};


protected:

	virtual void init();
};
}

#endif /*WE_SOCKET_MESSAGES_H_*/
