/**
 *	@file we_socket_linux.h
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2012
 *
 *	@brief linux implementation for the linux world
 */

#ifndef WE_SOCKET_LINUX_H_
#define WE_SOCKET_LINUX_H_

/*
 * MKA 20.01.2012
 *
 * class skeleton for socket support.
 * ... state machines and usage have to be defined by the application
 * which might be need the socket. Here you will only find a basic
 * structure.
 *
 * Maybe we can think about a basic server, but maybe this is not a good idea
 *
 * We should have an abstraction interface
 */
#include <arpa/inet.h>

#include "../include/we_socket.h"

namespace WE
{

class WESocketMessage;
class WESocketLinux : public WESocket
{
public:
	//we only override 3 of the 4 sendSocketData and receiveSocketData functions of WESocket
	using WESocket::sendSocketData;
	using WESocket::receiveSocketData;

	// This header could benefit from some DOCUMENTATION, function headers, DOCUMENTATION, comments, DOCUMENTATION ... and documentation
	// Ask the author if you should not consider names self-explanatory.


	WESocketLinux();
	virtual ~WESocketLinux();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WESocketLinux(const WESocketLinux &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WESocketLinux & operator=(const WESocketLinux &) = delete;

	//calls reset socket
	//<MKA 21.02.2012> Currently no special values are set by init
	void initSocket() override;
	//sets all default values for the class members.
	void resetSocket() override;


	//the connect socket methods are needed for the client site.
	//for the server you can use directly bind. The rest is done inside the class

	//if connect returns != 0 socket is not reachable
	//otherwise this method returns a socket handler
	int connectSocket(char* p_address, unsigned int p_port) override;
	//if connect returns != 0 socket is not reachable
	//otherwise this method returns a socket handler
	//Connect with default values ip 127.0.0.1 (localhost) port 15000
	//This can be used when using the implementation to do IPC
	int connectSocket() override;

	//the bind methods shall be used by the server.
	//returns: On success, zero is returned. On error, -1 is returned, and errno is set appropriately.
	//if bind returns != 0 socket is not reachable
	int bindSocket(char* p_address, unsigned int p_port) override;
	//if bind returns != 0 socket is not reachable
	//bind with default values
	int bindSocket() override;

	//this method is used by the server not needed by the client
	//return always 0
	int listenSocket() override;

	//this method is used by the server not needed by the client
	//please note: This is a blocking call and can only be ended by closing the socket!
	//returns the new socket handler or on error -1, and errno is set appropriately.
	int acceptSocket() override;

	//returns how many bytes are send (raw data!)
	//This method can be used if you only want to work with the WESocketMessage class. The class is serialized and deserialized by
	//the socket class. You do not need to take care.
	//Note: The data is serialized and send via the default socket. This is mostly used on the client site, where no new
	//socket is created for each new connection after acceptSocket()
	int sendSocketData(WESocketMessage *p_messageToSend) override;

	//the next two sendSocketData methods shall be used if you want to send raw(!) data through a socket.
	//here you have to take care on everything by your own.
	int sendSocketData(int* p_sendBuffer, unsigned int p_uiLength) override;
	//Use this on the server when needing to provide a new socket handler.
	int sendSocketData(unsigned int p_socket, int* p_sendBuffer, unsigned int p_uiLength) override;

	//returns the size of the received raw(!) Data
	//This method can be used if you only want to work with the WESocketMessage class. The class is serialized and deserialized by
	//the socket class. You do not need to take care.
	//Note: The data is deserialized by the class. The return value is the raw data send. Not the actual data of the WESocketMessage.
	//Note: This call is blocking and can only be broken by closing the socket when using SOCK_STREAM.
	//

	//We assume that *messageReceived is already allocated memory
	//In *messageReceived is the first received message stored.
	//All other message (if any) have to be retrieved by getNextInReceiveList();
	int receiveSocketData(char * &p_buffer) override;

	//the next two sendSocketData methods shall be used if you want to send raw(!) data through a socket.
	//here you have to take care on everything by your own.
	//Note: This call is blocking and can only be broken by closing the socket when using SOCK_STREAM.
	int receiveSocketData(int* p_receiveBuffer, unsigned int p_uiLength) override;
	//Use this on the server when needing to provide a new socket handler.
	int receiveSocketData(unsigned int p_socket, int* p_receiveBuffer, unsigned int p_uiLength) override;

	//These two methods are closing a socket.
	int closeSocket(unsigned int p_socket) override;
	//This closes the class socket.
	int closeSocket() override;




	/************************************************************************/
	//The next set methods shall only be used if you really know what you are doing.
	//Please refer before to the man page of extern int socket (int __domain, int __type, int __protocol) __THROW;
	//Any change to the defaults can have impact on the behaviour of several methods of this class.
	//If changing any of these values it can be that the described behaviour inside this header file do not match the actual
	//behaviour of your implementation.
	//Note: All comments in this header file refering to the defaults described in the class member section.
	void setSocketDomain(int p_domain) override
	{
		this->domain = p_domain;
	}

	int getSocketDomain() override
	{
		return this->domain;
	}

	void setSocketType(int p_type) override
	{
		this->type = p_type;
	}

	int getSocketType() override
	{
		return this->type;
	}

	void setSocketProtocol(int p_protocol) override
	{
		this->protocol = p_protocol;
	}

	int getSocketProtocol() override
	{
		return this->protocol;
	}

	//Note: use manpage listen()
	void setSocketMaxQueuedConnections(int p_conn) override
	{
		this->maxConnectionsQueued = p_conn;
	}

	int getSocketMaxQueuedConnections() override
	{
		return this->maxConnectionsQueued;
	}

	//Note: The next set methods have impact on the setsockopt method.
	//Note: Please refer to the setsockopt manpage for all options.
	//optval = 1 because SO_REUSEADDR is int. See setsockopt reference for details
	//SO_REUSEADDR is used for multiple requests (queued as long as no multithreading is used)
	//and if the app dies there is no 2 minute timeout until the socket can be reused.

	void setSocketOptionsLevel(int p_level) override
	{
		this->sockOptLevel = p_level;
	}

	int getSocketOptionsLevel() override
	{
		return this->sockOptLevel;
	}

	void setSocketOptionsName(int p_name) override
	{
		this->sockOptOptionName = p_name;
	}

	int getSocketOptionsName() override
	{
		return this->sockOptOptionName;
	}
	/************************************************************************/

	/* <MKA 21.02.2012> reserved for future use*/
	int exitReceiveSocketData(void) override;
	int resetSocketPort() override;

	int getSocketIdentifier(void) override
	{
		return (int)(this->sockConnectSocket);
	}


private:

	//creates the socket connection and set the socket options. Mainly used by connect() and bind()
	int createSocket();


	int maxConnectionsQueued; /*connection requests will be queued before further requests are refused, used by listen()*/
	int sockOptLevel; /* default = SOL_SOCKET*/
	int sockOptOptionName; /* default = SO_REUSEADDR */

	int domain; /*provides the default domain for the socket connection AF_INET*/
	int type; /* provides the socket type currently we are using SOCK_STREAM*/
	int protocol; /* this is currently 0 by default means the method choose one automatically, when using SOCK_STREAM it is TCP*/
	int sockConnectSocket;	/* Handle to socket */
	struct sockaddr_in addressInfo; /* Information about the Client */
};
}
#endif /* WE_SOCKET_LINUX_BASE_H_ */
