/**
 *	@file we_system.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Definition of a WESystem class
 *
 */

#ifndef WE_SYSTEM_H_
#define WE_SYSTEM_H_

#include <string>

namespace WE
{

enum CPUType
{
	CPU_TYPE_IMX6_SOLO,
	CPU_TYPE_IMX6_QUAD, //note that the dual (corny) behaves like the quad so we do not list it here
	CPU_TYPE_IMX35,
	CPU_TYPE_X86,
	CPU_TYPE_UNKNOWN
};

/**
 *	@class WESystem
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief A class that provides system functionality
 *
 *	@namespace WE
 *
 *	"A class without members doesn't need an instance" - Dirk Sterzel
 *	"One could argue if a class without members needs to be a class at all" - Dirk Sterzel
 */
class WESystem
{
public:
	/**
	 *	@fn static bool isProcessRunning(const std::string & p_processName)
	 *
	 *	@brief Function to check if a process identified by p_processName processName is currently running
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_processName String containing the name of a process
	 *
	 *	@returns true if process p_processName is running
	 *	@returns false if process p_processName is not running
	 *
	 */
	static bool isProcessRunning(const std::string & p_processName);

	/**
	 *	@fn static int runSystemCommand(const std::string & p_command)
	 *
	 *	@brief Execute an arbitrary system command
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_command String containing the command that shall be executed
	 *
	 *	@returns WE_SUCCESS (0) if the command was executed successfully and returned with no error
	 *	@returns The error code of the command that was executed
	 *	@returns -1 if the command did not exit normally
	 *
	 *	@warning KNOW WHAT YOU DO BEFORE RUNNING THIS FUNCTION!
	 *	@warning THE COMPLETE SYSTEM CAN BE DAMAGED BY EXECUTING WRONG COMMANDS!
	 */
	static int runSystemCommand(const std::string & p_command);

	/**
	 *	@fn static int getSystemCommandOutput(const std::string & p_command, std::string & p_output)
	 *
	 *	@brief Execute an arbitrary system command and retrieve it's output
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_command String containing the command that shall be executed
	 *	@param p_output Reference to a string where the command's output is stored in
	 *
	 *	@returns WE_SUCCESS (0) if the command was executed successfully and returned with no error. p_output will contain the output of the command
	 *	@returns WE_GENERAL_ERROR if executing the command failed (not the error return from the command itself as in runSystemCommand()
	 *
	 *	@warning KNOW WHAT YOU DO BEFORE RUNNING THIS FUNCTION!
	 *	@warning THE COMPLETE SYSTEM CAN BE DAMAGED BY EXECUTING WRONG COMMANDS!
	 */
	static int getSystemCommandOutput(const std::string & p_command, std::string & p_output);

	/**
	 *	@fn static int getProcessID()
	 *
	 *	@brief Get the process id for the process invoking this method
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *
	 *	@returns process id as returned by the OS.
	 *
	 */
	static int getProcessID();

	/**
	 *	@fn static std::string getSystemVersion()
	 *
	 *	@brief Get the version of the OS running on the terminal
	 *
	 *	@author ajay.bisht@aricent.com
	 *	@author Ajay Bisht
	 *
	 *
	 *	@returns OS version details.
	 *
	 */
	static const std::string getSystemVersion();

	/**
	 * Get the node name of the system. On Linux this is the node name returned by uname.
	 * Empty string on Windows;
	 */
	static const std::string getNodeName();

	/**
	 *	Get free system memory (RAM) in MegaBytes
	 *	@warning delete does not return memory allocated with new directly to the OS.
	 *		After a delete you will NOT get more free RAM (with top or free) directly.
	 */
	static unsigned int getFreeRAM();

	/**
	 *	Get total system memory (RAM) in MegaBytes
	 */
	static unsigned int getTotalRAM();

	/**
	 *	This function can be used to try to free RAM.
	 *	On Linux the following functions are used (see https://www.kernel.org/doc/Documentation/sysctl/vm.txt):
	 *
	 *	/proc/sys/vm/drop_caches:
	 *	From the Kernel documentation: "Writing to this will cause the kernel
	 *	to drop clean caches, as well as reclaimable slab objects like
	 *	dentries and inodes.  Once dropped, their memory becomes free."
	 *
	 *	/proc/sys/vm/compact_memory:
	 *	From the Kernel documentation: "Available only when CONFIG_COMPACTION is set.
	 *	When 1 is written to the file, all zones are compacted such that free memory
	 *	is available in contiguous blocks where possible. This can be important for
	 *	example in the allocation of huge pages although processes will also
	 *	directly compact memory as required."
	 *	This is only done if p_doDefrag is set to true
	 *
	 *	@param p_doDefrag Do compact memory operation (see above)
	 */
	static void tidyUpCaches(bool p_doDefrag = false);

	/**
	 * Get the type of CPU we are currently running on. See enum CPUType
	 */
	static int getCPUType(void);

	/**
	 *	@fn WESystem()
	 *
	 *	@brief Deleted constructor. No instance needed.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WESystem() = delete;
	~WESystem() = delete;

private:

	static int cpuType; //!< enum CPUType
};
} // namespace WE

#endif /* WE_SYSTEM_H_ */
