/*
 * watchdog_daemon_interface.h
 *
 *  Created on: Dec 6, 2021
 *      Author: dsterzel
 */

#ifndef SRC_OS_ABSTRACTION_WATCHDOG_DAEMON_INCLUDE_WATCHDOG_DAEMON_INTERFACE_H_
#define SRC_OS_ABSTRACTION_WATCHDOG_DAEMON_INCLUDE_WATCHDOG_DAEMON_INTERFACE_H_

#include <vector>

namespace WE {

/**
 * Interface class to communicate with the
 * watchdogDaemon.
 * Usage:
 * - Create an instance of this class and provide parameters via constructor or the setXXX functions
 * - Call subscribeToWatchdogDaemon()
 * - From now on call sendHeartbeat() if WE::WETick::getTickCount() is >= getNextFeedTick()
 * - When the application wants to quit gracefully (or if it knows that it cannot send a heartbeat anymore withing
 *   the timeout), call unsubscribeFromWatchdogDaemon()
 */
class WatchdogDaemonInterface
{
public:
	/**
	 * Constructor covering the mandatory parameters
	 */
	WatchdogDaemonInterface(std::string const & p_appName, std::string const & p_restartCommand, std::uint32_t p_timeoutInS);

	/**
	 * Default constructor. Make sure to set the mandatory options with the setXXX functions afterwards
	 */
	WatchdogDaemonInterface();

	/**
	 * Destructor. Only closes the socket to the watchdogDaemon
	 * but does NOT unsubscribe!
	 */
	~WatchdogDaemonInterface();

	/**
	 * No copy allowed
	 */
	WatchdogDaemonInterface(const WatchdogDaemonInterface &) = delete;

	/**
	 * No assignment allowed
	 */
	WatchdogDaemonInterface& operator=(const WatchdogDaemonInterface &) = delete;

	/**
	 * Set the name of the application that shall be monitored
	 */
	inline void setApplicationName(std::string const & p_appName)
	{
		this->appName = p_appName;
	}

	/**
	 * Set the command with which the application can be restarted
	 * (or provide a "reboot" to restart the whole device when the application times out)
	 */
	inline void setRestartCommand(std::string const & p_restartCommand)
	{
		this->restartCommand = p_restartCommand;
	}

	/**
	 * Set the timeout for the monitored application in seconds.
	 * If the application does not send a heartbeat within this timeout, the
	 * watchdogDaemon will execute the restartCommand (making sure the application got killed before that)
	 */
	inline void setTimeout(std::uint32_t p_timeout)
	{
		this->timeout = p_timeout;
	}

	/**
	 * Get the monotonic tick count (see WE::WETick) when the next heartbeat message should be sent.
	 * (this is ~ (tick of last feeding + (timeout / 4))
	 */
	std::uint32_t getNextFeedTick(void);

	/**
	 * Send a subscription request to the watchdogDaemon with the data provided
	 * in constructor or via the setXXX functions.
	 * This function will wait up to 500ms for a response from the watchdogDaemon.
	 *
	 * @return 	WE::WE_SOCKET_ERROR if the IPC connection to the watchdogDaemon could not be created
	 * 		WE::WE_WRITE_ERROR if the subscription request could not be sent to the watchdogDaemon
	 * 		WE::WE_RECEIVE_ERROR if the watchdogDaemon did not answer within 500ms
	 * 		WE::WE_READ_ERROR if the subscription response from watchdogDaemon could not be read from the socket
	 * 		WE::WE_CONNECT_ERROR if the subscription failed (e.g. due to missing data like app name, restart command, timeout)
	 * 		WE::WE_SUCCES if the subscription response from watchdogDaemon was positive
	 */
	std::int32_t subscribeToWatchdogDaemon(void);

	/**
	 * Use this function to send a heartbeat to the watchdogDaemon so it knows that this application
	 * is still alive. Use getNextFeedTick() to determine a good moment for sending the heartbeat.
	 * This function will NOT wait for a response from watchdogDaemon but will return directly
	 * after sending the heartbeat.
	 *
	 * @return 	WE::WE_SOCKET_ERROR if the connection to the watchdogDaemon was not established (use subscribeToWatchdogDaemon() before sending a heartbeat)
	 * 		WE::WE_SEND_ERROR if the heartbeat message could not be constructed properly
	 * 		WE::WE_WRITE_ERROR if the heartbeat message could not be written to the IPC socket
	 * 		WE::WE_SUCCESS if the heartbeat message was sent to the watchdogDaemon successfully
	 */
	std::int32_t sendHeartbeat(void);

	/**
	 * Use this function when the application does not want to be monitored anymore by the watchdogDaemon
	 * (e.g. when doing a graceful exit). If you fail to call this function before exit, the watchdogDaemon
	 * will execute the restart command.
	 * This function will wait for up to 500ms for a response from watchdogDaemon.
	 *
	 * @return 	WE::WE_SOCKET_ERROR if the IPC connection to the watchdogDaemon could not be created
	 * 		WE::WE_WRITE_ERROR if the subscription request could not be sent to the watchdogDaemon
	 * 		WE::WE_RECEIVE_ERROR if the watchdogDaemon did not answer within 500ms
	 * 		WE::WE_READ_ERROR if the subscription response from watchdogDaemon could not be read from the socket
	 * 		WE::WE_CONNECT_ERROR if the subscription failed (e.g. due to missing data like app name, restart command, timeout)
	 * 		WE::WE_SUCCES if the subscription response from watchdogDaemon was positive
	 */
	std::int32_t unsubscribeFromWatchdogDaemon(void);
private:
	std::int32_t connectToWatchdogDaemon(void);

	std::int32_t wddSocket{-1};
	std::vector<char> heartbeatBuffer{};
	std::string appName{};
	std::string restartCommand{};
	std::uint32_t timeout{0};
	std::uint32_t lastFeedTick{0};
	std::int32_t pid{-1};
};
}  // namespace WE



#endif /* SRC_OS_ABSTRACTION_WATCHDOG_DAEMON_INCLUDE_WATCHDOG_DAEMON_INTERFACE_H_ */
