/**
 *	@file we_can_open.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief THE NEW CANOPEN STACK!!!
 *
 */

#ifndef WE_CAN_OPEN_H_
#define WE_CAN_OPEN_H_

#include "../../../event_handling/include/we_signal.h"
#include "../../../dei/variables/include/we_variable.h"
#include "../../../dei/include/we_variable_owner.h"
#include "../../../os_abstraction/endpoint_manager/include/we_endpoint_handler.h"
#include "../../common/include/we_can_protocol.h"

#include "we_can_open_definitions.h"
#include "we_sdo_handler.h"
#include "we_pdo_handler.h"
#include "we_nmt_handler.h"
#include "we_sync_handler.h"
#include "we_emergency_handler.h"
#include "we_can_open_object_dictionary.h"
#include "we_can_open_state_machine.h"

namespace WE
{

typedef std::unordered_map <unsigned int, WEVariable*> CANopenVariableMap; //!< Map to be used for storing the Variables owned by this stack


class WEVariable;
class WECANRaw;
class WECANopenStateMachine;
class WEHeartbeatInterface;

/**
 *	@enum CANopenStackOptions
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Enum describing possible options for the CANopen stack
 *
 *	@namespace WE
 */
enum CANopenStackOptions
{
	CANOPEN_OPTION_AUTO_START,			//!< Don't wait for START_REMOTE_NODE NMT command but go to operational state directly
	CANOPEN_OPTION_EMERGENCY_CONSUMER,		//!< Should this stack be an emergency consumer?
	CANOPEN_OPTION_EMERGENCY_PRODUCER,		//!< Should this stack be an emergency producer?
	CANOPEN_OPTION_DONT_LISTEN_TO_MNT_BROADCAST,	//!< Listen only to MNT messages directly addressed to this stack's Node-ID
	CANOPEN_OPTION_ALLOW_APPLICATION_STATE_CHANGE	//!< The CANopen state is not only maintained by NMT commands but can be set by application
};

/**
 *	@class WECANopen
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief This class is the application interface for the new CANopen stack
 *
 *	@namespace WE
 *
 *	This class can be seen as the new CANopen stack.
 *	This should be the only class of all other CANopen classes floating around here
 *	to be used by an application.
 *	It implements the functions of WECANProtocol.
 *
 *	When using this class the following initialization sequence should be kept
 *	(see function documentation below for each mentioned function):
 *	- Create an instance: WECANopen * co = new WECANopen()
 *	- Set pointers and options:
 *	  - co->setDEI(WEDEI *)
 *	  - co->setCANPortNumber(unsigned int)
 *	  - co->setNodeID(unsigned char)
 *	  - co->setOption(CANopenStackOptions, bool) (for every option you want to set)
 *	- Call init function:
 *	  - co->init()
 *	- Now after the init, object dictionary related stuff can be set, following are examples:
 *	  - Add more SDO channels: co->addSDOChannel(unsigned int, unsigned int, unsigned char, WECANopenSDOChannelType)
 *	  - Add Variables owned by the stack: co->addVariable(WEVariable *)
 *	  - Assign a SDO server to a variable: co->setSDOServerID(unsigned short, unsigned char, unsigned int)
 *	  - Add PDOs: co->addPDO(WECANopenPDOType, unsigned short, unsigned int, unsigned short, unsigned short, unsigned char)
 *	  - Add PDO mappings: co->addPDOMapping(unsigned short, WECANopenPDOType, unsigned short, unsigned short)
 *	  - Add information about other variables to the stack: co->getApplicationVariables()
 *	- After finishing with this you have to subscribe to the WEEndpointManager:
 *	  - co->subscribeToEndpointManager(WEEndPointManager *)
 *	- After this the stack is ready to run
 *	- In your application's main loop or in a thread, you have to call these functions frequently to make the stack run:
 *	  - WEEndpointManager::waitForEndPoints(unsigned int, unsigned int)
 *	  - CANopen::process(unsigned int)
 */
class WECANopen : public WECANProtocol
{
public:
	/**
	 *	@fn WECANopen(const std::string & p_instanceName)
	 *
	 *	@brief Default constructor. Initializes pointers to nullptr and other members to sane start values. Als calls Constructor of WECANProtocol.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	explicit WECANopen(const std::string & p_instanceName);

	/**
	 *	@fn virtual ~WECANopen()
	 *
	 *	@brief Virtual destructor. Frees all occupied memory.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WECANopen();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WECANopen(const WECANopen &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WECANopen & operator=(const WECANopen &) = delete;

	/**
	 *	@fn int init(void)
	 *
	 *	@brief Initialize the CANopen stack
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@return WE_SUCCESS if initialization was successful
	 *	@return WE_INIT_ERROR if no DEI is present or the CAN port number is invalid
	 *
	 *	@see int init(unsigned int p_portNumber, WEDEI * p_dei)
	 *
	 *	Be sure that you have set DEI and the CAN port number before calling this init function.
	 *	This function creates all necessary modules of the stack so it can be run (message handlers
	 *	object dictionary, state machine, etc.)
	 */
	int init(void) override;

	/**
	 *	@fn int process(unsigned int p_time)
	 *
	 *	@brief Call this function frequently to make the stack work.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_time The tick count until which this function is allowed to run (e.g. currentTick + 10).
	 *
	 *	@returns WE_INTERFACE_ERROR If there is no state machine available
	 *	@returns the return value of int WECANopenStateMachine::run(unsigned int  p_time)
	 *
	 *	@see int WECANopenStateMachine::run(unsigned int  p_time)
	 *
	 *	This function implicitly calls the run() function of the state machine that will take care
	 *	of sending timed messages. Message reception works event driven using the WEEndPointManager.
	 */
	int process(unsigned int p_time) override;

	/**
	 *	@fn int addPDO(WECANopenPDOType p_type,	unsigned short p_pdoNumber, unsigned int p_cobID, unsigned short p_eventTime, unsigned short p_inhibitTime, unsigned char p_mode, bool p_active)
	 *
	 *	@brief Add a PDO to the object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_type Type of the PDO (Receive PDO (PDO_RX) or Transmit PDO (PDO_TX))
	 *	@param p_pdoNumber The PDO number. The will determine the object index of the PDO in the object dictionary
	 *	@param p_cobID The COB-ID with which a transmit PDO will be sent or the COB-ID that identifies the PDO on message reception
	 *	@param p_eventTime Depending on p_mode, this is a timer that indicates when a TxPDO shall be sent
	 *	@param p_inhibitTime The inhibit time of a TxPDO
	 *	@param p_mode The transmission mode for Rx- and TxPDOs (see WECANopenPDOTransmissionMode)
	 *	@param p_active is the PDO active or not?
	 *
	 *	@returns WE_INTERFACE_ERROR If no Object Dictionary is available
	 *	@returns WE_SUCCESS If adding the PDO works
	 *
	 *	@see addPDOMapping function
	 *
	 *	@warning Any existing PDO with the same PDO number gets overwritten!
	 *	@warning Adding a PDO while the stack already runs, may produce unexpected behavior.
	 *
	 *	This function adds a PDO with the given data to the object dictionary of this stack.
	 *	It will be ready for use after the CANopen state was in init state (when calling process() function the
	 *	the first time or after doing a reset Node or reset Communication via NMT).
	 */
	int addPDO(WECANopenPDOType p_type,
			unsigned short p_pdoNumber,
			unsigned int p_cobID,
			unsigned short p_eventTime,
			unsigned short p_inhibitTime,
			unsigned char p_mode,
			bool p_active,
			WESignalListener * p_eventListener);

	/**
	 *	@fn int addPDOMapping(unsigned short  p_pdoNumber,
			WECANopenPDOType p_pdoType,
			unsigned short  p_varIndex,
			unsigned short  p_varSubindex,
			int p_length)
	 *
	 *	@brief Add a single mapping entry to an existing PDO
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_pdoNumber The number of the PDO to which this mapping shall be added
	 *	@param p_pdoType The type of the PDO (RX_PDO or TX_PDO) (so that the correct entry in the object dictionary can be detrmined)
	 *	@param p_varIndex The index of an entry in the object dictionary that shall be mapped to the PDO
	 *	@param p_varSubindex Subindex of a sub entry in the object dictionary that shall be mapped to the PDO
	 *	@param p_length Length of the mapping entry in bits
	 *
	 *	@returns WE_INTERFACE_ERROR If no object dictionary exists
	 *	@returns WE_INVALID_PARAMETER_ERROR If the sub entry identified by p_varIndex/p_varSubindex does
	 *		not exist in the object dictionary or is not mappable to a PDO.
	 *	@returns WE_TYPE_ERROR If the sub entry has a data type that can't be mapped to a PDO
	 *	@returns WE_SUCCESS If adding the mapping to the PDO was successfull
	 *
	 *	@see addPDO() function
	 *
	 *	@warning Before calling this function, a PDO with the same number as p_pdoNumber and the same type
	 *		as p_pdoType had to be added to be added with the addPDO function
	 *	@warning Please don't add more mapping entries to the PDO than do fit in it. currently no length
	 *		check is done, so please know what you do!
	 *	@warning If this function does return other than WE_SUCCESS, the whole PDO will be marked as invalid and can never be enabled.
	 *		This prevents false OD entry => value associations.
	 *
	 *	This function adds a PDO mapping entry to the object dictionary. Please do not add more mapping
	 *	entries to a PDO than can fit in according to the mapped data types. These mapping entries will be used
	 *	by the PDO handler automatically at runtime.
	 *	Adding an entry while the stack is already running may lead to undefined behavior.
	 */
	int addPDOMapping(unsigned short  p_pdoNumber,
			WECANopenPDOType p_pdoType,
			unsigned short  p_varIndex,
			unsigned short  p_varSubindex,
			int p_length);

	/**
	 *	@fn int addSDOChannel(unsigned int p_clientID, unsigned int p_serverID, unsigned char p_nodeID, WECANopenSDOChannelType p_type)
	 *
	 *	@brief Add a new SDO channel to this CANopen stack
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_clientID The SDO Client COB-ID
	 *	@param p_serverID The SDO Server COB-ID
	 *	@param p_nodeID the Node ID of the external Server/Client
	 *	@param p_type the SDO channel type, either SDO_CLIENT_CHANNEL or SDO_SERVER_CHANNEL
	 *	@param p_ownerName Name of the variable owner of that channel
	 *
	 *	@returns WE_INTERFACE_ERROR if no object dictionary exists
	 *	@returns WE_WRITE_ERROR if all available object dictionary indices for the specified channel type are already in use
	 *
	 *	@see int setSDOServerID(unsigned short p_index, unsigned short p_subindex, unsigned int p_serverID)
	 *
	 *	With this function you can add additional SDO channels to this CANopen stack. Use this function before
	 *	setting a SDO server ID to a object dictionary sub entry.
	 */
	int addSDOChannel(unsigned int p_clientID, unsigned int p_serverID, unsigned char p_nodeID, WECANopenSDOChannelType p_type, const std::string & p_ownerName);

	/**
	 *	@fn int setSDOServerID(unsigned short p_index, unsigned short p_subindex, unsigned int p_serverID)
	 *
	 *	@brief Set a SDO server COB-ID to an existing object dictionary sub entry
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the sub entry for which the server ID shall be set
	 *	@param p_subindex Sbindex of the sub entry for which the server ID shall be set
	 *	@param p_serverID The SDO server COB-ID that shall be set to the sub entry
	 *
	 *	@returns WE_INTERFACE_ERROR if no object dictionary exists
	 *	@returns The return value of WECANopenObjectDictionary::setSDOServerID
	 *
	 *	@see int WECANopenObjectDictionary::setSDOServerID(unsigned short p_index, unsigned char p_subindex, unsigned int p_serverID)
	 *	@see int addSDOChannel(unsigned int p_clientID, unsigned int p_serverID, unsigned char p_nodeID, WECANopenSDOChannelType p_type);
	 *
	 *	@warning Be sure that the spcified sub entry already exists and that a SDO client channel with the
	 *		specified server ID was already created using the addSDOChannel function.
	 *
	 *	With this function you can specify a SDO server for a sub entry in the object dictionary.
	 *	Whenever one wants to write a value to that sub entry, a SDO Download request will be sent to
	 *	the specified SDO server. The proposed value will only be written to the corresponding sub entry
	 *	if the SDO server answers with a positive SDO Download response. If the server does not answer
	 *	within a certain time (default: 100ms), the value will discarded. Same when an SDO Abort for
	 *	that sub entry is received within the timeout.
	 */
	int setSDOServerID(unsigned short p_index, unsigned short p_subindex, unsigned int p_serverID);

	/**
	 *	@fn int addVariable(WEVariable * p_var)
	 *
	 *	@brief Add a variable to the CANopen stack
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_var Pointer to an existing WEVariable.
	 *
	 *	@returns WE_INVALID_PARAMETER_ERROR If p_var is nullptr
	 *	@returns WE_ALREADY_EXISTS_ERROR If p_var was added to the stack before
	 *	@returns WE_WRITE_ERROR If inserting p_var into the variable map failed
	 *	@returns WE_SUCCESS if adding the variable was successful
	 *
	 *	@warning Currently only 32 bit signed integers can be added. WEVaraible has to be enhanced to
	 *		use more data types.
	 *
	 *	This adds a WEVariable that is owned by this CANopen stack to the stack. The variable type will
	 *	automatically set to CANOPEN_VARIABLE and the owner of the variable is set to this instance.
	 *	The variable will also automatically be added as external varaible to the DEI.
	 */
	int addVariable(WEVariable * p_var) override;

	/**
	 *	@fn unsigned int getHeartbeatTime(void)
	 *
	 *	@brief Get the currently set heartbeat producer time (Index 0x1017, Subindex 0x00)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns The currently configured heartbeat producer time from object dictionary in milliseconds.
	 */
	unsigned int getHeartbeatTime(void);

	/**
	 *	@fn unsigned char getNodeID(void)
	 *
	 *	@brief Get the Node-ID of this CANopen stack
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns the Node-ID of this CANopen stack
	 */
	unsigned char getNodeID(void) const;

	/**
	 *	@fn void setNodeID(unsigned char p_nodeID)
	 *
	 *	@brief Set the node ID of this CANopen stack
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_nodeID The node ID with which this stack is associated
	 *
	 *	@warning Be sure to use only unique Node IDs in your application
	 */
	void setNodeID(unsigned char p_nodeID);

	/**
	 *	@fn virtual int updateRequest(unsigned int p_varKey, int p_intValue)
	 *
	 *	@brief Request a value change for a variable owned by this WEVariableOwner (=> this CANopen stack)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_varKey The combined Index/Subindex of a variable owned by this stack instance
	 *	@param p_intValue The value that shall be set
	 *
	 *	@returns WE_INTERFACE_ERROR if no SDOHandler exists
	 *	@returns The return value of int WESDOHandler::handleUpdateRequest(unsigned int p_varKey, int p_intValue)
	 *
	 *	@see int WESDOHandler::handleUpdateRequest(unsigned int p_varKey, int p_intValue)
	 *
	 *	Trying to write a new value to a variable owned by a CANopen stack makes only sense if
	 *	a SDO server ID was specified for that variable. A SDO Download Request will be sent to the specified
	 *	SDO server.
	 *	If values shall be transmitted with a PDO, the mapped variables should be internal variables
	 *	owned by the application.
	 */
	virtual int updateRequest(unsigned int p_varKey, int p_intValue) override;

	/**
	 *	@fn virtual int updateRequest(unsigned int p_varKey, const std::string & p_stringValue)
	 *
	 *	@brief Request a value change for a variable owned by this WEVariableOwner (=> this CANopen stack)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_varKey The combined Index/Subindex of a variable owned by this stack instance
	 *	@param p_stringValue Reference to the value that shall be set
	 *
	 *	@returns WE_INTERFACE_ERROR if no SDOHandler exists
	 *	@returns The return value of int WESDOHandler::handleUpdateRequest(unsigned int p_varKey, int p_intValue)
	 *
	 *	@see int WESDOHandler::handleUpdateRequest(unsigned int p_varKey, int p_intValue)
	 *
	 *	Trying to write a new value to a variable owned by a CANopen stack makes only sense if
	 *	a SDO server ID was specified for that variable. A SDO Download Request will be sent to the specified
	 *	SDO server.
	 *	If values shall be transmitted with a PDO, the mapped variables should be internal variables
	 *	owned by the application.
	 */
	virtual int updateRequest(unsigned int p_varKey, const std::string & p_stringValue) override;

	// JU
	/**
	 *	@fn virtual int getValueRequest(unsigned int p_varKey)
	 *
	 *	@brief Request to get a Value of a variable owned by a foreign stack
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim urbach
	 *
	 *	@param p_varKey The combined Index/Subindex of a variable owned by this stack instance
	 *
	 *	@returns WE_INTERFACE_ERROR if no SDOHandler exists
	 *	@returns The return value of int WESDOHandler::handlegetValueRequest(unsigned int p_varKey)
	 *
	 *	@see int WESDOHandler::handlegetValueRequest(unsigned int p_varKey)
	 *
	 *  Trying to read a value of a foreign object dictionary subentry. Makes only sense if
	 *	a SDO server ID was specified for that variable and the variable must have an external identifier.
	 *	A SDO Upload Request will be sent to the specified SDO server.
	 *
	 */
	virtual int getValueRequest(unsigned int p_varKey) override;

	/**
	 *	@brief Called when a variable mapped to a TxPDO has changed
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_signal WESignal containing information about the changed variable value
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR If event type is not a variable change event
	 *	@returns WE_INTERFACE_ERROR If neither the object dictionary nor the PDO handler exist
	 *	@returns the return value of int WEPDOHandler::handleVariableChange(unsigned int p_variableKey, int p_value)
	 *
	 *	@see WESignalListener
	 *
	 *	This function is only used to indicate the value change of a variable mapped to a TxPDO.
	 *	If the corresponding PDO is configured to be sent on value change, the sending of the PDO will be triggered.
	 */
	int signalOccurred(WESignal & p_signal) override;

	/**
	 *	@fn virtual void onEndPointReadable(WEEndPoint* endPoint, WEEndPointManager* endPointMgr)
	 *
	 *	@brief Signal that a message on the CAN port of this stack arrived
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param endPoint the endpoint (in this case CAN port) that got readable
	 *	@param endPointManager the corresponding endpoint manager (not used)
	 *
	 *	This function will be called by WEEndPointManager as soon as the CAN port of this stack gets
	 *	readable (a CAN message arrived). The CANopen stack then reads the message from CAN and starts
	 *	to process it. The message will be passed to the single message handlers until one of them identifies it
	 *	to belong to it and then processes it.
	 */
	virtual void onEndPointReadable(std::shared_ptr<WEEndPoint> p_endPoint, WEEndPointManager* p_endPointMgr) override;

	/**
	 *	@fn virtual void onEndPointReconnected(WEEndPoint* endpoint)
	 *
	 *	@brief Signal that a disconnected endpoint got connected again
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param endpoint the re-connected endpoint
	 *
	 *	This function is not used.
	 */
	virtual void onEndPointReconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;

	/**
	 *	@fn int setOption(int p_obtion, bool p_enable)
	 *
	 *	@brief Set various options to this CANopen stack
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_option the option to enable/disable
	 *	@param p_enable true for enabling an option, false for disabling an option
	 *
	 *	@returns WE_SUCCESS if setting an option was successful
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_option is not specified or supported
	 *
	 *	@see enum CANopenStackOptions for a list of available options
	 */
	int setOption(int p_obtion, bool p_enable);

	/**
	 *	@fn int getApplicationVariables(void)
	 *
	 *	@brief Populate object dictionary with variable entries from the application
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	Should return WE_SUCCESS if populating the object dictionary was successful - but never does.
	 *	@returns WE_INTERFACE_ERROR if this instance is not properly initialized or does not have an object dictionary yet.
	 *	@returns WE_GENERAL_ERROR if a copy of all variables can't be obtained from DEI ... and even in case of success :-(
	 *
	 *	@warning Seems to work as long as return value is ignored.
	 *
	 *	This function is used to ask DEI for all it's variable entries so that corresponding
	 *	object dictionary entries can be created.
	 */
	int getApplicationVariables(void);

	unsigned int getPortNumber(void) const
	{
		return this->resources->portNo;
	}

	bool isInitialized(void) const
	{
		return this->initialized;
	}

	int setCANopenState(int p_state);
	int getCANopenState(void) const;

	void setPaused(bool p_paused);
	bool isPaused(void) const;

	int setPDOAction(int p_PDONumber, WECANopenPDOType p_type, int p_action, unsigned int p_key);

	const WEVariable * getVariable(unsigned int p_key);

	const WEVariable * getVariableByExternalIdentifier(unsigned short p_index, unsigned short p_subindex, unsigned int p_clientID);

	int setSDOEventListener(WESignalListener * p_listener);

	void resetValue(unsigned int p_key);

	int addHeartbeatConsumer(unsigned char p_nodeID, unsigned short p_heartbeatTime);
	int addHeartbeatListener(WESignalListener * p_listener);
	int addStateListener(WESignalListener * p_listener);
	/*
	 * CAN thread interface method
	 */
	bool runProtocol(unsigned int p_time) override;

	void setHeartbeatProducerTime(int p_time);

	CANopenManSpecDataType * getManufacturerSpecificDataTypeDefinition(WECANopenDataType p_type);

	void addHeartbeatListener(WESignalListener * p_listener, int p_nodeID);

	void setSDOTimeout(unsigned int p_timeout);

	inline unsigned int getSDOTimeout(void) const
	{
		return this->sdoTimeout;
	}

	virtual int enableMessage(unsigned int p_objectID, bool p_enable) override;

	void sendPDOEvent(enum WEEventType p_eventType, unsigned short p_odIndex);

	void setPDOEventAction(unsigned short p_odIndex, int p_eventType, std::shared_ptr<WE::WEAction> p_varAction);

	static int debugSection;

private:
	WECANopenDataType calculateDataType(int p_numberOfBits, bool p_isSigned);
	WECANopenDataType createDataTypeIfNotExisting(int p_numberOfBits, bool p_isSigned);


	WECANopenStateMachine stateMachineCANopen{this};	//!< Pointer to this CANopen stack's state machine
	WEHeartbeatInterface * heartbeatInterface;	//!< Pointer to the heartbeat listener
	WECANopenObjectDictionary od{};			//!< The CANopen object dictionary
	WESDOHandler sdo{};				//!< The SDO handler
	WEPDOHandler pdo{this};				//!< The PDO handler
	WENMTHandler nmt{};				//!< The NMT handler
	WESyncHandler sync{this};				//!< The Sync handler
	WEEmergencyHandler emcy{};			//!< The Emergency handler

	unsigned char nodeID;		//!< The node ID of this CANopen stack
	//options:
	bool autoStart;			//!< If set to true, this stack will go directly from INIT to PRE-Operational to Operational without any NMT commands (default: false)
	bool emergencyProducer;		//!< If set to true, this stack shall be an emergency producer (default: false)
	bool emergencyConsumer;		//!< If set to true, this stack shall act as an Emergency consumer (default: false)
	bool onlyDirectNMTAddressing;	//!< If set to true, the stack shall ignore NMT broadcast messages and only
					//!< listen to NMT messages directly addressed to this stack's Node-ID (default: false)
	bool stateChangeAllowed;

	bool initialized;

	int heartbeatProducerTime;

	unsigned int sdoTimeout;	//how long we wait for a SDO Download request to time out

	CANopenManSpecDataType manSpecDataTypes[8];	//!< 8 manufacturer specific data types allowed
	CANopenVariableMap variableMap;			//!< A map of all variables owned by this stack
	std::unordered_map< unsigned short, WESignalListener * > pdoEventListener;
	std::unordered_map< unsigned short, std::map<int, std::shared_ptr<WE::WEAction>> > pdoEventActionMap;

	std::map<unsigned int, std::string> sdoClientIDtoOwnerName;
};
}

#endif /* WE_CAN_OPEN_H_ */
