/**
 *	@file we_can_open_definitions.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief This file contains definitions of types, error codes, constants, etc.
 *
 */


#ifndef WE_CAN_OPEN_DEFINITIONS_H_
#define WE_CAN_OPEN_DEFINITIONS_H_

namespace WE {

/**
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Transmit or receive PDO
 *
 *	@namespace WE
 */
typedef enum
{
	PDO_RX,
	PDO_TX
}WECANopenPDOType;


/**
 *	@enum WECANopenPDOTransmissionMode
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Possible PDO transmission modes (excluding the ones between 1 and 252, just use PDO_TX_MODE_CYCLIC_SYNCHRONOUS + X)
 *
 *	@namespace WE
 */
enum WECANopenPDOTransmissionMode
{
	PDO_TX_MODE_ACYCLIC_SYNCHRONOUS		= 0,	//!< Send with sync message but only if event happened before
	PDO_TX_MODE_CYCLIC_SYNCHRONOUS		= 1,	//!< Send with every sync message
	//Between 1 and 251 are the factors for the sync message (every Xth sync message)
	PDO_TX_MODE_RTR_SYNCHROONOUS		= 252,	//!< RTR NOT SUPPORTED!
	PDO_TX_MODE_RTR_ASYNCHROONOUS		= 253,	//!< RTR NOT SUPPORTED!
	PDO_TX_MODE_ACYCLIC_ASYNCHRONOUS_M	= 254,	//!< Send at any variable change and/or periodic
	PDO_TX_MODE_ACYCLIC_ASYNCHRONOUS_D	= 255	//!< Send at any variable change and/or periodic (actually not supported because this is device specific and we don't follow a device profile)
};

/**
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Possible kinds of SDO channels
 *
 *	@namespace WE
 */
typedef enum
{
	SDO_CLIENT_CHANNEL,
	SDO_SERVER_CHANNEL
}WECANopenSDOChannelType;


/**
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Possible CANopen object types (cuurently only RECORD and VAR supported in object dictionary)
 *
 *	@namespace WE
 */

typedef enum
{
	CANOPEN_OBJECT_TYPE_NULL	= 0,
	CANOPEN_OBJECT_TYPE_DOMAIN	= 2,
	CANOPEN_OBJECT_TYPE_DEFTYPE	= 5,
	CANOPEN_OBJECT_TYPE_DEFSTRUCT	= 6,
	CANOPEN_OBJECT_TYPE_VAR		= 7,
	CANOPEN_OBJECT_TYPE_ARRAY	= 8,
	CANOPEN_OBJECT_TYPE_RECORD	= 9
}WECANopenObjectType;

/**
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief CANopen data types
 *
 *	@namespace WE
 */
typedef enum
{
	CANOPEN_DATA_TYPE_UNDEFINED		= 0x00,
	CANOPEN_DATA_TYPE_BOOLEAN		= 0x01,
	CANOPEN_DATA_TYPE_SIGNED_INT_8		= 0x02,
	CANOPEN_DATA_TYPE_SIGNED_INT_16		= 0x03,
	CANOPEN_DATA_TYPE_SIGNED_INT_32		= 0x04,
	CANOPEN_DATA_TYPE_UNSIGNED_INT_8	= 0x05,
	CANOPEN_DATA_TYPE_UNSIGNED_INT_16	= 0x06,
	CANOPEN_DATA_TYPE_UNSIGNED_INT_32	= 0x07,
	CANOPEN_DATA_TYPE_REAL_32		= 0x08,
	CANOPEN_DATA_TYPE_VISIBLE_STRING	= 0x09,
	CANOPEN_DATA_TYPE_OCTET_STRING		= 0x0a,
	CANOPEN_DATA_TYPE_UNICODE_STRING	= 0x0b,
	CANOPEN_DATA_TYPE_TIME_OF_DAY		= 0x0c,
	CANOPEN_DATA_TYPE_TIME_DIFFERENCE	= 0x0d,

	CANOPEN_DATA_TYPE_DOMAIN		= 0x0f,
	CANOPEN_DATA_TYPE_SIGNED_INT_24		= 0x10,
	CANOPEN_DATA_TYPE_REAL64		= 0x11,
	CANOPEN_DATA_TYPE_SIGNED_INT_40		= 0x12,
	CANOPEN_DATA_TYPE_SIGNED_INT_48		= 0x13,
	CANOPEN_DATA_TYPE_SIGNED_INT_56		= 0x14,
	CANOPEN_DATA_TYPE_SIGNED_INT_64		= 0x15,
	CANOPEN_DATA_TYPE_UNSIGNED_INT_24	= 0x16,

	CANOPEN_DATA_TYPE_UNSIGNED_INT_40	= 0x18,
	CANOPEN_DATA_TYPE_UNSIGNED_INT_48	= 0x19,
	CANOPEN_DATA_TYPE_UNSIGNED_INT_56	= 0x1a,
	CANOPEN_DATA_TYPE_UNSIGNED_INT_64	= 0x1b,

	CANOPEN_DATA_TYPE_PDO_COM_PARA		= 0x20,
	CANOPEN_DATA_TYPE_PDO_MAPPING		= 0x21,
	CANOPEN_DATA_TYPE_SDO_PARAM		= 0x22,
	CANOPEN_DATA_TYPE_IDENTITY		= 0x23,

	CANOPEN_DATA_TYPE_MAN_SPEC_1		= 0x40,
	CANOPEN_DATA_TYPE_MAN_SPEC_2		= 0x41,
	CANOPEN_DATA_TYPE_MAN_SPEC_3		= 0x42,
	CANOPEN_DATA_TYPE_MAN_SPEC_4		= 0x43,
	CANOPEN_DATA_TYPE_MAN_SPEC_5		= 0x44,
	CANOPEN_DATA_TYPE_MAN_SPEC_6		= 0x45,
	CANOPEN_DATA_TYPE_MAN_SPEC_7		= 0x46,
	CANOPEN_DATA_TYPE_MAN_SPEC_8		= 0x47
}WECANopenDataType;



typedef struct
{
	int numberOfBits;
	bool isSigned;
	unsigned short odIndex;
}CANopenManSpecDataType;

/**
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Possible access types for object dictionary sub entries
 *
 *	@namespace WE
 */
typedef enum
{
	CANOPEN_ACCESS_TYPE_RO		= 0x00,
	CANOPEN_ACCESS_TYPE_WO		= 0x01,
	CANOPEN_ACCESS_TYPE_RW		= 0x02,
	CANOPEN_ACCESS_TYPE_RWW		= 0x03,
	CANOPEN_ACCESS_TYPE_RWR		= 0x04,
	CANOPEN_ACCESS_TYPE_CONST	= 0x05
}WECANopenAccessType;

/**
 *	@enum CANopenNMTStates
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Possible CANopen NMT states (only used internal, don't mix up with CANopenHeartbeatStates)
 *
 *	@namespace WE
 */
enum CANopenNMTStates
{
	CANOPEN_STATE_INITIALISATION,
	CANOPEN_STATE_RESET_APPLICATION,
	CANOPEN_STATE_RESET_COMMUNICATION,
	CANOPEN_STATE_PRE_OPERATIONAL,
	CANOPEN_STATE_OPERATIONAL,
	CANOPEN_STATE_STOPPED,

	CANOPEN_STATE_FATAL_ERROR
};

/**
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Possible CANopen NMT commands to change the state
 *
 *	@namespace WE
 */
typedef enum
{
	CANOPEN_NMT_SET_OPERATIONAL		= 0x01,
	CANOPEN_NMT_SET_STOPPED			= 0x02,
	CANOPEN_NMT_SET_PREOPERATIONAL		= 0x80,
	CANOPEN_NMT_SET_RESET_APPLICATION	= 0x81,
	CANOPEN_NMT_SET_RESET_COMMUNICATION	= 0x82
}CANopenNMTStateChangeCommands;

/**
 *	@enum CANopenHeartbeatStates
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Possible heartbeat message state values
 *
 *	@namespace WE
 */
enum CANopenHeartbeatStates
{
	CANOPEN_HEARTBEAT_BOOTUP 		= 0x0,
	CANOPEN_HEARTBEAT_STOPPED 		= 0x4,
	CANOPEN_HEARTBEAT_OPERATIONAL 		= 0x5,
	CANOPEN_HEARTBEAT_PRE_OPERATIONAL 	= 0x7f
};

/**
 *	@enum CANopenSDOClientCommandSpecifier
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Values for SDO CCS fields in a SDO message
 *
 *	@namespace WE
 */
enum CANopenSDOClientCommandSpecifier
{
	SDO_CCS_DOWNLOAD_SEGMENT		= 0,
	SDO_CCS_INITIATE_DOWNLOAD_REQUEST 	= 1,	//001
	SDO_CCS_INITIATE_UPLOAD_REQUEST 	= 2,	//010
	SDO_CCS_INITIATE_DOWNLOAD_RESPONSE 	= 3,	//011
	SDO_CCS_ABORT_SDO_TRANSFER 		= 4
};

// JU
enum CANopenSDOServerCommandSpecifier
{
	SDO_SCS_UPLOAD_SEGMENT			= 0,
	SDO_SCS_INITIATE_UPLOAD_RESPONSE 	= 2	//010
};

/**
 *	@enum CANopenSDOAbortCodes
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Pre-defined SDO abort codes
 *
 *	@namespace WE
 */
enum CANopenSDOAbortCodes
{
	SDO_ABORT_CODE_TOGGLE_BIT_NOT_ALTERED		= 0x05030000,
	SDO_ABORT_CODE_SDO_PROTOCOL_TIMED_OUT		= 0x05040000,
	SDO_ABORT_CODE_CS_NOT_VALID_OR_UNKNOWN		= 0x05040001,
	SDO_ABORT_CODE_INVALID_BLOCKSIZE		= 0x05040002,
	SDO_ABORT_CODE_INVALID_SEQUENCE_NUMBER		= 0x05040003,
	SDO_ABORT_CODE_CRC_ERROR			= 0x05040004,
	SDO_ABORT_CODE_OUT_OF_MEMORY			= 0x05040005,
	SDO_ABORT_CODE_UNSUPPORTED_ACCESS_TO_OBJECT	= 0x06010000,
	SDO_ABORT_CODE_READ_WRITEONLY_OBJECT		= 0x06010001,
	SDO_ABORT_CODE_WRITE_READONLY_OBJECT		= 0x06010002,
	SDO_ABORT_CODE_OBJECT_DOES_NOT_EXIST		= 0x06020000,
	SDO_ABORT_CODE_OBJECT_CANNOT_BEMAPPED_TO_PDO	= 0x06040041,
	SDO_ABORT_CODE_PDO_LENGTH_EXCEEDED		= 0x06040042,
	SDO_ABORT_CODE_PARAMETER_INCOMPATIBLE		= 0x06040043,
	SDO_ABORT_CODE_DEVICE_INCOMPATIBLE		= 0x06040047,
	SDO_ABORT_CODE_ACCESS_FAILED_HARDWARE_ERROR	= 0x06060000,
	SDO_ABORT_CODE_DATATYPE_SERVICE_PARAMETER	= 0x06070010,
	SDO_ABORT_CODE_DATATYPE_LENGTH_TOO_HIGH		= 0x06070012,
	SDO_ABORT_CODE_DATATYPE_LENGTH_TOO_LOW		= 0x06070013,
	SDO_ABORT_CODE_SUBINDEX_DOES_NOT_EXIST		= 0x06090011,
	SDO_ABORT_CODE_VALUE_RANGE_EXCEEDED		= 0x06090030,
	SDO_ABORT_CODE_VALUE_OF_PARMA_TOO_HIGH		= 0x06090031,
	SDO_ABORT_CODE_VALUE_OF_PARMA_TOO_LOW		= 0x06090032,
	SDO_ABORT_CODE_MAX_LESS_THAN_MIN		= 0x06090036,
	SDO_ABORT_CODE_GENERAL_ERROR			= 0x08000000,
	SDO_ABORT_CODE_APPLICATION_TRANSFER		= 0x08000020,
	SDO_ABORT_CODE_APPLICATION_TRANSFER_LOCAL	= 0x08000021,
	SDO_ABORT_CODE_APPLICATION_TRANSFER_DEVICE	= 0x08000022,
	SDO_ABORT_CODE_OBJECT_DICTIONARY_ERROR		= 0x08000023
};

/**
 *	@enum CANopenEmergencyCodes
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Pre-defined CANopen emergency message error codes
 *
 *	@namespace WE
 */
enum CANopenEmergencyCodes
{
	EMCY_ERROR_RESET_OR_NO_ERROR		= 0x00,

	EMCY_GENERIC_ERROR			= 0x10,

	EMCY_CURRENT				= 0x20,
		EMCY_CURRENT_DEVICE_INPUT_SIDE	= 0x21,
		EMCY_CURRENT_INSIDE_THE_DEVICE	= 0x22,
		EMCY_CURRENT_DEVICE_OUTPUT_SIDE	= 0x23,

	EMCY_VOLTAGE				= 0x30,
		EMCY_MAINS_VOLTAGE		= 0x31,
		EMCY_VOLTAGE_INSIDE_THE_DEVICE	= 0x32,
		EMCY_OUTPUT_VOLTAGE		= 0x33,

	EMCY_TEMPERATURE			= 0x40,
		EMCY_AMBIENT_TEMPERATURE	= 0x41,
		EMCY_DEVICE_TEMPERATURE		= 0x42,

	EMCY_DEVICE_HARDWARE			= 0x50,

	EMCY_DEVICE_SOFTWARE			= 0x60,
		EMCY_INTERNAL_SOFTWARE		= 0x61,
		EMCY_USER_SOFTWARE		= 0x62,
		EMCY_DATA_SET			= 0x63,

	EMCY_ADDITIONAL_MODULES			= 0x70,

	EMCY_MONITORING				= 0x80,
		EMCY_COMMUNICATION		= 0x81,
			EMCY_CAN_OVERRUN_OBJECT_LOST	= 0x8110,
			EMCY_CAN_IN_ERROR_PASSIVE_MODE	= 0x8120,
			EMCY_LIFEGUARD_HEARTBEAT_ERROR	= 0x8130,
			EMCY_RECOVERED_FROM_BUS_OFF	= 0x8140,
			EMCY_TRANSMIT_COB_ID_COLLISION	= 0x8150,
		EMCY_PROTOCOL_ERROR		= 0x82,
			EMCY_PDO_NOT_PROCESSED_DUE_TO_LENGTH_ERROR	= 0x8210,
			EMCY_PDO_LENGTH_EXCEEDED			= 0x8220,

	EMCY_EXTERNAL_ERROR			= 0x90,

	EMCY_ADDITIONAL_FUNCTIONS		= 0xf0,

	EMCY_DEVICE_SPECIFIC			= 0xff
};

enum PDOActions
{
	PDO_NO_ACTION,
	PDO_INCREMENT_BEFORE_SENT,
	PDO_INCREMENT_AFTER_SEND
};

static const int CANOPEN_WACHENDORFF_VENDOR_ID = 0x0000021f;	//!< The Wachendorff Elektronik CANopen vendor ID
}  // namespace WE

#endif /* WE_CAN_OPEN_DEFINITIONS_H_ */
