/**
 *	@file we_can_open_instance_manager.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of the WECANopenInstanceManager class
 *
 */


#ifndef WE_CAN_OPEN_INSTANCE_MANAGER_H_
#define WE_CAN_OPEN_INSTANCE_MANAGER_H_

#include <map>

namespace WE
{

class WECANopen;
class WETick;

/**
 *	@class WECANopenInstanceManager
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Manager of all CANopen instances.
 *
 *	@namespace WE
 *
 *	@see WECANopen
 *
 *	This singleton class can act as a manger of CANopen instances. A Pointer to a WECANopen instance
 *	can be added to this class and later be retrieved via the port number and (optional) the node ID of the
 *	instance.
 *	To clean all used memory, the deleteAllInstances function needs to be called at the exit of your
 *	application.
 */
class WECANopenInstanceManager
{
public:
	/**
	 *	@fn ~WECANopenInstanceManager()
	 *
	 *	@brief Destructor. Does NOT free memory of containing WECANopen instances. Use deleteAllInstances() first!
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WECANopenInstanceManager();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WECANopenInstanceManager(const WECANopenInstanceManager &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WECANopenInstanceManager & operator=(const WECANopenInstanceManager &) = delete;

	/**
	 *	@fn static WECANopenInstanceManager * instance(void)
	 *
	 *	@brief Retrieve the singleton instance of this class.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns The one and only instance of this class.
	 *
	 *	If no instance of this class existed, yet it will be automatically created and returned.
	 */
	static WECANopenInstanceManager * instance(void);

	/**
	 *	@fn void addCANopenInstance(WECANopen * p_instance)
	 *
	 *	@brief Add an existing WECANopen instance.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_instance Pointer to a existing WECANopen instance.
	 *
	 *	@warning Be sure that CAN port number and the node ID was set to the WECANopen instance before you
	 *		add it to the manager.
	 */
	void addCANopenInstance(WECANopen * p_instance);

	/**
	 *	@fn WECANopen * getCANopenInstance(unsigned int p_portNumber, unsigned char p_nodeID = 0x00)
	 *
	 *	@brief Retrieve a particular WECANopen instance
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_portNumber CAN port number of the WECANopen instance to retrieve.
	 *	@param p_nodeID Node ID of the WECANopen instance to retrieve (optional)
	 *
	 *	@returns A Pointer to the WECANopen instance identified by p_portNumber and (optional) p_nodeID
	 *
	 *	If multiple WECANopen instances per CAN port are used, be sure that the Node ID is provided. Else
	 *	the first found WECANopen instance for the given port will be returned.
	 */
	WECANopen * getCANopenInstance(unsigned int p_portNumber, unsigned char p_nodeID = 0x00);

	/**
	 *	@fn void updateAllData(unsigned int p_timeout)
	 *
	 *	@brief Call the "process" function of all WECANopen instances managed by this class.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timeout How long this function is allowed to run (needs some more work)
	 *
	 *	@returns Currently nothing, later probably WE_SUCCESS or TIMEOUT_ERROR. Need to figure out when updating the usage in PClient.
	 *
	 *	@see WECANopen::process()
	 *
	 *	This function can be used to automatically call the process function of all WECANopen stacks.
	 */
	void updateAllData(unsigned int p_timeout);

	/**
	 *	@fn void deleteAllInstances(void)
	 *
	 *	@brief Delete all WECANopen instances managed by this class. Free the memory.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	Use this function to delete all WECANopen instances managed by this class. The destructors will
	 *	be called and the internal map cleared.
	 */
	void deleteAllInstances(void);

private:
	WECANopenInstanceManager() = default;	//!< Private constructor

	std::multimap<unsigned short, WECANopen*> allInstances;	//!< Multimap containing all instances sorted by port number
};
}  // namespace WE

#endif /* WE_CAN_OPEN_INSTANCE_MANAGER_H_ */
