/**
 *	@file we_can_open_object_dictionary.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief The CANopen Object Dictionary class
 *
 *	@see we_canopen_definitions.h for type definitions
 */

#ifndef WE_CAN_OPEN_OBJECT_DICTIONARY_H_
#define WE_CAN_OPEN_OBJECT_DICTIONARY_H_


#include <string>
#include "we_can_open_definitions.h"
#include "../../../dei/variables/include/we_variable.h"

namespace WE
{

/**
 *	@struct ObjectDictionarySubEntry
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Representation of a sub-entry in the CANopen object dictionary
 *
 *	@namespace WE
 *
 *	@warning An entry with CANopen object type VAR also has exactly one sub entry (0)
 *
 *	This struct stores all necessary information of n sub entry in the Object Dictionary
 */
struct ObjectDictionarySubEntry
{
	unsigned char subindex{0};		//!< This sub-entries subindex
	WECANopenAccessType accessType{CANOPEN_ACCESS_TYPE_RO};	//!< The access type for this sub entry
	WECANopenDataType dataType{CANOPEN_DATA_TYPE_SIGNED_INT_32};	//!< The data type for this sub entry
	int intValue{0};			//!< The current value of this sub entry
	std::string strValue{""};		//!< The current value of this sub entry (if dataType = CANOPEN_DATA_TYPE_VISIBLE_STRING)
	unsigned int sdoServerID{0x000};	//!< If assigned to a different SDO server, this member contains the COB-ID of that server
	std::string name{""};			//!< The name (description) of this sub entry
	bool mappable{false};			//!< Can this sub-entry be mapped to a PDO?
	int defaultIntValue{0};			//!< The default value for this sub entry
	std::string defaultStringValue{""};	//!< The default value for this sub entry
	int lowLimit{0};			//!< The low limit for this entry (not for strings)
	int highLimit{0x7fffffff};		//!< The high limit for this entry (not for strings)
	bool valueManagedByDEI{false};		//!< All manufacturer specific entries are managed by DEI, communication specific stuff is managed by CANopen stack
	bool mappedInRxPDO{false};
	bool mappedInTxPDO{false};

	/**
	 *	@fn ObjectDictionarySubEntry()
	 *
	 *	@brief Default constructor to initialize members with (sane?) values
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	ObjectDictionarySubEntry() = default;
};

typedef std::unordered_map<unsigned char, ObjectDictionarySubEntry *> SubEntries;	//!< type definition for the sub entry map, used by every entry !!!DON'T USE UNORDERED MAP HERE!!!!

/**
 *	@struct ObjectDictionaryEntry
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief This represents an (Top-Level) entry of the CANopen object dictionary (=> Object)
 *
 *	@namespace WE
 *
 *	This struct stores the values needed to identify a top-level entry of the Object Dictionary.
 *	It has map of sub entries that are associated to this object
 */
struct ObjectDictionaryEntry
{
	unsigned short index{0};	//!< The index of this entry in the object dictionary
	SubEntries subindices{};	//!< All associated sub entries
	WECANopenObjectType type{CANOPEN_OBJECT_TYPE_VAR};	//!< The object type of this entry
	std::string name{""};	//!< The name of this entry

	/**
	 *	@fn ObjectDictionaryEntry()
	 *
	 *	@brief Just to initialize the members
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	ObjectDictionaryEntry() = default;
};

typedef std::unordered_map<unsigned short, ObjectDictionaryEntry *> ODEntries; //!< Type definition for the map that contains all OD entries !!!DON'T USE UNORDERED MAP HERE!!!!

//forward declarations
class WEDEI;
class WECANopen;
class WESDOHandler;
class WEPDOHandler;
class WENMTHandler;

/**
 *	@class WECANopenObjectDictionary
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief This is the CANopen Object Dictionary class
 *
 *	@namespace WE
 *
 *	This class is responsible for managing all Object Dictionary entries.
 *	It takes care of getting and setting values and also synchronizes the values
 *	in the dictionary with WEDEI. It also provides various information about
 *	the entries.
 *	Don't use this class from outside. It is intended to be used by CANopen class only.
 */
class WECANopenObjectDictionary
{
public:
	/**
	 *	@fn WECANopenObjectDictionary()
	 *
	 *	@brief Default constructor, initializes all pointers with nullptr
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WECANopenObjectDictionary();

	/**
	 *	@fn ~WECANopenObjectDictionary()
	 *
	 *	@brief Destructor, clears the object dictionary entries and frees memory
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WECANopenObjectDictionary();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WECANopenObjectDictionary(const WECANopenObjectDictionary &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WECANopenObjectDictionary & operator=(const WECANopenObjectDictionary &) = delete;

	/**
	 *	@fn int addEntry(unsigned short p_index, WECANopenObjectType p_type, const std::string & p_name)
	 *
	 *	@brief Adds an Object Dictionary Object to the object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the new entry
	 *	@param p_type Type of the object (VAR, RECORD, etc.)
	 *	@param p_name Name of the new object entry
	 *
	 *	@returns WE_ALREADY_EXISTS_ERROR If an object with p_index already exists in the object dictionary
	 *	@returns WE_SUCCESS if the entry was added successfully
	 *
	 *	@see addSubentry
	 *
	 *	@warning Be sure to call addSubentry function even for VAR types
	 *
	 *	By calling this function a new ObjectDictionaryEntry will be created and added to
	 *	the object dictionary.
	 */
	int addEntry(unsigned short p_index, WECANopenObjectType p_type, const std::string & p_name);

	/**
	 *	@fn int addSubentry(unsigned short p_index,
	 *		unsigned char p_subindex,
	 *		WECANopenAccessType p_access,
	 *		WECANopenDataType p_dataType,
	 *		const std::string & p_name,
	 *		bool p_mappable,
	 *		int p_defaultIntValue,
	 *		const std::string & p_defaultStringValue,
	 *		unsigned int p_sdoServerID = 0x0)
	 *
	 *	@brief Use this function to create a new ObjectDictionarySubEntry in the object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of an EXISTING object dictionary entry
	 *	@param p_subindex Subindex of the new sub entry
	 *	@param p_access Access type of the sub entry (RO; RW; etc.)
	 *	@param p_dataType Data type of the new sub entry (like UNSIGNED8, VISIBLE_STRING, etc.)
	 *	@param p_name Name of the new sub entry
	 *	@param p_mappable Is this entry mappable to a PDO?
	 *	@param p_defaultIntValue Default value (not changable later)
	 *	@param p_defaultStringValue Default value (not changable later)
	 *	@param p_sdoServerID If the entry shall be owned by another CANopen server, this shall be != 0 and contain the COB-ID of the SDO server
	 *
	 *	@returns WE_SUCCESS If adding the new sub entry was successful
	 *	@returns WE_NOT_FOUND_ERROR if no object with p_index exists in the object dictionary
	 *
	 *	@see addEntry
	 *
	 *	@warning An object with Index p_index needs to exist in the object dictionary before calling this function
	 *
	 *	This function creates a new sub entry in the object dictionary. The value of subindex 0 of VAR and
	 *	RECORD objects gets updated automatically.
	 */
	int addSubentry(unsigned short p_index,
			unsigned char p_subindex,
			WECANopenAccessType p_access,
			WECANopenDataType p_dataType,
			const std::string & p_name,
			bool p_mappable,
			int p_defaultIntValue,
			const std::string & p_defaultStringValue,
			unsigned int p_sdoServerID = 0x0);

	/**
	 *	@fn int getValue(unsigned int p_indexSubindex, int & p_value)
	 *
	 *	@brief Get a value from an object dictionary sub entry (int)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_indexSubindex Combined index and subindex of the object dictionary sub entry
	 *	@param p_value Reference to an integer variable. In this the retrieved value will be stored
	 *
	 *	@returns WE_SUCCESS If a value could be retrieved and stored to p_value
	 *	@returns WE_NOT_FOUND_ERROR If Object with p_indexSubindex is not present
	 *
	 *	This function retrieves the value of the object dictionary subentry
	 *	identified by p_indexSubindex. If the value is managed by DEI (index >= 0x2000)
	 *	the current value from DEI will be retrieved from DEI to make sure that the values are in sync.
	 *	For communication entries (index < 0x2000, the value is taken directly from this object dictionary.
	 */
	int getValue(unsigned int p_indexSubindex, int & p_value);

	/**
	 *	@fn int getValue(unsigned short p_index, unsigned char p_subindex, int & p_value);
	 *
	 *	@brief Get a value from an object dictionary sub entry (int)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the object dictionary entry
	 *	@param p_subindex Subindex of the object dictionary sub entry
	 *	@param p_value Reference to an integer variable. In this the retrieved value will be stored
	 *
	 *	@returns WE_SUCCESS If a value could be retrieved and stored to p_value
	 *	@returns WE_NOT_FOUND_ERROR If Object with p_indexSubindex is not present
	 *
	 *	This function retrieves the value of the object dictionary subentry
	 *	identified by p_index and p_subindex. If the value is managed by DEI (index >= 0x2000)
	 *	the current value from DEI will be retrieved from DEI to make sure that the values are in sync.
	 *	For communication entries (index < 0x2000, the value is taken directly from this object dictionary.
	 */
	int getValue(unsigned short p_index, unsigned char p_subindex, int & p_value);

	/**
	 *	@fn int getValue(unsigned int p_indexSubindex, std::string & p_value)
	 *
	 *	@brief Get a value from an object dictionary sub entry (string)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_indexSubindex Combined index and subindex of the object dictionary sub entry
	 *	@param p_value Reference to a string variable. In this the retrieved value will be stored
	 *
	 *	@returns WE_SUCCESS If a value could be retrieved and stored to p_value
	 *	@returns WE_NOT_FOUND_ERROR If Object with p_indexSubindex is not present
	 *
	 *	This function retrieves the value of the object dictionary subentry
	 *	identified by p_indexSubindex. If the value is managed by DEI (index >= 0x2000)
	 *	the current value from DEI will be retrieved from DEI to make sure that the values are in sync.
	 *	For communication entries (index < 0x2000, the value is taken directly from this object dictionary.
	 */
	int getValue(unsigned int p_indexSubindex, std::string & p_value);

	/**
	 *	@fn int getValue(unsigned short p_index, unsigned char p_subindex, std::string & p_value);
	 *
	 *	@brief Get a value from an object dictionary sub entry (string)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the object dictionary entry
	 *	@param p_subindex Subindex of the object dictionary sub entry
	 *	@param p_value Reference to a string variable. In this the retrieved value will be stored
	 *
	 *	@returns WE_SUCCESS If a value could be retrieved and stored to p_value
	 *	@returns WE_NOT_FOUND_ERROR If Object with p_indexSubindex is not present
	 *
	 *	This function retrieves the value of the object dictionary subentry
	 *	identified by p_index and p_subindex. If the value is managed by DEI (index >= 0x2000)
	 *	the current value from DEI will be retrieved from DEI to make sure that the values are in sync.
	 *	For communication entries (index < 0x2000, the value is taken directly from this object dictionary.
	 */
	int getValue(unsigned short p_index, unsigned char p_subindex, std::string & p_value);

	/**
	 *	@fn int setValue(unsigned int p_indexSubindex, int p_value, bool p_internal)
	 *
	 *	@brief Set a value to an object dictionary sub entry (int)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_indexSubindex Combined index and subindex identifing the sub entry of the object dictionary
	 *	@param p_value Value to be set to sub entry
	 *	@param p_internal Flag to tell if it is an internal "write request" or if it's triggered from outside (e.g. SDO, PDO)
	 *
	 *	@returns WE_SUCCESS if setting the value was successful
	 *	@returns WE_NOT_FOUND_ERROR If the sub entry identified by p_indexSubindex was not found in this object dictionary
	 *	@returns The return value of WEDEI::setVariableValue / WEDEI::forceSetVariableValue
	 *
	 *	@see int setSDOServerID(unsigned short p_index, unsigned char p_subindex, unsigned int p_serverID)
	 *
	 *	Using this function a value can be set to an object dictionary sub entry. If the sub entry's value
	 *	is managed by DEI, the according DEI function is called from inside (WEDEI::setVariableValue or
	 *	WEDEI::forceSetVariableValue (if the variable is owned by this CANopen instance)).
	 *	The p_internal flag indicates if the "write request" is triggered from inside the CANopen
	 *	stack (e.g. when resetting to default value), this makes it possible to write read-only objects
	 *	from inside the stacks. When using this function due to the reception of PDO or SDO, p_internal
	 *	needs to be false so that all necessary checks can be done.
	 */
	int setValue(unsigned int p_indexSubindex, int p_value, bool p_internal);

	/**
	 *	@fn int setValue(unsigned short p_index, unsigned char p_subindex, int p_value, bool p_internal)
	 *
	 *	@brief Set a value to an object dictionary sub entry (int)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index identifing the object entry in this object dictionary
	 *	@param p_subindex Subindex identifying the sub entry in this object dictionary
	 *	@param p_value Value to be set to sub entry
	 *	@param p_internal Flag to tell if it is an internal "write request" or if it's triggered from outside (e.g. SDO, PDO)
	 *
	 *	@returns WE_SUCCESS if setting the value was successful
	 *	@returns WE_NOT_FOUND_ERROR If the sub entry identified by p_indexSubindex was not found in this object dictionary
	 *	@returns The return value of WEDEI::setVariableValue / WEDEI::forceSetVariableValue
	 *
	 *	@see int setSDOServerID(unsigned short p_index, unsigned char p_subindex, unsigned int p_serverID)
	 *
	 *	Using this function a value can be set to an object dictionary sub entry. If the sub entry's value
	 *	is managed by DEI, the according DEI function is called from inside (WEDEI::setVariableValue or
	 *	WEDEI::forceSetVariableValue (if the variable is owned by this CANopen instance)).
	 *	The p_internal flag indicates if the "write request" is triggered from inside the CANopen
	 *	stack (e.g. when resetting to default value), this makes it possible to write read-only objects
	 *	from inside the stacks. When using this function due to the reception of PDO or SDO, p_internal
	 *	needs to be false so that all necessary checks can be done.
	 */
	int setValue(unsigned short p_index, unsigned char p_subindex, int p_value, bool p_internal);

	/**
	 *	@fn int setValue(unsigned short p_index, unsigned char p_subindex, const std::string & p_value, bool p_internal)
	 *
	 *	@brief Set a value to an object dictionary sub entry (string)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index identifing the object entry in this object dictionary
	 *	@param p_subindex Subindex identifying the sub entry in this object dictionary
	 *	@param p_value Value to be set to sub entry
	 *	@param p_internal Flag to tell if it is an internal "write request" or if it's triggered from outside (e.g. SDO, PDO)
	 *
	 *	@returns WE_SUCCESS if setting the value was successful
	 *	@returns WE_NOT_FOUND_ERROR If the sub entry identified by p_indexSubindex was not found in this object dictionary
	 *	@returns The return value of WEDEI::setVariableValue / WEDEI::forceSetVariableValue
	 *
	 *	@see int setSDOServerID(unsigned short p_index, unsigned char p_subindex, unsigned int p_serverID)
	 *
	 *	Using this function a value can be set to an object dictionary sub entry. If the sub entry's value
	 *	is managed by DEI, the according DEI function is called from inside (WEDEI::setVariableValue or
	 *	WEDEI::forceSetVariableValue (if the variable is owned by this CANopen instance)).
	 *	The p_internal flag indicates if the "write request" is triggered from inside the CANopen
	 *	stack (e.g. when resetting to default value), this makes it possible to write read-only objects
	 *	from inside the stacks. When using this function due to the reception of PDO or SDO, p_internal
	 *	needs to be false so that all necessary checks can be done.
	 */
	int setValue(unsigned short p_index, unsigned char p_subindex, const std::string & p_value, bool p_internal);

	/**
	 *	@fn void setDEI(WEDEI * p_dei)
	 *
	 *	@brief Provide a pointer to the WEDEI instance
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_dei Pointer to this processes WEDEI instance
	 */
	void setDEI(WEDEI * p_dei);

	/**
	 *	@fn void setCANopenInstance(WECANopen * p_instance)
	 *
	 *	@brief Set the pointer to the CANopen instance this object dictionary belongs to
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_instance Pointer to the instance of WECANopen to which this object dictionary belongs to
	 */
	void setCANopenInstance(WECANopen * p_instance);

	/**
	 *	@fn int createDataTypeEntries(void)
	 *
	 *	@brief Create object dictionary entries for the CANopen data types
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This function creates the object dictionary entries for the CANopen data types in
	 *	the object dictionary (Index (currently) 0x0001 to 0x001b).
	 */
	int createDataTypeEntries(void);

	/**
	 *	@fn int createPreDefinedConnectionSet(void)
	 *
	 *	@brief Create communication objects in the object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS If creating the entries was successful
	 *	@returns WE_INTERFACE_ERROR if the internal pointer to the CANopen instance is nullptr
	 *
	 *	@see void setCANopenInstance(WECANopen * p_instance)
	 *
	 *	This function creates default objects for the data types part and the communication part
	 *	of the object dictionary. After that a DCF file is created with all entries. You can find that
	 *	in the working directory of the application using this CANopen stack.
	 */
	int createPreDefinedConnectionSet(void);

	/**
	 *	@fn int getDataType(unsigned short p_index, unsigned char p_subindex, WECANopenDataType & p_type)
	 *
	 *	@brief Retrieve the CANopen data type of an sub entry in this object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the sub entry
	 *	@param p_subindex Subindex of the sub entry
	 *	@param p_type Reference to a WECANopenDataType variable in which the retrieved data type will be stored
	 *
	 *	@returns WE_SUCCESS if the data type of the sub entry was successfully written to p_type
	 *	@returns WE_NOT_FOUND_ERROR If the sub entry was not found in this object dictionary
	 *
	 *	@see WECANopenDataType
	 */
	int getDataType(unsigned short p_index, unsigned char p_subindex, WECANopenDataType & p_type);

	/**
	 *	@fn int getNumberOfBytesByDatatType(WECANopenDataType type)
	 *
	 *	@brief Find out, how many bytes a data type does occupy
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_type WECANopenDataType of which you want to know the size in bytes
	 *
	 *	@returns Number of data bytes occupied by the data type identified with p_type.
	 *	@returns -1 for the data types that have no specified fixed size (like VISIBLE_STRING)
	 *
	 *	@see int getNumberOfBitsByDatatType(WECANopenDataType p_type)
	 */
	int getNumberOfBytesByDatatType(WECANopenDataType p_type);

	/**
	 *	@fn int getNumberOfBitsByDatatType(WECANopenDataType p_type)
	 *
	 *	@brief Find out, how many bits a data type does occupy
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_type WECANopenDataType of which you want to know the size in bits
	 *
	 *	@returns Number of bits occupied by the data type identified with p_type.
	 *	@returns -1 for the data types that have no specified fixed size (like VISIBLE_STRING)
	 *
	 *	@see int getNumberOfBytesByDatatType(WECANopenDataType p_type)
	 */
	int getNumberOfBitsByDatatType(WECANopenDataType p_type);

	/**
	 *	@fn bool objectExists(unsigned short p_index)
	 *
	 *	@brief Check if an object exists in the object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the object that shall be checked for it's existence
	 *
	 *	@returns true if object exists
	 *	@returns false if the object does not exist
	 */
	bool objectExists(unsigned short p_index);

	/**
	 *	@fn bool subObjectExists(unsigned short p_index, unsigned char p_subindex)
	 *
	 *	@brief Check if a sub enty exists in the object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the sub entry that shall be checked for it's existence
	 *	@param p_subindex Subindex of the sub entry that shall be checked for it's existence
	 *
	 *	@returns true if sub entry exists
	 *	@returns false if sub entry does not exist
	 */
	bool subObjectExists(unsigned short p_index, unsigned char p_subindex);

	/**
	 *	@fn bool isSigned(unsigned int p_indexSubindex)
	 *
	 *	@brief Check if a sub entry is signed or unsigned
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_entryKey Combined index and subindex of the sub entry for which signedness shall bechecked
	 *
	 *	@returns true if the sub entry has a signed data type
	 *	@returns false if the sub entry has an unsigned data type
	 *
	 *	@see bool subObjectExists(unsigned short p_index, unsigned char p_subindex)
	 *
	 *	@warning Check if object exists before calling this function
	 */
	bool isSigned(unsigned int p_entryKey);
	bool isDataTypeSigned(WECANopenDataType p_dataType);

	/**
	 * @fn int getVariableStorageType(unsigned int p_combinedKey, WEVariableStorageType * p_storageType);
	 *
	 * @brief Reads the storage type of variable with index , subindex of p_combinedKey and sets value of p_storageType.
	 *
	 * @author ju@wachendorff.de
	 * @author Joachim Urbach
	 *
	 * @param p_combinedKey Combinde index and subindex of the sub entry for which WEStroageType shall be checked
	 * @param p_storageType StorageType of variable (STORAGE_TYPE_INTEGER, STORAGE_TYPE_STRING, STORAGE_TYPE_UNDEFINED)
	 *
	 * @return 	WE_SUCCESS 	 if variable can be found in the DEI and storage type could be read.
	 * 		WE_TYPE_ERROR	 if storage type of variable is UNDEFINED;
	 * 		WE_GENERAL_ERROR if variable can not be found in DEI
	 */
	int getVariableStorageType(unsigned int p_combinedKey, WEVariableStorageType * p_storageType);

	/**
	 *	@fn void setSDO(WESDOHandler * p_sdo)
	 *
	 *	@brief Set a pointer to the CANopen stacks SDO handler
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_sdo Pointer to the instance of the CANopen stack's SDO handler
	 */
	void setSDO(WESDOHandler * p_sdo)
	{
		this->sdo = p_sdo;
	}

	/**
	 *	@fn void setPDO(WEPDOHandler * p_pdo)
	 *
	 *	@brief Set a pointer to the CANopen stacks PDO handler
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_pdo Pointer to the instance of the CANopen stack's PDO handler
	 */
	void setPDO(WEPDOHandler * p_pdo)
	{
		this->pdo = p_pdo;
	}

	void setNMT(WENMTHandler * p_nmt)
	{
		this->nmt = p_nmt;
	}

	/**
	 *	@fn int setSDOServerID(unsigned short p_index, unsigned char p_subindex, unsigned int p_serverID)
	 *
	 *	@brief Set an SDO server COB-ID to a sub entry in the object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the sub entry
	 *	@param p_subindex Subindex of the sub entry
	 *	@param p_serverID The SDO server COB-ID to which SDO download requests shall be sent when trying
	 *		to write to the sub entry
	 *
	 *	@returns WE_SUCCESS if setting the SDO server COB-ID was successful
	 *	@returns WE_INTERFACE_ERROR if the sub entry was not found in the object dictionary
	 *		or there is no valid pointer to the SDO handler available
	 *
	 *	@see WESDOHandler
	 *
	 *	When setting a valid SDO server COB-ID to a sub entry of the object dictionary, a call to setValue
	 *	of this sub entry will initiate a SDO Download Request to the defined SDO server. The value will be
	 *	discarded until the SDO server answers with a positive SDO Download response. After that the value
	 *	will really be set. If the server sends an SDO abort for the Download request or a timeout
	 *	occurs, the value will be totally forgotten.
	 */
	int setSDOServerID(unsigned short p_index, unsigned char p_subindex, unsigned int p_serverID);

private:

	friend class WENMTHandler;		//!< NMT handler may access object dictionary directly
	friend class WECANopenStateMachine;	//!< The CANopen state machine may access object dictionary directly
	friend class WECANopen;			//!< The CANopen instance may access object dictionary directly
	friend class WESyncHandler;		//!< Sync handler may access object dictionary directly
	friend class WEEmergencyHandler;	//!< Emergency handler may access object dictionary directly

	ODEntries entries;	//!< This is the map in which ALL object entries of this object dictionary are stored
	WEDEI * dei;		//!< Pointer to the processes DEI instance
	WECANopen * canopenInstance;	//!< Pointer to the CANopen instance of this object dictionary
	WEPDOHandler * pdo;	//!< Pointer to the PDO handler of this CANopen stack
	WESDOHandler * sdo;	//!< Pointer to the SDO handler of this CANopen stack
	WENMTHandler * nmt;	//!< NMT handler

	/**
	 *	@fn const int * getPointerToValue(unsigned short p_index, unsigned char p_subindex)
	 *
	 *	@brief Retrieve the pointer to the int value of an object dictionary sub entry
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the object dictionary sub entry
	 *	@param p_subindex Subindex of the object dictionary sub entry
	 *
	 *	@returns Const pointer to the int value of the object dictionary sub entry.
	 *	@returns nullptr if sub entry with p_index/p_subindex does not exist
	 */
	const int * getPointerToValue(unsigned short p_index, unsigned char p_subindex);

	/**
	 *	@fn ObjectDictionarySubEntry * getSubentry(unsigned short p_index, unsigned char p_subindex)
	 *
	 *	@brief Get a pointer to a sub entry in this object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the sub entry
	 *	@param p_subindex Sub index of the sub entry
	 *
	 *	@returns Pointer to the sub entry if it exists
	 *	@returns nullptr if the sub entry does not exist
	 *
	 *	@see ObjectDictionaryEntry * getEntry(unsigned short p_index)
	 */
	ObjectDictionarySubEntry * getSubentry(unsigned short p_index, unsigned char p_subindex);

	/**
	 *	@fn ObjectDictionaryEntry * getEntry(unsigned short p_index)
	 *
	 *	@brief Get a pointer to a top-level entry in this object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the object entry
	 *
	 *	@returns Pointer to the object entry if it exists
	 *	@returns nullptr if the object entry does not exist in this object dictionary
	 *
	 *	@see ObjectDictionarySubEntry * getSubentry(unsigned short p_index, unsigned char p_subindex)
	 */
	ObjectDictionaryEntry * getEntry(unsigned short p_index);

	/**
	 *	@fn int handleValueChangeRequest(unsigned short p_index,
			ObjectDictionarySubEntry * p_subentry,
			int & p_value,
			bool p_internal)
	 *
	 *	@brief Function to check if a value that shall be set to a sub entry is valid or not
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_entry Pointer to the top level object entry object to which the sub entry belongs
	 *	@param p_subentry Pointer to the sub entry which's value shall be changed
	 *	@param p_value The value that shall be set to the sub entry
	 *	@param p_internal Flag that tells if it is an internal setValue request or if it was triggered from outside
	 *
	 *	@returns WE_SUCCESS if p_value has a valid value for the sub entry
	 *	@returns Other error code if the value is not valid. If the error code is > WE_MAX_ERROR, the error code is an SDO Abort code
	 *
	 *	@see enum CANopenSDOAbortCodes
	 *
	 *	This function is used to check if values for the communication parameter object dictionary entries are valid or not
	 */
	int handleValueChangeRequest(unsigned short p_index,
			ObjectDictionarySubEntry * p_subentry,
			int & p_value,
			bool p_internal);

	/**
	 * Used by handleValueChangeRequest() if SDO server parameter shall be changed
	 */
	int handleSDOServerParameterValueChangeRequest(unsigned short p_index, ObjectDictionarySubEntry * p_subentry, int & p_value, bool p_internal);

	/**
	 * Used by handleValueChangeRequest() if SDO client parameter shall be changed
	 */
	int handleSDOClientParameterValueChangeRequest(ObjectDictionarySubEntry * p_subentry, int & p_value);

	/**
	 * Used by handleValueChangeRequest() if PDO parameter shall be changed
	 */
	int handlePDOParameterValueChangeRequest(unsigned short p_index, ObjectDictionarySubEntry * p_subentry, int & p_value, bool p_internal);

	/**
	 *	@fn int handleValueChangeRequest(unsigned short p_index,
			ObjectDictionarySubEntry * p_subentry,
			const std::string & p_value,
			bool p_internal)
	 *
	 *	@brief Function to check if a value that shall be set to a sub entry is valid or not
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_entry Pointer to the top level object entry object to which the sub entry belongs
	 *	@param p_subentry Pointer to the sub entry which's value shall be changed
	 *	@param p_value The value that shall be set to the sub entry
	 *	@param p_internal Flag that tells if it is an internal setValue request or if it was triggered from outside
	 *
	 *	@returns WE_SUCCESS if p_value has a valid value for the sub entry
	 *	@returns Other error code if the value is not valid. If the error code is > WE_MAX_ERROR, the error code is an SDO Abort code
	 *
	 *	@see enum CANopenSDOAbortCodes
	 *
	 *	This function is used to check if values for the communication parameter object dictionary entries are valid or not
	 */
	int handleValueChangeRequest(unsigned short p_index,
			ObjectDictionarySubEntry * p_subentry,
			const std::string & p_value,
			bool p_internal);

	/**
	 *	@fn int storeParameters(const std::string & p_filePath)
	 *
	 *	@brief Store the complete object dictionary in DCF format
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_filePath Path to the file in which the DCF shall be stored. If p_filePath == cout, the DCF will be printed to stdout
	 *
	 *	@returns WE_SUCCESS if everything was successful
	 *	@returns WE_FILE_ERROR if p_filePath points to file that can't be written to
	 *
	 *	Currently this function creates a DCF file with the data from the object dictionary.
	 *	Later it shall also serialize the object dictionary so it can be loaded again later.
	 */
	int storeParameters(const std::string & p_filePath);

	/**
	 *	@fn std::string getAccessTypeString(WECANopenAccessType p_type)
	 *
	 *	@brief Retrieve the name of a CANopen access type (used only for DCF creation)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_type CANopen data type
	 *
	 *	@returns String describing the data type
	 */
	static std::string getAccessTypeString(WECANopenAccessType p_type);

	/**
	 *	@fn int setLimits(unsigned short p_index, unsigned char p_subindex, int p_lowLimit, int p_highLimit)
	 *
	 *	@brief Setting high and low limit of an object dictionary sub entry
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the sub entry
	 *	@param p_subindex Subindex of the sub entry
	 *	@param p_lowLimit Low limit for the sub entry
	 *	@param p_highLimit High limit for the sub entry
	 *
	 *	@returns WE_SUCCESS if setting the limits was successful
	 *	@returns WE_INVALID_PARAMETER_ERROR If the subentry was not found
	 */
	int setLimits(unsigned short p_index, unsigned char p_subindex, int p_lowLimit, int p_highLimit);

	/**
	 *	@fn int performReset(bool p_complete)
	 *
	 *	@brief Reset the object dictionary (complete or only comm entries) to the default values
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_complete if this is true, the complete object dictionary will be set to the default values,
	 *		else only the Communication aprameters will be reset
	 *
	 *	@returns WE_SUCCESS if the reset was successfull
	 *
	 *	Some work needs to be done in this function since ALL entries will be asked to get default values again
	 *	even entries that are owned by other protocols or internal variables. I don't think that this is
	 *	the intention, is it?
	 */
	int performReset(bool p_complete);

	void resetValue(unsigned int p_key);
};
}

#endif /* WE_CAN_OPEN_OBJECT_DICTIONARY_H_ */
