/**
 *	@file we_sdo_handler.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of the SDO handler class
 *
 */

#ifndef WE_SDO_HANDLER_H_
#define WE_SDO_HANDLER_H_

#include <deque>

#include "../../../common/include/we_error.h"
#include "we_can_open_definitions.h"
#include "we_can_open_message_handler.h"

namespace WE
{

class WECANRaw;
class WECANopenObjectDictionary;
class WECANopenStateMachine;
class WETick;
class WECANopen;
class WESignalListener;
class WEDEI;

/**
 *	@struct SDODownloadResponseStorage
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Information needed to identify a pending SDO download response
 *
 *	@namespace WE
 */
struct SDODownloadResponseStorage
{
	unsigned int serverID{0};	//!< The SDO server ID with which the download request was sent
	unsigned int clientID{0};	//!< The SDO client ID that the response will be addressed to
	int intValue{0};		//!< The pending value of the request (int)
	std::string strValue{};	//!< The pending value of the request (string)
	unsigned int timeout{0};	//!< The tick count when the SDO download request times out
	unsigned int internalKey{0};
};

struct SDODownloadRequestParameter
{
	SDODownloadRequestParameter() = default;
	SDODownloadRequestParameter(unsigned int p_varKey,
			unsigned int p_serverID,
			unsigned int p_clientID,
			int p_value,
			const std::string & p_stringValue,
			WECANopenDataType p_type)
	:varKey(p_varKey),
	 serverID(p_serverID),
	 clientID(p_clientID),
	 value(p_value),
	 stringValue(p_stringValue),
	 type(p_type)
	{
		//nothing more to do here
	}
	unsigned int varKey{std::numeric_limits<unsigned int>::max()};
	unsigned int serverID{0};
	unsigned int clientID{0};
	int value{-1};
	std::string stringValue{""};
	WECANopenDataType type{CANOPEN_DATA_TYPE_UNDEFINED};
};

struct SDOSegmentedTransferStorage
{
	unsigned int serverID{0};
	unsigned int clientID{0};
	std::string strValue{};
	int bytesAlreadySent{0};
	int bytesToSend{0};
	int lastToggleValue{0};
	unsigned short index{0};
	unsigned char subindex{0};
	bool waitForLastResponse{false};
	unsigned int lastTick{0};
	bool iAmClient{false};

	SDOSegmentedTransferStorage() = default;
};

/**
 *	@class WESDOHandler
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief The message handler responsible for SDO messages
 *
 *	@namespace WE
 *
 *	The WESDOHandler class is responsible for handling all SDO specific CAN messages and behavior.
 *	It supports SDO server and SDO client functionality.
 */
class WESDOHandler : public WECANopenMessageHandler
{
public:
	/**
	 *	@fn WESDOHandler()
	 *
	 *	@brief Constructor. Calls WECANopenMessageHandler() and initializes tick
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WESDOHandler();

	/**
	 *	@fn virtual ~WESDOHandler()
	 *
	 *	@brief Clears the containers to free memory
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WESDOHandler();

	/**
	 *	@fn int handleMessage(WECANMessage * p_msg)
	 *
	 *	@brief Check for SDO message and handle the request/response/abort
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_msg pointer to a CAN message that might be a SDO message
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR if the message does not caontain SDO data
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_msg is nullptr
	 *	@returns WE_SUCCESS if the message was identified as a SDO message for this stack
	 *
	 *	This message is called by the CANopen state machine whenever a CAN message is received that was not
	 *	handled by higher priority message handlers (like NMT).
	 *	It checks if the message contains SDO requests/responses/aborts that are addressed to this stack.
	 *	If correct SDO data is detected, it processes the containing request/response.
	 */
	int handleMessage(WECANMessage * p_msg) override;

	/**
	 *	@fn int init(void)
	 *
	 *	@brief Initialize internal data with data from CANopen object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_INTERFACE_ERROR if the object dictionary can't be accessed
	 *	@returns WE_SUCCESS if initialization was successful
	 *
	 *	The init function reads SDO client and server settings from the CANopen object dictionary to configure
	 *	the internal members so that SDO messages for this stack can be detected / SDO requests can be sent.
	 */
	int init(void);

	/**
	 *	@fn int handleUpdateRequest(unsigned int p_varKey, int p_intValue)
	 *
	 *	@brief Handle a request to change a value of a variable owned by this CANopen stack
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_varKey Index and subindex of a variable / object dictionary sub entry
	 *	@param p_intValue The integer value that shall be set
	 *
	 *	@returns WE_NOT_FOUND_ERROR If the variable identified by p_varKey is not configured for SDO download requests (no server ID set)
	 *	@returns the returnValue of int WESDOHandler::sendSDODownloadRequest(unsigned int p_varKey, unsigned int p_serverID, unsigned int p_clientID, int p_value, bool p_expedited)
	 *
	 *	@see int WESDOHandler::sendSDODownloadRequest(unsigned int p_varKey, unsigned int p_serverID, unsigned int p_clientID, int p_value, bool p_expedited)
	 *
	 *	This function gets called every time a variable owned by this CANopen stack shall get a new value.
	 *	The handler will check if a SDO server COB-ID is configured for the matching object dictionary entry.
	 *	If an SDO server ID is configured, a SDO download request is sent to the server. If the server does not
	 *	answer within a timeout or sends an abort, the value will be discarded. If a positive response is
	 *	received, the value will be set to the object dictionary / DEI.
	 */
	int handleUpdateRequest(unsigned int p_varKey, int p_intValue);
	int handleUpdateRequest(unsigned int p_varKey, const std::string & p_stringValue);
	//JU
	/**
	 *	@fn int handleGetValueRequest(unsigned int p_varKey)
	 *
	 *	@brief initiates a request to get a value of a foreign stack
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *  @param p_varKey Index and subindex of a variable / object dictionary sub entry owned by our Stack
	 *
	 *  @returns WE_NOT_FOUND_ERROR If the variable identified by p_varKey is not configured for SDO download requests (no server ID set)
	 *	@returns the returnValue of sendSDOUploadRequest(p_varKey, serverID, clientID)
	 *
	 *	Initiates a request to read a Object Dictionary Subentry value of a foreign stack.
	 *	Calls function sendSDOUploadRequest()
	 */
	int handleGetValueRequest(unsigned int p_varKey);

	// JU
	/**
	 *	@fn int sendSDOUploadRequest(unsigned int p_varKey, unsigned int p_serverID, unsigned int p_clientID)
	 *
	 *	@brief send a upload request to a foreign stack
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *  @param p_varKey Index and subindex of a variable / object dictionary sub entry owned by our Stack
	 *  @param p_serverID COB-ID of the server from which we want to read a object dictionary subentry value
	 *
	 *  @returns WE_INIT_ERROR If canPort or our object dictionary aren't initialized
	 *	@returns the returnValue of this->sendCANMessage(this->msg)
	 *
	 *  This function sends a SDO upload request to the commited p_serverID. The index of the foreign
	 *  object dictionary subentry is stored in the external identifiert of the variable(p_varKey).
	 *  The index and subindex of the variable(p_varkey) will be stored in the storage "currentSDOUpload".
	 *  So later we can write the received value to variable with forceSetVariableValue().
	 *
	 *  @see handleSegmentedSDOUploadSegment()
	 *
	 *  @ warning:
	 */
	int sendSDOUploadRequest(unsigned int p_varKey, unsigned int p_serverID, unsigned int p_clientID);

	/**
	 *	@fn int addVariableSDOServerMapping(unsigned short p_index, unsigned char p_subindex, unsigned int p_serverID)
	 *
	 *	@brief Assign a SDO server ID to a variable / object dictionary entry
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the object dictionary subentry
	 *	@param p_subindex Subindex of the object dictionary subentry
	 *	@param p_serverID COB-Id of the SDO server
	 *
	 *	@returns WE_SUCCESS
	 */
	int addVariableSDOServerMapping(unsigned short p_index, unsigned char p_subindex, unsigned int p_serverID);

	/**
	 *	@fn void checkSDOTimeouts(unsigned int* p_timeLeft = nullptr)
	 *
	 *	@brief Checks if a SDO download request timed out and sends abort message
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	void checkSDOTimeouts(unsigned int* p_timeLeft = nullptr);

	void setStateMachine(WECANopenStateMachine * p_sm)
	{
		this->sm = p_sm;
	}



	int setEventListener(WESignalListener * p_listener)
	{
		this->eventListener = p_listener;
		return WE_SUCCESS;
	}

	/**
	 * If someone wants to send a SDO downlaod request via WECANopen's updateRequest() function
	 * these requests are put in a map and can later be sent out via this function.
	 * (thread synchronization)
	 */
	void sendQueuedDownloadRequests(void);

	/**
	 * Any SDO Upload Request is saved in a map and will only be sent in the "main loop" of the CANopen
	 * stack with this function. Only one queued SDO Upload Request is sent and then we wait for a response (100ms timeout).
	 * After response was received or after timeout, the next request is sent.
	 */
	void sendNextQueuedUploadRequest(void);

private:

	std::map<unsigned int, unsigned int> serverIDMapping; //!< Mapping of my server IDs and the corresponding client IDs
	std::map<unsigned int, unsigned int> clientIDMapping; //!< Mapping of my client IDs and the corresponding server IDs
	std::map<unsigned int, unsigned int> clientIDMappingReverse;
	std::map<unsigned int, unsigned int> variableToServerID; //!< Mapping of variable key with corresponding SDO server COB-ID
	std::map<unsigned int, SDODownloadResponseStorage> downloadResponseStorage; //!< Mapping between variable key and the download request data
	std::map<unsigned int, SDODownloadRequestParameter> downloadRequestStorage; //!< Mapping between variable key and the download request data. See sendSDODownloadRequest()
	std::deque<SDODownloadRequestParameter> uploadRequestStorage; //!< Mapping between variable key and the upload request data.

	WECANopenStateMachine * sm; //!< Pointer to the CANopen state machine

	bool segemntedUploadInProgress;
	bool segemntedDownloadInProgress;

	/**
	 *	@fn int sendUploadResponse(unsigned short p_index, unsigned char p_subindex, unsigned int p_clientID, unsigned int p_serverID)
	 *
	 *	@brief Send a response to a SDO upload request
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the requested object dictionary subentry
	 *	@param p_subindex Subindex of the requested object dictionary subentry
	 *	@param p_clientID the COB-ID of the requesting SDO client
	 *
	 *	@returns WE_INTERFACE_ERROR if the object dictionary is not accessible
	 *	@returns return value of sendSDOAbort(p_index, p_subindex, p_clientID, SDO_ABORT_CODE_SUBINDEX_DOES_NOT_EXIST)
	 *		if the subentry was not found in the object dictionary
	 *	@returns WE_SUCCESS if the upload response could be sent successfully
	 *	@returns return value of this->sendSDOAbort(p_index, p_subindex, p_clientID, SDO_ABORT_CODE_DATATYPE_SERVICE_PARAMETER)
	 *		if the length of the subentry is more than 4 byte.
	 *
	 *	@warning Can't be used for an answer to segmented or block requests
	 *
	 *	This function is used to send a response for an SDO upload request. Currently it is only called
	 *	for expedited transfer. If segmented/block transfer will be supported later, this function needs
	 *	to be reworked.
	 */
	int sendUploadResponse(unsigned short p_index, unsigned char p_subindex, unsigned int p_clientID, unsigned int p_serverID);

	/**
	 *	@fn int sendSDOAbort(unsigned short p_index, unsigned char p_subindex, unsigned int p_clientID, unsigned int p_abortCode)
	 *
	 *	@brief Send a SDO abort message
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Object dictionary index related to the abort
	 *	@param p_subindex Object dictionary subindex related to the SDO abort
	 *	@param p_clientID The COB-ID with which the abort shall be sent
	 *	@param p_abortCode the SDO abort code of the SDO abort message
	 *
	 *	@returns WE_SUCCESS if the message was sent successfully
	 *	@returns WE_INTERFACE_ERROR if the CAN port was not accessible
	 *
	 *	@see enum CANopenSDOAbortCodes
	 *
	 *	This function is used to send SDO abort messages.
	 */
	int sendSDOAbort(unsigned short p_index, unsigned char p_subindex, unsigned int p_clientID, unsigned int p_abortCode);

	/**
	 *	@fn int handleExpeditedSDODownloadRequest(unsigned short p_index,
			unsigned char p_subindex,
			bool p_sizeIndicated,
			unsigned char p_size,
			unsigned int p_clientID,
			unsigned char p_byte5,
			unsigned char p_byte6,
			unsigned char p_byte7,
			unsigned char p_byte8 )
	 *
	 *	@brief Handle an incomming SDO download request
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the object dictionary subentry
	 *	@param p_subindex Subindex of the object dictionary subentry
	 *	@param p_sizeIndicated flag that tells if the object size was indicated in the download request
	 *	@param p_size Indicated object size in byte
	 *	@param p_clientID COB-Id of the SDO client that shall be used for sending the SDO download response
	 *	@param p_byte5 1st byte containing the value
	 *	@param p_byte5 2nd byte containing the value
	 *	@param p_byte5 3rd byte containing the value
	 *	@param p_byte5 4th byte containing the value
	 *
	 *	@returns WE_NOT_FOUND_ERROR if the subentry with p_index/p_subindex was not found in the object dictionary
	 *	@returns the SDO abort code if setting the value in the object dictionary was not successful
	 *	@returns WE_SUCCESS if setting the value in the object dictionary was successful and the SDO download response could be sent
	 *
	 *	This message is used to handle an expedited SDO download request. It extracts the value from the data
	 *	bytes of the SDO message and tries to set it to the object dictionary. If everything was successful,
	 *	a positive SDO download response is sent. If not, a SDO abort message is sent.
	 */
	int handleExpeditedSDODownloadRequest(unsigned short p_index,
			unsigned char p_subindex,
			bool p_sizeIndicated,
			unsigned char p_size,
			unsigned int p_clientID,
			unsigned char p_byte5,
			unsigned char p_byte6,
			unsigned char p_byte7,
			unsigned char p_byte8 );

	int handleSegmentedSDODownloadRequest(unsigned short p_index,
			unsigned char p_subindex,
			bool p_sizeIndicated,
			unsigned char p_size,
			unsigned int p_clientID,
			unsigned int p_serverID,
			unsigned char p_byte5,
			unsigned char p_byte6,
			unsigned char p_byte7,
			unsigned char p_byte8 );

	int handleSegmentedSDODownloadSegment(unsigned int p_clientID,
			unsigned int p_serverID,
			unsigned char p_byte1,
			unsigned char p_byte2,
			unsigned char p_byte3,
			unsigned char p_byte4,
			unsigned char p_byte5,
			unsigned char p_byte6,
			unsigned char p_byte7,
			unsigned char p_byte8 );

	/**
	 *	@fn int handleSDODownloadResponse(unsigned short p_index, unsigned char p_subindex, unsigned int p_serverID)
	 *
	 *	@brief Handle a SDO download response message
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the object dictionary subentry
	 *	@param p_subindex Subindex of the object dictionary subentry
	 *	@param p_serverID the COB-ID of the SDO server that did send the response
	 *
	 *	@returns WE_NOT_FOUND_ERROR If there is no pending request for the subentry identified by p_index/p_subindex
	 *	@returns WE_INVALID_RESPONSE_ERROR If we are waiting for a response but from another SDO server (that should not be possible in theory)
	 *	@returns WE_INTERFACE_ERROR if the object dictionary can't be accessed
	 *	@returns WE_TYPE_ERROR If the data type of the subentry can't be detrmined
	 *	@returns the returnValue of WECANOpenObjectDictionary::setValue
	 *
	 *	This function is called if a SDO download response message is received. It checks if we
	 *	are waiting for a download response for that subentry from that server. If yes, the proposed value
	 *	of the request is set to the subentry in the object dictionary. If not, the response is ignored
	 *	and an error code is returned.
	 */
	int handleSDODownloadResponse(unsigned short p_index, unsigned char p_subindex, unsigned int p_serverID);

	int handleSegmentedSDODownloadSegmentResponse(unsigned int p_serverID, unsigned int p_clientID, unsigned char p_byte1, bool p_firstResponse = false);

	void handleSDOAbortMessage(unsigned short p_index, unsigned char p_subindex, unsigned int p_serverID);

	int sendInitiateSegmentedUploadResponse(unsigned short p_index, unsigned char p_subindex, unsigned int p_clientID, unsigned int p_serverID, const std::string & p_value);
	void handleSDOUploadSegmentRequest(unsigned int p_clientID, unsigned short p_serverID, int p_toggle);

	int updateValueAccordingToDataType(unsigned short p_index, unsigned char p_subindex, int &p_value);

	// JU
	/**
	 *	@fn int handleSDOUploadResponse(unsigned int p_serverID,
			unsigned int p_clientID,
			unsigned int p_index,
			unsigned int p_subIndex,
			unsigned char p_byte_0,
			unsigned char p_byte_4,
			unsigned char p_byte_5,
			unsigned char p_byte_6,
			unsigned char p_byte_7 )
	 *
	 *	@brief Handle the SDO init upload response message
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *  @param p p_serverID the COB-ID of the SDO server that did send the response
	 *  @param p_clientID the COB-ID
	 *	@param p_index Index of the object dictionary subentry
	 *	@param p_subindex Subindex of the object dictionary subentry
	 *	@param p_byte0 CMD byte
	 *	@param p_byte4 data byte 1
	 *	@param p_byte5 data byte 2
	 *	@param p_byte6 data byte 3
	 *	@param p_byte7 data byte 4
	 *
	 *	@returns WE_RECEIVE_ERROR If we can't handle the message
	 *	@returns the returnValue of sendUploadSDOSegmentRequest(p_clientID)
	 *	@returns the returnValue of handleExpeditedSDOUploadResponse(p_index, p_subIndex, p_byte_4, p_byte_5, p_byte_6, p_byte_7, n)
	 *
	 *	This function is called if a SDO upload response message is received. It checks if it is a
	 *	segmented or expedited response and deligates it to the next handle-function.
	 *  If it is a segmented response the value segemntedUploadInProgress is set to TRUE to get
	 *  the next segmented message.
	 */
	int handleSDOUploadResponse(unsigned int p_serverID,
			unsigned int p_clientID,
			unsigned int p_index,
			unsigned int p_subIndex,
			unsigned char p_byte0,
			unsigned char p_byte4,
			unsigned char p_byte5,
			unsigned char p_byte6,
			unsigned char p_byte7 );

	// JU
	/**
	 *	@fn int sendUploadSDOSegmentRequest(unsigned int p_clientID)
	 *
	 *	@brief sends a SDO segmented upload request message to the server
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *  @param p_clientID the COB-ID of the SDO server which we want to send the message
	 *
	 *	@returns the returnValue of sendCANMessage(this->msg) respectively WE_SEND_ERROR
	 *
	 *	This function will called from handleSDOResponse() or handleSegmentedSDOUploadSegment()
	 *	to send a request to the server to get the next segmeted SDO upload message.
	 */
	int sendUploadSDOSegmentRequest(unsigned int p_clientID);

	// JU
	/**
	 *	@fn int handleSegmentedSDOUploadSegment(unsigned int p_serverID,
			unsigned int p_clientID,
			unsigned char p_byte1,
			unsigned char p_byte2,
			unsigned char p_byte3,
			unsigned char p_byte4,
			unsigned char p_byte5,
			unsigned char p_byte6,
			unsigned char p_byte7,
			unsigned char p_byte8 )
	 *
	 *	@brief handle the received SDO segmented upload message
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *  @param p_clientID the COB-ID of the SDO server which we want to send the message
	 *  @param p_serverID the COB-ID of the SDO client from which we the sdo segmented upload messages
	 *  @param p_byte1-7 transmitted data
	 *
	 *  @returns WE_RECEIVE_ERROR if the order of the messages is wrong
	 *  @returns WE_PERMISSION_ERROR if want to write to a variable which is not owned by our Stack
	 *	@returns the returnValue of sendUploadSDOSegmentRequest(p_clientID)
	 *	@returns the returnValue of dei->forceSetVariableValue()
	 *
	 *	This function checkes if we received the expected message from the server. If the message is
	 *	ok and we have to receive more messages, we send the next SDO segmented upload request to the server.
	 *	If we received the last message the received (string value) is stored at the variable which initiate
	 *	the upload process. The index and subindex of this variable is stored in "currentSDOUpload"
	 */
	int handleSegmentedSDOUploadSegment(unsigned int p_serverID,
			unsigned int p_clientID,
			unsigned char p_byte1,
			unsigned char p_byte2,
			unsigned char p_byte3,
			unsigned char p_byte4,
			unsigned char p_byte5,
			unsigned char p_byte6,
			unsigned char p_byte7,
			unsigned char p_byte8 );

	int handleExpeditedSDOUploadResponse(unsigned int p_index,
			unsigned int p_subindex,
			unsigned char p_byte4,
			unsigned char p_byte5,
			unsigned char p_byte6,
			unsigned char p_byte7,
			int p_numberOfBytesWithNoData);

	/**
	 *	@fn int sendSDODownloadRequest(unsigned int p_varKey, unsigned int p_serverID, unsigned int p_clientID, int p_value, const std::string & p_stringValue, WECANopenDataType = CANOPEN_DATA_TYPE_UNDEFINED)
	 *
	 *	@brief Send a SDO download request
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_varKey Combined index/subindex of an object dictionary subentry
	 *	@param p_serverID The COB-ID of the SDO server (we wait for a response with that COB-ID)
	 *	@param p_clientId The SDO client COB-ID (the COB-ID used in this request message)
	 *	@param p_value The value that shall be downloaded to the server
	 *	@param p_expeditet Flag that indicates if this shall be a expedited transfer
	 *
	 *	@returns WE_INTERFACE_ERROR if the object dictionary or the CAN port are not accessible
	 *	@returns WE_NOT_FOUND_ERROR if the subentry with index/subindex (p_varKey) was not found in the object dictionary
	 *	@returns WE_SUCCESS if sending the request was successful
	 *
	 *	A download request is sent whenever a variable associated with a SDO-Server COB-ID shall be changed.
	 *	If the request was sent successfully, a SDODownloadResponseStorage is stored so that the handler knows
	 *	that it has to wait for a SDO download response. The request may time out or the server can answer with
	 *	a SDO abort message. If the server answers with a positive SDO download response, the value is passed
	 *	to the object dictionary, else it is discarded.
	 *	Note that the download request is not sent directly but it is queued so that it is guaranteed that the
	 *	CANopen thread sends it out and we don't run into thread synchronization issues.
	 */
	int sendSDODownloadRequest(unsigned int p_varKey, unsigned int p_serverID, unsigned int p_clientID, int p_value, const std::string & p_stringValue, WECANopenDataType = CANOPEN_DATA_TYPE_UNDEFINED);

	WEDEI * dei;
	WESignalListener * eventListener;

	SDOSegmentedTransferStorage currentSDOUpload;
	SDOSegmentedTransferStorage currentSDODownload;

	std::mutex sdoDownloadResponseLocker;
	std::mutex sdoUploadRequestLocker;
};
}

#endif /* WE_SDO_HANDLER_H_ */
