/**
 *	@file we_sync_handler.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of the WESyncHandler class
 *
 */

#ifndef WE_SYNC_HANDLER_H_
#define WE_SYNC_HANDLER_H_

#include <map>
#include "we_can_open_message_handler.h"

namespace WE
{

//forward declarations
class WEPDO;
class WEPDOHandler;
class WECANopenStateMachine;
class WECANopen;

/**
 *	@class WESyncHandler
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Message handler responsible for everything related to synchronous RX/TX messages
 *
 *	@namespace WE
 *
 *	The main responsibility of this class is to synchronize PDO transfers.
 *	It is used to queue synchronous TxPDOs and to send them ate the correct arrival of the sync message.
 *	It also stores synchronous RxPDOs to pass the values to the object dictionary at the correct time.
 */
class WESyncHandler  : public WECANopenMessageHandler
{
public:
	/**
	 *	@fn WESyncHandler()
	 *
	 *	@brief Constructor. Implicitly calls constructor of WECANopenMessageHandler and initializes members with nullptr and the syncCounter to 0
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	explicit WESyncHandler(WECANopen * p_canopenInstance);

	/**
	 *	@fn virtual ~WESyncHandler()
	 *
	 *	@brief Destructor. Empties the containers.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WESyncHandler();

	/**
	 *	@fn int init(void)
	 *
	 *	@brief Initialize the sync handler
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_INIT_ERROR if the pointer to the object dictionary was not set before or the pointer
	 *		to the Sync-Object COB-ID could NOT be retrieved from the object dictionary
	 *	@returns WE_SUCCESS if the pointer to the Sync-Object COB-ID could be retrieved from the object dictionary
	 *
	 *	This function only fetches the pointer to the value of subentry 0x1005/0x00 from the object
	 *	dictionary so that a fast access to that value is guaranteed and also changes for the COB-ID will
	 *	be noticed immediately
	 */
	int init(void);

	/**
	 *	@fn int handleMessage(WECANMessage * p_msg)
	 *
	 *	@brief Handle a CAN message that could be a sync message
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_msg CAN message that shall be handled
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR if p_msg is not a valid sync message
	 *	@returns WE_SUCCESS if p_msg is the sync message and everything synchronous was done successfully
	 *
	 *	This function is called by the CANopen state machine whenever a CAN message is received and was
	 *	not ahndled by higher priority message handler.
	 *	It checks if it is a valid sync message and then passes the synchronous pending RxPDO values to the
	 *	object dictionary and sends the synchronous TxPDOs (if the transmission type matches the sync counter)
	 */
	int handleMessage(WECANMessage * p_msg) override;

	/**
	 *	@fn int addReceivePDO(WEPDO * p_pdo, WECANMessage * p_msg)
	 *
	 *	@brief Add a synchronous receive PDO to a container so that the value can be passed to the object dictionary with next sync
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_pdo The received synchronous RxPDO
	 *	@param p_msg Pointer to the corresponding CAN message with all PDO data
	 *
	 *	@returns WE_SUCCESS if adding the PDO was successful
	 *	@returns WE_INVALID_PARAMETER_ERROR if either p_pdo or p_msg is nullptr
	 *
	 *	This function just adds a received synchronous RxPDO to a map together with a copy of the corresponding
	 *	received CAN message. If the same PDO is received multiple times, the entry in the map gets overwritten.
	 *	At occurrence of a sync message, all PDOs in this map will be processed and the values set to the object
	 *	dictionary. This can help to reduce CPU load since the user can set a higher transmission mode for
	 *	slower value processing.
	 */
	int addReceivePDO(WEPDO * p_pdo, WECANMessage * p_msg);

	/**
	 *	@fn int addTxPDO(WEPDO * p_pdo)
	 *
	 *	@brief Add a synchronous TxPDO
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_pdoe Pointer to a cyclic synchronous PDO that shall be added
	 *
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_pdo is nullptr
	 *	@returns WE_SUCCESS if the
	 *
	 *	Acyclic synchronous PDOs will be added to a queue that will be emptied at the next reception of a
	 *	sync message. Cyclic synchronous PDOs will be added to a container that will be checked at every
	 *	reception of a sync message. If the configured number of sync messages for that PDO arrived, the PDO
	 *	will be transmitted.
	 */
	int addTxPDO(WEPDO * p_pdo);

	/**
	 *	@fn void setPDOHandler(WEPDOHandler * p_pdo)
	 *
	 *	@brief Set pointer to the PDO handler
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_pdo Pointer to the WEPDOHandler of this stack
	 */
	void setPDOHandler(WEPDOHandler * p_pdo)
	{
		this->pdo = p_pdo;
	}

	/**
	 *	@fn void setStateMachine(WECANopenStateMachine * p_sm)
	 *
	 *	@brief Set a pointer to the CANopen state machine
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_sm Pointer to the state machine
	 */
	void setStateMachine(WECANopenStateMachine * p_sm)
	{
		this->sm = p_sm;
	}

	void removeTxPDO(unsigned short p_index);

	void reset(void);

private:
	std::map<WEPDO*, WECANMessage> syncRXPDOs;	//!< Map containing a pointer to all received synchronous RxPDOs
	std::set<WEPDO*> syncTXPDOs;	//!< All cyclic synchronous TxPDOs
	std::set<WEPDO*> txPDOsForNextTransmit;	//!< Cyclic and acyclic TxPDOs that shall be sent at the next reception of sync
	WEPDOHandler * pdo;	//!< Pointer to the PDO handler
	WECANopenStateMachine * sm;	//!< Pointer to the CANopen state machine
	WECANopen * canopenInstance;
	const int * syncCOBID;	//!< Pointer to the value of the COB-ID of the sync object
	unsigned char syncCounter;	//!< Counter counting sync messages, will be reset after reaching 251

	/**
	 *	@fn void passPDOValuesToApplication(void)
	 *
	 *	@brief Pass the values of allsynchronous RxPDOs to the object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	This function just iterates through all PDOs that were received since last reception of the sync message
	 *	and starts extraction of the values of the CAN messages associated to the PDOs. The values will then be
	 *	passed to the object dictionary.
	 */
	void passPDOValuesToApplication(void);
};
}

#endif /* WE_SYNC_HANDLER_H_ */
