/**
 *	@file we_can_ecu.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief This file provides the basic information of an ECU.
 *
 */

#ifndef WE_CAN_ECU_H_
#define WE_CAN_ECU_H_

#include <string>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>

namespace WE
{

/**
 *	@enum ecu_state
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief Enum describing possible states of an ECU
 *
 *	@namespace WE
 */
enum ecu_state
{
	CAN_ECU_STATE_NEVER_SEEN,	//!< ECU was never seen on bus.
	CAN_ECU_STATE_ACTIVE,		//!< ECU is active on bus.
	CAN_ECU_STATE_OFFLINE,		//!< ECU was seen, but is not communication (e.g. did not get an address).
	CAN_ECU_STATE_TIMEOUT		//!< ECU did not send messages since a specified time.
};

/**
 *	@struct WECanEcu
 *
 *	@brief Struct that defines basic ECU properties.
 *
 */
class WECanEcu
{
public:
	WECanEcu() = default;

	virtual ~WECanEcu() = default;

	//seems like we neither need to copy nor assign an ECU
	WECanEcu(const WECanEcu &p_ecu) = delete;
	WECanEcu & operator=(const WECanEcu &p_ecu) = delete;

	void reset(void)
	{
		this->ID = 0;
		this->name.clear();
		this->address = 0;
		this->state = CAN_ECU_STATE_NEVER_SEEN;
		this->lastSeen = 0;
		this->cookie = nullptr;
	}

	int		ID{0};		//!< ID e.g. generated by Projektor tool
	std::string	name{""};		//!< The ECU string name (normally used to fill in the name defined by Projektor)
	unsigned int	address{0};	//!< CAN Identifier or protocol specific interpretation
	int		state{CAN_ECU_STATE_NEVER_SEEN};		//!< State of the ECU
	unsigned int	lastSeen{0};	//!< tick value when this ECU was last seen on CAN
	void *		cookie{nullptr};		//!< save pointer to whatever you want to associate with this ECU - and don't expect it to be restored from serialization

private:

	// SERIALIZATION
	// This class is exported in we_j1939_message_object.cpp, did not find any other suitable class/file in which to export.
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int )
	{
		// no parent here: ar & boost::serialization::base_object<WECanEcu> (*this);	// identify parent class
		p_ar & ID;	//!< ID e.g. generated by Projektor tool
		p_ar & name;	//!< The ECU string name (normally used to fill in the name defined by Projektor)
		p_ar & address;	//!< CAN Identifier or protocol specific interpretation
		p_ar & state;	//!< State of the ECU
		p_ar & lastSeen;	//!< tick value when this ECU was last seen on CAN
	}
};
}//End of namespace WE

#endif /* WE_CAN_ECU_H_ */
