/**
 *	@file we_can_mapping_object.h
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2010
 *
 *	@brief Definitions of data structures used for CAN protocol variable mapping
 *
 *	This file was originally designed for CAN Freestyle implementation but we figured out
 *	that it is useful for other protocols such as J1939 also.
 *
 *	To be more generic mapping is extended beyond the 64 bit boundary of a single CAN message.
 *
 *	This part only describes the data part of a CAN transmit and handles everything related to DEI variables.
 *	Because of the DEI relation this also handles the transmit type and period handling.
 *	The CAN ID and bus related handling is done in WECANMessageObject.
 *
 *	@see WECANMessageObject
 *	@see WECANFreestyle
*/

#ifndef WE_CAN_MAPPING_OBJECT_H_
#define WE_CAN_MAPPING_OBJECT_H_

#include <string>
#include <vector>
#include <set>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>

#include "../../../dei/variables/include/we_variable.h"

namespace WE
{

class WESignalListener;
struct WECANMessage;
class WEDEI;
class WEVariableOwner;
class WECANMessageObject;
class WECANProtocolResources;

/**
 *	@enum WETransmitType
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Definition of CANFreestyle transmit types for WECANMappingObject
 *
 *	@namespace WE
 */
typedef enum
{
	WE_CAN_TRANSMIT_TYPE_UNDEFINED			 = 0x0000, // 0000 0000 0000 0000

	// Transmit settings
	WE_CAN_TRANSMIT_TYPE_ANY_VALUE_CHANGED 	 	 = 0x0001, // 0000 0000 0000 0001
	// Transmit Mapping triggered by specific value
	WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_CHANGED 	 = 0x0002, // 0000 0000 0000 0010
	// Periodic transmit mapping
	WE_CAN_TRANSMIT_TYPE_PERIODIC 			 = 0x0004, // 0000 0000 0000 0100
	// Request message enabled
	WE_CAN_TRANSMIT_TYPE_BY_REQUEST 		 = 0x0008, // 0000 0000 0000 1000
	// Sequence message
	WE_CAN_TRANSMIT_TYPE_SEQUENCE			 = 0x0010, // 0000 0000 0001 0000

	WE_CAN_TRANSMIT_TYPE				 = 0x0080, // 0000 0000 1000 0000

	// Mask with all possible Transmit settings	   if one of these bits is set, it is a transmit mapping
	WE_CAN_TRANSMIT_TYPE_TRANSMIT_MASK		 = 0x00ff,	// figure out transmit/receive mapping


	// Predefined transmit settings
	WE_CAN_TRANSMIT_TYPE_ANY_VALUE_AND_PERIODIC 	 = WE_CAN_TRANSMIT_TYPE_ANY_VALUE_CHANGED | WE_CAN_TRANSMIT_TYPE_PERIODIC,	// 0000 0000 0000 0101
	WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_AND_PERIODIC	 = WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_CHANGED | WE_CAN_TRANSMIT_TYPE_PERIODIC,	// 0000 0000 0000 0110
	WE_CAN_TRANSMIT_TYPE_BY_REQUEST_AND_PERIODIC 	 = WE_CAN_TRANSMIT_TYPE_BY_REQUEST | WE_CAN_TRANSMIT_TYPE_PERIODIC,		// 0000 0000 0000 1100

	WE_CAN_TRANSMIT_TYPE_ON_WRITE_REQUEST		 = 0x8000, // 1000 0000 0000 0000 - send out if value of external variable should be changed

	// Receive settings
	//  we expect to receive this message
	WE_CAN_TRANSMIT_TYPE_RECEIVE			 = 0x00010000, 		// 0000 0001 0000 0000 0000 0000
	//  we expect to receive this message periodically - timeout handling will be applied
	WE_CAN_TRANSMIT_TYPE_RECEIVE_PERIODIC		 = 0x00020000,		// 0000 0010 0000 0000 0000 0000
	//  send out request message to trigger receive mapping
	WE_CAN_TRANSMIT_TYPE_RECEIVE_BY_REQUEST		 = 0x00040000,		// 0000 0100 0000 0000 0000 0000
	//  receive request message triggered by specific value
	WE_CAN_TRANSMIT_TYPE_RECEIVE_CERTAIN_VALUE_CHANGED = 0x00080000,	// 0000 1000 0000 0000 0000 0000
	//  send out receive request message periodically
	WE_CAN_TRANSMIT_TYPE_PERIODIC_REQUEST		 = 0x00100000		// 0001 0000 0000 0000 0000 0000
} WETransmitType;


typedef enum
{
	WE_MAPPING_VARIABLE_TYPE_UNDEFINED	= 0x0000,
	// LENGTH based mapping
	WE_MAPPING_VARIABLE_TYPE_LENGTH		= 0x0001,
	// DELIMITER based mapping
	WE_MAPPING_VARIABLE_TYPE_DELIMITER	= 0x0002,
	// LENGTH and DELIMITER base mapping
	WE_MAPPING_VARIABLE_TYPE_MIXED		= 0x0003
} WEMappingVariableType;

/**
 *	@struct WECANMappingData
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Definition of the data for a single variable mapping inside a WECANMappingObject
 *
 *	@namespace WE
 *
 *	@see WECANMappingObject
 *
 *	This struct stores the data that is/can be necessary to map a variable into a WECANMappingObject.
 *	It does not necessarily contain data to map a variable but can also represent a static value.
 *	This data structure is used for receive and transmit objects.
 *	Important for receive objects:
 *	Only messages are excepted if constantIntData matches!
 */
struct WECANMappingData
{
	/**
	 *	@fn WECANMappingData()
	 *
	 *	@brief Standard constructor. All members are initialized.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WECANMappingData() = default;

	virtual ~WECANMappingData()
	{
		delete [] this->constantByteData;
		this->constantByteData	= nullptr;
		delete [] this->constantByteMask;
		this->constantByteMask	= nullptr;
	}

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WECANMappingData(const WECANMappingData &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WECANMappingData & operator=(const WECANMappingData &) = delete;

	int 		byteNumber{0}; 		//!< The number of the byte of the CAN message where the data is mapped to (> 0)
						// OR: negative number of "separator/delimiter chars" defining the END of the mapping
						// e.g. byteNumber == -2: data is between delimiter 1 and 2; *constantByteMask is the delimiter char then
	unsigned int 	startPosition{0};	//!< The start position inside the first byte (0..7)
	unsigned int 	numberOfBits{8};	//!< The number of bits
	unsigned int 	mappedKey{0};		//!< If a variable is mapped, this contains it's DEI ((index << 16) | subIndex) (else: 0)
	bool 		isMapped{false};	//!< If a variable is mapped, this flag is true
	WEVariableStorageType storageType{STORAGE_TYPE_UNDEFINED}; //!< Storage type information (String or Integer)
	unsigned int 	constantIntData{0};	//!< If no variable is mapped, this shall contain a constant integer value
	unsigned char *	constantByteData{nullptr}; //!< If constant data of more than 4 bytes of constantIntData should be compared otherwise nullptr
	unsigned int 	mask{0};		//!< A bit mask that is applied to the value or delimiter char if byteNumber < 0
	unsigned char * constantByteMask{nullptr}; //!< If constant data of more than 4 bytes of mask should be compared, otherwise nullptr
	bool 		isSigned{false};	//!< Flag that indicates if the value shall be interpreted as a signed value
	float 		offset{0.0};		//!< Positive or negative offset
	float 		offset2{0.0};		//!< Positive or negative offset
	float		scale{1.0};		//!< Scale that will be applied to the variable value
	int		shiftToRight{0};	//!< To extract only parts of a value, a shift factor can be applied (negative should shift to left)
	unsigned int	numberOfCANBytes{1};	//!< The number of CAN data bytes that this mapping occupies
	WESignalListener * 	timeoutListener{nullptr}; //!< Pointer to a WESignalListener. To be called if mapped variable could not be received in time
	bool		forwardDontCare{false};	//!< Only relevant for J1939. If set to true a don't care value (e.g. all bits set) will be set to the mapped variable. If false (default) the value will be ignored.

private:
	// SERIALIZATION
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int )
	{
		unsigned int i = 0;
		// no parent class here: ar & boost::serialization::base_object<AppBaseDDO> (*this);
		p_ar & byteNumber; 	//!< The number of the byte of the CAN message where the data is mapped to (> 0)
						// OR: negative number of "separator/delimiter chars" defining the END of the mapping
						// e.g. byteNumber == -2: data is between delimiter 1 and 2; *constantByteMask is the delimiter char then
		p_ar & startPosition;	//!< The start position inside the first byte (0..7)
		p_ar & numberOfBits;	//!< The number of bits
		p_ar & mappedKey;	//!< If a variable is mapped, this contains it's DEI ((index << 16) | subIndex) (else: 0)
		// cppcheck-suppress bitwiseOnBoolean
		p_ar & isMapped;	//!< If a variable is mapped, this flag is true
		p_ar & storageType; //!< Storagetype information (String or Integer)
		p_ar & constantIntData;//!< If no variable is mapped, this shall contain a constant integer value
		p_ar & numberOfCANBytes; //!< The number of CAN data bytes that this mapping occupies (moved up to have it deserialized for the loop)
		if (nullptr != this->constantByteData)
		{
			for (i = 0; i < this->numberOfCANBytes; ++i)
			{
				p_ar & constantByteData[i]; //!< If constant data of more than 4 bytes of constantIntData should be compared otherwise nullptr
			}
		}

		p_ar & mask;		//!< A bit mask that is applied to the value or delimiter char if byteNumber < 0

		if (nullptr != this->constantByteMask)
		{
			for (i = 0; i < this->numberOfCANBytes; ++i)
			{
				p_ar & constantByteMask[i]; //!< If constant data of more than 4 bytes of mask should be compared, otherwise nullptr
			}
		}
		// cppcheck-suppress bitwiseOnBoolean
		p_ar & isSigned;	//!< Flag that indicates if the value shall be interpreted as a signed value
		p_ar & offset;		//!< Positive or negative offset
		p_ar & offset2;		//!< Positive or negative offset
		p_ar & scale;		//!< Scale that will be applied to the variable value
		p_ar & shiftToRight;	//!< To extract only parts of a value, a shift factor can be applied (negative should shift to left)
		// should not need to serialize this: timeoutListener; //!< Pointer to a WESignalListener. To be called if mapped variable could not be received in time
		// cppcheck-suppress bitwiseOnBoolean
		p_ar & forwardDontCare;
	}
};


/**
 *	@class WECANMappingObject
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief WECANMappingObject stores data and provides some functionality for WECANFreestyle variable mapping
 *
 *	@namespace WE
 *
 *	This class is e.g. used for CANFreestyle variable mapping functionality.
 *	Up to 64 variables can be mapped for a single CAN message (64 variables all representing one bit).
 *	Beyond the size of a CAN message buffers can be provided to map byte aligned data.
 *	A WECANMappingObject can be a receive or transmit object.
 *	A transmit object can cause the sending of a CAN message containing the values of all mapped variables. The
 *	trigger for this can be any entry in  WETransmitType.
 *	A receive object receives the values of the mapped variables from a CAN message and updates the variables
 *	in WEDEI.
 *
 *	Calculation of value of mapped integers is done in the following order:
 *
 *	receive: |	(1) extract value		transmit:	 .	(5) apply value
 *		 |	(2) limit to # of bits				/|\	(4) avoid sign extension if shift right and unsigned
 *		 |	(3) shift right					 |	(3) shift left
 *		 |	(4) mask					 |	(2) mask
 *		 |	(5) sign extension				 |	(1) reverse scale, offset
 *		 |	(6) scale, offset				 |
 *		\|/							 |
 *		 .
 *
 *	NOTE: Sign extension checks the most significant bit based on the number of mapped bits.
 *	NOTE: I.e. is shift factor is > 0 the bit checked will be 0 in any case, so we will never detect negative values.
 *	NOTE: If shift factor < 0 one of the bits of the mapped value will be considered as the sign bit.
 *
 *	Assume 16 bits are mapped and the value is 0x3fff == 16383.
 *	Applying shift factor -2 would convert it to 0xfffc == 65532 == -4.
 *
 *	So one should be very careful when combining "shift factor" and "isSigned" property in mappings.
 *
 */
class WECANMappingObject
{
public:
	/**
	 *	@fn WECANMappingObject(const WECANMappingObject & p_mapping)
	 *
	 *	@brief Copy contructor providing copying all data to a new instance of this class.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_mapping - object to be copied
	 *
	 *	@warning This does NOT copy the vector of pointers to mapping data to avoid trouble with memory allocation and inheritance.
	 */
	WECANMappingObject(const WECANMappingObject & p_mapping);
	WECANMappingObject & operator=(const WECANMappingObject &p_mapping);


	/**
	 *	@fn WECANMappingObject(
			unsigned int p_messageLength,
			unsigned int p_transmitType,
			unsigned int p_transmitTriggerKey,
			unsigned int p_transmitPeriod,
			int p_mappingType,
			bool p_littleEndian,
			char p_fillByte,
			bool p_useInvalidTransmitPattern = false,
			unsigned int p_invalidTransmitPattern = 0
			)
	 *
	 *	@brief Contructor providing all necessary data for creating an instance of this class
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_messageLength Number of bytes of the CAN message (DLC)
	 *	@param p_transmitType trigger definition for a transmit object or WE_CAN_TRANSMIT_TYPE_RECEIVE
	 *	@param p_transmitTriggerKey if p_transmitType includes XXX_CERTAIN_VALUE_XXX, this contains the key of the variable
	 *	@param p_transmitPeriod if p_transmitType is XXX_PERIODIC_XXX, this contains the transmit period in milliseconds
	 *	@param p_mappingType	- type of this mapping (i.e. length, delimiter or mixed mapping type)
	 *	@param p_littleEndian If true, received and sent CAN message data bytes will be treated as little endian values
	 *	@param p_fillByte The fill byte which is used to fill unused CAN message bytes if configured mapping length is > than the length of the mapped variables
	 *	@param p_useInvalidTransmitPattern If false the value from DEI will be used for transmit even if marked as "invalid"
	 *	@param p_invalidTransmitPattern pattern to replace value if it is reported "invalid" by DEI
	 */
	WECANMappingObject(
			unsigned int p_messageLength,
			unsigned int p_transmitType,
			unsigned int p_transmitTriggerKey,
			unsigned int p_transmitPeriod,
			int p_mappingType,
			bool p_littleEndian,
			unsigned char p_fillByte,
			bool p_useInvalidTransmitPattern = false,
			unsigned int p_invalidTransmitPattern = 0
			);


	/**
	 *	@fn ~WECANMappingObject()
	 *
	 *	@brief Standard destructor. Frees all occupied and managed memory.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 */
	virtual ~WECANMappingObject();


	/**
	 *	@fn int addMappingData(WECANMappingData * p_mappingData)
	 *
	 *	@brief Use this function to add a variable mapping or constant data to this mapping object
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_mappingData Pointer to a mapping data object to be copied into this mapping object
	 *
	 *	@returns WE_SUCCESS if mapping was accepted
	 *	@returns Any other value int addMappingData(unsigned char p_byteNumber,
	 *			unsigned char p_startPosition,
	 *			unsigned int p_numberOfBits,
	 *			unsigned short p_mappedIndex,
	 *			unsigned short p_mappedSubindex,
	 *			bool p_signed,
	 *			unsigned int p_constantIntData,
	 *			unsigned int p_mask,
	 *			float p_offset,
	 *			float p_scale,
	 *			char p_shift
	 *			WESignalListener * p_timeoutListener) may return.
	 *
	 *	When calling this function, a new WECANMappingData object gets created and stored in this object.
	 *	If buffers are provided in p_mappingData they will be copied as well.
	 *
	 *	Internally this method calls the version with multiple parameters - please check its documentation also.
	 *
	 *	@see int addMappingData(unsigned char p_byteNumber,
	 *			unsigned char p_startPosition,
	 *			unsigned int p_numberOfBits,
	 *			unsigned short p_mappedIndex,
	 *			unsigned short p_mappedSubindex,
	 *			bool p_signed,
	 *			unsigned int p_constantIntData,
	 *			unsigned int p_mask,
	 *			float p_offset,
	 *			float p_scale,
	 *			char p_shift
	 *			WESignalListener * p_timeoutListener)
	 *
	 */
	int addMappingData(WECANMappingData * p_mappingData);


	/**
	 *	@brief Use this function to add a variable mapping or constant data to this mapping object
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_byteNumber Number of the byte where the mapped (or constant) value is stored; negative for delimiter char mapping
	 *	@param p_startPosition Start bit of the data inside p_byteNumber
	 *	@param p_numberOfBits The "length" of the data/value in bits
	 *	@param p_mappedKey Combined Index of the mapped variable (if constant data is defined, this is 0)
	 *	@param p_storageType Storage type of the mapped variable (String or Integer)
	 *	@param p_signed Flag if the data shall be interpreted a signed or unsigned
	 *	@param p_constantIntData If no variable shall be mapped, but the value shall be constant, provide the constant value with this parameter
	 *	@param p_constByteData If more than 4 bytes of constant data should be mapped this points to the buffer containing them
	 *	@param p_mask The mask of the mapping entry or delimiter char if p_byteNumber < 0
	 *	@param p_offset Offset that is applied to the mapped value
	 *	@param p_scale Scaling factor with which the mapped value will be multiplied with
	 *	@param p_shift Number of bits that the mapped value will be shifted to the right
	 *	@param p_timeoutListener Listener to be informed if the mapped variable becomes (in)valid
	 *	@param p_forwardDontCare true: Even all bits set to one is written to variable, false: in J1939 'all bits == 1' is ignored
	 *
	 *	@returns WE_SUCCESS if mapping was accepted
	 *	@returns WE_INVALID_PARAMETER_ERROR if mapping for a buffer >8 bytes (not a single CAN message) isn't byte aligned.
	 *	@returns WE_INVALID_PARAMETER_ERROR if mapping for more than 4 bytes of const data isn't byte aligned.
	 *
	 *	When calling this function, a new WECANMappingData object gets created and stored in this object.
	 *	If buffers are provided (i.e. p_constByteData != nullptr or p_constByteMask != nullptr) they will be copied.
	 *
	 *	This method uses p_byteNumber and p_startPosition to identify the least significant bit of a constant or variable
	 *	(if the value is <= 32 bits). Only for these values endianess and shift factors are taken into account.
	 *
	 *	If values > 32 bit should be compared, mapping should be compared against buffers > 64 bit (more than a single CAN message)
	 *	or strings should be extracted or compared then all data is expected to be byte aligned.
	 *
	 *	If instead of position based mapping a delimiter char is used then p_bytenumber defines the negated occurrence of the
	 *	terminating delimiter char.
	 *	The delimiter char is provided in p_mask in this case.
	 *	Assume "*" is the delimiter char and the mapping should find the data between the first and second occurrence of this char.
	 *	Then "p_byteNumber" would be set to -2 and p_mask to '*'.
	 *	Delimiter based constants will be recognized if at least p_numberOfBits/8 bytes match with the constant provided.
	 *
	 *	@warning This method will NOT check if a variable key provided is valid and/or points to a variable of correct type.
	 *
	 */
	int addMappingData(int p_byteNumber,
		unsigned int p_startPosition,
		unsigned int p_numberOfBits,
		unsigned int p_mappedKey,
		WEVariableStorageType p_storageType,
		bool p_signed,
		unsigned int p_constantIntData,
		unsigned char * p_constByteData,
		unsigned int p_mask,
		unsigned char * p_constByteMask,
		float p_offset,
		float p_offset2,
		float p_scale,
		int p_shift,
		WESignalListener * p_timeoutListener,
		bool p_forwardDontCare);


	/**
	 *	@fn bool containsVariable(unsigned int p_key)
	 *
	 *	@brief Returns true if variable identified by p_key is mapped to this object
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_key A combination of index/subindex of a variable
	 *
	 *	@returns true if a variable identified by p_key is mapped to this object
	 *	@returns false if a variable identified by p_key is NOT mapped to this object
	 *
	 *	The key to be provided consists of (varIndex << 16) | (varSubIndex)
	 *
	 *
	 */
	bool containsVariable(unsigned int p_key) const;


	/**
	 *	@brief Stores all its variable keys in the multimap provided and adds a pointer to its CAN message object as data.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_map Pointer to the map to be used.
	 *	@param p_mapMsg Pointer to be added to the map as data
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_map is nullptr or p_mapMsg is nullptr
	 *
	 *	This method is intended to feed a multimap of all variables maintained in a CAN stack.
	 *	With this map the stack can directly trigger the CAN messages using a variable if a new value arrives.
	 *
	 *
	 */
	int storeVariablesInMultimap(std::unordered_multimap< unsigned int, WECANMessageObject * > * p_map, WECANMessageObject * p_mapMsg) const;


	/**
	 *	@fn bool isReceiveMapping( void )
	 *
	 *	@brief Check if this is a receive (or a send) mapping.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns true if this mapping is a receive mapping
	 *	@returns false otherwise
	 *
	 *
	 */
	bool isReceiveMapping( void ) const
	{
		return (0 == (WE_CAN_TRANSMIT_TYPE_TRANSMIT_MASK & this->transmitType) );
	}


	/**
	 *	@fn bool isLittleEndian( void )
	 *
	 *	@brief Check if if byte order is little endian for this mapping.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param none
	 *
	 *	@returns true if this mapping has little endian byte order
	 *	@returns false otherwise
	 *
	 *
	 */
	bool isLittleEndian( void ) const
	{
		return this->littleEndian;
	}


	/**
	 *	@fn unsigned int getTransmitType( void )
	 *
	 *	@brief Get transmit type of this mapping.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns Transmit type of this mapping
	 *
	 *	This method is e.g. used to figure out if this mapping is a part of a transmit sequence.
	 *
	 *	@see WETransmitType
	 */
	unsigned int getTransmitType( void ) const
	{
		return this->transmitType;
	}


	/**
	 *	@fn std::vector< WECANMappingData * > getMappingObjects( void )
	 *
	 *	@brief Get all mapping objects of the mapping.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param NONE
	 *
	 *	@returns vector with all mappingObjects
	 *	@returns an empty vector if no mapping objects exists
	 *
	 *
	 */
	std::vector< WECANMappingData * > getMappingObjects( void ) const
	{
		return this->mappingObjects;
	}


	/**
	 *	@fn int setTransmitType(WETransmitType p_transmitType );
	 *
	 *	@brief Set transmit type for data mapping object.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_transmitType transmit type to set.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	@see WETransmitType in WECANMappingObject
	 *
	 */
	int setTransmitType(WETransmitType p_transmitType);

	/**
	 *	@fn unsigned int getTransmitPeriod( void )
	 *
	 *	@brief Get transmit period of this mapping.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns Transmit period of this mapping
	 *
	 *	This method is e.g. used to calculate when this mapping has to be sent out next time.
	 *
	 *	@see WETick
	 */
	unsigned int getTransmitPeriod( void ) const
	{
		return this->transmitPeriod;
	}


	/**
	 *	@fn void setMappingType( int p_mappingType );
	 *
	 *	@brief Set mapping type for this CAN message.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_mapping - int type of mapping (delimiter, length or mixed)
	 *
	 */
	void setMappingType( int p_mappingType );


	/**
	 *	@fn int getMappingType( void );
	 *
	 *	@brief Get mapping type of this message. (delimiter, length or mixed)
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param none
	 *
	 *	@returns mapping type
	 *	@returns WE_MAPPING_VARIABLE_TYPE_UNDEFINED if it is not set.
	 *
	 *
	 */
	int getMappingType( void ) const;


	/**
	 *	@fn unsigned int getTransmitTriggerKey( void )
	 *
	 *	@brief Get transmit combined key of trigger variable of this mapping.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns Combined key identifying DEI variable which should trigger this (transmit) mapping.
	 *
	 *	This method is e.g. used to set a trigger that is not part of the mapping's data part.
	 *
	 *	@see WEDEI
	 */
	unsigned int getTransmitTriggerKey( void ) const
	{
		return this->triggerVariableKey;
	}


	/**
	 *	@fn bool WECANMappingObject::fitsConstData(WECANMessage * p_msg)
	 *
	 *	@brief Check if the data provided fits this mapping's constant data.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg CAN message to be checked
	 *
	 *	@returns true if this mapping fits the p_msg
	 *	@returns false otherwise
	 *
	 *	This method checks the message length and const data of the message provided.
	 *	It is assumed that the CAN ID of the CAN message was checked before.
	 *	Internally this method uses fitsConstData(unsigned char * p_data, unsigned int p_size).
	 *
	 *	@see fitsConstData(unsigned char * p_data, unsigned int p_size)
	 */
	bool fitsConstData(WECANMessage * p_msg) const;


	/**
	 *	@fn bool fitsConstData(unsigned char * p_data, unsigned int p_size)
	 *
	 *	@brief Check if the data provided fits this mapping's constant data.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_data CAN data to be checked
	 *	@param p_size length of p_data in bytes
	 *
	 *	@returns true if this mapping fits p_data and p_length
	 *	@returns false otherwise
	 *
	 *	This method checks the const data of the data provided and succeeds when the the shorter
	 *	one of both ends and patterns matched.
	 *	It is assumed that the CAN ID was checked before.
	 */
	bool fitsConstData(unsigned char * p_data, unsigned int p_size) const;


	/**
	 *	@fn unsigned int getMessageLength( void )
	 *
	 *	@brief Get number of bytes expected for this mapping.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns number of bytes this mapping consists of
	 *
	 */
	unsigned int getMessageLength( void ) const;


	/**
	 *	@fn int extractMappedValues(WECANMessage * p_msg, WEDEI * p_dei, WEVariableOwner * p_owner, bool p_checkMessageLength = true)
	 *
	 *	@brief Extract variable values and transfer them to p_dei.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg CAN message to be checked.
	 *	@param p_dei Pointer to the DEI instance which should get the values.
	 *	@param p_owner Caller is variable owner "p_owner" - so we can force setting his variables from here
	 *	@param p_checkMessageLength default is true but some protocols (like CANopen) also accept shorter messages
	 *
	 *	@returns WE_INVALID_MESSAGE_RECEIVED_ERROR if size of message does not match mapping
	 *	@returns errorcode as specified for extractMappedValues(unsigned char * p_data, unsigned int p_size) otherwise
	 *
	 *	When this method is called we expect that this mapping is the one to be used.
	 *	So we just extract all the values we can find and transfer them to DEI directly.
	 *
	 *	@see extractMappedValues(unsigned char * p_data, unsigned int p_size, WEDEI * p_dei, WEVariableOwner * p_owner)
	 */
	int extractMappedValues(WECANMessage * p_msg, WEDEI * p_dei, WEVariableOwner * p_owner, bool p_checkMessageLength = true) const;


	/**
	 *	@fn int extractMappedValues(unsigned char * p_data, unsigned int p_size, WEDEI * p_dei, WEVariableOwner * p_owner)
	 *
	 *	@brief Extract variable values and transfer them to p_dei.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_data CAN data to be checked
	 *	@param p_size length of p_data in bytes
	 *	@param p_dei Pointer to the DEI instance which should get the values.
	 *	@param p_owner Caller is variable owner "p_owner" - so we can force setting his variables from here
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	When this method is called we expect that this mapping is the one to be used.
	 *	So we just extract all the values we can find and transfer them to DEI directly.
	 *
	 *	If more than 32 bits are mapped or delimiter based mapping is used this method expects them to be written
	 *	to a string variable.
	 *
	 *	For delimiter based mapping masks are ignored - it just copies byte by byte between delimiters.
	 *
	 *	p_dei is only forwarded to handleExtractedValue() so it is allowed to be nullptr here. If handleExtractedValue() is
	 *	not overwritten it forwards the values to the DEI provided.
	 *
	 *	@warning Currently this implementation ignores return values from DEI - returning WE_SUCCESS only means that it tried to set everything.
	 *
	 *	@see handleExtractedValue()
	 *
	 */
	int extractMappedValues(unsigned char * p_data, unsigned int p_size, WEDEI * p_dei, WEVariableOwner * p_owner) const;


	/**
	 *	@fn virtual int checkRawValue(unsigned int p_rawValue, WECANMappingData * p_mapping)
	 *
	 *	@brief Check if the raw value extracted from CAN message may have a dedicated meaning (such as "don't care")
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_rawValue Value extracted from CAN message (assembled according to endianess but no scale, offset etc. applied)
	 *	@param p_mapping The WECANMappingData used to fetch p_rawValue from CAN message
	 *
	 *	@returns WE_SUCCESS if p_rawValue is a normal, valid value
	 *	@returns WE_INVALID_MESSAGE_RECEIVED_ERROR if the value should be set to "invalid" (e.g. in DEI)
	 *	@returns WE_NOT_SUPPORTED_ERROR if the value should just be ignored (i.e. indicates "don't care")
	 *	@returns The default implementation returns WE_SUCCESS in any case
	 *
	 *	This function gets called by the extractMappedValues function for each mapped variable.
	 *	The default implementation considers every raw value provided as "valid".
	 *	The function can be overwritten in inheriting classes to implement custom behavior such as "all bits set indicates 'don't care'"
	 *	for J1939 protocol.
	 *
	 *	p_mapping can be used to get more information about p_rawValue e.g. the number of bits actually extracted from the CAn message.
	 *
	 *
	 *	@see int extractMappedValues(WECANMessage * p_msg, WEDEI * p_dei, WEVariableOwner * p_owner)
	 *	@see int handleExtractedValue(unsigned int p_key, int p_value, WEDEI * p_dei, WEVariableOwner * p_owner, int p_valueValid)
	 *	@see WEPDO
	 */
	virtual int checkRawValue(unsigned int p_rawValue, WECANMappingData * p_mapping) const;


	/**
	 *	@fn virtual int checkRawValue(const std::string & p_rawValue, WECANMappingData * p_mapping)
	 *
	 *	@brief Check if the raw value extracted from CAN message may have a dedicated meaning (such as "don't care)
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_rawValue Value extracted from CAN message (assembled according to endianess but no scale, offset etc. applied)
	 *	@param p_mapping The WECANMappingData used to fetch p_rawValue from CAN message
	 *
	 *	@returns WE_SUCCESS if p_rawValue is a normal, valid value
	 *	@returns WE_INVALID_MESSAGE_RECEIVED_ERROR if the value should be set to "invalid" (e.g. in DEI)
	 *	@returns WE_NOT_SUPPORTED_ERROR if the value should just be ignored (i.e. indicates "don't care")
	 *	@returns The default implementation returns WE_SUCCESS in any case
	 *
	 *	This function gets called by the extractMappedValues function for each mapped variable.
	 *	The default implementation considers every raw value provided as "valid".
	 *	The function can be overwritten in inheriting classes to implement custom behavior such as "all bits set indicates 'don't care'"
	 *	for J1939 protocol.
	 *
	 *	p_mapping can be used to get more information about p_rawValue e.g. the number of bits actually extracted from the CAn message.
	 *
	 *
	 *	@see int extractMappedValues(WECANMessage * p_msg, WEDEI * p_dei, WEVariableOwner * p_owner)
	 *	@see int int handleExtractedValue(unsigned int p_key, std::string & p_value, WEDEI * p_dei, WEVariableOwner * p_owner, int p_valueValid)
	 *	@see WEPDO
	 */
	virtual int checkRawValue(const std::string & p_rawValue, WECANMappingData * p_mapping) const;


	/**
	 *	@fn virtual int handleExtractedValue(unsigned int p_key, int p_value, WEDEI * p_dei, WEVariableOwner * p_owner, int p_valueValid)
	 *
	 *	@brief Handle a value extracted by "extractMappedValues" function
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_key A combination of index/subindex of a variable
	 *	@param p_value The extracted (integer) value with mask, scale and offset applied
	 *	@param p_dei Pointer to the DEI instance
	 *	@param p_owner Pointer to the owner of the variable
	 *	@param p_valueValid result of checkRawValue()
	 *
	 *	@returns The default implementation returns the return value of WEDEI::setVariableValue / WEDEI::forceSetVariableValue
	 *	@returns The return values of of overwritten methods are described in the corresponding classes
	 *
	 *	@see int extractMappedValues(WECANMessage * p_msg, WEDEI * p_dei, WEVariableOwner * p_owner)
	 *	@see WEPDO
	 *
	 *	This function gets called by the extractMappedValues function for each mapped variable.
	 *	The default implementation here just passes the value to WEDEI if p_valueValid is WE_SUCCESS.
	 *	The function can be overwritten in inheriting classes to implement custom behavior.
	 *
	 *	The key to be provided consists of (varIndex << 16) | (varSubIndex)
	 *
	 */
	virtual int handleExtractedValue(unsigned int p_key, int p_value, WEDEI * p_dei, WEVariableOwner * p_owner, int p_valueValid) const;


	/**
	 *	@fn virtual virtual int handleExtractedValue(unsigned int p_key, const std::string & p_value, WEDEI * p_dei, WEVariableOwner * p_owner, int p_valueValid)
	 *
	 *	@brief Handle a value extracted by "extractMappedValues" function
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_key A combination of index/subindex of a variable
	 *	@param p_value The extracted (string) value
	 *	@param p_dei Pointer to the DEI instance
	 *	@param p_owner Pointer to the owner of the variable
	 *	@param p_valueValid result of checkRawValue()
	 *
	 *	@returns The default implementation returns the return value of WEDEI::setVariableValue / WEDEI::forceSetVariableValue
	 *	@returns The return values of of overwritten methods are described in the corresponding classes
	 *
	 *	@see int extractMappedValues(WECANMessage * p_msg, WEDEI * p_dei, WEVariableOwner * p_owner)
	 *	@see WEPDO
	 *
	 *	This function gets called by the extractMappedValues function for each mapped variable.
	 *	The default implementation here just passes the value to WEDEI if p_valueValid is WE_SUCCESS.
	 *	The function can be overwritten in inheriting classes to implement custom behavior
	 *
	 *	The key to be provided consists of (varIndex << 16) | (varSubIndex)
	 *
	 */
	virtual int handleExtractedValue(unsigned int p_key, const std::string & p_value, WEDEI * p_dei, WEVariableOwner * p_owner, int p_valueValid) const;


	/**
	 *	@fn bool isTriggeredByVariableValueChange(unsigned int p_varKey)
	 *
	 *	@brief Check if a modified value of p_varKey should trigger to send this mapping.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_varKey - Key of the variable whose value changed
	 *
	 *	@returns true if mapping should be sent
	 *	@returns false otherwise
	 *
	 *	This method returns "true" if this is a transmit mapping which contains the variable
	 *	identified by p_varKey and has trigger "WE_CAN_TRANSMIT_TYPE_ANY_VALUE_CHANGED"
	 *	or "WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_CHANGED" where certain value has to be the one of
	 *	p_varKey.
	 *
	 *
	 *	@see int getTransmitType( void )
	 *
	 *
	 */
	bool isTriggeredByVariableValueChange(unsigned int p_varKey) const;


	/**
	 *	@fn bool isTriggeredByWriteRequest( unsigned int p_varKey )
	 *
	 *	@brief Check if this mapping should be sent out for a write request on variable identified by p_varKey.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_varKey - Key of the variable whose value is requested to be changed
	 *
	 *	@returns true if mapping should be sent
	 *	@returns false otherwise
	 *
	 *	This method returns "true" if this is a transmit mapping which contains the variable
	 *	identified by p_varKey and has trigger "WE_CAN_TRANSMIT_TYPE_ON_WRITE_REQUEST".
	 *
	 *
	 *	@see int getTransmitType( void )
	 *
	 *
	 */
	bool isTriggeredByWriteRequest( unsigned int p_varKey ) const;

	/**
	 *	@fn int fillMessageFromDEI(WECANMessage * p_msg, WEDEI * p_dei, unsigned int p_varKey = 0, int p_varValue = 0)
	 *
	 *	@brief Fill CAN message data part with values from DEI and constants.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg - CAN message to be filled with data
	 *	@param p_dei - pointer to DEI instance where variable values should come from
	 *	@param p_varKey - key of variable which should NOT be sent with DEI value but with p_varValue.
	 *	@param p_varValue - value to be used for variable identified by "p_varKey"
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_dei == nullptr
	 *	@returns WE_RECEIVE_ERROR if this is a receive mapping object
	 *
	 *	Compose data part of a CAN message as defined by this mapping.
	 *	Optionally replace value of variable identified by p_varKey with p_varValue.
	 *	This could be used to use the mapping as a "write request" for this variable.
	 *
	 *	@see addSingleMappingDataToCANMsgData()
	 *
	 *
	 */
	int fillMessageFromDEI(WECANMessage * p_msg, WEDEI * p_dei, unsigned int p_varKey = 0, int p_varValue = 0) const;

	/**
	 *	@fn int fillMessageFromDEI(
			unsigned char * p_data,
			unsigned int p_size,
			unsigned int & p_filledSize,
			WEDEI * p_dei,
			unsigned int p_varKey = 0,
			int p_varValue = 0)
	 *
	 *	@brief Fill data buffer with values to be sent from DEI and constants.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_data - buffer to be filled with data
	 *	@param p_size - size of p_data in bytes
	 *	@param p_filledSize - amount of bytes that were actually filled with data
	 *	@param p_dei - pointer to DEI instance where variable values should come from
	 *	@param p_varKey - key of variable which should NOT be sent with DEI value but with p_varValue.
	 *	@param p_varValue - value to be used for variable identified by "p_varKey"
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_dei == nullptr or p_data == nullptr
	 *	@returns WE_RECEIVE_ERROR if this is a receive mapping object
	 *	@returns any error code addSingleMappingDataToCANMsgData() could return
	 *
	 *	Compose data part to be sent out into a data buffer as defined by this mapping.
	 *	Optionally replace value of variable identified by p_varKey with p_varValue.
	 *	This could be used to use the mapping as a "write request" for this variable.
	 *
	 *	Don't remember what that should mean: 'There could be a version to create write requests for strings.'
	 *
	 *	@see addSingleMappingDataToCANMsgData()
	 *
	 */
	int fillMessageFromDEI(
			unsigned char * p_data,
			unsigned int p_size,
			unsigned int & p_filledSize,
			WEDEI * p_dei,
			unsigned int p_varKey = 0,
			int p_varValue = 0) const;


	/**
	 *	@fn int addSingleMappingDataToCANMsgData(
					unsigned char * p_data,
					unsigned int p_size,
					unsigned int & p_filledSize,
					int p_value,
					WECANMappingData * p_mapping,
					bool p_applyRawValue = false)
	 *
	 *	@brief Fill one int value into data buffer as defined by mapping.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_data - buffer to be filled with data
	 *	@param p_size - size of p_data in bytes
	 *	@param p_filledSize - amount of bytes that were actually filled with data
	 *	@param p_value - value to be mapped into p_data
	 *	@param p_mapping - mapping which defines how p_value will be applied to p_data
	 *	@param p_applyRawValue - if true shifting, scale and offset will not be applied
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_NOT_SUPPORTED_ERROR if p_size > 8 but p_mapping is not byte aligned
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if mapping the value would cross boundaries of p_data
	 *
	 *	Reverse apply scale, offset, mask, shift and then copy the value into p_data.
	 *	Value to send is calculated as
	 *	1) value = (value - offset) / scale
	 *	2) value = value & mask
	 *	3) value = value << shift (negative "shift" is allowed)
	 *	4) Limit size of value to number of bits mapped
	 *
	 *	If p_size is greater than 8 this internally calls the byte array version of this method.
	 *
	 *	p_applyRawValue can be set to "true" in order to apply p_value without any modification except
	 *	limitation to the mapped number of bits.
	 *	This is useful to e.g. set a value to an "invalid" pattern/value.
	 *
	 *	@warning Internally the implementation assumes that size of p_data is at least 8 bytes even if p_size indicates less.
	 *
	 */
	int addSingleMappingDataToCANMsgData(
					unsigned char * p_data,
					unsigned int p_size,
					unsigned int & p_filledSize,
					int p_value,
					WECANMappingData * p_mapping,
					bool p_applyRawValue = false) const;


	/**
	 *	@fn int addSingleMappingDataToCANMsgData(
					unsigned char * p_data,
					unsigned int p_size,
					unsigned int & p_filledSize,
					std::string & p_value,
					WECANMappingData * p_mapping)
	 *	@fn int addSingleMappingDataToCANMsgData(
					unsigned char * p_data,
					unsigned int p_size,
					unsigned int & p_filledSize,
					const unsigned char * p_value,
					unsigned int p_valSize,
					WECANMappingData * p_mapping)
	 *
	 *	@brief Fill one string or char array value into data buffer as defined by mapping.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_data - buffer to be filled with data
	 *	@param p_size - size of p_data in bytes
	 *	@param p_filledSize - amount of bytes that were actually filled with data
	 *	@param p_value - value to be mapped into p_data
	 *	@param p_valSize - number of bytes of "p_value"
	 *	@param p_mapping - mapping which defines how p_value will be applied to p_data
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_NOT_SUPPORTED_ERROR if p_size > 8 but p_mapping is not byte aligned
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if mapping the value would cross boundaries of p_data
	 *
	 *	This method is expected to be used for data that exceeds the 4 bytes of an integer and therefore ignores
	 *	any shifting that may be defined in the mapping.
	 *
	 *	For delimiter based mapping masks are ignored - it just appends byte by byte after the appropriate
	 *	delimiter and finally adds another delimiter char.
	 *	Therefore this method has to be called in the order data should appear in p_data - it will not move
	 *	any data in p_data which should appear behind p_value.
	 *
	 *	@warning Internally the implementation assumes that size of p_data is at least 8 bytes even if p_size indicates less.
	 *
	 */
	int addSingleMappingDataToCANMsgData(
					unsigned char * p_data,
					unsigned int p_size,
					unsigned int & p_filledSize,
					const std::string & p_value,
					WECANMappingData * p_mapping) const;
	static int addSingleMappingDataToCANMsgData(
					unsigned char * p_data,
					unsigned int p_size,
					unsigned int & p_filledSize,
					const unsigned char * p_value,
					unsigned int p_valSize,
					WECANMappingData * p_mapping);


	/**
	 *	@fn std::uint64_t getCANdata(WECANMessage * p_msg, bool isLittleEndian = true);
	 *
	 *	@brief Copy data provided by p_msg into return value taking endianess into account.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg CAN Message containing the data
	 *	@param p_isLittleEndian true if p_msg is transferred in little endian format
	 *
	 *	@returns Data of p_msg copied according to its length and endianess.
	 *
	 *
	 */
	std::uint64_t getCANdata(WECANMessage * p_msg, bool p_isLittleEndian = true) const;


	/**
	 *	@fn std::uint64_t getCANdata(unsigned char * p_data, unsigned int p_size, bool p_isLittleEndian = true)
	 *
	 *	@brief Copy data provided into return value taking endianess into account.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_data data to be copied
	 *	@param p_size length in bytes of the buffer provided by p_data
	 *	@param p_isLittleEndian true if p_msg is transferred in little endian format
	 *
	 *	@returns Data of p_data copied according to its length (<= 8) and endianess.
	 *	@returns 0 if conditions for parameters don't match.
	 *
	 *	The value returned will contain the data bytes as they were transmitted in CAN message.
	 *	I.e. if e.g. two bytes are transferred they will be shifted to the two most significant
	 *	bytes of the std::uint64_t returned.
	 *
	 */
	std::uint64_t getCANdata(unsigned char * p_data, unsigned int p_size, bool p_isLittleEndian = true) const;


	/**
	 *	@fn int sendTimeoutEvent(bool p_engaged, WEVariableOwner * p_owner = nullptr, WEDEI * p_dei = nullptr)
	 *
	 *	@brief This function is called whenever a receive timeout occurs. The event will be forwarded to timoutListeners of every variable in this mapping.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_engaged true: timeout occurred, false: timeout resolved
	 *	@param p_owner This owner's variables will be set to invalid in p_dei if p_engaged is true
	 *	@param p_dei DEI instance where mapped variables can be found
	 *
	 *	@returns The last return value != WE_SUCCESS of WESignalListner::signalOccurred()
	 *	@returns The last return value != WE_SUCCESS of p_dei->markValueAsInvalid()
	 *	@returns WE_NOT_SUPPORTED_ERROR if this is a send mapping.
	 *	@returns WE_SUCCESS in case of success.
	 *
	 *	@see WESignalListener
	 *
	 *	When a receive timeout happens, timeoutListener of every mapped variable will receive a TIMEOUT_EVENT event.
	 *	If the appropriate message is received again, a TIMEOUT_EVENT is sent to timeoutListener.
	 *	Both events differ in parameter "ep3.isValid" which is set to p_engaged.
	 *	Additionally all variables of this mapping will be set to "invalid" in DEI if p_engaged is true.
	 *
	 *	For transmit mappings this method does nothing.
	 *
	 */
	int sendTimeoutEvent(bool p_engaged, WEVariableOwner * p_owner = nullptr, WEDEI * p_dei = nullptr) const;

	/**
         *	@fn int copyMappingObjectData(WECANMappingObject * toCopyFromMappingObject)
         *
         *	@brief Copies the mappingObject from the reference toCopyFromMappingObject to this->mappingObjects.
         *
         *	@author ju@wachendorff.de
         *	@author Joachim Urbach
         *
         *	@param p_toCopyFromMappingObject reference of WECANMappingObject to copy data from
         *
         *	@returns 	WE_SUCCES if copy was successfull
         *			WE_INVALID_PARAMETER_ERROR if parameter reference is nullptr
         *
         */
	int copyMappingObjectData(WECANMappingObject * p_toCopyFromMappingObject);

protected:

	unsigned int	messageLength;				//!< Length of the message / number of data (value) bytes
	std::vector< WECANMappingData * > mappingObjects;	//!< Vector containing all mapped data information
	std::set< unsigned int > mappedVariables;		//!< A set containing the combined keys for all mapped variables (for faster searching)
	unsigned int	transmitType;				//!< The transmit type of this mapping object
	unsigned int	triggerVariableKey;			//!< if transmitType is XXX_CERTAIN_VALUE_XXX, this contains the key of the variable
	unsigned int	transmitPeriod;				//!< if transmitType is XXX_PERIODIC_XXX, this contains the transmit period in milliseconds
	// transmit period can be used
	// (a) if a transmit mapping should be sent out periodically
	// (b) if a request for a receive mapping should be sent out periodically
	// To check if a receive mapping is received periodically please use "timeoutPeriod" in WECANMessageObject
	bool		littleEndian;				//!< Flag to indicate which endianess to use
	bool		useInvalidTransmitPattern;		//!< true: Use invalidTransmitPattern instead of invalid value in transmit pattern
	unsigned int	invalidTransmitPattern;			//!< Pattern replacing value value reported as "invalid" from DEI

	int 		mappingType;				//!< Mapping type (i.e. delimiter, length or mixed mapping)
	unsigned char	fillByte;
};
} // namespace WE

#endif /* WE_CAN_MAPPING_OBJECT_H_ */
