/**
 *	@file we_can_open_state_machine.h
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Header file for WECANProtocolStateMachine class
 *
 */

#ifndef WE_CAN_PROTOCOL_STATE_MACHINE_H_
#define WE_CAN_PROTOCOL_STATE_MACHINE_H_

#include "../../../common/include/we_error.h"

namespace WE
{

//forward declarations
class WECANRaw;
class WETick;
struct WECANMessage;
class WEDEI;

class WECANProtocolResources;
class WECANProtocol;



/**
 *	@enum WECANProtocolStates
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@brief Possible CAN protocol states (considered the ones from CANopen to be generic)
 *
 *	@namespace WE
 */
enum WECANProtocolStates
{
	CAN_PROTOCOL_STATE_NEW,				//!< existing but not initialized
	CAN_PROTOCOL_STATE_INITIALISATION,		//!< Init the stack
	CAN_PROTOCOL_STATE_RESET_COMMUNICATION,		//!< Terminate all TP sessions, reclaim address
	CAN_PROTOCOL_STATE_PRE_OPERATIONAL,		//!< Currently claiming an address
	CAN_PROTOCOL_STATE_OPERATIONAL,			//!< normal operation
	CAN_PROTOCOL_STATE_STOPPED,			//!< be quiet on the bus, don't receive messages, accept messages to be written until queues are full

	CAN_PROTOCOL_STATE_FATAL_ERROR,
	CAN_PROTOCOL_SPECIFIC_STATES			//!< use this as start value for your own additional states
};



/**
 *	@class WECANProtocolStateMachine
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@brief Class for a CAN protocol state machine
 *
 *	@namespace WE
 *
 *	This class represents a CAN protocol state machine.
 *	Don't use this class from outside. It is intended to be used by CAN protocols only.
 *
 *	This class is separated from the protocol stack itself so that the state machine implementation can be
 *	set at runtime. Some protocols such as CANopen leave some freedom on the behaviour of a stack.
 *	Others e.g. J1939 allow proprietary definitions which may need proprietary behaviour i.e. a modified state
 *	machine.
 *	With this separate class one could implement new behaviour while keeping the rest of the implementation.
 *	Additionally one could even switch behaviour at startup e.g. for a service session.
 *	The goal is to keep the stack implementation common for all applications while implementing modified
 *	state machine based on specific requirements.
 *
 *	The criteria deciding whether functionality goes to the stack or the state machine are:
 *	Protocol stack:
 *		- invariant definitions from the standard
 *		- "given" interfaces e.g. to socket CAN
 *		- service methods with predefined functionality
 *	State machine:
 *		- define specific behaviour by calling the service methods of the protocol stack (and possibly reorder these)
 *		- optional functionality
 *		- specific implementation of an alternative allowed by the standard
 *		- proprietary behaviour
 *
 *
 *
 *
 */
class WECANProtocolStateMachine
{
public:




	// ************************************************************************************************
	// ************************ (de)initialization related ********************************************
	// ************************************************************************************************


	/**
	 *	@fn WECANProtocolStateMachine()
	 *
	 *	@brief Constructor
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 */
	WECANProtocolStateMachine();

	/**
	 *	@fn virtual ~WECANProtocolStateMachine()
	 *
	 *	@brief Destructor, frees memory, generally the class reserving memory should also delete it.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 */
	virtual ~WECANProtocolStateMachine();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WECANProtocolStateMachine(const WECANProtocolStateMachine &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WECANProtocolStateMachine & operator=(const WECANProtocolStateMachine &) = delete;

	/**
	 *	@fn virtual int create(WECANProtocolResources * p_res)
	 *
	 *	@brief Create everything that requires memory allocation for this class.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_res pointer to filled resources instance to be used by the stack and this state machine
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if any of the pointers provided in p_res is nullptr
	 *
	 *	This methods creates instances of everything the state machine needs to work.
	 *
	 *	@see release()
	 *
	 */
	virtual int create(WECANProtocolResources * p_res);

	/**
	 *	@fn virtual int init()
	 *
	 *	@brief Initialize members of the state machine to defaults and values provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This method initializes the state machine to defaults and the values provided.
	 *	Default for the current state is CAN_PROTOCOL_STATE_INITIALISATION.
	 *
	 *
	 *	@warning The state machine will only delete memory allocated by itself i.e. care has to be taken to free the pointers provided here.
	 */
	virtual int init();

	/**
	 *	@fn virtual int reset()
	 *
	 *	@brief Initialize members of the state machine to "equivalent to new" state.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	This method cleans values the state machine but does not consider any memory that may be
	 *	allocated (e.g. it will reset pointers to nullptr without calling delete before).
	 *	Current state is set to CAN_PROTOCOL_STATE_NEW.
	 *
	 *
	 */
	virtual int reset();

	/**
	 *	@fn virtual int reset()
	 *
	 *	@brief Delete everything this instance of state machine allocated before.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	This method deletes all instances of classes created by this state machine instance
	 *	- especially those in method create().
	 *
	 *
	 *	@see create()
	 *
	 */
	virtual int release();




	// ************************************************************************************************
	// ************************ state machine related *************************************************
	// ************************************************************************************************


	/**
	 *	@fn virtual int run(unsigned int p_time)
	 *
	 *	@brief Switch state machine until p_time is reached or current work is done.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_time WETick where this method should be finished or interrupt itself.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_TIMEOUT_ERROR in case execution was terminated prematurely because p_time expired
	 *
	 *	This method executes the periodically driven state machine of a stack.
	 *	It should be called frequently via the "process()" method of the CAN stack in order to perform all
	 *	the stack activities NOT driven by (external) events.
	 *	This includes (but is not limited to) e.g.
	 *	- intialization: getting a bus address
	 *	- normal operation: sending messages, checking timeouts,
	 *	- termination: send log off, cleaning up memory
	 *
	 *	This method should NOT be used for activities which could directly be driven by external events
	 *	such as
	 *	- receiving a CAN message
	 *	- receiving an event
	 *
	 *	When implementing this method one should be aware that it blocks the event handling methods.
	 *	So similar to them it should do things quickly and not perform any waiting by itself.
	 *	If longer lasting tasks are implemented they should be interrupted when p_time is reached.
	 *	This requires a design keeping the current state when time is up and continuing when this method
	 *	is called next time.
	 *
	 *	For longer lasting tasks it could be considered starting threads to execute them although this
	 *	requires careful synchronization of shared resources.
	 *
	 *	When this method is not running it is expected that the stack is either
	 *	waiting for events/endpoints or executing handlers for these.
	 *
	 *
	 */
	virtual int run(unsigned int p_time) = 0;


	/**
	 *	@fn virtual int setPause(bool p_pause = true)
	 *
	 *	@brief Pause or un-pause state machine.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_pause true for pause, false for continue
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	If p_pause == true, all incoming CAN messages are ignored and no messages will be sent out.
	 *	Otherwise stack should keep its state and not consider any incoming messages to be timed out
	 *	i.e. timeout periods should be restarted when pause ends.
	 *
	 *	@see setPassive()
	 *
	 */
	virtual int setPause(bool p_pause = true);



	/**
	 *	@fn virtual bool getPause(void)
	 *
	 *	@brief Figure out whether state machine is paused or not.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_pause true for pause, false for continue
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	If p_pause == true, all incoming CAN messages are ignored and no messages will be sent out.
	 *	Otherwise stack should keep its state and not consider any incoming messages to be timed out
	 *	i.e. timeout periods should be restarted when pause ends.
	 *
	 *	@see setPassive()
	 *
	 */
	virtual bool getPause(void) const;




	/**
	 *	@fn int setPassive(bool p_isPassive)
	 *
	 *	@brief Switch this instance to passive (or active).
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_isPassive = true to switch stack to passive mode
	 *	@param p_isPassive = false to switch stack to active mode
	 *
	 *	@returns an error code indicating if setting was accepted
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	The base implementation never fails - but other stacks may have conditions to be
	 *	met before switching between active and passive mode.
	 *	The base stack defines "passive" as "keep quiet on CAN" which could result in timeouts
	 *	of mappings since no requests will be sent any more.
	 *
	 *	In contrast to pause the stack should still receive messages while in "passive" mode.
	 *
	 *
	 *	@see setPause()
	 *
	 */
	virtual int setPassive(bool p_isPassive = true);


	/**
	 *	@fn bool getPassive(void)
	 *
	 *	@brief Returns whether this instance is passive or active.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns true if stack is in passive mode
	 *	@returns false if stack is in active mode
	 *
	 *	@returns an error code indicating if setting was accepted
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	The base implementation never fails - but other stacks may have conditions to be
	 *	met before switching between active and passive mode.
	 *	The base stack defines "passive" as "keep quiet on CAN" which could result in timeouts
	 *	of mappings since no requests will be sent any more.
	 *
	 */
	virtual bool getPassive(void) const;




	// ************************************************************************************************
	// ************************ event related *********************************************************
	// ************************************************************************************************


	/**
	 *	@fn virtual int handleCANmsg(WECANMessage * p_msg)
	 *
	 *	@brief Method to handle and distribute all incoming CAN messages.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg Pointer to the message received which should be evaluated/handled here.
	 *
	 *	@returns WE_SUCCESS if the received message could be handled (or the state machine is paused)
	 *	@returns WE_NOT_SUPPORTED_ERROR if the received message was not intended for this CAN protocol instance
	 *
	 *	This function get's called whenever a CAN message is read from the CAN bus by WECANProtocol::onEndPointReadable().
	 *
	 *
	 *	@warning "p_msg" is only valid in this method - if (parts of) the message should be stored one has to copy them/it
	 *
	 *	@see WECANProtocol::onEndPointReadable()
	 *
	 */
	virtual int handleCANmsg(WECANMessage * p_msg) = 0;

	/**
	 *	@fn int getCurrentState(void)
	 *
	 *	@brief Return the current state of the state machine
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns the current state of the state machine
	 *
	 *	@see enum CANopenNMTStates for available states
	 */
	int getCurrentState(void) const;


protected:



	/**
	 *	@fn void setState(int p_state)
	 *
	 *	@brief Set state of state machine to p_state.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_state state to switch to
	 *
	 *	@returns nothing
	 *
	 *	The main advantage of this method is that it can be used for breakpoints and debug output
	 *	(assuming it is used for all state changes).
	 *
	 *	Additionally it could be used to check the range of the state set or do other plausibility checks.
	 *
	 */
	virtual void setState(int p_state);


	WECANProtocolResources * resources{nullptr};	//!< pointer to data that will not change while stack is running and can be shared with state machine.

	int currentState{CAN_PROTOCOL_STATE_FATAL_ERROR};	//!< The current state of the state machine
	bool isPassive{true};		//!< if true stack should only listen but not send messages
	bool isPaused{true};		//!< if true stack should not receive or send messages
};
} //end of namespace WE

#endif /* WE_CAN_PROTOCOL_STATE_MACHINE_H_ */
