/**
 *	@file we_j1939.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief J1939 stack completely written at WE (in contrast to its predecessor).
 *
 */

#ifndef WE_J1939_H_
#define WE_J1939_H_


#include "../../../os_abstraction/endpoint_manager/include/we_endpoint_handler.h"
#include "../../common/include/we_can_protocol.h"
#include "../../common/include/we_can_message_listener.h"

#include "we_j1939_definitions.h"
#include "we_j1939_resources.h"
#include "we_j1939_message_object.h"

namespace WE
{

class WEJ1939CANMessage;
class WEJ1939StateMachine;

/**
 *	@class WEJ1939
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief This class is the application interface for the new J1939 stack
 *
 *	@namespace WE
 *
 *	This class can be seen as the new J1939 stack.
 *	This should be the only class of all other J1939 classes floating around here
 *	to be used by an application.
 *	It implements the functions of WECANProtocol.
 *
 *	When using this class the following initialization sequence should be kept
 *	(see function documentation below for each mentioned function):
 *	- Create an instance: WEJ1939 * j1939 = new WEJ1939()
 *	- Set pointers and options:
 *	- Before calling init function, you need to create a WE_CAN object, and initialize the baud rate
 *	      and interface name for the CANportNumber that is used by J1939 stack.
 *	- Call create function:
 *	  - j1939->create(dei, portNumber, preferredAddress, ECUname)
 *	- After finishing with this you have to subscribe to the WEEndpointManager:
 *	  - j1939->subscribeToEndpointManager(WEEndPointManager *)
 *	- After this the stack is ready to run
 *	- In your application's main loop or in a thread, you have to call these functions frequently to make the stack run:
 *	  - WEEndpointManager::waitForEndPoints(unsigned int, unsigned int)
 *	  - J1939::process(unsigned int)
 */
class WEJ1939 : public WECANProtocol
{
public:

	// ************************************************************************************************
	// ************************ (de)initialization related ********************************************
	// ************************************************************************************************


	/**
	 *	@fn WEJ1939( std::string p_instanceName)
	 *
	 *	@brief Default constructor. Initializes pointers to nullptr and other members to sane
	 *		start values. Also calls Constructor of WECANProtocol.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_instanceName name this instance should use as a WEVariableOwner
	 *
	 */
	explicit WEJ1939(const std::string & p_instanceName);


	/**
	 *	@fn virtual ~WEJ1939()
	 *
	 *	@brief Virtual destructor. Frees all occupied memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	virtual ~WEJ1939();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEJ1939(const WEJ1939 &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEJ1939 & operator=(const WEJ1939 &) = delete;

	/**
	 *	virtual int create(WEDEI * const p_dei,
	 *			WEVariableEventDispatcher * const p_varEventDispatcher,
	 *			unsigned int p_portNumber,
	 *			WEJ1939StateMachine * p_machine,
	 *			unsigned char p_prefAddress,
	 *			unsigned long long int p_name);
	 *
	 *	@brief Initialize the J1939 stack and allocates needed memory
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_dei Pointer to the DEI instance of this process
	 *	@param p_portNumber The CAN port number which J1939 shall use (starting with 1, not 0)
	 *	@param p_machine State machine to be used to run this stack instance.
	 *	@param p_prefAddress Preferred address to use.
	 *	@param p_name The J1939 ECU name used in address claim to set.
	 *
	 *	@returns WE_SUCCESS if a valid WETick, DEI pointer and a CAN port number are present
	 *	@returns WE_INIT_ERROR if setting the CAN port number or the DEI pointer failed
	 *
	 *	@see int init(void)
	 *
	 *	You can use this create function if you don't want to set the port number and DEI before.
	 *	This function implicitly calls the init() function without parameter.
	 *
	 *	This is the preferred way to initialize the stack. If one has to create the instance before
	 *	p_prefAddress and p_name are available (e.g. on PClient) they should be set to 0.
	 *	In this case they have to be set before the "process()" method is called the first time i.e.
	 *	before the state machine is used.
	 *
	 *
	 *
	 */
	virtual int create(	WEDEI * const p_dei,
				WEVariableEventDispatcher * const p_varEventDispatcher,
				unsigned int p_portNumber,
				WEJ1939StateMachine * p_machine,
				unsigned char p_prefAddress,
				unsigned long long int p_name);

	/**
	 *	@fn virtual int init(void)
	 *
	 *	@brief Initialize all class attributes with meaningfull default values. Does not allocate memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int init(void) override;

	/**
	 *	@fn virtual int reset();
	 *
	 *	@brief Reset all class attributes with default values. Does not free allocated memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int reset() override;

	/**
	 *	@fn virtual int release();
	 *
	 *	@brief Frees allocated memory and sets all class attributes to default values.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int release() override;

	/**
	 *	@fn virtual int clearECUs()
	 *
	 *	@brief Clear all the ecus known to the VT during "ignition off" i.e when the VT is in "low power" mode.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@return WE_SUCCESS
	 *
	 */
	virtual int clearECUs();

	/**
	 *	@fn int createResources(
				WEDEI * p_dei,
				WEVariableEventDispatcher * p_varEventDispatcher,
				unsigned int p_portNo,
				unsigned char p_prefAddress,
				unsigned long long int p_name,
				WEJ1939Resources ** p_resources
			)
	 *
	 *
	 *	@brief Set resources for stack (and possibly state machine) in case you don't want to use the init method.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_dei pointer to DEI instance to be used
	 *	@param p_varEventDispatcher pointer to dispatcher providing events from p_dei
	 *	@param p_portNo no of CAN port to be used (starting by 1)
	 *	@param p_prefAddress The preferred J1939 address
	 *	@param p_name The J1939 name of the ECU.
	 *	@param p_resources Returns the pointer of the created resources
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if any of the pointers provided is nullptr or p_portNo is 0
	 *
	 *	This method checks the content and creates a new instance of WECANProtocolResources for the stack to use.
	 *
	 *	@warning The stack will delete both of the instances provided when it is destroyed itself.
	 */
	int createResources(
				WEDEI * p_dei,
				WEVariableEventDispatcher * p_varEventDispatcher,
				unsigned int p_portNo,
				unsigned char p_prefAddress,
				unsigned long long int p_name,
				WEJ1939Resources ** p_resources
			);

	// ************************************************************************************************
	// ************************ address claim related *************************************************
	// ************************************************************************************************


	/**
	 *	@fn unsigned char getPreferredAddress(void)
	 *
	 *	@brief This function delivers the current used preferred address.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns the preferred address
	 */
	unsigned char getPreferredAddress(void) const;

	/**
	 *	@fn int setPreferredAddress(unsigned char p_prefAddress)
	 *
	 *	@brief This function sets the preferred address.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_prefAddress The preferred address to set.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_GENERAL_ERROR if this->resources is nullptr
	 *	@returns WE_GENERAL_ERROR if this->resources->addressHandler is nullptr
	 *
	 *	This method internally calls WEJ1939AddressHandler::setPreferredAddress()
	 *
	 *	@warning Preferred address is only used if set before method process() is called the first time.
	 *
	 *	@see WEJ1939AddressHandler::setPreferredAddress()
	 *
	 */
	int setPreferredAddress(unsigned char p_prefAddress);


	/**
	 *	@fn unsigned char getOwnAddress(void)
	 *
	 *	@brief This function delivers the address claimed by this stack instance.
	 *
	 *	@returns The address currently used by this stack instance.
	 */
	unsigned char getOwnAddress(void) const;

	/**
	 *	@fn unsigned long long int getOwnECUName(void)
	 *
	 *	@brief This function delivers the ECU name used by this stack instance to claim its own address.
	 *	@returns the ECU name.
	 */
	unsigned long long int getOwnECUName(void) const;


	/**
	 *	@fn int setOwnECUName(unsigned long long int p_j1939Name)
	 *
	 *
	 *	@brief Set J1939 name to be used by this stack instance for address claiming.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_j1939Name J1939 name to be used by this stack instance for address claiming.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_GENERAL_ERROR if this->resources is nullptr
	 *
	 *
	 *	@warning Own name is only used if set before method process() is called the first time.
	 */
	int setOwnECUName(unsigned long long int p_j1939Name);
	int setIsProjektConfigured(bool p_isProjektConfigured);

	/**
	 *	@fn int addEcuInfo(WEJ1939CanEcu * p_ecu)
	 *
	 *
	 *	@brief Add ECU definition so that we can identify sender or receiver of PGNs as this ECU.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_ecu Pointer to ECU description to add.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR
	 *	@returns
	 *
	 *	This method checks the content and creates a new instance of WEJ1939CanEcu in the stack's address handler.
	 *
	 */
	int addEcuInfo(WEJ1939CanEcu * p_ecu);
	int addEcuInfo(int p_id, const std::string & p_name, unsigned long long p_ecuName, unsigned long long p_ecuMask, unsigned int p_address, bool p_addressIsStatic);

	/**
	 *	@fn WEJ1939CanEcu * getEcuInfoByName(unsigned long long p_name)
	 *
	 *	@brief Get pointer to ECU Data identified by J1939 ECU name
	 *
	 *	@author mohd.faisa@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@param p_name The J1939 ECU name used to identify the ECU data.
	 *
	 *	@return Pointer to the ECU data identified by the J1939 ECU name.
	 */
	WEJ1939CanEcu * getEcuInfoByName(unsigned long long p_name) const;

	/**
	 *	@fn bool isOnlySentToMe(WEJ1939CANMessage * p_msg)
	 *
	 *
	 *	@brief Return true if p_msg is NOT a broadcast and is addressed to this stack instance.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg Pointer to massage to check.
	 *
	 *	@returns true if message is sent to this stack only
	 *	@returns false if message is a broadcast or sent to another estination
	 *
	 *	This method can be used in order to decide if a message requires a NACK.
	 *
	 */
	bool isOnlySentToMe(WEJ1939CANMessage * const p_msg) const;

	/**
	 *	@fn bool hasValidAddress(void) const
	 *
	 *	@brief This function checks if a valid address is claimed.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *
	 *	@returns true if address is claimed and messages can be send.
	 *	@returns false if the no address is claimed.
	 */
	bool hasValidAddress() const;


	// ************************************************************************************************
	// ************************ variable mapping related **********************************************
	// ************************************************************************************************


	/**
	 *	@fn int addMappingMessage(WEJ1939MessageObject * p_object)
	 *
	 *	@brief Add (a static copy of) p_object to this stack.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_object	Mapping message object to be copied into this stack instance.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_ALREADY_EXISTS_ERROR if a mapping message with the same ID is already stored.
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_object does not contain a mapping object
	 *
	 *	This method adds the mapping message p_object to the stack by creating a copy of it.
	 *	This copy does NOT contain the mapping data which the mapping object attached to p_object
	 *	may have.
	 *
	 *	This method tries to add a reference to the owner ECU based on the ECU ID coming with p_object.
	 *	This will only succeed if the stack already knows the ECUs. If not (e.g. in PClient) call
	 *	assignECUsToMappings() whenever all ECUs are added.
	 *
	 *	The mapping data has to be added separately using addMappingDataToID().
	 *
	 *	@see addMappingDataToID( unsigned int p_ID, WECANMappingData * p_data )
	 *	@see int assignECUsToMappings(void)
	 *	@see int addEcuInfo(WEJ1939CanEcu * p_ecu)
	 *
	 */
	int addMappingMessage(WEJ1939MessageObject * p_object);


	/**
	 *	@fn int addMappingDataToID( unsigned int p_ID, WECANMappingData * p_data, bool p_isTriggerOnly = false )
	 *
	 *	@brief Add (a copy of) p_data to this stack's mapping message identified by p_ID.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_ID	ID of the mapping message object which should get the new mapping data.
	 *	@param p_data	Pointer to mapping data which will be copied into the mapping object of p_ID
	 *	@param p_isTriggerOnly Don't map into CAN message but consider as trigger to send the mapping
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INIT_ERROR if p_data is nullptr
	 *	@returns WE_NOT_FOUND_ERROR if a mapping message object with ID p_ID could not be found
	 *	@returns WE_INVALID_PARAMETER_ERROR if the mapping message object with ID p_ID did not contain a mapping object
	 *
	 *	This method adds the mapping data described by p_data to the mapping object attached to
	 *	the mapping message identified by p_ID.
	 *	Do do so it creates a copy of p_data.
	 *
	 *	The mapping message has to be added to the stack before by using addMappingMessage().
	 *
	 *	@see addMappingMessage(WEJ1939MessageObject * p_object)
	 *
	 *	This method (interface) could also be part of the common stack. Even though it would have to be overwritten to maintain protocol specific data.
	 *
	 */
	int addMappingDataToID( unsigned int p_ID, WECANMappingData * p_data, bool p_isTriggerOnly = false );


	/**
	 *	@fn int assignECUsToMappings(void)
	 *
	 *	@brief Assign references to ECU descriptions to all mapping message objects of this stack.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_NOT_FOUND_ERROR if at least one mapping message was left without an ECU reference (i.e. ECU not found by ID)
	 *
	 *	This method adds connections between mapping message objects and the "owner" ECUs inside the stack.
	 *	It should be called if ECUs were not defined while mapping messages were added to the stack.
	 *
	 *	The mapping messages have to be added to the stack before by calling addMappingMessage().
	 *	The ECU descriptions have to be added to the stack before by using addEcuInfo()
	 *
	 *	@see addMappingMessage(WEJ1939MessageObject * p_object)
	 *	@see int addEcuInfo(WEJ1939CanEcu * p_ecu)
	 *
	 *	@warning: Currently only receive mappings are handled.
	 *	Somehow we got away since 2011 without doing the same for transmit mappings
	 *
	 */
	int assignECUsToMappings(void);


	/**
	 *	@fn int handleReceiveMappings( WEJ1939CANMessage & p_j1939msg, unsigned int p_timeNow )
	 *
	 *	@brief Check message against mappings and transfer values to DEI if one or more mapping(s) fit(s).
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_j1939msg	Message to be checked.
	 *	@param p_timeNow	Time to use for timestamp in case message fits a mapping
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INIT_ERROR if p_data is nullptr
	 *	@returns WE_NOT_FOUND_ERROR if a mapping message object could not be found for the message provided
	 *
	 *	This method checks if the message provided fits any mapping stored and if it does it forwards the
	 *	value using WECANMapping::extractMappedValues().
	 *	In this case the timestamp of the respective mapping message object is set to p_timeNow.
	 *
	 *	The mapping message has to be added to the stack before by calling addMappingMessage().
	 *
	 *	@see WECANMapping::extractMappedValues()
	 *
	 *
	 */
	int handleReceiveMappings( WEJ1939CANMessage & p_j1939msg, unsigned int p_timeNow );


	/**
	 *	@fn int handleRequestsReceived( WEJ1939CANMessage & p_j1939msg, unsigned int p_timeNow )
	 *
	 *	@brief Check if we have a transmit mapping to fit the request and send it out.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_j1939msg	Message to be checked if it contains a request.
	 *	@param p_timeNow	Time to use for timestamp in case request fits a mapping
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_NOT_FOUND_ERROR if a mapping message object could not be found for the message provided
	 *
	 *	This method checks if the message provided is a standard or proprietary request for one of our
	 *	transmit mappings. If one of the transmit mapping messages fits it is sent out.
	 *	If it is sent to the configured destination the timestamp of the respective mapping message object is set to p_timeNow.
	 *
	 *	The mapping message has to be added to the stack before by calling addMappingMessage().
	 *
	 *
	 *
	 */
	int handleRequestsReceived( WEJ1939CANMessage & p_j1939msg, unsigned int p_timeNow );



	/**
	 *	@fn int handleReceiveTimeouts(unsigned int p_timeNow)
	 *
	 *	@brief Check receive messages timestamps and trigger handlers if they timed out.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_timeNow	Current time to use for timeout calculation
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	This method checks all the receive mappings with a timeout period defined if they have timed out.
	 *	For all those that have time out the mapped variables are set to "invalid" in DEI and handlers
	 *	for the event are triggered.
	 *
	 *
	 *
	 *
	 */
	int handleReceiveTimeouts(unsigned int p_timeNow);


	/**
	 *	@fn int sendPeriodicTransmitMappings(unsigned int p_time)
	 *
	 *	@brief Send out transmit mappings with repeat rate.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_time - current time tick in ms
	 *
	 *
	 *
	 *	@see
	 *
	 */
	int sendPeriodicTransmitMappings(unsigned int p_time);


	/**
	 *	@fn int sendPeriodicRequests(unsigned int p_time)
	 *
	 *	@brief Send out requests with repeat rate.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_time - current time tick in ms
	 *
	 *
	 *
	 *	@see
	 *
	 */
	int sendPeriodicRequests(unsigned int p_time);

	/**
	 *	@fn int sendRequestForPGN(unsigned int p_pgn, WEJ1939CanEcu * p_destEcu)
	 *
	 *	@brief Send out a standard request for the specified PGN to the specified destination ECU.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@param p_pgn		PGN for which the request needs to be sent
	 *	@param p_destEcu	Destination ECU to which the request needs to be sent
	 *
	 */
	int sendRequestForPGN(unsigned int p_pgn, WEJ1939CanEcu * p_destEcu);


	// ************************************************************************************************
	// ************************ variable owner related ************************************************
	// ************************************************************************************************


	/**
	 *	@fn virtual int updateRequest(unsigned int p_varKey, int p_intValue)
	 *
	 *	@brief Request a value change for a variable owned by this WEVariableOwner (=> this J1939 stack)
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_varKey The combined Index/Subindex of a variable owned by this stack instance
	 *	@param p_intValue The value that shall be set
	 *
	 *	@returns WE_NOT_FOUND_ERROR if no mapping could be found to be used as write request for the variable
	 *	@returns WE_SUCCESS if at least one write request was sent
	 *
	 *	This method checks if there are transmit mappings containing the variable and if they are marked
	 *	to be used as write requests also.
	 *	All those mappings where both conditions are met will be sent out with the mapped variable value
	 *	set to p_intValue.
	 *
	 *	@see WEJ1939MappingObject::getTransmitType()
	 *
	 */
	virtual int updateRequest(unsigned int p_varKey, int p_intValue) override;


	/**
	 *	@fn virtual int updateRequest(unsigned int p_varKey, const std::string & p_stringValue)
	 *
	 *	@brief Request a value change for a variable owned by this WEVariableOwner (=> this J1939 stack)
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_varKey The combined Index/Subindex of a variable owned by this stack instance
	 *	@param p_stringValue Reference to the value that shall be set
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_NOT_FOUND_ERROR in case p_varKey couldn't be found in any transmit PGN
	 *
	 *	If sending one request should fail 'WE_SUCCESS' is returned and an error message is written to std::cout.
	 */
	virtual int updateRequest(unsigned int p_varKey, const std::string & p_stringValue) override;

	/**
	 *	@fn virtual int getValueRequest(unsigned int p_varKey)
	 *
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_varKey The combined Index/Subindex of a variable e
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR
	 *
	 *	@ warning Function not implemented for J1939
	 */
	virtual int getValueRequest(unsigned int p_varKey) override;




	// ************************************************************************************************
	// ************************ variable subscriber related *******************************************
	// ************************************************************************************************


	/**
	 *	@brief Called when a variable has changed.
	 *
	 *	@param p_signal WESignal containing information about the changed variable value
	 *
	 *	@returns WE_SUCCESS in case of success (even if event is NOT a variable change event (and therefore ignored)
	 *	@returns WE_NOT_FOUND_ERROR If we don't know the variable
	 *
	 *	If sending one transmit PGN or request should fail 'WE_SUCCESS' is returned and an error message is written to std::cout.
	 *	Error message is only written though if message causing the issue or error code changed in order to avoid
	 *	permanent repetition of the same error report.
	 *
	 *	@see WESignalListener
	 *
	 */
	int signalOccurred(WESignal & p_signal) override;

	/**
	 *	@brief The function registers an event handler for transport protocol events.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_eventListener The listener to call when the event occurs.
	 *
	 *	@returns WE_SUCCESS.
	 *
	 *	For unregistering the event Listener set p_eventListener to nullptr.
	 *	@WARNING: Only one listener can be registered.
	 *
	 *	The event send will have following content:
	 *	  eventType = J1939_TP_MESSAGE_SEND
	 *	  eventParameter1 = PGN of the message
	 *	  eventParameter2 = error code
	 *	  ep3.timestampMilliseconds = tick when the event occurred
	 *	Following error codes are supported:
	 *		WE_SUCCESS: Message was send successful
	 *		WE_OPERATION_IN_PROGRESS: The other ECU did not accept the message, because it is already busy with another TP message
	 *		WE_TIMEOUT_ERROR: A timeout has occurred.
	 *		WE_STATE_NOT_SUPPORTED_ERROR: the other ECU does not hold the resources needed for receiving the message.
	 *		WE_INVALID_MESSAGE_RECEIVED_ERROR: This ECU is already sending a TP message and is requested to send another one at the same time. WARNING: in this situation, sending both messages will be aborted.
	 */
	int registerTPEventListener(WESignalListener* p_tpEventListener);




	// ************************************************************************************************
	// ************************ CAN port related ******************************************************
	// ************************************************************************************************


	/**
	 *	@fn int sendMessage(WEJ1939CANMessage *p_msg)
	 *
	 *	@brief Sends the message using J1939 stack and protocol.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_msg Pointer to the message to send
	 *
	 *	@returns WE_SUCCESS if message was sent successfully.
	 *	@returns WE_SEND_ERROR if internal resources are missing
	 *	@returns WE_NOT_SUPPORTED_ERROR if we don't have a valid address (yet)
	 *	@returns Any error code WECANRaw::send() or WEJ1939TransportProtocolHandler::sendTPMessage() may return.
	 *
	 *	This function is used for sending p_msg via J1939 stack.
	 *
	 *	@see WECANRaw::send()
	 *	@see WEJ1939TransportProtocolHandler::sendTPMessage()
	 */
	int sendMessage(WEJ1939CANMessage *p_msg);




	// ************************************************************************************************
	// ************************ protocol utility related **********************************************
	// ************************************************************************************************


	/**
	 *	@fn 	int WEJ1939::fillACK(
	 *			WEJ1939CANMessage * p_msg,
	 *			unsigned char p_destAddress,
	 *			unsigned char p_controlByte,
	 *			unsigned char p_groupFunction,
	 *			unsigned int p_pgn,
	 *			bool p_setDestAddress = false )
	 *
	 *
	 *	@brief Request a value change for a variable owned by this WEVariableOwner (=> this J1939 stack)
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg Message to be filled with ACK data
	 *	@param p_destAddress Address to be used as "Address Acknowledged"
	 *	@param p_controlByte Fill with value defined in enum J1939AckControlByteValues
	 *	@param p_groupFunction Could not find proper values for this yet - suggest to add them in enum J1939AckGroupFunctionByteValues
	 *	@param p_pgn PGN to ACK
	 *	@param p_setDestAddress Default is "false" as required by J1939-21 (2006), ISO 11783-03 (2007) suggests "true"
	 *
	 *	@returns WE_NOT_FOUND_ERROR if no mapping could be found to be used as write request for the variable
	 *	@returns WE_SUCCESS in case of success
	 *	@returns Any error code the methods of WEJ1939CANMessage may return
	 *
	 *	This method fills "p_msg" with an acknowledgment message as defined in J1939-21 (2006) and ISO 11783-03 (2007).
	 *	Parameter p_setDestAddress defines if parameter p_destAddress is only used to define "Address Acknowledged" (false, J1939 definition)
	 *	or if it is additionally used as destination address in the CAN ID (true, ISO 11783-21 definition).
	 *
	 *	@see enum J1939AckControlByteValues
	 *	@see enum J1939AckGroupFunctionByteValues
	 *
	 */
	static int fillACK(
			WEJ1939CANMessage * p_msg,
			unsigned char p_destAddress,
			unsigned char p_controlByte,
			unsigned char p_groupFunction,
			unsigned int p_pgn,
			bool p_setDestAddress = false);


private:

	/**
	 *	@struct pgnListener
	 *	@brief Struct keeping information to identify a listener entry in pgn2ReceiveListener.
	 */
	struct PGNListener
	{
		PGNListener(WECANMessageListener * p_listener, WEJ1939CanEcu * p_originator)
		{
			this->listener = p_listener;
			this->originator = p_originator;
		}
		WECANMessageListener * listener;
		WEJ1939CanEcu * originator;
	};

public:



	/**
	 *	@fn int addPgnListener(unsigned int p_pgn, WECANMessageListener * p_listener, WEJ1939CanEcu * p_ecu)
	 *
	 *
	 *	@brief Request to be called if a certain PGN is received.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_pgn PGN to be received
	 *	@param p_listener Class instance to be called if p_pgn arrives
	 *	@param p_ecu Originator ECU of p_pgn - nullptr if p_pgn of any ECU should be forwarded
	 *
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_listener is nullptr
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_pgn consists of more than 18 bits
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	This method adds p_listener to the map of listeners as the one for p_pgn.
	 *	Set p_ecu to nullptr if the originator address is not relevant.
	 *	Set p_ecu to an ECU instance known by the stack if only messages sent by a certain ECU are of interest.
	 *
	 *	@see
	 *
	 */
	int addPgnListener(unsigned int p_pgn, WECANMessageListener * p_listener, WEJ1939CanEcu * p_ecu = nullptr);


	/**
	 *	@fn int handleReceiveListeners( WEJ1939CANMessage & p_j1939msg )
	 *
	 *	@brief Check message against listeners and call them if appropriate.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_j1939msg	Message to be checked.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_NOT_FOUND_ERROR if a listener could not be found for the message provided
	 *
	 *	This method checks if the message provided fits any listener registration and calls the listener
	 *	if p_j1939msg fits its registration.
	 *
	 *
	 *	@see int addPgnListener(unsigned int p_pgn, WECANMessageListener * p_listener, WEJ1939CanEcu * p_ecu)
	 *
	 */
	int handleReceiveListeners( WEJ1939CANMessage & p_j1939msg );

	virtual int enableMessage(unsigned int p_objectID, bool p_enable) override;

	static int debugSection;


protected:




	// ************************************************************************************************
	// ************************ (de)initialization related ********************************************
	// ************************************************************************************************

	WEJ1939Resources * resourcesJ1939{nullptr};	//!< pointer to data that will not change while stack is running and can be shared with state machine.

	// Create data structures as suggested in WECANProtocol:

	// one map keeping all mapping messages accessible by their IDs
	std::unordered_map< unsigned int, WEJ1939MessageObject * >		id2Pgn{};

	// (1) (masked) CAN ID -> receive message description (one list/map per ID mask; to forward new values)
	// --> for J1939 we have to store the PGN rather than the CAN ID - and we may expect the same PGN from multiple ECUs
	// When populating this map we could also set ID filters for the CAN socket - assuming the socket is exclusively used for this stack instance
	std::unordered_multimap< unsigned int, WEJ1939MessageObject * >		pgn2ReceivePgn{};

	std::unordered_multimap< unsigned int, PGNListener >			pgn2ReceiveListener{};


	// (1a) PGN --> transmit mapping PGN to answer requests quickly
	std::unordered_multimap< unsigned int, WEJ1939MessageObject * >		pgn2TransmitPgn{};

	// (2) Variable key -> transmit message description (to satisfy "updateRequest" and variable change event)
	std::unordered_multimap< unsigned int, WEJ1939MessageObject * >		varKey2TransmitPgn{};

	// (3) list of receive message descriptions with timeout (to set variables to "invalid") if message expires
	std::set < WEJ1939MessageObject * >					receivePgnsWithTimeout{};

	// (4) list of transmit messages with repeat time (to send them out in time)
	std::set < WEJ1939MessageObject * >					transmitPgnsWithPeriod{};

	// (4a) list of transmit messages with proprietary requests (to check them quickly)
	std::set < WEJ1939MessageObject * >					transmitPgnsWithRequests{};

	// (4b) list of transmit messages with repeat time (to send them out in time)
	std::set < WEJ1939MessageObject * >					receivePgnsWithPeriodicRequest{};

	// (5) list of transmit sequences (to send out their messages in time)
	// --> not supported for J1939 (yet)

	// (6) list of receive messages with request messages which we have if specific variable changed
	std::unordered_multimap <unsigned int, WEJ1939MessageObject * >		varKey2ReceiveMappingPGN{};

	int errorObjectID{-1};	// Replace 'static' variable in 'triggerOccurred()' - make sure we only print error message if error changed
	int errorReturnValue{WE_GENERAL_ERROR};	// Replace 'static' variable in 'triggerOccurred()' - make sure we only print error message if error changed

	int errorObjectID2{-1};	// Replace 'static' variable in 'triggerOccurred()' - make sure we only print error message if error changed
	int errorReturnValue2{WE_GENERAL_ERROR};	// Replace 'static' variable in 'triggerOccurred()' - make sure we only print error message if error changed
};
}

#endif /* WE_J1939_H_ */
