/**
 *	@file we_j1939_address_handler.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief This file contains definitions of types, error codes, constants, etc.
 *
 */

#ifndef WE_J1939_ADDRESS_HANDLER_H_
#define WE_J1939_ADDRESS_HANDLER_H_

#include <string>
#include <map>

#include "we_j1939_can_message.h"
#include "we_j1939_can_ecu.h"

namespace WE
{

class WEJ1939;
class WECANRaw;
class WETick;
class WEJ1939Resources;

/**
 *	@class WEJ1939AddressHandler
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief This class provides the address claiming for J1939 stack.
 *
 *	@namespace WE
 *
 *	This class controls the address claiming of the new J1939 stack.
 */
class WEJ1939AddressHandler
{
public:
	/**
	 *	@fn WEJ1939AddressHandler()
	 *
	 *	@brief Default constructor. Initializes pointers to nullptr and other members to sane
	 *		start values.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 */
	WEJ1939AddressHandler();

	/**
	 *	@fn virtual ~WEJ1939AddressHandler()
	 *
	 *	@brief Virtual destructor. Frees all occupied memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	virtual ~WEJ1939AddressHandler();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEJ1939AddressHandler(const WEJ1939AddressHandler &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEJ1939AddressHandler & operator=(const WEJ1939AddressHandler &) = delete;

	/**
	 *	@fn virtual int create(WEJ1939Resources * p_resources)
	 *
	 *	@brief Initialize the J1939 address handler and allocated needed memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_resources pointer to J1939 stack resources
	 *
	 *	@returns WE_SUCCESS
	 */
	virtual int create(WEJ1939Resources * p_resources);

	/**
	 *	@fn virtual int init();
	 *
	 *	@brief Initialize all class attributes with meaningful default values. Does not allocate memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int init();

	/**
	 *	@fn virtual int reset( bool p_lowPower = false )
	 *
	 *	@brief Reset all class attributes with default values. Does not free allocated memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_lowPower if the release is being done due to lowPower on the VT then we should keep the WEJ1939Resources data member
	 *	as it is and not modify our initial address which was updated to our successfully claimed address so that the we can re-claim
	 *	the same address during power on.
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int reset( bool p_lowPower = false );

	/**
	 *	@fn virtual int release( bool p_lowPower = false )
	 *
	 *	@brief Frees allocated memory and sets all class attributes to default values.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_lowPower if the release is being done due to lowPower on the VT then we should keep the WEJ1939Resources data member
	 *	as it is and not modify our initial address which was updated to our successfully claimed address so that the we can re-claim
	 *	the same address during power on.
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int release( bool p_lowPower = false );

	/**
	 *	@fn int handleMessage(WEJ1939CANMessage &p_msg)
	 *
	 *	@brief Method to handle and distribute all incoming J1939 CAN messages.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_msg Pointer to the J1939 message received which should be evaluated/handled here.
	 *
	 *	@returns WE_SUCCESS if the received message could be handled (or the state machine is paused)
	 *	@returns WE_NOT_FOUND_ERROR if the p_msg does not belong to address claiming i.e. was not handled here
	 *	@returns WE_NOT_SUPPORTED_ERROR if the received message was not intended for J1939 instance
	 *
	 *	This function get's called whenever a J1939 CAN message is read from the CAN bus by WEJ1939StateMAchine::handleCANmsg()
	 *
	 *
	 *	@warning "p_msg" is only valid in this method - if (parts of) the message should be stored one has to copy them/it
	 *
	 *	@see WEJ1939StateMachine::handleCANmsg()
	 *
	 */
	int handleMessage(WEJ1939CANMessage &p_msg);

	/**
	 *	@fn int run(unsigned int p_time)
	 *
	 *	@brief The run function needs to be called frequently to process all address claim tasks
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_time The tick count until which this function is allowed to run
	 *
	 *	@returns If everything went OK, WE_SUCCESS is returned. Everything else means error but is hard to define at the moment
	 *
	 *	This functions needs to be called frequently so that all address claiming tasks of the J1939 stack
	 *	can be triggered at the right time.
	 */
	int run(unsigned int p_time);


	/**
	 *	@fn int setOwnName( unsigned long long int p_j1939Name )
	 *
	 *
	 *	@brief Set j1939 name to be used by this stack instance for address claiming.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_addr J1939 address to be used by this stack instance for address claiming.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_GENERAL_ERROR if this->resources is nullptr
	 *
	 *
	 *	@warning This method is a workaround - preferred method to set the J1939 name is "create()".
	 *
	 *	@see create()
	 *
	 */
	int setOwnName( unsigned long long int p_j1939Name );
	int setIsProjektConfigured(bool p_isProjektConfigured);

	/**
	 *	@fn int setPreferredAddress( unsigned char p_addr)
	 *
	 *
	 *	@brief Set preferred address to be used by this stack instance for address claiming.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_addr J1939 address to be used by this stack instance for address claiming.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_GENERAL_ERROR if this->resources is nullptr
	 *
	 *
	 *	@warning Preferred address only has an effect in static address claiming or when no initial address is stored.
	 *	@warning J1939 name has to be set before in order to set dynamic/static address claiming mode.
	 */
	int setPreferredAddress( unsigned char p_addr);


	/**
	 *	@fn bool hasValidAddress(unsigned char * p_address = nullptr) const
	 *
	 *	@brief This function checks if a valid address is claimed - i.e. contradicted our address claim.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_address Optional pointer to a variable that would get the current address.
	 *
	 *	@returns true if valid address is claimed.
	 *	@returns false if the no address is claimed or claimed address has not become valid yet.
	 *
	 *	The address is considered valid if nobody contradicted for 250 ms after address claim.
	 *
	 *	@see hasAddress(unsigned char *)
	 */
	bool hasValidAddress(unsigned char * p_address = nullptr) const;


	/**
	 *	@fn bool hasValidAddress(unsigned char * p_address = nullptr) const
	 *
	 *	@brief This function checks if an address is claimed.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_address Optional pointer to a variable that would get the current address.
	 *
	 *	@returns true if address is claimed. Messages should only be sent if hasValidAddress() returns true.
	 *	@returns false if the no address is claimed.
	 *
	 *	@see hasValidAddress(unsigned char *)
	 */
	bool hasAddress(unsigned char * p_address = nullptr) const;


	/**
	 *	@fn unsigned char getAddress() const
	 *
	 *	@brief This function delivers the current J1939 address
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	unsigned char getAddress() const;

	/**
	 *	@fn struct WEJ1939CanEcu * const getECUDataByID(int p_id) const
	 *
	 *	@brief Get pointer to ECU Data identified by ID
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_id The ECU address used to identify the ECU data.
	 *
	 *	@return Pointer to the ECU data identified by the ECU ID.
	 */
	WEJ1939CanEcu * getECUDataByID(int p_id) const;

	/**
	 *	@fn struct WEJ1939CanEcu * const getECUDataByAddress(unsigned char p_addr) const
	 *
	 *	@brief Get pointer to ECU Data identified by ECU address
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_addr The ECU address used to identify the ECU data.
	 *
	 *	@return Pointer to the ECU data identified by the ECU address addr.
	 */
	WEJ1939CanEcu * getECUDataByAddress(unsigned char p_addr) const;

	/**
	 *	@fn struct WEJ1939CanEcu * getECUDataByName(unsigned long long p_name) const
	 *
	 *	@brief Get pointer to ECU Data identified by J1939 ECU name
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_name The J1939 ECU name used to identify the ECU data.
	 *
	 *	@return Pointer to the ECU data identified by the J1939 ECU name.
	 */
	WEJ1939CanEcu * getECUDataByName(unsigned long long p_name) const;


	/**
	 *	@fn WEJ1939CanEcu * getECUDataByMaskedName(unsigned long long p_name, unsigned long long p_mask) const
	 *
	 *	@brief Get pointer to ECU Data identified by J1939 ECU name and mask
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_name The J1939 ECU name used to identify the ECU data.
	 *	@param p_mask The J1939 ECU mask used to identify the ECU data.
	 *
	 *	@return Pointer to the ECU data identified by the J1939 ECU name and mask.
	 *	@return nullptr if no ECU was found
	 */
	WEJ1939CanEcu * getECUDataByMaskedName(unsigned long long p_name, unsigned long long p_mask) const;


	/**
	 *	@fn WEJ1939CanEcu * getECUDataByMaskedName(unsigned long long p_name) const
	 *
	 *	@brief Get pointer to ECU Data identified by J1939 ECU name and mask of stored ECU
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_name The J1939 ECU name used to identify the ECU data.
	 *
	 *	@returns Pointer to the ECU data identified by the J1939 ECU name and mask.
	 *	@returns nullptr if no ECU was found
	 *
	 *	This method checks the name provided against all the ECUs stored.
	 *	The ECU returned is the first one found where (0 != storedECU->j1939nameMask) AND
	 *	(storedECU->j1939name & storedECU->j1939NameMask) == (p_name & storedECU->j1939NameMask)
	 *	are both "true".
	 *
	 *	ECUs with (0 == storedECU->j1939nameMask) are those who should only be identified by their
	 *	static address or their unmasked name.
	 *
	 *	@see getECUDataByName(unsigned long long p_name)
	 */
	WEJ1939CanEcu * getECUDataByMaskedName(unsigned long long p_name) const;


	/**
	 *	@fn int createECU (
			const unsigned int p_ID,
			const std::string & p_description,
			const unsigned long long & p_name,
			const unsigned long long & p_nameMask,
			bool p_hasStaticAddress,
			unsigned int p_staticAddress = J1939_NULL_ADDRESS)
	 *
	 * *	@brief Adds ECU data to the ECU list.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_ID Mappings initially identify the ECU by ID provided by Projektor.
	 *	@param p_description Description of the ECU (mainly for debugging).
	 *	@param p_name J1939 name of the ECU.
	 *	@param p_name J1939 name mask of the ECU.
	 *	@param p_staticAddress in case we should run on a network with static addresses we may accept messages even without address claim
	 *
	 *	@return WE_SUCCESS.
	 *
	 *	@WARNING creating an ECU that already exists (name is already in nameMap),
	 *		the second call to create will overwrite the data of the existing ECU data.
	 *
	 *	This reports an error to std::cout in case the static address for the ECU should already be occupied.
	 */
	int createECU (
			const unsigned int p_ID,
			const std::string & p_description,
			const unsigned long long & p_name,
			const unsigned long long & p_nameMask,
			bool p_hasStaticAddress,
			unsigned int p_staticAddress = J1939_NULL_ADDRESS);

protected:
	/**
	 *	@fn void setState(int p_state)
	 *
	 *	@brief Set the current state of the address claim state machine
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_state State to set (see J1939AddressClaimStates)
	 *
	 *	This function could even be private - since 2011 we didn't see any reason to call it externally.
	 *	To pause the machine or use the auto start feature use the appropriate methods.
	 */
	void setState(int p_state);

	/**
	 *	@fn int sendRequestForAddressClaimed(unsigned char p_srcAddr, unsigned char p_destAddr)
	 *
	 *	@brief Sends a 'Request for Address Claimed' message to destination destAddr.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_srcAddr The source address to send the message from.
	 *	@param p_destAddr The destination address to send the message to.
	 *
	 *	@return Returns the result of this->canPort->send(&msg)
	 *
	 *	@see WECANRaw
	 *
	 *	This function triggers sending of a 'Request for Address Claimed' message.
	 */
	int sendRequestForAddressClaimed(unsigned char p_srcAddr, unsigned char p_destAddr);

	/**
	 *	@fn int sendAddressClaimed(unsigned char p_srcAddr)
	 *
	 *	@brief Sends a 'Address Claimed' message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_srcAddr The source address to claim.
	 *	@param p_destAddr The destination address for this (broadcast by default)
	 *
	 *	@return Returns the result of this->canPort->send(&msg)
	 *
	 *	@see WECANRaw
	 *
	 *	This function triggers sending of a 'Address Claimed' message that claims
	 *	the address p_srcAddr.
	 */
	int sendAddressClaimed(unsigned char p_srcAddr, unsigned char p_destAddr = J1939_BROADCAST_ADDRESS);

	/**
	 *	@fn int sendCannotClaimSourceAddress(void)
	 *
	 *	@brief Sends a 'Cannot Claim Source Address' message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return Returns the result of this->canPort->send(&msg)
	 *
	 *	@see WECANRaw
	 *
	 *	This function triggers sending of a 'Cannot Claim Source Address' message.
	 */
	int sendCannotClaimSourceAddress(void);

	/**
	 *	@brief Need to send a DM1 in case of address spoofing
	 *	@param p_spn
	 *	@param p_fmi
	 *	@param p_occurenceCount default == 0x7f == max value, also used for 'not supported'
	 *	@return error code provided by the final 'send(WEJ1939CANMessage * p_msg)'
	 *
	 *	Assemble and send DM1 (one single active trouble code) according to
	 *	ISO 11783-12:2019, B.6 Active diagnostic trouble codes (DM1)
	 *
	 *	@see send(WEJ1939CANMessage * p_msg)
	 *
	 *	@warning This is a very rudimentary DM1 support consisting of this one method only.
	 *	I.e. we will still NACK any request for active trouble codes.
	 *	And the message would belong in a separate class once we really support DTCs.
	 */
	int sendDM1(unsigned int p_spn, unsigned char p_fmi, unsigned char p_occurenceCount = 0x7f);


	/**
	 *	@fn int send(WEJ1939CANMessage *p_msg)
	 *
	 *	@brief Sends the message using J1939 stack and protocol.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_msg Pointer to the message to send
	 *
	 *	@returns WE_SUCCESS if message was send successfull.
	 *	@returns Whatever WECANProtocol::sendMessage() may return.
	 *
	 *	This function is used for sending messages from address claim handler.
	 *	It uses the send method from the WECANProtocol class instead of the one from
	 *	the WEJ1939 class. Doing that allows sending a message, even when J1939 send
	 *	method wouldn't allow sending when the stack is in the wrong state.
	 *	But at least the address claimer needs to send messages for beeing able to change
	 *	the state in future ;)
	 *
	 *	@see WECANProtocol::sendMessage()
	 */
	int send(WEJ1939CANMessage * p_msg);

	/**
	 *	@fn unsigned char getFreeAddress(void) const
	 *
	 *	@brief Delivers a free address
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return nullptr address if no free address was found.
	 *	@return an address if a free address was found.
	 *
	 *
	 */
	unsigned char getFreeAddress(void) const;

	/**
	 *	@fn int updateEcuList(WEJ1939CANMessage &p_msg)
	 *
	 *	@brief Updates the ECU list with the information contained in p_addrClaimMsg
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_addrClaimMsg Address claim message i.e. PGN J1939_PGN_CLAIMED_ADDRESS == 60928 == 0xee00
	 *
	 *	@return WE_SUCCESS
	 *
	 *	This function updates the ECU list with the data given my p_addrClaimMsg.
	 */
	int updateEcuList(const WEJ1939CANMessage & p_addrClaimMsg);

	/**
	 *	@fn int resetEcuListTimestamps()
	 *
	 *	@brief This function sets lastSeen attribute of all known ECUs to zero.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	int resetEcuListTimestamps();

	/**
	 *	@fn int cleanupAddressList()
	 *
	 *	@brief This function removes all ECU that do not have claimed an address from addrMap
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	int cleanupAddressList();

	/**
	 *	@fn int copyData(unsigned char *p_dst, unsigned long long p_src) const
	 *
	 *	@brief Copies an unsigned long long int into a byte array, independent of the endianness.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_dst The destination character array to copy p_src into.
	 *	@param p_src The source value to compy into p_dst.
	 *
	 *	@return WE_SUCCESS on success
	 *	@return WE_INVALID_PARAMETER_ERROR when p_dest == nullptr
	 *
	 *	This function can be used for copying unsigned long long integers into the data part of CAN
	 *	messages. E.g. to copy the ECU name into the data part of message 'Address Claimed'.
	 *
	 *	@WARNING The caller needs to ensure, that the destination array is at least as large as the sizeof(p_src). No
	 *		size checking is done inside of this function.
	 */
	static int copyData(unsigned char *p_dst, unsigned long long p_src);

	/**
	 *	@fn int readInitialAddressFromDEI(unsigned char *p_initialAddress) const
	 *
	 *	@brief Copies an unsigned long long int into a byte array, independent of the endianness.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[out] p_initialAddress The initial address read from DEI
	 *
	 *	@return WE_SUCCESS If initial address was read successful from DEI.
	 *	@return WE_GENERAL_ERROR If reading initial address from DEI failed.
	 */
	int readInitialAddressFromDEI(unsigned char *p_initialAddress) const;

	/**
	 *	@fn int writeInitialAddressToDEI(unsigned char p_initialAddress) const
	 *
	 *	@brief Copies an unsigned long long int into a byte array, independent of the endianness.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[out] p_initialAddress The initial address to write to DEI
	 *
	 *	@return WE_SUCCESS If initial address was written successful to DEI.
	 *	@return WE_GENERAL_ERROR If writing initial address to DEI failed.
	 */
	int writeInitialAddressToDEI(unsigned char p_initialAddress) const;

	WEJ1939Resources * resources{nullptr};	//!< pointer to data that will not change while stack is running and can be shared with state machine.

	int addressHandlerState{J1939_ADDRESS_CLAIM_STATE_DISCOVER};	//!< Current state of the address claiming.
	unsigned int discoveryTimeout{0};	//!< Contains the tick count of a running 'Request for Address Claimed'
	bool discoveryTimoutWaitForOverflow{false};//!< set to true when the discoveryTimout points to a time that is after an tick overflow

	unsigned char initialAddress{J1939_NULL_ADDRESS};	//!< Initial address of the ECU.
	unsigned char address{J1939_NULL_ADDRESS};		//!< Current address of the ECU.

	bool doStaticAddressing{false};		//!< true if static addressing is used. Otherwise dynamic addressing is used
	bool isProjektConfigured{false};		//!< Indicate whether the ecus that are added here are configured via a projekt or not
	struct WEJ1939CanEcu * addrMap[J1939_NUM_ADDRESSES]{};	//!< Address list containing pointers to the ECUs that have claimed an address.
	std::map<unsigned long long, struct WEJ1939CanEcu *> nameMap{};	//!< ECU list containing ECU descriptions using the ECU J1939 name as key. All known ECU can be found in this list.
	bool ecuMapInitialized{false};					//!< true when initialization (request for address claimed) was done

	// the following three members were "static" inside "run()" methods which did not work if more than one instance was started
	unsigned int runTimeout{0};	// the same as "this->discoveryTimeout" but triggered by our own "request for address claimed"
	bool runTimeoutWaitForOverflow{false};	// assume we don't have to wait for an overflow before evaluating "timeout"
	unsigned int runLastTick{0};	// discover WETick overflow between two calls of this method

	unsigned int nextRequestForAddressClaimTick{0};	//!< Replace static member from "handleMessage()" - we want to limit the requests for address claimed we send for unknown originators
};
}

#endif /* WE_J1939_ADDRESS_HANDLER_H_ */
