/**
 *	@file we_j1939_can_message.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief This file provides the WECANMessage specialization for J1939.
 *
 */

#ifndef WE_J1939_CAN_MESSAGE_H_
#define WE_J1939_CAN_MESSAGE_H_

#include "../../../hardware/can_raw/include/we_can_message.h"

namespace WE
{
class WEJ1939CanEcu;

/**
 *	@class WEJ1939CANMessage
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief Class for handling J1939 specific CAN data.
 *
 *	@namespace WE
 *
 *	@warning WECANMessage is inherited public, to allow automatic conversion from WEJ1939CANMessage into
 *		WECANMessage. Anyhow, never access the public attributes of WECANMessage directly. Always use
 *		the functions provided by WEJ1939CANMessage to get proper behavior of this class!
 *
 *	This class specializes WECANMessage with special J1939 attributes.
 */
class WEJ1939CANMessage : public WECANMessage
{
	friend class WEJ1939AddressHandler;	//!< address handler is allowed to set "originator"

public:
	/**
	 *	@fn WEJ1939CANMessage(void)
	 *
	 *	@brief Default constructor.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	WEJ1939CANMessage();

	/**
	 *	@fn WEJ1939CANMessage(const WECANMessage &p_msg)
	 *
	 *	@brief Copy constructor. Initialize data using a WECANMessage.
	 *
	 *	@param p_msg WECANMessage that contains the initial data.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	explicit WEJ1939CANMessage(const WECANMessage &p_msg);

	/**
	 *	@fn WEJ1939CANMessage(const WEJ1939CANMessage &p_msg)
	 *
	 *	@brief Copy constructor. Initialize data using a WEJ1939CANMessage.
	 *
	 *	@param p_msg WEJ1939CANMessage that contains the initial data.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	WEJ1939CANMessage(const WEJ1939CANMessage &p_msg);

	/**
	 *	@fn ~WEJ1939CANMessage()
	 *
	 *	@brief Destructor. Frees all occupied memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	virtual ~WEJ1939CANMessage();

	/**
	 *	@fn WEJ1939CANMessage & operator=(const WECANMessage &p_msg)
	 *
	 *	@brief Assignment operator. Initializing data with a WECANMessage
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_msg WECANMessage containing the data to use.
	 *
	 *	@returns Returns a reference to 'this' object.
	 *
	 *	This function can be used to convert a WECANMessage into a WEJ1939CANMessage.
	 *	This functions is creating a deep copy of the source object.
	 */
	WEJ1939CANMessage & operator=(const WECANMessage &p_msg);

	/**
	 *	@fn WEJ1939CANMessage & operator=(const WEJ1939CANMessage &p_msg)
	 *
	 *	@brief Assignment operator. Initializing data with a WEJ1939CANMessage
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_msg WEJ1939CANMessage containing the data to use.
	 *
	 *	@returns Returns a reference to 'this' object.
	 *	@returns nullptr if allocation of memory failed.
	 *
	 *	This function can be used to copy the content of an other WEJ1939CANMessage.
	 *	This function is creating a deep copy of the source object.
	 */
	WEJ1939CANMessage & operator=(const WEJ1939CANMessage &p_msg);

	/**
	 *	@fn bool isExtendedIDFormat(void) const
	 *
	 *	@brief Delivers information about normal or extended CAN Id format.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns true for extended CAN ID format (29 Bit).
	 *	@returns Returns false for normal CAN ID format (11 Bit).
	 */
	bool isExtendedIDFormat(void) const;

	/**
	 *	@fn int setIDFormat(bool p_isExendedFormat)
	 *
	 *	@brief This function has to be used to set if the ID is in normal or extended Format.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_isExetendedForamt Set to true for 29 Bit ID format, otherwise false (11 Bit ID format).
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This function has to be used for setting the CAN ID format.
	 */
	int setIDFormat(bool p_isExendedFormat);

	/**
	 *	@fn unsigned int getID(void) const
	 *
	 *	@brief Delivers the CAN ID of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns the CAN ID of the message.
	 */
	unsigned int getID(void) const;

	/**
	 *	@fn int setID(const unsigned int p_ID)
	 *
	 *	@brief This function has to be used to change the CAN ID of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_ID The CAN ID to set.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This function has to be used for setting or changing the CAN ID. All related
	 *	variables (that are part of the CAN ID field) are automatically extracted from
	 *	p_ID and updated.
	 */
	int setID(unsigned int p_ID);

	/**
	 *	@fn unsigned int getLength(void) const
	 *
	 *	@brief Delivers the length of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns the length of the message.
	 */
	unsigned int getLength(void) const;

	/**
	 *	@fn int setLength(const unsigned int p_length)
	 *
	 *	@brief This function has to be used to change the length of the message data.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_length The message data length to set.
	 *
	 *	@returns WE_SUCCESS on success.
	 *	@returns WE_GENERAL_ERROR when allocation of memory failed.
	 *
	 *	This function has to be used for setting or changing the message length.
	 *	WARNING: When calling this function, the memory location of the data might change and
	 *		all stored data might be lost! For getting a new valid pointer to the data
	 *		use member function getData().
	 */
	int setLength(unsigned int p_length);

	/**
	 *	@fn unsigned int getPriority(void) const
	 *
	 *	@brief Delivers the priority of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns the priority of the message.
	 */
	unsigned int getPriority(void) const;

	/**
	 *	@fn int setPriority(const unsigned int p_priority)
	 *
	 *	@brief This function has to be used to change the priority of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_priority The priority to set.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This function has to be used for setting or changing the priority.
	 *	When using this function, the CAN ID is automatically updated, so that
	 *	it fits to the new priority value.
	 */
	int setPriority(unsigned int p_priority);

	/**
	 *	@fn unsigned int getExtendedDataPage(void) const
	 *
	 *	@brief Delivers the extended data page of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns the extended data page of the message.
	 */
	unsigned int getExtendedDataPage(void) const;

	/**
	 *	@fn int setExtendedDataPage(const unsigned int p_extDataPage)
	 *
	 *	@brief This function has to be used to change the extended data page of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_extDataPage The extended data page to set.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This function has to be used for setting or changing the extended data page.
	 *	When using this function, the CAN ID and the PGN are automatically updated, so that
	 *	they fit to the new extended data page value.
	 */
	int setExtendedDataPage(unsigned int p_extDataPage);

	/**
	 *	@fn unsigned int getDataPage(void) const
	 *
	 *	@brief Delivers the data page of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns the data page of the message.
	 */
	unsigned int getDataPage(void) const;

	/**
	 *	@fn int setDataPage(const unsigned int p_dataPage)
	 *
	 *	@brief This function has to be used to change the data page of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_dataPage The data page to set.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This function has to be used for setting or changing the data page.
	 *	When using this function, the CAN ID and the PGN are automatically updated, so that
	 *	they fit to the new data page value.
	 */
	int setDataPage(unsigned int p_dataPage);

	/**
	 *	@fn unsigned int getPDUFormat(void) const
	 *
	 *	@brief Delivers the PDU Format of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns the PDU Format of the message.
	 */
	unsigned int getPDUFormat(void)const;

	/**
	 *	@fn int setPDUFormat(const unsigned int p_pduFormat)
	 *
	 *	@brief This function has to be used to change the PDU format of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_pduFormat The PDU format to set.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This function has to be used for setting or changing the PDU format.
	 *	When using this function, the CAN ID and the PGN are automatically updated, so that
	 *	they fit to the new PDU format value.
	 */
	int setPDUFormat(unsigned int p_pduFormat);

	/**
	 *	@fn unsigned int getPDUSpecific(void) const
	 *
	 *	@brief Delivers the PDU specific of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns the PDU specific of the message.
	 *
	 *	If PDU format is < 240 this returns the destination address while PDU specific of PGN is 0.
	 *
	 *
	 */
	unsigned int getPDUSpecific(void) const;

	/**
	 *	@fn int setPDUSpecific(const unsigned int p_pduSpecific)
	 *
	 *	@brief This function has to be used to change the PDU specific of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_pduSpecific The PDU specific to set.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This function has to be used for setting or changing the PDU specific.
	 *	When using this function, the CAN ID and the PGN are automatically updated, so that
	 *	they fit to the new PDU specific value.
	 *
	 *	If PDU format is < 240 this is used to set the destination address but keep PGN unchanged.
	 *
	 */
	int setPDUSpecific(unsigned int p_pduSpecific);

	/**
	 *	@fn unsigned int getSourceAddress(void) const
	 *
	 *	@brief Delivers the source address of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns the source address of the message.
	 */
	unsigned int getSourceAddress(void) const;

	/**
	 *	@fn int setSourceAddress(const unsigned int p_srcAddress)
	 *
	 *	@brief This function has to be used to change the source address of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_srcAddress The source address to set.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This function has to be used for setting or changing the source address.
	 *	When using this function, the CAN ID is automatically updated, so that
	 *	it fits to the new source address value.
	 */
	int setSourceAddress(unsigned int p_srcAddress);

	/**
	 *	@fn unsigned int getDestinationAddress(void) const
	 *
	 *	@brief Delivers the destination address of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns the destination address of the message.
	 *
	 *	If (240 <= this->getPDUFormat()) the result will be J1939_BROADCAST_ADDRESS in any case.
	 *
	 */
	unsigned int getDestinationAddress(void) const;

	/**
	 *	@fn int setDestinationAddress(const unsigned int p_dsetAddress)
	 *
	 *	@brief This function has to be used to change the destination address of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_dstAddress The destination address to set.
	 *
	 *	@returns WE_SUCCESS if (240 > this->getPDUFormat())
	 *	@returns WE_INVALID_PARAMETER_ERROR otherwise
	 *
	 *	If (240 > this->getPDUFormat()) this method will set the destination address in the
	 *	CAN ID of this message. Otherwise it will ignore it.
	 *
	 *	This function has to be used for setting or changing the destination address.
	 *	When using this function, the CAN ID is automatically updated, so that
	 *	it fits to the new destination address value.
	 *
	 *
	 */
	int setDestinationAddress(unsigned int p_dstAddress);

	/**
	 *	@fn unsigned int getPGN(void) const
	 *
	 *	@brief Delivers the PGN of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns the PGN of the message.
	 */
	unsigned int getPGN(void) const;

	/**
	 *	@fn int setPGN(const unsigned int p_pgn)
	 *
	 *	@brief This function has to be used to change the PGN of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_pgn The PGN to set.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if PDU specific != 0 while PDU format < 240
	 *
	 *	This function has to be used for setting or changing the PGN.
	 *	When using this function, the CAN ID, extended data page, data page,
	 *	PDU format and PDU specific are automatically updated, so that they
	 *	fits to the new PGN value.
	 */
	int setPGN(unsigned int p_pgn);

	/**
	 *	@fn unsigned char* getData(void) const
	 *
	 *	@brief Delivers a pointer to the objects data of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns a pointer to the objects data of the message.
	 */
	unsigned char* getData(void) const;

	/**
	 *	@fn int setData(const unsigned char * p_data, const unsigned int p_length)
	 *
	 *	@brief This function has to be used to change the data and data length of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_data The data to set.
	 *	@param p_length The length of the data. This length will also be set.
	 *
	 *	@returns WE_SUCCESS on success
	 *	@returns WE_GENERAL_ERROR when allocation of memory fails.
	 *
	 *	This function has to be used for setting or changing the data.
	 *	This function also updates the message data length. Because dynamic
	 *	memory allocation is used, this function can only be used for setting all
	 *	data at once.
	 *	Same functionality can be achieved by using setLength() and
	 *	getData() member functions and writing the data manually to the returned
	 *	data pointer.
	 */
	int setData(const unsigned char * const p_data, unsigned int p_length);

	/**
	 *	@fn unsigned int getTimeRx(void) const
	 *
	 *	@brief Delivers the receive time stamp of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Returns the receive time stamp of the message.
	 */
	unsigned int getTimeRx(void) const;

	/**
	 *	@fn int setTimeRx(const unsigned int p_timeRx)
	 *
	 *	@brief This function has to be used to set the receive time of the message.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_priority The receive time to set.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This function has to be used for setting the receive time of the message.
	 */
	int setTimeRx(unsigned int p_timeRx);


	/**
	 *	@fn bool isValidJ1939msg( void )
	 *
	 *	@brief Check if p_msg follows basic requirements for J1939 messages.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns true if this message is a valid J1939 message by the criteria listed below
	 *	@returns false otherwise
	 *
	 *	This method considers a CAN message to be a valid J1939 message if
	 *	* it has an extended CAN ID,
	 *	* destination is NOT the nullptr address and
	 *	* sender address is NOT the broadcast address
	 *
	 */
	bool isValidJ1939msg( void ) const;


	/**
	 *	@fn bool isSentToaddress( unsigned char p_address )
	 *
	 *	@brief Check if this message is a broadcast or is sent to p_address.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_address Address of the intended receiver.
	 *
	 *	@returns true if message is a broadcast or sent peer to peer to p_address
	 *	@returns false otherwise
	 *
	 */
	bool isSentToaddress( unsigned char p_address ) const;


	/**
	 *	@fn void reset(void)
	 *
	 *	@brief This function resets the attributes of this class to default values.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	void reset(void);


	/**
	 *	@fn WEJ1939CanEcu * getOriginatorECU( void ) const
	 *
	 *	@brief Return pointer to description of ECU which sent this message.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns pointer to description of ECU which sent this message
	 *	@returns nullptr if WEJ1939AddressHandler did not set the address (yet).
	 *
	 *	The returned value is only useful for received messages where WEJ1939AddressHandler already
	 *	added a valid value.
	 *
	 *	@see WEJ1939AddressHandler
	 *
	 */
	WEJ1939CanEcu * getOriginatorECU( void ) const;


	/**
	 *	@fn int copyOriginatorECU( const WEJ1939CANMessage * p_msg )
	 *
	 *	@brief Copy originator, e.g. as destination for response.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg where destination ECU should be copied to this.
	 *
	 *	@returns An error code - WE_SUCCESS in case of success.
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_msg is nullptr.
	 *
	 *	@see WEJ1939AddressHandler
	 *
	 */
	int copyOriginatorECU( const WEJ1939CANMessage * p_msg );


private:

	// special members of a J1939 CAN ID
	mutable unsigned char * longData;	//!< used to store data when message length is greater than 8 bytes. (mutable so a pointer to it can be returned in a const function (getData()))
	WEJ1939CanEcu * originator;
};
} /* namespace WE */

#endif /* WE_J1939_CAN_MESSAGE_H_ */
