/**
 *	@file we_j1939_message_object.h
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Implementation of PGN mapping objects.
 *
 *	Based on we_can_message_object.[h|cpp] this file implements what is needed to receive and send PGNs from a
 *	CAN protocol stack.
 *
 *
 *
 *
 *	This file extends the generic description of a CAN message  to a PGN definition for J1939.
 *	This is NOT implementing the raw CAN message to be sent on the bus (see WECANMessage and J1939CANMessage for that) but
 *	providing some interfaces for a J1939 stack to maintain such messages.
 *
 *
 *	@see WECANMessage
 *	@see WEJ1939MappingObject
 *	@see J1939CANMessage
 *
 */

#ifndef WE_J1939_MESSAGE_OBJECT_H_
#define WE_J1939_MESSAGE_OBJECT_H_

#include "we_j1939_mapping_object.h"	// WETransmitType, mapping object
#include "../../common/include/we_can_message_object.h"	// will need the base class  (and WECANRequestData) in any case

namespace WE
{

class WECANRaw;
struct WECANMessage;
class WEJ1939CanEcu;
class WEJ1939CANMessage;

/**
 *	@class WECANRequestData
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@brief Struct that stores information of how a request looks like for a WEJ1939MappingObject
 *
 *	@warning Here we assume that no request needs more than one CAN message.
 *
 *
 *	@namespace WE
 */
class WEJ1939RequestData : public WECANRequestData
{
public:
	/**
	 *	@fn WEJ1939RequestData()
	 *
	 *	@brief Standard constructor that initializes all member variables.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 */
	WEJ1939RequestData()
	:WECANRequestData()
	{
		//other members are initialized at declaration
	}

	virtual ~WEJ1939RequestData() = default;

	unsigned char priority{6};		//!< priority to be used in case we should ever send this request.
	unsigned int pgn{0xea00};		//!< pgn of the request - including data page and extended data page
	int ecuID{0};			//!< while parsing the Projekt we are using the same structure but only know ECU ID
	WEJ1939CanEcu *ecu{nullptr};		//!< ECU which has to send the request (this is only to recognize the request
					// - mapping will be sent as configured)
					// or for receive mapping the ECU which should get the request
	std::string pgnType{"0x00EF00(A)"};

	// SERIALIZATION
	// This class is exported in we_j1939_message_object.cpp, did not find any other suitable class/file in which to export.
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive &p_ar, const unsigned int)
	{
		p_ar & boost::serialization::base_object<WECANRequestData>(*this);	// don't forget to serialize parent class
		p_ar & priority;		//!< priority to be used in case we should ever send this request.
		p_ar & pgn;		//!< pgn of the request - including data page and extended data page
		p_ar & ecuID;		//!< while parsing the Projekt we are using the same structure but only know ECU ID
		// should not need to serialize this: ar & ecu;		//!< ECU which has to send the request (this is only to recognize the request		ar & byteNumber; 	//!< The number of the byte of the CAN message where the data is mapped to (> 0)
		// OR: negative number of "separator/delimiter chars" defining the END of the mapping
		// e.g. byteNumber == -2: data is between delimiter 1 and 2; *constantByteMask is the delimiter char then
		p_ar & pgnType;
	}
};

class WEJ1939MessageObject : public WECANMessageObject
{
public:
	//we need to use this using mechanism if we are only overriding one
	//of the two existing matchesMsg functions from base object
	using WECANMessageObject::matchesMsg;
	/**
	 *	@fn WEJ1939MessageObject(void)
	 *
	 *	@brief Default constructor initializing members with default values.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 */
	WEJ1939MessageObject() = default;

	/**
	 *	@fn WEJ1939MessageObject(unsigned int p_objectID)
	 *
	 *	@brief Constructor initializing members with default values - except object ID.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_objectID - ID of this object (NOT the CAN ID of this object)
	 *
	 */
	explicit WEJ1939MessageObject(unsigned int p_objectID);

	/**
	 *	@fn WEJ1939MessageObject(const WEJ1939MessageObject & p_object)
	 *
	 *	@brief Copy constructor.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhartwe_can_message_object
	 *
	 *	@param p_object - object to be copied
	 *
	 */
	WEJ1939MessageObject(const WEJ1939MessageObject &p_object);
	WEJ1939MessageObject& operator=(const WEJ1939MessageObject &p_object);

	/**
	 *	@brief Constructor initializing members with values provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_objectID - ID of this object (NOT the CAN ID of this object)
	 *	@param p_priority - priority to be used in CAN message
	 *	@param p_pgn - pgn to be sent/received here (including [ext] data page bits)
	 *	@param p_ownerID - ID of ECU description of ECU who should send/receive this PGN
	 *	@param p_timeoutPeriod - period to be applied for timeout of receive message or as sending period of transmit message
	 *	@param p_timeoutListener - this will get events if receive message times out or becomes valid again
	 *	@param p_mapping	- mapping describing the data part of this message
	 *	@param p_requestData	- request to be sent out to get receive message or to be answered with transmit message
	 *
	 */
	WEJ1939MessageObject(
				unsigned int p_objectID,
				unsigned int p_priority,
				unsigned int p_pgn,
				unsigned int p_ownerID,
				unsigned int p_timeoutPeriod,
				WESignalListener *p_timeoutListener,
				WEJ1939MappingObject *p_mapping,
				WEJ1939RequestData *p_requestData
				);

	/**
	 *	@fn virtual ~WEJ1939MessageObject()
	 *
	 *	@brief Default destructor destroying members and freeing allocated memory.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 */
	virtual ~WEJ1939MessageObject() = default;

	/**
	 *	@fn int createRequest(WEJ1939RequestData * p_request)
	 *
	 *	@brief Copy request provided and use it for this object.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_request - pointer to request to be copied into this message.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This method overwrites the one of the base class using the extended request type for J1939.
	 *
	 *	@warning This will in any case delete the existing request of this message.
	 */
	int createRequest(WEJ1939RequestData *p_request);

	/**
	 *	@fn WEJ1939RequestData * getRequest(void)
	 *
	 *	@brief Get pointer to request attached to this mapping.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns A pointer to the request of this mapping
	 *	@returns nullptr if there is no request defined
	 *
	 *	@warning This does not copy the request.
	 */
	WEJ1939RequestData* getRequest(void) const;

	/**
	 *	@fn bool matchesRequest(WEJ1939CANMessage * p_msg)
	 *
	 *	@brief Check if the request of this message description fits the message provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg - message to be compared with this object's request data
	 *
	 *	@returns true if this is a transmit object and the CAN message provided fits to it
	 *	@returns false if this belongs to a receive object or definition doesn't fit.
	 *
	 *	This method also requires the size to match.
	 *
	 *	This method overwrites the one of the base class using the extended request type for J1939.
	 *
	 *	@see fitsConstRequestData()
	 */
	bool matchesRequest(WEJ1939CANMessage *p_msg) const;

	/**
	 *	@fn bool matchesMsg(WECANMessage * p_msg, unsigned int p_timeNow = 0)
	 *
	 *	@brief Check if this message description fits the message provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg - message to be compared with this object
	 *	@param p_timeNow - value for timestamp in case return value is "true"
	 *
	 *	@returns true if this is a receive object and the CAN message provided fits to it
	 *	@returns false if this belongs to a transmit object
	 *	@returns false if definition doesn't fit
	 *	@returns false if this has no mapping attached
	 *
	 *	Time stamp of this object is set to "p_timeNow" if this method returns true.
	 *
	 *	@warning Originator ECU has to be checked before since this class has no access to address handler.
	 *
	 *	@see WEJ1939MappingObject
	 *	@see WETick
	 */
	bool matchesMsg(WECANMessage *p_msg, unsigned int p_timeNow) override;

	/**
	 *	@fn bool matchesMsg(WECANMessage * p_msg, unsigned int p_timeNow = 0)
	 *
	 *	@brief Check if this message description fits the message provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg - message to be compared with this object
	 *	@param p_originatorEcu pointer to be compared with this->ecu - or "nullptr" for "don't care"
	 *	@param p_timeNow - value for timestamp in case return value is "true"
	 *
	 *	@returns true if this is a receive object and the CAN message provided fits to it
	 *	@returns false if this belongs to a transmit object
	 *	@returns false if definition doesn't fit
	 *	@returns false if this has no mapping attached
	 *
	 *	Time stamp of this object is set to "p_timeNow" if this method returns true.
	 *
	 *	@see getOwnerEcu()
	 *	@see WEJ1939MappingObject
	 *	@see WETick
	 */
	bool matchesMsg(const WEJ1939CANMessage &p_msg, WEJ1939CanEcu *p_originatorEcu, unsigned int p_timeNow);

	/**
	 *	@fn bool matchesMsg(unsigned int p_ID, unsigned char * p_data, unsigned int p_size, unsigned int p_timeNow = 0)
	 *
	 *	@brief Check if this message description fits the (extended) message provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_ID - CAN ID of message to be compared with this object
	 *	@param p_data - data buffer of message to be compared with this object
	 *	@param p_size - size of p_data in bytes
	 *
	 *	@returns true if this is a receive object and the CAN message provided fits to it
	 *	@returns false if this belongs to a transmit object
	 *	@returns false if definition doesn't fit
	 *	@returns false if this has no mapping attached
	 *
	 *	This could be used e.g. to check if this object matches a sequence of more than one CAN message
	 *	- if the protocol implementation supports something like transport protocol.
	 *
	 *	Time stamp of this object is set to "p_timeNow" if this method returns true.
	 *	If "0 == p_timeNow" timestamp is set to WETick::getInstance()->getTickCount() instead.
	 *
	 *
	 *	@see WEJ1939MappingObject
	 */
	virtual bool matchesMsg(unsigned int p_ID, unsigned char *p_data, unsigned int p_size, unsigned int p_timeNow = 0) const;

	/**
	 *	@fn int fillReqForPGN( WEJ1939CANMessage * p_msg, unsigned int p_timeNow )
	 *
	 *	@brief Fill p_msg with request for this mapping PGN.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg - message which should get the ID
	 *	@param p_timeNow - Time to apply to this->requestTimeStamp if filling p_msg succeeds
	 *
	 *	@returns WE_SUCCESS if message could be written to send buffer
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_msg is nullptr
	 *	@returns WE_CONNECT_ERROR if this' ECU isn't active (i.e. didn't receive address claim of dynamic ECU)
	 *
	 *	Convert p_msg into a request for PGN (0x00ea00) for this PGN to be sent to the ECU we expect to receive this PGN from.
	 *	This method does not consider if this instance is a send or receive message.
	 *
	 *	@see createRequest()
	 */
	int fillReqForPGN(WEJ1939CANMessage *p_msg, unsigned int p_timeNow);

	/**
	 *	@fn int fillID( WECANMessage * p_msg )
	 *
	 *	@brief Add ID of this object to p_msg.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg - message which should get the ID
	 *
	 *	@returns WE_SUCCESS if message could be written to send buffer
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_msg is nullptr
	 *	@returns WE_DEVICE_ERROR if the ecu who should receive the message did not claim an address yet
	 *
	 *	Just set the CAN ID of this object in p_msg.
	 *	In the base implementation this method does not consider if this is a send or receive message.
	 *	It also does not consider the ID mask in any way (since it's intended for receiving anyways).
	 *
	 *	WE_DEVICE_ERROR is only returned if the ECU has to be identified by its name rather than by a
	 *	static address.
	 *
	 *	@see WECANRaw
	 */
	virtual int fillID(WECANMessage *p_msg) const override;

	/**
	 *	@fn void createDataMapping( WECANMappingObject * p_mapping )
	 *
	 *	@brief Create a copy of p_mapping and set it as the data mapping for this CAN message.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_mapping - pointer to mapping which will be copied
	 *
	 */
	void createDataMapping(WECANMappingObject *p_mapping) override;

	/**
	 *	@fn unsigned int getPgn( void )
	 *
	 *	@brief Get PGN of this message object.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns Content of pgn member variable i.e. the PGN this object is configured for
	 *
	 *
	 */
	unsigned int getPgn(void) const;

	/**
	 *	@fn WEJ1939CanEcu * getOwnerEcu( void )
	 *
	 *	@brief Get ECU for this message object.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns Content of ecu member variable i.e. the pointer to the ECU who should send or get this mapping message.
	 *
	 *
	 */
	WEJ1939CanEcu* getOwnerEcu(void) const;

	/**
	 *	@fn void setOwnerEcu( WEJ1939CanEcu * p_ecu )
	 *
	 *	@brief Set ECU for this message object.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_ecu Ecu member variable i.e. the pointer to the ECU who should send or get this mapping message.
	 *
	 *
	 *	@returns nothing
	 *
	 *	This method is intended to add a pointer to an ECU description after the object is created.
	 *	This is useful e.g. if the original definition only knew the ID of this ECU.
	 *
	 *	@warning Currently this method accepts every pointer provided.
	 *
	 */
	void setOwnerEcu(WEJ1939CanEcu *p_ecu);

	/**
	 *	@fn int * getOwnerEcuID( void )
	 *
	 *	@brief Get ECU ID for this message object.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns Content of ownerEcuID member variable i.e. the ID of the ECU who should send or get this mapping message.
	 *
	 *
	 */
	int getOwnerEcuID(void) const;

	/**
	 *	@fn virtual int fillSendMessage(
	 WEJ1939CANMessage * p_msg,
	 WEDEI * p_dei,
	 unsigned int p_timeNow = 0,
	 unsigned int p_varKey = 0,
	 int p_varValue = 0,
	 bool p_setTimeStamp = true)
	 *
	 *	@brief If this is a transmit message populate p_msg with its data.
	 *
	 *	@author ajay.bisht@aricent.com
	 *	@author Ajay Bisht
	 *
	 *	@param p_msg - Pointer to the CAN message carrying the PGN information for which response message has to be filled.
	 *	@param p_actualLength - Actual length of the data filled in the message
	 *	@param p_dei - DEI which should provide variable values
	 *	@param p_timeNow - Tick in milliseconds to be set to the timestamp of this object
	 *	@param p_varKey - key of variable which should NOT be sent with DEI value but with p_varValue.
	 *	@param p_varValue - value to be used for variable identified by "p_varKey"
	 *	@param p_setTimeStamp - true if timestamp should be set in case of success - false otherwise
	 *
	 *	@returns WE_SUCCESS if message could be filled
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_msg == nullptr
	 *	@returns WE_READ_ERROR if this is a read and not a write message
	 *	@returns any value WECANMappingObject::fillMessageFromDEI() returns
	 *
	 *	If this is a send message it is determined by the mapping object attached.
	 *	If message could be filled successfully with DEI data only, timestamp of this object is set to p_timeNow.
	 *	If p_varKey = 0 AND p_varValue = 0 timestamp is set (otherwise this is considered a write request)
	 *	If p_timeNow == 0 and timestamp should be set timestamp is set to WETick::getInstance()->getTickCount();
	 *	p_setTimeStamp can be used if message is filled to answer a request but periodic sending should continue
	 *	ignoring this.
	 *
	 *	@see WECANMappingObject
	 *	@see WECANMappingObject::fillMessageFromDEI()
	 *	@see WETick
	 */
	virtual int fillSendMessage(
					WEJ1939CANMessage *p_msg,
					WEDEI *p_dei,
					unsigned int p_timeNow = 0,
					unsigned int p_varKey = 0,
					int p_varValue = 0,
					bool p_setTimeStamp = true);

	/**
	 *	@fn int fillMsgWithRequest(WEJ1939CANMessage * p_msg, unsigned int p_timeNow)
	 *
	 *	@brief Fill p_msg with data defined in this->requestData
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_msg - message which should get the ID
	 *	@param p_timeNow - Time to apply to this->requestTimeStamp if filling p_msg succeeds
	 *
	 *	@returns WE_SUCCESS if message could be written to send buffer
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_msg is nullptr
	 *	@returns WE_INVALID_PARAMETER_ERROR if this->requestData is nullptr
	 *	@returns WE_CONNECT_ERROR if this' ECU isn't active (i.e. didn't receive address claim of dynamic ECU)
	 *
	 *	Just set the CAN ID, DLC and data of this->requestData in p_msg.
	 *
	 *
	 *	@see WECANMessageObject::fillMsgWithRequest(WECANMessage * p_msg, unsigned int p_timeNow)
	 */
	virtual int fillMsgWithRequest(WEJ1939CANMessage *p_msg, unsigned int p_timeNow);

private:
	int ownerEcuID{-1};	//!< ID if the ECU who should provide or get this PGN; used to set "ecu" pointer later
	WEJ1939CanEcu *ecu{nullptr};		//!< ECU which should provide or get this PGN, "nullptr" for "any ECU"
	unsigned int priority{6};	//!< priority for transmit PGNs
	unsigned int pgn{0};		//!< PGN including data page and extended data page
	WEJ1939RequestData *requestDataJ1939{nullptr}; //!< overwrite the member of the parent class here to avoid all the type casting
};
}	// namespace

#endif /* WE_J1939_MESSAGE_OBJECT_H_ */
