/**
 *	@file we_j1939_transport_protocol_handler.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief This file contains definitions of types, error codes, constants, etc.
 *
 *
 *	Definitions for TP come from SAE J1939-21:2006, Section 5.10 and Annex C.
 *	Definitions for ETP come from ISO 11783-6:2010, Annex K.
 *
 */

#ifndef WE_J1939_TRANSPORT_PROTOCOL_HANDLER_H_
#define WE_J1939_TRANSPORT_PROTOCOL_HANDLER_H_

#include <set>
#include <list>

#include "we_j1939_definitions.h"
#include "we_j1939_can_message.h"


namespace WE
{

class WEJ1939Resources;
class WESignalListener;

const size_t TP_REORDER_BUFFER_SIZE = 5;	//!< Size of buffer used in order to re-arrange received messages if required

/**
 *	@struct messageAssembly
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief This struct hold all data needed to reassemble a large message.
 *
 *	This struct is mainly used to assemble (E)TP messages received.
 *	Additionally we maintain an instance with every ECU for the sending direction.
 *
 *	@namespace WE
 */
struct MessageAssembly
{
	MessageAssembly()
	:msg(new WEJ1939CANMessage()),
	 destAddress(J1939_BROADCAST_ADDRESS),
	 writePtr(nullptr),
	 nextPacketNo(0),
	 bytesReceived(0),
	 timeout(0),
	 numPackagesUntilCTS(0xff),
	 numPackagesForOffset(0x0),
	 currentNumPackagesUntilCTS(0xff),
	 cm_pgn(0)
	{
		for (size_t i = 0; i < TP_REORDER_BUFFER_SIZE; ++i)
		{
			this->reOrderBuf[i] = nullptr;
		}
	}

	MessageAssembly(const MessageAssembly & p_ma)
	:msg(new WEJ1939CANMessage()),
	 destAddress(p_ma.destAddress),
	 writePtr(p_ma.writePtr),
	 nextPacketNo(p_ma.nextPacketNo),
	 bytesReceived(p_ma.bytesReceived),
	 timeout(p_ma.timeout),
	 numPackagesUntilCTS(p_ma.numPackagesUntilCTS),
	 numPackagesForOffset(p_ma.numPackagesForOffset),
	 currentNumPackagesUntilCTS(p_ma.currentNumPackagesUntilCTS),
	 cm_pgn(p_ma.cm_pgn)
	{
		for (size_t i = 0; i < TP_REORDER_BUFFER_SIZE; ++i)
		{
			this->reOrderBuf[i] = nullptr;
		}
	}
	MessageAssembly& operator=(const MessageAssembly & p_ma) = delete;

	void init( void )
	{
		if (nullptr != this->msg)
		{
			*(this->msg) = WEJ1939CANMessage();
		}
		this->destAddress = J1939_BROADCAST_ADDRESS;
		this->writePtr = nullptr;
		this->nextPacketNo = 0;
		this->bytesReceived = 0;
		this->timeout = 0;
		this->numPackagesUntilCTS = 0xff;
		this->numPackagesForOffset = 0x0;
		this->currentNumPackagesUntilCTS = 0xff;
		this->cm_pgn = 0;
		for (size_t i = 0; i < TP_REORDER_BUFFER_SIZE; ++i)
		{
			this->reOrderBuf[i] = nullptr;
		}
	}

	virtual ~MessageAssembly()
	{
		delete msg;
		for (size_t i = 0; i < TP_REORDER_BUFFER_SIZE; ++i)
		{
			delete this->reOrderBuf[i];
		}
	}

	WEJ1939CANMessage * msg;		//!< Message that is assembled (is a pointer since this exact message is provided to the subscriber)
						//   This message is also used as the one to send (i.e. it contains all the data to transfer via (E)TP)
	unsigned char destAddress;		//!< Even a broadcast PGN may be sent by a directed TP - so we have to remember the address separately here
	unsigned char * writePtr;		//!< points to the next byte were a received byte has to be written
	unsigned int nextPacketNo;		//!< holds the next package number that has to be received
	unsigned int bytesReceived;
	unsigned int timeout;			//!< holds a timeout value. If tick counter reaches the (timeout + msg->getTimeRx()) value, the message has to be destroyed
	// NOTE: when changing the following data type you need to uncomment some code in function handleRequestToSend()
	unsigned char numPackagesUntilCTS;	//!< holds number of packages that needs to be received until a CTS has to be send (calculated value when requestToSend is received)
	unsigned char numPackagesForOffset; 	//!< holds the number of packages that have to be applied to the offset that was send with the last data package offset message
	unsigned char currentNumPackagesUntilCTS;//!< holds current number of packages that needs to be received until a CTS has to be send.
	unsigned int cm_pgn;			//!< holds the pgn of the connection management request message. This allows deciding between BAM, TP and ETP.
	WEJ1939CANMessage * reOrderBuf[TP_REORDER_BUFFER_SIZE];	//!< Buffer used to re-order received messages if required
};

/**
 *	@class WEJ1939TransportProtocolHandler
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief This class provides the transport protocol for J1939 stack.
 *
 *	@namespace WE
 *
 *	This class controls the transport protocol of the new J1939 stack.
 */
class WEJ1939TransportProtocolHandler
{
public:
	/**
	 *	@fn WEJ1939TransportProtocolHandler()
	 *
	 *	@brief Default constructor. Initializes pointers to nullptr and other members to sane
	 *		start values.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 */
	WEJ1939TransportProtocolHandler();

	/**
	 *	@fn virtual ~WEJ1939TransportProtocolHandler()
	 *
	 *	@brief Virtual destructor. Frees all occupied memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	virtual ~WEJ1939TransportProtocolHandler();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEJ1939TransportProtocolHandler(const WEJ1939TransportProtocolHandler &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEJ1939TransportProtocolHandler & operator=(const WEJ1939TransportProtocolHandler &) = delete;

	/**
	 *	@fn virtual int create(WEJ1939Resources * p_resources)
	 *
	 *	@brief Initialize the J1939 transport protocol handler and allocated needed memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_resources pointer to J1939 stack resources
	 *
	 *	@returns WE_SUCCESS
	 */
	virtual int create(WEJ1939Resources * p_resources);

	/**
	 *	@fn virtual int init();
	 *
	 *	@brief Initialize all class attributes with meaningful default values. Does not allocate memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int init();

	/**
	 *	@fn virtual int reset();
	 *
	 *	@brief Reset all class attributes with default values. Does not free allocated memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int reset();

	/**
	 *	@fn virtual int release();
	 *
	 *	@brief Frees allocated memory and sets all class attributes to default values.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int release();

	/**
	 *	@fn int handleMessage(WEJ1939CANMessage &p_msg, WEJ1939CANMessage **p_longMsg)
	 *
	 *	@brief Method to handle and distribute all incoming J1939 CAN messages.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_msg[in] Pointer to the J1939 message received which should be evaluated/handled here.
	 *	@param p_longMsg[out] If a transport protocol message is completed, this pointer points to this message. Otherwise it is nullptr.
	 *
	 *	@returns WE_SUCCESS if the received message could be handled (or the state machine is paused)
	 *	@returns WE_NOT_FOUND_ERROR if PGN of the received message was not intended for J1939 transport protocol
	 *	@returns WE_INVALID_MESSAGE_RECEIVED_ERROR if message contains a TP PGN but fails in more detailed checks
	 *
	 *	This function get's called whenever a J1939 CAN message is read from the CAN bus by WEJ1939StateMAchine::handleCANmsg()
	 *
	 *	Some examples why this method could return WE_INVALID_MESSAGE_RECEIVED_ERROR:
	 *	* wrong data page or extended data page
	 *	* length != 8 data bytes
	 *	* first data byte of the message is not recognized as a control byte (i.e. not a value defined in the standard)
	 *	* message is not a follow up in the sequence
	 *
	 *	@warning "p_msg" is only valid in this method - if (parts of) the message should be stored one has to copy them/it
	 *
	 *	@see WEJ1939StateMachine::handleCANmsg()
	 *
	 */
	int handleMessage(WEJ1939CANMessage &p_msg, WEJ1939CANMessage **p_longMsg);

	/**
	 *	@fn int run(unsigned int p_time)
	 *
	 *	@brief The run function needs to be called frequently to process all transport protocol tasks
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_time The tick count until which this function is allowed to run
	 *
	 *	@returns If everything went OK, WE_SUCCESS is returned. Everything else means error but is hard to define at the moment
	 *
	 *	This functions needs to be called frequently so that all transport protocol tasks of the J1939 stack
	 *	can be triggered at the right time.
	 */
	int run(unsigned int p_time);

	/**
	 *	@fn int sendTPMessage(WEJ1939CANMessage * p_msg, unsigned char p_destAddress = J1939_NULL_ADDRESS)
	 *
	 *	@brief The function triggers sending of J1939 CAN messages that need transport protocol for beeing send
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_msg The J1939 message to send by using the transport protocol.
	 *	@param p_destAddress Destination the TP should be sent to
	 *
	 *	@returns WE_SUCCESS on success
	 *	@returns WE_INVALID_PARAMETER_ERROR if the parameter p_msg is invalid
	 *	@returns WE_NOT_SUPPORTED_ERROR if the message length is invalid or the ECU does not have a valid address
	 *	@returns WE_WRITE_ERROR if this->sendMsg is invalid
	 *	@returns WE_SEND_ERROR if sending the message failed because destination ECU is unknown
	 *
	 *	J1939 messages with more then 8 bytes data have to be send by using this function. This function uses the J1939
	 *	transport protocol to send large messages over the CAN bus.
	 *
	 *	If p_destAddress is set to J1939_NULL_ADDRESS (0xfe) the destination address is taken from p_msg.
	 *	Setting p_destAddress allows to even transfer broadcast PGN in a dedicated (E)TP session.
	 *
	 */
	int sendTPMessage(WEJ1939CANMessage * p_msg, unsigned char p_destAddress = J1939_NULL_ADDRESS);

	/**
	 *	@brief The function registers an event handler that gets the result events when sending transport protocol messages
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_eventListener The listener to call when the event occurs.
	 *
	 *	@returns WE_SUCCESS.
	 *
	 *	For unregistering the event Listener set p_eventListener to nullptr.
	 *	@WARNING: Only one listener can be registered.
	 *
	 *	The event send will have following content:
	 *	  eventType = J1939_TP_MESSAGE_SEND
	 *	  eventParameter1 = PGN of the message
	 *	  eventParameter2 = error code
	 *	  ep3.timestampMilliseconds = tick when the event occurred
	 *	Following error codes are supported:
	 *		WE_SUCCESS: Message was send successful
	 *		WE_OPERATION_IN_PROGRESS: The other ECU did not accept the message, because it is already busy with another TP message
	 *		WE_TIMEOUT_ERROR: A timeout has occurred.
	 *		WE_STATE_NOT_SUPPORTED_ERROR: the other ECU does not hold the resources needed for receiving the message.
	 *		WE_INVALID_MESSAGE_RECEIVED_ERROR: This ECU is already sending a TP message and is requested to send another one at the same time. WARNING: in this situation, sending both messages will be aborted.
	 */
	int registerEventListener(WESignalListener* p_eventListener);

private:
	/**
	 *	@fn int sendNextMessagePacket(WEJ1939CanEcu * p_ecu);
	 *
	 *	@brief The functions sends the next part of a large message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_ecu Send next message of TP session to this ECU.
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int sendNextMessagePacket(WEJ1939CanEcu * p_ecu);


	/**
	 *	@fn int sendRequestToSendMessage(WEJ1939CanEcu * p_ecu)
	 *
	 *	@brief The functions sends a request to send message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_ecu The ECU to which the request to send message has to be sent.
	 *
	 *	@returns WE_SUCCESS.
	 *
	 *	The functions sends a 'request to send' message. It automatically takes care if the
	 *	normal or extended transport protocol version of the request to send message has
	 *	to be sent.
	 */
	int sendRequestToSendMessage(WEJ1939CanEcu * p_ecu);


	/**
	 *	@fn int sendClearToSendMessageTP(unsigned int p_pgn,
				     unsigned char p_dstAdr,
				     unsigned char p_maxNumPackages,
				     unsigned char p_nextPacketNumber)
	 *
	 *	@brief The functions sends an TP clearToSend message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_pgn The PGN of the clearToSend message
	 *	@param p_dstAdr The destination address of the clearToSend message
	 *	@param p_maxNumPackages The number of packages that shall be send after this CTS message
	 *	@param p_nextPacketNumber The next packet number expected
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int sendClearToSendMessageTP(unsigned int p_pgn,
				     unsigned char p_dstAdr,
				     unsigned char p_maxNumPackages,
				     unsigned char p_nextPacketNumber);

	/**
	 *	@fn int sendClearToSendMessageETP(unsigned int p_pgn,
				   unsigned char p_dstAdr,
				   unsigned char p_maxNumPackages,
				   unsigned int p_nextPacketNumber)
	 *
	 *	@brief The functions sends an ETP clearToSend message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_pgn The PGN of the clearToSend message
	 *	@param p_dstAdr The destination address of the clearToSend message
	 *	@param p_maxNumPackages The number of packages that shall be send after this CTS message
	 *	@param p_nextPacketNumber The next packet number expected
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int sendClearToSendMessageETP(unsigned int p_pgn,
				   unsigned char p_dstAdr,
				   unsigned char p_maxNumPackages,
				   unsigned int p_nextPacketNumber);

	/**
	 *	2021 version of J1939-21 section 5.10.4.1 requires to have byte 3, bits 1 and 2 of an abort
	 *	message to contain information about the initiator of the transfer that is aborted.
	 *	This can either be this J1939 stack, another ECU or unspecified.
	 *
	 *	The abort message now contains the role the sender of the abort message has in the (E)TP
	 *	session to be aborted.
	 */
	enum AbortMessageRoleOfSender
	{
		ABORT_BY_SENDER_OF_DATA 	= 0b00,
		ABORT_BY_RECEIVER_OF_DATA 	= 0b01,
		ABORT_ROLE_RESERVED		= 0b10,
		ABORT_ROLE_NOT_SPECIFIED	= 0b11		// for compatibility with older ECUs
	};

	/**
	 *	@brief This function sends a transport protocol abort message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_pgn The PGN of the abort message
	 *	@param p_dstAdr The destination address of the abort message
	 *	@param p_errorCode The reason why the message transfer was aborted
	 *	@param p_isETP false if normal transport protocol message. true if extended transport protocol message.
	 *	@param p_roleOfSender indicator if this J1939 stack or another ECU started the transfer, see enum AbortMessageRoleOfSender
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int sendAbortMessage(unsigned int p_pgn,
			     unsigned char p_dstAdr,
			     unsigned char p_errorCode,
			     bool p_isETP,
			     enum AbortMessageRoleOfSender p_roleOfSender);

	/**
	 *	@fn int sendACKMessage(unsigned int p_pgn,
			   unsigned char p_dstAdr,
			   unsigned int p_length,
			   unsigned char p_numPackages,
			   bool p_isETP)
	 *
	 *	@brief The functions sends an acknowledge message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_pgn The PGN of the acknowledged message
	 *	@param p_dstAdr The destination address of the acknowledge message
	 *	@param p_length The length of the message that is acknowledged
	 *	@param p_numPackages The number of packages into which the acknowledged message was splited.
	 *	@param p_isETP false if normal transport protocol message. true if extended transport protocol message.
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int sendACKMessage(unsigned int p_pgn,
			   unsigned char p_dstAdr,
			   unsigned int p_length,
			   unsigned char p_numPackages,
			   bool p_isETP);

	/**
	 *	@fn int sendDataPacketOffsetMessage(unsigned int p_pgn,
						    unsigned char p_dstAdr,
						    unsigned char p_maxNumPackages,
						    unsigned int p_packageOffset)
	 *
	 *	@brief The functions sends an data package offset message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_pgn The PGN of the acknowledged message
	 *	@param p_dstAdr The destination address of the acknowledge message
	 *	@param p_maxNumPackages The number of packages that have to be applied to the offset.
	 *	@param p_packageOffset The package offset to be send
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int sendDataPacketOffsetMessage(unsigned int p_pgn,
					unsigned char p_dstAdr,
					unsigned char p_maxNumPackages,
					unsigned int p_packageOffset);

	/**
	 *	@fn int WEJ1939TransportProtocolHandler::notifySendResult(unsigned int p_pgn, int p_result, unsigned int p_timestamp, const WEJ1939CANMessage * p_message)
	 *
	 *	@brief The functions notifies the result of sending a transport protocol message via event handler
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_pgn PGN of the message that was send
	 *	@param p_result Result of sending the message
	 *	@param p_timestamp Timestamp when the message was send
	 *	@param p_msg Pointer to a CAN message. The databytes will be extracted and written into
	 *		the event as a 64Bit value
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int notifySendResult(unsigned int p_pgn, int p_result, unsigned int p_timestamp, const WEJ1939CANMessage * p_msg);

	/**
	 *	@fn int handleRequestToSend(const WEJ1939CANMessage &p_msg)
	 *
	 *	@brief Function that has to be called for handling a request to send message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int handleRequestToSend(const WEJ1939CANMessage &p_msg);

	/**
	 *	@fn int handleClearToSend(const WEJ1939CANMessage &p_msg)
	 *
	 *	@brief Function that has to be called for handling a clear to send message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int handleClearToSend(const WEJ1939CANMessage &p_msg);

	/**
	 *	@fn int handleDataPackageOffset(const WEJ1939CANMessage &p_msg)
	 *
	 *	@brief Function that has to be called for handling a data package offset message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int handleDataPackageOffset(const WEJ1939CANMessage &p_msg);

	/**
	 *	@fn int handleACK(const WEJ1939CANMessage &p_msg)
	 *
	 *	@brief Function that has to be called for handling an acknowledgement message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int handleACK(const WEJ1939CANMessage &p_msg);

	/**
	 *	@fn int handleBAM(const WEJ1939CANMessage &p_msg)
	 *
	 *	@brief Function that has to be called for handling a BAM message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS.
	 *
	 */
	int handleBAM(const WEJ1939CANMessage &p_msg);

	/**
	 *	@fn int handleAbort(const WEJ1939CANMessage &p_msg)
	 *
	 *	@brief Function that has to be called for handling an abort message
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_OPERATION_IN_PROGRESS when the other node is already engaged with another connection
	 *	@returns WE_STATE_NOT_SUPPORTED_ERROR when the other node does not have the necessary resources
	 *	@returns WE_TIMEOUT_ERROR when a timeout has occurred
	 *	@returns WE_REQUEST_REJECTED_ERROR in all other cases
	 */
	int handleAbort(const WEJ1939CANMessage &p_msg);

	/**
	 *	@fn bool isSameProtocol(unsigned int pgnA, unsigned int pgnB)
	 *
	 *	@brief Returns if pgnA and pgnB belongs to the same transport protocol (BAM/TP or ETP)
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	We support three types of transport protocol: broadcast (BAM),
	 *	transport protocol (TP) and extended transport protocol (ETP).
	 *	BAM/TP and ETP are using different PGNs. But each is using more
	 *	than one PGN. For an easy check if two given PGNs, and thus
	 *	the belonging CAN messages, belong to the same protocol, just call this
	 *	function and provide the two PGNs. The return value indicates if the
	 *	function PGNs belong to the same protocol type.
	 *
	 *	@WARNING: This function can not distinguish between BAM and TP, because
	 *		  they are using same PGNs. For distinguishing BAM from TP it is
	 *		  recommended to verify the destination address.
	 *
	 *	@returns true if the given PGNs belong to the same protocol.
	 *	@returns false if the given PGNs do not belong to the same protocol.
	 */
	static bool isSameProtocol(unsigned int p_pgnA, unsigned int p_pgnB);


	/**
	 *	@fn bool isTP(unsigned int p_pgn)
	 *
	 *	@brief Figure out if PGN provided is TP connection management or data transfer PGN.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns true if "p_pgn" is a TP PGN
	 *	@returns false otherwise
	 */
	static bool isTP(unsigned int p_pgn);


	/**
	 *	@fn bool isETP(unsigned int p_pgn)
	 *
	 *	@brief Figure out if PGN provided is ETP connection management or data transfer PGN.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns true if "p_pgn" is a ETP PGN
	 *	@returns false otherwise
	 */
	// cppcheck-suppress unusedPrivateFunction
	static bool isETP(unsigned int p_pgn);


protected:
	WEJ1939Resources * resources;	//!< pointer to data that will not change while stack is running and can be shared with state machine.

	// send Attributes

	unsigned char numPackagesAllowedForCTS;	//!< Number of packages allowed in response to CTS - seems this is more or less a constant of what we allow.
						// only seems to be relevant for ETP whereas in TP same byte is "reserved"
						// meanwhile we allow 255 packages so that it could also be a constant set to 0xff
						// We could consider having different limits for TP and for ETP - but we anyways allow the maximum 255 for both.

	unsigned int sentProgress;	// Remove 'static' variable from 'handleMessage()'
	unsigned int lastSendTime;	// Remove 'static' variable from 'run()'

/*
 *
 * Translation message assembly			-->	old members
 *
 * *(this->msg) = WEJ1939CANMessage();			sendMsg
 * msg.setTimeRx()					sendTime
 * ecu->address						sendDestAddress
 * this->destAddress = J1939_BROADCAST_ADDRESS; --> unused
 * this->writePtr = nullptr;				sendPackageNo (counting packages of 7 bytes already sent)
 * this->nextPacketNo = 0;				sendSequenceNo
 * this->bytesReceived = 0;
 * this->timeout = 0;					timeout
 * this->numPackagesUntilCTS = 0xff;			numPackagesClearToSend
 * this->numPackagesForOffset = 0x0;			packetOffset
 * this->currentNumPackagesUntilCTS = 0xff;		sendPackageNoAtCTS
 * this->cm_pgn = 0;
 *
 */
	std::set< WEJ1939CanEcu * > tpSendList;		//!< List of ECUs with active (E)TPs sent to them currently.

	// receive Attributes
	std::list<struct MessageAssembly *> assemblyList;	//!< This List contains fragments of all message that are still not received completely

	// pointer to a registered TP event listener
	WESignalListener * eventListener;
};
}

#endif /* WE_J1939_TRANSPORT_PROTOCOL_HANDLER_H_ */
