/**
 *	@file we_aes.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Class that handles AES encryption.
 *
 *	This class provides an AES encryption algorithm. It uses EBC block mode and PKCS5 padding.
 *	Further its output is a base64 encoded string.
 */

#ifndef WE_AES_H_
#define WE_AES_H_

#include <string>

#include "aes_algo.h"

namespace WE
{

static const unsigned int keyLength = 16;

/**
 *	@class WEAES
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief Provides functions to encrypt a string using teh AES 128 algorithm. The
 *	       result is a base64 encoded string.
 *
 *	@namespace WE
 *
 *	To encrypt a string you need to do the following calls:
 *	  AES aes;
 *	  aes.startEncryption(secretKey);
 *	  aes.encrypt(plainText);
 *	  cipherText = aes.getEncryptedData();
 */
class WEAES : private AES
{
public:

	/**
	 *	@fn WEAES()
	 *
	 *	@brief Default constructor. Initializes pointers to nullptr and other members to sane
	 *		start values. Also calls Constructor of AES class.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	*/
	WEAES();

	/**
	 *	@fn ~WEAES()
	 *
	 *	@brief Destructor. Frees all occupied memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	*/
	virtual ~WEAES() = default;

	/**
	 *	@fn int startEncryption(const std::string &p_key)
	 *
	 *	@brief Initialize class for encryption
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	call this before any encryption with the key to use
	 *
	 *	@param p_key The key string. Its length has to be 16.
	 *
	 *	@return WE_SUCCESS on successful initialization of the encryption
	 *	@return WE_INVALID_PARAMETER_ERROR when the key length is not 16 characters.
	*/
	int startEncryption(const std::string &p_key);

	/**
	 *	@fn void encrypt(const std::string &p_plainText)
	 *
	 *	@brief encrypt a single block (default 128 bits, or unsigned char[16]) of data
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	call this to encrypt the given text.
	 *
	 *	@param p_plainText The text that has to be encrypted.
	 *
	 *	@return WE_SUCCESS when the encryption was done successful
	 *	@return WE_GENERAL_ERROR when the encryption failed.
	*/
	int encrypt(const std::string &p_plainText);

	/**
	 *	@fn const std::string& getEncryptedData()
	 *
	 *	@brief Delivers the AES encrypted and base64 encoded data string
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return a reference to the AES encrypted and base64 encoded data string.
	*/
	const std::string& getEncryptedData();

	// NOTE: Decryption currently not implemented

private:
	/**
	 *	@fn unsigned char * convertPlainText(const std::string &p_text, int &p_length)
	 *
	 *	@brief Converts a string into a AES algorithm conform c-style string.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	This function copies p_text into a c-style string and adds padding bytes.
	 *	The resulting sting length will be a multiple of the AES lbock size (16 Bytes).
	 *
	 *	Padding Rules (PKCS5/PKCS7 padding):
	 *	  - If number of characters in text is a multiple of 16, add 16 padding characters.
	 *	  - If number of characters in text is not a multiple of 16, add n padding characters, so that
	 *	    text.length() + n is a multiple of 16 and n < 16.
	 *
	 *	WARNING: The resulting c-style string is not 0 terminated!
	 *
	 *	@param p_text The string to be converted
	 *	@param p_length The length of the converted string
	 *
	 *	@returns Pointer to the converted string. WARNING: The result is a non \0 terminated string!
	*/
	unsigned char * convertPlainText(const std::string &p_text, int &p_length);

	/**
	 *	@fn int encodeBase64(std::string &p_encodedCipher, const unsigned char *p_cipher, int p_cipherLength)
	 *
	 *	@brief Converts a c-style string into a base64 encoded string.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[out] p_encodedCipher The base64 encoded cipher text.
	 *	@param[in] p_cipher The c-style cipher text
	 *	@param[in] p_cipherLength The length of the cipher text p_cipher
	 *
	 *	@returns WE_SUCCESS on successful converting the cipher text.
	 *	@returns WE_GERNERAL_ERROR when encoding fails (This case should never happen - see comments in implementation).
	*/
	int encodeBase64(std::string &p_encodedCipher, const unsigned char *p_cipher, int p_cipherLength);

	unsigned char key[keyLength];	// the secret key used for encryption or decryption
	std::string text;		// contains the cipher text after encryption or the plain Text after decryption
};
}

#endif /* WE_AES_H_ */
