/*
 * we_debug.h
 *
 *  Created on: 23.02.2012
 *      Author: mka
 *
 *
debug.ini example:

//general section
[debug_defines]
we_log_to_console=0 	//If you want to log to console put a 1 here
we_log_to_file=1 		//If you want to log to a file put a 1 here
we_log_max_file_size=24599 //Max file size. If size is reached the file is reset or a new file is created when we_log_continues_logging = 1
we_log_continues_logging=0 //If you want to have several files with a max filesize of we_log_max_file_size put a 1 here

//debug section
[debug]
we_general_log_level=5 //log level according to the levels in the debug defines header
//Here starts the sections
we_socket_linux=1
we_multiple_socket_server=1
we_simple_socket_client=1

//General usage:
 * this->myDebugKey = WEDebug::insertNewDebugSection("we_socket_linux");

	you can create new debug sections on the fly using this call. If the section already exists you get back the found key.
	You get back the current log section key which have to be used later when
	you want to log something:

	WEDebug::debugPrintError(WE_DEBUG_TRACE,this->myDebugKey,"WESocketLinux::getNextInReceiveList");

	This call is made with level of the message, the key, and the string.
	The key can be held as class memeber or however you want it.

	You can do something like that, if you do not want to store the key by yourself:

	WEDebug::debugPrintError(WE_DEBUG_TRACE, WEDebug::getDebugSectionKey("we_socket_linux"), "WESocketMessage::WESocketMessage CopyCon create paylod size %d ",p_message.dataPayloadSize);

	This can have impact on the performance!! Because the getDebugSectionKey uses a find call to a map.
 */

#ifndef WE_DEBUG_H_
#define WE_DEBUG_H_

#include <sstream>
#include <cstring>
//included in header: So you need only to include debug.h to use the debug out
#include "we_debug_defines.h"

namespace WE
{

class WEDebug
{
public:

	virtual ~WEDebug() = default;

	/**
	 *	@fn static void debugPrintError(int p_level, unsigned int p_section, const char* p_format, ...)
	 *
	 *	@brief call to debug print out
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param p_level debug level where the current message belongs too. see we_debug_defines.h for some hints on levels
	 *	@param p_section key to the section which the message belongs too. Sections can be defined as you need them for your modules.
	 *	@param p_format const string. For syntax see sprintf
	 *	@param ... various args which are representing variables which are build into the const char. See sprintf for details
	 *
	 *	@returns N/A
	 */
	static void debugPrintError(int p_level, unsigned int p_section, const char* p_format, ...);

	//wrapper method so you need only to include we_debug.h. See debug_config.h
	static int insertNewDebugSection(const std::string & p_section);

	//wrapper method so you need only to include we_debug.h. See debug_config.h
	static int getDebugSectionKey(const std::string & p_section);

	//wrapper method to get the actual ini value of that debug section
	static int getIniValPerSection(int p_section);

	/**
	 *	@fn static int getDebugLevelOfSection(int p_section)
	 *
	 *	@brief Figure out the debug level for p_section
	 *
	 *	@param p_section key to the section/module whose current debug level should be returned
	 *
	 *	@returns The debug level of p_section.
	 *
	 *
	 *	This method can be used to e.g. figure out if a complex debug message should really be
	 *	created or if WEDebug would throw it away anyways.
	 *
	 *	If level returned is greater or equal than the level used for the message than it would
	 *	be added to the debug output.
	 *
	 *	Could be used like this:
	 *
	 *	if (WE::WE_DEBUG_INFO <= WE::WEDebug::getDebugLevelOfSection(weDebugSection))
	 *	{
	 *		std::string complexDebug;
	 *		// do some complex collection of information into string "complexDebug"
	 *		WE::WEDebug::debugPrintError(WE::WE_DEBUG_INFO, weDebugSection, "complex debug: '%s'.", complexDebug.c_str());
	 *	}
	 */
	static int getDebugLevelOfSection(int p_section);

private:
	WEDebug() = delete;
};
}

#endif /* WE_DEBUG_H_ */
