/**
 *	@file we_files_and_folders.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2010
 *
 *	@brief Helper functions for file and directory operations
 */

#ifndef WE_FILES_AND_FOLDERS_H_
#define WE_FILES_AND_FOLDERS_H_

#include <string>
#include <vector>

namespace WE
{

enum FileFilter
{
	FILE_FILTER_NONE,
	FILE_FILTER_FILES_ONLY,
	FILE_FILTER_DIR_ONLY
};

enum FileTime
{
	TIME_OF_LAST_ACCESS = 1,
	TIME_OF_LAST_MODIFICATION,
	TIME_OF_LAST_STATUS_CHANGE
};
/**
 *	@class WEFilesAndFolders
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Class interface that provides helper functions for different file and directory operations.
 *
 *	@namespace WE
 *
 *	@warning Functions isDirectoryOnExternalMedium, mountDevice and unMountDevice will
 *		only work properly on OPUS terminals and not on a PC.
 *	@warning Be very careful when using these functions and think twice if the end user shall
 *		be able to access these functions. Especially when running an application with "root" permissions.
 *		Either use these functions only internally or restrict the inputs to pre-defined files and folders.
 *		Very bad things can happen if a user has free access to these function. Don't complain! I warned you!
 *
 *	This class consists only of static functions and no members. Therefore it is not
 *	necessary to create an instance of this class (actually it is not possible because the constructor is private).
 *	Just include this header file and call the functions like this: WEFileAndFolders::functionName(...)
 */
class WEFilesAndFolders
{
public:
	/**
	 *	@fn WEFilesAndFolders()
	 *
	 *	@brief Deleted constructor of WEFilesAndFolders. An instance of this class is not necessary.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WEFilesAndFolders() = delete;
	~WEFilesAndFolders() = delete;

	/**
	 *	@fn static bool isDirectory(const std::string & p_path)
	 *
	 *	@brief Test if provided path is a directory
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_path Filesystem path that shall be tested if it is a directory
	 *
	 *	@returns true if provided path is a directory
	 *	@returns false if the provided path is a file or can not be found
	 */
	static bool isDirectory(const std::string & p_path);


	/**
	 *	@fn static bool isFile(const std::string & p_path)
	 *
	 *	@brief Test if provided path is a file
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_path Filesystem path that shall be tested if it is a file
	 *
	 *	@returns true if provided path is a file
	 *	@returns false if the provided path is a directory or can not be found
	 */
	static bool isFile(const std::string & p_path);


	/**
	 *	@fn static int isDirectoryOnExternalMedium(const std::string & p_path)
	 *
	 *	@brief Test if a provided file system path is located on an external medium
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_path File system path that shall be checked if it is located on an external medium
	 *
	 *	@returns WE_SUCCESS if provided file system path is located on an external medium
	 *	@returns other error code if the provided path is a local directory
	 *
	 *	@warning Only works properly on supported OPUS terminals!
	 *	@warning Only works for ONE USB stick!
	 *
	 *	This function checks if a provided file system path is located on an external medium
	 *	like USB stick or SDCard.
	 */
	static int isDirectoryOnExternalMedium(const std::string & p_path);


	/**
	 *	@fn static long long getFileSize(const std::string & p_filePath, std::string* p_error = nullptr)
	 *
	 *	@brief Reurns the file size in bytes of the provided file name
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_filePath Path to the file that's size shall be returned
	 *	@param p_error contains error string if return value is -1 and a buffer is given <MKA 06.03.2014>
	 *		 *
	 *	@returns File size in bytes
	 *	@returns -1 if an error occurred
	 */
	static long long getFileSize(const std::string & p_filePath, std::string* p_error = nullptr);


	static unsigned int getFileCreationTime(const std::string & p_filePath);

	/**
	 *	@fn static unsigned int getFileTime(const std::string & p_filePath, int p_time, std::string* p_error = nullptr)
	 *
	 *	@brief returns file time based on the FileTime enum. You can choose last access, last modified or last status change
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param p_filePath Path to the file that's size shall be returned
	 *	@param p_time based on FileTime enum
	 *	@param p_error contains error string if return value is -1 and a buffer is given <MKA 06.03.2014>
	 *
	 *	@returns File time or on error -1 / 0
	 *
	 *	@see FileTime
	 */
	static unsigned int getFileTime(const std::string & p_filePath, int p_time, std::string* p_error = nullptr);

	/**
	 *
	 * The getFileCountAndOldestFile(const std::string & p_folderName, const std::string & p_fileExtension, std::string* p_oldestFile, long long* p_totalFileSize) is platform independed.
	 * It might be a good idea to have a common files and folders class, but for only one method??
	 * Well therefore we have copied code inside the Linux and Windows files_and_folders class.
	 *
	 *	@fn static int getFileCountAndOldestFile(const std::string & p_folderName, const std::string & p_fileExtension, std::string* p_oldestFile, long long* p_totalFileSize)
	 *
	 *	@brief returns file count of files with specific file extension and oldest file in specified folder. No wildcards please!
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param p_folderName folder of files to count
	 *	@param p_fileExtension for example ".txt", ".log"
	 *	@param p_oldestFile used for writing in the oldestfile name
	 *	@param p_totalFileSize size in bytes of all count files
	 *
	 *	@returns filecount
	 */
	static int getFileCountAndOldestFile(const std::string & p_folderName, const std::string & p_fileExtension, std::string* p_oldestFile, long long* p_totalFileSize);


	/**
	 *	@fn static std::string getParentDirectoryPath(const std::string & p_filepath)
	 *
	 *	@brief Return the parent directory of the specified filepath.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@returns std::string that contains the parent working directory without the ending slash
	 */
	static std::string getParentDirectoryPath(const std::string & p_filepath);


	/**
	 *	@fn static long long getFreeSpaceOfDirectory(const std::string & p_directoryPath)
	 *
	 *	@brief Return the available free space of a partition by passing any directory path as parameter
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_directoryPath path to a directory on a partition that's free space shall be returned
	 *
	 *	@returns Available disk space in kilobytes
	 *	@returns -1 if an error occurred (e.g. directory does not exist)
	 */
	static long long getFreeSpaceOfDirectory(const std::string & p_directoryPath);

	/**
	 * Get the size of a directory in bytes
	 * @warning This function might run for a long time if p_directoryPath has a very large directory hierarchy (recursion is used)
	 * @param p_directoryPath Path of the directory
	 * @returns Size of the directory in bytes
	 */
	static long long getDirectorySize(const std::string & p_directoryPath);


	/**
	 *	@fn static long long getTotalSpaceOfPartition(const std::string & p_partition)
	 *
	 *	@brief Returns the total storage space of a partition
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_partition path of the partition that's total storage space shall be returned
	 *
	 *	@returns Total storage space of a partition in kilobytes
	 *	@returns -1 if an error occurred (e.g. partition does not exist)
	 */
	static long long getTotalSpaceOfPartition(const std::string & p_partition);


	/**
	 *	@brief Write p_fileContent to p_filePath. Create file if required.
	 *
	 *	@param p_filePath - variable whose value should be written to file
	 *	@param p_filePath - variable whose value should be written to file
	 *
	 *	@return WE::WE_SUCCESS in case of success
	 *	@return WE::WE_ALREADY_EXISTS_ERROR if file already contains p_fileContent
	 *	@return WE::WE_FILE_ERROR in case file could not be truncated, written or closed
	 *	@return WE::WE_INVALID_RESPONSE_ERROR if final sync didn't work
	 *
	 *	This method writes the value p_fileContent to file p_filePath
	 *	if it differs from the existing value.
	 *	If required, file is created first.
	 *	On Linux it is made sure that the file content is written to disk (flash) immediately
	 *	in order to avoid loss of data in case of power drop.
	 *
	 *	This method was introduced when BSP 2.x started using UBIFS and we figured out that
	 *	we lost data when writing the same value to file again and did not sync after writing.
	 *	This was caused by UBIFS directly truncating existing files but using write buffers
	 *	for the data. Like that a file could be reduced to size 0 by just writing the same
	 *	content again and then switching off power.
	 *	If power is switched off in the right moment this could still happen with this method
	 *	but it's way less likely than before - so it's highly recommended to shut down gracefully.
	 */
	static int writeStringToFile(const std::string & p_filePath, const std::string & p_fileContent);


	/**
	 *	@fn static int copyFileOrFolder(const std::string & p_source, const std::string & p_destination)
	 *
	 *	@brief Copy a file or folder from p_source to p_destination
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_source File or folder that shall be copied (don't make it a const reference: will be modified in function)
	 *	@param p_destination File or folder name that identifies the copy destination (don't make it a const reference: will be modified in function)
	 *
	 *	@returns WE_SUCCESS if copying was successful
	 *	@returns WE_FILE_ERROR if source file or folder was not found
	 *	@returns WE_WRITE_ERROR if copying to p_destination failed
	 *
	 *	Use this file to copy a file or a folder to another location
	 */
	static int copyFileOrFolder(std::string p_source, std::string p_destination);


	/**
	 *	@fn static int linkFile(const std::string & p_source, const std::string & p_destination)
	 *
	 *	@brief Create link of a file p_source to file p_destination.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_source File or folder that shall be copied (don't make it a const reference: will be modified in function)
	 *	@param p_destination File or folder name that identifies the copy destination (don't make it a const reference: will be modified in function)
	 *	@param p_softLink Don't do a hard link but create a softlink
	 *
	 *	@returns WE_SUCCESS if copying was successful
	 *	@returns WE_FILE_ERROR if source file or folder was not found
	 *	@returns WE_WRITE_ERROR if copying to p_destination failed (default to false => create hardlink)
	 *
	 *	This method creates links for one or more files.
	 *	The default is to create a hard link (if p_softLink is false or empty).
	 *	Please note that hard links to folders are typically not allowed.
	 *	Set p_softLink to true to create a softlink (only supported on Linux operating system)
	 *
	 *	@Warning Depending on the underlying file system this may fail or fall back to copying the file(s) instead of creating a hard link.
	 *
	 */
	static int linkFile(std::string p_source, std::string p_destination, bool p_softLink = false);


	/**
	 *	@fn static int moveFileOrFolder(const std::string & p_source, const std::string & p_destination)
	 *
	 *	@brief Move a file or folder from p_source to p_destination
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_source File or folder that shall be moved
	 *	@param p_destination File or folder name that identifies the move destination
	 *
	 *	@returns WE_SUCCESS if moving the file or folder was successful
	 *	@returns return value of copyFileOrFolder() if the copying process failed
	 *	@returns return value of eraseFileOrFolder() if erasing the source file failed
	 *
	 *	@see copyFileOrFolder
	 *	@see eraseFileOrFolder
	 */
	static int moveFileOrFolder(const std::string & p_source, const std::string & p_destination);


	/**
	 *	@fn static int eraseFileOrFolder(const std::string & p_fileToErase, bool p_onlyContents = false, const std::string & p_filterSubString = "")
	 *
	 *	@brief Erase a file or a folder on the file system
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_fileToErase Path to the file/folder that shall be erased completely along with contents (don't make it a const reference: will be modified in function)
	 *	@param p_onlyContents True, deletes only the contents of p_fileToErase folder but keep the p_fileToErase folder, Default is False.
	 *	@param p_filterSubString Delete files that have this as substring in their name (e.g; .pool). Applicable only if p_onlyContents is True. Default is ""
	 *
	 *	@returns WE_SUCCESS if erasing the file was successful
	 *	@returns WE_FILE_ERROR if p_fileToErase was not found
	 *	@returns WE_WRITE_ERROR if removing the file failed
	 *
	 *	NOTE: This does not handle wildcards in the path
	 */
	static int eraseFileOrFolder(std::string p_fileToErase, bool p_onlyContents = false, const std::string & p_filterSubString = "");


	/**
	 *	@fn static int mountDevice(const std::string & p_deviceFile, const std::string & p_mountPath)
	 *
	 *	@brief Mount a device to the local file system (like USB stick or SD Card if supported)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_deviceFile Path to the device file that shall be mounted
	 *	@param p_mountPath Path to an existing directory of the filesystem to where the device shall be mounted
	 *
	 *	@returns WE_SUCCESS if mounting the device was successful
	 *	@returns WE_READ_ERROR if the mount table could not be opened for reading
	 *	@returns WE_DEVICE_ERROR if mounting the device failed
	 *
	 *	@warning Only mount devices formatted with FAT(32) file system!
	 */
	static int mountDevice(const std::string & p_deviceFile, const std::string & p_mountPath);


	/**
	 *	@fn static int unMountDevice(const std::string & p_deviceFile)
	 *
	 *	@brief With this function, a mounted device can be unmounted
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_deviceFile Path of the device file that shall be unmounted
	 *
	 *	@returns WE_SUCCESS if unmounting the device was successful
	 *	@returns WE_READ_ERROR if the mount table could not be opened for reading
	 *	@returns WE_DEVICE_ERROR if the device is not mounted or if unmounting the device failed
	 *
	 *	@warning Only succeeds if the device file is really mounted
	 */
	static int unMountDevice(const std::string & p_deviceFile);


	/**
	 *	@fn static std::string getCurrentWorkingDirectory(void)
	 *
	 *	@brief Return the working directory of the process this function was called in
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns std::string that contains the current working directory
	 */
	static std::string getCurrentWorkingDirectory(void);


	/**
	 *	@fn static int createDirectory(const std::string & p_path)
	 *
	 *	@brief Create a directory
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_path Path of the directory that shall be created
	 *
	 *	@returns WE_SUCCESS if creating the directory was successful
	 *	@returns WE_FILE_ERROR if the directory could not be created
	 *	@returns WE_ALREADY_EXISTS_ERROR if the directory already existed
	 *
	 *	If required this also creates parent directories.
	 */
	static int createDirectory(const std::string & p_path);


	/**
	 *	@fn static std::string changeSlashes(const std::string & p_path)
	 *
	 *	@brief Change slashes from \ to / , if any
	 *
	 *	@param p_path Path of the directory that shall be changed
	 *
	 *	@returns changed or unchanged path
	 *	We assume that we are here in a class for Linux so we want to change \\ to /
	 */
	static std::string changeSlashes(const std::string & p_path);


	/**
	 *	@fn static std::string addSlash(std::string p_path)
	 *
	 *	@brief Check if at the end is a slash, if not add one.
	 *
	 *	@param p_path string of the directory that shall be changed
	 *
	 *	@returns changed or unchanged path
	 */
	static std::string addSlash(const std::string & p_path);


	/**
	 *	@fn static std::string trimSlash(std::string p_path)
	 *
	 *	@brief Check if at the beginning is a slash, if so kill it.
	 *
	 *	@param p_path string of the directory that shall be changed
	 *
	 *	@returns changed or unchanged path
	 */
	static std::string trimSlash(const std::string & p_path);


	/**
	 *	@fn static std::string stripSlash(std::string p_path)
	 *
	 *	@brief Check if at the end is a slash, if so kill it.
	 *
	 *	@param p_path string of the directory that shall be changed
	 *
	 *	@returns changed or unchanged path
	 */
	static std::string stripSlash(const std::string & p_path);


	/**
	 *	@fn static std::string getDeviceFromPath(std::string p_path)
	 *
	 *	@brief check the mount table if you find the path and return device
	 *
	 *	@param p_path string of the directory that shall be changed
	 *
	 *	@returns device name
	 *
	 *	Note: THE PATH TO BE CHECKED SHALL BE WITHOUT / AT THE END! USE stripSlash before
	 */
	static std::string getDeviceFromPath(const std::string & p_path);


	/**
	 *	@fn static char getDirDelimiter()
	 *
	 *	@returns current valid folder delimiter for that system
	 */
	static char getDirDelimiter();


	/**
	 *	@fn static int getAllFilesOrFoldersInFolder(
			std::string p_folderName,
			std::vector<std::string> & p_files,
			unsigned int p_maxLength = 0,
			FileFilter p_filter = FILE_FILTER_NONE)
	 *
	 *	@brief Get all file or folder names in p_folderName which consist of less than p_maxLength characters
	 *
	 *	@param p_folderName string of the directory that shall be investigated
	 *	@param p_files Vector of strings to be filled with names
	 *	@param p_maxLength Max length a name may have to be added to p_files. 0 for "don't care"
	 *
	 *	@returns WE::WE_SUCCESS in case of success
	 *	@returns WE::WE_NO_DIRECTORY if p_folderName does not point to a folder
	 *
	 *	Note: THE PATH TO BE CHECKED SHALL BE WITHOUT / AT THE END! USE stripSlash before
	 */
	static int getAllFilesOrFoldersInFolder(
			const std::string & p_folderName,
			std::vector<std::string> & p_files,
			unsigned int p_maxLength = 0,
			FileFilter p_filter = FILE_FILTER_NONE);

	/**
	 *	@fn static std::string getApplicationTempDirectory(void)
	 *
	 *	@brief Get the path to the application's temp directory
	 *
	 *	@returns A string containing the path to the tmp directory
	 */
	static const std::string & getApplicationTempDirectory(void);

	/**
	 * Get the application data partition that was defined at compile time.
	 * The application data partition can be configured by setting the APP_DATA_PARTITION
	 * environment variable before executing make. If it is not set the default "/opt" is used.
	 */
	static const std::string & getDataPartition(void);

	static const std::string USB_DEVICE_PREFIX; //!< const string that holds the default prefix for USB stick device files, only usable on OPUS terminals and for the first USB stick
	static const std::string SD_DEVICE_PREFIX; //!< const string that holds the default prefix for SDCard device files, only usable on OPUS terminals (currently only OPUS A5)

	static const char WE_LINUX_DIR_DELIMITER; //!< Definition of the Linux directory delimiter ('/')
	static const char WE_WINDOWS_DIR_DELIMITER; //!< Definition of the Windows directory delimiter ('\')

	/*static int getAllFilesWithTimestampsInFolder(
			std::string p_folderName,
			std::vector<std::string> & p_files,
			std::vector<long> & p_timestamps,
			unsigned int p_maxLength = 0,
			FileFilter p_filter = FILE_FILTER_NONE);*/


	static int sync();

	static int syncFileToDisk(const std::string & p_filePath);

	static int debugSection;
private:


	static const std::string DATA_PARTITION;
};
} // namespace WE

#endif /* WE_FILES_AND_FOLDERS_H_ */
