/**
 *	@file helper_functions.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 */

#ifndef HELPER_FUNCTIONS_H_
#define HELPER_FUNCTIONS_H_

#include <string>
#include <algorithm>
#include <cctype>
#include <locale>

namespace WE
{

/**
 *	@enum StorageInfo
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Definition of file storage information
 *
 *	@namespace WE
 */
typedef enum
{
	STORAGE_TOTAL,	//!< Total storage (used + free)
	STORAGE_FREE,	//!< Free storage available
	STORAGE_USED	//!< Currently used storage
}StorageInfo;

/**
 *	@fn int readStringFromFile(std::string p_filePath, std::string * p_storeString)
 *
 *	@brief Function to read exactly one line from a file and store this line into a string
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@param p_filePath Absolute file path of the file that shall be read
 *	@param p_storeString A pointer to a string in which the read line will be stored
 *
 *	@returns WE_SUCCESS if reading the line was successful
 *	@returns WE_FILE_ERROR if reading failed
 */
int readStringFromFile(const std::string & p_filePath, std::string * p_storeString);

int writeToFile(const std::string & p_filePath, const std::string & p_writeString);
int writeToFile(const std::string & p_filePath, int p_numberToWrite);

/**
 *	@fn int readIntegerFromFile(std::string p_filePath, int * p_storeInt)
 *
 *	@brief Function to read exactly one integer from a file and store it into a integer variable
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@param p_filePath Absolute file path of the file that shall be read
 *	@param p_storeInt A pointer to a integer in which the read number will be stored
 *
 *	@returns WE_SUCCESS if reading the line was successful
 *	@returns WE_FILE_ERROR if reading failed
 */
int readIntegerFromFile(const std::string & p_filePath, int * p_storeInt);

/**
 *	@fn int getMountPath(std::string p_device, std::string * p_resultString)
 *
 *	@brief Retrieve the file system path on which a device is mounted
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@param p_device The full path of a mass storage device (e.g. "/dev/sda1")
 *	@param p_resultString Pointer to a string in which the resulting file system path will be stored
 *
 *	@returns WE_SUCCESS if retrieving the information was successful
 *	@returns WE_FILE_ERROR if the mount table could not be opened (Linux)
 *	@returns WE_NOT_FOUND_ERROR if the device is NOT mounted
 *
 */
int getMountPath(const std::string & p_device, std::string * p_resultString);

/**
 *	@fn bool isAnUSBStickMounted(std::string * p_mountPathToReturn = nullptr)
 *
 *	@brief Check if an USB drive is mounted
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@param p_mountPathToReturn if this pointer is not nullptr, the mount path of the USB stick will be stored there
 *
 *	@returns true if an USB stick is mounted
 *	@returns false if no USB stick is mounted
 */
bool isAnUSBStickMounted(std::string * p_mountPathToReturn = nullptr);

/**
 *	@fn int getDiskSpace(StorageInfo what, std::string fileSystemPath)
 *
 *	@brief Retrieve the free/used/total disk space on a file system
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@param what either STORAGE_TOTAL, STORAGE_FREE or STORAGE_USED
 *	@param fileSystemPath A path on any partition of the system
 *
 *	@returns -1 if any error occurred
 *	@returns Either the used, free or total disk space used/available on the specified file path (in MByte)
 */
int getDiskSpace(StorageInfo p_what, const std::string & p_fileSystemPath);

/**
 *	@fn std::string getLineFromFile(std::string p_file)
 *
 *	@brief Read first line of a file and return it
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@param p_file Absolute path of a file that's first line shall be read
 *
 *	@returns The first line of p_file or an empty string
 */
std::string getLineFromFile(const std::string & p_file);

/**
 *	@fn static void printBacktrace(int p_length = 32, int p_suppress = 1)
 *
 *	@brief Prints the current stack to stderr when calling the function
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@param p_length defines the number of lines printed for the backtrace
 *	@param p_suppress defines how much many lines of the printout are suppressed (e.g. to hide the backtrace function itself)
 */
void printBacktrace(int p_length = 32, int p_suppress = 1);

const std::string toString(int p_number);
int toInt(const std::string & p_string);
double toDouble(const std::string & p_string);

/**
 * Trim (remove) white space at left side of string.
 * @param s Reference to a string to remove the white space from.
 */
static inline void ltrim(std::string &p_s)
{
	p_s.erase(p_s.begin(), std::find_if(p_s.begin(),
					p_s.end(),
					[](unsigned char p_ch)
					{//AXIVION Line CodingStyle-AllmanBraces : Ignore for lambdas
						return !std::isspace(p_ch);
					})
	);
}

/**
 * Trim (remove) white space at right side of string.
 * @param s Reference to a string to remove the white space from.
 */
static inline void rtrim(std::string &p_s)
{
	p_s.erase(std::find_if(p_s.rbegin(),
			     p_s.rend(),
			     [](unsigned char p_ch)
			     {//AXIVION Line CodingStyle-AllmanBraces : Ignore for lambdas
				return !std::isspace(p_ch); //AXIVION Line CodingStyle-AllmanBraces : Ignore for lambdas
			     }).base(),
			     p_s.end()
	);
}

/**
 * Trim (remove) white space at both sides of string.
 * @param s Reference to a string to remove the white space from.
 */
static inline void trim(std::string &p_s)
{
	ltrim(p_s);
	rtrim(p_s);
}

/**
 * Helper function to print a line to kernel log ("/dev/kmsg") which can
 * then be read via dmesg or /var/log/messages. After calling this function
 * once, the file stream to /dev/kmsg stays open.
 *
 * @WARNING use only for debugging purposes!
 */
void writeDMESGLogEntry(const std::string & p_str);
}

#endif /* HELPER_FUNCTIONS_H_ */
