/**
 *	@file thread_safe_queue.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of a thread safe queue
 *
 */


#ifndef THREAD_SAFE_QUEUE_H_
#define THREAD_SAFE_QUEUE_H_

#include <condition_variable>
#include <queue>

namespace WE {

/**
 *	@class WEThreadSafeQueue
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief A thread safe wrapper around a std::queue
 *
 *	@namespace WE
 *
 *	This class is a wrapper around a std::queue that should make it thread safe.
 *	Except for the isEmpty() and size() methods, this code was copied from some
 *	website that I do not remember :(
 *
 *	@warning: link against boost_thread when using this header!
 *
 */
template <typename T>
class WEThreadSafeQueue
{
private:
	std::queue<T> queue{};
	std::mutex mtx{};
	std::condition_variable conditionVariable{};

public:
	WEThreadSafeQueue() = default;
	~WEThreadSafeQueue() = default;
	void push(const T& p_data)
	{
		std::lock_guard<std::mutex> l(mtx);
		queue.push(p_data);
		conditionVariable.notify_one();
	}

	T pop()
	{
		std::unique_lock<std::mutex> l(mtx);
		if(queue.size() == 0)
		{
			conditionVariable.wait(l);
		}

		T res = queue.front();
		queue.pop();
		return res;
	}

	bool isEmpty(void) const
	{
		return queue.empty();
	}

	int size(void) const
	{
		return queue.size();
	}
};
}  // namespace WE


#endif /* THREAD_SAFE_QUEUE_H_ */
