/**
 *	@file we_time.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief WETime provides static functions for retrieving time and date information from the real time clock.
 *
 */

#ifndef WE_TIME_H_
#define WE_TIME_H_

#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>

namespace WE
{

typedef enum
{
	WE_DAYTIME_AM = 0,
	WE_DAYTIME_PM = 1
} WEDayTime;


/**
 *	@struct WETimeInfo
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Struct containing time information
 *
 *	@namespace WE
 */
struct WETimeInfo
{
	int hours{0};	//!< hours of the day 0 to 23
	int minutes{0};	//!< minutes 0 to 59
	int seconds{0};	//!< seconds 0 to 59

	virtual ~WETimeInfo() = default;

private:
	//SERIALIZATION
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int )
	{
		p_ar & hours;
		p_ar & minutes;
		p_ar & seconds;
	}
};

/**
 *	@struct WEDateInfo
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Struct containing date information
 *
 *	@namespace WE
 */
struct WEDateInfo
{
	int year{0};	//!< year >= 2010 (in 32bit Linux: < 2036)
	int month{0};	//!< month of year 1 to 12
	int days{0};	//!< day of month 1 to 31

	virtual ~WEDateInfo() = default;
private:
	//SERIALIZATION
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int )
	{
		p_ar & year;
		p_ar & month;
		p_ar & days;
	}
};

/**
 *	@struct WETimeAndDateInfo
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Struct containing time and date information
 *
 *	@namespace WE
 */
struct WETimeAndDateInfo
{
	int year{0};	//!< year >= 2010 (in 32bit Linux: < 2036)
	int month{0};	//!< month of year 1 to 12
	int days{0};	//!< day of month 1 to 31
	int hours{0};	//!< hours of the day 0 to 23
	int minutes{0};	//!< minutes 0 to 59
	int seconds{0};	//!< seconds 0 to 59

	virtual ~WETimeAndDateInfo() = default;

private:
	//SERIALIZATION
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int )
	{
		p_ar & year;
		p_ar & month;
		p_ar & days;
		p_ar & hours;
		p_ar & minutes;
		p_ar & seconds;
	}
};

/**
 *	@class WETime
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief WETime provides static functions for retrieving time information from the real time clock.
 *
 *	@namespace WE
 *
 *	@warning Do not try to instantiate this class, only use the static functions.
 *
 *	The functions of WETime can be called like this:
 *	int seconds = WE::WETime::getSeconds();
 */
class WETime
{
public:
	/**
	 *	@fn static int getSeconds(unsigned long long p_timestamp = 0)
	 *
	 *	@brief Retrieve second information of the real time clock (RTC).
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to retrieve the wanted information
	 *
	 *	@returns Integer value representing the "seconds" of the real time clock.
	 */
	static int getSeconds(unsigned long long p_timestamp = 0);

	/**
	 *	@fn static int getMinutes(unsigned long long p_timestamp = 0)
	 *
	 *	@brief Retrieve minute information of the real time clock (RTC).
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to retrieve the wanted information
	 *
	 *	@returns Integer value representing the "minutes" of the real time clock.
	 */
	static int getMinutes(unsigned long long p_timestamp = 0);

	/**
	 *	@fn static int getHours(unsigned long long p_timestamp = 0)
	 *
	 *	@brief Retrieve hour information of the real time clock (RTC).
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to retrieve the wanted information
	 *
	 *	@returns Integer value representing the "hours" of the real time clock.
	 */
	static int getHours(unsigned long long p_timestamp = 0);

	/**
	 *	@fn static int getHours12h(unsigned long long p_timestamp = 0)
	 *
	 *	@brief Retrieve hour information of the real time clock (RTC) at 12h mode.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to retrieve the wanted information
	 *
	 *	@returns Integer value representing the "hours" (1 -12) of the real time clock.
	 */
	static int getHours12h(unsigned long long p_timestamp = 0);

	/**
	 *	@fn static int getDays(unsigned long long p_timestamp = 0)
	 *
	 *	@brief Retrieve day information of the real time clock (RTC).
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to retrieve the wanted information
	 *
	 *	@returns Integer value representing the "days" (day of month) of the real time clock.
	 */
	static int getDays(unsigned long long p_timestamp = 0);

	/**
	 *	@fn static int getMonth(unsigned long long p_timestamp = 0)
	 *
	 *	@brief Retrieve month information of the real time clock (RTC).
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to retrieve the wanted information
	 *
	 *	@returns Integer value representing the "month" (1 to 12) of the real time clock.
	 */
	static int getMonth(unsigned long long p_timestamp = 0);

	/**
	 *	@fn static int getYear(unsigned long long p_timestamp = 0)
	 *
	 *	@brief Retrieve year information of the real time clock (RTC).
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to retrieve the wanted information
	 *
	 *	@returns Integer value representing the "year" (current year, e.g. 2009) of the real time clock.
	 */
	static int getYear(unsigned long long p_timestamp = 0);

	/**
	 *	@fn static WETimeAndDateInfo getTimeAndDate()
	 *
	 *	@brief Returns a WETimeAndDateInfo struct filed with current date and time
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WETimeAndDateInfo struct filled with current date and time
	 *
	 *	@see WETimeAndDateInfo
	 */
	static WETimeAndDateInfo getTimeAndDate();

	/**
	 *	@fn static WETimeInfo getTime()
	 *
	 *	@brief Returns current time in a WETimeInfo struct
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WETimeInfo struct filled with the current time (hours, minutes, seconds)
	 */
	static WETimeInfo getTime();

	/**
	 *	@fn static WEDateInfo getDate()
	 *
	 *	@brief Returns current date in a WEDateInfo struct
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WEDateInfo struct filled with the current date (year, month, days)
	 */
	static WEDateInfo getDate();

	/**
	 *	@fn static int getDayTime()
	 *
	 *	@brief Retrieve day time information (AM or PM) of hour real time clock (RTC).
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@returns Integer value 0 = AM or 1 = PM
	 */
	static int getDayTime();

	/**
	 *	@fn static std::string getTimeAndDateAsString(bool p_withSeperator, char p_dateSeparator = '-', char p_space = '_', char p_timeSeparator = '-')
	 *
	 *	@brief Return the current date and time as a formatted string
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_withSeperator Flag to indicate if a seperator shall be used between parts of date
	 *		and time components or not.
	 *		false: YYYYMMDDHHmmSS
	 *		true: YYYY.MM.DD HH:mm:SS
	 *
	 *	@returns std::string containing current date and time
	 *
	 *	This function can be used to retrieve a formatted string containing current date and time.
	 *	Could e.g. be used in a logger class
	 */
	static std::string getTimeAndDateAsString(bool p_withSeperator, char p_dateSeparator = '-', char p_space = '_', char p_timeSeparator = '-');

	/**
	 *	@fn static int setSeconds(int p_valueToSet)
	 *
	 *	@brief Set the current seconds of hardware RTC
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_valueToSet Seconds to set
	 *
	 *	@returns WE_SUCCESS if setting the time worked
	 *	@returns WE_WRITE_ERROR if it was not possible to write the new system time
	 */
	static int setSeconds(int p_valueToSet);

	/**
	 *	@fn static int setMinutes(int p_valueToSet)
	 *
	 *	@brief Set the current minutes of hardware RTC
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_valueToSet Minutes to set
	 *
	 *	@returns WE_SUCCESS if setting the time worked
	 *	@returns WE_WRITE_ERROR if it was not possible to write the new system time
	 */
	static int setMinutes(int p_valueToSet);

	/**
	 *	@fn static int setHours(int p_valueToSet)
	 *
	 *	@brief Set the current hours of hardware RTC
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_valueToSet Hours to set
	 *
	 *	@returns WE_SUCCESS if setting the time worked
	 *	@returns WE_WRITE_ERROR if it was not possible to write the new system time
	 */
	static int setHours(int p_valueToSet);

	/**
	 *	@fn static int setHours12h(int p_valueToSet)
	 *
	 *	@brief Set the current hours of hardware RTC at 12h mode
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_valueToSet Hours to set (1 -12)
	 *
	 *	@returns WE_SUCCESS if setting the time worked
	 *	@returns WE_WRITE_ERROR if it was not possible to write the new system time
	 */
	static int setHours12h(int p_valueToSet);

	/**
	 *	@fn static int setDays(int p_valueToSet)
	 *
	 *	@brief Set the current days of hardware RTC
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_valueToSet Days to set
	 *
	 *	@returns WE_SUCCESS if setting the time worked
	 *	@returns WE_WRITE_ERROR if it was not possible to write the new system time
	 */
	static int setDays(int p_valueToSet);

	/**
	 *	@fn static int setMonth(int p_valueToSet)
	 *
	 *	@brief Set the current month of hardware RTC
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_valueToSet Month to set
	 *
	 *	@returns WE_SUCCESS if setting the time worked
	 *	@returns WE_WRITE_ERROR if it was not possible to write the new system time
	 */
	static int setMonth(int p_valueToSet);

	/**
	 *	@fn static int setYear(int p_valueToSet)
	 *
	 *	@brief Set the current year of hardware RTC
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_valueToSet Year to set
	 *
	 *	@returns WE_SUCCESS if setting the time worked
	 *	@returns WE_WRITE_ERROR if it was not possible to write the new system time
	 */
	static int setYear(int p_valueToSet);

	/**
	 *	@fn static int setTimeAndDate(WETimeAndDateInfo p_valueToSet)
	 *
	 *	@brief Sets both time and date to hardware RTC
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_valueToSet Struct containing complete time and date information
	 *
	 *	@returns WE_SUCCESS if setting system time and date was successful
	 *	@returns WE_WRITE_ERROR if writing the new system time failed
	 *
	 *	@see WETimeAndDateInfo
	 */
	static int setTimeAndDate(WETimeAndDateInfo p_valueToSet);

	/**
	 *	@fn static int setTime(WETimeInfo p_valueToSet)
	 *
	 *	@brief Sets time to hardware RTC
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_valueToSet Struct containing time information (hours, minutes, seconds)
	 *
	 *	@returns WE_SUCCESS if setting system time was successful
	 *	@returns WE_WRITE_ERROR if writing the new system time failed
	 *
	 *	@see WETimeInfo
	 */
	static int setTime(WETimeInfo p_valueToSet);

	/**
	 *	@fn static int setDate(WEDateInfo p_valueToSet)
	 *
	 *	@brief Sets date to hardware RTC
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_valueToSet Struct containing date information (year, month, day)
	 *
	 *	@returns WE_SUCCESS if setting system date was successful
	 *	@returns WE_WRITE_ERROR if writing the new system date failed
	 *
	 *	@see WEDateInfo
	 */
	static int setDate(WEDateInfo p_valueToSet);

	/**
	 *	@fn static int setDayTime(int p_dayTime)
	 *
	 *	@brief Sets day time (AM or PM) and recalculate hours
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urach
	 *
	 *	@param p_dayTime 0 = WE_DAYTIME_AM or 1 = WE_DAYTIME_PM
	 *
	 *	@returns WE_SUCCESS if setting system date was successful
	 *	@returns WE_WRITE_ERROR if writing the new system date failed
	 *
	 *	@see WEDateInfo
	 */
	static int setDayTime(int p_dayTime);


	/**
	 *	@fn static int calcHoursIn12hFormat(int p_hours24h)
	 *
	 *	@brief Calculates hours from 24h format to hours 12h format
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_hours24 hours format 24h (0 -23)
	 *
	 *	@returns Integer value representing the "hours" (1 -12) in 12h format.
	 */
	static int calcHoursIn12hFormat(int p_hours24h);

	/**
	 *	@fn static int calcHoursIn24hFormat(int p_hours12h, int p_dayTime);
	 *
	 *	@brief Calculates hours from 12h format to hours 24h format, with the information of p_dayTime
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_hours12h hours in 12h format (1 -12)
	 *	@param p_dayTime  0 = WE_DAYTIME_AM or 1 = WE_DAYTIME_PM
	 *
	 *	@returns Integer value representing the "hours" (0 -23) in 24h format.
	 */
	static int calcHoursIn24hFormat(int p_hours12h, int p_dayTime);

	/**
	 *	@fn static int calcHoursForNewDayTime(int p_hours24h, int p_dayTime)
	 *
	 *	@brief Calculates hours in 24h format for new dayTime.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_hours24h hours in 24h format (current hours)
	 *	@param p_dayTime  0 = WE_DAYTIME_AM or 1 = WE_DAYTIME_PM
	 *
	 *	@returns Integer value representing the new "hours" (0 -23) in 24h format.
	 */
	static int calcHoursForNewDayTime(int p_hours24h, int p_dayTime);

	static int debugSection;

	/**
	 *	@fn WETime()
	 *
	 *	@brief Deleted constructor of WETime. No instantiation is possible.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WETime() = delete;
	~WETime() = delete;
};
} // end of namespace WE

#endif /* WE_TIME_H_ */
