/**
 *
 *	@file we_stateful.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file defines the class for having a state driven architecture
 *	for the applications.
 *
*/

#ifndef WE_STATEFUL_H_
#define WE_STATEFUL_H_

namespace WE
{

/**
 *	@class WEStateful
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief This class defines the abstraction for an application for having a state driven architecture.
 *	This architecture allows an application to be:
 *	- event driven in short running critical parts, such as handling CAN messages
 *	- state driven in less critical parts like screen refresh
 *	The states of the state driven application can be updated by the update() method of this class
 *	The applications can have a single main loop which should be invoked whenever time is available
 *	(i.e. IDLE processing).	This main loop handles the events from various sources and refreshes the GUI
 *	in different states	of an application. In order to react on asynchronous events within main loop,
 *	an entity can queue	the events until they can be handled. In order to prevent the queue(s) from
 *	overflowing the runtime	of the main loop should be limited.
 *
 *	@see WEApplication
 *
*/
class WEStateful
{
public:
	/**
	 *	@fn updateState(unsigned int p_timeTicks)
	 *
	 *	@brief This method allows an application to run through it's state machine and switch it
	 *	further based on the timeTicks counter provided. The transitions performed in this method
	 *	have to be limited - otherwise it will block the main loop of the application. These transitions
	 *	should be performed within the time ticks as specified in the parameter to avoid blocking of the
	 *	main loop. If an object is not able to update itself due to the timeout then a central entity
	 *	should know how to reach all the WEStateful objects and to remember if and for which object it
	 *	should continue calling updateState() because time out was reached in the last run.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param p_timeTicks This is the point of time where this method should return even
	 *	if it is not finished.
	 *
	 *	@returns int A successful return value (WE_SUCCESS) should indicate the object is updated.
	 *	@returns int A unsuccessful return value (WE_CALL_AGAIN) indicates that the object is not up to
	 *	date and should be called again since update was suspended due to timeout.
	 *
	 *	@see WEApplication
	 *	@see WETick
	*/
	virtual int updateState(unsigned int p_timeTicks) = 0;
	virtual ~WEStateful() = default;
};
}
#endif /* WE_STATEFUL_H_ */
