/**
 *	@file toolkit_we_application.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file contains the interfaces for toolkit dependent definition of the WEApplication.
 *
 *	@see WEApplication
 */


#ifndef TOOLKIT_WE_APPLICATION_H_
#define TOOLKIT_WE_APPLICATION_H_

#include <string>

namespace WE
{

class WEApplication;
class WESize;
class WEWidget;
class WEColor;

/**
 *	@class ToolkitWEApplication
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief This class contains the interfaces that a WEApplication shall have. Any toolkit
 *	specific implementation must provide these interfaces.
 *
 *	@namespace WE
 *
 *	@see QTWEApplication
 *	@see WEApplication
 *
*/
class ToolkitWEApplication
{

public:

	virtual ~ToolkitWEApplication() = default;

	/**
	 *	@fn execute()
	 *
	 *	@brief This method of is used to start event handling. The main
	 *	It receives paint event from the underlying window system and
	 *	dispatches these to the application widgets.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@returns Returns the value that was set to exit().
	 *
	 *	@warning It is the responsibility of the user creating the instance
	 *	to delete the instance.
	 *
	 *	This method must be called for the widgets to receive the events.
	 */
	virtual int execute()=0;

	/**
	 *	@fn virtual void exit(int returnCode)=0
	 *
	 *	@brief This method tells the application to exit with a return code.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param returnCode The execute() function returns this returnCode.
	 *
	 */
	virtual void exit(int p_returnCode)=0;

	/**
	 *	@fn virtual void enablePainting(bool p_paint) = 0
	 *
	 *	@brief Enable or disable painting
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@warning Has no effect on a Linux PC
	 *
	 *	@param p_paint true if painting shall be enabled, false if painting shall be disabled
	 */
	virtual void enablePainting(bool p_paint) = 0;

	 /*	@fn virtual void setTouchInput(bool p_setTouch) = 0
	 *
	 *	@brief Enable or disable touch input events on device
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@warning Has no effect on a Linux PC
	 *
	 *	@param p_paint true if touch input events shall be enabled, false if touch input events shall be disabled
	 */
	virtual void setTouchInput(bool p_setTouch) = 0;

	/**
	 * Return the flag indicating if touch shall be enabled or not (set via setTouchInput())
	 */
	virtual bool isTouchEnabled(void) = 0;

	virtual void reloadMouse(void) = 0;

	/**
	 *	@fn virtual void setCursorVisible(bool p_visible) = 0
	 *
	 *	@brief Show or hide the mouse cursor
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_visible true if the mouse cursor shall be shown, false if the mouse cursor shall be hidden
	 *
	 *	@warning Has no effect on a windows PC
	 */
	virtual void setCursorVisible(bool p_visible) = 0;

	/**
	 *	@fn virtual WESize getDisplayResolution(void) = 0
	 *
	 *	@brief Retrieve the pixel resolution of the display on which the application runs on
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns A WESize instance that contains the width and height in pixel of the display
	 */
	virtual WESize getDisplayResolution(void) const = 0;

	virtual void setApplicationName(const std::string &) = 0;
	virtual const std::string & getApplicationName( void ) const = 0;

	virtual void setApplicationVersion(const std::string &) = 0;
	virtual const std::string & getApplicationVersion( void ) const = 0;

	virtual void setActiveWindow(WEWidget * p_widget) = 0;
	/**
	 *	@fn bool isServerAppln()
	 *
	 *	@brief This method check if the application context in which
	 *	this method is called, is a Qt Server application or Client application.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@returns TRUE if application is server
	 *
	 */
	virtual bool isServerAppln() = 0;

	/**
	 *	@fn bool isApplicationFocussed()
	 *
	 *	@brief This method check if the application currently has focused an
	 *	it is an active application
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@returns TRUE if application is focused, FALSE otherwise
	 *
	 */
	virtual bool isApplicationFocussed() = 0;

	/**
	 *	@fn virtual void setExitFunction(void (*functionPointer)(void*), bool p_terminateWithLastWindow = true) = 0
	 *
	 *	@brief Set function to be called when quit signal (from WM) is received
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param functionPointer Pointer to a function that will be called when the
	 *		ToolkitApplication receives the quit event.
	 *	@param p_terminateWithLastWindow Set to "true" if "functionPointer" should be called when this' last window is closed
	 *
	 *	The exit function will be called when the application is quit via
	 *	window manager.
	 *	This function will be automatically called when the user sets the
	 *	termination function with WEApplication::setTerminationListenerFunction
	 *
	 *	p_terminateWithLastWindow may not be implemented if there should be no support from the toolkit used.
	 *	It is supported for Qt though.
	 */
	virtual void setExitFunction(void (*p_functionPointer)(void*), bool p_terminateWithLastWindow = true) = 0;

	/**
	 *	@fn void processEvents()
	 *
	 *	@brief This method can be called for immediate handling of events in current context.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@note: This function was introduced to give Qt more chances to process input events while
	 *	       our large stage engine is running in Qt idle processing. Result is mainly the faster
	 *	       response on key, mouse and encoder events.
	 */
	virtual void processEvents() = 0;

	virtual void setCursorBlank() = 0;

	/**
	 *	@fn void setBackgroundColor(WEColor &p_color)
	 *
	 *	@brief Sets the background color and transparency for the application.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_color The color (red, green, blue and transparency) to apply to for the background
	 */
	virtual void setBackgroundColor(WEColor &p_color) = 0;

	/**
	 * Retrieve the rotation of the screen.
	 * Uses window system specific code from the respective servers (X11 or QWS).
	 * @return 0, 90, 180, 270
	 */
	virtual int getScreenRotation(void) const = 0;

	/**
	 *	@brief Introduced to rotate display at runtime.
	 *
	 *	@param p_rotation Set to 0, 90, 180 or 270. All other values may be considered invalid.
	 *
	 *	@returns WE_SUCCESS in case of success.
	 *	@returns WE_NOT_SUPPORTED_ERROR if parameter is not set to one of the supported values.
	 *
	 *	For real implementation
	 *	@see QTWEServer::setScreenRotation(p_rotation)
	 *	for your platform.
	 */
	virtual int setScreenRotation(int p_rotation) = 0;

	/**
	 *	@fn void tryEnableSquish(int p_port);
	 *
	 *	@brief 	Try to enable Squish (load Squish Plugin) for Test-Automation.
	 *
	 *	@param p_port Port for Squish connection
	 *
	 *	NOTE:
	 *	Currently this method is only available for Topcon OS  (OE Angstroem).
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 */
	virtual void tryEnableSquish(int p_port) = 0;
};
}
#endif /* TOOLKIT_WE_APPLICATION_H_ */
