/**
 *	@file qt_we_application.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora

 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file contains the definition of the Qt specific definition of the
 *	WEApplication.
 *
 *	@see qt_we_application.cpp
 *
 */

#ifndef QT_WE_APPLICATION_H_
#define QT_WE_APPLICATION_H_

#include <QDataStream>
#include <QApplication>
#include <QIcon>
#include <QTimer>

#include "../../../include/toolkit_we_application.h"

#include <iostream>


namespace WE
{

class WEWidget;


/**
 *	@class QTWEApplication
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief This class inherits from ToolkitWEApplication and QApplication
 *			for the Qt specific definition of the WEApplication
 *
 *	@namespace WE
 *
 *	@see WEApplication
 */
class QTWEApplication: public QApplication, public ToolkitWEApplication
{
	//Required for the connect call of QObject
Q_OBJECT
public:
	friend class QTWEServer; // does everything platform specific regarding window managing
	/**
	 *	@fn QTWEApplication(int& argc, char *argv[], WEApplication* application)
	 *
	 *	@brief Cconstructs an instance of the QTWEApplication
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] argc Command Line parameter argc
	 *	@param[in] argv Command Line parameter argv
	 *	@param[in] application Reference to the WEApplication
	 */
	QTWEApplication(int& p_argc, char *p_argv[], WEApplication* p_application);


	/**
	 *	@fn ~QTWEApplication
	 *
	 *	@brief Destructor of QTWEApplication
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 */
	virtual ~QTWEApplication();


	/**
	 *	@fn execute()
	 *
	 *	@brief This method of is used to start event handling. The main
	 *	It receives paint event from the underlying window system and
	 *	dispatches these to the application widgets.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	This method must be called for the widgets to receive the paint
	 *	event. It is the pure virtual method of WApplication.
	 */
	int execute() override;


	/**
	 *	@fn exit(int returnCode)
	 *
	 *	@brief This method tells the application to exit with a return code.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param returnCode The execute() function returns this returnCode.
	 *
	 */
	void exit(int p_returnCode) override;


	/**
	 *	@fn void enablePainting(bool p_paint)
	 *
	 *	@brief Enable or disable painting
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@warning Has no effect on a Linux PC
	 *
	 *	@param p_paint true if painting shall be enabled, false if painting shall be disabled
	 */
	void enablePainting(bool p_paint) override;


	/*
	 *	@fn virtual void setTouchInput(bool p_setTouch) = 0
	 *
	 *	@brief Enable or disable touch input events on device
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@warning Has no effect on a Linux PC
	 *
	 *	@param p_paint true if touch input events shall be enabled, false if touch input events shall be disabled
	 */
	void setTouchInput(bool p_setTouch) override;

	/**
	 * Return the flag indicating if touch shall be enabled or not (set via setTouchInput())
	 */
	bool isTouchEnabled(void) override;

	void reloadMouse(void) override;


	/**
	 *	@fn void setCursorVisible(bool p_visible)
	 *
	 *	@brief Show or hide the mouse cursor
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_visible true if the mouse cursor shall be shown, false if the mouse cursor shall be hidden
	 *
	 *	@warning Has no effect on a windows PC
	 */
	void setCursorVisible(bool p_visible) override;


	/**
	 *	@fn WE::WESize getDisplayResolution(void)
	 *
	 *	@brief Retrieve the pixel resolution of the display on which the application runs on
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns A WESize instance that contains the width and height in pixel of the display
	 */
	WE::WESize getDisplayResolution(void) const override;


	/**
	 *	@fn void setApplicationName(std::string p_name)
	 *
	 *	@brief This method sets the version of the application in the
	 *	QApplication.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@param[in] p_name name of the application
	 *
	 * 	This application name is useful when we get the list of applications in WM.
	 * 	We can use it to identify the WE Applications but mechanism needs to be devised.
	 */
	void setApplicationName(const std::string & p_name) override
	{
		// std::cout << "setApplicationName: " << p_name << std::endl;
		QApplication::setApplicationName(QString::fromStdString(p_name));
		//setObjectName is required when we get the list of applications in WM and
		// from the list we have to identify the WE Applications
		QApplication::setObjectName(QString::fromStdString(p_name));
	}

	const std::string & getApplicationName( void ) const override
	{
		static std::string name = QApplication::applicationName().toStdString();
		return name;
	};


	/**
	 *	@fn void setApplicationVersion(std::string p_version)
	 *
	 *	@brief This method sets the version of the application in the
	 *	QApplication.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@param[in] p_version version of the application
	 *
	 */
	void setApplicationVersion(const std::string & p_version) override
	{
		QApplication::setApplicationVersion(QString::fromStdString(p_version));
	}

	const std::string & getApplicationVersion( void ) const override
	{
		static std::string version = QApplication::applicationVersion().toStdString();
		return version;
	};


	/**
	 *	@fn void setActiveWindow(WEWidget * p_widget)
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@param[in] p_widget (window) widget to become the active window for this application
	 *
	 *	Define the active window for this application.
	 *	Set p_widget to nullptr if there should be no active window for the application.
	 *	This could be helpful in order to receive another window active event once the
	 *	user (or the window manager) bring the window to front.
	 *	Bringing the active window to front again does not raise another window active event.
	 *
	 *	Required in the iso vt application to activate main windows of the pool when
	 *	pools are updated or load version (not by the user) etc.
	 *
	 */
	void setActiveWindow(WEWidget * p_widget) override;


	/**
	 *	@fn bool isServerAppln()
	 *
	 *	@brief This method check if the application context in which
	 *	this method is called, is a Qt Server application or Client application.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@returns TRUE if application is server
	 *
	 */
	bool isServerAppln() override
	{
#if QT_VERSION >= 0x050000
		// This seems mainly to be used in order to identify a single application ...
		// ... and assumes the window manager is a good way to avoid things are done twice
		// ... so we should set this to 'true' even for our X11 window manager although it
		// ... will never be the windows (X) server.
		return this->isWinManager;
#else
		bool retValue = false;
		if (QApplication::type() == QApplication::GuiServer)
		{
			retValue = true;
		}
		return retValue;
#endif
	}


	/**
	 *	@fn bool isApplicationFocussed()
	 *
	 *	@brief This method check if the application currently has focused an
	 *	it is an active application
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@returns TRUE if application is focused, FALSE otherwise
	 *
	 */
	bool isApplicationFocussed() override
	{
		return this->applicationFocussed;
	}


	// This is a callback method which is called when any QEvent is generated
	bool event(QEvent* p_e) override;


	QWidgetList topLevelWidgets();


	/**
	 *	@fn void setExitFunction(void (*functionPointer)(void*), bool p_terminateWithLastWindow = true)
	 *
	 *	@brief Set function to be called when quit signal (from WM) is received
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param functionPointer Pointer to a function that will be called when the
	 *		ToolkitApplication receives the quit event.
	 *	@param p_terminateWithLastWindow Set to "true" if "functionPointer" should be called when this' last window is closed
	 *
	 *	The exit function will be called when the application is quit via
	 *	window manager.
	 *	The problem was, that on PC, when you closed the PClient window, the
	 *	window disappeared but PClient kept running. This should be fixed now.
	 *
	 *	p_terminateWithLastWindow may not be implemented if there should be no support from the toolkit used.
	 *	It is supported for Qt though.
	 */
	void setExitFunction(void (*p_functionPointer)(void*), bool p_terminateWithLastWindow = true) override;


	/**
	 *	@fn void processEvents()
	 *
	 *	@brief This method can be called for immediate handling of events in current context.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@note: This function was introduced to give Qt more chances to process input events while
	 *	       our large stage engine is running in Qt idle processing. Result is mainly the faster
	 *	       response on key, mouse and encoder events.
	 */
	void processEvents() override;


	void setCursorBlank() override;


	/**
	 *	@fn void setBackgroundColor(WEColor &p_color)
	 *
	 *	@brief Sets the background color and transparency for the application.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_color The color (red, green, blue and transparency) to apply to for the background
	 */
	void setBackgroundColor(WEColor &p_color) override;


	/**
	 *	@fn int setAlarmPriority(unsigned int p_winID, int p_priority)
	 *
	 *	@brief Set alarm priority for a certain window
	 *
	 *	@author mlenhart@topcon.com
	 *	@author Michael Lenhart
	 *
	 *	@param p_winID ID of the window whose alarm priority changed
	 *	@param p_priority priority of the alarm, -1 for no alarm
	 *
	 *	@return An error code: WE_SUCCESS in case of success.
	 *	@return WE_SOCKET_ERROR if there is no connection to window manager.
	 *	@return WE_SEND_ERROR if no message could be sent.
	 *	@return WE_GENERAL_ERROR for other errors.
	 *
	 *	On QWS based systems this will contact the window manager/server via QCopChannel.
	 *	On X11 based systems a message will be sent via X server connection.
	 *
	 *	@see QTWEServer::setAlarmPriority(unsigned int p_winID, int p_priority)
	 */
	int setAlarmPriority(unsigned int p_winID, int p_priority);


	/**
	 *	@fn int sendToggleMessage(void)
	 *
	 *	@brief Send message to window manager in order to switch to next window
	 *
	 *	@author mlenhart@topcon.com
	 *	@author Michael Lenhart
	 *
	 *	@return An error code: WE_SUCCESS in case of success.
	 *	@return WE_SOCKET_ERROR if there is no connection to window manager.
	 *	@return WE_SEND_ERROR if no message could be sent.
	 *	@return WE_GENERAL_ERROR for other errors.
	 *
	 *	On QWS based systems this will contact the window manager/server via QCopChannel.
	 *	On X11 based systems a message will be sent via X server connection.
	 *
	 *	@see QTWEServer::sendToggleMassage(void)
	 */
	int sendToggleMessage(void);

	int getScreenRotation(void) const override;
	int setScreenRotation(int p_rotation) override;

	/**
	 *	@fn void tryEnableSquish(int p_port);
	 *
	 *	@brief 	Try to enable Squish (load Squish Plugin) for Test-Automation.
	 *
	 *	@param p_port Port for Squish connection
	 *
	 *	NOTE:
	 *	Currently this method is only available for Topcon OS  (OE Angstroem).
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 */
	void tryEnableSquish(int p_port) override;


private slots:
	/**
	 *	@fn timerHandler
	 *
	 *	@brief This method is the slot for the timeout event of the QTimer and is
	 *	invoked when the timeout has occurred on the timer, here with a QTimer of 0
	 *	timeout the method is invoked during IDLE processing of QT's event handling.
	 *	This method internally invokes the mainLoop method of the application.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 */
	void timerHandler();


private:
	QTimer timer; //!< Timer to invoke the mainLoop method during IDLE processing
	WEApplication* application; //Reference to the WEApplication
	bool applicationFocussed;	//!< Flag which tells if application is currently visible to the user or not. True, if active and visible to the user, false otherwise
	bool isWinManager;		//!< Flag indicating if this is the window manager (Qt >= 5.x) or server (QT <= 4.x)
	bool touchEnabled;

	void (*exitFunctionFP)(void*);	//!< Function pointer to be called when receiving "exit" event from Qt.
};
} //end of namespace WE

#endif /* QT_WE_APPLICATION_H_ */
