/**
 *	@file boost_assertion_handler.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Implement boost's assertion_failed function to be able to handle errors without quitting
 *
 */


#ifndef BOOST_ASSERTION_HANDLER_H_
#define BOOST_ASSERTION_HANDLER_H_

#include <iostream>

namespace boost
{

/**
 *	@fn void assertion_failed(char const * expr, char const * function, char const * file, long line)
 *
 *	@brief Implemented to handle boost assertions without quitting
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@param p_expr The expression where the assertion occurred
 *	@param p_function The function where the assertion occurred
 *	@param p_file The source file where the assertion occurred
 *	@param p_line The line inside file where the assertion ocurred
 *
 *	@see boost assertion documentation
 *
 *	@warning If the thrown exception is caught, the application will not quit
 *		like it would do after an assertion.
 *
 *	To make this work, the application has to be compiled using the following option:
 *	-DBOOST_ENABLE_ASSERT_HANDLER
 *	And this makefile has to be included after including the boost includes.
 */
void assertion_failed(char const * p_expr, char const * p_function, char const * p_file, long p_line)
{
	// AXIVION Routine Generic-NoFunctionDefinitionInHeader : This function needs to be defined here to make it work for boost
	try
	{
		std::string text = "Boost Exception:\n";
		text += "Expression: ";
		text += p_expr;
		text += "\n";
		text += "Function: ";
		text += p_function;
		text += "\n";
		text += "In File: ";
		text += p_file;
		text += "\n";
		text += "At Line: ";
		text += p_line;
		text += "\n";

		throw std::runtime_error(text);
	}
	catch (...)
	{
		std::cout << "Error during boost assertion handling" << std::endl;
	}
}

/**
 *	@fn void assertion_failed_msg(char const * expr, char const * msg, char const * function, char const * file, long line)
 *
 *	@brief Implemented to handle boost assertions without quitting
 *
 *	@author preths@topcon.com
 *	@author Patrick Reths
 *
 *	@param p_expr The expression where the assertion occurred
 *	@param p_msg	A description of the assertion occurred
 *	@param p_function The function where the assertion occurred
 *	@param p_file The source file where the assertion occurred
 *	@param p_line The line inside file where the assertion ocurred
 *
 *	@see void assertion_failed(char const * expr, char const * function, char const * file, long line)
 */
void assertion_failed_msg(char const * p_expr, char const * p_msg, char const * p_function, char const * p_file, long p_line)
{
	// AXIVION Routine Generic-NoFunctionDefinitionInHeader : This function needs to be defined here to make it work for boost
	try
	{
		std::string text = "Boost Exception:\n";
		text += "Message: ";
		text += p_msg;
		text += "\n";
		text += "Expression: ";
		text += p_expr;
		text += "\n";
		text += "Function: ";
		text += p_function;
		text += "\n";
		text += "In File: ";
		text += p_file;
		text += "\n";
		text += "At Line: ";
		text += p_line;
		text += "\n";

		throw std::runtime_error(text);
	}
	catch (...)
	{
		std::cout << "Error during boost assertion handling" << std::endl;
	}
}
}

#endif /* BOOST_ASSERTION_HANDLER_H_ */
