/**
 *	@file we_error.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Error types and error functions used by WEAPI
 *
 *	@see we_error.cpp
 *
 *	This file contains error codes and error handling function definitions.
 */

#ifndef WE_ERROR_H_
#define WE_ERROR_H_

#include <string>

namespace WE
{

/**
 *	@enum we_error_codes
 *
 *	@brief Definitions of WEAPI error codes
 */
enum we_error_codes
{
	WE_SUCCESS = 0, //!< To be returned if everything worked as expected 0
	WE_GENERAL_ERROR = 1, //!< Should not be returned, better use more speaking error codes
	WE_DEVICE_ERROR = 2, //!< A device function failed
	WE_INIT_ERROR = 3, //!< Initialization failed
	WE_SOCKET_ERROR = 4, //!< Socket error
	WE_READ_ERROR = 5, //!< Reading failed
	WE_WRITE_ERROR = 6, //!< Writing failed
	WE_SEND_ERROR = 7, //!< Sending something was not successful
	WE_RECEIVE_ERROR = 8, //!< Receiving something was not successful
	WE_SELECT_ERROR = 9, //!< The select function failed
	WE_INTERFACE_ERROR = 10, //!< Interface error
	WE_NOT_SUPPORTED_ERROR = 11, //!< Function not supported on this platform
	WE_GET_VARIABLE_VALUE_ERROR = 12, //!< Get variable value request failed
	WE_SET_VARIABLE_VALUE_ERROR = 13, //!< Set variable value request failed
	WE_CONNECT_ERROR = 14, //!< Connecting not successful
	WE_INVALID_RESPONSE_ERROR = 15, //!< Received unexpected response
	WE_FILE_ERROR = 16, //!< File operation not successful (e.g. open)
	WE_TIMEOUT_ERROR = 17, //!< Operation times out
	WE_ACCEPT_ERROR = 18, //!< accept connection fails
	WE_CREATE_THREAD_ERROR = 19, //!< Thread creation fails
	WE_INVALID_PARAMETER_ERROR = 20, //!< Parameter passed to a function is invalid
	WE_EVENT_SUBSCRIPTION_ERROR = 21, //!< Event subscription request fails.
	WE_EVENT_NOTIFY_ERROR = 22, //!< Event notification fails.
	WE_NOT_FOUND_ERROR = 23, //!< E.g. when a variable is not found in variable map
	WE_OUT_OF_BOUNDS_ERROR = 24, //!< If a value was intended to be set that is too high or too low
	WE_ALREADY_EXISTS_ERROR = 25, //!< If something shall be created or added that already exists
	WE_INVALID_MESSAGE_RECEIVED_ERROR = 26, //!< If the received message type is not what is expected.
	WE_REQUEST_REJECTED_ERROR = 27, //!< If a request is rejected by the server.
	WE_NEGATIVE_STATUS_RECEIVED_ERROR = 28, //!< If message status FALSE is received in the message.
	WE_TYPE_ERROR = 29, //!< Trying to set a value to a variable with wrong type (or similar errors)
	WE_PERMISSION_ERROR = 30, //!< When trying to do something that needs (not present) permission
	WE_ADD_WIDGET_ERROR = 31, //!< Added a child widget failed.
	WE_SET_PROPERTY_VALUE_ERROR = 32, //!< Setting the property Value
	WE_EVENT_NOT_SUPPORTED_ERROR = 33, //!<< When trying to set an Event Listener for an event that is not supported by the widget.
	WE_ADD_PROPERTY_ERROR = 34, //!< Adding the property
	WE_STATE_NOT_SUPPORTED_ERROR = 35, //!< When trying to set a state to a widget that is not recognized by the widget.
//	WE_CANOPEN_OBJ_SET_TO_DEF=36,//!< When initializing the stack and NO PDO mapping object found.
//	WE_CANOPEN_VAR_LESS_THEN_MAPPABLE_OBJECTS=37, //!< When CANOpen variables are less than the maapable application objects.
//	WE_CANOPEN_PDO_DEF_NOT_FULL=38, //!< When CANOpen PDO Mapping and Communication params are not equal
	WE_EVENT_CONFIGURATION_FAILED=39,
//	WE_GD_VARIABLE_DOES_NOT_EXIST=40, //!< When GD Variable is not found in the GD Stack
	WE_NO_DIRECTORY=41, //!< When we try to copy and move we test for DIR
	WE_DIRECTORY_EMPTY=42, //!< If we have a valid directory but no content
	WE_OPERATION_IN_PROGRESS=43, //!< If an operation is in progress and stopped by user.
	WE_MEMORY_NOT_ALLOCATED=44,	//!< If new does not allocate memory to the instance.
	// RESERVED: WE_EXECUTE_VAR_ACTION_SET_VALUE=45, //!< Execute the Set Value Variable Action
	WE_CALL_NOT_COMPLETED = 46,
	// NOTE:	If new types are created please append them so that existing numbers are not changed.
	// 		This avoids need for recompilation of existing programs not affected by the new definitions.

	// NOTE:	Please make sure to add a human readable description for new errors in weGetErrorString().
	WE_MAX_ERROR = 200
};

/**
 *	@fn std::string weGetErrorString(int errorNumber)
 *
 *	@brief Return error description by error number
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@param[in] errorNumber Error number as defined in enum we_error_codes
 *
 *	@returns The string that matches the error number
 *
 *	This function takes an error number a sinput and returns
 *	the string that matches the error number.
 */
std::string weGetErrorString(int p_errorNumber);
} //end of namespace WE

#endif /* WE_ERROR_H_ */
