/*
 * dei_ipc_data.h
 *
 * This header file contains the data definitions needed to communicate with
 * a WEapi DEI application via IPC (such as hardware daemon or PClient).
 *
 * Copyright 2019 Topcon Electronics GmbH & Co. KG
 *
 * This API and it's functions are provided as is in source code without any guarantee.
 * You are allowed to use this code to transfer data from any custom application to a WEapi
 * DEI application such as PClient or hardware daemon. The sources are provided as source code
 * to make custom application connections easier and to reduce dependencies to other libraries
 * (you don't need to link to anything, just use the headers and source files in your application).
  * Although most of the implementation of this API uses C code (because of Linux system calls), you need to use a
 * C++ compiler to use this API.
 * Please do not modify the data parts of the API since this might break compatibility with existing applications.
 * You are free to modify the function code to your liking (e.g. memory management, return values, etc.).
 * You are however NOT free to spread this code anywhere. Only for direct Topcon Electronics customers.
 * Future releases of WEapi and PClient may use an updated version of this IPC API. Make sure to check
 * for changes and update your application accordingly.
 */

#ifndef DEI_DEI_IPC_API_SRC_DEI_IPC_DATA_H_
#define DEI_DEI_IPC_API_SRC_DEI_IPC_DATA_H_

#include <new>

/**
 * The IPCMessageType enum lists all possible message types (see messageType in IPCMessage).
 * Use the different message type to define the request for sending or to identify the response when receiving.
 */
typedef enum
{
	VARIABLE_GET_REQUEST		= 0,	//!< Request the value of a variable.
	VARIABLE_GET_RESPONSE		= 1,	//!< The response to the above request with the latest value of the requested variable.
	VARIABLE_SET_REQUEST		= 2,	//!< Request to set a new value to a variable.
	VARIABLE_SET_RESPONSE		= 3,	//!< The response to the above request (check messageStatus for success or fail).
	EVENT_SUBSCRIPTION_REQUEST	= 4,	//!< Request to get updates for each value change of a variable.
	EVENT_SUBSCRIPTION_RESPONSE	= 5,	//!< The response to the above request (check messageStatus for success or fail). Already contains the value.
	VARIABLE_UPDATE_EVENT		= 6,	//!< After subscribing to a variable this message type will be used value change messages.
	IPC_MESSAGE_TYPE_INVALID	= 0xff	//!< Marks an invalid message. Should be discarded.
}IPCMessageType;

/**
 * IPCMessageHeader contains the data that defines the header of each IPCMessage.
 * See header member in IPCMessage struct.
 */
struct IPCMessageHeader
{
	IPCMessageType messageType{IPC_MESSAGE_TYPE_INVALID};	//!< Type of the message, see IPCMessageType above
	bool messageStatus{false};		//!< Is usually set in a response type message: true for success, false for failure.  Set to true when sending a VARIABLE_UPDATE_EVENT with a valid value!
	unsigned int dataLength{0};	//!< Length of the data array in the message (see getBufferSize in dei_ipc_functions.h)

	IPCMessageHeader() = default;
};

const int messageHeaderLength = sizeof(IPCMessageHeader); //!<< IPCMessageHeader length (should be 12 byte including padding for the bool type)

/**
 * The IPCMessage struct is used to send and receive variable data via Linux domain
 * sockets to a WEapi DEI application (e.g. hardware daemon, PClient, etc.).
 * Make sure to always fill all properties properly (including the header).
 * Note that the data part is a serialized version of the IPCData struct.
 * Use "unsigned int getBufferSize(IPCData * data)" to determine the value for dataLength and
 * "unsigned int convertToBuffer(unsigned char * buffer, IPCData * data)" to write the serialized bytes to the data pointer.
 * When receiving a message use "unsigned int dataFromBuffer(unsigned char * buffer, unsigned int bufferLength, IPCData * data)"
 * to convert the serialized data bytes to a IPCData struct.
 */
struct IPCMessage
{
	struct IPCMessageHeader header;	//!< Each IPCMessage has a header with fixed length preceding the data part
	unsigned char* data;		//!< Data of size header.dataLength. Serialized IPCData. See convertToBuffer and dataFromBuffer in dei_ipc_functions.h)

	IPCMessage()
	:header(),
	 data(nullptr)
	{
		//std::cout << "CONSTRUCT IPC MESSAGE" << std::endl;
	}
	~IPCMessage()
	{
		if (nullptr != data)
		{
			//std::cout << "DELETE IPC MESSAGE" << std::endl;
			delete [] data;
		}
	}

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	IPCMessage(const IPCMessage &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	IPCMessage & operator=(const IPCMessage &) = delete;
};

/**
 * The DataBufferPositions enum can be used to determine where in the serialized data array of an IPCMessage,
 * the information of the IPCData struct are stored. This is only interesting if the data shall be extracted or set manually.
 * Always prefer to use dataFromBuffer and convertToBuffer in dei_ipc_functions.h
 */
enum DataBufferPositions
{
	POSITION_INDEX_LSB			= 0,	//!< position of the least significant byte of the index
	POSITION_INDEX_MSB			= 1,	//!< position of the most significant byte of the index
	POSITION_SUBINDEX_LSB			= 2,	//!< position of the least significant byte of the subindex
	POSITION_SUBINDEX_MSB			= 3,	//!< position of the most significant byte of the subindex
	POSITION_TYPE				= 4,	//!< position of the data type (dataType is transfered as only one byte!)
	POSITION_INT_OR_STRING_LENGTH_LLSB 	= 5,	//!< position of the least significant byte of the integer value or string length (depends on data type)
	POSITION_INT_OR_STRING_LENGTH_LSB 	= 6,	//!< position of the 2nd least significant byte of the integer value or string length (depends on data type)
	POSITION_INT_OR_STRING_LENGTH_MSB 	= 7,	//!< position of the 2nd most significant byte of the integer value or string length (depends on data type)
	POSITION_INT_OR_STRING_LENGTH_MMSB 	= 8,	//!< position of the most significant byte of the integer value or string length (depends on data type)
	POSITION_STRING_VALUE_START 		= 9	//!< position of the first byte of the string value
};

/**
 * The data type a variable can have. It is either integer or string.
 */
typedef enum
{
	DATA_TYPE_INT,
	DATA_TYPE_STRING,
	DATA_TYPE_INVALID
}data_type;

/**
 * A string always consists of the length of the string and the corresponding character array.
 */
typedef struct
{
	unsigned int length;
	unsigned char * value;
}stringData;

/**
 * A value in the IPCData struct can either be an integer or a string.
 * Use a union to save 4 bytes of data.
 */
typedef union
{
	int intValue;
	stringData stringValue;
}valueData;

/**
 * The IPCData struct contains various informations that are sent in a serialized way in
 * the IPMessage's data array.
 * For pre-defined variables of hardware daemon or PClient check the properties of the variables in Projektor
 * or in Projektor help.
 * All members of this struct need to be filled properly to make it work.
 * Use dataFromBuffer and convertToBuffer in dei_ipc_functions.h to assign IPCData to a IPCMessage.
 */
struct IPCData
{
	unsigned short index;		//!< index of the variable
	unsigned short subIndex;	//!< subIndex of the variable
	data_type type;			//!< data type of the variable => important so anyone knows if there is a string or integer value in the valueData union!
	valueData data;			//!< The integer or string value => depends on data type of the variable

	IPCData()
	:index(0),
	 subIndex(0),
	 type(DATA_TYPE_INVALID)
	{
		data.intValue = 0;
		data.stringValue.length = 0;
		data.stringValue.value = nullptr;
		//std::cout << "CONSTRUCT IPC DATA" << std::endl;
	}
	~IPCData()
	{
		//std::cout << "DELETE IPC DATA" << std::endl;
		if (nullptr != data.stringValue.value)
		{
			delete [] data.stringValue.value;
		}
	}

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	IPCData(const IPCData &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	IPCData & operator=(const IPCData &) = delete;
};

#endif /* DEI_DEI_IPC_API_SRC_DEI_IPC_DATA_H_ */
