/**
 *	@file we_dei.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Interface for using all kinds of variables.
 *
 *	@see WEDEIIPCInterface
 *	@see WEVariable
 *	@see WEInternalVariable
 *	@see WECANOpenVariable
 *	@see WEGDVariable
 *	@see WECFVariable
 *	@see WEJ1939Variable
 *	@see Manual: "Developing applications using WEAPI"
 *
 *	DEI, or Data Element Interface, is the interface to all variables
 *	used by WEAPI. Applications do not access variables directly but
 *	use the DEI interface to change or retrieve the value of all kinds
 *	of Variables.
 */

#ifndef WE_DEI_H_
#define WE_DEI_H_

#include <memory> //unique_ptr
#include <string>
#include <map>
#include <unordered_map>
#include <set>

#include "../../os_abstraction/ipc/include/we_ipc_server.h"
#include "../variables/include/we_variable_definitions.h"
#include "../variables/include/we_variable.h"
#include "we_dei_ipc_client_handler.h"
#include "we_dei_ipc_server_handler.h"

namespace WE
{
//some forward declarations
class WEIPCEndPoint;
class WEEndPointManager;
class DEIIPCVariableObserver;
class WEVariableOwner;
class WEEventQueue;
class WEVariableAction;
class WETick;

typedef std::unordered_map<unsigned int, WEVariable> DEIVariableMap;	//!< hashmap <ulong, WEVariable*>
typedef std::unordered_map<std::string, unsigned int> DEIVariableNameMap;	//!< hashmap <string, ulong>
typedef std::unordered_map<unsigned int, WEVariableOwner*> DEIVariableOwnerMap;	//!< //!< hashmap <ulong, WEVariableOwner*>

typedef std::map<std::string, std::set<unsigned int>> OutstandingSubscriptionRequestsMap; //!< key contains IPC socket name of variable owner, the set contains all keys that were subscribed for but no response was received, yet.


const std::string HARDWARE_VARIABLE_OWNER = "/tmp/hardwareDaemon"; //!< This string defines the IPC address of WEHardwareDaemon, that owns all hardware variables

/**
 *	@class WEDEI
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Class that defines the interface of the Wachendorff DEI interface
 *
 *	@namespace WE
 *
 *	@see WEDEIIPCInterface
 *	@see All variable implementations
 *
 *	Another approach would avoid synchronization that could cause deadlocks:
 *	Use a message queue so that process internal requests are similar to those
 *	from other processes.
 *
 *	The DEI interface provides functions for adding, accessing and
 *	removing variables. Applications do not access variables directly but
 *	use the DEI interface to change or retrieve the value of all kinds
 *	of Variables. Applications can also subscribe to variable events to
 *	get informed when a variable value changes.
 *	WEDEI is implemented as a singleton, a user of this class always has to
 *	call the instance() method to access the public methods of the class
 *	(e.g. WEDEI::instance()->addVariable(pointerToVariable);)
 *	This reflects the fact that each process should have only one DEI as data
 *	repository but processes may exchange data by accessing each others DEI via IPC.
 *	The first call of WEDEI (probably in an initialize method) has to contain
 *	a name with which the "owner" of a variable can be identified.
 */
class WEDEI
{
private:
	/**
	 *	@fn WEDEI(WE::WEEndPointManager * p_endPointManager, bool startIPCThread = true)
	 *
	 *	@brief Private constructor (singleton). Not allowed to make an instance
	 *		of this class from outside
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *	modified by:
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	explicit WEDEI(WE::WEEndPointManager * p_endPointManager, bool p_startIPCThread = true);

	/**
	 *	@fn WEVariable * getVariable(unsigned short p_varIndex, unsigned short p_varSubIndex)
	 *
	 *	@brief Returns a pointer to a variable identified by index and subindex
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_varIndex Index of the variable that shall be returned
	 *	@param[in] p_varSubIndex Subindex of the variable that shall be returned
	 *
	 *	@returns Reference to a variable identified by index and subindex.
	 *	
	 *	@throws std::runtime_error if variable was nor found
	 *
	 *	This function searches the variable map for a combined key that
	 *	matches the index and subindex provided as parameter. If a matching variable
	 *	is found, a pointer to this variable will be returned. If no matching variable
	 *	is found, nullptr will be returned.
	 *	The returned variable pointer can then be casted to the concrete subclass
	 *	of WEVariable (use var->getType() to determine the correct type before casting).
	 */
	WEVariable & getVariable(unsigned short p_varIndex, unsigned short p_varSubIndex);


	/**
	 *	@fn WEVariable * getVariable(unsigned int p_combinedKey)
	 *
	 *	@brief Returns a pointer to a variable identified by index and subindex provided as p_combinedKey.
	 *
	 *	@param[in] p_combinedKey Combination of index and subindex of the variable that shall be returned
	 *
	 *	@returns Reference to a variable identified by index and subindex.
	 *
	 *	@throws std::runtime_error if variable was nor found
	 *
	 *	This function searches the variable map for a combined key that
	 *	matches the index and subindex provided as parameter. If a matching variable
	 *	is found, a pointer to this variable will be returned. If no matching variable
	 *	is found, nullptr will be returned.
	 *	The returned variable pointer can then be casted to the concrete subclass
	 *	of WEVariable (use var->getType() to determine the correct type before casting).
	 *
	 *	@see getVariable(unsigned short, unsigned short)
	 *	@see calculateCombinedKey(unsigned short, unsigned short)
	 */
	WEVariable & getVariable(unsigned int p_combinedKey);

	/**
	 * See getVariable() above
	 * 
	 * @returns const reference to the variable identified by p_varIndex and p_varSubIndex
	 * 
	 * @throws std::runtime_error if variable was nor found
	 */
	const WEVariable & getConstVariable(unsigned short p_varIndex, unsigned short p_varSubIndex) const;

	/**
	 * See getVariable() above
	 * 
	 * @returns const reference to the variable identified by p_varIndex and p_varSubIndex
	 * 
	 * @throws std::runtime_error if variable was nor found
	 */
	const WEVariable & getConstVariable(unsigned int p_combinedKey) const;


	/**
	 *	@fn int initializeRemanentVariable(unsigned short p_index, unsigned short p_subindex)
	 *
	 *	@brief When a remanent variable is added to the DEI, this function tries
	 *		to read the stored value from file or creates a new file.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_var Reference to the variable which's remanent value shall be initialized
	 *
	 *	@returns WE_SUCCESS if everything worked
	 *	@returns WE_NOT_FOUND_ERROR if variable was not found in variable map
	 *	@returns WE_PERMISSION_ERROR if this DEI is not the owner of the variable
	 *	@returns WE_FILE_ERROR if creating a new storage file for a remanent variable failed
	 *	@returns WE_GET_VARIABLE_VALUE_ERROR if the current value of the variable can not be retrieved
	 *	@returns any error code returned by the variable's setValue() function.
	 *
	 *	@see WEVariable
	 *
	 *	When a remanent variable is added to the DEI, the DEI tries to open the
	 *	file with stored information of the variable. If the file is present,
	 *	it reads the value stored in the file and sets it to the variable.
	 *	If no storage file is found, a new file gets created and the current
	 *	value gets written to the file.
	 */
	int initializeRemanentVariable(WEVariable & p_var);


	/**
	 *	@brief Write current value of p_variable to file.
	 *
	 *	@param p_variable - variable whose value should be written to file
	 *	@return WE::WE_SUCCESS in case of success
	 *	@return WE::WE_INVALID_PARAMETER_ERROR in case nullptr == p_variable OR !(p_variable->isRemanent())
	 *	@return WE::WE_SET_VARIABLE_VALUE_ERROR if p_variable is not owned by this instance of DEI
	 *	@return WE::WE_GET_VARIABLE_VALUE_ERROR if p_variable->getValue() fails
	 *	@return any value possibly returned by WE::WEFilesAndFolders::writeStringToFile()
	 *
	 *	This method checks if this DEI owns p_variable and then writes the value to file
	 *	if it differs from the existing value.
	 *	If required, file is created first.
	 *
	 *	@warning When the variable writes the value to the file it locks it's mutex!
	 *
	 *	This method was introduced when BSP 2.x started using UBIFS and we figured out that
	 *	we lost data when writing the same value to file again and did not sync after writing.
	 *
	 *	@see WE::WEFilesAndFolders::writeStringToFile()
	 *	@see WE::WEVariable::writeRemanentValue()
	 */
	static int writeRemanentVariableToFile(WEVariable & p_variable);


	/**
	 *	@fn bool isVariableObserved(unsigned int p_key)
	 *
	 *	@brief Checks if a variable has an event listener assigned.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_key combined key of a variable
	 *
	 *	@returns true if the combined key of the variable is in the set of observed variables
	 *	@returns false if combined key was not found in the set of observed variables
	 *
	 *	@see observedVariables
	 *	@see addVariableToEventList()
	 *	@see removeVariableFromEventList()
	 *
	 *	Checks if someone is interested in events linked to this variable.
	 *	A variable gets added to the list of observed variables with addVariableToEventList() and removed
	 *	with removeVariablefromEventList.
	 */
	bool isVariableObserved(unsigned int p_key) const;


	/**
	 *	@fn WE::std::int32_t forceNotifyVariableChange(WE::WEVariable* p_variable)
	 *
	 *	@brief This methos notify all the IPC observers and Variable Observers of the variable
	 *	@brief about the value change.
	 *
	 *	@author herpreet.sura@aricent.com
	 *	@author Herpreet Sura
	 *
	 *	@param[in] p_variable variable, whose value change is to be notified
	 *
	 *	@returns WE_SUCCESS for successful execution of the notification
	 *
	 *	@see DEIIPCClientHandler::sendUpdateToIPCObservers for error codes returned
	 *
	 *	@warning The value of the variable must be successfully set before calling this method.
	 *
	 */
	std::int32_t forceNotifyVariableChange(WE::WEVariable & p_variable);

	void createRemanentVariablesPath(void) const;

	/**
	 * Check if p_owner is one of the registered local variable
	 * owners. See registerLocalVariableOwner().
	 */
	bool isLocalVariableOwner(WEVariableOwner * p_owner);


	static inline std::unique_ptr<WEDEI> theInstance{}; //!< The one and only instance of a DEI that exists per process
	static inline std::string remanentVariablesPath{};
	static inline const std::string emptyString{};	//!< Empty string is returned when getStringValue() fails

	std::string instanceName; //!< The name of this DEI, used to determine if this DEI is owner of a certain variable

	//Variable stuff
	DEIVariableMap variableMap; //!< HashMap where all the (pointer to) variables are stored
	DEIVariableNameMap variableNameMap; //!< To be able to access variables via their names
	DEIVariableOwnerMap variableOwnerMap; //!< Mapping between variable key and a pointer to the variable's owner

	//IPC stuff
	std::unique_ptr<DEIIPCServerHandler> ipcServerHandler{}; //!< Server handler for IPC messages
	std::unique_ptr<DEIIPCClientHandler> ipcClientHandler{}; //!< Client handler for IPC messages
	WEIPCServer ipcServer{}; //!< The IPC Server
	WEEndPointManager * ipcEndPointManager; //!< Manager for all used endpoints, owned by PClient. Never destroy inside DEI!
	bool isIPCServerInitialized; //!< Indicates if the IPCServer was already initialized to avoid double initialization.

	//Event stuff:
	std::unique_ptr<WEEventQueue> sendQueue; //!< Event queue for outgoing events (like INT_VARIABLE_VALUE_CHANGED_EVENT)
	std::set<unsigned int> observedVariables; //!< Set of variable combined keys that are being observed from event listeners

	WEVariableOwner * hardwareVariableOwner;

	std::map<std::string, std::shared_ptr<WEIPCEndPoint>> peerAddressToIPCEndPointMap;	//!< a Map of peerAddress to WEIPCEndPoint*, used to find the endpoint for the variable owners

	bool allVariablesAreObserved;

	OutstandingSubscriptionRequestsMap outstandingSubscriptionRequests;

	std::set<WEVariableOwner*> localVariableOwners; //!< set of owners for local variables (e.g. project variable manager, ut plugin, ...)

public:
	/**
	 *	@fn WEDEI(WEDEI const&)
	 *
	 *	@brief Deleted copy constructor. No copies are allowed (singleton). Implemented to keep static code check quiet
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WEDEI(const WEDEI & p_dei) = delete; //no copy allowed


	/**
	 *	@fn WEDEI& operator=(WEDEI const&)
	 *
	 *	@brief Deleted assignment operator. No copying allowed (singleton)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WEDEI& operator=(WEDEI const&) = delete; //no assignment allowed

	/**
	 *	@fn void release()
	 *
	 *	@brief deallocates the memory allocated by an DEI instance.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 */
	void release();


	/**
	 *	@fn ~WEDEI()
	 *
	 *	@brief Destructor of the DEI. Does nothing!
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	Because of the singleton implementation call the release() function to free
	 *	memory. This will also destroy the singleton instance which will call this
	 *	destructor (which does nothing)
	 */
	virtual ~WEDEI();


	/**
	 *	@fn static WEDEI * create(WE::WEEndPointManager * p_endPointManager, bool startIPCServer = true)
	 *
	 *	@brief The create() method creates and returns the instance of the DEI of this process.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *	modified by:
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_endPointManager The end point manager which will be used to manage DEI events.
	 *
	 *	@returns pointer to the created instance of the DEI
	 *
	 *	@warning When using other methods they have to be called by calling the instance method! (see below)
	 *		e.g. WEDEI::instance()->otherMethod();
	 *	@warning The first time WEDEI::instance() is called, the EndPointManager shall be provided.
	 *		E.g. WEDEI::instance(myEndPointManager);
	 *		It is mandatory to call startIPCServer() method before using the DEI.
	 *
	 *	Since WEDEI is a singleton, there is only one instance per process allowed.
	 *	Therefore this function shall only be called once within the process.
	 */
	static WEDEI * create(WE::WEEndPointManager * p_endPointManager, bool p_startIPCServer = true);


	/**
	 *	@fn int startIPCServer(std::string p_instanceName)
	 *
	 *	@brief Starts the IPC server and creates the corresponding socket.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_instanceName The name of the DEI. Used for variable ownership observation.
	 *			This is used as a filename to create a socket on Linux - so the user needs permissions
	 *			to create a file at the path used here.
	 *
	 *	@returns WE_SUCCESS on success.
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_instanceName is invalid.
	 *	@returns any other error delivered by WEIPCServer::initialize()
	 *
	 *	Since the p_instanceName is used as a filename it is suggested to use a location where
	 *	every user has write permissions, e.g. /tmp/my_application_instance_name
	 */
	int startIPCServer(const std::string & p_instanceName);


	/**
	 *	@fn static WEDEI * instance()
	 *
	 *	@brief The instance() method returns the instance of the DEI of this process.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns pointer to the existing instance of the DEI. If no DEI instance exists, it gives
	 *		 some error message and abort the process.
	 *
	 *	@warning When using other methods they have to be called by calling the instance method!
	 *		e.g. WEDEI::instance()->otherMethod();
	 *
	 *	Since WEDEI is a singleton, there is only one instance per process allowed.
	 *	Therefore this static method has to be used to call any other methods of
	 *	the WEDEI.
	 */
	static WEDEI * instance();


	/**
	 *	@fn int addVariable(WEVariable * p_variableToAddToMap)
	 *
	 *	@brief Adds a WEVariable to DEI's variable repository
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_variableToAddToMap The variable that shall be added
	 *
	 *	@returns RETURN DESCRIPTION
	 *
	 *	@see WEVariable 
	 *
	 *	@warning Do not add an observer to the variable before adding the variable to DEI. It will get lost!
	 *
	 *	With this function a variable can be added to the DEI's variable repository.
	 *	When passing p_variableToAddToMap in this function it will be copied and the copy
	 *	will be moved to DEI's storage. After the variable was added all DEI functions
	 *	that manipulate a variable via index/subindex can be used.
	 */
	int addVariable(WEVariable p_variableToAddToMap);


	/**
	 *	@fn int removeVariable(unsigned short p_index, unsigned short p_subindex);
	 *
	 *	@brief Removes a variable from DEI's repository
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the variable that shall be removed
	 *	@param p_subindex Subindex of the variable that shall be removed
	 *
	 *	@returns WE_NOT_FOUND_ERROR if the variable identified by p_index/p_subindex was not found
	 *	@returns WE_SUCCESS if the variable was removed successfully
	 */
	int removeVariable(unsigned short p_index, unsigned short p_subindex);


	/**
	 *	@fn int getIndicesByName(std::string p_variableName,
	 *		unsigned short * p_indexToReturn,
	 *		unsigned short * p_subindexToReturn)
	 *
	 *	@brief With this function it is possible to retrieve the index and subindex
	 *		of a variable using only it's name.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_variableName Name of the variable
	 *	@param[out] p_indexToReturn To this address the index of the variable gets stored.
	 *		Result will be 0 if variable was not found.
	 *	@param[out] p_subindexToReturn To this address the subindex of the variable gets stored.
	 *		Result will be 0 if variable was not found.
	 *
	 *	@returns WE_SUCCESS if index and subindex were successfully stored to the pointer's address
	 *	@returns WE_NOT_FOUND_ERROR if the variable with the provided name was not found.
	 *
	 *	This function makes it possible to retrieve the index and subindex of a variable if only
	 *	the name of the variable is known. All subsequent function calls of DEI will only be available
	 *	by addressing a variable with index and subindex.
	 */
	int getIndicesByName(const std::string & p_variableName,
			unsigned short * p_indexToReturn,
			unsigned short * p_subindexToReturn) const;

	/**
	 *	@fn int getVariableStorageType(unsigned short p_index,
	 *		unsigned short p_subindex,
	 *		WEVariableStorageType * p_storageType)
	 *
	 *	@brief Retrieve the storage type (integer or string) from a variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_index Index of the variable
	 *	@param[in] p_subindex Subindex of the variable
	 *	@param[out] p_storageType Pointer containing address to where the storage type shall be written
	 *
	 *	@returns WE_SUCCESS if everything worked
	 *	@returns WE_NOT_FOUND_ERROR if the variable was not found in DEI's variable repository
	 *	@returns WE_TYPE_ERROR If the storage type of the variable was not defined correctly
	 *
	 *	@see WEVariable and all it's subclasses
	 *	@see WEVariableStorageType
	 *
	 *	This functions allows the retrieval of a variable's storage type.
	 */
	int getVariableStorageType(unsigned short p_index,
			unsigned short p_subindex,
			WEVariableStorageType * p_storageType) const;
	int getVariableStorageType(unsigned int p_combinedKey, WEVariableStorageType * p_storageType) const;

	/**
	 *	@fn WEVariableAccessType getVariableAccessType(unsigned short p_index, unsigned short p_subindex)
	 *
	 *	@brief Retrieves the access type of a variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_index Index of the variable
	 *	@param[in] p_subindex Subindex of the variable
	 *
	 *	@returns The access type of the variable
	 *
	 *	@see WEVariableAccessType
	 *	@see WEVariable
	 */
	WEVariableAccessType getVariableAccessType(unsigned short p_index, unsigned short p_subindex) const;


	/**
	 *	@fn int setVariableValue(unsigned short p_varIndex, unsigned short p_varSubIndex, int p_valueToSet, bool p_isValid = true, bool p_calledBySimulation = false)
	 *
	 *	@brief Set a new value to a variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_varIndex Index of the variable to which a new value shall be set
	 *	@param[in] p_varSubIndex Subindex of the variable to which a new value shall be set
	 *	@param[in] p_valueToSet New value of the variable
	 *
	 *	@returns WE_SUCCESS if setting the value worked as expected
	 *	@returns WE_PERMISSION_ERROR if this DEI has no permission to set the value
	 *	@returns WE_INTERFACE_ERROR if the variable owner is not set in DEI's repository
	 *	@returns WE_NOT_FOUND_ERROR if the variable was not found in DEI's variable repository
	 *	@returns WE_TYPE_ERROR if the variable is not an integer variable
	 *	@returns WE_ALREADY_EXISTS_ERROR if the value and/or its validity would not change
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if value is below minimal or above maximum value for this variable
	 *	@returns Implementation dependent return code from the methods called to actually change the value.
	 *
	 *	@see WEVariable and all classes that implement that interface
	 *
	 *	@warning Implementation not finished!
	 *
	 *	This function allows a user to try to set a new integer value to an existing variable.
	 *	The attempt to change the variable value will only be made if the var->minIntValue <= p_valueToSet <= var->maxIntValue.
	 *	If variable is owned by DEI (mainly applies to user defined variables) the variable
	 *	value is changed directly and all subscribers are informed.
	 *	Otherwise DEI will forward a write request to the owner of the variable who in turn
	 *	has to accept by calling forceSetVariableValue() with the new value.
	 *	Before setting the value or forwarding the request the action configured for the corresponding
	 *	event is executed. In case the action changes value of another variable, subscribers for the other
	 *	variable are notified before the subscribers for this variable are notified.
	 *
	 *	@see forceSetVariableValue()
	 *	@see WEVariable::setValue(int)
	 *	@see WEVariableOwner::updateRequest()
	 *	@see DEIIPCClientHandler::sendIPCMessage()
	 *	@see DEIIPCClientHandler::sendUpdateToIPCObservers()
	 */
	int setVariableValue(unsigned short p_varIndex, unsigned short p_varSubIndex, int p_valueToSet, bool p_isValid = true, bool p_calledBySimulation = false);


	/**
	 *	@fn int setVariableValue(unsigned short p_varIndex, unsigned short p_varSubIndex, std::string p_valueToSet, bool p_isValid = true, bool p_calledBySimulation = false)
	 *
	 *	@brief Set a new value to a variable
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_varIndex Index of the variable to which a new value shall be set
	 *	@param[in] p_varSubIndex Subindex of the variable to which a new value shall be set
	 *	@param[in] p_valueToSet New value of the variable
	 *
	 *	@returns WE_SUCCESS if setting the value worked as expected
	 *	@returns WE_PERMISSION_ERROR if this DEI has no permission to set the value
	 *	@returns WE_INTERFACE_ERROR if the variable owner is not set in DEI's repository
	 *	@returns WE_NOT_FOUND_ERROR if the variable was not found in DEI's variable repository
	 *	@returns WE_TYPE_ERROR if the variable is not a string variable
	 *	@returns WE_ALREADY_EXISTS_ERROR if the value and/or its validity would not change
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if value is below minimal or above maximum value for this variable
	 *	@returns Implementation dependent return code from the methods called to actually change the value.
	 *
	 *	@see WEVariable and all classes that implement that interface
	 *
	 *	@warning Implementation not finished!
	 *
	 *	This function allows a user to try to set a new string value to an existing variable.
	 *	If variable is owned by DEI (mainly applies to user defined variables) the variable
	 *	value is changed directly and all subscribers are informed.
	 *	Otherwise DEI will forward a write request to the owner of the variable who in turn
	 *	has to accept by calling forceSetVariableValue() with the new value.
	 *	Before setting the value or forwarding the request the action configured for the corresponding
	 *	event is executed. In case the action changes value of another variable, subscribers for the other
	 *	variable are notified before the subscribers for this variable are notified.
	 *
	 *	@see forceSetVariableValue()
	 *	@see WEVariable::setValue(std::string)
	 *	@see WEVariableOwner::updateRequest()
	 *	@see DEIIPCClientHandler::sendIPCMessage()
	 *	@see DEIIPCClientHandler::sendUpdateToIPCObservers()
	 */
	int setVariableValue(unsigned short p_varIndex, unsigned short p_varSubIndex, const std::string & p_valueToSet, bool p_isValid = true, bool p_calledBySimulation = false);


	/**
	 *	@fn getVariableValueFromOwner(unsigned short p_varIndex, unsigned short p_varSubIndex)
	 *
	 *	@brief Get a new value of an external variable (owned by a foreign stack)
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param[in] p_varIndex Index of the variable where the value which we want to read shall be stored
	 *	@param[in] p_varSubIndex Subindex of the variable where the value which we want to read shall be stored
	 *
	 *	@returns WE_SUCCESS if setting the value worked as expected
	 *	@returns WE_PERMISSION_ERROR if this DEI has no permission to set the value
	 *	@returns WE_NOT_FOUND_ERROR if the variable was not found in DEI's variable repository
	 *	@returns WE_ALREADY_EXISTS_ERROR if the value and/or its validity would not change
	 *	@returns Implementation dependent return code from the variable
	 *
	 *	@see WEVariable and all classes that implement that interface
	 *
	 *	This function allows a user to try to get a value of an Object Dictionary of a of a foreign stack.
	 *	If variable is owned by a foreign stack(The variable must have an external identifier)
	 *	All subscribers are informed when we get the new value.
	 *
	 *	This seems exclusively triggered via PClient's JS function with the same name.
	 *
	 * @warning Implementation not finished!
	 */
	int getVariableValueFromOwner(unsigned short p_varIndex, unsigned short p_varSubIndex) const;

	/**
	 *	@fn const std::string & getInstanceName(void) const
	 *
	 *	@brief Simple get function that returns the name of this DEI
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Name of the DEI
	 */
	const std::string & getInstanceName(void) const
	{
		return instanceName;
	}


	/**
	 *	@fn void decomposeCombinedKey(unsigned int p_combinedKey,
	 *		unsigned short * p_indexToReturn,
	 *		unsigned short * p_subindexToReturn)
	 *
	 *	@brief Decomposes a combined key and stores value of index and subindex in the provided pointers
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_combinedKey The combined key of the variable map that shall be decomposed
	 *	@param[out] p_indexToReturn Pointer to an integer where the extracted index
	 *		of the combined key gets stored to
	 *	@param[out] p_subindexToReturn Pointer to an integer where the extracted subindex
	 *		of the combined key gets stored to
	 *
	 *	@returns void
	 *
	 *	@see calculateCombinedKey(unsigned short, unsigned short)
	 */
	static void decomposeCombinedKey(unsigned int p_combinedKey,
			unsigned short * p_indexToReturn,
			unsigned short * p_subindexToReturn)
	{
		*p_indexToReturn = p_combinedKey >> 16;
		*p_subindexToReturn = p_combinedKey;
	}


	/**
	 *	@fn int registerVariableOwner(unsigned short p_index,
	 *		unsigned short p_subindex,
	 *		WEVariableOwner * p_owner)
	 *
	 *	@brief Adds a pointer to the owner of a variable to the variable owner map
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_index Index of the owned variable
	 *	@param[in] p_subindex Subindex of the owned variable
	 *	@param[in] p_owner Pointer to the owner of the variable
	 *
	 *	@returns WE_SUCCESS if adding the owner to the map worked as expected
	 *	@returns WE_ALREADY_EXISTS_ERROR if an owner of the variable identified by
	 *		index/subindex already existed in the map
	 *
	 *	@see WEVariableOwner
	 *	@see WEVariable
	 */
	int registerVariableOwner(unsigned short p_index, unsigned short p_subindex, WEVariableOwner * p_owner);


	/**
	 *	@fn WEIPCEndPoint * getIPCEndPointByPeerAddress(const std::string& p_peerAddress)
	 *
	 *	@brief Retrieve an IPC endpoint identified by it's peer address
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@param p_peerAddress A string containing the peer address of an IPC endpoint
	 *
	 *	@returns A pointer to the IPC endpoint identified by p_peerAddress
	 *	@returns nullptr if no IPC endpoint with the specified peer address was found
	 */
	std::shared_ptr<WEIPCEndPoint> getIPCEndPointByPeerAddress(const std::string& p_peerAddress) const;


	/**
	 *	@fn int addIPCEndPoint(WEIPCEndPoint* endpoint)
	 *
	 *	@brief add an IPC endpoint with the DEI where it could be retrieved based
	 *			on its peer address through  getIPCEndPointByPeerAddress().
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@param endpoint endpoint to be added
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	Although the endpoints are added to the endpoint manager also but since now
	 *	the endpoint manager deals with a generic interface for endpoints it could not
	 *	retrieve the endpoint by its peer address.
	 */
	int addIPCEndPoint(std::shared_ptr<WEIPCEndPoint> p_endpoint);


	int removeIPCEndPoint(std::shared_ptr<WEIPCEndPoint> p_endpoint);

	bool isEndpointConnected(const std::string& p_peerAddress) const;

	/**
	 *	@fn WEEndPointManager * getEndPointManager(void)
	 *
	 *	@brief Returns the pointer to DEI's endpoint manager instance
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns pointer to DEI's endpoint manager instance
	 */
	WEEndPointManager * getEndPointManager(void) const
	{
		return ipcEndPointManager;
	}


	/**
	 *	@fn DEIIPCClientHandler * getIPCClientHandler(void)
	 *
	 *	@brief Returns the pointer to DEI's IPC client handler  instance
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns pointer to DEI's IPC client handler instance
	 */
	DEIIPCClientHandler * getIPCClientHandler(void) const
	{
		return ipcClientHandler.get();
	}


	/**
	 *	@fn int forceSetVariableValue(unsigned short p_index, unsigned short p_subindex, const int p_intValue, const std::string& p_owner, bool p_isValid = true, bool p_calledBySimulation = false)
	 *
	 *	@brief Owner of variables may call this function to set the value of a variable they own.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the variable who's value shall be set
	 *	@param p_subindex Subindex of the variable who's value shall be set
	 *	@param p_intValue Integer value that shall be set
	 *	@param p_owner Name of the owner of the variable
	 *	@param p_isValid True: Variable value should be valid, false otherwise
	 *
	 *	@returns WE_NOT_FOUND_ERROR If the variable was not found in DEI's variable map
	 *	@returns WE_PERMISSION_ERROR If p_owner is not matching the variable's owner information
	 *	@returns the return value of WEVariable::setValue()
	 *
	 *	This function should only be called by variable owners who really want to change the current value
	 *	of their variables. This set value will be distributed to all subscriber or via the events.
	 *	Before setting the value the action configured for the corresponding event is executed.
	 *	In case the action changes value of another variable, subscribers for the other
	 *	variable are notified before the subscribers for this variable are notified.
	 */
	int forceSetVariableValue(unsigned short p_index, unsigned short p_subindex, const int p_intValue, const std::string& p_owner, bool p_isValid = true, bool p_calledBySimulation = false);


	/**
	 *	@fn int forceSetVariableValue(unsigned short p_index, unsigned short p_subindex, const std::string& p_stringValue, const std::string& p_owner, bool p_isValid = true, bool p_calledBySimulation = false)
	 *
	 *	@brief Owner of variables may call this function to set the value of a variable they own.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the variable who's value shall be set
	 *	@param p_subindex subindex of the variable who's value shall be set
	 *	@param p_stringValue String value that shall be set
	 *	@param p_owner Name of the owner of the variable
	 *	@param p_isValid True: Variable value should be valid, false otherwise
	 *
	 *	@returns WE_NOT_FOUND_ERROR If the variable was not found in DEI's variable map
	 *	@returns WE_PERMISSION_ERROR If p_owner is not matching the variable's owner information
	 *	@returns the return value of WEVariable::setValue()
	 *
	 *	This function should only be called by variable owners who really want to change the current value
	 *	of their variables. This set value will be distributed to all subscriber or via the events.
	 *	Before setting the value the action configured for the corresponding event is executed.
	 *	In case the action changes value of another variable, subscribers for the other
	 *	variable are notified before the subscribers for this variable are notified.
	 */
	int forceSetVariableValue(unsigned short p_index, unsigned short p_subindex, const std::string& p_stringValue, const std::string& p_owner, bool p_isValid = true, bool p_calledBySimulation = false);


	/**
	 *	@fn int markValueAsInvalid(unsigned short p_index, unsigned short p_subindex, bool p_forceEvent = false)
	 *
	 *	@brief Marks the value of a variable as invalid
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the variable which's value shall be marked as invalid
	 *	@param p_subindex Subindex of the variable which's value shall be marked as invalid
	 *	@param p_forceEvent - send out event even if variable already is invalid - to be used when subscription response arrives
	 *
	 *	@returns WE_SUCCESS If the value was marked as invalid
	 *	@returns WE_NOT_FOUND_ERROR If the variable identified by p_index/p_subindex was not found
	 */
	int markValueAsInvalid(unsigned short p_index, unsigned short p_subindex, bool p_forceEvent = false);


	/**
	 *	@fn int markOwnerVariablesInvalid(std::string p_owner)
	 *
	 *	@brief Mark all variables of an owner identified by p_owner as invalid
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_owner String identifying an owner of one or more variables
	 *
	 *	@returns WE_SUCCESS
	 */
	int markOwnerVariablesInvalid(const std::string & p_owner);


	/**
	 *	@fn int markValueAsValid(unsigned short p_index, unsigned short p_subindex)
	 *
	 *	@brief Marks the value of a variable as valid
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the variable which's value shall be marked as valid
	 *	@param p_subindex Subindex of the variable which's value shall be marked as valid
	 *
	 *	@returns WE_SUCCESS If the value was marked as valid
	 *	@returns WE_NOT_FOUND_ERROR If the variable identified by p_index/p_subindex was not found
	 */
	int markValueAsValid(unsigned short p_index, unsigned short p_subindex);


	/**
	 *	@fn unsigned int calculateCombinedKey(unsigned short p_varIndex, unsigned short p_varSubIndex) const
	 *
	 *	@brief Calculates the combined key for the variableMap using index and subindex.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_varIndex Index of the variable for which the combined key shall be calculated
	 *	@param[in] p_varSubIndex Sub-Index of the variable for which the combined key shall be calculated
	 *
	 *	@returns The combined key calculated using Index and Sub-Index of a variable
	 *
	 *	To uniquely identify a variable inside the DEI's variable map, an unsigned int
	 *	key is used. This function combines two unsigned short indices to one combined key.
	 *	E.g.:
	 *	Index:     0x2000
	 *	Sub-Index: 0x0010
	 *	Resulting combined key: 0x20000010
	 *
	 *	@see decomposeCombinedKey(unsigned int, unsigned short *, unsigned short *)
	 */
	static unsigned int calculateCombinedKey(unsigned short p_varIndex, unsigned short p_varSubIndex)
	{
		//self explaining...
		return ( (p_varIndex << 16) | (p_varSubIndex) ) & 0xffffffff;
	}


	/**
	 *	@fn WEEventQueue * getEventQueueOutgoing(void)
	 *
	 *	@brief Returns a pointer to the outgoing event queue
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Pointer to outgoing event queue
	 */
	WEEventQueue * getEventQueueOutgoing(void) const
	{
		return this->sendQueue.get();
	}


	/**
	 *	@fn int addVariableToEventList(unsigned short p_index, unsigned short p_subindex)
	 *
	 *	@brief Adds a variable's indices to the list of variables that
	 *		cause an event to be fired if it's value changes.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the variable that shall be added to the list of variables that causes events
	 *	@param p_subindex Subindex of the variable that shall be added to the list of variables that causes events
	 *
	 *	@returns Always WE_SUCCESS
	 *
	 *	This function adds the index/subindex of a variable to list of variables that cause
	 *	an event to be fired if it's value changes.
	 */
	int addVariableToEventList(unsigned short p_index, unsigned short p_subindex);


	/**
	 *	@fn int addVariableToEventList(unsigned int p_varKey)
	 *
	 *	@brief Adds a variable's key to the list of variables that
	 *		cause an event to be fired if it's value changes.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_varKey Key of the variable that shall be added to the list of variables that causes events
	 *
	 *	@returns Always WE_SUCCESS
	 *
	 *	This function adds the key of a variable to list of variables that cause
	 *	an event to be fired if it's value changes.
	 *	p_varKey shall be calculated as ( index << 16 | (subindex & 0xffff) ).
	 *
	 *
	 *	@see WEDEI::calculateCombinedKey()
	 */
	int addVariableToEventList(unsigned int p_varKey);


	/**
	 *	@fn int removeVariableFromEventList(unsigned short p_index, unsigned short p_subindex)
	 *
	 *	@brief Removes a variable's indices from the list of variables that
	 *		cause an event to be fired if it's value changes.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the variable that shall be removed from the list of variables that causes events
	 *	@param p_subindex Subindex of the variable that shall be removed from the list of variables that causes events
	 *
	 *	@returns WE_SUCCESS if the indices were removed
	 *	@returns WE_NOT_FOUND_ERROR if no entry was found
	 *
	 *	@warning After calling this function no events will be triggered if the variable's value change!
	 *
	 *	This function removes the index/subindex of a variable from the list of variables that cause
	 *	an event to be fired if it's value changes.
	 */
	int removeVariableFromEventList(unsigned short p_index, unsigned short p_subindex);


	/**
	 *	@fn int dumpVariables(std::string p_filePath)
	 *
	 *	@brief Debug function to print all variables to a file
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_filePath File name/path to dump the variables into
	 *
	 *	@returns WE_SUCCESS if everything worked
	 *	@returns WE_FILE_ERROR else
	 *
	 *	Calling this function will result in the creation of a file that contains
	 *	information about all variables contained in this DEI in the following format:
	 *	Index	Subindex	Name	Value	Owner
	 */
	int dumpVariables(const std::string & p_filePath) const;


	/**
	 *	@fn int setHardwareVariableOwner(WEVariableOwner * p_owner)
	 *
	 *	@brief With this function the pointer to the owner of hardware variables can be set
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_owner Pointer to the variable owner
	 *
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_owner is nullptr
	 *	@returns WE_SUCCESS else
	 */
	int setHardwareVariableOwner(WEVariableOwner * p_owner);


	/**
	 *	@fn int getVariableName(unsigned int p_combinedKey, std::string * p_nameToReturn)
	 *
	 *	@brief Function that stores the name of a variable identified by p_combinedKey into p_nameToReturn
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_combinedKey Combination of index/subindex of a variable
	 *	@param p_nameToReturn Address of a std::string to which the name of the variable will be stored
	 *
	 *	@returns WE_SUCCESS if retreiving the name worked
	 *	@returns WE_NOT_FOUND_ERROR if a variable with the provided combined key was not found
	 */
	int getVariableName(unsigned int p_combinedKey, std::string * p_nameToReturn) const;


	/**
	 *	@fn unsigned int getMaxStringLength(unsigned int p_combinedKey)
	 *
	 *	@brief Get maximum length defined for variable identified by p_combinedKey.
	 *
	 *	@param p_combinedKey Combination of index/subindex of a variable
	 *
	 *	@returns The length in bytes set for the variable
	 *	@returns 0 if p_combinedKey does not identify a string variable
	 *
	 *	@see calculateCombinedKey(unsigned short, unsigned short)
	 */
	unsigned int getMaxStringLength(unsigned int p_combinedKey) const;


	/**
	 *	@fn void resubscribeVariables(WEIPCEndPoint *variableOwnerEndPoint)
	 *
	 *	@brief Resubscribe variables to an owner (IPC)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param variableOwnerEndPoint Pointer to the owner endpoint
	 *
	 *	When the IPC connection to a variable owner was lost and has been re-established,
	 *	this function can be used to subscribe to the owner's variables again.
	 */
	void resubscribeVariables(std::shared_ptr<WEIPCEndPoint> p_variableOwnerEndPoint);


	/**
	 *	@fn bool isVariableValid(unsigned short p_index, unsigned short p_subindex)
	 *
	 *	@brief Check if a value of a variable is valid or not
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of a variable
	 *	@param p_subindex Subindex of a variable
	 *
	 *	@returns true if the value of the variable identified by p_index/p_subindex is valid
	 *	@returns false if value is invalid
	 */
	bool isVariableValid(unsigned short p_index, unsigned short p_subindex) const;


	/**
	 *	@fn std::string * getVariableOwner(unsigned short p_index, unsigned short p_subindex)
	 *
	 *	@brief Retrieve the name of a variable's owner
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the variable
	 *	@param p_subindex Subindex of the variable
	 *
	 *	@returns A pointer to a string containing the name of the variable owner
	 *	@returns nullptr if variable was not found
	 */
	const std::string & getVariableOwner(unsigned short p_index, unsigned short p_subindex) const;

	WEVariableOwner * getVariableOwnerPointer(unsigned short p_index, unsigned short p_subindex) const;

	/**
	 * Get the copy of the value of an integer variable.
	 * If you are not sure if the variable exist or if it is of storage type integer, check that before
	 * calling this function. If it is an integer variable, it is guaranteed that the current value of it will be returned.
	 * If it is not an integer variable or if it does not exist, -1 will be returned (which might also be a valid value,
	 * so do the tests before!).
	 * @param p_index index of the variable
	 * @param p_subindex of the variable
	 * @return Copy of the value of the variable
	 */
	int getIntValue(unsigned short p_index, unsigned short p_subindex) const;

	/**
	 * Get a const reference to the value of a string variable.
	 * If you are not sure if the variable exist or if it is of storage type string, check that before
	 * calling this function. If it is an string variable, it is guaranteed that the current value of it will be returned.
	 * If it is not a string variable or if it does not exist, a reference to an ampty string will be returned
	 * (which might also be a valid value, so do the tests before!).
	 * @param p_index index of the variable
	 * @param p_subindex of the variable
	 * @return Copy of the value of the variable
	 */
	const std::string & getStringValue(unsigned short p_index, unsigned short p_subindex) const;

	/**
	 *	Ask Dirk if name of the method should not be sufficient as documentation.
	 *
	 *	Seems to be used for CANopen only.
	 */
	const DEIVariableMap * getCopyOfVariableMap(void) const;


	/**
	 *	Ask Dirk if name of the method should not be sufficient as documentation.
	 *
	 *	Seems to be used for hardware daemon only.
	 */
	void startIPCThread(void);


	/**
	 *	Ask Dirk if name of the method should not be sufficient as documentation.
	 *
	 *	Seems this was used used for old CAN Freestyle only.
	 */
	WEVariableType getVariableType(unsigned int p_key) const;


	/**
	 *	Ask Dirk if name of the method should not be sufficient as documentation.
	 *
	 *	Seems this was used used for old CAN Freestyle only.
	 */
	int getInternalVariableType(unsigned int p_key) const;


	/**
	 *	check if variable is signed or unsigned
	 *	@param p_key combined key of variable
	 *	@return true for signed variables, false for unsigned variables
	 *
	 *	Note that string variables can also have the sign information...
	 */
	bool isVariableSigned(unsigned int p_key) const;


	/**
	 *	Ask Dirk if name of the method should not be sufficient as documentation.
	 *
	 *	Seems to be used for CANopen only.
	 */
	int getVariableLength(unsigned int p_key) const;

	/**
	 *	Ask Dirk if name of the method should not be sufficient as documentation.
	 */
	void setAllVariablesAreObserved(bool p_observed);

	/**
	 *	Ask Dirk if name of the method should not be sufficient as documentation.
	 */
	void allowValueChangeOnlyBySimulation(unsigned short p_index, unsigned short p_subindex, bool p_simOnly);

	/**
	 *	Ask Dirk if name of the method should not be sufficient as documentation.
	 */
	void removeAllSimulationValueChangeSettings(void);

	/**
	 *	Returns the path where remanent variables owned by this DEI store their value files
	 */
	static const std::string & getRemanentVariablesPath(void);

	/**
	 *	Check if all subscription responses were received from a particular owner.
	 *	@param ownerString IPC socket name of the owner (e.g. /tmp/hardwareDaemon)
	 *	@return true if all pending request responses were received
	 *		false if there are pending responses
	 */
	bool allSubscriptionResponsesReceived(const std::string & p_ownerString) const;

	/**
	 *	Call this function to signal DEI that a subscription request response was received
	 *	@param ownerString IPC socket name of the owner (e.g. /tmp/hardwareDaemon)
	 *	@param combinedKey The combined key of the variable for which a response was received
	 *	@see allSubscriptionResponsesReceived() above
	 */
	void eventSubscriptionResponseReceived(const std::string & p_ownerString, unsigned int p_combinedKey);

	/**
	 * Check if variable with p_index and p_subindex exists
	 * @param p_index Index of the variable
	 * @param p_subindex subindex of the variable
	 * @return true if the variable exists
	 */
	bool variableExists(unsigned short p_index, unsigned short p_subindex) const;

	/**
	 * Check if variable with p_key (as combinde index/subindex) exists
	 * @param p_key Combined key of the variable (index/subindex)
	 * @return true if the variable exists
	 */
	bool variableExists(unsigned int p_key) const;

	/**
	 * Check if the variable identified by p_index and p_subindex is of storage type integer
	 * @param p_index Index of the variable
	 * @param p_subindex subindex of the variable
	 * @return true if the variable exists and is of storage type integer
	 */
	bool isIntVariable(unsigned short p_index, unsigned short p_subindex) const;

	/**
	 * Check if the variable identified by p_index and p_subindex is of storage type integer
	 * @param p_key Combined key of the variable (index/subindex)
	 * @return true if the variable exists and is of storage type integer
	 */
	bool isIntVariable(unsigned int p_key) const;

	/**
	 * Check if the variable identified by p_index and p_subindex is of storage type string
	 * @param p_index Index of the variable
	 * @param p_subindex subindex of the variable
	 * @return true if the variable exists and is of storage type string
	 */
	bool isStringVariable(unsigned short p_index, unsigned short p_subindex) const;

	/**
	 * Check if the variable identified by p_key is of storage type string
	 * @param p_key Combined key of the variable (index/subindex)
	 * @return true if the variable exists and is of storage type string
	 */
	bool isStringVariable(unsigned int p_key) const;

	/**
	 * Registers a WEVariableOwner as an owner for local variables.
	 * When variables belonging to a registered local owner are added
	 * to DEI, these are handled like normal internal variables
	 * (e.g. DEI will take care of the remanent values etc. but updateRequest()
	 * of the owner will be called if a value is set to the variable.
	 */
	void registerLocalVariableOwner(WEVariableOwner * p_owner);

	static int debugSection;
};
} // namespace WE

#endif /* WE_DEI_H_ */
