/**
 *	@file we_dei_ipc_server_handler.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief The only purpose of this class is to add a new DEIIPCClientHandler for every new IPC connection
 *
 *	@see DEIIPCClientHandler
 */

#ifndef WE_DEI_IPC_SERVER_HANDLER_H_
#define WE_DEI_IPC_SERVER_HANDLER_H_

#include "../../os_abstraction/endpoint_manager/include/we_endpoint_handler.h"

namespace WE
{

class WEIPCServer;

/**
 *	@class DEIIPCServerHandler
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Implementing the interface of WEEndPointHandler
 *
 *	@namespace WE
 *
 *	@see WEEndPointHandler
 *	@see DEIIPCClientHandler
 *
 *	The only purpose of this class is to get called whenever a new IPC connection
 *	to the DEI is incoming. The new DEIIPCClientHandler gets registered with the new connection
 *	at the WEEndPointManager.
 */
class DEIIPCServerHandler : public WEEndPointHandler
{
private:
	WEIPCServer* server; //!< Pointer to an IPC server, for accepting connections
public:
	/**
	 *	@fn DEIIPCServerHandler(WEIPCServer * p_server)
	 *
	 *	@brief Constructor of DEIIPCServerHandler
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_server Pointer to an IPC server with which connections shall be accepted.
	 */
	explicit DEIIPCServerHandler(WEIPCServer * p_server);

	/**
	 *	@fn ~DEIIPCServerHandler()
	 *
	 *	@brief Standard destructor of DEIIPCServerHandler
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~DEIIPCServerHandler() = default;

	/**
	 *	@fn void onEndPointReadable(WEEndPoint* p_endPoint, WEEndPointManager* p_endPointMgr)
	 *
	 *	@brief Registers new connections to the end point manager.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_endPoint This parameter is not used in this function.
	 *	@param[in] p_endPointMgr Pointer to the end point manager that manages the new connection.
	 *
	 *	@returns void
	 *
	 *	@see WEEndPointManager
	 */
	void onEndPointReadable(std::shared_ptr<WEEndPoint> p_endPoint, WEEndPointManager* p_endPointMgr) override;

	/**
	 *	@fn void onEndPointReconnected(WEEndPoint* p_endpoint);
	 *
	 *	@brief Call this function if a IPC connection was lost and could be re-established
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endpoint Endpoint of the connection that was re-connected
	 *
	 *	Internally re-subscribes the variables owned by the variable owner identified by p_endpoint
	 */
	void onEndPointReconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;

	/**
	 *	@fn void onEndPointDisconnected(WEEndPoint* p_endpoint)
	 *
	 *	@brief Call this function if a IPC connection was lost
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endpoint Endpoint of the connection that was disconnected
	 *
	 */
	void onEndPointDisconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;
};
} // namespace WE




#endif /* WE_DEI_IPC_SERVER_H_ */
