/**
 *	@file we_variable_logger.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file contains the class definition of WEVariableLogger
 *
 *	@see Class comment for more information
 */

#ifndef WE_VARIABLE_LOGGER_H_
#define WE_VARIABLE_LOGGER_H_

#include <vector>
#include <fstream>
#include <map>

#include "../../../event_handling/include/we_signal.h"
#include "../../variables/include/we_variable_definitions.h"


namespace WE
{

class WEVariableListElement
{
public:
	WEVariableListElement() = default;

	virtual ~WEVariableListElement() = default;

	//public members
	WEVariableStorageType storageType{STORAGE_TYPE_UNDEFINED};

	WEVariableListElement(const WEVariableListElement & p_element) = default;
	WEVariableListElement & operator=(const WEVariableListElement &p_element) = default;
};


typedef std::multimap<int, unsigned int> MMIntToULong;	//!< multi hashmap <int, ulong>
typedef std::map<unsigned int, int*> MULongToIntP;	//!< hashmap <ulong, int*>
typedef std::map<unsigned int, WEVariableListElement*> CombindedKeyToVariable;
typedef std::map<unsigned int, std::string*> MULongToStringP;	//!< hashmap <ulong, string*>

class WEDEI; //!< Forward declaration of WEDEI

/**
 *	@class WEVariableLogger
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Class that is responisble for logging variable data to a file
 *
 *	@namespace WE
 *
 *	@warning: Call the instance() function of WEVariableLogger ONLY AFTER an instance of WEDEI has been created.
 *
 *	This class can be configured for event based variable logging or time based variable logging.
 *	Event based logging: Only log the value of a variable due to certain events (e.g. value change
 *	or when entering a page).
 *	Time based logging: When a given timer expires, all values of configured variables are logged
 *	to the log file.
 *	This class is implemented as a singleton: Only one instance of this class exists per process.
 *	Create a instance of this class by calling the WEVariableLogger::instance() method. When this
 *	method is called after an instance has been created, it returns a pointer to this instance.
 *
 *	WEVariableLogger is capable of replacing the following tags in strings:
 *	\<YYYY\> - current year (e.g. 2010)
 *	\<YY\> - last two digits of current year (e.g. 10)
 *	\<MM\> - current month
 *	\<DD\> - current day
 *	\<hh\> - current hour
 *	\<mm\> - current minute
 *	\<ss\> - current seconds
 *	\<s/10\> - current milliseconds
 *	\<LIST_OF_LOGGED_VARAIBLES\> - A list of all variables sorted by index/subindex of the variable and the storage
 *		type (first integer variables, then string variables)
 *	\<any variable name\> - The name of the variable will be replaced by the current value of the variable
 */
class WEVariableLogger : public WESignalListener
{
public:
	/**
	 *	@fn ~WEVariableLogger()
	 *
	 *	@brief Standard destructor of WEVaraibleLogger
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	Only sets pointer to myDEI = nullptr
	 */
	virtual ~WEVariableLogger();

	/**
	 *	@fn static WEVariableLogger * instance(void)
	 *
	 *	@brief Singleton instance function that reurns the (one and only) instance to the variable logger
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns The only existing instance of WEVariableLogger
	 *
	 *	Call this function to get or create the singleton-instance of WEVariableLogger
	 */
	static WEVariableLogger * instance(void);

	/**
	 *	@fn int init(void)
	 *
	 *	@brief Initialize WEVariableLogger with sane default values and get the pointer to WEDEI
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS if initialization was successful
	 *	@returns WE_INIT_ERROR if initialization failed (because of problems with WEDEI pointer)
	 *
	 *	@warning This function needs to be called before you can use WEVariableLogger
	 */
	int init(void);

	/**
	 *	@fn int setTimeBasedLogging(bool p_timeBased)
	 *
	 *	@brief Choose between time based or event based logging
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_timeBased If p_timeBased is set to true, WEVariableLogger will do time based logging.
	 *		If p_timeBased is set to false, WEVariableLogger will do event based logging
	 *
	 *	@returns WE_SUCCESS if setting this option was successful.
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if this function was called while logging is in progress.
	 *
	 *	@warning Only call this function when logging is stopped
	 *
	 *	With this function the general option can be set if time or event based logging shall be done.
	 *	This will have a major impact on how data is logged.
	 */
	int setTimeBasedLogging(bool p_timeBased);

	/**
	 *	@fn int setTimeBasedLoggingTimer(unsigned int p_timer)
	 *
	 *	@brief Set the timer of time based logging. When this timer expires, data will be logged.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_timer Timer for time based logging in milliseconds
	 *
	 *	@returns WE_SUCCESS if setting the timer was successful
	 *	@returns WE_OUT_OF_BOUNDS_ERROR If the timer was set too low
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if this function was called when logging is in progress
	 *
	 *	@warning Do not call this function while logging is in progress
	 *
	 *	If the timer set by this function expires, the value of all registered variables will be
	 *	logged to the log file.
	 */
	int setTimeBasedLoggingTimer(unsigned int p_timer);

	/**
	 *	@fn bool isTimeBased(void)
	 *
	 *	@brief Is logger configured for time based or event based logging?
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns true if WEVariableLogger is configured for time based logging
	 *	@returns false if WEVariableLogger is configured for event based logging
	 */
	bool isTimeBased(void) const;

	/**
	 *	@fn int setLogFileName(std::string p_fileName)
	 *
	 *	@brief Set the name of the log file
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_fileName Name of the log file as a string. May contain tags.
	 *
	 *	@returns WE_SUCCESS if setting the log file name was successful
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if this function was called while logging is in progress
	 *
	 *	The desired log file name may continue tags such as \<YYYY\> \<hh\>, etc. See class description for more
	 *	information.
	 */
	int setLogFileName(const std::string & p_fileName);

	/**
	 *	@fn const std::string & getLogFileName(void)
	 *
	 *	@brief Returns the log file name like it was set by the user (tags not replaced)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Name of the log file like the user has set it. Tags are not replaced
	 */
	const std::string & getLogFileName(void) const;

	/**
	 *	@fn const std::string & getCurrentLogFileName(void)
	 *
	 *	@brief Returns the current log file name complete with replaced tags
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns the current log file name complete with replaced tags
	 */
	const std::string & getCurrentLogFileName(void) const;

	/**
	 *	@fn int setLogPath(std::string p_path)
	 *
	 *	@brief Set the file system path to where the log file shall be saved
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_path file system path to where the log file shall be saved to. (Directory must exist)
	 *
	 *	@returns WE_SUCCESS if the log path could be set
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if this function was called while logging is in progress
	 *	@returns WE_FILE_ERROR if p_path is not a valid directory
	 *
	 *	@warning Directory p_path MUST exist on the file system.
	 *	@warning Do not use tags in p_path, they won't be replaced!
	 *
	 *	The log path can only be set to an existing directory on the file system.
	 *	Any tags included in p_path will NOT be replaced and will probably lead to an unsuccessful
	 *	execution of this function.
	 *	If p_path is a valid directory but a '/' is missing at the end, a '/' will be appended.
	 */
	int setLogPath(const std::string & p_path);

	/**
	 *	@fn const std::string & getLogPath(void)
	 *
	 *	@brief Retrieve the currently set log path
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Current log path
	 */
	const std::string & getLogPath(void) const;

	/**
	 *	@fn int setDelimiter(char p_delimiter)
	 *
	 *	@brief Set the delimiter with which entries in the log file are separated
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_delimiter character that will be used to separate entries in the log file
	 *
	 *	@returns WE_SUCCESS if setting the delimiter was successful
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if this function was called while logging is in progress
	 *
	 *	@warning Be sure to stop logging before you set the delimiter
	 *
	 *	The default delimiter is a ';'
	 *	Each entry in the log file will be separated by this delimiter.
	 *	Example for time based logging (delimiter ';'):
	 *	2010-01-27 13:37:42.4: 42;My String Value;1337;
	 *	2010-01-27 13:37:43.4: 43;My Changed String Value;1400;
	 *
	 *	Example for event based logging (delimiter: ','):
	 *	2010-01-27 13:37:42.0: \@RTC_Seconds,OnValueChange(int),42
	 *	2010-01-27 13:37:59.6: \@ProjektCurrentPage,OnPageInit,12
	 */
	int setDelimiter(char p_delimiter);

	/**
	 *	@fn char getDelimiter(void)
	 *
	 *	@brief Retrieve the currently set delimiter
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Currently set delimiter
	 *
	 *	@see setDelimiter(char p_delimiter)
	 */
	char getDelimiter(void) const;

	/**
	 *	@fn int setEventTimer(unsigned int p_timerNumber, unsigned int p_timerValue)
	 *
	 *	@brief Set one of the event timers for event based logging
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_timerNumber Number of the timer that shall be set (1, 2 or 3)
	 *	@param[in] p_timerValue The timer value that shall be set in milliseconds (>= 500ms)
	 *
	 *	@returns WE_SUCCESS if setting the timer was successful
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if the timer value is \< MINIMUM_LOG_INTERVAL
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if this function is called while logging is in progress
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if p_timerNumber is not 1, 2 or 3
	 *
	 *	@warning Call stopLogging() before you call this function, else it fails.
	 *
	 *	Every time one of the set timers expires, an event will be generated that logs all variable values
	 *	that where configured for one of the timers.
	 */
	int setEventTimer(unsigned int p_timerNumber, unsigned int p_timerValue);

	/**
	 *	@fn unsigned int getEventTimer(unsigned int p_timerNumber)
	 *
	 *	@brief Retrieve one of the event timers
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timerNumber the number of the timer that shall be returned (1, 2 or 3)
	 *
	 *	@returns The value of the timer identified by p_timerNumber
	 *
	 *	@see setEventTimer(unsigned int p_timerNumber, unsigned int p_timerValue)
	 */
	unsigned int getEventTimer(unsigned int p_timerNumber) const;

	/**
	 *	@fn int setLinePrefix(std::string p_prefix)
	 *
	 *	@brief Set a line prefix that will be printed at the beginning of each log line
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_prefix A string that represents the line prefix
	 *
	 *	@returns WE_SUCCESS if setting the prefix was successful
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if this function is called while logging is in progress
	 *
	 *	@warning Do not call this function while logging is in progress
	 *
	 *	The string p_prefix may contain tags that will be replaced every time a log line is printed.
	 *	Example: p_prefix is "\<YYYY\>-\<MM\>-\<DD\> \<hh\>:\<mm\>:\<ss\>: "
	 *	The output will be:
	 *	2010-01-27 13:37:42: \<variable values\>
	 *	2010-01-27 13:37:43: \<variable values\>
	 *
	 *	2010-01-27 01:38:00: \<variable values\>
	 */
	int setLinePrefix(const std::string & p_prefix);

	/**
	 *	@fn const std::string & getLinePrefix(void)
	 *
	 *	@brief Return the line prefix
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Currently set line prefix (including not replaced tags)
	 *
	 *	@see setLinePrefix(std::string p_prefix)
	 */
	const std::string & getLinePrefix(void) const;

	/**
	 *	@fn int setFileHeaderText(std::string p_header)
	 *
	 *	@brief Set the file header text that will be printed on top of each log file
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_header String containing the text of the file header (may include tags)
	 *
	 *	@returns WE_SUCCESS if setting the file header text was successful
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if this function is called while logging is in progress
	 *
	 *	@warning Do not call this function while logging is in progress
	 *
	 *	The file header text will appear on top of every log file. If continuous logging is selected,
	 *	the header will also be printed into the log file if logging was stopped and then started again.
	 *	Any supported tag inside p_header will be replaced when printing.
	 */
	int setFileHeaderText(const std::string & p_header);

	/**
	 *	@fn std::string getFileHeaderText(void)
	 *
	 *	@brief Return file header with replaced tags
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Log file header text with all tags replaced
	 *
	 *	@see setFileHeaderText(std::string p_header)
	 *
	 *	This function returns the set file header with all tags replaced.
	 */
	std::string getFileHeaderText(void) const;

	/**
	 *	@fn int setLogLevel(int p_level)
	 *
	 *	@brief Set the current log level
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_level Log level to set (MIN_LOG_LEVEL \<= p_level \<= MAX_LOG_LEVEL)
	 *
	 *	@returns WE_SUCCESS if setting the log level was successful
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if p_level \> MAX_LOG_LEVEL or p_level \< MIN_LOG_LEVEL
	 *
	 *	This function CAN be called while logging is in progress.
	 *	If log level is set to MIN_LOG_LEVEL (normally 0), logging will be stopped.
	 */
	int setLogLevel(int p_logLevel);

	/**
	 *	@fn int getLogLevel(void)
	 *
	 *	@brief Return the current log level
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns The current log level
	 *
	 *	@see setLogLevel(int p_level)
	 */
	int getLogLevel(void) const;

	/**
	 *	@fn int setContinuous(bool p_continuous)
	 *
	 *	@brief Choose between continuous logging or "one file logging"
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_continuous True to set continuous logging, false to set "one file logging"
	 *
	 *	@returns WE_SUCCESS if setting this option was successful
	 *	@returns WE_PERMISSION_ERROR if logging is in progress while this function is called
	 *
	 *	@warning Do not call this function while logging is in progress
	 *
	 *	If this option is set to true, no new file will be generated when re-starting logging.
	 *	If this option is set to false, every time logging is started, a new file is generated
	 *	(be sure to insert tags into filename, else content will still be appended).
	 */
	int setContinuous(bool p_continuous);

	/**
	 *	@fn int registerVariable(unsigned short p_index, unsigned short p_subindex, int p_eventType = 0, int p_logLevel = 0, bool p_timeBased = false)
	 *
	 *	@brief Register a varaible for varaible logging
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] p_index Index of the variable that shall be logged
	 *	@param[in] p_subindex Subindex of the variable that shall be logged
	 *	@param[in] p_eventType Event that causes the value of the varaible to be logged
	 *		For time based logging this parameter may be left blank
	 *	@param[in] p_logLevel Configured log level for the variable that shall be logged
	 *		For time based logging this parameter may be left blank
	 *	@param[in] p_timeBased indicates that the variable is a time based one, needed for keeping order of variables from Projektor <MKA 05.12.2013>
	 *
	 *	@returns WE_SUCCESS if registering the variable was successful
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if this function is called while logging is in progress
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if No event was specified or p_logLevel is out of bounds
	 *	@returns WE_NOT_FOUND_ERROR if the specified variable is not present in DEI
	 *
	 *	@warning Do not call this function while logging is in progress
	 *
	 *	With this function a variable identified by index/subindex can be configured for logging.
	 *	If event based logging is used and the variable shall be configured for more than one event,
	 *	this function has to be called for every event.
	 */
	int registerVariable(unsigned short p_index,
		unsigned short p_subindex,
		int p_eventType = 0,
		int p_logLevel = 0,
		bool p_timeBased = false);

	/**
	 *	@fn int getNamesOfLoggedVariables(std::vector<std::string> * p_vector)
	 *
	 *	@brief Retrieve the names of all logged variables and store the names into p_vector
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_vector A vector that will store all names of all registered variables
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	The vector is sorted like this: First all integer variables sorted by index/subindex, after that all
	 *	string variables sorted by index/subindex.
	 */
	int getNamesOfLoggedVariables(std::vector<std::string> * p_vector) const;

	/**
	 *	@fn int getCurrentLogFileSize(void)
	 *
	 *	@brief Retrieve the size of the current log file
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns The size of the current log file in bytes or -1 if an error occurred
	 */
	long long getCurrentLogFileSize(void) const;

	/**
	 *	@fn int getFreeSpaceOfLogPath(void)
	 *
	 *	@brief Retrieve the free space of the current log directory
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns The free space of the current log directory in bytes
	 */
	long long getFreeSpaceOfLogPath(void) const;

	/**
	 *	@fn bool isLogging(void)
	 *
	 *	@brief Check if logging is in progress or stopped
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns true if logging is in progress
	 *	@returns false if logging is stopped
	 */
	bool isLogging(void) const;

	/**
	 *	@fn int stopLogging(void)
	 *
	 *	@brief Stop data logging
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS if logging was running and could be successfully stopped
	 *	@returns WE_PERMISSION_ERROR if logging was not in progress (was stopped before)
	 *
	 *	@see startLogging()
	 *	@see isLogging()
	 *
	 *	@warning Check if logging is in progress before calling this function
	 *
	 *	Calling this function stops variable logging and the log file will be closed
	 *	(it can be copied or moved after the call).
	 */
	int stopLogging(void);

	/**
	 *	@fn int startLogging(void)
	 *
	 *	@brief Starts variable logging
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS if logging could be started successfully
	 *	@returns WE_PERMISSION_ERROR if logging was already started before
	 *	@returns WE_FILE_ERROR if the log file could not be opened
	 *
	 *	@see stopLogging()
	 *	@see isLogging()
	 *	@see run()
	 *
	 *	@warning Do not call this function while logging is in progress
	 *	@warning Only start logging if you are sure that all settings were made and all
	 *		variables are registered
	 *
	 *	This function opens the configured log file and logging is started. To make timers work,
	 *	the run() method has to be called frequently.
	 */
	int startLogging(void);

	/**
	 *	@brief Send an event to WEVariableLogger, log variable values configured for this event
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_signal WESignal struct containing information about the event
	 *
	 *	@returns an error code - WE_SUCCESS in case the event was successfully handled.
	 *
	 *	@see WESignal
	 *
	 *	This function implements the WESignalListener interface.
	 *	If this function is called, it will be checked variables are configured for that event and
	 *	all variable values for that event will be logged.
	 */
	int signalOccurred(WESignal & p_signal) override;

	/**
	 *	@fn int run(unsigned int p_tick)
	 *
	 *	@brief Call this function frequently so that timers are working
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_tick Your applications current tick count in milliseconds
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	@see WETick
	 *
	 *	This function checks if any of the configured timers are expired and takes appropriate action.
	 *	This function should be called frequently (often enough to match your smalles timer amount).
	 */
	int run(unsigned int p_tick);

	/**
	 *	@fn std::vector<unsigned int> getVariablesByEventType(int p_eventType)
	 *
	 *	@brief Return a vector with variable IDs configured for p_eventType
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_eventType Event type for which the variable IDs shall be returned
	 *
	 *	@returns A vector containing variable IDs (index << 16 | subindex)
	 *		of all variables configured for p_eventType
	 */
	std::vector<unsigned int> getVariablesByEventType(int p_eventType) const;

	/**
	 *	@fn int getMinimumLogLevel(void)
	 *
	 *	@brief This function returns MIN_LOG_LEVEL
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns MIN_LOG_LEVEL (This log level means: stop logging)
	 */
	static int getMinimumLogLevel(void)
	{
		return MIN_LOG_LEVEL;
	}


	int setLogVariableName(bool p_logVarName);
	bool getLogVariableName() const;

	int setLogEventName(bool p_logEventName);
	bool getLogEventName() const;

	static int countLogFilesInFolder(const std::string & p_logPath, std::string* p_oldestFile, long long* p_totalSize);
	static int deleteFile(const std::string & p_logPath, const std::string & p_oldestFile);

	/*
	 * <MKA 14.03.2014> The following four methods are implemented here but currently not used
	 * reserved for future implementations
	 */
	int setNumberOfFiles(int p_tmpValue);
	int getNumberOfFiler() const;
	int setOptLogPath(const std::string & p_path);
	const std::string & getOptLogPath() const;

private:
	/**
	 *	@fn WEVariableLogger()
	 *
	 *	@brief Private constructor of WEVariableLogger. Only called by instance() method
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	The constructor sets values to the constant tag definitions in it's initializer list.
	 */
	WEVariableLogger();

	/**
	 *	@fn int printFileHeader()
	 *
	 *	@brief Prints the configured file header to the log file
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS If writing to the log file was successful
	 *	@returns WE_FILE_ERROR If the log file is not accessible
	 *
	 *	@see setLogFileHeaderText()
	 *
	 *	@warning Only call this function while logging is in progress
	 *
	 *	This function prints the contents of the configured header text into the log file.
	 *	All tags will be replaced by current values.
	 */
	int printFileHeader(void);

	/**
	 *	@fn int printTimeBasedLogLine(void)
	 *
	 *	@brief Print a line to the log file containing all current variable values
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS If printing the line was successful
	 *	@returns WE_PERMISSION_ERROR If this function was called when logging is stopped
	 *
	 *	@see printLinePrefix()
	 *
	 *	@warning Only call this function while logging is in progress
	 *
	 *	This function internally calls printLinePrefix() and then prints all configured variable values
	 *	to the log file; separated with the configured delimiter.
	 */
	int printTimeBasedLogLine(void);

	/**
	 *	@fn int printStopLine(void)
	 *
	 *	@brief Every time logging is stopped, this function gets called to log that logging was stopped
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS If printing the line was successful
	 *	@returns WE_PERMISSION_ERROR If this function was called when logging is stopped
	 *	@returns WE_FILE_ERROR If log file is not present
	 */
	int printStopLine(void);

	/**
	 *	@fn int printStartLine(void)
	 *
	 *	@brief Every time logging is started, this function gets called to log that logging was started
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS If printing the line was successful
	 *	@returns WE_PERMISSION_ERROR If this function was called when logging is stopped
	 *	@returns WE_FILE_ERROR If log file is not present
	 */
	int printStartLine(void);

	/**
	 *	@fn int printEventBasedLogLine(std::string p_variableName, std::string p_eventName, std::string p_value)
	 *
	 *	@brief Print exactly one variable value with one event name and one value to the log file
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_variableName Name of the variable
	 *	@param p_eventName Description of the event that caused the print
	 *	@param p_value String that contains the value that shall be printed
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to print date/time in the logfile
	 *
	 *	@returns WE_SUCCESS if printing was successful
	 *	@returns WE_PERMISSION_ERROR if this function was called while logging is stopped
	 *	@returns WE_FILE_ERROR if the log file is not accessible
	 *
	 *	@see printEventBasedLogLinesForEvent()
	 *
	 *	@warning Only call this function when logging is in progress
	 *
	 *	This function is used only to print exactly one line for one variable and one event.
	 *
	 */
	int printEventBasedLogLine(const std::string & p_variableName, const std::string & p_eventName, const std::string & p_value, unsigned long long p_timestamp);

	/**
	 *	@fn int printEventBasedLogLinesForEvent(int p_eventType)
	 *
	 *	@brief Print all log lines for one event type
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_eventType Type of event that shall cause the logging
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	@warning Only call this function while logging is in progress
	 *	@warning Only call this function for "global" events, see detailed description
	 *
	 *	This function is internally used to print log lines if one of the three event timers expires.
	 *	This function should only be used for "global" events (like the timers, project start, etc.).
	 *	If you call this function with event type VARIABLE_VALUE_CHANGED_EVENT_INT or similar, ALL
	 *	variables configured for this event will be logged even their value did not change.
	 */
	int printEventBasedLogLinesForEvent(int p_eventType);

	/**
	 *	@fn int printLinePrefix(unsigned long long p_timestamp = 0)
	 *
	 *	@brief Prints the configured line prefix to the log file
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to print date/time in the logfile
	 *
	 *	@returns WE_SUCCESS if printing the line prefix to the log file was successful
	 *	@returns WE_PERMISSION_ERROR if this function was called while logging is stopped
	 *	@returns WE_FILE_ERROR if an error occurred while accessing the log file
	 *
	 *	@warning Only call this function while logging is in progress
	 *
	 *	This function prints the line prefix at the beginning of a line in the log file.
	 *	All tags included in the prefix will be replaced with current values.
	 */
	int printLinePrefix(unsigned long long p_timestamp = 0);

	/**
	 *	@fn bool isVariableConfiguredForEvent(const int p_eventType, const unsigned int p_key)
	 *
	 *	@brief Check if variable identified by p_key is configured for event p_eventType
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_eventType Type of event to check if the variable is configured for
	 *	@param p_key Combination of index/subindex of the variable
	 *
	 *	@returns true if variable is configured for the event p_eventType
	 *	@returns false if the variable is not configured for the event p_eventType
	 */
	bool isVariableConfiguredForEvent(const int p_eventType, const unsigned int p_key) const;

	/**
	 *	@fn bool isVariableConfiguredForCurrentLogLevel(const unsigned int p_variableKey)
	 *
	 *	@brief Checks if a variable is configured for the current log level
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_variableKey combination of index/subindex of a varaible
	 *
	 *	@returns true if variable identified by p_varaibleKey is configured for currently active log level
	 *	@returns false if variable identified by p_varaibleKey is NOT configured for currently active log level
	 *
	 *	@warning This functions modifies this->logLevel if it is out of range therefore this can't be const
	 */
	bool isVariableConfiguredForCurrentLogLevel(const unsigned int p_variableKey);

	/**
	 *	@fn int replaceTagsInString(std::string * p_string, unsigned long long p_timestamp = 0)
	 *
	 *	@brief Replace all tags contained in a string
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_string pointer to a string with or without tags. Tags will be replaced.
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to print date/time in the logfile
	 *
	 *	@returns returnValue of replaceVariableNameTagsWithVariableValues()
	 *
	 *	This function internally calls all replaceXXX functions listed below to replace all
	 *	supported tags in a string. The result will be stored in p_string
	 */
	int replaceTagsInString(std::string * p_string, unsigned long long p_timestamp = 0) const;

	/**
	 *	@fn void replaceYearTags(std::string * p_string, unsigned long long p_timestamp = 0)
	 *
	 *	@brief Replace any \<YYYY\> tags contained in a string with the current year
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_string Pointer to a string which tags shall be replaced.
	 *		Result will be stored in this pointer
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to print date/time in the logfile
	 *
	 *	@returns nothing. Either tags are replaced or not, it's not an error to not replace non existing tags.
	 *
	 *	@see replaceTagsInString()
	 */
	void replaceYearTags(std::string * p_string, unsigned long long p_timestamp = 0) const;

	/**
	 *	@fn void replaceShortYearTags(std::string * p_string, unsigned long long p_timestamp = 0)
	 *
	 *	@brief Replace any \<YY\> tags contained in a string with the last two digits of the current year
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_string Pointer to a string which tags shall be replaced.
	 *		Result will be stored in this pointer
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to print date/time in the logfile
	 *
	 *	@returns nothing. Either tags are replaced or not, it's not an error to not replace non existing tags.
	 *
	 *	@see replaceTagsInString()
	 */
	void replaceShortYearTags(std::string * p_string, unsigned long long p_timestamp = 0) const;

	/**
	 *	@fn void replaceMonthTags(std::string * p_string, unsigned long long p_timestamp = 0)
	 *
	 *	@brief Replace any \<MM\> tags contained in a string with the current month
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_string Pointer to a string which tags shall be replaced.
	 *		Result will be stored in this pointer
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to print date/time in the logfile
	 *
	 *	@returns nothing. Either tags are replaced or not, it's not an error to not replace non existing tags.
	 *
	 *	@see replaceTagsInString()
	 */
	void replaceMonthTags(std::string * p_string, unsigned long long p_timestamp = 0) const;

	/**
	 *	@fn void replaceDayTags(std::string * p_string, unsigned long long p_timestamp = 0)
	 *
	 *	@brief Replace any \<DD\> tags contained in a string with the current day
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_string Pointer to a string which tags shall be replaced.
	 *		Result will be stored in this pointer
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to print date/time in the logfile
	 *
	 *	@returns nothing. Either tags are replaced or not, it's not an error to not replace non existing tags.
	 *
	 *	@see replaceTagsInString()
	 */
	void replaceDayTags(std::string * p_string, unsigned long long p_timestamp = 0) const;

	/**
	 *	@fn void replaceHourTags(std::string * p_string, unsigned long long p_timestamp = 0)
	 *
	 *	@brief Replace any \<hh\> tags contained in a string with the current hour
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_string Pointer to a string which tags shall be replaced.
	 *		Result will be stored in this pointer
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to print date/time in the logfile
	 *
	 *	@returns nothing. Either tags are replaced or not, it's not an error to not replace non existing tags.
	 *
	 *	@see replaceTagsInString()
	 */
	void replaceHourTags(std::string * p_string, unsigned long long p_timestamp = 0) const;

	/**
	 *	@fn void replaceMinuteTags(std::string * p_string, unsigned long long p_timestamp = 0)
	 *
	 *	@brief Replace any \<mm\> tags contained in a string with the current minute
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_string Pointer to a string which tags shall be replaced.
	 *		Result will be stored in this pointer
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to print date/time in the logfile
	 *
	 *	@returns nothing. Either tags are replaced or not, it's not an error to not replace non existing tags.
	 *
	 *	@see replaceTagsInString()
	 */
	void replaceMinuteTags(std::string * p_string, unsigned long long p_timestamp = 0) const;

	/**
	 *	@fn void replaceSecondTags(std::string * p_string, unsigned long long p_timestamp = 0)
	 *
	 *	@brief Replace any \<ss\> tags contained in a string with the current second
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_string Pointer to a string which tags shall be replaced.
	 *		Result will be stored in this pointer
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to print date/time in the logfile
	 *
	 *	@returns nothing. Either tags are replaced or not, it's not an error to not replace non existing tags.
	 *
	 *	@see replaceTagsInString()
	 */
	void replaceSecondTags(std::string * p_string, unsigned long long p_timestamp = 0) const;

	/**
	 *	@fn void replaceMillisecondTags(std::string * p_string, unsigned long long p_timestamp = 0)
	 *
	 *	@brief Replace any \<s/10\> tags contained in a string with the first digit of the current millisecond
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_string Pointer to a string which tags shall be replaced.
	 *		Result will be stored in this pointer
	 *	@param p_timestamp If p_timestamp is == 0, the current system time is used, if it is > 0,
	 *		the time from from p_timestamp is used to print date/time in the logfile
	 *
	 *	@see replaceTagsInString()
	 *
	 *	@returns nothing. Either tags are replaced or not, it's not an error to not replace non existing tags.
	 *
	 *	e.g. if current milliseconds are 642, the \<s/10\> tag will be replaced with 6
	 */
	void replaceMillisecondTags(std::string * p_string, unsigned long long p_timestamp = 0) const;

	/**
	 *	@fn int replaceListOfVariablesTags(std::string * p_string)
	 *
	 *	@brief Replace the \<LIST_OF_LOGGED_VARIABLES\> tag with a delimiter seperated list of variable names
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_string Pointer to a string which tags shall be replaced.
	 *		Result will be stored in this pointer
	 *
	 *	@see replaceTagsInString()
	 */
	int replaceListOfVariablesTags(std::string * p_string) const;

	/**
	 *	@fn int replaceVariableNameTagsWithVariableValues(std::string * p_string)
	 *
	 *	@brief Replace any tag that contains a variable's name with it's value
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_string Pointer to a string which tags shall be replaced.
	 *		Result will be stored in this pointer
	 *
	 *	@returns WE_SUCCESS if all existing tags were replaced with varaible names
	 *
	 *	@see replaceTagsInString()
	 *
	 *	@warning Only call this function if all other tags already has been replaced!
	 *
	 *	This function checks every tag if it contains a valid variable name (variable needs to be
	 *	available in WEDEI) and then replaces the name tag with the variable's value.
	 */
	int replaceVariableNameTagsWithVariableValues(std::string * p_string) const;

	MMIntToULong eventVariables{}; //!< Map containing all variables configured for event based logging, key: eventType, mapped value index/subindex of varaible
	MMIntToULong eventVariablesByLogLevel{}; //!< Map containing all variables configured for event based logging sorted by log level
	MULongToIntP integerVariables{}; //!< Map containing all integer variable keys and a pointer to their value
	MULongToStringP stringVariables{}; //!< Map containing all string variable keys and a pointer to their value
	//--------------------------
	std::map<unsigned int, WEVariableStorageType> storageTypeByCombinedKey{};
	std::vector<unsigned int> variablesInOrderOfAppearance{};

	bool timeBased{false}; //!< True if time based logging is configured, false if event based logging is configured
	bool continuousLogging{true}; //!< True if only one log file shall be used, false if at every restart a new log file shall be generated
	std::string logFileName{}; //!< Name of the log file, tags not replaced
	std::string currentLogFileName{}; //!< Current name of the log file with replaced tags
	std::string logPath{}; //!< The filesystem path where the log file is located
	std::string logOptPath{}; //!<
	std::string originalFileHeader{}; //!< The log file header text, tags not replaced
	std::string originalLinePrefix{}; //!< The current log file header text with replaced tags
	bool logEventName{true};
	bool logVariableName{true};
	int maxNumberOfFiles{5};

	unsigned int timerValue{0}; //!< The timer for time based logging
	unsigned int timerOne{0}; //!< First timer for event based logging
	unsigned int timerTwo{0}; //!< Second timer for event based logging
	unsigned int timerThree{0}; //!< Third timer for event based logging
	unsigned int oldTickCount{0}; //!< The tick count at the time when the time based timer has expired the last time
	unsigned int oldTickCountOne{0}; //!< The tick count at the time when the first event based timer has expired the last time
	unsigned int oldTickCountTwo{0}; //!< The tick count at the time when the second event based timer has expired the last time
	unsigned int oldTickCountThree{0}; //!< The tick count at the time when the third event based timer has expired the last time

	char delimiter{';'}; //!< The configured delimiter with which all entries in the log file are separated
	bool logging{false}; //!< True if logging is in progress, false when logging is stopped
	int logLevel{0}; //!< The current log level (MIN_LOG_LEVEL <= logLevel <= MAX_LOG_LEVEL)
	bool firstStart{true}; //!< Indicates if the start function is called the first time

	WEDEI * myDEI{nullptr}; //!< Pointer to the WEDEI instance


	const std::string yearTag{"<YYYY>"}; //!< Definition of the year tag, initialized in constructor
	const std::string yearTagShort{"<YY>"}; //!< Definition of the short year tag, initialized in constructor
	const std::string monthTag{"<MM>"}; //!< Definition of the month tag, initialized in constructor
	const std::string dayTag{"<DD>"}; //!< Definition of the day tag, initialized in constructor
	const std::string hourTag{"<hh>"}; //!< Definition of the hour tag, initialized in constructor
	const std::string minuteTag{"<mm>"}; //!< Definition of the minute tag, initialized in constructor
	const std::string secondsTag{"<ss>"}; //!< Definition of the seconds tag, initialized in constructor
	const std::string millisecondTag{"<s/10>"}; //!< Definition of the milliseconds tag, initialized in constructor
	const std::string variableListTag{"<LIST_OF_LOGGED_VARIABLES>"}; //!< Definition of the logged variable name list tag, initialized in constructor

	std::ofstream logFile{}; //!< The log file! (or better the stream to the log file)

	static const unsigned int MINIMUM_LOG_INTERVAL{500}; //!< Minimum log interval in milliseconds

	static const int MAX_LOG_LEVEL{10}; //!< Definition of the maximum log level
	static const int MIN_LOG_LEVEL{0}; //!< Definition of the minimum log level

	static int debugSection;
};
} // namespace WE

#endif /* WE_VARIABLE_LOGGER_H_ */
