/**
 *	@file we_action_queue.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Definition of WEActionQueue class
 *
 *	@see WEAction
 */

#ifndef WE_ACTION_QUEUE_H_
#define WE_ACTION_QUEUE_H_

#include <set>
#include <thread> //for thread ID

#include "we_event.h"	// need the event type as trigger information
#include "../../utility/thread_safe_queue/include/thread_safe_queue.h"

namespace WE
{

class WEAction;



/**
 *	@class WEActionQueue
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief A queue to place WEActions in and execute them.
 *
 *	@namespace WE
 *
 *	@see WEAction
 *
 *	This class is responsible to store AppActions (and their corresponding DDOs)
 *	and execute them all with executeActions (until queue is empty or timeout expires)
 *	WEActionQueue is implemented as a singleton. Use the instance() method to return the
 *	instance of this class.
 *
 *	We could consider to have one action queue per ISO-VT pool. Then we could detect loops on a 'per pool' basis.
 *
 */
class WEActionQueue
{
public:
	/**
	 *	@fn static WEActionQueue * instance(void)
	 *
	 *	@brief Return the instance of WEActionQueue
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns the one and only instance of this class
	 */
	static WEActionQueue * instance(void);


	/**
	 *	@brief put a WEAction into the queue
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_action Action that shall be executed
	 *	@param p_event Event which triggered the action, nullptr if loop detection is not required
	 *	@param p_origin Origin of the action, nullptr if loop detection is not required
	 *
	 *	@returns WE::WE_SUCCESS;
	 *
	 *	@see int executeActions(unsigned int p_timeout, bool p_breakLoop = false)
	 *
	 */
	int postAction(std::shared_ptr<WE::WEAction> p_action, WEEvent * p_event = nullptr , void * p_origin = nullptr);


	/**
	 *	@fn int executeActions(unsigned int p_timeout, bool p_breakLoop = false)
	 *
	 *	@brief Execute all actions in the queue
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timeout time in milliseconds when the function shall return
	 *	@param p_breakLoop Set to true if execution should terminate if an action is triggered again by the same action
	 *
	 *	@returns WE::WE_SUCCESS if at least one action was found
	 *	@returns WE::WE_NOT_FOUND_ERROR if no action was found in queue
	 *	@returns WE::WE_CALL_NOT_COMPLETED in case p_timeout is reached before queue is empty
	 *	@returns WE::WE_ALREADY_EXISTS_ERROR in case a loop was terminated
	 *
	 *	This function returns if the queue is empty or if timeout expires.
	 *
	 *	It also performs loop detection triggered by the following condition:
	 *	a) The same start (top level) action is still valid.
	 *	b) An action is triggered is triggered a second time.
	 *	c) The same object (DDO) triggered the action.
	 *	d) The same event triggered the action.
	 *	e) The counter of the old trigger found has to be lower than in the suspicious action (otherwise it's just the next command of the same macro)
	 *
	 *	To make this work all the parameters have to be provided in "postAction()".
	 *
	 *	If a loop is found then only the one action where we detect the loop is skipped.
	 *	This makes terminating the loop longer but it should at least keep valid actions from other sources.
	 *	Since action queue is designed as a singleton e.g. for ISO-VT there is only one action queue for all pools.
	 *	So this approach tries to keep all valid actions while breaking the loop multiple times for the bad source.
	 *
	 *	@see int postAction(std::shared_ptr<WE::WEAction> p_action, WEEvent * p_event = nullptr , void * p_origin = nullptr)
	 *
	 *	@warning For loop detection to work properly this method always has to be called from the same thread.
	 */
	int executeActions(unsigned int p_timeout, bool p_breakLoop = false);


	/**
	 *	@fn ~WEActionQueue()
	 *
	 *	@brief Destructor. Empties queue.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WEActionQueue();


private:

	/**
	 *	@fn WEActionQueue()
	 *
	 *	@brief Private constructor => Singleton
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WEActionQueue();


	/**
	 *	@fn WEActionQueue& operator=(WEActionQueue const&)
	 *
	 *	@brief Private assignment operator. No copying allowed (singleton)
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	WEActionQueue& operator=(WEActionQueue const&) = delete; //no assignment allowed

	typedef struct ActionItem
	{
		std::shared_ptr<WE::WEAction> execAction{};
		std::shared_ptr<WE::WEAction> triggerAction{};

		// WEEvent triggerEvent;	// Seems VT aux events were the only ones using the full event
		WEEventType triggerEvent{UNKNOWN_EVENT};	// Everywhere else it was either ignored completely or the type was sufficient.
		void * origin{nullptr};
		unsigned int counter{0};	// only required to distinguish between a simple loop (macro directly triggering itself)
					// and a macro containing the same command more than once
					// therefore value is NOT compared but only used to decide when history is updated or loop is detected
		std::thread::id threadID{};	// loop is only valid if new macro/action is posted from same thread as "WEActionQueue::executeActions()"
					// is running in - otherwise encoder move events could trigger loop detection when turning encoder fast enough
					// because execution of action was running while new event arrived.

		// we have to check if the same item is in our list already - not really used anymore
		bool operator==(const ActionItem &p_c) const
		{
			// std::cout << "operator==(const actionItem" << std::endl;
			return	   (p_c.execAction	== execAction)
				&& (p_c.triggerAction	== triggerAction)
				&& (UNKNOWN_EVENT	!= triggerEvent)
				&& (p_c.triggerEvent	== triggerEvent)
				&& (nullptr		!= origin)
				&& (p_c.origin		== origin);
		}

		// std::set needs a "less" operator to sort items:
		bool operator<(const ActionItem &p_c) const
		{
			// std::cout << "operator<(const actionItem" << std::endl;
			// When 'comparing' pointers 'unsigned long' has the right size on all (our) platforms
			// whereas 'unsigned int' is 32 bit on all of them (including 64 bit Linux PC)
			return	   ((unsigned long)execAction.get()	< (unsigned long)p_c.execAction.get()	)
				|| (  ((unsigned long)execAction.get()   == (unsigned long)p_c.execAction.get())
				   && ((unsigned long)triggerAction.get() < (unsigned long)p_c.triggerAction.get()) )
				|| (  ((unsigned long)execAction.get()    == (unsigned long)p_c.execAction.get())
				   && ((unsigned long)triggerAction.get() == (unsigned long)p_c.triggerAction.get())
				   && ((unsigned int)triggerEvent   < (unsigned int)p_c.triggerEvent) )
				|| (  ((unsigned long)execAction.get()    == (unsigned long)p_c.execAction.get())
				   && ((unsigned long)triggerAction.get() == (unsigned long)p_c.triggerAction.get())
				   && ((unsigned int)triggerEvent   == (unsigned int)p_c.triggerEvent)
				   && ((unsigned long)origin         < (unsigned long)p_c.origin) );
		}
	} ActionQueueItem;

	WE::WEThreadSafeQueue< ActionQueueItem > actions; //!< The actual queue that stores all actions

	std::set < ActionQueueItem > triggerHistory;	//!< Remember active action and the one triggering it. Used to break infinite loops.
	std::shared_ptr<WE::WEAction> activeAction;			//!< Action currently executed (to be used as trigger if another action is enqueued)
	std::shared_ptr<WE::WEAction> activeTriggerAction;			//!< Action which triggered the currently active actions - when this one changes we have update triggerHistory
	std::set < ActionQueueItem > triggerNow;	//!< Remember active action and the one triggering it. Used to update triggerHistory when activeTriggerAction changes.
	unsigned int triggerCounter;			//!< Criteria to know a new active trigger action has started even if it should be the same as the old
};
} // namespace WEApp

#endif /* WE_ACTION_QUEUE_H_ */
