/**
 *	@file we_event_queue_as_queue.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 */


#ifndef WE_EVENT_QUEUE_AS_QUEUE_H_
#define WE_EVENT_QUEUE_AS_QUEUE_H_

#include "../../utility/thread_safe_queue/include/thread_safe_queue.h"
#include "we_event_queue_base.h"
#include "we_some_event.h"


namespace WE
{

/**
 *	@class WEEventQueueAsQueue
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd. Faisal
 *
 *	@brief The event queue to store the events in the order of arrival and fetch the events in
 *		oldest first manner.
 *
 *	@namespace WE
 *
*/
class WEEventQueueAsQueue : public WEEventQueue
{
public:
	/**
	 *	@fn WEEventQueueAsQueue()
	 *
	 *	@brief Default constructor to create a WEEventQueue instance.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	*/
	WEEventQueueAsQueue() = default;

	/**
	 *	@fn virtual ~WEEventQueueAsQueue()
	 *
	 *	@brief standard desctructor for a WEEventQueue instance.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	*/
	virtual ~WEEventQueueAsQueue() = default;

	/**
	 *	@brief Add the specified event to the end of the queue.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@param[in] p_event Event to be posted to the event queue.
	 *
	*/
	void postEvent(WEEvent & p_event) override;

	/**
	 *	@brief Retrieve the oldest event from the queue.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns Event at the front of the queue.
	 *	@returns A bogus instance if the queue is empty
	 *
	 *	Also removes the event from the queue.
	*/
	WESomeEvent getNextEvent() override;

	/**
	 *	@fn getSize()
	 *
	 *	@brief Retrieve the no of events currently present in the queue .
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns The no of events in the queue.
	 *	@returns Zero if the queue is empty.
	 *
	*/
	int getSize() override;

	/**
	 *	@fn isEmpty()
	 *
	 *	@brief Check whether the queue is empty or not.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns TRUE if the queue is empty, FALSE otherwise.
	*/
	bool isEmpty() override;

private:
	WEThreadSafeQueue< WESomeEvent > queue{};
};
}


#endif /* WE_EVENT_QUEUE_AS_QUEUE_H_ */
