/**
 *	@file we_event_queue_base.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief General interface for a event queue class
 *
 *	@see WEEventQueueAsPipe and WEEventQueueAsQueue
 *
 *	DESCRIPTION
 */

#ifndef WE_EVENT_QUEUE_BASE_H_
#define WE_EVENT_QUEUE_BASE_H_

#include <memory> //unique_ptr

namespace WE
{
struct WEEvent;
struct WESomeEvent;

typedef enum
{
	EVENT_QUEUE_TYPE_QUEUE	//The type for an event queue that uses a WEThreadSafeQueue
}eventQueueType;


class WEEventQueue
{
public:
	/**
	 * Function to create a instance of WEEventQueue.
	 * @param p_type Type of event queue that shall be created,
	 * 	either one that uses a pipe (which provides also the interfaces from WEPipie e.g. endpoints)
	 * 	or one that uses a WEThreadSafeQueue.
	 * @return Either an instance of WEEventQueueAsPipe, WEEventQueueAsQueue or nullptr if p_type is unknown.
	 */
	static std::unique_ptr<WEEventQueue> create(eventQueueType p_type);

	virtual ~WEEventQueue() = default;

	/**
	 *	@brief Add the specified event to the end of the queue.
	 *
	 *	@param[in] event Event to be posted to the event queue.
	 *
	 *	Internally queue will copy 'event' into a WESomeEvent which
	 *	is big enough to carry any event as it inherits them all.
	 */
	virtual void postEvent(WEEvent & p_event) = 0;

	/**
	 *	@brief Retrieve the oldest event from the queue.
	 *
	 *	@returns Event at the front of the queue.
	 *	@returns Event with unknown event type if queue is empty
	 *
	 *	Also removes the event from the queue.
	 *	Make sure to cast the returned element to the (base) type
	 *	for the queue you are using.
	 */
	virtual WESomeEvent getNextEvent() = 0;

	/**
	 *	@fn getSize()
	 *
	 *	@brief Retrieve the no of events currently present in the queue .
	 *
	 *	@returns The no of events in the queue.
	 *	@returns Zero if the queue is empty.
	 */
	virtual int getSize() = 0;

	/**
	 *	@fn isEmpty()
	 *
	 *	@brief Check whether the queue is empty or not.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns TRUE if the queue is empty, FALSE otherwise.
	 */
	virtual bool isEmpty() = 0;

	/**
	 *	@fn void setAfterPostFunction(void (*functionPointer)(void))
	 *
	 *	@brief Set function pointer to be called after posting event to queue
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param functionPointer Pointer to a function that shall be called after an event was posted to the event queue
	 *
	 *	With this function a function pointer can be set that is called directly
	 *	after an event was posted to the event queue.
	 */
	void setAfterPostFunction(void (*p_functionPointer)(void));

protected:
	WEEventQueue();
	void (*postFunctionPointer)(void);	//!< Function pointer to be called directly after an event was posted to event queue
	static const int eventQueueDebugKey;
};
}


#endif /* WE_EVENT_QUEUE_BASE_H_ */
