/**
 *	@file we_input_event_manager.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2012
 *
 *	@brief Handler and manager of keyboard, encoder and mouse events.
 *
 *	This class is implemented as singleton. All Key, encoder and mouse events are handled by this class.
*/


#ifndef WE_INPUT_EVENT_MANAGER_H_
#define WE_INPUT_EVENT_MANAGER_H_

#include <array>
#include <map>
#include <mutex>

#include "../../os_abstraction/thread/include/we_thread.h"
#include "../../os_abstraction/endpoint_manager/include/we_endpoint_mgr.h"
#include "we_can_remote_handler.h"
#include "we_event_queue_base.h"
#include "we_input_event.h"

namespace WE
{

class WECANRaw;
class WEEncoderDynamic;
struct WEToolkitEvent;

/**
 * Struct containing the informations about the configuration of one key for the CAN remote
 */
struct KeyOptionsCANRemote
{
	int keyNumber{-1};	//!< Key number (0 to x) where 0 is always the encoder
	int keyCode{-1};	//!< WE key code (like 0 for ESC, 1 for HOME, etc.)
	int byte{-1};	//!< The byte in the CAN message where the status of that key can be found (start counting with 0)
	int bit{-1};	//!< The bit inside the byte that represents press or release of the key (start counting with 0)
	int CANID{-1};	//!< The CAN ID of the CAN message that contains this key state
	int oldKeyState{-1};	//!< The last key state that was previously received
};

struct CANRemoteOptions
{
	bool enabled{false};	//!< CAN remote is enabled
	int encoderUp{-1};	//!< WE keycode for encoder turn up
	int encoderDown{-1};	//!< WE keycode for encoder turn down
	std::multimap<int, KeyOptionsCANRemote*> keyOptionsByCANID{}; //!< mapping between CAN ID and the respective key options
	std::array<KeyOptionsCANRemote, 20> keyOptions{}; //all keys
	unsigned int CANID{0};	//!< CAN ID for encoder turns only (keys might be different)
	unsigned char encoderLSB{0};	//!< Least significant byte of the encoder turn counter inside the CAN message
	unsigned char canPort{0};	//!< CAN port to use
	int encoderMask{-1};	//!< Read encoder count is masked with this. It also represents the maximum of the encoder counter
	bool encoderLittleEndian{true};	//!< Is the encoder counter transmitted as little or big endian?
};

/**
 *	@class WEInputEventManager
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief The class has to be called for handling all key, encoder and mouse events.
 *
 *	@namespace WE
 *
 *	This class is implemented as a singleton, to have a single entry point for all
 *	key, encoder and mouse events.
 *	It further provides a special thread taking care about creating events for double
 *	press and long press. If the used toolkit-framework supports these kind of events,
 *	you can move the handling of double press and long press to the framework itself.
 */
class WEInputEventManager
{
	friend class CANRemoteHandler;

public:
	static const int DEFAULT_DOUBLE_KEY_PRESS_DURATION = 300;	//!< Default Double Key press duration. Half a second was too long.
									// 200ms are fine for a mouse but a little too short on terminal.
	static const int DEFAULT_LONG_KEY_PRESS_DURATION = 2000;	//!< Default Long Key press duration. Two seconds.


	/**
	 *	@fn ~WEInputEventManager()
	 *
	 *	@brief Destructor of WEInputEventManager()
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	virtual ~WEInputEventManager();

	/**
	 *	@fn static WEInputEventManager* getInstance(void)
	 *
	 *	@brief static method which returns the pointer to the current instance of this class
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns Pointer to the existing instance, otherwise nullptr.
	 *
	 *	This function returns a pointer to the current available instance (singleton) of this class. It does not
	 *	create a new instance when no instance is available. When using carefully (i.e. being sure that
	 *	an instance exists, this function can be used for doing single line function calls to the instance.
	 */
	static WEInputEventManager* getInstance(void);

	/**
	 *	@fn int setDoubleKeyPressDuration(unsigned int p_milliSeconds)
	 *
	 *	@brief This function is used to change the double key press duration
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_milliSeconds The maximum time in milliseconds between two key press events that is allowed for a double key press.
	 *
	 *	@returns WE_SUCCESS if setting double key press duration to p_milliseconds was successful.
	 *	@returns WE_INVALID_PARAMENTER_ERROR if p_milliseconds is greater or equal the long key press duration.
	 *			In this case the double key press duration is not changed.
	 */
	int setDoubleKeyPressDuration(unsigned int p_milliSeconds);

	/**
	 *	@fn int setLongKeyPressDuration(unsigned int p_milliSeconds)
	 *
	 *	@brief This function is used to change the long key press duration
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_milliSeconds The minimum time in milliseconds to not receive a key release event after key press event to detect the key press event as long press event.
	 *
	 *	@returns WE_SUCCESS if setting long key press duration to p_milliseconds was successful.
	 *	@returns WE_INVALID_PARAMENTER_ERROR if p_milliseconds is smaller or equal the double key press duration.
	 *			In this case the long key press duration is not changed.
	 */
	int setLongKeyPressDuration(unsigned int p_milliSeconds);

	/**
	 *	@fn void setKeyboardLayout(int p_layout)
	 *
	 *	@brief This function is used to change the keyboard layout
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_layout The keyboard layout to use. Use KEYS_OPUS_44_1 to enable opus44 key layout.
	 *
	 *	This function is used to switch between the standard opus keyboad layout and special layouts
	 *	like the CAN-connected opus44 keyboard.
	 */
	void setKeyboardLayout(int p_layout);

	/**
	 *	@fn WE::WEEventQueue* getInputEventsQueue()
	 *
	 *	@brief This function returns the input event queue.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return Pointer to the input event queue.
	 *
	 *	This function delivers a pointer to the input event queue. This queue needs to be processed by the
	 *	event handler of the application. All key, encoder and mouse events will be send to the application
	 *	via this queue.
	 */
	WE::WEEventQueue* getInputEventsQueue();

	/**
	 *	@brief This function delivers the next available event from the input event queue
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return Removes and delivers the next event from the input event queue. If no event is
	 *		available WEInputEvent::eventType is set to UNKNOWN_EVENT. All other parameters are set to 0;
	 */
	WEInputEvent getNextEvent();

	/**
	 *	@brief This function checks if an event can be read from the input event queue.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return true if there are elements in the input event queue.
	 *	@return false if there are no elements in the input event queue.
	 */
	bool hasEvents();

	/**
	 *	@brief Function for setting special opus44 options.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_options The CANRemoteOptions options to set.
	 */
	void setCANRemoteOptions(struct CANRemoteOptions p_options);


	int setKeyMapping(std::map<int, int>  p_osCodeToWeCodeMap);

	/**
	 *	@brief Function that handles key, encoder and mouse events.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_inputEvent The key, encoder or mouse event,
	 *
	 *	@return WE_SUCCESS if a valid event was received
	 *	@return WE_INVALID_PARAMETER_ERROR if an unknown event was provided.
	 *
	 *	This function has to be called for every incoming key, encoder or mouse event. It
	 *	will convert the incoming key event into a WEInputEvent of the 'correct' type and
	 *	send it to the application via input event queue.
	 *
	 *	Currently this is typically called directly from the toolkit widget receiving the event.
	 *	Therefore toolkit widget directly uses the fact that this class is a singleton.
	 *
	 *	@WARNING p_inputEvent is modified by this function. We do NOT make a copy of this parameter.
	 *		This is done for performance improvement.
	 *
	 *	Not sure if it helps to create encoder events or if we should just distinguish them
	 *	via the key codes (at least for encoder press/release).
	 */
	virtual int inputEventOccurred(WEInputEvent & p_inputEvent);


	/**
	 *	@brief Receive raw events, convert them to WE events and call 'inputEventOccurred()'..
	 *
	 *	@see int inputEventOccurred(WEInputEvent &)
	 *
	 *	Since WEToolkitEvent already is a WEInputEvent, we only modify a few parameters
	 *	before calling inputEventOccurred() with the same event.
	 *	I.e. we mainly convert the event type and add a generic WE keycode.
	 */
	virtual int toolkitEventOccurred(WE::WEToolkitEvent & p_event);



	/**
	 *	@fn int postESCtoEventQueue(bool p_press)
	 *
	 *	@brief Allow creation of ESC events e.g. from mouse events.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_press True: press event, false: release event is posted to queue
	 *
	 *	@returns An error code - WE::WE_SUCCESS in case of success.
	 *
	 *
	 */
	int postESCtoEventQueue(bool p_press);


	static std::map<int, int> * getWEtoOSCodeMap(void);


	/**
	 *	Function to generate a simulated QKeyEvent.
	 *	This generated event should just work like a real, physical key event.
	 *	@param p_event A WEInputEvent that is a key or encoder event
	 *
	 *	E.g. used by CANRemoteHandler to convert CAN messages into events.
	 */
	static void generateKeyEvent(WEInputEvent p_event);


	/**
	 *	@fn int acceptInput(bool p_accept)
	 *
	 *	@brief Allow blocking input event acceptance.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_accept true: Accept user input, false: ignore user input
	 *
	 *	@returns An error code - WE::WE_SUCCESS in case of success.
	 *
	 *
	 */
	int acceptInput(bool p_accept);

	/**
	 * If this->possibleLongPressEvent was caused by a mouse event,
	 * calling this function will reset the timestamp of this event to 0
	 * so no long press will be generated (handy when eg. changing a page 
	 * via onPress and the widget that caused the press will not be available
	 * on the new page)
	 */
	void resetPossibleLongMousePressTimer(void);

private:
	/**
	 *	@fn WEInputEventManager()
	 *
	 *	@brief Default constructor
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	WEInputEventManager();

	/**
	 *	@fn int initialize()
	 *
	 *	@brief This function is called by the constructor to initialize the object.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	int initialize();

	/**
	 *	@fn int deinitialize()
	 *
	 *	@brief This function is called by the destructor to deinitialize the object.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	int deinitialize();

	int initializeOSCodeToWECodeMap();

	/**
	 *	@brief Converts toolkit event type into WEInputEvent type.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_event[in][out] The toolkit event that will be covnerted into a(nother) WEInputEvent
	 *
	 *	@return WE_SUCCESS if p_event is converted into a(nother) WEInputEvent
	 *	@return WE_INVALID_PARAMETER if p_event can not be converted
	 *	@return
	 *
	 *	@see WETookitEvent
	 *	@see WEInputEvent
	 */
	int converToolkitToWEEvent(WEInputEvent & p_event);

	/**
	 *	@brief If the given key event belongs to the encoder, this function converts the key event into an encoder event.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_event The key event that might be converted into an encoder event.
	 *
	 *	@return WE_SUCCESS when conversion was successful and the event has to be send to event queue.
	 *	@return WE_NOT_SUPPORTED_ERROR if the event should not be posted to event queue.
	 */
	int convertToEncoderEvents(WEInputEvent & p_event);

	/**
	 *	@brief Accumulate several key encoder event to one single event
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_event The event that might be accumulated
	 *
	 *	@return WE_SUCCESS when the event has to be send to event queue.
	 *	@return WE_NOT_SUPPORTED_ERROR if the event should not be posted to event queue.
	 *
	 *	@NOTE Part of the job of accumulating events is handled by the input event manager thread.
	 */
	int bundleKeyEncoderEvents(WEInputEvent & p_event);

	/**
	 *	@brief Converts a given event into a double press .
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_event The key event that might be converted into an encoder event.
	 *
	 *	@return WE_SUCCESS
	 */
	int convertToDoublePressEvent(WEInputEvent & p_event);

	/**
	 *	@brief Remembers key press events for detection of long press events
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_event The key event to remember for long press detection.
	 *
	 *	@return WE_SUCCESS
	 */
	int markPossibleLongPress(const WEInputEvent & p_event);

	/**
	 *	@brief This function is used to push events to the input event queue
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_event The event to push to the queue.
	 *
	 *	@return WE_SUCCESS
	 */
	int postToEventQueue(WEEvent & p_event);

	/**
	 *	@brief Checks if a long press has occurred. If so, a long press event is sent.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_currentTick The current WETick
	 *
	 *	@return WE_SUCCESS
	 */
	int sendLongPressEvent(unsigned int p_currentTick);

	/**
	 *	@fn static int sendEncoderMoveEvent(unsigned int p_currentTick)
	 *
	 *	@brief Checks if an encoder move event has to be sent. If so, an encoder move event is sent.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_currentTick The current WETick
	 *
	 *	@return WE_SUCCESS
	 */
	int sendEncoderMoveEvent(unsigned int p_currentTick);

	/**
	 *	@brief This function is executed by the input event manager thread.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_args The argument contains the pointer to the WEInputEventManager instance.
	 *
	 *	@return WE_SUCCESS
	 */
	static void inputEventManagerThreadFunc(void * p_args);

	static std::map< int, int > weCodeToOSCodeMap;


	bool acceptInputState;				//<! Define if user input is currently accepted or not.

	std::map<int, int> osCodeToWECodeMap;

	std::unique_ptr<WEEventQueue> inputEventsQueue{};	//!< Queue holds all key, encoder and mouse events that need to be processed by the application
	WEEndPointManager key_encoder_endpoints_mgr_OP44{};	//!< End point manager for handling CAN-messages from opus44 keyboard.
	unsigned int doubleKeyPressDurationMilliSec;	//!< The time to detect double key presses.
	unsigned int longKeyPressDurationMilliSec;	//!< The time to detect long key presses.

	WEThread inputEventManagerThread;		//!< Pointer to the input event manager thread for handling opus44 endpoint and detection of long and double key presses.
	bool terminateManagerThread;			//!< terminateManagerThread is set by deinitialize to indicate that the input event manager thread should die

	int keyboardLayout;				//!< Hold the used keyboard layout (e.g. opus44 keyboard)
	std::unique_ptr<CANRemoteHandler> opus44Handler;		//!< Pointer to the event handler for opus44 events.
	std::shared_ptr<WECANRaw> canPort;				//!< We need a can port for receiving opus44 key and encoder messages
	CANRemoteOptions canRemoteOptions;		//!< special options for opus44 keyboard.
	WEEncoderDynamic * encoderDynamic;		//!< Pointer to the class that computes the encoder dynamics

	WEInputEvent possibleDoublePressEvent;		//!< holds the previous event for double press and long press detection.
	WEInputEvent possibleLongPressEvent;		//!< holds the previous event for double press and long press detection.
	WEInputEvent bundleEncoderMoveEvent;		//!< holds the previous key-encoder event for accumulating the steps like it is available for mouse wheels
	std::mutex mutex;				//!< mutex to coordinate access align sending events to event handler from event thread and input event manager thread
};
}
#endif /* WE_INPUT_EVENT_MANAGER_H_ */
