/*
 *	we_signal.h
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2020
 *
 *	@brief Based on WETrigger and intended for hardware, OS and CAN stack events.
 *
 *	Should be handled in same context as WEIncident but originate 'outside'.
 *	Inherits 'WETrigger' so that it can be forwarded to the GUI (and terminate there).
 */

#ifndef SRC_EVENT_HANDLING_INCLUDE_WE_SIGNAL_H_
#define SRC_EVENT_HANDLING_INCLUDE_WE_SIGNAL_H_

#include "we_trigger.h" //contains <cstring> for memset, memcpy

namespace WE
{

// TIMER_TIMEOUT_EVENT
//struct timerTimeoutParameter
//{
//	// use WEEvent::value: std::int32_t type;	//!< the type of the timeout (single shot/repeating timer)
//};

// HEARTBEAT_TIMEOUT_EVENT
//struct heartbeatTimeoutParameter
//{
//	// use WEEvent::value: std::int32_t result;		//!< the result of the timeout
//	// use WEEvent::uvalue: std::int32_t messageId;	//!< the message ID that belongs to the timeout
//};

// STATE_CHANGED_EVENT
//struct stateChangeParameter
//{
//	// use WEEvent::value: std::int32_t state;		//!< the state
//	// use WEEvent::uvalue: unsigned char nodeID;	//!< the node id
//};

// CAN_MAPPING_TIMEOUT_EVENT
//struct CANMappingTimeoutParameter
//{
//	// use WEEvent::value: std::int32_t variableID;	//!< variable id
//	// use WEEvent::object: void *mappingObject;	//!< the mapping object
//	// use WEEvent:uvalue: bool engaged;		//!< true: timeout active, false: timeout resolved
//};

// J1939_TP_MESSAGE_SEND
struct J1939TpParameter
{
	// use WEEvent::object: void * ecu;	//!< ECU this TP comes from (or goes to)
	// use WEEvent::uvalue: std::uint32_t pgn;	//!< the J1939 PGN of the transport protocol message
	// use WEEvent::value:  std::int32_t result;	//!< the result of sending the transport protocol message
	unsigned long long dataBytes; //!< The 8 data bytes of a CAN message as 64Bit value
};


// POWER_MANAGEMENT_TIMEOUT_EVENT
// received in slow thread - without any queue
//struct powerManagementTimeoutParameter
//{
//	// use WEEvent::value: std::int32_t state;	//!< the power management state
//	// use WEEvent::uvalue: std::int32_t timeout;	//!< the timeout value
//};

// POWER_MANAGEMENT_DEVICE_STATE_EVENT
//struct powerManagementDeviceStateParameter
//{
//	// use WEEvent::uvalue: std::int32_t device;	//!< the hardware device id
//	// use WEEvent::value: std::int32_t state;	//!< the state of the hardware device
//};

// POWER_MANAGEMENT_POWER_STATE_EVENT
//	we_power_management_interface_a3.cpp
//struct powerManagementPowerStateParameter
//{
//	// use WEEvent::value: std::int32_t state;	//!< the state of the hardware device
//};

// POWER_MANAGEMENT_POWER_STATE_CHANGE_EVENT
//struct powerManagementPowerStateChangeParameter
//{
//	// use WEEvent::value: std::int32_t state;	//!< the state of the hardware device
//};

// IO_TOGGLE_EVENT
//struct ioToggleParameter
//{
//	// use WEEvent::value: std::int32_t value;	//!< the value of the hardware device
//	// use WEEvent::uvalue: std::int32_t device;	//!< the hardware device id
//};


// MEDIA_EVENT
struct MediaEventParameter
{
	// use WEEvent::pointer: std::string * filePath;
	unsigned int totalTimeInSeconds;
	unsigned int currentTimeInSeconds;
	// use WEEvent::value: int state;
	// use WEEvent::uvalue: int oldState;
	// use WEEvent::object: void * listener;
	bool metadataChanged;
};


// AppButtonDDO::incidentOccurred() creates and directly executes action
// int AppIntDDO::setDDOProperty()  creates and directly executes action
// AppIntDDO::incidentOccurred() creates and directly executes action
// PClient::initInitializeAlarms() creates and directly executes action to initialize language and theme
// otherwise sent to variable event dispatcher read out in slow thread --> WETrigger
// Meanwhile it's a WESignal, changing project/object pool first and then possibly GUI as WETrigger
// VARIABLE_VALUE_CHANGED_EVENT_INT
// VARIABLE_VALUE_CHANGED_EVENT_STRING
// VARIABLE_VALIDITY_CHANGED (uses same event as the two above)

// For INT only created by WESDOHandler who directly calls the listeners.
// Nobody seems to create it for STRING.
// Consumers consider it the same as VARIABLE_VALUE_CHANGED_EVENT_INT/VARIABLE_VALUE_CHANGED_EVENT_STRING
// VARIABLE_GET_VALUE_RESPONSE_EVENT_INT
// VARIABLE_GET_VALUE_RESPONSE_EVENT_STRING
struct VariableParameter
{
	//union
	//{
		// use WEEvent::value: std::int32_t integer;		//!< integer variable value
		// use WEEvent::pointer: std::string *string;	//!< string pointer variable value
	//} value;			//!< the value of the variable
	// use WEEvent::uvalue: unsigned long combinedKey;	//!< variable combined index value
	bool isValid;			//!< valid flag of variable
	// unused/not evaluated: bool isString;			//!< true if the value is a string value
	// unused/not evaluated: bool isInt;			//!< true if the value is an integer value
	// use WEEvent::object: void * object;			//!< used by ISO-VT only
	unsigned long long timestamp64;	//!< timestamp as local time in milliseconds (not tick count!)
};




union signalParameter
{
	struct J1939TpParameter j1939TP;
	struct MediaEventParameter mediaEvent;
	struct VariableParameter variable;
};



struct WESignal:virtual WETrigger
{

	union signalParameter signalPara;

	/**
	 *	@brief Constructor directly defining the event type - UNKNOWN_EVENT if none provided
	 *
	 *	@param p_type The type of event to be created
	 *	@param p_class Keep default - is only modified by classes inheriting this one
	 *
	 *	@warning Logically p_type should clearly define what event class to generate but nothing is enforced there.
	 *
	 *	Marked as 'explicit' i.e. not allowed for implicit casting.
	 */
	explicit WESignal(WEEventType p_type = UNKNOWN_EVENT, WEEventClass p_class = WE_SIGNAL)
	:WEEvent(p_type, p_class),	// otherwise the empty constructor is called
	 WETrigger(p_type, p_class)
	{
		memset(&(this->signalPara), 0, sizeof(this->signalPara));
	}

	virtual ~WESignal()
	{
		// nothing to clean up
	}


	inline void setHeartbeatTimeoutParameter(std::int32_t p_result, std::int32_t p_messageId)
	{
		this->value  = p_result;
		this->uvalue = p_messageId;
	}

	inline void setCANMappingTimeoutParameter(std::int32_t p_variableID, void * const p_mappingObject, bool p_engaged)
	{
		this->value  = p_variableID;		//!< variable id
		this->object = p_mappingObject;		//!< the mapping object
		this->uvalue = p_engaged;		//!< true: timeout active, false: timeout resolved
	}

	inline void setCANMappingReceivedParameter(void * const p_mappingObject)
	{
		this->value = 0;			//!< variable id
		this->object = p_mappingObject;		//!< the mapping object
		this->uvalue = true;			//!< true: timeout active, false: timeout resolved
	}

	inline void setTimerTimeoutParameter(std::int32_t p_type)
	{
		this->value = p_type;	//!< the type of the timeout (single shot/repeating timer)
	}

	inline void setStateChangeParameter(int p_state, unsigned char p_nodeID)
	{
		this->value  = p_state;	//!< the state
		this->uvalue = p_nodeID;	//!< the node id
	}

	inline void setJ1939TPParameter(void * p_ecu, std::uint32_t p_pgn, std::int32_t p_result, unsigned long long p_dataBytes)
	{
		this->object = p_ecu;				//!< ECU this TP comes from (or goes to)
		this->uvalue = p_pgn;				//!< the J1939 PGN of the transport protocol message
		this->value  = p_result;			//!< the result of sending the transport protocol message
		this->signalPara.j1939TP.dataBytes = p_dataBytes;	//!< The 8 data bytes of a CAN message as 64Bit value
	}

	inline void setPowerManagementTimeoutParameter(std::int32_t p_state, std::uint32_t p_timeout)
	{
		this->value	= p_state;		//!< the power management state
		this->uvalue	= p_timeout;		//!< the timeout value
	}

	inline void setPowerManagementDeviceStateParameter(std::uint32_t p_device, std::int32_t p_state)
	{
		this->value	= p_state;		//!< the state of the hardware device
		this->uvalue	= p_device;		//!< the hardware device id
	}

	inline void setPowerManagementPowerStateParameter(std::int32_t p_state)
	{
		this->value	= p_state;		//!< the state of the hardware device
	}

	inline void setPowerManagementPowerStateChangeParameter(std::int32_t p_state)
	{
		this->value	= p_state;		//!< the state of the hardware device
	}

	inline void setIOToggleParameter(std::uint32_t p_device, std::int32_t p_value)
	{
		this->value	= p_value;		//!< the value of the hardware device
		this->uvalue	= p_device;		//!< the hardware device id
	}

	inline void setMediaEventParameter(const std::string & p_filePath,
			unsigned int p_totalTimeInSeconds,
			unsigned int p_currentTimeInSeconds,
			int p_state,
			int p_oldState,
			void * p_listener,
			bool p_metadataChanged = false)
	{
		this->value				= p_state;
		this->uvalue				= p_oldState;
		this->pointer				= nullptr;
		this->strValue				= p_filePath;
		this->object				= p_listener;
		this->signalPara.mediaEvent.totalTimeInSeconds = p_totalTimeInSeconds;
		this->signalPara.mediaEvent.currentTimeInSeconds = p_currentTimeInSeconds;
		this->signalPara.mediaEvent.metadataChanged	= p_metadataChanged;
	}

	inline void setVariableParameter(std::int32_t p_value,
					 unsigned int p_combinedKey,
					 bool p_isValid,
					 void * const p_object = nullptr,
					 unsigned long long p_timestamp = 0)
	{
		this->value = p_value;
		this->pointer = nullptr;	// just in case somebody want to read a string value from this
		this->uvalue = p_combinedKey;
		this->object = p_object;
		this->signalPara.variable.isValid = p_isValid;
		this->signalPara.variable.timestamp64 = p_timestamp;
	}

	inline void setVariableParameter(const std::string & p_strValue,
					 unsigned int p_combinedKey,
					 bool p_isValid,
					 void * const p_object = nullptr,
					 unsigned long long p_timestamp = 0)
	{
		this->value = -1;	// just in case somebody tries to read the int value from this
		this->strValue = p_strValue;
		this->pointer = nullptr;
		this->uvalue = p_combinedKey;
		this->object = p_object;
		this->signalPara.variable.isValid = p_isValid;
		this->signalPara.variable.timestamp64 = p_timestamp;
	}
};



/**
 *	@brief Event listener interface that needs to be implemented to get notified of input events.
 *
 *	Subscription to events has to be implemented separately but they are expected to be handled in
 *	'signalOccurred()'.
 */
class WESignalListener
{
public:
	/**
	 *	@brief Handle events in this method.
	 *
	 *	@param[in] p_signal The event that has occurred.
	 *
	 *	@returns an error code - WE_SUCCESS in case the event was successfully handled.
	 *
	 *	p_signal should potentially be completed in this call but it should be carefully considered
	 *	to e.g. alter the event type since the event may be forwarded to other entities
	 *	after this method terminated. They are probably expecting the original content then.
	 *	I.e. generally p_signal should be copied before the existing content is modified.
	 *
	 *	It is suggested to implement event handling in the data context (fast thread) in this
	 *	method.
	 */
	virtual int signalOccurred(WESignal & p_signal) = 0;
	virtual ~WESignalListener() = default;
};
}	// namespace WE

#endif /* SRC_EVENT_HANDLING_INCLUDE_WE_SIGNAL_H_ */
