/**
 *	@file we_gui_math_utils.h
 *
 *	@author preths@topcon.com
 *	@author Patrick Reths
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2020
 *
 */

#ifndef WE_GUI_MATH_UTILS_H
#define WE_GUI_MATH_UTILS_H

namespace WE
{
class WERect;
struct WEPoint;

/**
 *
 * This namespace contains some mathematical utility functions which are used in the GUI classes.
 *
 */
namespace WEGUIMathUtils
{


/**
 *	@fn float gradToRad(float angle)
 *
 *	@brief calculate the grad of the angle to the radian
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@param angle is the angle for the calculation
 *
 *	@returns the result of the calculation
 */
float gradToRad(float p_angle);


/**
 *	@fn float fixEllipseAngle(float width, float height, float inputAngle)
 *
 *	@brief Converts real ellipse angles used in ISO 11783-6 to eccentric angles used by Qt
 *
 *	Some people (e.g. those who created ISO 11783-6) expect real angles
 *	rather than what happens if putting angles into ellipse equations (as Qt does):
 *	https://bugreports.qt-project.org/browse/QTBUG-2644
 *	http://mathworld.wolfram.com/EccentricAngle.html
 *	So we borrowed this method from Brisbane.
 *
 *	@param p_width the width of the ellipse
 *	@param p_height the height of the ellipse
 *	@param p_inputAngle the input angle in degrees
 *
 *	@returns the converted angle in degrees
 */
float fixEllipseAngle(float p_width, float p_height, float p_inputAngle);


/**
 *	@fn WEPoint getEllipsePoint(const WERect& rect, float angle)
 *
 *	@brief Calculates the position of a point in an ellipse
 *
 *	@author preths@topcon.com
 *	@author Patrick Reths
 *
 *	@param p_rect the rectangle of the ellipse
 *	@param p_angle the angle of the point in degrees
 *	@param p_isoAngle convert the angle with the fixEllipseAngle method if true
 *
 *	@returns the calculated position
 */
WEPoint calculateEllipsePoint(const WERect& p_rect, float p_angle, bool p_isoAngle=false);


/**
 *	@fn bool isPointInEllipse(float px, float py, float radiusX, float radiusY, float centerX, float centerY)
 *
 *	@brief Checks if a point is inside a ellipse
 *
 *	@author preths@topcon.com
 *	@author Patrick Reths
 *
 *	@param p_px the x position of the point
 *	@param p_py the y position of the point
 *	@param p_radiusX the horizontal radius of the ellipse
 *	@param p_radiusY the vertical radius of the ellipse
 *	@param p_centerX the center of the ellipse on the x axis
 *	@param p_centerY the center of the ellipse on the y axis
 *
 *	@returns true, if the point is inside the ellipse
 */
bool isPointInEllipse(float p_px, float p_py, float p_radiusX, float p_radiusY, float p_centerX, float p_centerY);
} //End of namespace WEGUIMathUtils
} //End of namespace WE
#endif // WE_GUI_MATH_UTILS_H
