/**
 *	@file we_inplace_editing_listener.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 */

#ifndef WE_INPLACE_EDITING_LISTENER_H_
#define WE_INPLACE_EDITING_LISTENER_H_

#include <variant>
#include <string>

namespace WE
{

/**
 *	@class WEInplaceEditingListener
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief This is an abstract class which requires the listener to implement the interfaces defined here.
 *	A Listener can register itself using the setInplaceEditingListener() of WEWidget. The inplace editing
 *	widget can fire these events and notify the registered listener (e.g. the DDO being edited).
 *
 *	@namespace WE
 *
 */
class WEInplaceEditingListener
{
public:
	// virtual destructor gets me rid of all the warnings - and has to provide an empty implementation here
	// in order to avoid a missing reference in final application.
	/**
	 * @fn ~WEInplaceEditingListener()
	 *
	 * @brief	Virtual Destructor
	 */
	virtual ~WEInplaceEditingListener() = default;

	/**
	 *	@brief	This function will be called when editing is confirmed e.g. by pressing encoder.
	 *
	 *	@param p_value A std::variant containing the confirmed value either as int, double or string
	 *		=> use "std::holds_alternative<TYPE>(p_value)" to check if data type TYPE is stored in the variant
	 *		and then "std::get<TYPE>(p_value)" to extract that value
	 *
	 *	Once the inplace editing widget receives an event that confirms the edited value it will
	 *	call this method of the WEInplaceEditingListener. The listener can then decide
	 *	if it wants to accept the value and it can close or hide the editing widget.
	 */
	virtual void onEditingConfirmed(std::variant<int, double, std::string> p_value) = 0;

	/**
	 *	@brief This method will be called when the editing is cancelled.
	 *
	 *	@param oldValue The initial value with which the editing was initiated.
	 *
	 *	Once the inplace editing widget receives the WE::ESCAPE_EVENT it will cancel editing
	 *	and call this method of the WEInplaceEditingListener.
	 *
	 *	@WARNING This explanation does not really fit since every event is sent to DDO first.
	 *
	 */
	virtual void onEditingCancelled(void * p_oldValue) = 0;

	/**
	 *	@fn onEditingValueChanged(void *value)
	 *
	 *	@brief This method will be called when the value being edited is changed.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param value The changed value
	 *
	 *	Once the inplace editing widget receives an event that changes the edited value it will
	 *	call this method of the WEInplaceEditingListener. The listener can then decide
	 *	if it wants to do something with the intermediate value.
	 *
	 *
	 *	This method could have a return value allowing the listener to modify the value.
	 *	This could be helpful to restrict it to a (currently) valid range whose criteria may
	 *	be too complex to be checked by the editing widget itself.
	 *
	 */
	// unused: virtual void onEditingValueChanged(void *value)=0;
};
}

#endif /*WE_INPLACE_EDITING_LISTENER_H_*/
