/**
 *	@file we_font.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief WEFont Class
 *
 *	@see toolkit_we_font.h
 *
 *	The WEFont class specifies a font used for drawing text.
 */
#ifndef WE_FONT_H_
#define WE_FONT_H_

#include <memory>
//Included for the WEFontStyle, WEColor
#include "../../../common/include/we_color.h"
#include "../../include/we_datatypes_gui.h"

//Included headers for Data Members
#include "toolkit_we_font.h"

namespace WE
{


/**
 *	@enum WEFontFamily
 *
 *	@brief This enum describes the different fonts families that are used to display text
 *
 */
typedef enum
{
	Helvetica,
	SansSerif = Helvetica,
	Times,
	Serif = Times,
	Courier
} WEFontFamily;


/**
 *	@enum WEFontStyle
 *
 *	@brief This enum describes the different styles that are used to display text
 *	@brief Normal or Italic
 */
typedef enum
{
	STYLE_NORMAL 		= 0x0000,	// 0000 0000	Default
	STYLE_BOLD 		= 0x0001,	// 0000 0001
	STYLE_CROSSED_OUT 	= 0x0002,	// 0000 0010
	STYLE_UNDERLINED 	= 0x0004,	// 0000 0100
	STYLE_ITALIC		= 0x0008,	// 0000 1000
	STYLE_INVERTED 		= 0x0010,	// 0001 0000	Inverting to exchange background and pen color
	STYLE_FLASH_INVERTED	= 0x0020,	// 0010 0000	Flash between inverted and styles 0-3
	STYLE_FLASH_HIDDEN 	= 0x0040,	// 0100 0000	Flash both background and foreground between hidden and styles 0-4
	STYLE_PROPORTIONAL 	= 0x0080,	// 1000 0000	Proportional font rendering supported

	//STYLE_OUTLINED		= 0x1000	// 1000 ... 0000 Draw outline around characters
	STYLE_FIXED_LINE_HEIGHT	= 0x2000,	// 0010 0000 0000 0000 ISO 11783-6 requires font height to be line height

	STYLE_MAX_LIMIT		= 0xffff	// use as plausibility check
} WEFontStyle;


class WEPainter;

class WEFont
{
private:
	/*--------------------------------------------------------*/
	/**
	 *	@fn WEFont()
	 *
	 *	@brief Standard constructor of WEFont - only for serialization
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	Constructs a font object that uses the application's default font.
	 */
	/*--------------------------------------------------------*/
	WEFont() = default;


public:
	/*--------------------------------------------------------*/
	/**
	 *	@fn WEFont()
	 *
	 *	@brief constructor of WEFont with initialisation
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_family - the font family of the font
	 *	@param[in] p_size - the size of the font
	 *	@param[in] p_style - the style of the font, which should be set of the datatypes WEFontStyle
	 *	@param[in] p_color - the color of the font
	 *
	 *	Constructs a font object that uses parameter for the initialization
	 */
	/*--------------------------------------------------------*/
	WEFont(const std::string & p_family, int p_size, int p_style, WEColor p_color, WEEncodeType p_encodeType = UTF_8, int p_width = 0);
	WEFont(const WEFont&);
	/*--------------------------------------------------------*/
	/**
	 *	@fn ~WEFont()
	 *
	 *	@brief Standard destructor of WEFont
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	Destroys the font object and frees all allocated resources
	 */
	/*--------------------------------------------------------*/
	virtual ~WEFont();


	/**
	 * @brief Set all parameters of an existing font.
	 *
	 * @param p_family
	 * @param p_size
	 * @param p_style
	 * @param p_color
	 * @param p_EncodeType default: UTF_8
	 * @param p_width > 0: fixed width font with this amount of pixels character width, else proportional font (default)
	 *
	 * @return An error code, WE::WE_SUCCESS in case of success
	 *
	 *
	 * This method can be called externally to avoid creating an instance just to assign it
	 * and is used internally for constructor, copy constructor and assignment operator.
	 *
	 */
	int assignAllMembers(
			const std::string & p_family,
			const int p_size,
			const int p_style,
			const WEColor p_color,
			const WEEncodeType p_EncodeType = UTF_8,
			const int p_width = 0);


	/*--------------------------------------------------------*/
	/**
	 *	@fn void setFontFamily(std::string p_family)
	 *
	 *	@brief Function to set the font family
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_fontFamily for the font family name (for Example "Times")
	 *
	 *	Set the family name of the font.
	 */
	/*--------------------------------------------------------*/
	void setFontFamily(const std::string & p_familyName);

	/*--------------------------------------------------------*/
	/**
	 *	@fn std::string getFontFamily() const
	 *
	 *	@brief Function to get the font family
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@return the font family name
	 *
	 *	Returns the requested font family name,
	 *	i.e. the name set in the constructor or the last setFontName() call
	 */
	/*--------------------------------------------------------*/
	std::string getFontFamily() const;

	/*--------------------------------------------------------*/
	/**
	 *	@fn void setSize(int p_size)
	 *
	 *	@brief Function to set the size of the font
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_fontSize for the font size
	 *
	 *	Set the font size to p_size pixel
	 */
	/*--------------------------------------------------------*/
	void setSize(int p_fontSize);

	/*--------------------------------------------------------*/
	/**
	 *	@fn int getSize()
	 *
	 *	@brief Function to get the size of the font
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@return the pixel size of the font
	 *
	 *	Returns the pixel size of the font
	 */
	/*--------------------------------------------------------*/
	int getSize() const;

	/*--------------------------------------------------------*/
	/**
	 *	@fn void setFontStyle(int p_fontStyle)
	 *
	 *	@brief Function to set the style of the font.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_fontStyle should be set from the WEFontStyle enums
	 *
	 *	Set the font style to the value of the WEFontStyle
	 */
	/*--------------------------------------------------------*/
	void setFontStyle(int p_fontStyle);

	/*--------------------------------------------------------*/
	/**
	 *	@fn int getFontStyle()
	 *
	 *	@brief Function to get the style of the font
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@return the mask for the font style (of the type WEFontStyle)
	 */
	/*--------------------------------------------------------*/
	int getFontStyle() const;

	/*--------------------------------------------------------*/
	/**
	 *	@fn void setFontColor(WEColor p_fontColor)
	 *
	 *	@brief Function to set the color of the font.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_fontColor should be from tha type WEColor
	 *
	 *	Set the font color
	 */
	/*--------------------------------------------------------*/
	void setFontColor(WEColor p_fontColor);

	/*--------------------------------------------------------*/
	/**
	 *	@fn WEColor getFontColor()
	 *
	 *	@brief Function to get the color of the font
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@return the WEColor object with the color of the font
	 */
	WEColor getFontColor() const;

	/**
	 *	@fn int getStringWidth(WEPainter* p_painter, std::string p_text)
	 *
	 *	@brief Function to measure the width of the string
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_text for the text, which width should be calculated
	 *
	 *	@return the width of a string in pixel
	 */
	int getStringWidth(WEPainter* p_painter, std::string p_text) const;
	int getStringWidth(WEPainter* p_painter, std::wstring p_text) const;

	/**
	 *	@fn void setFontType(WEEncodeType p_EncodeType)
	 *
	 *	@brief Function to set the font encoding type
	 *
	 *	@author satheeskumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *	@param[in] p_EncodeType Encoding Type
	 *
	 *	Set the font type
	 */
	void setFontType(WEEncodeType p_EncodeType);

	/**
	 *	@fn WEEncodeType getFontType()
	 *
	 *	@brief Function to get the Font Type
	 *
	 *	@author satheeskumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *	@return the Encode Type
	 */
	WEEncodeType getFontType() const;

	/**
	 *	@fn void setFontWidth(int p_width)
	 *
	 *	@brief Function to set the font width
	 *
	 *	@author satheeskumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *	@param[in] p_width font width
	 *
	 *	Set the font width to 0 to let the underlying toolkit render the font.
	 *	Set it to another positive value to define the width of a single character in pixels.
	 *
	 *	@see setSize()
	 */
	void setFontWidth(int p_width);

	/**
	 *	@fn int getFontWidth()
	 *
	 *	@brief Function to get the Font Width
	 *
	 *	@author satheeskumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *	@return the font width
	 *
	 *	@see getSize()
	 */

	int getFontWidth() const;

	/**
	 *	@fn WEFont& operator=(const WEFont& p_font)
	 *
	 *	@brief Assignment operator to allow the assigning of LHS WEFont
	 *	of "=" to RHS WEFont.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_font WEFont to be assigned to LHS of "=".
	 *
	*/
	WEFont& operator=(const WEFont& p_font);

	ToolkitWEFont* getToolkitWEFont() const;

//	int setOutlined(bool p_outlined)
//	{
//		this->outlined = p_outlined;
//		return 0;
//	}
//
//	bool isOutlined(void)
//	{
//		return this->outlined;
//	}


	/**
	 *	@fn int getStringWidthWithoutPainter(std::string p_text)
	 *
	 *	@brief Get width of the string provided using latin 1 encoding.
	 *
	 *	@param p_text Text to be evaluated
	 *
	 *	@returns The width in pixels of p_text rendered for the screen with this font.
	 *
	 */
	int getStringWidthWithoutPainter(std::string p_text) const;
	int getStringWidthWithoutPainter(std::wstring p_text) const;


	/**
	 *	@fn WERect getBoundingRect(WERect p_rect, WEAlignment p_align, bool p_autowrap, std::string p_text, bool p_wrapOnHyphen = false)
	 *
	 *	@brief Get bounding rectangle of the string provided using latin 1 encoding.
	 *
	 *	@param p_rect	Rectangle where p_text should be shown, e.g. widget boundaries.
	 *	@param p_align	Flags defining how p_text should be aligned.
	 *	@param p_autowrap True if p_text should be automatically wrapped
	 *	@param p_text	Text to be evaluated.
	 *	@param p_wrapOnHyphen	False: hyphen character (0x2D) is not considered line breaking
	 *
	 *	@returns The bounding rect of p_text rendered for the screen with this font and the parameters given.
	 *
	 */
	WERect getBoundingRect(const WERect & p_rect, WEAlignment p_align, bool p_autowrap, std::string p_text, bool p_wrapOnHyphen = false) const;


	/**
	 *	@fn WERect getBoundingRect(WERect p_rect, WEAlignment p_align, bool p_autowrap, std::wstring p_text, bool p_wrapOnHyphen = false)
	 *
	 *	@brief Get bounding rectangle of the string provided using unicode encoding.
	 *
	 *	@param p_rect	Rectangle where p_text should be shown, e.g. widget boundaries.
	 *	@param p_align	Flags defining how p_text should be aligned.
	 *	@param p_autowrap True if p_text should be automatically wrapped
	 *	@param p_text	Text to be evaluated.
	 *	@param p_wrapOnHyphen	False: hyphen character (0x2D) is not considered line breaking
	 *
	 *	@returns The bounding rect of p_text rendered for the screen with this font and the parameters given.
	 *
	 *	Qt consider the hyphen character (0x2D) also a line separator character for the text wrapping , so wrapping
	 *	occurs at the hyphen character by default. So when the wrapOnHyphen option is disabled then wrap should not
	 *	happen at the hyphen and this is achieved by replacing the hyphen character as non breaking hyphen character.
	 *	This option is basically required by ISO-VT
	 */
	WERect getBoundingRect(const WERect & p_rect, WEAlignment p_align, bool p_autowrap, std::wstring p_text, bool p_wrapOnHyphen = false) const;


	/**
	 *	@fn WERect getBoundingRectUTF8(WERect p_rect, WEAlignment p_align, bool p_autowrap, std::string p_text, bool p_wrapOnHyphen = false)
	 *
	 *	@brief Get bounding rectangle of the string provided using unicode encoding.
	 *
	 *	@param p_rect	Rectangle where p_text should be shown, e.g. widget boundaries.
	 *	@param p_align	Flags defining how p_text should be aligned.
	 *	@param p_autowrap True if p_text should be automatically wrapped
	 *	@param p_text	Text to be evaluated.
	 *	@param p_wrapOnHyphen	False: hyphen character (0x2D) is not considered line breaking
	 *
	 *	@returns The bounding rect of p_text rendered for the screen with this font and the parameters given.
	 *
	 */
	WERect getBoundingRectUTF8(const WERect & p_rect, WEAlignment p_align, bool p_autowrap, std::string p_text, bool p_wrapOnHyphen = false) const;


	/**
	 *	@fn bool operator==(const WEFont& o) const
	 *
	 *	@brief  Comparison operator for WEFont.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] o Reference to the other WEFont object.
	 *
	 *	@returns true if the objects are equal
	 *
	 */
	bool operator==(const WEFont& p_o) const;


	/**
	 *	@fn bool operator!=(const WEFont& o) const
	 *
	 *	@brief  Comparison operator for WEFont.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] o Reference to the other WEFont object.
	 *
	 *	@returns true if the objects are not equal
	 *
	 */
	bool operator!=(const WEFont& p_o) const;


	/**
	 * Method to add a font to the application that is not installed on the OS
	 * @param p_fontPath Path to the font (including font file)
	 */
	static void addFontToApplication(std::string & p_fontPath);

	static void printAvailableFonts();

	static int debugSection;

private:
	void create();

	std::unique_ptr<ToolkitWEFont> toolkitFont;

	std::string fontFamily{"DejaVu Sans Serif"};	//!<Font Family Name
	int fontSize{12};			//!< Font Size - Default font size s 12
	int fontStyle{WE::STYLE_NORMAL};		//!< Font Style Mask with the values of the WEFontStyle
	int fontWidth{0};
	WEColor fontColor{WE::black};		//!< Font Color

	//bool outlined;
	WEEncodeType encodeType{UTF_8}; //!< Text Encoding Type

	//SERIALIZATION
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int );
};
} // end of namespace WE

BOOST_CLASS_EXPORT_KEY(WE::WEFont)

#endif /* WE_FONT_H_ */
