/**
 *	@file toolkit_we_image.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch

 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief The Toolkit Interface  for the Image class
 *
 */

#ifndef TOOLKIT_WE_IMAGE_H_
#define TOOLKIT_WE_IMAGE_H_


#include <vector>
#include <QSGTexture>

namespace WE
{

class WEWidget;
class WEColor;

namespace WETextureCache
{
struct Data;
}


/**
 *	@class ToolkitWEImage
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@brief This class specifies a image to be used for drawing text.
 *
 *	@namespace WE
 *
 *
 */
class ToolkitWEImage
{
public:

	/**
	 *	@fn ToolkitWEImage()
	 *
	 *	@brief Default Constructor
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 */
	ToolkitWEImage() = default;

	/**
	 *	@fn ~ToolkitWEImage()
	 *
	 *	@brief Default Destructor
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 */
	virtual ~ToolkitWEImage() = default;

	/**
	 *	@fn int createImage(const std::string & p_strFile)
	 *
	 *	@brief creates an image
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_strFile File + full path
	 *
	 *	@returns WE_error
	 *
	 */
	virtual int createImage(const std::string & p_strFile) = 0;

	/**
	 *	@fn int createImage(QPixmap *p_pixmap)
	 *
	 *	@brief creates an image with the given pixmap
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] p_pixmap the QPixmap
	 *
	 *	@returns WE_error
	 */
	virtual int createImage(QPixmap *p_pixmap) = 0;

	/**
	 *	@fn int createImage(QSGTexture *p_texture, WETextureCache::Data *p_cacheData=nullptr)
	 *
	 *	@brief creates an image with the given texture
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] p_texture the QSGTexture
	 *	@param[in] p_cacheData optional pointer to the WETextureCache::Data object of the texture
	 *
	 *	@returns WE_error
	 */
	virtual int createImage(QSGTexture *p_texture, WETextureCache::Data *p_cacheData=nullptr) = 0;


	/**
	 *	@fn QSGTexture * getTexture(bool createIfNeeded)
	 *
	 *	@brief Gets the texture of this image or creates a new texture if it needs an update and createIfNeeded is true.
	 *
	 *	@warning: This method should only be called from the Qt quick scene graph rendering thread.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param p_createIfNeeded create a texture of this image if there is no texture or the texture is invalid
	 *
	 *	@see void releaseTexture()
	 *
	 *	@returns the texture
	 */
	virtual QSGTexture * getTexture(bool p_createIfNeeded=true) = 0;


	/**
	 *	@fn void releaseTexture()
	 *
	 *	@brief Releases the texture created by createTexture()
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@see QSGTexture * getTexture(bool createIfNeeded=true)
	 */
	virtual void releaseTexture() = 0;


	/**
	 *	@fn void invalidateTexture()
	 *
	 *	@brief Invalidates the texture of this image
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@see bool textureNeedsUpdate()
	 */
	virtual void invalidateTexture() = 0;


	/**
	 *	@fn bool textureNeedsUpdate()
	 *
	 *	@brief Returns true if the texture of this image needs an update.
	 *	       This is the case if the texture was never created or invalidateTexture was called before.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@see void invalidateTexture()
	 *
	 *	@returns true or false
	 */
	virtual bool textureNeedsUpdate() const = 0;


	/**
	 *	@fn void setCacheData(WETextureCache::Data *p_cacheData)
	 *
	 *	@brief Setter for the WETextureCache::Data object of this image
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] p_cacheData pointer to the WETextureCache::Data object of the texture of this image
	 */
	virtual void setCacheData(WETextureCache::Data *p_cacheData) = 0;


	/**
	 *	@fn virtual void setSize(int p_iWidth, int p_iHeight, bool p_resize) = 0
	 *
	 *	@brief overwrites the size read from the file
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_iWidth width
	 *	@param[in] p_iHeight height
	 * 	@param[in] p_resize yes or no
	 *
	 */
	virtual void setSize(int p_iWidth, int p_iHeight, bool p_resize) = 0;

	/**
	 *	@fn release()
	 *
	 *	@brief releases memory
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 */
	virtual void release() = 0;

	/**
	 *	@fn getRootObject()
	 *
	 *	@brief creates a WEImage
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns pointer to the according toolkit object
	 *
	 */
	virtual void* getRootObject() const = 0;

		/**
	 *	@fn getWidth()
	 *
	 *	@brief get image width
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns width of image
	 *
	 */
	virtual int getWidth() const = 0;

	/**
	 *	@fn getHeight()
	 *
	 *	@brief get image height
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns height of image
	 *
	 */
	virtual int getHeight() const = 0;


	/**
	 *	@fn int createTransparent(bool p_bTransparent, WEColor & p_transparencyColor)
	 *
	 *	@brief Makes image transparent at all pixels of color p_transparencyColor.
	 *
	 *	@param[in] p_bTransparent create = true or false for reset.
	 *	@param[in] p_transparencyColor Color to create mask if p_bTransparent is true.
	 *
	 *	@returns WE_SUCCESS or failure
	 *
	 *	Makes image transparent at all pixels of color p_transparencyColor. Alpha channels of the image are not
	 *	overwritten. Alpha value of p_transparencyColor is ignored.
	 */
	virtual int createTransparent(bool p_bTransparent, const WEColor & p_transparencyColor) = 0;

	/**
	 *	@fn int setTransparent(bool p_bTransparent, int p_colorIndex)
	 *
	 *	@brief Makes image transparent at all pixels of color index.
	 *
	 *	@param[in] p_bTransparent create = true or false for reset.
	 *	@param[in] p_colorIndex index for transparent color
	 *
	 *	@returns WE_SUCCESS or failure
	 */
	virtual int setTransparent(bool p_bTransparent, int p_colorIndex) = 0;


	/**
	 * 	@brief This should be a static method used to store images e.g. from an ISO-VT pool to file system.
	 *
	 *	This mainly exists to store images coming from an ISO-VT object pool with the original
	 *	references to the color table.
	 *	Refer to ISO-VT pool parser to see how the different formats defined in ISO 11783-6 are
	 *	converted into the common format expected here.
	 *	It is not really static because we cannot define it as 'static' and 'virtual' in this interface.
	 *
	 *	@param indexed8ImageData pointer to the raw image data - one byte per pixel, new lines start at next 32 bit boundary
	 *	@param imageWidth	width in pixels
	 *	@param imageHeight	height in pixels
	 *	@param colorTable	e.g. obtained from WEApp::ColorTable::getAllColors()
	 *	@param imageFileFormat	image file format (e,g; PNG, JPG etc)
	 *	@param imageFilePath	absolute path of the image file which shall be created
	 *	@returns		an error code - WE_SUCCESS in case of success
	 *
	 *	@see WEApp::PoolParser::saveRawDataAsImageFile()
	 *	@see WEApp::ColorTable::getAllColors()
	 */
	virtual int saveRawDataAsImageFile(
			unsigned char * p_indexed8ImageData,
			unsigned int p_imageWidth,
			unsigned int p_imageHeight,
			std::vector<unsigned int> & p_colorTable,
			const std::string & p_imageFileFormat,
			const std::string & p_imageFilePath) const = 0;


	/**
	 *	@fn bool hasAlphaChannel()
	 *
	 *	@brief pure virtual method, must be implemented by a class
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 */
	virtual bool hasAlphaChannel() const = 0;

	/**
	 *	@fn bool isTransparent()
	 *
	 *	@brief pure virtual method, must be implemented by a class
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 */
	virtual bool isTransparent() const = 0;


	/**
	 *	@fn void setWidget(ToolkitWEWidget * widget)
	 *
	 *	@brief pure virtual method, must be implemented by a class
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	Assign a toolkit widget to this toolkit image.
	 *
	 *
	 */
	virtual void setWidget(WEWidget * p_widget) = 0;

	virtual bool isSVG(void) const = 0;
	virtual bool isGIF(void) const = 0;
};
}
#endif /* TOOLKIT_WE_IMAGE_H_ */
