/**
 *	@file we_image_list.h
 *
 *
 *	@author For the concept and the initial version:
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@author For the migration from list to map (2011-01-10):
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Class definition of we_image_list
 *
 *	This class keeps a list of all images which are currently loaded.
 *	Its purpose is to maintain only one instance of each image in memory even if it
 *	should be displayed more than once.
 *
 *
 *	Requirements coming from different applications:
 *
 *	PClient:
 *
 *	a) Images are object properties and hence the same image can appear on multiple DDOs (e.g. as background image)
 *	b) It speeds up PClient if images persist until the application is destroyed.
 *	c) Typically images do not change while PClient is running.
 *
 *
 *	ISO-VT:
 *	a) Most images come from object pools and an ECU is allowed to overwrite an existing image with another.
 *	   Images are stored based on their object ID - the replacement gets the same filename.
 *	b) Consequence of "a)" is that we have to make sure that the VT can delete images when an image DDO is deleted.
 *	c) Some generated Images are shared among pools (e.g. those on the softkey level switches) and should persist
 *	   at least until the last pool is erased.
 *	d) Aux screen also uses internally generated images on its keys which should persist.
 *	e) Like PClient it could help if images would persist as long as the DDO does and not just destroyed
 *	   because the last widget gets deleted.
 *
 *
 *	--> Seems one solution for both would be a list working as initially intended for ISO-VT
 *	and an option "keep images forever" to be used by PClient.
 *	--> Modify list so that it
 *	1) keeps correct reference count
 *	2) never deletes an image automatically
 *	3) on request deletes an image if reference count == 0
 *	4) On destruction deletes all images
 *	5) On request deletes all images whose reference count == 0 (to save some memory)
 *
 *	Alternatively we could add an option to every image identifying its lifetime.
 *
 *	Or we could make the whole list optional somehow (if e.g. the DDO would create a WEImage just to be used by WEImageWidget
 *	then the DDO could easily control the lifetime of the image).
 *	This would require some modifications on WEImageWidget probably to take an WEImage pointer rather than a filename.
 *	--> from my point of view that would be the best solution - for ISO-VT at least.
 *	--> Let's do that - mainly for ISO-VT. Image widget will then NOT delete the image (and not even have the path and filename).
 *
 *	One additional problem may be that "release()" currently only takes the filename - maybe we should require the pointer
 *	obtained from a previous call to "create()".
 *
 */

#ifndef WE_IMAGE_LIST_H_
#define WE_IMAGE_LIST_H_

#include <memory>
#include <string>
#include <unordered_map>

namespace WE
{

class WEWidget;
class WEImageContainer;
class ToolkitWEImage;

// for terminals other than Opus A5 we expect access via hash values to be more efficient:
typedef std::unordered_map< std::string, WEImageContainer* > ImageList;	//!< hashmap <string, pointer>



/**
 *	@class WEImageList
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@brief Ask MKA
 *
 *	@namespace WE
 */
class WEImageList
{
public:

	/**
	 *	@fn WEImageList
	 *
	 *	@brief Constructor of image list
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 */
	WEImageList() = default;


	/**
	 *	@fn virtual ~WEImageList()
	 *
	 *	@brief Destructor.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 */
	virtual ~WEImageList();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEImageList(const WEImageList &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEImageList & operator=(const WEImageList &) = delete;

	/**
	 *	@fn int release(std::string & p_strFilename, bool p_forceRelease = false)
	 *
	 *	@brief release a instance of a image.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_strFilename filename
	 *	@param[in] p_forceRelease If true assume that the file was exchanged so that the stored copy shows the wrong image
	 *
	 *	@returns weerror code
	 *
	 *	checks count of images - unless p_forceRelease is true
	 *	returns WE_SUCCESS
	 */
	int release(const std::string & p_strFilename, bool p_forceRelease = false);


	/**
	 *	@fn void releaseAll()
	 *
	 *	@brief for safety therefore private
	 *
	 *
	 *	if the class gets killed
	 *	checks all images, gives error if not all images are released and release them.
	 *	could lead into problems when a application still wants to use it.
	 */
	void releaseAll();


	/**
	 *	@fn ToolkitWEImage * create(const std::string & p_strFilename, ToolkitWEWidget * p_widget)
	 *
	 *	@brief creates instance of an image.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param[in] p_strFilename filename
	 *	@param p_widget Pointer to the corresponding toolkit widget
	 *
	 *	@returns pointer to toolkit interface
	 *
	 *	If image is found in the list it returns the pointer and
	 *	increases the internal counter.
	 *
	 *	If p_bTransparent the transparency color OF THIS LIST is applied to the image.
	 *
	 *	@Warning Changing image size later on will also apply to all instances derived from the same file.
	 *
	 */
	std::shared_ptr<ToolkitWEImage> create(const std::string & p_strFilename, WEWidget * p_widget);

	/**
	 *	@fn static bool getKeepUnusedImages( void )
	 *
	 *	@brief Check if every unused WEtoolkitImage instances will be kept or not.
	 *
	 *	@param None
	 *
	 *	@returns true if image list should keep ToolkitWEImage instances even if they are not used currently
	 *	@returns false if image list will delete each instance of ToolkitWEImage when reference count is reset to 0
	 *
	 *	This method returns the value of the respective private static member.
	 *
	 *	@see static int setKeepUnusedImages( bool )
	 */
	static bool getKeepUnusedImages( void );


	/**
	 *	@fn static int setKeepUnusedImages( bool p_keepThem )
	 *
	 *	@brief Define if every unused WEtoolkitImage instances will be kept or not.
	 *
	 *	@param[in] p_keepThem - false if image list should be deleted, true otherwise
	 *
	 *	@returns WE::WE_SUCCESS
	 *
	 *	This method sets the value of the respective private static member.
	 *
	 *	This is intended to stay unchanged during the lifetime of an application
	 *	i.e. the idea is to define the policy at application initialization.
	 *
	 *	@warning release() will use the current setting whenever it is called.
	 *
	 *	@see static bool getKeepUnusedImages( void )
	 */
	static int setKeepUnusedImages( bool p_keepThem );


private:

	/**
	 *	@fn releaseSingleEntry(ImageList::iterator & p_imageIterator)
	 *
	 *	@brief Remove single entry identified by "p_imageIterator" from the list of images
	 *
	 *	@param[in] p_imageIterator - iterator into imageList entry to be removed
	 *
	 *	This method frees the image, the image container and the list entry.
	 *
	 *
	 */
	void releaseSingleEntry(ImageList::iterator & p_imageIterator);


	/**
	 *	@fn WEImageContainer * find(const std::string & p_strFilename, ImageList::iterator * p_imageIterator = nullptr)
	 *
	 *	@brief finds a image in the list
	 *
	 *	@param[in] p_strFilename - filename of the image to be found
	 *	@param[in/out] p_imageIterator - if != nullptr this will be set to the list element where returnContainer is stored.
	 */
	WEImageContainer * find(const std::string & p_strFilename, ImageList::iterator * p_imageIterator = nullptr);



	ImageList imageList{};	//!<internal list of images (which is actually a map now)

	static inline bool keepUnusedImages{true};	//!<Define if a ToolkitWEImage is deleted when reference count is 0. Default: assume speed is more important than memory consumption

	static int debugSection;
};
} //Namespace WE

extern WE::WEImageList	theWEImageList; //!< WE Image List class singleton

#endif /*WE_IMAGE_LIST_H_*/
