/**
 *	@file we_multimedia_player.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Class definition of WEMultimediaPlayer and some enums belonging to it.
 */


#ifndef WE_MULTIMEDIA_PLAYER_H_
#define WE_MULTIMEDIA_PLAYER_H_

#include <mutex>
#include <QMimeDatabase>
#include "../../../../event_handling/include/we_signal.h"
#include "../../../../common/include/we_size.h"
#include "../../../../os_abstraction/thread/include/we_thread.h"

class QQuickItem;

namespace WE {

class WEWidget;
class WEContainer;
class WEWindowWidget;
class WEEventQueue;

/**
 * Enum defining possible audio sinks (outputs)
 */
typedef enum
{
	WE_AUDIO_SINK_NONE				= 0,
	WE_AUDIO_SINK_INTERNAL_SPEAKER			= 1,//!< Internal speaker built in to the device
	WE_AUDIO_SINK_LINE_OUT				= 2,         //!< The audio line out
	WE_AUDIO_SINK_INTERNAL_SPEAKER_AND_LINE_OUT	= 3
}WEAudioSink;

/**
 * These media states are a 1:1 copy (except for the names) of Qt Phonon::State
 */
enum MediaStates
{
	MEDIA_STATE_LOADING	= 0,  //!< MEDIA_STATE_LOADING
	MEDIA_STATE_STOPPED	= 1,  //!< MEDIA_STATE_STOPPED
	MEDIA_STATE_PLAYING	= 2,  //!< MEDIA_STATE_PLAYING
	MEDIA_STATE_BUFFERING	= 3,//!< MEDIA_STATE_BUFFERING
	MEDIA_STATE_PAUSED	= 4,   //!< MEDIA_STATE_PAUSED
	MEDIA_STATE_EOS		= 5,    //!< End of stream reached
	MEDIA_STATE_ERROR	= 6,    //!< MEDIA_STATE_ERROR
	MEDIA_STATE_UNKNOWN	= 0xff    //!< Unknown state
};

/**
 * This enum lists the available categories for any media. Any source needs to be identified by a
 * category so that the different volume variables can have an influence on the volume that is
 * used when playing back these files.
 */
typedef enum
{
	MEDIA_CATEGORY_MULTIMEDIA	= 0x00,//!< MEDIA_CATEGORY_MULTIMEDIA
	MEDIA_CATEGORY_ALARM		= 0x01,    //!< MEDIA_CATEGORY_ALARM
	MEDIA_CATEGORY_EVENT		= 0x02,    //!< MEDIA_CATEGORY_EVENT
	MEDIA_CATEGORY_CLICK		= 0x03,    //!< MEDIA_CATEGORY_CLICK
	MEDIA_CATEGORY_UNKNOWN		= 0xff  //!< MEDIA_CATEGORY_UNKNOWN
}MediaCategory;

/**
 * The WEMultimediaPlayer is a singleton and is responsible for playing back all multimedia
 * files of one application (this includes multimedia such as MP3s or video files but also alarm and event sounds).
 * It can store several media sources and play them back at request. The media sources are uniquely identified by their file path.
 *
 * The basic concept is:
 * 1) Add a media file with the add() function
 * 2) Set the audio sink (speaker or line out)
 * 3) Set a WEWidget as video widget (video will be rendered in front of all other content) for video files.
 *    If the file is an audio file, you can also set a WEWidget. WEMultimediaPlayer then tries to extract album art
 *    from the media files and renders it in the provided WEWidget
 * 4) Optional: use setVolume to set the volume of the media file, if you don't do this, it will play at 100% volume
 * 5) use play(), stop(), pause() etc. to control playback of the media
 *
 * If you are interested, you can subscribe to events for certain media file. You will then get events
 * whenever the state of the media or the current playback time changes.
 *
 * @Warning Adding the same WEWidget to more than one media file will cause errors and undefined behaviour.
 * 	Using the same widget multiple times at the same time is not possible.
 * 	If you want to re-use the same widget, remove the old media file with the remove() function,
 * 	after that the widget can be set to the new media file.
 *
 */
class WEMultimediaPlayer : public WESignalListener
{
public:
	/**
	 * Destructor. Deletes all added source files.
	 */
	virtual ~WEMultimediaPlayer();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEMultimediaPlayer(const WEMultimediaPlayer &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEMultimediaPlayer & operator=(const WEMultimediaPlayer &) = delete;

	/**
	 * The singleton getInstance function.
	 * If no instance of the player exists, it will be created.
	 * @return The one and only instance of this calss.
	 */
	static WEMultimediaPlayer * getInstance(WEEventQueue * p_eventQueue = nullptr);

	/**
	 * Add a new media file to the player.
	 * Mandantory first step if a media file shall be played.
	 * @param p_path file path of media source
	 * @param p_category the media category to which the media belongs (see enum MediaCategory for details)
	 * @param p_listener Listener for events for the played file from this player
	 * @return WE_SUCCESS if adding the media was successful
	 * 	WE_ALREADY_EXISTS_ERROR if a media file with the same path already exists in the player
	 * 	WE_NOT_FOUND_ERROR if the file was not found on the file system
	 */
	static int add(const std::string & p_path, MediaCategory p_category, WESignalListener * p_listener = nullptr);

	/**
	 * With this function a media file can be removed from the player.
	 * This is mandatory if a video widget that was already used by a media file shall be re-used by a new media file
	 * @param p_path file path of media source
	 * @return WE_SUCCESS if removing the media was successful
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int remove(const std::string & p_path);

	/**
	 * Play back the media file identified by path.
	 * @Warning: See class desription for basic concept
	 * @param p_path file path identifier
	 * @return WE_SUCCESS if play was called successfully (this does not mean that it will be played back correctly
	 * 	(only phonon's play() function could be successfully called, if there are some problems e.g. with the video codec
	 * 	you can set yourself as event listener (see subscribeForEvents()) and be informed about the state of the media.)
	 * 	WE_STATE_NOT_SUPPORTED if you try to play a video file without setting the video widget before
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int play(const std::string & p_path);

	/**
	 * Pause the media file identified by path.
	 * @Warning: See class desription for basic concept
	 * @param p_path file path identifier
	 * @return WE_SUCCESS if pause was called successfully (this does not mean that it will be paused correctly
	 * 	(only phonon's pause() function could be successfully called, if there are some problems e.g. with the video codec
	 * 	you can set yourself as event listener (see subscribeForEvents()) and be informed about the state of the media.)
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int pause(const std::string & p_path);

	/**
	 * Stop the media file identified by path. Current time of the media file will be set to 0.
	 * @Warning: See class desription for basic concept
	 * @param p_path file path identifier
	 * @return WE_SUCCESS if stop was called successfully (this does not mean that it will be stopped correctly
	 * 	(only phonon's stop() function could be successfully called, if there are some problems e.g. with the video codec
	 * 	you can set yourself as event listener (see subscribeForEvents()) and be informed about the state of the media.)
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int stop(const std::string & p_path);

	/**
	 * Set the volume of the media file that shall be played back. This is always a percentage from 0 to 100.
	 * The media player will then set the "real" volume based on the value of the volume variables (e.g. @MaxVolume and @AlarmVolume)
	 * depending to which media category the media file belongs.
	 * @Warning: See class desription for basic concept
	 * @param p_path file path identifier
	 * @parm p_volume volume (from 0 to 100) that shall be set
	 * @return WE_SUCCESS if the intended volume was set successfully (this does not mean that it was successfully set to hardware)
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int setVolume(const std::string & p_path, unsigned char p_volume);

	/**
	 * This function is used to increase or decrease the currently set volume for a file.
	 * @param p_path file path identifier
	 * @param p_volumeOffset volume offset that shall be applied to the current volume (-100 to 100)
	 * @return WE_SUCCESS if the intended offset was set successfully (this does not mean that it was successfully set to hardware)
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int increaseVolume(const std::string & p_path, int p_volumeOffset);

	/**
	 * If the media file that shall be played is a video or if the audio file's album art shall be displayed,
	 * you MUST set a widget before calling the play() function. The widget can be any WEWidget.
	 * The video or album art will be displayed in the foreground of the widget. All other content of the widget will
	 * not be visible (advise: use a WEContainer)
	 * @param p_path file path identifier
	 * @param p_widget The WEContainer that shall be used to render contents into it
	 * @return WE_SUCCES if widget was set successfully and a valid path was created between media file and video widget
	 * 	WE_INTERFACE_ERROR If creating the path between media file and video widget failed
	 * 	WE_STATE_NOT_SUPPORTED_ERROR if the provided widget does not have a QWidget as toolkit widget
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int setVideoWidget(const std::string & p_path, WEContainer * p_widget);

	/**
	 * Remove a video widget that was set to the media p_path before.
	 * If the media really is a video and is currently not stopped, this function will make it stop.
	 * @param p_path file path identifier
	 * @return WE_SUCCESS if the pointer to the widget was removed from media
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path
	 */
	static int removeVideoWidget(const std::string & p_path);

	/**
	 * Set the loop count for a certain media file. See parameter description for details
	 * @param p_path file path identifier
	 * @param p_maxLoopCount 0: Don't loop. value > 0: loop value times. value < 0: infinite loop
	 * @return WE_SUCCESS if setting the loop count was successful
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int setMaxLoopCount(const std::string & p_path, int p_maxLoopCount);

	/**
	 * With this function a playback gap can be set. This only has an effect if max loop count is < 0 or > 0.
	 * @param p_path file path identifier
	 * @param p_gapInMilliseconds playback gap in milliseconds between each loop
	 * @return WE_SUCCESS if playback gap was set successfully
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int setPlaybackGap(const std::string & p_path, int p_gapInMilliseconds);

	/**
	 * This is a relative seek function. Set the current time of the media file to (currentTime + p_offsetInSeconds).
	 * This function only has an effect if the media file is either playing, paused or buffering.
	 * @param p_path file path identifier
	 * @param p_offsetInSeconds Seek offset in seconds. Positive: seek forward. Negative: seek backward.
	 * @return WE_SUCCESS if setting the new playback time was successful
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int seek(const std::string & p_path, int p_offsetInSeconds);

	/**
	 * This is a absolute seek function. Set the current time of the media file to p_absolutePositionInSeconds.
	 * This function only has an effect if the media file is either playing, paused or buffering.
	 * @param p_path file path identifier
	 * @param p_absolutePositionInSeconds Absolute time in seconds to jump to.
	 * @return WE_SUCCESS if setting the new playback time was successful
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int jumpTo(const std::string & p_path, int p_absolutePositionInSeconds);

	/**
	 * Get the total playback length of a media file
	 * @param p_path file path identifier
	 * @return The total playback time in seconds or -1 if no file identified by p_path is present
	 */
	static int getTotalTime(const std::string & p_path);

	/**
	 * This function returns the preffered volume for a certain file (from 0 to 100). This is not the volume that
	 * was set to hardware because the "hardware volume" is influenced by the volume variables
	 * @param p_path file path identifier
	 * @return intended volume from 0 to 100 if p_path was found. -1 is returned if p_path was not found.
	 */
	static int getPrefferedVolume(const std::string & p_path);

	/**
	 * Overwritten event listener method. The WEMultimediaPlayer shall subscribe to all predefined volume variables
	 * so that it can adjust the volume of each media category automatically.
	 * @Warning Subscribing to the volume variables is not done automatically, the application is responsible for that!
	 * @param p_signal. The WEMultimediaPlayer only handles integer value change events and also only the volume variables
	 * @return WE_SUCCESS if the event and variable was handled
	 * 	WE_NOT_SUPPORTED_ERROR if either event type or variable in the value change event is not handled here
	 */
	virtual int signalOccurred(WESignal & p_signal) override;

	/**
	 * If you need to know the current state of a media object or the current playback time, you should subscribe
	 * for events of this object. Currently only one subscriber is possible for each media object (and the last ones wins).
	 * Each time the current time, the total time or the state of the media object changes, a media event is sent to the subscriber
	 * @param p_path file path identifier
	 * @param p_listener Pointer to a WESignalListener that will receive the media events for the selected object
	 * @return WE_SUCCESS if listener was correctly set
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	// obsolete?
//	static int subscribeForEvents(const std::string & p_path, void * p_listener);

	/**
	 * Remove an event subscription that was added with subscribeForEvents()
	 * @param p_path file path identifier
	 * @param p_listener Pointer to the listener that shall be removed
	 * @return WE_SUCCESS if after the call p_listener is not the listener of p_path (but maybe another listener)
	 * 	WE_NOT_FOUND_ERROR if there is no media file stored with the given path (call add() before this function)
	 */
	static int removeSubscription(const std::string & p_path, WESignalListener * p_listener);

	/**
	 * Remove any subscriptions to p_listener from all present media objects
	 * @param p_listener Pointer to listener that will be removed from all media objects
	 * @return WE_SUCCESS if after the call p_listener is not the listener of p_path (but maybe another listener) (actually it's always WE_SUCCESS :) )
	 */
	static int removeAllSubscriptions(WESignalListener * p_listener);

	/**
	 * Retreive the title of a media file from it's meta data
	 * @param p_path file path of the media file
	 * @return Either an empty string if there is no media file stored with the given path (call add() before this function)
	 * 	or the file path itself if the title could not be extracted from meta data (e.g. if file format for meta data not supported)
	 * 	or the title extracted from the meta data
	 */
	static const std::string & getTitle(const std::string & p_path);

	/**
	 * Retreive the error description for the media object identified by p_path.
	 * This should only be called if a media event with state MEDIA_STATE_ERROR was received.
	 * @param p_path file path of the media file
	 * @return Either an empty string if there is no error or the error description provided by phonon.
	 */
	static const std::string & getErrorString(const std::string & p_path);

	/**
	 * This function adjusts the intended volume that was set to a media object with setVolume() with respect to
	 * the media category of the media object nd the last received volume variable values.
	 * @param p_source Pointer to source object for which the volume adjustment should take place
	 */
	void adjustVolume(void * p_source) const;

	static int getState(const std::string & p_path);

	void toggleFullscreen(const std::string & p_path);
	static void setFullscreen(bool p_enable, const std::string & p_path);

	static void setMainWindow(WEWindowWidget * p_mainWindow);

	static bool isVideo(const std::string & p_path);

	void fullscreenChanged(bool p_enabled);

	void setScreenResolution(WESize p_resolution);
	void setDisplayRotation(int p_rotation);

	static void doQuit(void);

private:
	/**
	 * Private constructor of WEMultimediaPlayer.
	 * Private because it is a singleton. Will get called by getInstance() exactly once.
	 */
	WEMultimediaPlayer();

	/**
	 * If p_widget is set with setVideoWidget() this functions makes sure that all other connections
	 * to this particular widget are disconnected before really adding the widget.
	 * One video widget may only belong to one media file!
	 * @param p_widget Pointer to the widget for which all connections should be cloase.
	 */
	void disconnectAllSourcesFromWidget(WEWidget * p_widget);

	void positionVideo(WEContainer * p_widget, void * p_videoSink, bool p_fullscreen);

	static void busListenerFunction(void*);

	static inline std::unique_ptr<QMimeDatabase> db{}; //!< static unique_ptr to the mime database that helps to determine if a source file has video based on mime type

	//static QQuickItem * createVideoItem(void);

	static inline std::unique_ptr<WEMultimediaPlayer> theInstance{};	//!< The one and only instance of this class (singleton)

	unsigned char maxVolume;	//!< Last received value of "@MaxVolume"
	unsigned char alarmVolume;	//!< Last received value of "@AlarmVolume"
	unsigned char eventVolume;	//!< Last received value of "@EventVolume"
	unsigned char minAlarmVolume;	//!< Last received value of "@MinEmergencyAlarmVolume"

	bool fullscreenEnabled;

	static WEWindowWidget * mainWindow;

	WE::WESize screenResolution;
	int displayRotation;
	static const std::string emptyString;

	static inline WEThread busThread{};
	static inline std::mutex sourceLock{};
	static inline bool quit{false};
};
}  // namespace WE


#endif /* WE_MULTIMEDIA_PLAYER_H_ */
