/**
 *	@file we_navigation_object.h
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd. Faisal
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	Interface for an object that is to be used with the navigation scheme.
 *	It represents as object in a hierarchy of objects.
 */


#ifndef WE_NAVIGATION_OBJECT_H_
#define WE_NAVIGATION_OBJECT_H_

#include "../../../event_handling/include/we_navigation_event.h"
#include "../../../gui/include/we_datatypes_gui.h"

namespace WE
{

class WEWidget;
class WESize;

// a class implementing WENavigationObject is expected to also implement event listeners
// so that it provides a methods "inputEventOccurred()" and "navigationEventOccurred()"
// TODO: Wondering if we should drop this definition now that we do navigation from the apps.
class WENavigationObject : public WEInputEventListener, public WENavigationEventListener
{
public:

	virtual ~WENavigationObject() = default;

	virtual bool isInputNavigationObject() const = 0;


	virtual WEBorderState getNavigationBorderState() const = 0;


	virtual WENavigationObject* getNavigationParent() const = 0;


	/**
	 *	@fn WE::WENavigationObject * getNavigationInputParent();
	 *
	 *	@brief Return parent that does not have to be skipped in navigation.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param None
	 *
	 *	@returns Pointer to first parent that survives "skipTypeInNavigation()" test
	 *	@returns Pointer to highest non-nullptr parent if on the way there is none that passes the test (helps e.g. to get page)
	 *	@returns nullptr otherwise
	 *
	 *	@see skipTypeInNavigation(int)
	 */
	virtual WE::WENavigationObject * getNavigationInputParent() const = 0;


	virtual bool hasChildNavigationObjects() const = 0;


	virtual bool hasSiblingNavigationObjects() const = 0;


	virtual WENavigationObject* getFirstNavigationChild() const = 0;


	virtual WENavigationObject* getFirstInputNavigationChild() const = 0;


	/**
	 *	@fn WENavigationObject * getLastInputNavigationChild()
	 *
	 *	@brief Provide last input navigation child of this parent.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns A pointer to the last input object which is child of this one.
	 *
	 *
	 *	This method is mainly intended for the navigation scheme - otherwise
	 *	getLastInputNavigationSibling() may be more appropriate.
	 *
	 *	@see getFirstInputNavigationChild()
	 *	@see getLastInputNavigationSibling()
	 *
	 */
	virtual WENavigationObject * getLastInputNavigationChild() const = 0;


	virtual WENavigationObject* getNextNavigationSibling() const = 0;


	virtual WENavigationObject* getNextInputNavigationSibling() const = 0;


	virtual WENavigationObject* getPreviousNavigationSibling() const = 0;


	virtual WENavigationObject* getPreviousInputNavigationSibling() const = 0;


	virtual WENavigationObject* getFirstInputNavigationSibling() const = 0;


	virtual WENavigationObject* getLastInputNavigationSibling() const = 0;


	virtual void repaint() = 0;


	virtual bool isInplaceEditingSupported() const = 0;


	virtual bool isDirectlyConnectedToEncoder() const = 0;


	/**
	 *	Option to invert encoder evaluation while being 'open for input'.
	 *
	 *	Mainly of interest for lists where higher values are on the bottom of the list.
	 */
	virtual bool higherValueIsKeyDownWhileEditing() const = 0;


	virtual bool isRollOverActive() const = 0;


	/**
	 *	@fn bool hasChildInSubTree(WE::WENavigationObject * p_child)
	 *
	 *	@brief Check if a pointer to a navigation object is still valid.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_child - pointer to child we are looking for
	 *
	 *	@returns true if it should be safe to still use this pointer
	 *
	 *
	 *	This method can be used e.g. to check if the navigation object which received a mouse
	 *	event still is existing.
	 */
	virtual bool hasChildInSubTree(WENavigationObject * p_child) const = 0;


	/**
	 *	@fn virtual int getTopLeftPosition(WE::WEPoint * posToReturn) = 0
	 *
	 *	@brief Get RELATIVE top left position of the navigation object on screen
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_posToReturn Pointer to a WEPoint struct that on successful return
	 *		contains the top left coordinate of the navigation object relative to it's parent
	 *
	 *	@returns WE_SUCCESS if successful ;)
	 *	@returns Other implementation dependent error codes
	*/
	virtual int getTopLeftPosition(WE::WEPoint * p_posToReturn) const = 0;


	/**
	 *	@fn virtual int getSize(WE::WESize * sizeToReturn) = 0
	 *
	 *	@brief Get size of the navigation object
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[out] p_sizeToReturn A Pointer to a WESize instance. Will contain the size
	 *		of the navigation object on successful return.
	 *
	 *	@returns WE_SUCCESS if successful
	 *	@returns Other implementation dependent error code
	*/
	virtual int getSize(WE::WESize * p_sizeToReturn) const = 0;


	/**
	 *	@fn virtual WEWidget * getWidget(void) = 0
	 *
	 *	@brief Get a pointer to the associated WEWidget
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Pointer to the WEWidget that is associated to this navigation object
	 *	@returns nullptr if no widget is associated
	*/
	virtual WEWidget * getWidget(void) const = 0;


	/**
	 *	@fn void * getBaseObject( void )
	 *
	 *	@brief Return base object of this navigation object.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns A pointer to the base object
	 *
	 *
	 *	For ISO-VT and PClient we use this to return a pointer to the DDO.
	 *
	 *	@see WEApp::AppWidgetController::getDDO()
	 */
	virtual void * getBaseObject( void ) const
	{
		return nullptr;
	}


	/**
	 *	@fn virtual int getTypeID(void) = 0
	 *
	 *	@brief Retrieve the type  ID of the navigation object
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Implementation specific type ID of the navigation object
	*/
	virtual int getTypeID(void) const = 0;

	// debugging only:
	virtual const std::string & getName() const = 0;
};
}
#endif /* WE_NAVIGATION_OBJECT_H_ */
