/**
 *	@file we_navigation_scheme.h
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd. Faisal
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Navigation Scheme entity used to change focus of widgets based on key and mouse inputs.
 *
 *	Main method of this class is 'navigate' which triggers WENavigationEvents based on the
 *	WEInputEvent it gets as parameter.
 *
 *	WENavigationEvents replace the 'forward events' used in the past.
 *
 *
 *
 *	For a configurable navigation we probably
 *	* keep the state in every navigation object (no real change)
 *	* keep navigation actions for every input event in every navigation object
 *	* provide default actions so that the navigation object only has to maintain deviations from the default
 *	* coordinate actions here (e.g. remove focus from one object when another one gets focus exclusively)
 *	* i.e. also keep basic setting here like 'navigation root', 'tab-order', 'tab-order-table'
 *
 *
 *	Navigation objects have to support the following states (no change):
 *	* output only (en-/disabled is only optical design not related to navigation)
 *	* disabled input
 *	* enabled input without focus
 *	* enabled input with focus --> all input goes to the object(s) 'with focus', for 'navigation'
 *	* open for input --> all input goes to the object(s) 'open for input', no more 'navigation'
 *
 *
 *	The following events could be used to change navigation states
 *	* touch press to specific object (top one on display position touched)
 *	* touch long press to specific object (top one on display position touched)
 *	* touch long release to specific object (top one on display position touched)
 *	* touch release to specific object (top one on display position touched? Or touch released?)
 *	* touch enter to specific object (top one on display position)
 *	* touch exit to specific object (top one on display position)
 *		* position within widget?
 *		* position close to widget?
 *
 *	* key press
 *	* key long press
 *	* key long release
 *	* key release
 *
 *	* encoder move
 *	( encoder push is a key )
 *
 *
 *	Navigation actions are
 *	* setFocussed(navObject, exclusively)
 *	* removeFocus(navObject)
 *	* openForInput(navObject, exclusively)
 *	* confirmInput(navObject)
 *	* escInput(navObject)
 *
 *	Helper functions are
 *	* setNavigationOrder(z-order/tab-order)
 *	* createTabOrderFromPosition(rowsOrColumns)
 *	* getFirstNavObject(ddoId) --> assuming it's sufficient to find one instance of the DDO
 *	* get/setNavigationRoot(navObject) --> only [grand-]children will be considered in navigation order
 *	* getNextNavObject(navObject)	--> based on current navigation order
 *	* getPrevNavObject(navObject)	--> based on current navigation order
 *	* getParent(navObject)
 *	* getFirstInputChild(navObject)
 *	* getLastInputChild(navObject)
 *	* getNextInputChild(navObject)
 *	* getFirstInputSibling(navObject)
 *	* getLastInputSibling(navObject)
 *	* getNextInputSibling(navObject)
 *
 *
 *
 *	Can we keep that separate from other script actions?
 *	Or will navigation scripts be the same? Should we e.g. allow to get/set DEI variable values?
 *	Or just keep the 'setFocussedObject()' and encoder simulation we already have? And keep navigation separate?
 *	Problem: They are operating in different contexts (data vs. GUI).
 */
#ifndef WE_NAVIGATION_SCHEME_H_
#define WE_NAVIGATION_SCHEME_H_

#include "../../../common/include/we_color.h"
#include "../../../event_handling/include/we_input_event.h"
#include "../../../event_handling/include/we_signal.h"
#include "../../common/include/we_border_profile.h"
#include "../../include/we_datatypes_gui.h"

namespace WE
{

class WEWidget;
class WEContainer;
class WEColor;
class WESignal;
class WETimer;
class WEBorderProfile;
class WENavigationEvent;
class WENavigationObject;

/**
 *	@class WENavigationScheme
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd. Faisal
 *
 *	@brief Navigation Scheme implementation to navigate through all the widgets at all levels of
 *		hierarchy.
 *
 *	@namespace WE
 *
 *	It maintains a root widget from where the navigation starts, e.g; Page Widget. And the current
 *	focussed widget which changes when focus changes from one widget to another.
 */
class WENavigationScheme : public WESignalListener
{

public:
	/**
	 *	@fn WENavigationScheme()
	 *
	 *	@brief Default constructor to create an instance of the WENavigationScheme.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 */
	WENavigationScheme();


	/**
	 *	@fn ~WENavigationScheme()
	 *
	 *	@brief Standard destructor to destroy an instance of the WENavigationScheme.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	Reset the member variable pointers to nullptr.
	 */
	virtual ~WENavigationScheme();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WENavigationScheme(const WENavigationScheme &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WENavigationScheme & operator=(const WENavigationScheme &) = delete;


	/**
	 *	@fn WENavigationObject * getNavigationRoot(void)
	 *
	 *	@brief Get current navigation root object.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns Navigation root object.
	 *
	 *
	 *	This could e.g. be used in order to check if root object should be reset
	 *	after an alarm vanished.
	 *
	 *	@see setNavigationRoot()
	 *
	 */
	WENavigationObject * getNavigationRoot(void);


	/**
	 *	@fn int setNavigationRoot(WENavigationObject * rootWidget, WENavigationObject * objectToBeFocussed = nullptr, bool p_resetOnly = false)
	 *
	 *	@brief Initialise the navigation scheme with the specified root from where the
	 *		navigation will start, e.g; the Page Widget.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@param[in] p_rootWidget - Root of the hierarchy of widgets that are to be navigated.
	 *	@param[in] p_objectToBeFocussed - object to get focus (again), e.g. after an alarm was closed
	 *	@param[in] p_resetOnly - only set focus to first input again
	 *
	 *	@returns WE_SUCCESS When navigation scheme is successfully initialized with the
	 *		specified root.
	 *	@returns WE_INVALID_PARAMETER_ERROR When rootWidget is nullptr.
	 *
	 *	The state of the root widget is calculated based on the following:
	 *	1. If it has input children then its state is "OPEN_FOR_INPUT" and first child input
	 *	widget's state is "ENABLED_WITH_FOCUS" This is repeated for all children as long as
	 *	there is nothing to select (e.g.: If there is only one frame with one string input
	 *	on a page then the frame is open for input and the sting widget has focus)
	 *	2. If it has no child input widget then its state is "DISABLED"
	 *
	 *	If both parameters are nullptr this method will only reinitialize navigation.
	 */
	int setNavigationRoot(WENavigationObject * p_rootWidget, WENavigationObject * p_objectToBeFocussed = nullptr, bool p_resetOnly = false);


	/**
	 *	@fn int updateNavigation(void)
	 *
	 *	@brief Based on known root object set focus to first input object (again) if required.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns WE::WE_SUCCESS in case of success.
	 *
	 *
	 *	Internally this calls currentFocussedObjectIsValid(void) and setNavigationRoot(void).
	 *
	 *	@see currentFocussedObjectIsValid(void)
	 *	@see setNavigationRoot()
	 *
	 */
	int updateNavigation(void);


	/**
	 *	@fn int currentFocussedObjectIsValid(void)
	 *
	 *	@brief Check if the object currently focussed still is an active input.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns true if the focussed object is a valid active input object.
	 *	@returns false otherwise
	 *
	 *	@warning This assumes that navigation root is still valid.
	 */
	bool currentFocussedObjectIsValid( void ) const;


	/**
	 *	@fn bool objectIsOnScreen( WE::WENavigationObject * p_obj )
	 *
	 *	@brief Check if the p_obj can still be found on screen.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_obj Object to be found on screen
	 *
	 *	@returns true if p_obj still is on screen.
	 *	@returns false otherwise
	 *
	 *	This method was introduced in order to figure out if it makes sense
	 *	to forward a touch release event to a DDO (by assuming that the DDO
	 *	should still exist if the respective WE::WENavigationObject does).
	 */
	bool objectIsOnScreen( WE::WENavigationObject * p_obj );


	/**
	 *	@fn WENavigationObject * getFocussedObject()
	 *
	 *	@brief Retrieve the currently focussed navigation object.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns Pointer to the currently focussed navigation object.
	 *	@returns nullptr if no widget has focus.
	 *
	 */
	WENavigationObject * getFocussedObject();


	/**
	 *	@fn void * getFocussedBaseObject(bool & p_isOpenForInput)
	 *
	 *	@brief Make focussed base object available even to other thread.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[out] p_isOpenForInput true if navigation object is open for input, false otherwise
	 *
	 *	@returns Pointer to base object.
	 *	@returns true if object is open for input (in parameter)
	 *
	 *	This method was introduced in order to allow checking if operator is currently editing
	 *	an object that should be modified by command.
	 *	For ISO-VT this is handled in two different threads - so this method provides buffered
	 *	information which is safe to read as long as the navigation scheme exists.
	 *
	 *	For WE's PClient and ISO-VT implementation this will return a pointer to an AppGraphicDDO.
	 */
	void * getFocussedBaseObject(bool & p_isOpenForInput);


	/**
	 *	@fn int setFocussedObject(WENavigationObject * p_objectToBeFocussed, bool p_openForInput)
	 *
	 *	@brief Set the focus to the specified object and also open it for input if specified
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@param p_objectToBeFocussed	the object to be selected, use nullptr to remove focus from the currently focussed object.
	 *	@param p_openForInput		TRUE open the specified for input, FALSE simply show it as focussed.
	 *
	 *	@returns WE::WE_SUCCESS		if object is selected (focussed/opened for input) successfully.
	 *	@returns WE::ALREDAY_EXISTS_ERROR if trying to change focus when another object is already being edited.
	 *
	 *	It is assumed that this is used when the system wants to change
	 *	the focussed object (e.g; via select input object command) and
	 *	NOT the operator.
	 */
	int setFocussedObject(WENavigationObject * p_objectToBeFocussed, bool p_openForInput);

	/**
	 *	@brief Send p_confirmEvent to p_objectToConfirm and close p_objectToConfirm.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_EVENT_NOT_SUPPORTED_ERROR if (CONFIRM_INPLACE_VALUE_EVENT != p_confirmEvent.eventType)
	 *	@returns WE_STATE_NOT_SUPPORTED_ERROR if p_objectToConfirm doesn't have focus currently
	 *	@returns WE_PERMISSION_ERROR if p_objectToConfirm is not open for input or 'connected to encoder'
	 *	@returns
	 */
	int confirmInput(const WENavigationEvent & p_confirmEvent, WENavigationObject * p_objectToConfirm);


	/**
	 *	@fn int escapeInput(unsigned int p_eventOrigin = WE::EVENT_ORIGIN_OPERATOR)
	 *
	 *	@brief Escape/Abort the editing being done by the operator.
	 *		This could be used when operator presses escape or
	 *		when a ESC command arrives or
	 *		when a pclient escape action is executed.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@param p_eventOrigin	The origin of the event (Operator, command or system).
	 *
	 *	@returns WE::WE_SUCCESS		if object is selected (focussed/opened for input) successfully.
	 *
	 *
	 */
	int escapeInput(unsigned int p_eventOrigin = WE::EVENT_ORIGIN_OPERATOR);


	/**
	 *	@brief Perform the navigation between widgets in the hierarchy based on the specified
	 *		event.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@param[in] p_event Event for which the navigation is to be performed.
	 *
	 *	@returns WE_SUCCESS if navigation is successfully done.
	 *	@returns WE_EVENT_NOT_SUPPORTED_ERROR If the type of the specified event is not recognized
	 *		by the navigation scheme.
	 *	@returns WE_INVALID_PARAMETER_ERROR If function is called with invalid parameters.
	 *
	 *	This is the actual logic of the navigation scheme, where it decided based on the state
	 *	of the currently focused widget and the incoming event what actions are to be taken,
	 *	which will be new focused widget and what state should be returned to the application
	 *	for further action.
	 */
	int navigate(const WEInputEvent & p_event);


	/**
	 *	@fn int enableNavigationFrameWidget(bool p_yes,	int p_lineWidth, WEColor p_enabledColor, WEColor p_inputColor, unsigned int p_timeout = 0, WE::WETimer * p_timer = nullptr)
	 *
	 *	@brief Enable an object independant navigation frame to draw around focused navigation objects
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_yes true for enabling the navigation frame. False for disabling it.
	 *	@param p_lineWidth The width of the lines drawn around the navigation object
	 *	@param p_enabledColor The color that is used if the navigation object is enabled with focus
	 *	@param p_inputColor The color that is used if the navigation object is open for input
	 *	@param p_timeout The time in ms after which the frame will disappear if unused (0 == never)
	 *	@param p_timer Instance of WETimer to be used for p_timeout. Can be nullptr if p_timeout == 0.
	 *
	 *	@returns WE_SUCCESS if either activating or deactivating worked and a
	 *		frame is drawn around the current focused object
	 *	@returns WE_NOT_FOUND_ERROR If no object is currently focused. The frame will be drawn ass soon as
	 *		a navigation object gets focus.
	 *	@returns WE_OUT_OF_BOUNDS_ERROR If it was not possible to determine the current focused object's
	 *		size or position
	 *
	 *	@warning This feature is only interesting for rectangular navigation objects.
	 *		If there are non-rectangular navigation objects, still a rectangular frame will be drawn around them.
	 *
	 *	NavigationScheme's default setting is to not use a navigation frame widget. If this function is called
	 *	wit p_yes == true, a frame in the defined color with the defined line width is drawn around
	 *	the current focused object.
	 *	The color of this frame changes from p_enabledColor when navigating from one object to the other to
	 *	p_inputColor when modifying the value of an object.
	 */
	int enableNavigationFrameWidget(
			bool p_yes,
			int p_lineWidth,
			WEColor p_enabledColor,
			WEColor p_inputColor,
			unsigned int p_timeout = 0,
			WE::WETimer * p_timer = nullptr);


	/**
	 *	@brief Handle events - currently only the one from WE::WETimer.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_event Event to be handled
	 *
	 *	@returns WE::WE_SUCCESS in case of success.
	 *
	 */
	int signalOccurred(WE::WESignal & p_event) override;


	/**
	 *	@fn void suspendNavigation(bool p_suspend)
	 *
	 *	@brief Switch navigation on/off.
	 *
	 *	@param p_suspend Set to true to switch off
	 *
	 *	@returns nothing.
	 *
	 */
	void suspendNavigation(bool p_suspend);


	/**
	 *	@fn bool isNavigationSuspended( void )
	 *
	 *	@brief Check if this instance is currently suspended or not.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns true if this instance is currently disabled.
	 *
	 */
	bool isNavigationSuspended( void ) const;


	/**
	 *	@fn int setNavigationFrameWidget(WENavigationObject * p_obj, WEBorderState p_state);
	 *
	 *	@brief Set position, size and state of the navigation frame widget
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_obj Pointer to a navigation object around which the navigation frame shall be drawn
	 *	@param p_state Naviagtion state of the focused object (currently only OPEN_FOR_INPUT
	 *		and ENABLED_WITH_FOCUS are supported)
	 *
	 *	@returns WE_SUCCESS if everything works and the navigation frame is drawn correctly around the object
	 *	@returns WE_NOT_FOUND_ERROR If there is no instance of the navigation frame widget
	 */
	int setNavigationFrameWidget(WENavigationObject * p_obj, WEBorderState p_state);


	/**
	 *	@fn int refreshNavigationFrameWidget(WENavigationObject * p_obj)
	 *
	 *	@brief refresh the navigation frame widget to adjust to the change in size and location on the focussed widget.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@param[in] p_obj - Object whose size or location have changed
	 *
	 *	@returns An error code - WE::WE_SUCCESS in case of success
	 *
	 *	This function does nothing if the p_obj is not the focussed widget.
	 *	This function is useful e.g; when a list changes its size when open for input
	 *	or any focussed object changes its size or location due to a ISO VT command.
	 */
	int refreshNavigationFrameWidget(WENavigationObject * p_obj);


	/**
	 *	@fn void setStandardNavigationDirection(bool p_stdDirection);
	 *
	 *	@brief Setter for member stdNavigationDirection.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_stdDirection	TRUE   use STANDARD navigation direction
	 *				FALSE  use REVERTED navigation direction
	 *
	 *	This method sets the ENCODER MOVEMENT DIRECTION and the global variables ENCODER_UP / ENCODER_DOWN.
	 *	This affects the navigate() method.
	 *
	 *	@see int navigate(const WEInputEvent & p_event);
	 */
	void setStandardNavigationDirection(bool p_stdDirection);

	/**
	 *	@fn bool isStandardNavigationDirection( void );
	 *
	 *	@brief Getter for member stdNavigationDirection.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@preturn	TRUE   STANDARD navigation direction is used
	 *			FALSE  REVERTED navigation direction is used
	 *
	 */
	bool isStandardNavigationDirection( void ) const;

protected:


	/**
	 *	@fn int setNavigationBorderState(WENavigationObject * p_borderObject, WEBorderState p_borderState, unsigned int p_eventOrigin = WE::EVENT_ORIGIN_OPERATOR, WEBorderState p_oldState = WE::BORDER_STATE_UNDEFINED )
	 *
	 *	@brief Trigger sending a navigation state event to the object specified.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_borderObject - Object whose state should be set
	 *	@param[in] p_borderState - state it should be set to
	 *	@param[in] p_eventOrigin - the origin of the event e.g; Operator, System or Command.
	 *	@param[in] p_oldState - state to be left for the new one.
	 *
	 *	@returns An error code - WE::WE_SUCCESS in case of success
	 *
	 *	@see WEBorderState
	 *
	 *	This function sends out a navigation event based on the state the object should change into.
	 *	It is also used to place the navigation frame widget with the correct size, position and state
	 *	to the current focused navigation object.
	 *	Currently p_oldStaten is only used to add it to a "focus gained" event in order to decide if we should
	 *	create a CAN message for it on ISO-VT.
	 */
	int setNavigationBorderState(WENavigationObject * p_borderObject,
				    WEBorderState p_borderState,
				    unsigned int p_eventOrigin = WE::EVENT_ORIGIN_OPERATOR,
				    WEBorderState p_oldState = WE::BORDER_STATE_UNDEFINED );


	/**
	 *	@fn int setNavigationBorderStateToHierarchy(WENavigationObject * p_borderObject, WEBorderState p_borderState, unsigned int p_eventOrigin = WE::EVENT_ORIGIN_SYSTEM)
	 *
	 *	@brief Trigger sending a navigation state event to the object and all its children.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_borderObject - Object (and it's children ) whose state should be set
	 *	@param[in] p_borderState - state it should be set to
	 *
	 *	@returns An error code - WE::WE_SUCCESS in case of success
	 *
	 *	@see WEBorderState
	 *
	 *	This function is used to reset the states of the object and its children to the specified state.
	 */
	int setNavigationBorderStateToHierarchy(WENavigationObject * p_borderObject, WEBorderState p_borderState, unsigned int p_eventOrigin = WE::EVENT_ORIGIN_SYSTEM);

	/**
	 *	@fn void resetFrameTimer(void)
	 *
	 *	@brief Restart time after which the navigation frame should disappear.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns nothing
	 *
	 */
	void resetFrameTimer(void);

	/**
	 *	@fn void updateKeyboardFocus(void)
	 *
	 *	@brief Set the keyboard focus to the currently focused widget
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param none
	 *
	 *	@returns nothing
	 *
	 */
	void updateKeyboardFocus(void) const;


	/**
	 *	@brief This function sends navigation events to the given navigation object
	 *
	 *	@param p_navObj The object to send the event too.
	 *	@param p_forwardEvent The event to send
	 *
	 *	@returns return value of currentFocusedObject->nevigationEventOccurred()
	 *
	 *	@see WENavigationObject
	 *	@see AppWidgetController
	 */
	static int forwardNavigationEvent(WENavigationObject * p_navObj, WENavigationEvent p_forwardEvent);


	/**
	 *	@brief This function sends input events to the given navigation object
	 *
	 *	@param p_navObj The object to send the event too.
	 *	@param p_forwardEvent The event to send
	 *
	 *	@returns return value of currentFocusedObject->inputEventOccurred()
	 *
	 *	Basically navigation scheme should transform input events into navigation events.
	 *	There are some 'inplace editing widgets' though which operate directly on input
	 *	events (which they should probably NOT get from navigation scheme).
	 *
	 *	@see WENavigationObject
	 *	@see AppWidgetController
	 */
	static int forwardInputEvent(WENavigationObject * p_navObj, WEInputEvent p_forwardEvent);


	/**
	 *	@fn int setNavFramePositionSize(WENavigationObject * p_obj)
	 *
	 *	@brief Move and resize navigation frame widget to enclose p_obj.
	 *
	 *	@author mlenhart@topcon.com
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_obj - Object to be enclosed by navigation frame widget
	 *
	 *	@returns An error code - WE::WE_SUCCESS in case of success
	 *	@returns WE::WE_INVALID_PARAMETER_ERROR if p_obj is nullptr
	 *
	 *	This method only sets the position and size of the frame widget.
	 *	It does not care for hide, show, repaint etc.
	 */
	int setNavFramePositionSize(WENavigationObject * p_obj);

	static int debugSection;

	bool navigationFrameWidgetEnabled;	//!< Flag to tell if the usage of the separate widget to draw the navigation is enabled
						// Not to be confused with the usage of hierarchical "frame navigation" in PClient
	WE::WEContainer * naviFrameWidget;	//!< Currently a WEContainer is used as the navigation frame widget

	WE::WEColor navigationFrameColorEnabled;	//!< The color of the navigation frame for eanbled with focus state
	WE::WEColor navigationFrameColorInput;		//!< The color of the navigation frame for open for input state
	int navigationFrameLineWidth;		//!< Line width of the navigation frame widget's rectangle that is drawn around navigation objects
	WE::WEBorderProfile borderProfile{};	//!< The border profile used for the navigation frame widget

	WE::WETimer * timer;			//!< Timer instance to be used here will be created and destroyed outside of this class (currently it's a singleton anyways)
	unsigned int frameTimeout;		//!< Time of inactivity after which the navigation frame widget should disappear
	bool hideFrame;				//!< Indicate if the navigation frame is currently hidden or not.
	bool navigationSuspended;		//!< Complete navigation is suspended, ignore all events

	/*
	 *  "true" 	= navigate down => clockwise
	 *		= navigate up	=> anticlockwise
	 *  "false"  	= navigate down	=> anticlockwise
	 *		= navigate up	=> clockwise
	 */
	bool stdNavigationDirection;		//!< Indicates the direction of Encoder movement

private:
	WENavigationObject * pressedObject;		//!< We only want to navigate "on release" if the same object received the "press".
	WENavigationObject * rootObject;		//!< Root of the navigation tree of object.
	WENavigationObject * currentFocussedObject;	//!< Currently focussed object.
	WENavigationObject * currentParentObject;	//!< Current parent object - needed to decide on which object to "roll-over" even in frame navigation.

	// Need some properties we can ask for from another thread - mainly when executing commands in ISO-VT.
	// Therefore we should store them when we know that accessing the navigation object is safe.
	// Like that one can read them as long as the navigation scheme exists:
	void * navigationBaseObject;			//!< For ISO-VT and PClient this will be a pointer to a DDO
	bool isOpenForInput;				//!< Indicate if focussed object only has focus or is even open for input
							// This information could be incorrect for inplace editing objects maintaining their own state.
};
}
#endif /* WE_NAVIGATION_SCHEME_H_ */
