/**
 *	@file we_base_node.h
 *
 *	@author preths@topcon.com
 *	@author Patrick Reths
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2020
 *
 */

#ifndef WE_BASE_NODE_H
#define WE_BASE_NODE_H

#include <QSGGeometryNode>

namespace WE
{

// Forward declarations:
class WENodeImageSupport;


/**
 *
 *	This class is the base class for all WENodes
 *	It is NOT a QSGNode itself but stores a reference to one.
 *	Classes inheriting this class typically are QSGNodes
 *	e.g. WEBasicGeometryNode is a QSGGeometryNode
 *
 */
class WEBaseNode
{

public:

	virtual ~WEBaseNode() = default;


	/**
	 *	@brief Suspend usage while 'updateNode()' modifies content.
	 *
	 *	@param p_drawNode true: Draw this node, false: Don't draw
	 *
	 *	WEControllerNode::begin() sets this to 'false' for all its controlled nodes.
	 *	WEControllerNode::draw(WEBaseNode *p_node) sets it to 'true' for p_node after it
	 *	repositioned it in node tree/scene graph
	 *
	 *
	 *	node/widget's 'updateNode() updates content and calls --> WENodeUtils::drawNode()
	 *	--> WEControllerNode::draw()
	 *
	 *	@see WEControllerNode::begin()
	 *	@see WEControllerNode::draw(WEBaseNode *p_node)
	 */
	void draw(bool p_drawNode=true);


	/**
	 *	@brief Define if this node should be visible in the sense of visibility variables.
	 *
	 *	@param p_shouldBeVisible true: node should appear on screen, false: hide it
	 *
	 */
	void setVisible(bool p_shouldBeVisible);


	/**
	 *	@brief Get current setting made via setVisible(bool)
	 *
	 *	@return true if node should appear on screen, false if it should be hidden
	 *
	 *	Note that the node could be hidden even if 'true' is returned if its parent
	 *	is hidden.
	 */
	bool shouldBeVisible() const
	{
		return this->shouldBeOnScreen;
	}


	/**
	 *	@brief Implement visibility by adding/removing the node to/from node tree/scene graph.
	 *
	 *	Mainly called in 'end()' method of WEControllerNode after 'updateNode()' of the controlled node.
	 *
	 *	@see WEControllerNode::end()
	 */
	void applyVisibility();


	bool isNodeHidden() const
	{
		return !this->finalVisibility;
	}


	QSGNode & getQSGNode() const
	{
		return this->qsgNode;
	}

	QSGNode * getParent() const
	{
		return this->qsgNode.parent();
	}

	// This seems to be an override from the Qt universe:
	virtual bool isSubtreeBlocked() const = 0;


protected:

    bool onInit{true};
    bool dirty{true};

    // If both following flags are 'true', applyVisibility() marks our subtree as 'dirty' in scenegraph
    bool shouldBeOnScreen{true};	// visibility setting
    bool drawNode{false};		// render pending

    QSGNode & qsgNode;

    explicit WEBaseNode(QSGNode & p_qsgNode);

    WEBaseNode(const WEBaseNode&) = delete;


private:

    // This should be (this->visible && this->drawNode).
    // ("drawNode" is used by the WEControllerNode to realize a QPainter like usage of the nodes.)
    bool finalVisibility{true};

    friend WENodeImageSupport;
};
} //End of namespace WE
#endif // WE_BASE_NODE_H
