/**
 *	@file we_text_node.h
 *
 *	@author preths@topcon.com
 *	@author Patrick Reths
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2020
 *
 */

#ifndef WE_TEXT_NODE_H
#define WE_TEXT_NODE_H

#include <vector>

#include <QSGNode>
#include <QQuickItem>
#include <QTextDocument>
#include <QSGFlatColorMaterial>

#include <private/qsgcontext_p.h>

#include "we_base_node.h"
#include "we_rectangle_node.h"
#include "../../widgets/include/we_widget.h"
#include "../../font/include/we_font.h"
#include "../../common/include/we_rect.h"
#include "../../../common/include/we_color.h"
#include "../../include/we_datatypes_gui.h"

namespace WE
{

class WEControllerNode;

/**
 *
 * This class can be used to draw text with the Qt quick scene graph.
 *
 */
class WETextNode : public QSGNode, public WEBaseNode
{

public:

	explicit WETextNode(QQuickItem *p_ownerElement);
	virtual ~WETextNode();

	void setText(const QString& p_text);
	void setRect(const WERect& p_rect);
	void setAlignment(Qt::Alignment p_alignment);
	void setColor(const WEColor& p_color);
	void setBackgroundColor(const WEColor& p_color);
	void setFont(const WEFont& p_font);
	void setAutoWrapEnabled(bool p_enabled);
	void setWrapOnHyphenEnabled(bool p_enabled);
	void setTrimEnabled(bool p_enabled);
	void apply();

	QString getText() const
	{
		return this->text;
	}

	const WERect& getRect() const
	{
		return this->rect;
	}

	Qt::Alignment getAlignment() const
	{
		return this->alignment;
	}

	const WEColor& getColor() const
	{
		return  this->color;
	}

	const WEColor& getBackgroundColor() const
	{
		return  this->backgroundColor;
	}

	const WEFont& getFont() const
	{
		return this->font;
	}

	bool isAutoWrapEnabled() const
	{
		return this->autoWrap;
	}

	bool isWrapOnHyphenEnabled() const
	{
		return this->wrapOnHyphen;
	}

	bool isTrimEnabled() const
	{
		return this->trim;
	}

	bool isSubtreeBlocked() const override
	{
		return this->isNodeHidden();
	}

	static void updateNode(WETextNode * & p_node, WEControllerNode *p_controller, WEWidget *p_widget, const std::string& p_text, WEEncodeType p_encodeType, const WERect& p_rect, WEAlignment p_alignment, const WEColor& p_color, const WEColor& p_backgroundColor, WEFont *p_font, bool p_autoWrap=false, bool p_wrapOnHyphen=false, bool p_trim=true);
	static void updateNode(WETextNode * & p_node, WEControllerNode *p_controller, WEWidget *p_widget, const std::wstring& p_text, const WERect& p_rect, WEAlignment p_alignment, const WEColor& p_color, const WEColor& p_backgroundColor, WEFont *p_font, bool p_autoWrap=false, bool p_wrapOnHyphen=false, bool p_trim=true);
	static void updateNode(WETextNode * & p_node, WEControllerNode *p_controller, WEWidget *p_widget, const QString& p_text, const WERect& p_rect, WEAlignment p_alignment, const WEColor& p_color, const WEColor& p_backgroundColor, WEFont *p_font, bool p_autoWrap=false, bool p_wrapOnHyphen=false, bool p_trim=true);

	static void initDecoders();
	static void deleteDecoders();

	static const WEColor transparentColor;


private:

	static std::map<std::string,QTextDecoder*> decoders;	// Map of Encoding Scheme name to corresponding QTextDecoder's to be used for converting to unicode (UTF-8) format

	QSGContext *sceneGraphContext;
	QSGRenderContext *renderContext;
	WERectangleNode *backgroundRectangleNode;
	QSGNode *glyphParentNode;
	QQuickItem *ownerElement;
	QTextOption textOption;
	Qt::Alignment alignment;
	QString text;
	QStringList textList;
	std::vector<QTextLayout*> layoutList;
	WERect rect;
	WEColor color;
	WEColor backgroundColor;
	WEFont font;
	const QFont *qFont;
	int fontAscent;
	int fontLineSpacing;
	bool invertedColors;
	bool autoWrap;
	bool wrapOnHyphen;
	bool trim;
	bool dirtyText;


	void applyText();
};
} //End of namespace WE
#endif // WE_TEXT_NODE_H
