/**
 *	@file we_texture_cache.h
 *
 *	@author preths@topcon.com
 *	@author Patrick Reths
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2020
 *
 */

#ifndef WE_TEXTURE_CACHE_H
#define WE_TEXTURE_CACHE_H

#include <vector>
#include <iostream>

#include <QMap>
#include <QSGTexture>

#include "we_widget_texture_cache_key.h"
#include "we_svg_texture_cache_key.h"
#include "we_gif_texture_cache_key.h"
#include "we_meter_texture_cache_key.h"
#include "we_gauge_texture_cache_key.h"
#include "we_arched_bg_texture_cache_key.h"
#include "../../../utility/tick/include/we_tick.h"

class QQuickWindow;

namespace WE
{

namespace WETextureCache
{

// AXIVION Next Line CWE-Bad-Coding-Practices-1126 Generic-LocalScope : could be defined in the only function that uses it but this is a better place...
static constexpr unsigned int textureCacheMinFreeRAM = 200; // in MB (if less RAM is available, all unused textures will be removed)
static constexpr unsigned int textureCacheLowFreeRAM = 250; // in MB (if less RAM is available, the oldest unused textures will be removed)


struct Data
{
	QSGTexture * texture;
	int usageCount;
	unsigned long lastUsage; // WETick of the last usage of this texture
	//		unsigned int sizeInMemory;

	Data(QSGTexture * p_texture, int p_usageCount) : texture(p_texture), usageCount(p_usageCount)
	{
		this->lastUsage = (p_usageCount > 0) ? WETick::getTickCount() : 0;

//		if (nullptr != texture)
//		{
//			QSize s = texture->textureSize();
//			this->sizeInMemory = s.width() * s.height() * 4 + sizeof(QSGTexture) + sizeof(Data);
//		}
//		else
//		{
//			this->sizeInMemory = 0;
//		}
	}

	virtual ~Data()
	{
		if (nullptr != texture)
		{
			delete texture;
		}
	}

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	Data(const Data &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	Data & operator=(const Data &) = delete;

	void use()
	{
		this->usageCount++;
		this->lastUsage = WETick::getTickCount();
		//std::cout << "WETextureCache::Data::use: usageCount: " << this->usageCount << std::endl;
	}

	void release()
	{
		this->usageCount--;
		//std::cout << "WETextureCache::Data::release: usageCount: " << this->usageCount << std::endl;

		if (this->usageCount < 0)
		{
			std::cerr << "WETextureCache::Data::release: usageCount < 0" << std::endl;
		}
	}
};


struct GIFData : Data
{
	std::vector<QSGTexture*> textures;
	std::vector<int> frameDelay;

	GIFData(const std::vector<QSGTexture*>& p_textures, const std::vector<int>& p_frameDelay, int p_usageCount)
	:Data(nullptr, p_usageCount), textures(p_textures), frameDelay(p_frameDelay)
	{
//		this->sizeInMemory = (width * height * 4 + sizeof(QSGTexture) + sizeof(int)) * frameCount + sizeof(GIFData);
	}

	virtual ~GIFData()
	{
		for (QSGTexture *t: textures)
		{
			delete t;
		}
	}

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	GIFData(const GIFData &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	GIFData & operator=(const GIFData &) = delete;
};


template <class K, class V>
class Implementation
{

public:
	Implementation() = default;
	~Implementation() = default;

	V * addTexture(const K & p_key, QSGTexture *p_texture, int p_usageCount=1)
	{
		V *d = new V(p_texture, p_usageCount);
		cache.insert(p_key, d);
		return d;
	}

	V * addGIFTextures(const K & p_key, const std::vector<QSGTexture*>& p_textures, const std::vector<int>& p_frameDelay, int p_usageCount=1)
	{
		V *d = new V(p_textures, p_frameDelay, p_usageCount);
		cache.insert(p_key, d);
		return d;
	}

	V * getCacheData(const K & p_key)
	{
		V *ret = nullptr;

		auto it = cache.find(p_key);
		if (it != cache.end())
		{
			ret = it.value();
		}

		return ret;
	}

	bool contains(const K & p_key) const
	{
		return cache.contains(p_key);
	}

	QSGTexture * useTexture(const K & p_key)
	{
		QSGTexture *ret = nullptr;
		V *d = getCacheData(p_key);

		if (nullptr != d)
		{
			d->use();
			ret = d->texture;
		}

		return ret;
	}

	void releaseTexture(const K & p_key)
	{
		V *d = getCacheData(p_key);

		if (nullptr != d)
		{
			d->release();
		}
	}

	// call only on rendering thread!
	void releaseUnusedTextures()
	{
		auto i = cache.begin();
		while (i != cache.end())
		{
			V *d = i.value();

			if (d->usageCount == 0)
			{
				//std::cout << "WETextureCache::releaseUnusedTextures: delete texture" << std::endl;
				delete d;
				i = cache.erase(i);
			}
			else
			{
				++i;
			}
		}
	}

	// call only on rendering thread!
	void releaseOldestUnusedTexture()
	{
		V * oldestUnusedData{nullptr};
		K oldestUnusedKey{};

		auto i = cache.begin();
		while (i != cache.end())
		{
			V *d = i.value();

			if (d->usageCount == 0)
			{
				if (nullptr == oldestUnusedData || d->lastUsage < oldestUnusedData->lastUsage)
				{
					oldestUnusedData = d;
					oldestUnusedKey = i.key();
				}
			}
			++i;
		}

		if (nullptr != oldestUnusedData)
		{
			//std::cout << "WETextureCache::releaseOldestUnusedTexture: Delete texture" << std::endl;
			delete oldestUnusedData;
			cache.remove(oldestUnusedKey);
		}
	}


private:

	QMap<K, V*> cache;
};


extern Implementation<WESVGTextureCacheKey, Data> svgTextureCache;
extern Implementation<WEGIFTextureCacheKey, GIFData> gifTextureCache;
extern Implementation<WEMeterTextureCacheKey, Data> meterTextureCache;
extern Implementation<WEGaugeTextureCacheKey, Data> gaugeTextureCache;
extern Implementation<WEArchedBGTextureCacheKey, Data> archedBGTextureCache;


void scheduleRenderJob(void (*p_func)(), QQuickWindow *p_window, bool p_scheduleRendering);

// call only on rendering thread!
void releaseAllUnusedTextures();

void releaseAllUnusedTextures(QQuickWindow *p_window, bool p_scheduleRendering=true);

// call only on rendering thread!
void releaseOldestUnusedTextures();

void releaseOldestUnusedTextures(QQuickWindow *p_window, bool p_scheduleRendering=true);
} //End of namespace WETextureCache
} //End of namespace WE
#endif // WE_TEXTURE_CACHE_H
