/**
 *	@file we_archedbargraph_widget.h
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This files defines a archedbargraph widget
 *
 *	@see WEWidget
 *
 *
 */
#ifndef WE_ARCHEDBARGRAPH_WIDGET_H_
#define WE_ARCHEDBARGRAPH_WIDGET_H_


#include <QPainterPath>
#include "we_int_widget.h"


class QPainter;
class QPixmap;

namespace WE
{

class WEImage;
class WELineNode;


class WEArchedbargraphWidget : public WEIntWidget
{
	Q_OBJECT
	Q_PROPERTY(int direction READ getDirection)
	Q_PROPERTY(bool fillMode READ isFillMode)
	Q_PROPERTY(bool drawEnclosingBorder READ isDrawEnclosingBorder)
	Q_PROPERTY(int graphWidth READ getGraphWidth)
	Q_PROPERTY(int lineWidth READ getLineWidth)

public:

	/**
	 *	@fn WEArchedbargraphWidget(WEWidget *parent, void * p_controller = nullptr)
	 *
	 *	@brief constructor of the archedbargraph
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] parent pointer to parent object, if nullptr then top widget
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 *
	 *	@returns nothing
	*/
	explicit WEArchedbargraphWidget(WEWidget *parent, void * p_controller = nullptr);

	/**
	 *	@fn ~WEArchedbargraphWidget()
	 *
	 *	@brief deconstructor
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns nothing
	*/
	virtual ~WEArchedbargraphWidget();


	/**
	 *	@fn WEPainterPath createArchedBargraphPath(
			int p_topLeftX,
			int p_topLeftY,
			int p_width,
			int p_height,
			int p_startAngle,
			int p_endAngle,
			int p_graphWidth,
			bool p_antialiasing)
	 *
	 *	@brief Create a path that is a segment of an arched bargraph.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_x X-position of the top left corner of the enclosing rectangle defining the outer ellipse
	 *	@param p_y Y-position of the top left corner of the enclosing rectangle defining the outer ellipse
	 *	@param p_width Width of the top left corner of the enclosing rectangle defining the outer ellipse
	 *	@param p_height Height of the top left corner of the enclosing rectangle defining the outer ellipse
	 *	@param p_startAngle Angle in degrees, counter-clockwise defining the start of the ellipse segment; 0° points to the right
	 *	@param p_endAngle Angle in degrees, counter-clockwise defining the end of the ellipse segment; 0° points to the right
	 *	@param p_graphWidth Width of the line that will be limited by the outer and the inner ellipse
	 *
	 *	@returns A path describing the arched bargraph.
	 *
	 *
	 *	The path can be used to define a clip region e.g. for the background image or to be drawn directly.
	 *
	 */
	static QPainterPath createArchedBargraphPath(
			int p_x,
			int p_y,
			int p_width,
			int p_height,
			double p_startAngle,
			double p_endAngle,
			int p_graphWidth,
			bool p_antialiasing);

	WERect createEllipseRect(
			int p_x,
			int p_y,
			int p_width,
			int p_height,
			int p_graphWidth);


	/**
	 *	@fn void paint(WEPainter* painter)
	 *
	 *	@brief draws the meter
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *  	@param[in] p_painter pointer to painter object
	 *
	 *  	@returns nothing
	*/
	void paint(WEPainter* p_painter) override;

	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data) override;

	/**
	 *	@fn int setStartAngle(unsigned int p_startAngle)
	 *
	 *	@brief set the start Angle
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_startAngle start angle of the meter
	 *
	 *	@returns WE_ERROR code
	*/
	int setStartAngle(int p_startAngle) override;


	/**
	 *	@fn int setEndAngle(unsigned int p_endAngle)
	 *
	 *	@brief set the end Angle
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_endAngle end angle of the meter
	 *
	 *	@returns WE_ERROR code
	*/
	int setEndAngle(int p_endAngle) override;


	/**
	 *	@fn int setTargetLineColor(WEColor p_targetLineColor)
	 *
	 *	@brief set the target line color
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_targetLineColor color used for drawing the line
	 *
	 *	@returns WE_ERROR code
	*/
	int setTargetLineColor(const WEColor & p_targetLineColor);

	/**
	 *	@fn int setFillMode(bool p_fillMode)
	 *
	 *	@brief set fill mode to true or false
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_fillMode sets if only the value line should be drawn or the ellipse till the value should be filled with color
	 *
	 *	@returns WE_ERROR code
	*/
	int setFillMode(bool p_fillMode);


	/**
	 *	@fn int setDeflectionDirection(WEDeflectionDirectionType p_dir)
	 *
	 *	@brief set the deflection direction type (NO_DEFLECTION = 0, CLOCKWISE,	ANTICLOCKWISE)
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_dir direction
	 *
	 *	@see WEDeflectionDirectionType
	 *
	 */
	int setDeflectionDirection(WEDeflectionDirectionType p_dir);


	/**
	 *	@fn int int setGraphWidth(unsigned int p_width)
	 *
	 *	@brief set the width of the graph
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_width the width of the bargraph
	 *
	 *	@returns WE_ERROR code
	*/
	int setGraphWidth(int p_width);


	/**
	 *	@fn int int setLineWidth(unsigned int p_width)
	 *
	 *	@brief set the width of the line for the value and target value
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_width the width of the line
	 *
	 *	@returns WE_ERROR code
	 *
	 *	@warning the parameter is checked to be <= as a member variables graphWidth and graphHeight
	 *	@warning so the this->graphWidth and this->graphHeight should be set before
	*/
	int setLineWidth(int p_width);





	/**
	 *	@fn setEnclosingBorderColor(WEColor p_enclosingBorderColor)
	 *
	 *	@brief set the Enclosing Border color
	 *
	 *	@author satheeshkumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *  @param[in] p_enclosingBorderColor color used for drawing the Enclosing Border
	 *
	 *  @returns WE_ERROR code
	*/
	int setEnclosingBorderColor(const WEColor & p_enclosingBorderColor);


	/**
	 *	@fn int setDrawEnclosingBorder(bool p_drawEnclosingBorder)
	 *
	 *	@brief set drawing of Enclosing Border to true or false
	 *
	 *	@author satheeshkumar.p@aricent.com
	 *	@author Satheeskumar Palanivel
	 *
	 *  @param[in] p_drawEnclosingBorder sets if the Enclosing Border is drawn or not
	 *
	 *  @returns WE_ERROR code
	*/
	int setDrawEnclosingBorder(bool p_drawEnclosingBorder);

	int getDirection(void) const;
	bool isFillMode(void) const;
	bool isDrawEnclosingBorder(void) const;
	int getGraphWidth(void) const;
	int getRealGraphWidth(void) const;
	int getLineWidth(void) const;
	virtual void repaint() override;

	virtual bool amIHit(int p_globalX, int p_globalY) const override;

	//overwritten from WEIntWidget (to trigger foreground repaint)
	virtual int setValue(int p_value, bool p_updateScrollbar = true) override;
	int setAbsMinMax(int p_min, int p_max) override;
	int setAbsMin(int p_min) override;
	int setAbsMax(int p_max) override;
	int setTargetValue(int p_targetValue) override;
	int setDrawTargetLine(bool p_drawTargetLine) override;
	int setColorMinMax(
		const WEColor & p_minColor,
		const WEColor & p_normalColor,
		const WEColor & p_maxColor) override;
	int setColorNormal(const WEColor & p_normalColor) override;
	int setColorMin(const WEColor & p_minColor) override;
	int setColorMax(const WEColor & p_maxColor) override;
	//overwritten from WEWidget (to trigger foreground repaint)
	int setBorderState(int state) override;
	int setBorderWidth(int p_borderWidth) override;
	void setDrawBorder(bool p_is) override;
	void setValueToValid(bool p_valid) override;

	bool isInputRoot() const override;

protected:
	WEArchedBGTextureCacheKey getBackgroundTextureCacheKey() const;

	virtual void paintBackgroundGraphics(WEPainter* p_painter) override;

	void updateBackgroundGraphicNodes(WEControllerNode * controller);

	void paintForegroundGraphics(WEPainter * p_painter);

	void updateForegroundGraphicNodes(WEControllerNode * controller);

	void repaintForegroundPixmap();

	void createForegroundPixmap();

	void clearForeroundPixmap();

private:

	/**
	 *	@fn void init()
	 *
	 *	@brief initialize the members
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *  	@returns no
	*/
	void init();


	/**
	 *	@fn float calculateValueAngle(int p_value)
	 *
	 *	@brief calculate the angle on the arc for the parameter p_value
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns float angle on the arc, where the line for the value should be drawn
	*/
	//float calculateValueAngle(int p_value);


	/**
	 *	@fn void drawValueLine(float p_angle, QPainter * p_painter, QPainterPath & p_clipRegion)
	 *
	 *	@brief Draw the line at p_anglefrom the center to the outer ellipse representing a value
	 *
	 *	@param[in] p_angle	Angle where line is drawn - should fit to the value to show
	 *	@param[in] p_painter	Painter used for drawing
	 *	@param[in] p_clipRegion	Use this clip region to limit the length of the line
	 *
	 *	@returns nothing
	 *
	 *	When drawing the line the current pen is used and the line is limited by setting p_clipRegion.
	 *	p_painter's clip region is reset to full widget area before returning.
	 */
	void drawValueLine(float p_angle, QPainter * p_painter, QPainterPath & p_clipRegion) const;

	void updateLineNode(WELineNode * & lineNode, WEControllerNode *controller, WEColor color, float p_angle, int p_lineWidth, int p_graphWidth);

	void drawBorder( WEPainter* p_painter );

	void updateBorderNodes(WEControllerNode *controller);


	WEDeflectionDirectionType direction; 	//!< stores current direction of the needle.

	int graphWidth;			//!< should be <= width / 2
	int lineWidth;			//!< the width of the lines for the value and target value

	bool fillMode;			//!< Should the graph be filled (true/false)
	bool drawEnclosingBorder; //!< Draw Enclosing Border true/false

	WE::WEColor enclosingBorderColor; //!< color of the Enclosing Border
	WEColor targetLineColor;	//!< The color of the target line

	QPainterPath fullBargraphDrawPath;
	QPainterPath bargraphFill;
	QRegion paintingRegion;

	QTWEPainter *foregroundPainter;	//!< painter that draws the foreground graphics onto a pixmap
	QPixmap *foregroundPixmap;	//!< The pixmap with the foreground graphics
	WEImage *foregroundPixmapImage;	//!< The image with the foreground pixmap
	bool updateForegroundPixmap;	//!< Flag that indicates that the foreground pixmap needs to be repainted

	bool fillChanged;	//!< Flag that indicates that the painter path for the foreground (fill) of the bargraph needs to be re-calculated
	bool backgroundPathChanged; //!< Flag that indicates that the painter path for the background of the bargraph needs to be re-calculated

	WEImageNode * backgroundPixmapImageNode;	//!<Node of the Widget which is used to draw the background pixmap
	WEImageNode * foregroundPixmapImageNode;	//!<Node of the Widget which is used to draw the foreground pixmap
	WEEllipseNode * backgroundEllipseNode;		//!<Node of the Widget which is used to draw the background ellipse
	WEEllipseNode * backgroundImageEllipseNode;	//!<Node of the Widget which is used to draw the background ellipse
	WEEllipseNode * borderImageEllipseNode;		//!<Node of the Widget which is used to draw the border ellipse
	WEEllipseNode * valueEllipseNode;		//!<Node of the Widget which is used to draw the value ellipse
	WELineNode * valueLineNode;			//!<Node of the Widget which is used to draw the value line
	WELineNode * targetLineNode;			//!<Node of the Widget which is used to draw the target line

	static const bool USE_BACKGROUND_PIXMAP = false; //!< Set to true to use a pixmap to draw the background.
	static const bool USE_FOREGROUND_PIXMAP = false; //!< Set to true to "optimize" foreground painting with a pixmap. When having many value changes, this costs MORE cpu!
};
}
#endif /* WE_ARCHEDBARGRAPH_WIDGET_H_ */
