/**
 *	@file we_container.h
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file contains the definition of the WEContainer.
 *
 *	The WEContainer is used to contain other widgets. Widgets like WENumericWidget,
 *	WEStringWidget etc can be added to the WEContainer, WEContainer is a top
 *	level widget.
 *	WEContainer can be configured to show a video picture using WEAnalogVideo.
 */

#ifndef WE_CONTAINER_H_
#define WE_CONTAINER_H_

#include "we_widget.h"
#include "../../../common/include/we_rotation.h"
#include "../../../common/include/we_size.h"

namespace WE
{

class WEImageWidget;
class WEPainter;

/**
 *	@class WEContainer
 *
 *	@author maneesh.arora@aricent.com
 *	@author Maneesh Arora
 *
 *	@brief The Container widget is used to contain other widgets. Widgets like WENumericWidget,
 *	WEStringWidget etc can be added to the WEContainer at the specified location. A widget
 *	can only have a single parent container at a time. The WEContainer may also be added as
 *	a child to other containers.
 *
 *	@namespace WE
 *
 */
class WEContainer : public WEWidget
{
	Q_OBJECT
public:
	/**
	 *	@fn WEContainer(WEWidget * p_parent = nullptr, void * p_controller = nullptr, int p_typeID = WIDGET_TYPE_ID_CONTAINER)
	 *
	 *	@brief Default constructor.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_cameraNumber - 0 for "no camera"
	 *	@param[in] p_parent - pointer to parent widget to create a tree of widgets
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 *
	 */
	WEContainer(WEWidget * p_parent = nullptr, void * p_controller = nullptr, int p_typeID = WIDGET_TYPE_ID_CONTAINER);

	/**
	 *	@fn ~WEContainer()
	 *
	 *	@brief Destructor of the WEContainer.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 */
	virtual ~WEContainer();


	/**
	 *	@fn init()
	 *
	 *	@brief This method is used to initialise the container by setting the
	 *	default values of its' properties (like Name, Background Colour).
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 */
	void init();


	void showAlbumart(WEImageWidget * p_art);

private:


	/**
	 *	@fn paint(WEPainter *painter)
	 *
	 *	@brief This is the paint method of the Container. This method is
	 *	overridden from the WEWidget. The method uses it's properties
	 *	to define the look and feel of the Container. Please refer to
	 *	the WEWidget paint() for more details.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param p_painter Reference to the Graphical Context which is used for drawing
	 *	the container. Please see WEPainter for further details.
	 *
	 *	@see WEWidget paint()
 	 *	@see WEPainter
 	 *
	 *	@warning Do not use repaint() with in paint which can cause	recursion.
	 */
	void paint(WEPainter *p_painter) override;

	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData *p_data) override;


	int drawBorderForChildren(WEPainter * p_painter);

	WEImageWidget * albumArt;
};
}	// End of Namespace WE
#endif /* WE_CONTAINER_H_ */
