/**
 *	@file we_gauge_widget.h
 *
 *	@author Alexander Faust
 *
*/

#ifndef WE_GAUGE_WIDGET_H_
#define WE_GAUGE_WIDGET_H_

#include "we_meter_widget.h"
#include "we_text_widget.h"
#include "../../../common/include/we_error.h"

namespace WE
{

class WERectangleNode;
class WETextNode;

class WEGaugeWidget : public WEMeterWidget, public WETextWidget
{

	Q_OBJECT
	Q_PROPERTY(bool drawMinorTicks READ getDrawMinorTicks)
	Q_PROPERTY(unsigned int numberOfMinorTicks READ getNumberOfMinorTicks)
	Q_PROPERTY(double minorTickHeightFactor READ getMinorTickHeightFactor)
	Q_PROPERTY(double minorTickWidthFactor READ getMinorTickWidthFactor)
	Q_PROPERTY(bool drawValues READ getDrawValues)
	Q_PROPERTY(double valuesDistanceFactor READ getValuesDistanceFactor)
	Q_PROPERTY(bool fontAutoScaling READ getFontAutoScaling)
	Q_PROPERTY(double fontScalingFactor READ getFontScalingFactor)
	Q_PROPERTY(bool drawBoundaryBar READ getDrawBoundaryBar)
	Q_PROPERTY(double boundaryBarWidthFactor READ getBoundaryBarWidthFactor)
	Q_PROPERTY(bool drawDigitalValue READ getDrawDigitalValue)
	Q_PROPERTY(double digitalValueFontScalingFactor READ getDigitalValueFontScalingFactor)
	Q_PROPERTY(double digitalValueBackgroundXOffsetFactor READ getDigitalValueBackgroundXOffsetFactor)
	Q_PROPERTY(double digitalValueBackgroundYOffsetFactor READ getDigitalValueBackgroundYOffsetFactor)
	Q_PROPERTY(double digitalValueBackgroundWidthFactor READ getDigitalValueBackgroundWidthFactor)
	Q_PROPERTY(double digitalValueBackgroundHeightFactor READ getDigitalValueBackgroundHeightFactor)

public:

	/**
	 *	@fn WEGaugeWidget(WEWidget *parent, void * p_controller = nullptr)
	 *
	 *	@brief constructor of the gauge widget
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@param[in] parent pointer to parent object, if nullptr then top widget
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 *
	 *	@returns nothing
	*/
	explicit WEGaugeWidget(WEWidget *parent, void * p_controller = nullptr);


	/**
	 *	@fn ~WEGaugeWidget()
	 *
	 *	@brief deconstructor of the gauge widget
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  @returns nothing
	*/
	virtual ~WEGaugeWidget();


	/**
	 *	@fn void paint(WEPainter* painter)
	 *
	 *	@brief draws the gauge widget
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *  @param[in] painter pointer to painter object
	 *
	 *  @returns nothing
	*/
	void paint(WEPainter* painter) override;


	int setColorBetween(WEColor colorBetween);


	int setDrawMinorTicks(bool drawMinorTicks);


	int setMinorTickColor(WEColor minorTickColor);


	int setNumberOfMinorTicks(std::uint16_t numberOfMinorTicks);


	int setMinorTickHeightFactor(double minorTickHeightFactor);


	int setMinorTickWidthFactor(double minorTickWidthFactor);


	int setDrawValues(bool drawValues);


	int setValuesDistanceFactor(double valuesDistanceFactor);


	int setFontAutoScaling(bool fontAutoScaling);


	int setFontScalingFactor(double fontScalingFactor);


	int setDrawBoundaryBar(bool drawBoundaryBar);


	int setBoundaryBarWidthFactor(double boundaryBarWidthFactor);


	int setDrawDigitalValue(bool drawDigitalValue);


	int setDigitalValueFontColor(WEColor digitalValueFontColor);


	int setDigitalValueBackgroundColor(WEColor digitalValueBackgroundColor);


	int setDigitalValueFontScalingFactor(double digitalValueFontScalingFactor);


	int setDigitalValueBackgroundXOffsetFactor(double digitalValueBackgroundXOffsetFactor);


	int setDigitalValueBackgroundYOffsetFactor(double digitalValueBackgroundYOffsetFactor);


	int setDigitalValueBackgroundWidthFactor(double digitalValueBackgroundWidthFactor);


	int setDigitalValueBackgroundHeightFactor(double digitalValueBackgroundHeightFactor);


	WEColor getColorBetween(void) const;


	bool getDrawMinorTicks(void) const;


	WEColor getMinorTickColor(void) const;


	unsigned short getNumberOfMinorTicks(void) const;


	double getMinorTickHeightFactor(void) const;


	double getMinorTickWidthFactor(void) const;


	bool getDrawValues(void) const;


	double getValuesDistanceFactor(void) const;


	bool getFontAutoScaling(void) const;


	double getFontScalingFactor(void) const;


	bool getDrawBoundaryBar(void) const;


	double getBoundaryBarWidthFactor(void) const;


	bool getDrawDigitalValue(void) const;


	WEColor getDigitalValueFontColor(void) const;


	WEColor getDigitalValueBackgroundColor(void) const;


	double getDigitalValueFontScalingFactor(void) const;


	double getDigitalValueBackgroundXOffsetFactor(void) const;


	double getDigitalValueBackgroundYOffsetFactor(void) const;


	double getDigitalValueBackgroundWidthFactor(void) const;


	double getDigitalValueBackgroundHeightFactor(void) const;


	//overwritten from WEIntWidget because it triggers background repaint in gauge:
	int setAbsMinMax(int p_min, int p_max) override;


	int setAbsMin(int p_min) override;


	int setAbsMax(int p_max) override;


	int setMinMax(int p_min, int p_max) override;


	int setMin(int p_min) override;


	int setMax(int p_max) override;


	int setColorMin(const WEColor & p_minColor) override;


	int setColorMax(const WEColor & p_maxColor) override;


	int setOffset(double p_offset) override;


	int setOffset2(double p_offset2) override;


	int setScale(double p_scale) override;


	int setNumberOfDecimals(int p_noOfDecimals) override;


protected:

	/**
	 * Overridden because we use a different WETextureCache type than the meter widget
	 */
	virtual void createBackgroundPixmap() override;


	WEGaugeTextureCacheKey getBackgroundTextureCacheKey() const;


	/**
	 *	@fn paintBackgroundGraphics(WEPainter* painter)
	 *
	 *	@brief draws static background graphics on the gauge widget (background, ticks, tick values, ...)
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@param[in] painter pointer to painter object
	 *
	 *	@returns nothing
	 */
	void paintBackgroundGraphics(WEPainter* painter) override;


	/**
	 *	@fn void paintForegroundGraphics(WEPainter* painter)
	 *
	 *	@brief draws high frequently updated foreground graphics on the gauge widget (e.g. needle, digital value, ...)
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@param[in] painter pointer to painter object
	 *
	 *	@returns nothing
	 */
	void paintForegroundGraphics(WEPainter* painter) override;


	void updateForegroundGraphicsNodes(WEControllerNode *controller) override;


	/**
	 *	@fn void drawGaugeValues(int value, WEPainter* painter)
	 *
	 *	@brief draws the values for every major tick
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@param[in] value value to drawn on Gauge
	 *	@param[in] painter pointer to painter object
	 *
	 *	@returns nothing
	 */
	void drawGaugeValues(double value, WEPainter* painter);


	/**
	 *	@fn void drawGaugeMinorTicks(int value, double stepWidth, WEPainter* painter)
	 *
	 *	@brief draws the minor ticks
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@param[in] value value for start the drawing of minor ticks
	 *	@param[in] stepWidth represents the step width to the next major tick.
	 *	@param[in] painter pointer to painter object
	 *
	 *	@returns nothing
	 */
	void drawGaugeMinorTicks(double value, double stepWidth, WEPainter* painter);


	/**
	 *	@fn void drawGaugeBoundaryBar(WEPainter* painter)
	 *
	 *	@brief draws the different boundary bars (below min, between, above ,max)
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@param[in] painter pointer to painter object
	 *
	 *	@returns nothing
	 */
	void drawGaugeBoundaryBar(WEPainter* painter);


	/**
	 *	@fn void drawGaugeDigitalValueBackground(WEPainter* painter)
	 *
	 *	@brief draws the background of the digital value
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] painter pointer to painter object
	 *
	 *	@returns nothing
	 */
	void drawGaugeDigitalValueBackground(WEPainter* painter);


	/**
	 *	@fn void drawGaugeDigitalValue(WEPainter* painter)
	 *
	 *	@brief draws the digital value with an label on the gauge. The value represents the preview value.
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@param[in] painter pointer to painter object
	 *
	 *	@returns nothing
	 */
	void drawGaugeDigitalValue(WEPainter* painter);


	void updateGaugeDigitalValueNodes(WEControllerNode *controller);


	/**
	 *	@fn void createDigitalValueTextRect()
	 *
	 *	@brief Creates the rectangle which is used to render the digital value on the gauge.
	 *	       Don't call this directly. Instead set updateDigitalValueTextRect to true so updateGaugeDigitalValueNodes will call this method.
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@returns nothing
	 */
	void createDigitalValueTextRect();


	/**
	 *	@fn double convertValueToAngle(double value)
	 *
	 *	@brief converts a given value to an angle under considering the current other properties like "Deflection Direction, ..."
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@param[in] value value to convert
	 *
	 *	@returns Angle of the value
	 */
	double convertValueToAngle(double value) const;


private:

	/**
	 *	@fn void init()
	 *
	 *	@brief initialize the members
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 * 	@returns void
	 */
	void init();


	WE::WEColor colorBetween;
	bool drawMinorTicks;
	WE::WEColor minorTickColor;
	unsigned int numberOfMinorTicks;
	double minorTickHeightFactor;
	double minorTickWidthFactor;
	bool drawValues;
	double valuesDistanceFactor;
	bool fontAutoScaling;
	double fontScalingFactor;
	bool drawBoundaryBar;
	double boundaryBarWidthFactor;
	bool drawDigitalValue;
	WE::WEFont * digitalValueFont;
	WE::WEColor digitalValueFontColor;
	WE::WEColor digitalValueBackgroundColor;
	double digitalValueFontScalingFactor;
	double digitalValueBackgroundXOffsetFactor;
	double digitalValueBackgroundYOffsetFactor;
	double digitalValueBackgroundWidthFactor;
	double digitalValueBackgroundHeightFactor;
	WERect digitalValueTextRect;
	bool updateDigitalValueTextRect;

	WERectangleNode * digitalValueBackgroundNode;
	WETextNode * digitalValueTextNode;
};
}
#endif /* WE_GAUGE_WIDGET_H_ */
