/**
 *	@file we_line_widget.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This files defines a widget for the drawing a line
 *	and provides interfaces to configure the WELineWidget.
 *
 *	@see we_line_widget.cpp, WEWidget
 *
 */

#ifndef WE_LINE_WIDGET_H_
#define WE_LINE_WIDGET_H_

#include "we_widget.h"

namespace WE
{


/**
 *	@class WELineWidget
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@brief This widget provides line display.
 *
 *	@namespace WE
 *
 *	This class includes the functions, which allows
 *	the user to display the specified Lines at the given location. It inherits
 *	from WEWidget, which inherits from WEPaintListener to draw its line in the paint() event.
 */
class WELineWidget : public WEWidget
{
	Q_OBJECT
	Q_PROPERTY(int lineDirection READ getLineDirection)

public:
	/**
	 *	@fn WELineWidget(WEWidget *p_parent, void * p_controller = nullptr)
	 *
	 *	@brief Constructs a WELineWidget Widget.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] parent If parent IS nullptr then this widget becomes a window with the same properties.
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 *
	 *	This widget can be created as a separate window by specifying the
	 *	parent as nullptr or it can be included as a child widget of the parent.
	 */
	explicit WELineWidget(WEWidget* p_parent, void * p_controller = nullptr);

	/**
	 *	@fn ~WELineWidget()
	 *
	 *	@brief Destructor of the widget.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 */
	virtual ~WELineWidget() = default;

	/**
	 *	@fn init()
	 *
	 *	@brief the method to initialize the member variable of the class
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 */
	void init();

	/**
	 *	@fn paint(WEPainter* painter)
	 *
	 *	@brief The drawing for this widget is performed in this method.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param p_painter pointer to the corresponding WEPainter
	 *
	 *	@see WEPainter
	 *	@see repaint()
	 *
	 *	@warning Only one instance of the WEPainter can be created for a Widget.
	 *	@warning Do not invoke this method directly, call repaint() to redraw
	 *	the contents of the widget.
	 *
	 */
	void paint(WEPainter* p_painter) override;

	LineDirection getLineDirection() const;
	void setLineDirection(LineDirection lineDirection);

private:
	LineDirection lineDirection;
};
}
#endif /* WE_LINE_WIDGET_H_ */
