/**
 *	@file we_scaled_container.h
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2017
 *
 */

#ifndef WE_SCALED_CONTAINER_H_
#define WE_SCALED_CONTAINER_H_

#include <QMouseEvent>

#include "we_widget.h"
#include "../../../common/include/we_size.h"



namespace WE
{

class WEPainter;

/**
 *	@class WEScaledContainer
 *
 *	Scaled container internally draws a widget and its children to a QPixmap.
 *	The QPixmap has the size of the original widget and gets scaled to the
 *	size of this widget while this one is painted.
 *
 *	@namespace WE
 *
 */
class WEScaledContainer : public WEWidget
{
	Q_OBJECT
public:
	/**
	 *
	 *	@brief Default constructor.
	 *
	 *	@param[in] p_parent - pointer to parent widget to create a tree of widgets
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController()
	 *
	 */
	WEScaledContainer(WEWidget * p_parent = nullptr, void * p_controller = nullptr, int p_typeID = WIDGET_TYPE_ID_SCALED_CONTAINER);


	/**
	 *	@brief Destructor of the WEScaledContainer.
	 *
	*/
	virtual ~WEScaledContainer() = default;


	/**
	 *	@brief Define which widget should be shown and scaled.
	 *
	 *	@param p_unscaledWidget Pointer to a widget, e.g. a page
	 *	@param p_clear Clear the internal image if true
	 *
	 *	@returns An error code, WE_SUCCESS in case of success
	 *
	 *	This method was introduced to support 'miniview' of a page.
	 *
	 *	This scaled widget will NOT accept any input to its children.
	 *
	 */
	int scaleFrom(WEWidget * p_unscaledWidget, bool p_clear=true);


	/**
	 *	@fn void mousePressEvent(QMouseEvent* p_qtMouseEvent)
	 *
	 *	@brief Qt Mouse Press Event
	 *
	 *	@param p_qtMouseEvent The event data
	 *
	 *
	 *	Try to forward event to the right (fullsize) widget.
	 *
	 *	@warning processMouseEvents() can prevent creation of mouse events.
	 *
	 *	@see processMouseEvents(bool)
	 */
	virtual void mousePressEvent(QMouseEvent * p_qtMouseEvent) override;


	/**
	 *	@fn void mouseReleaseEvent(QMouseEvent* p_qtMouseEvent)
	 *
	 *	@brief Qt Mouse Release Event
	 *
	 *	@param p_qtMouseEvent The event data
	 *
	 *
	 *	Try to forward event to the right (fullsize) widget.
	 *
	 *	@warning processMouseEvents() can prevent creation of mouse events.
	 *
	 *	@see processMouseEvents(bool)
	 */
	virtual void mouseReleaseEvent(QMouseEvent * p_qtMouseEvent) override;


	/**
	 *	@fn void mouseMoveEvent(QMouseEvent* p_qtMouseEvent)
	 *
	 *	@brief Qt Mouse Move Event
	 *
	 *	@param p_qtMouseEvent The event data
	 *
	 *
	 *	Try to forward event to the right (fullsize) widget.
	 *
	 *	@warning processMouseEvents() can prevent creation of mouse events.
	 *
	 *	@see processMouseEvents(bool)
	 */
	virtual void mouseMoveEvent(QMouseEvent * p_qtMouseEvent) override;


private:

	/**
	 *
	 *	@brief This method is used to initialise the container by setting the
	 *	default values .
	 *
	 */
	void init();


	/**
	 *	@fn paint(WEPainter *painter)
	 *
	 *	@brief This is the paint method of the Container. This method is
	 *	overridden from the WEWidget. The method uses it's properties
	 *	to define the look and feel of the Container. Please refer to
	 *	the WEWidget paint() for more details.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param painter Reference to the Graphical Context which is used for drawing
	 *	the container. Please see WEPainter for further details.
	 *
	 *	@see WEWidget paint()
 	 *	@see WEPainter
 	 *
 	 */
	void paint(WEPainter *p_painter) override;

	/**
	 *	Try to find a child of this->widgetToScale to receive the mouse event
	 *
	 *	@param p_qtMouseEvent Event as provided by Qt
	 *
	 *	@returns WEWidget which would get the event in full size view.
	 *
	 */
	WEWidget * getWidgetToBeTouched(QMouseEvent * p_qtMouseEvent) const;

	WEWidget * widgetToScale;

	int unscaledWidth;
	int unscaledHeight;

	QImage scaledImage;
};
}	// End of Namespace WE
#endif /* WE_SCALED_CONTAINER_H_ */
