/**
 *	@file we_scrollbar_widget.h
 *
 *	@author dsterzel@topcon.com
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2016
 *
 *	@brief WEScrollbarWidget class definition
 */

#ifndef SRC_GUI_WIDGETS_INCLUDE_WE_SCROLLBAR_WIDGET_H_
#define SRC_GUI_WIDGETS_INCLUDE_WE_SCROLLBAR_WIDGET_H_

#include "../qt/qt_quick_we_scrollbar.h"
#include "we_int_widget.h"
#include "we_button_widget.h"
#include "we_string_widget.h"

namespace WE {

class WENavigationEventListener;

enum
{
	SCROLLBAR_AUTO = 0,
	SCROLLBAR_ON,
	SCROLLBAR_OFF,
	SCROLLBAR_MAX_VALUE
};

/**
 * This class is a middle layer between int widget and widgets that want to use a scroll bar for editing.
 */
class WEScrollbarWidget : public WEIntWidget, public WEToolkitEventListener
{
	Q_OBJECT
	Q_PROPERTY(bool useTouchInputMethod READ getUseTouchInput)
	Q_PROPERTY(int originalWidth READ getOriginalWidth)
	Q_PROPERTY(unsigned  widthEditMode READ getWidthEditMode)

public:
	WEScrollbarWidget(WEWidget * p_parent, void * p_controller, int p_typeID);
	virtual ~WEScrollbarWidget();

	int setUseTouchInput(bool p_touchIn);
	bool getUseTouchInput() const;

	/*
	 *	@brief Overwrite method of WEWidget to set our own 'shouldBeTransparent' flag.
	 */
	virtual int setTransparent(bool p_isTransparent) override;

	/**
	 *	@fn int setScrollBarMode(unsigned int p_scrollBarMode)
	 *
	 *	@brief Switch scroll bar mode to on/off/auto.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_scrollBarMode SCROLLBAR_AUTO / SCROLLBAR_ON / SCROLLBAR_OFF
	 *
	 *	@returns WE_ERROR code - WE::WE_SUCCESS in case of success.
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if (SCROLLBAR_MAX_VALUE <= p_scrollBarMode)
	 *
	 *	SCROLLBAR_AUTO:	Scrollbar only appears for input list which is open for input
	 *	SCROLLBAR_ON:	Scrollbar is always shown
	 *	SCROLLBAR_OFF:	Scrollbar is never shown
	 *
	 *	If setting to "on" or "off" the scroll bar is also shown/hidden by this method.
	 *
	 *	@see int showScrollBar(bool p_show)
	 */
	int setScrollBarMode(unsigned int p_scrollBarMode);

	/**
	 *	@fn int showScrollBar(bool p_show)
	 *
	 *	@brief Switch scroll bar on or off.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_show "true" to show, "false" to hide scrollbar
	 *
	 *	@returns WE_ERROR code - WE::WE_SUCCESS in case of success.
	 *
	 *	This method only shows the scrollbar if (WEWidget::scrollBarWidth > 0).
	 *
	 *	It is used by method "setMode to show and hide scrollbar for input list automatically.
	 *
	 *	As public method it can additionally be used in order to show the scrollbar for output lists.
	 *
	 *	@see int setMode(WEListModeType p_mode)
	 */
	int showScrollBar(bool p_show);

	void setScrollListener(WE::WENavigationEventListener * p_l)
	{
		this->scrollListener = p_l;
	}


	//documentation see WEToolkitWidget::hideScrollBar
	void hideScrollBar( void );
	//documentation see WEToolkitWidget::showScrollBar
	void showScrollBar(int p_x, int p_y, int p_width, int p_height, int p_initialValue, int p_numValues, bool p_showButtons);

	/**
	 *	@fn paint(WEPainter* p_painter)
	 *
	 *	@brief paints the scrollbar
	 *
	 *	@param[in] p_painter Pointer to we painter object
	 *
	 *	@returns void
	 */
	void paint(WEPainter* p_painter) override;

	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData *p_data) override;

	//documentation see WEToolkitWidget::ssetScrollBarValue
	void setScrollbarValue(int p_value, bool p_ignoreTrigger = false);

	virtual void mousePressEvent(QMouseEvent* p_qtMouseEvent) override;

	virtual void mouseReleaseEvent(QMouseEvent* p_qtMouseEvent) override;

	virtual void mouseMoveEvent(QMouseEvent* p_qtMouseEvent) override;

	int toolkitEventOccurred(WEToolkitEvent & p_event) override;

	void scrollValueChanged(unsigned int p_newIndex);

	int getOriginalWidth(void) const
	{
		return this->originalWidth;
	}

	int getWidthEditMode(void) const
	{
		return this->widthEditMode;
	}


	//Implement the following functions in subclasses:
	/**
	 * Change the size of the widget based on the given new width
	 * @param p_width New width of the widget
	 */
	virtual void updateSizeWithWidth(int p_width) = 0;

	/**
	 * Get the number of visible items i.e. all items known minus
	 * those currently invalid or defined invisible.
	 * This is unrelated to widget size it only refers to data.
	 * @return the value range of of the scroll bar.
	 */
	virtual unsigned int getNumberOfVisibleItems(void) const = 0;

	/**
	 * Get mapping between value of scrollbar and widget value
	 * @param p_index scrollbar value
	 * @return value of widget
	 */
	virtual unsigned int getItemAtIndex(int p_index) const = 0;

	/**
	 * Get scrollbar value at pixel position.
	 * Only important if ignoreMouse is false (e.g. if you can select a item in a list widget)
	 * @param p_x x position
	 * @param p_y y position
	 * @return scroll bar value at given position
	 */
	virtual int getIndexUnderPosition(int p_x, int p_y) const = 0;

	/**
	 * Get the scroll bar value for a given widget value
	 * @param p_value widget value
	 * @return scroll bar value
	 */
	virtual unsigned int getIndexForValue(unsigned int p_value) const = 0;

	public slots:

		/**
		 *	@fn void scrollValueChanged(void)
		 *
		 *	@brief Called when scroll bar slider was moved
		 *
		 *	@author dsl@wachendorff.de
		 *	@author Dirk Sterzel
		 *
		 *	@warning Currently only works for WEListWidget!
		 *
		 *	Called when user changes scrollbar value so that the widget can
		 *	forward the new value to the DDO
		 */
		void scrollValueChanged(void);

		/**
		 * Slot that is connected to enterButton's clicked() signal.
		 * Checks if the widget is a list widget and calls it's enterClicked() method to confirm input.
		 */
		void enterClicked(void);

		/**
		 * Slot that is connected to cancelButton's clicked() signal.
		 * Checks if the widget is a list widget and calls it's cancelClicked() method to confirm input.
		 */
		void cancelClicked(void);



	protected:
		unsigned int	scrollBarMode;	//!< SCROLLBAR_AUTO / SCROLLBAR_ON / SCROLLBAR_OFF
		int		originalWidth;	//!< Since the width of the widget could change due to the scroll bar, this is where we keep the original width of the widget.
		int		widthEditMode;	//!< The widget should be this much wide (excluding the scroll bar) in the edit mode. This is mainly required to show aux object where we need to expand the list to include the aux pool and aux object icon in the list.
		bool		useTouchInputMethod;	//!< Improve usability by adding 'OK' and 'Cancel' buttons to the list and invert encoder navigation
		bool		shouldBeTransparent;	//!< useTouchInputMethod renders list opaque while editing - remember what it should be in normal mode.
		QtQuickWEScrollBar * scrollBar;
		bool		ignoreMouse;


	private:
		WE::WENavigationEventListener * scrollListener;

		WEButtonWidget * enterButtonWidget;
		WEStringWidget * enterStringWidget;
		WEFont * enterStringWidgetFont;

		WEButtonWidget * cancelButtonWidget;
		WEStringWidget * cancelStringWidget;
		WEFont * cancelStringWidgetFont;


		bool nextEventIsNotTriggeredByScrollbarItself;
};
}  // namespace WE

#endif /* SRC_GUI_WIDGETS_INCLUDE_WE_SCROLLBAR_WIDGET_H_ */
