/**
 *	@file we_table_widget.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 */


#ifndef WE_TABLE_WIDGET_H_
#define WE_TABLE_WIDGET_H_

#include <vector>
#include<map>

#include "we_widget.h"
#include "we_scrollbar_widget.h"
#include "we_text_widget.h"

namespace WE {

class WETableWidget : public WEScrollbarWidget, public WETextWidget
{
	Q_OBJECT
	Q_PROPERTY(int highlightedRow READ getHighlightedRow)
	Q_PROPERTY(int numberOfColumns READ getNumberOfColumns)
	Q_PROPERTY(bool headingUsed READ isHeadingUsed)
	Q_PROPERTY(QString heading READ getHeadingAsQString)
	Q_PROPERTY(char delimiter READ getDelimiter)

public:
	/**
	 *	@fn WETableWidget(WEWidget *p_parent, void * p_controller = nullptr)
	 *
	 *	@brief Constructor
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_parent The parent widget of this table widget
	 *	@param p_controller Used in WEApp applications (AppWidgetController)
	*/
	explicit WETableWidget(WEWidget *p_parent, void * p_controller = nullptr);

	/**
	 *	@fn ~WETableWidget()
	 *
	 *	@brief Destructor. Free memory (of container members)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	*/
	virtual ~WETableWidget();


	/**
	 *	@fn void paint(WEPainter * p_painter)
	 *
	 *	@brief Paint function. Draw borders, heading, content and scrollbar
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_painter WEPainter used for drawing
	 *
	 *	The row height is determined by the set font size.
	 *	To make sure that the contents will fit in without touching a border,
	 *	the row height is font size + 2.
	 *	(if font size is 12, the row height is 14)
	*/
	void paint(WEPainter * p_painter) override;


	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData *p_data) override;


	/**
	 *	@fn void setNumberOfColumns(int p_columns)
	 *
	 *	@brief Set the number of columns of this table
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_columns Number of columns
	 *
	 *	Note: Number of rows is determined by the numbers of data sets (see setData())
	*/
	void setNumberOfColumns(int p_columns);

	/**
	 *	@fn void setData(std::vector<std::string> & p_data)
	 *
	 *	@brief Set the content data that shall be displayed
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_data A vector with strings representing the data that shall be shown.
	 *		Each string contains the row-data. The columns of the row data has to be
	 *		delimited by the defined delimiter character.
	*/
	void setData(const std::vector<std::string> & p_data);

	/**
	 *	@fn void setDelimiter(char p_delimiter)
	 *
	 *	@brief Set the delimiting character
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_delimiter. Every column of the table is delimited by this character
	 *
	 *	When processing the row-data set with setData(), each row string is checked for
	 *	this delimiting character. The row-string gets cut into peaces with which the
	 *	single columns are then filled.
	*/
	void setDelimiter(char p_delimiter);

	/**
	 *	@fn void setColumnSpacing(int p_column, int p_sizeInPercent)
	 *
	 *	@brief Set the width for each table column.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_column Number of the column for which the spacing shall be set.
	 *	@param p_sizeInPercent The with of the column in percent
	*/
	void setColumnSpacing(int p_column, int p_sizeInPercent);

	/**
	 *	@fn void setHeadingColors(WEColor & p_bg, WEColor & p_font)
	 *
	 *	@brief Set the background and font color for the heading row
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_bg Background color for the heading row.
	 *	@param p_font Font color for the heading row
	 *
	 *	@see useHeading()
	 *
	 *	The header is not drawn by default.
	*/
	void setHeadingColors(const WEColor & p_bg, const WEColor & p_font);

	/**
	 *	@fn void setHeading(std::string p_heading)
	 *
	 *	@brief Set the text for the heading of the table
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_heading String containing a string delimited with the delimiter character
	 *		The format needs to be the same as the rows set with setData()
	*/
	void setHeading(const std::string & p_heading);

	/**
	 *	@fn void useHeading(bool p_yes)
	 *
	 *	@brief Shall the heading be drawn or not.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_yes True if heading shall be drawn, false if not.
	*/
	void useHeading(bool p_yes);

	/**
	 * Use this function to mark a row as highlighted. The row will be highlighted in the following way:
	 * - Font color is inverted
	 * - The background of each cell is painted in font color (with 50% translucency)
	 * Along with the highlighting this function also calculates a suggested value that should be set to the
	 * table widget after highlighting a row (e.g. if the highlighted row is outside of the visible are the value
	 * of the list should be changed to scroll the list to the highlighted row)
	 * @param p_rowNumber The row number that shall be highlighted. Count begins at 0. Set negative value to remove highlight.
	 * @return suggested value that should be set to the table to make the table scroll to the highlighted row
	 */
	int setHighlightedRow(int p_rowNumber);

	int getHighlightedRow(void) const
	{
		return this->highlightedRow;
	}

	int getNumberOfColumns(void) const
	{
		return this->numberOfColumns;
	}

	bool isHeadingUsed(void) const
	{
		return this->headingUsed;
	}

	QString getHeadingAsQString(void) const
	{
		return QString::fromStdString(this->heading);
	}

	char getDelimiter(void) const
	{
		return this->delimiter;
	}

	void updateSizeWithWidth(int p_width) override;
	unsigned int getNumberOfVisibleItems(void) const override;
	unsigned int getItemAtIndex(int p_index) const override;
	int getIndexUnderPosition(int p_x, int p_y) const override;
	unsigned int getIndexForValue(unsigned int p_value) const override;

	virtual int setBorderState(int p_state) override;

	int setValue(int p_value, bool p_updateScrollbar = true) override;


	/**
	 *	@fn setSize(int p_width, int p_height)
	 *
	 *	@brief Sets the size of the Widget.
	 *
	 *	@param[in] p_width Width to be set for the widget.
	 *	@param[in] p_height Height to be set for the widget.
	 *
	 *	Implemented here in order to trigger some internal calculations.
	 */
	int setSize(int p_width, int p_height) override;


private:

	/**
	 *	@fn void drawTableBorders(WEPainter * p_painter)
	 *
	 *	@brief Draws the table borders between each row and each column
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_painter WEPainter used for drawing
	*/
	void drawTableBorders(WEPainter * p_painter);

	void updateTableBorderNodes(WEControllerNode *p_controllerNode);

	/**
	 *	@fn void drawHeadings(WEPainter * p_painter)
	 *
	 *	@brief Draw the heading line (if set with useHeading() function)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_painter WEPainter used for drawing
	*/
	void drawHeadings(WEPainter * p_painter);

	void updateHeadingNodes(WEControllerNode *p_controllerNode);

	/**
	 *	@fn void drawContent(WEPainter * p_painter)
	 *
	 *	@brief Extract each cell value from the set data and draw it on screen
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_painter WEPainter used for drawing
	*/
	void drawContent(WEPainter * p_painter);

	void updateContentNodes(WEControllerNode *p_controllerNode);

	/**
	 *	@fn void drawScrollbar(WEPainter * p_painter)
	 *
	 *	@brief Draw a scrollbar on the right side of the table but only if not all content does fit
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	*/
	void drawScrollbar(WEPainter * p_painter);

	void updateScrollbarNodes(WEControllerNode *p_controllerNode);

	/**
	 *	@fn void calculateXCoordinates(void)
	 *
	 *	@brief Calculate the x-coordinates of the columns based on column spacing settings
	 *	Consider drawable area only i.e. exclude border.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	*/
	void calculateXCoordinates(void);

	/**
	 *	@brief Calculate this->rowHeight and this->visibleRows
	 *
	 */
	void calculateRows(void);

	static const int TABLE_BORDER_WIDTH;

	int highlightedRow;	//!< Row number that should be highlighted
	int numberOfColumns;	//!< The number of columns of this table
	int rowHeight;	//!< Height of a single row guessed from the font size.
				// WARNING: Many local variables with the same name and purpose exist.
	int visibleDataRows;// Rows fitting into the widget (depends on data size, widget size, font size and 'headingUsed')
	bool headingUsed;	//!< Flag if the heading shall be drawn or not
	WEColor headingBackgroundColor;	//!< The background color for the heading
	WEColor headingFontColor;	//!< The font color for each heading
	std::string heading;		//!< The heading text itself. Columns have to be separated with the delimiter character.
	std::vector<std::string> data;	//!< Vector containing the cell data (without heading). Cell data has to be delimited within each string.
	std::map<int, int> columnSpacing;	//!< Setting for the width of each column in percent
	std::vector<int> xCoordinates;	//!< xCoordinates for each column
	char delimiter;		//!< Delimiter character, see data


	WERectangleNode * headingBackgroundRectangleNode;	//!<Node of the Widget which is used to draw the headings
	WERectangleNode * borderRectangleNode;			//!<Node of the Widget which is used to draw the border
	WEGeometryNode * borderGeometryNode;			//!<Node of the Widget which is used to draw the border
	WERectangleNode * scrollbarRectangleNode;		//!<Node of the Widget which is used to draw the scrollbar
	WERectangleNode * scrollbarBorderRectangleNode;		//!<Node of the Widget which is used to draw the scrollbar
	std::vector<WETextNode*> headingsTextNodes;		//!<Headings Text Nodes of the Widget
	std::vector<std::vector<WETextNode*>> contentTextNodes;	//!<Content Text Nodes of the Widget
};
}  // namespace WE

#endif /* WE_TABLE_WIDGET_H_ */
