/**
 *	@file we_virtual_keyboard_widget.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of virtual keyboard widget
 */


#ifndef WE_VIRTUAL_KEYBOARD_WIDGET_H_
#define WE_VIRTUAL_KEYBOARD_WIDGET_H_

#include <vector>
#include <set>

//Qt 4.x MOC versions can have problems with some boost headers.
//Here we tell MOC to ignore the following includes
#ifndef Q_MOC_RUN
#include <boost/bimap/bimap.hpp>
#include <boost/bimap/set_of.hpp>
#endif

#include "../../../common/include/we_color.h"
#include "../../../event_handling/include/we_input_event.h"
#include "../../../event_handling/include/we_toolkit_event.h"

#include"we_widget.h"

namespace WE {

class WEButtonWidget;
class WEContainer;
class WEStringWidget;
class WELinearBargraphWidget;
class WEInplaceEditingListener;

/**
 *	@class WEVirtualKeyboardWidget
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Widget for displaying a keyboard and editing a certain value
 *
 *	@namespace WE
 *
 *	The layered string keyboard has to deal with a variety of different string formats:
 *	a) It's supposed to accept input attributes based on character encoding such as ISO 8895-1
 *	b) Later on it's supposed to also work with wide chars (for ISO-VT version 4)
 *	c) The string widgets operate on UTF-8 by default - but also carry the encoding in their font which is considered by the painter
 *	The approach to cover these different formats are:
 *	1) Mainly use plain strings which are typically the same as UTF-8 (at least for the predefined buttons)
 *	2) Convert all input attributes to wide strings (as it is already done in inplace editor)
 *	3) Convert the shown string to UTF-8.
 *	4) Use Unicode wide strings to calculate the length of the string (which should also be correct for character table based strings)
 *	5) When creating special buttons convert the wide char to UTF-8 for display on screen.
 *	6) Generally operate on wide chars once the keyboard is running.
 *
 */
class WEVirtualKeyboardWidget : public WEWidget, WEInputEventListener, WEToolkitEventListener
{
	Q_OBJECT
public:
	WEVirtualKeyboardWidget(WEWidget * p_parent, bool p_isNumeric, void * p_controller = nullptr);
	virtual ~WEVirtualKeyboardWidget();


	/**
	 *	@fn void setColors(	unsigned int p_bgColor,
			unsigned int p_buttonColor,
			unsigned int p_buttonShadowColor,
			unsigned int p_valueColor,
			unsigned int p_warningColor,
			unsigned int p_bargraphColor,
			unsigned int p_alpha,
			unsigned int p_focus = 0x0000ff00,
			unsigned int p_open = 0x00ff0000,
			unsigned int p_borderAlpha = 0xa0,
			unsigned int p_borderWidth = 5 )
	 *
	 *	@brief Set colors to be used for buttons, background
	 *
	 *	@param[in] p_bgColor Background color for the whole keyboard.
	 *	@param[in] p_buttonColor Color of the button surfaces.
	 *	@param[in] p_buttonShadowColor Color of button shadow indicating if button is pressed or not
	 *	@param[in] p_valueColor Color used for the background of the value shown on top of the keyboard
	 *	@param[in] p_warningColor Background color of the value field which is used if the entered value is invalid
	 *	@param[in] p_bargraphColor Color of the bargraph visualizing min/max range of numeric VKB
	 *	@param[in] p_alpha) Opaqueness applied to the keyboard and the buttons (0: completely transparent, 0xff: no transparency)
	 *	@param[in] p_focus Color for "focus" border around buttons (optional, default is "lime")
	 *	@param[in] p_open Color for "open for input" border around other widgets (optional, default is "red")
	 *	@param[in] p_borderAlpha Alpha value to be used for navigation frame (optional, default is 0xa0 == 160)
	 *	@param[in] p_borderWidth Width to be used for navigation frame (optional, default is 5)
	 *
	 *	@returns nothing
	 *
	 *
	 *	This method sets different colors to be applied to the elements of the virtual keyboard.
	 *	The intention is to make them configurable e.g. from a config file.
	 *	createLayout() will use these and changing them later will not affect existing instances.
	 *
	 *	@see createLayout(unsigned int p_width, unsigned int p_height)
	 */
	void setColors(	unsigned int p_bgColor,
			unsigned int p_buttonColor,
			unsigned int p_buttonShadowColor,
			unsigned int p_valueColor,
			unsigned int p_warningColor,
			unsigned int p_bargraphColor,
			unsigned int p_alpha,
			unsigned int p_focus = 0x0000ff00,
			unsigned int p_open = 0x00ff0000,
			unsigned int p_borderAlpha = 0xa0,
			unsigned int p_borderWidth = 5 );


	/**
	 *	@fn void activateButton(WEButtonWidget * p_button, bool p_active = true)
	 *
	 *	@brief Switch VKB button to active or inactive.
	 *
	 *	@param[in] p_button Button to be switched.
	 *	@param[in] p_active State the button should get.
	 *
	 *	@returns nothing
	 *
	 *	@see createSingleButton()
	 */
	void activateButton(WEButtonWidget * p_button, bool p_active = true);


	/**
	 *	@fn void setBorderProperties( WEWidget * p_widget );
	 *
	 *	@brief Create border for input object shown on VKB.
	 *
	 *	@param p_widget Pointer widget that should get the properties.
	 *
	 *	@returns nothing
	 *
	 *
	 */
	void setBorderProperties( WEWidget * p_widget );


	/**
	 *	@fn 	WEButtonWidget * createNewlineButton( WEWidget * p_parent );
	 *
	 *	@brief Create a single button with the new line functionality.
	 *
	 *	@param p_parent Pointer to layer which is parent of the returned new line button.
	 *
	 *	@returns A pointer to the button created
	 *
	 *
	 *	This method is used to add the new line button to every layer of the string VKB.
	 *	Formerly there was only one common button of this type - but it broke the tab order.
	 *
	 *	The created button is also added to the "newLineButtons" set.
	 *
	 *	@see createLayout(unsigned int p_width, unsigned int p_height)
	 *	@see createSingleButton()
	 */
	WEButtonWidget * createNewlineButton( WEWidget * p_parent );


	/**
	 *	@fn WEButtonWidget * createSingleButton(
			unsigned int p_width,
			unsigned int p_height,
			std::string & p_caption,
			bool p_isEnabled = true,
			WEWidget * p_parent = nullptr,
			WEColor p_captionColor = WEColor(black),
			unsigned int p_fontSize  = 0)
	 *
	 *	@brief Create a single button as child of this VKB.
	 *
	 *	@param[in] p_width Width to be set to the button.
	 *	@param[in] p_height Height to be set for the button.
	 *	@param[in] p_caption Text to appear on the button.
	 *	@param[in] p_isEnabled True: Button should be enabled, false otherwise
	 *	@param[in] p_parent Parent widget to assign to the new button - "nullptr" is replaced by "this"
	 *	@param[in] p_captionColor Color of text to appear on the button.
	 *	@param[in] p_fontSize Size in pixels of text to appear on the button (if 0 p_height/2 is used).
	 *
	 *	@returns A pointer to the button created
	 *
	 *
	 *	This method is used when creating the layout of this keyboard to create the single buttons.
	 *
	 *	@see createLayout(unsigned int p_width, unsigned int p_height)
	 */
	WEButtonWidget * createSingleButton(
			unsigned int p_width,
			unsigned int p_height,
			const std::string & p_caption,
			bool p_isEnabled = true,
			WEWidget * p_parent = nullptr,
			WEColor p_captionColor = WEColor(black),
			unsigned int p_fontSize  = 0);


	/**
	 *	@fn void createLayout(unsigned int p_width, unsigned int p_height)
	 *
	 *	@brief Create everything needed to display this virtual keyboard on screen.
	 *
	 *	@param[in] p_width Width to be set to the whole keyboard (both 0 to keep current size).
	 *	@param[in] p_height Height to be set for the whole keyboard (both 0 to keep current size).
	 *
	 *	This method sets the size members of this keyboard and then creates all the buttons and other widgets required.
	 *	Once this has run through VKB can be used for editing.
	 *
	 *	@Warning VKB will not work if this method isn't called before trying to use it.
	 */
	void createLayout(unsigned int p_width, unsigned int p_height);


	/**
	 *	@fn void deactivateOldValidationString( void )
	 *
	 *	@brief Reset string VKB to virgin state ;-).
	 *
	 *	@param none
	 *
	 *	@returns nothing
	 *
	 *
	 *	Remove all the layers created for the recent validation string and disable predefined buttons.
	 *
	 */
	void deactivateOldValidationString( void );


	/**
	 *	@fn void applyValidationString( void )
	 *
	 *	@brief Enable buttons selectively and create additional ones if required.
	 *
	 *	@param none
	 *
	 *	@returns nothing
	 *
	 *
	 *	Enable default buttons already created or create new ones and add them to additional layers.
	 *
	 */
	void applyValidationString( void );


	void setNumericOptions(const std::string & p_value, bool p_resultIsDouble, double p_min, double p_max, double p_offset1, double p_offset2, double p_scale, int p_noOfDecimals, bool p_rollOver, WEInplaceEditingListener * p_listener);


	/**
	 *	@fn int setEditedValueAndListener(std::string & p_value, WEInplaceEditingListener * p_listener)
	 *
	 *	@brief Set the value to be edited and identify target for result.
	 *
	 *
	 *	@param[in] p_value Start value for editing (can't be const => line endings will be fixed!)
	 *	@param[in] p_listener Inplace editing listener to be called when editing is finished
	 *
	 *	@returns An error code, WE::WE_SUCCESS in case of success
	 *
	 */
	int setEditedValueAndListener(const std::string & p_value, WEInplaceEditingListener * p_listener = nullptr)
;


	/**
	 *	@fn int setNoValidationString(WEEncodeType p_encodeType)
	 *
	 *	@brief Allow choosing every printable character supported by p_encodeType.
	 *
	 *	@param[in] p_encodeType Code table to be used.
	 *
	 *	@returns An error code, WE::WE_SUCCESS in case of success
	 *
	 *
	 *	This method sets the internal validation wide string with all characters allowed that are supported
	 *	by code table p_encodeType.
	 *
	 *	@Warning Although this method internally creates a list of wide chars it only works for ASCII code tables.
	 *
	 *	@see int setValidationString(WEEncodeType p_encodeType, const std::string & p_string, bool p_validationType)
	 */
	int setNoValidationString(WEEncodeType p_encodeType);


	/**
	 *	@fn int setValidationString(WEEncodeType p_encodeType, const std::string & p_string, bool p_validationType)
	 *
	 *	@brief Set the string of characters which operator is allowed to enter here.
	 *
	 *	@param[in] p_encodeType Code table to be used.
	 *	@param[in] p_string Characters as defined by DDO.
	 *	@param[in] p_validationType false if p_string contains a positive list, true otherwise
	 *
	 *	@returns An error code, WE::WE_SUCCESS in case of success
	 *
	 *
	 *	This method sets the internal validation wide string with only allowed characters based on the
	 *	parameters provided. I.e. if p_validationType is false this will convert the list of forbidden
	 *	characters provided in p_string into a list of allowed characters used to set up the VKB.
	 *
	 *	@Warning Although this method internally creates a list of wide chars it only works for ASCII code tables.
	 */
	int setValidationString(WEEncodeType p_encodeType, const std::string & p_string, bool p_validationType);


	/**
	 *	@fn unsigned int getMaxStringLength( void )
	 *
	 *	@brief Get the max string length that can be entered here.
	 *
	 *	@param none
	 *
	 *	@returns The max string length currently set for this VKB.
	 *
	 *	String length is only enforced for string VKB.
	 */
	unsigned int getMaxStringLength( void ) const;


	/**
	 *	@fn unsigned int setMaxStringLength( unsigned int p_length )
	 *
	 *	@brief Set the max string length that can be entered here.
	 *
	 *	@param Max string length that should be allowed when editing.
	 *
	 *	@returns The max string length now set for this VKB.
	 *
	 *	String length is only enforced for string VKB.
	 */
	unsigned int setMaxStringLength( unsigned int p_length );


	int inputEventOccurred(WEInputEvent & p_event) override;

	int toolkitEventOccurred(WEToolkitEvent & p_event) override;

	void paint(WEPainter* p_painter) override;

	/**
	 *	@fn QSGNode * updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *data)
	 *
	 *	@brief Updates the nodes for this Qt quick item
	 *
	 *	@see QSGNode *QQuickItem::updatePaintNode(QSGNode *oldNode, QQuickItem::UpdatePaintNodeData *updatePaintNodeData)
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param oldNode the previously returned value of this function
	 *	@param data contains a pointer to the QSGTransformNode of this item
	 *
	 *	@returns the root node of this item
	 */
	QSGNode * updatePaintNode(QSGNode *p_oldNode, QQuickItem::UpdatePaintNodeData *p_data) override;


	/**
	 *	@fn setSize(int p_width, int p_height)
	 *
	 *	@brief Sets the VISIBLE size of the Widget.
	 *
	 *	@param[in] p_width Width to be set for the widget.
	 *	@param[in] p_height Height to be set for the widget.
	 *
	 *	This method sets the visible size of the VKB.
	 *	String VKB widget should anyways fill the whole screen but for
	 *	numeric VKB we want to block all underlying inputs.
	 *
	 *	Consequently the VKB has to fill the whole screen to catch all mouse
	 *	(i.e. touch) events.
	 *
	 *	We assume that using the parent widget's size will be the right one to use
	 *	because typically the parent has to be the window itself.
	 *
	 *	@see WE::WEWidget::setSize(int p_width, int p_height)
	 */
	int setSize(int p_width, int p_height) override;



	/**
	 *	@fn setLocation(int p_x, int p_y)
	 *
	 *	@brief Sets the VISIBLE location of the Widget relative to its parent.
	 *
	 *	@param[in] p_x X-Coordinate of the location to be set for the widget.
	 *	@param[in] p_y Y-Coordinate of the location to be set for the widget.
	 *
	 *	This method sets the visible position of the VKB.
	 *	String VKB widget should anyways fill the whole screen but for
	 *	numeric VKB we want to block all underlying inputs.
	 *
	 *	Consequently the VKB has to fill the whole screen to catch all mouse
	 *	(i.e. touch) events.
	 *
	 *	We assume that using the parent widget's size will be the right one to use
	 *	because typically the parent has to be the window itself.
	 *
	 *	@see WE::WEWidget::setLocation(int p_x, int p_y)
	 */
	void move(int p_x, int p_y);

	/**
	 * Set new style/theme information that will be applied to each button of the VKB
	 */
	void updateButtonStyle(const WE::WEColor & p_bgColor,
			       const std::string & p_bgImage,
			       const std::string & p_borderImage,
			       int p_borderWidth,
			       bool p_transparent,
			       const WE::WEColor & p_shadowColor,
			       bool p_drawShadow,
			       const std::string & p_pressedImage,
			       const std::string & p_releasedImage);


private:
	void handleNewStringInput(const std::string & p_value);


	/**
	 *	@brief Apply edited value to string widget and set its properties accordingly.
	 *
	 *	@param p_cursor Increment for cursor, 0: keep cursor where it is
	 *	@param p_inverted true: value is displayed inverted
	 *
	 *	@returns The amount of printable characters in this->currentValue.
	 *
	 *	Depending in the amount of characters the ok button is switched on or off.
	 */
	unsigned int applyStringValue( int p_cursor = 0, bool p_inverted = false );


	/**
	 *	@fn bool isStringValid( const std::string & p_value )
	 *
	 *	@brief Check UTF-8 string p_value against the current validation string
	 *
	 *	@param p_value UTF-8 string to be checked.
	 *
	 *	@returns true if p_value is allowed, false if it contains characters that are not allowed
	 *
	 *	@see validationString
	 *	@see bool isStringValid( const std::wstring & )
	 */
	bool isStringValid( const std::string & p_value ) const;


	/**
	 *	@fn bool isStringValid( const std::wstring & p_value )
	 *
	 *	@brief Check Unicode string p_value against the current validation string
	 *
	 *	@param p_value Unicode (UTF-16) string to be checked.
	 *
	 *	@returns true if p_value is allowed, false if it contains characters that are not allowed
	 *
	 *	@see validationString
	 *	@see bool isStringValid( const std::string & )
	 */
	bool isStringValid( const std::wstring & p_value ) const;


	void handleNewNumericInput(const std::string & p_value);


	/**
	 *	@fn void calculateCurrentNumericValue(bool p_useInt = false)
	 *
	 *	@brief Check int value against settings of DDO we are editing here.
	 *
	 *	@param p_useNumber false: use this->currentValue (the string), true: use this->currentValueAsInt/Double
	 *	@param p_inverted true: value is displayed inverted
	 *
	 *	@returns nothing
	 *
	 *	In case we are getting a double (i.e. the value to show) calculate the int value.
	 *	Which not too sensible as we then typically also get the limits as calculated double.
	 */
	void calculateCurrentNumericValue(bool p_useNumber = false, bool p_inverted = false);


	/**
	 *	@fn int focusObject(int p_offset, WEWidget * p_objectTofocus = nullptr)
	 *
	 *	@brief Set border states and internal member so that another (e.g.) button get focus.
	 *
	 *	@param p_offset 1: next object, -1: previous object, 0: set focus to p_objectTofocus
	 *	@param p_objectTofocus (touched) object to get focus
	 *
	 *	@returns An error code; WE_SUCCESS in case of success.
	 *
	 *	This is used to set focus for encoder navigation.
	 */
	int focusObject(int p_offset, WEWidget * p_objectTofocus = nullptr);


	/**
	 *	@fn int clickButton(WEButtonWidget * p_button)
	 *
	 *	@brief Do what is required when button is clicked (by mouse or encoder).
	 *
	 *	@param p_button Button to be clicked
	 *
	 *	@returns An error code; WE_SUCCESS in case of success.
	 *
	 *	This is used to set focus for encoder navigation as well as touch screen.
	 */
	int clickButton(WEButtonWidget * p_button);


	/**
	 *	@fn int incrementValue(int p_step)
	 *
	 *	@brief Increment/decrement raw value so that a change is visible on screen.
	 *
	 *	@param p_step Direction and amount of steps to modify.
	 *
	 *	@returns An error code; WE_SUCCESS in case of success.
	 *
	 *	This is used to modify value with plus or minus button or with encoder.
	 */
	int incrementValue(int p_step);

	/**
	 * A style/theme for all buttons can be set with updateButtonStyle(). These settings
	 * are stored in lastButtonStyle member and applied with this function. If a new button is created,
	 * this function will also apply the last set style to that new button.
	 */
	void applyLastStyleToButton(WEButtonWidget * p_buttonWidget);

	/**
	 * trims or adds decimal places from/to the given string based on noOfDecimals member
	 */
	void trimNumericStringToNumberOfDecimals(std::string & p_stringToTrim);

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// members for all types of virtual keyboards supported here:
	///////////////////////////////////////////////////////////////////////////////////////////////////
	static const unsigned int charScaleMul = 1;	//!< Normal characters as button caption should have font size 1/2 of button height.
	static const unsigned int charScaleDiv = 2;

	static const unsigned int specialCharScaleMul = 3;	//!< Special characters as button caption should have font size 3/4 of button height.
	static const unsigned int specialCharScaleDiv = 4;	// Note: This makes all special characters fit on PC but needs to be checked on different devices

	static const unsigned int maxCustomButtonsPerLayer = 27;	//!< We can add this many buttons to every custom layer

	enum
	{
		LAYER_LOWER_CASE	= 0,
		LAYER_UPPER_CASE	= 1,
		LAYER_NUMBERS		= 2,
		LAYER_HIGHEST_PREDEFINED = LAYER_NUMBERS,
		LAYER_FIRST_CUSTOMIZED	= 3
	};

	// members to be taken from config file modifying the look:
	WEColor buttonColor{WE::silver};	//!< Color of button's background
	WEColor buttonShadowColor{WE::black};	//!< Color of button's shadow (identifying pressed/released)
	WEColor valueColor{WE::silver};		//!< Color of the character shown in shownValueString
	WEColor invertedValueColor{WE::white};	//!< Color of the character shown in shownValueString while it's inverted
	WEColor warningColor{WE::magenta};	//!< Background color of the value field if value is invalid
	WEColor barGraphColor{WE::green};	//!< Fill color of the bargraph showing min/max range
	unsigned int alpha{128};		//!< Opaqueness applied to the background and the buttons
	WEColor focusColor{WE::lime};		//!< Color for "focus" border around buttons
	WEColor openColor{WE::red};		//!< Color for "open for input" border around other widgets
	unsigned int borderAlpha{160};		//!< Alpha value to be used for navigation frame
	unsigned int borderFrameWidth{5};	//!< Width to be used for navigation frame

	unsigned int childWidth{10};		//!< Maintain button size base values for generation of buttons later on
	unsigned int childHeight{10};		//!< Maintain button size base values for generation of buttons later on
	unsigned int childGapVertical{5};	//!< Maintain button size base values for generation of buttons later on
	unsigned int childGapHorizontal{5};	//!< Maintain button size base values for generation of buttons later on


	// members required in control flow:
	int vkbX{0};				//!< Visible x position of the VKB (the widget itself will have screen size)
	int vkbY{0};				//!< Visible y position of the VKB (the widget itself will have screen size)
	int vkbWidth{120};			//!< Visible width of the VKB (the widget itself will have screen size)
	int vkbHeight{120};			//!< Visible height of the VKB (the widget itself will have screen size)

	bool isNumeric{false};				//!< Switch between string and numeric keyboard layout and functionality
	std::vector< WEButtonWidget * > buttons{};	//!< List of all the buttons currently shown on screen - or used for all layers?
	WEStringWidget * shownValueString{nullptr};	//!< Current value as edited be operator (shown above the button area)
	WEStringWidget * minMaxString{nullptr};		//!< Small indicator for min/max value of numeric keyboard painted above shownValueString
	std::string currentValue{""};		//!< Edited value shown in "shownValueString" widget
	std::string initialValue{""};		//!< Value as it was before starting to edit (i.e. the one to provide to the DDO on "cancel")
	WEButtonWidget * okButton{nullptr};	//!< For every user input we have to consider hiding this - so avoid searching all the time.


	WEWidget * currentFocussedObject{nullptr};	//!< Currently focussed object.
	int currentFocussedButtonIndex{-1};		//!< Index into "buttons" vector pointing to currentFocussedObject,
						//   negative if focussed object is not a button
	WEWidget * initiallyFocussedObject{nullptr};	//!< Encoder navigation should start here - preferably on "Cancel"

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// members for the string keyboard only:
	///////////////////////////////////////////////////////////////////////////////////////////////////
	WEEncodeType charSet{ISO_8859_1};			//!< Character set to be used (VT version <= 3 only supports Latin 1 (ISO_8859_1) and 9 (ISO_8859_15)

	std::vector< WEContainer * > layers{};	//!< List of all the layers we have
	unsigned int layerShown{LAYER_LOWER_CASE};	//!< Layer currently active
	bool shifted{false};				//!< true: Show upper case, false: lower case
	std::wstring validationString{L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890/*-+=?!§$%&[]{}().:,;<>"};		//!< Characters allowed to enter according to input attributes object
	unsigned int maxStringLength{25};		//!< Consider value invalid if longer than that

	std::set < WEButtonWidget * > newLineButtons{};	//!< We have a very different caption than CR or LF but it has to be allowed by validation string
	WEButtonWidget * extraCharsButton{nullptr};	//!< This button has to be (de)activated based on whether we have additional layers or not

	typedef boost::bimaps::bimap
	<
		boost::bimaps::set_of< wchar_t > ,
		boost::bimaps::set_of< WEButtonWidget * >
	> charToButtonBimap;

	charToButtonBimap supportedChars{};	//!< Mapping between supported chars and their buttons

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// members for the numeric keyboard only:
	///////////////////////////////////////////////////////////////////////////////////////////////////
	int listenerValue{0};			// !< this just exists so that we can return a pointer to the listener
	signed long long currentValueAsInt{0};
	double currentValueAsDouble{0.0};	// !< Value shown, should be the same as (currentValueAsInt + offset) * scale

	double min{0};
	double max{0};
	double offset1{0.0};
	double offset2{0.0};			//!< Seems this is an option for the future.
	double scale{1.0};
	unsigned int noOfDecimals{0};
	bool rollOver{false};			//!< Should value overflow when using encoder or "+" and "-" button?
	bool truncate{false};			//!< Define if value is truncated or rounded after scale and offset are applied (currently always "false")
	char decimalChar{'.'};		//!< Make sure we use the decimal char suggested by numeric widget
	bool numericResultIsDouble{false};

	WEButtonWidget * decimalButton{nullptr};
	WEButtonWidget * plusButton{nullptr};	//!< Needed to include bargraph in navigation (otherwise would have to compare all the labels of buttons)
	WEButtonWidget * minusButton{nullptr};	//!< Needed to include bargraph in navigation (otherwise would have to compare all the labels of buttons)
	WELinearBargraphWidget * barGraph{nullptr};	// bargraph showing min and max value range (and allowing editing with encoder)

	struct ButtonStyle
	{
		WE::WEColor bgColor{WE::silver};
		std::string bgImage{};
		std::string borderImage{};
		int borderWidth{3};
		bool transparent{false};
		WE::WEColor shadowColor{WE::black};
		bool drawShadow{true};
		std::string pressedImage{};
		std::string releasedImage{};
	};

	ButtonStyle lastButtonStyle{};
};
}  // namespace WE

#endif /* WE_VIRTUAL_KEYBOARD_WIDGET_H_ */
