/**
 * 	@file we_widget.h
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file defines the base Widget of all the user visible widgets.
 *
 *	@see we_widget.cpp
 *
 *	The base Widget defined in this file allows the user to create user visible
 *	Widgets.
 */

#ifndef WE_WIDGET_H_
#define WE_WIDGET_H_

#include <map>

//Qt 4.x MOC versions can have problems with some boost headers.
//Here we tell MOC to ignore the following includes
#ifndef Q_MOC_RUN
#include <boost/bimap/bimap.hpp>
#include <boost/bimap/set_of.hpp>
#endif

#include <QtGlobal>

#include <QQuickItem>

#include <QPushButton>
#include <QPainter>
#include <QTouchEvent>

// Included header files for the data members only.
#include "../../../event_handling/include/we_toolkit_event.h"
#include "../../painter/include/we_pen.h"
#include "../../painter/include/we_brush.h"
#include "../../painter/include/we_painter.h"
#include "../../common/include/we_border_profile.h"
#include "../../common/include/we_inplace_editing_listener.h"
#include "../../../common/include/we_error.h"
#include "../../../gui/image/include/we_image.h"
#include "../../texture_cache/include/we_texture_cache.h"
#include "../../../gui/image/include/toolkit_we_image.h"

class QEvent;

namespace WE
{

class WEWidget;
class QTWEPainter;
class QTWEApplication;

class WERectangleNode;
class WEControllerNode;
class WEGeometryNode;
class WEEllipseNode;
class WEImageNode;


typedef boost::bimaps::bimap
<
	boost::bimaps::set_of< unsigned int > ,
	boost::bimaps::set_of< WEWidget* >
> childrenBiMap;

/**
 *	@enum widgetTypeID
 *
 *	@brief This contains the defined TYPEIDs of widgets
 *
 *	@see WEApp::TYPE_ID
 */
enum widgetTypeID
{
	// Up to 255 they should be identical to those defined in ISO 11783-6 higher values are proprietary
	// To avoid confusion make sure these are in sync with WEApp::TYPE_ID
	WIDGET_TYPE_ID_WORKING_SET	=    0,	//!< Page
	WIDGET_TYPE_ID_PAGE		=    1,	//!< Page
	WIDGET_TYPE_ID_CONTAINER	=    3,	//!< Container
	WIDGET_TYPE_ID_SOFTKEY_FRAME	=    4,	//!< SoftkeyFrame
	WIDGET_TYPE_ID_SOFTKEY		=    5,	//!< Softkey
	WIDGET_TYPE_ID_BUTTON		=    6,	//!< Button
	WIDGET_TYPE_ID_BOOLEAN		=    7, //!< Boolean Widget
	WIDGET_TYPE_ID_STRING		=    8,	//!< String Field
	WIDGET_TYPE_ID_NUMERIC		=    9,	//!< Numeric Field
	WIDGET_TYPE_ID_LIST		=   10,	//!< List Object
	WIDGET_TYPE_ID_LINE		=   13,	//!< Line Object
	WIDGET_TYPE_ID_RECTANGLE	=   14,	//!< Rectangle Object
	WIDGET_TYPE_ID_ELLIPSE		=   15,	//!< Ellipse Object
	WIDGET_TYPE_ID_POLYGON		=   16,	//!< Polygon Object
	WIDGET_TYPE_ID_METER		=   17,	//!< Meter
	WIDGET_TYPE_ID_LINEAR_BARGRAPH	=   18,	//!< Linear Bargraph
	WIDGET_TYPE_ID_ARCHED_BARGRAPH	=   19,	//!< Arched Bargraph
	WIDGET_TYPE_ID_IMAGE		=   20,	//!< Image
	WIDGET_TYPE_ID_NUMBER_VAR	=   21,	//!< Number Variable
	WIDGET_TYPE_ID_STRING_VAR	=   22,	//!< String Variable
	WIDGET_TYPE_ID_POINTER		=   27,	//!< Object pointer
	WIDGET_TYPE_ID_OUTPUT_LIST	=   37,
	WIDGET_TYPE_ID_FRAME		=  256,
	WIDGET_TYPE_ID_SCALED_CONTAINER	=  257,
	WIDGET_TYPE_ID_TABLE		=  258,	//!< Table
	WIDGET_TYPE_ID_GRAPH		=  259, //!< Guess what that means...
	WIDGET_TYPE_ID_GAUGE		=  262,
	WIDGET_TYPE_ID_STRING_INPLACE_EDIT = 308,	//!< simple String inplace editing widget
	WIDGET_TYPE_VT_PRIMARY		=  310,	//!< Container widget showing BNE VT and accepting operator input to the pool
	WIDGET_TYPE_VT_VIEW		=  311,	//!< Container widget showing BNE VT
	WIDGET_TYPE_ID_STRING_VIRTUAL_KEYB = 408,	//!< virtual keyboard inplace editing widget
	WIDGET_TYPE_ID_VIRTUAL_KEYBOARD	= 409,
	WIDGET_TYPE_ID_VIRTUAL_WALL	= 500,
	WIDGET_TYPE_ID_WINDOW		= 1000,	//!< virtual keyboard inplace editing widget
	WIDGET_TYPE_ID_UNDEFINED	= 0xffff

	// Note: There's a similar list in app_base_ddo.h
	// Note: If type IDs are added, make sure to (maybe) add them in AppBaseDDO::getDDOTypeName() also
};


// Forward declarations:
class WEImage;
class QtWEScrollBar;
class QtWEWindowStyle;
class QTWEScrollArea;

/**
 *	@class WEWidget
 *
 *	@author kw@wachendorff.de
 *	@author Katsiaryna Walesch
 *
 *	@brief This class is the base class for all user visible objects (Widgets and Windows).
 *
 *	@namespace WE
 *
 */
class WEWidget : public QQuickItem
{
Q_OBJECT
Q_PROPERTY(int typeID READ getTypeID)
Q_PROPERTY(bool active READ isActive)
Q_PROPERTY(bool drawBorder READ getDrawBorder)
Q_PROPERTY(bool transparentBorder READ isBorderTransparent)
Q_PROPERTY(int borderWidth READ getBorderWidth)
Q_PROPERTY(int borderState READ getBorderState)
Q_PROPERTY(bool transparent READ isTransparent)
Q_PROPERTY(bool setAsInput READ isSetAsInput)
Q_PROPERTY(int startAngle READ getStartAngle)
Q_PROPERTY(int endAngle READ getEndAngle)
Q_PROPERTY(bool weVisible READ shouldBeVisible)
Q_PROPERTY(bool drawBorderImage READ getDrawBorderImage)
Q_PROPERTY(bool drawBackgroundImage READ getDrawBackgroundImage)
Q_PROPERTY(bool hasValidValue READ isValueValid)
Q_PROPERTY(int spanAngle READ getSpanAngle)

public:

	/**
	 *	@fn WEWidget(int p_typeID, WEWidget * p_parent, void * p_controller = nullptr)
	 *
	 *	@brief Constructor for a (general) widget - only to be called from derived classes.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_typeID - to identify which kind of widget this really is
	 *	@param[in] p_parent - pointer to parent widget to create a tree of widgets
	 *	@param[in] p_controller - pointer to controller - just to get it back via getController() or with an event
	 *
	 *	@returns nothing
	 *
	 *
	 *	If called by a class derived from WEWidget this will initialize all the general
	 *	properties of a widget.
	 *
	 *	If p_controller should be nullptr e.g. a mouse event will not contain a valid pointer to the origin that triggered it.
	 */
	WEWidget(int p_typeID, WEWidget * p_parent, void * p_controller = nullptr);


	/**
	 *	@fn ~WEWidget()
	 *
	 *	@brief Destructor from the abstract widget class
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	WEWidget is a completely abstract class
	 *	Therefore it has no Constructor and only the virtual Destructor, which must be defined
	 */
	virtual ~WEWidget();


	/**
	 *	@fn void release()
	 *
	 *	@brief release the members
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  	@returns void
	 */
	void release();


	/**
	 *	@fn void deleteSubtree( void )
	 *
	 *	@brief Recursively delete all children of this widget.
	 *
	 *	@param none.
	 *
	 *	@returns nothing
	 *
	 *	This method deletes all the child widgets of this one recursively.
	 *	Its main purpose is cleaning up in destructor if the application didn't do it
	 *	and supporting designs with children that are not controlled by application
	 *	(such as the WEVirtualKeyboardWidget).
	 */
	void deleteSubtree( void );


	void create(WEWidget * p_parentWidget);

	/**
	 *	@fn int getTypeID(void)
	 *
	 *	@brief Get the type ID of this widget as set in constructor.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *
	 *	@returns Type ID of this widget.
	 *
	 *	@see widgetTypeID
	 *
	 */
	int getTypeID(void) const
	{
		return this->typeID;
	}


	/**
	 *	@fn void * getController( void )
	 *
	 *	@brief Get the controller registered at construction time.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns pointer to controller as provided to constructor
	 *
	 *	This method is mainly intended to use the tree of widgets e.g. for navigation.
	 *	It provides a pointer that could/should point to the instance controlling this widget
	 *	on application side.
	 *
	 *	If this pointer was not provided to constructor nullptr will be returned.
	 */
	virtual void * getController( void ) const
	{
		return this->controller;
	}

	/**
	 *	@fn void getScreenLocation(int & p_x, int & p_y, WEWidget * p_parentWidget = nullptr)
	 *
	 *	@brief Get the coordinates of the top left corner of this widget with respect to origin of the screen
	 *		or the specified parent widget.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@param p_x Reference to be filled with x coordinate
	 *	@param p_y Reference to be filled with y coordinate
	 *	@param p_parentWidget a parent (may not be immediate) of this widget in whose respect the coordinates are expected,
	 *				nullptr means screen(global) coordinates are expected.
	 *
	 *	@returns nothing (except in reference parameters)
	 *
	 *	This method is useful e.g. to compare mouse location with widget position or
	 *	to know where the widget is located within a certain parent (e.g. datamask) widget.
	 *
	 *	This is a wrapper for the ToolkitWEWidget::getScreenLocation(),
	 *	which is required when we need to call it from the application/ddos
	 *	where we dont have access to ToolkitWEWiget.
	 */
	void getScreenLocation(int & p_x, int & p_y, WEWidget * p_parentWidget = nullptr) const;

	/**
	 *	@fn setX(int p_x)
	 *
	 *	@brief Sets the X-location of the Widget.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_x X-Coordinate of the location to be set for the widget.
	 *
	 *	@see setLocation(int p_x, int p_y)
	 */
	void setX(int p_x);

	/**
	 *	@fn setY(int p_y)
	 *
	 *	@brief Sets the Y-location of the Widget.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_y Y-Coordinate of the location to be set for the widget.
	 *
	 *	@see setLocation(int p_x, int p_y)
	 */
	void setY(int p_y);


	/**
	 *	@fn setSize(int p_width, int p_height)
	 *
	 *	@brief Sets the size of the Widget.
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@param[in] p_width Width to be set for the widget.
	 *	@param[in] p_height Height to be set for the widget.
	 *
	 *	@see setWidth(int p_width), setHeight(int p_height)
	 */
	virtual int setSize(int p_width, int p_height);

	// Setter and Getter methods for the foreground color of the object
	virtual void setForeground(WEColor &p_fgColor);


	/**
	 *	@fn getForeground()
	 *
	 *	@brief Returns the value of the Foreground Color
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns Returns the reference of the foreground color
	 *
	 */
	const WEColor& getForeground() const;


	/**
	 *	@fn setBackground(WEColor &bgColor)
	 *
	 *	@brief Sets the background color of the widget to bgColor.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] bgColor Background color to be set for the widget.
	 */
	virtual int setBackground(const WEColor &p_bgColor);


	/**
	 *	@fn getBackground()
	 *
	 *	@brief Returns the value of the Background Color
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 */
	const WEColor& getBackground() const;


	virtual void setDrawBorder(bool p_is);

	bool getDrawBorder() const
	{
		return this->drawBorderFlag;
	}

	/**
	 * Check if the current color (active border profile for current state) is completely
	 * transparent (i.e. the chosen transparent color)
	 */
	bool isCurrentBorderColorTransparent(void) const;

	virtual int setName(const std::string & p_name);

	const std::string & getName() const;

	virtual int setTransparentBorder(bool p_isTransparent);

	bool isBorderTransparent() const;

	virtual void setCornerRadius(int p_cornerRadius);

	int getCornerRadius() const
	{
		return this->cornerRadius;
	}

	/**
	 *	@fn setTransparent(bool p_isTransparent)
	 *
	 *	@brief This method is a setter for the transparency property of the WEWidget
	 *	This method is independent of the method setOpaque() which is used to set the
	 *	transparency of the surface of the WEWidget. For e.g. for the Background Color
	 *	the user can set the brush and draw a rectangle and to achieve transparency
	 *	he can set the value of this property and with in the paint() he may/may not
	 *	clip the inner area of the drawn rectangle according to the value of this property.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param p_isTransparent bool Sets the transparency for the WEWidget.
	 *
	 *	@returns WE_SUCCESS int on Successfully setting the value property.
	 *	@returns WE_SET_PROPERTY_VALUE_ERROR if the property cannot be set.
	 *
	 *	@see setOpaque
	 */
	virtual int setTransparent(bool p_isTransparent);


	bool isTransparent() const;

	/**
	 *	@fn bool isDrawBorderImage()
	 *
	 *	@brief Gets the information that the border image is drawn or not
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@returns True if the border image is drawed
	 */
	bool isDrawBorderImage() const
	{
		return drawBorderImageFlag;
	}

	/**
	 *	@fn bool isDrawBackgroundImage()
	 *
	 *	@brief Gets the information that the background image is drawn or not
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@returns True if the background image is drawed
	 */
	bool isDrawBackgroundImage() const
	{
		return drawBackgroundImageFlag;
	}

	virtual int setTransparentBorderColor(const WEColor & p_transparentBorderColor);


	const WEColor& getTransparentBorderColor() const;


	virtual int setBorderWidth(int p_borderWidth);

	virtual int setStartAngle(int p_startAngle);

	virtual int setEndAngle(int p_endAngle);

	int getBorderWidth() const;

	void update(const QRect &p_rect = QRect());

	virtual void paint(WEPainter* p_painter) = 0;


	/**
	 *	@brief Set the event listener for the specified event type.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@param[in] eventType event for which the eventListener is set.
	 *	@param[in] eventListener eventListener that is to be invoked on occurrence of an event.
	 *
	 *	@returns WE_SUCCESS when the eventType is supported by the widget.
	 *	@returns WE_EVENT_NOT_SUPPORTED_ERROR if the eventType is not supported by the widget.
	 *
	 *	@warning DANGEROUS STUFF
	 *
	 *	By default WE::mousePressEvent, WE::mouseReleaseEvent, WE::mouseMoveEvent,
	 *	WE::mouseEnterEvent, WE::mouseExitEvent, WE::keyPressEvent, WE::keyReleaseEvent
	 *	are supported.
	 *	The sequence of generation of mouse events depends on the underlying toolkit.
	 *	For Qt:
	 *	A single click produces following sequence of events:
	 *	a. mouseMoveEvent
	 *	b. mousePressEvent click count = 1
	 *	c. mouseMoveEvent
	 *	d. mouseReleaseEvent
	 *	All with the same mouse position unless the user's hand shakes.
	 *
	 *	A double click produces the following sequence of events:
	 *	a. mouseMoveEvent
	 *	b. mousePressEvent click count = 1
	 *	c. mouseMoveEvent
	 *	d. mouseReleaseEvent
	 *	e. mouseMoveEvent
	 *	f. mousePressEvent click count = 2
	 *	g. mouseMoveEvent
	 *	h. mouseReleaseEvent
	 *	All with the same mouse position unless the user's hand shakes.
	 *
	 *	mouseEnterEvent and mouseExitEvent are generated only once when the mouse enters and
	 *	leaves respectively.
	 *
	 *	No mousePressEvent/mouseReleaseEvent is generated when Alt modifer key is pressed
	 *	while clicking, only mouseMoveEvent are generated.
	 *
	 */
	int setEventListener(int p_eventType, WEToolkitEventListener * p_eventListener);

	/**
	 * @fn void keyPressEvent(QKeyEvent* p_qtKeyEvent)
	 *
	 * @brief Qt Key Press Event
	 *
	 * @author hm@wachendorff.de
	 * @author Hendrik Mueller
	 *
	 * @param p_qtKeyEvent The event data
	 */
	virtual void keyPressEvent(QKeyEvent* p_qtKeyEvent) override;


	/**
	 * @fn void keyReleaseEvent(QKeyEvent* p_qtKeyEvent)
	 *
	 * @brief Qt Key Release Event
	 *
	 * @author hm@wachendorff.de
	 * @author Hendrik Mueller
	 *
	 * @param p_qtKeyEvent The event data
	 */
	virtual void keyReleaseEvent(QKeyEvent* p_qtKeyEvent) override;

	/**
	 * @fn void mousePressEvent(QMouseEvent* p_qtMouseEvent)
	 *
	 * @brief Qt Mouse Press Event
	 *
	 * @author hm@wachendorff.de
	 * @author Hendrik Mueller
	 *
	 * @param p_qtMouseEvent The event data
	 *
	 * @warning processMouseEvents() can prevent creation of mouse events.
	 *
	 * @see processMouseEvents(bool)
	 */
	virtual void mousePressEvent(QMouseEvent* p_qtMouseEvent) override;

	/**
	 * @fn void mouseReleaseEvent(QMouseEvent* p_qtMouseEvent)
	 *
	 * @brief Qt Mouse Release Event
	 *
	 * @author hm@wachendorff.de
	 * @author Hendrik Mueller
	 *
	 * @param p_qtMouseEvent The event data
	 *
	 * @warning processMouseEvents() can prevent creation of mouse events.
	 *
	 * @see processMouseEvents(bool)
	 */
	virtual void mouseReleaseEvent(QMouseEvent* p_qtMouseEvent) override;

	/**
	 * @fn void mouseMoveEvent(QMouseEvent* p_qtMouseEvent)
	 *
	 * @brief Qt Mouse Move Event
	 *
	 * @author hm@wachendorff.de
	 * @author Hendrik Mueller
	 *
	 * @param p_qtMouseEvent The event data
	 *
	 * @warning processMouseEvents() can prevent creation of mouse events.
	 *
	 * @see processMouseEvents(bool)
	 */
	virtual void mouseMoveEvent(QMouseEvent* p_qtMouseEvent) override;

	/**
	 * @brief Converts Qt wheel events into WEToolkitEvent.
	 *
	 * @author hm@wachendorff.de
	 * @author Hendrik Mueller
	 *
	 * @param p_eventType The event type that occurred
	 * @param p_steps Steps the mouse wheel was moved.
	 *
	 * @return The created WEToolkitEvent.
	 */
	static WEToolkitEvent createWheelEvent(WEEventType p_eventType, int p_steps);

	/**
	 * @brief Converts Qt mouse events into WEToolkitEvent.
	 *
	 * @author hm@wachendorff.de
	 * @author Hendrik Mueller
	 *
	 * @param p_eventType The event type that occurred
	 * @param p_x Global X-Position of the mouse.
	 * @param p_y Global Y-Position of the mouse.
	 * @param p_isMousePressed set to true if the mouse was pressed
	 * @param p_modifiers Qt event modifiers.
	 * @param p_relX x position in parent
	 * @param p_relY y position in parent
	 *
	 * @return The created WEToolkitEvent.
	 *
	 * @warning processMouseEvents() can prevent creation of mouse events.
	 *
	 * @see processMouseEvents(bool)
	 */
	WEToolkitEvent createMouseEvent(WEEventType p_eventType, int p_x, int p_y, bool p_isMousePressed, unsigned int p_modifiers, int p_relX, int p_relY) const;

	/**
	 * @brief Converts Qt key events into WEToolkitEvent.
	 *
	 * @author hm@wachendorff.de
	 * @author Hendrik Mueller
	 *
	 * @param p_eventType The event type that occurred
	 * @param p_keyCode The key code of the key that triggered the event.
	 * @param p_modifiers Qt event modifiers.
	 *
	 * @return The created WEToolkitEvent. If p_keyCode is unknown, the eventType is set to WE_KEY_UNKNOWN.
	 */
	static WEToolkitEvent createKeyEvent(WEEventType p_eventType, int p_keyCode, unsigned int p_modifiers);

	/**
	 *	@fn setInplaceEditingListener(WE::WEInplaceEditingListener* listener);
	 *
	 *	@brief This method is used to register the Listener for this Inplace Editing Widget. The
	 *	listener can listen to/recieve the events onEditingValueChanged(),onEditingConfirmed() and onEditingCancelled().
	 *
	 *	@see WEInplaceEditingListener for further Details
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param[in] p_listener WEInplaceEditingListener Listener for the Inplace Editing events onEditingValueChanged(),
	 *	onEditingConfirmed() and onEditingCancelled().
	 *
	 *	@returns WE_SUCCESS on successfully registering a listener.
	 *
	 */
	int setInplaceEditingListener(WE::WEInplaceEditingListener* p_listener);


	/**
	 *	@brief Get the event listener associated with the specified event type.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@param[in] eventType event for which the eventListener is requested.
	 *
	 *	@returns EventListener pointer for the listener associated with the eventType.
	 *	@returns nullptr if no listener is set for the eventType or event is not supported.
	 *
	 *	By default no listener is set for any of the default supported events.
	 */
	WEToolkitEventListener * getEventListener(int p_eventType) const;

	/**
	 *	@brief Send the specified event to the widget. Triggers the corresponding event
	 *		listener if present.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@param[in] p_event event to be sent to the widget.
	 *
	 *	@returns WE_SUCCESS if event listener is found and invoked from within the method.
	 *	@returns WE_EVENT_NOT_SUPPORTED_ERROR if no listener is set for the eventType or event is not supported.
	 *
	 *	This method returns after the corresponding eventListener has been invoked and
	 *	finished its toolkitEventOccurred() call.
	 *	This is mainly used when borderstate is set - qt_we_widget for example calls
	 *	toolkitEventOccurred() of the respective listener by itself.
	 *
	 */
	virtual int sendEvent(WEToolkitEvent p_event);


	/**
	 *	@fn int add(WEWidget *widget)
	 *
	 *	@brief Add widget as a child of this widget and set it's parent pointer accordingly.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_widget - pointer to widget that should become a child of this one
	 *
	 *	@returns WE_ADD_WIDGET_ERROR if nullptr == p_widget
	 *	@returns WE_SUCCESS otherwise
	 *	<MKA 19.03.2010 made virtual so could use it in WEListWidget
	 */
	virtual int add(WEWidget * p_widget);


	/**
	 *	@fn void updateClipping(WEWidget * child)
	 *
	 *	@brief Enables clipping for this widget if the child needs to be clipped
	 *	       for the same visual result as before with QWidgets
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param[in] child - A child of this widget
	 */
	void updateClipping(WEWidget * p_child);


	//void calculateBorderPosition();



	/**
	 *	@fn bool pointWithinWidget(int p_x, int p_y);
	 *
	 *	@brief Figure out if position e.g. of a touch event is inside this widget.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_x - Horizontal screen coordinate of the point to be checked.
	 *	@param[in] p_y - Vertical screen coordinate of the point to be checked.
	 *
	 *	@returns true if the point is inside the bounding rect of the widget
	 *	@returns false otherwise
	 *
	 *	@see bool pointCloseToWidget(int p_x, int p_y, unsigned int p_distance)
	 */
	bool pointWithinWidget(int p_x, int p_y) const;


	/**
	 *	@fn bool pointCloseToWidget(int p_x, int p_y, int p_distance)
	 *
	 *	@brief Figure out if position e.g. of a touch event is inside this widget or within p_distance around.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_x - Horizontal screen coordinate of the point to be checked.
	 *	@param[in] p_y - Vertical screen coordinate of the point to be checked.
	 *	@param[in] p_distance - Points within this distance around the widget still return "true".
	 *
	 *	@returns true if the point is inside the bounding rect or not more than p_distance outside of the widget
	 *	@returns false otherwise
	 *
	 *	@see bool pointWithinWidget(int p_x, int p_y)
	 */
	bool pointCloseToWidget(int p_x, int p_y, int p_distance) const;


	/**
	 *	@fn WEWidget * getBetterWidgetUnderCursor(int p_x, int p_y)
	 *
	 *	@brief Find the highest widget below this one which should get a touch or mouse event.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_x - Horizontal screen coordinate of the point to be checked.
	 *	@param[in] p_y - Vertical screen coordinate of the point to be checked.
	 *
	 *	@returns Pointer to the highest input widget below this one whose bounding rect covers p_x, p_y.
	 *	@returns nullptr if there is no better input object
	 *
	 *	This method finds the touched object e.g. below the pointer we create for every pointer DDO.
	 *	To find the widget a recursion of getInputWidgetUnderCursor() is triggered starting from the lowest
	 *	widget i.e. the main window and searching up until this widget is reached.
	 *
	 *	@see void getInputWidgetUnderCursor(WEWidget *, WEWidget * *, int, int)
	 */
	WEWidget * getBetterWidgetUnderCursor(int p_x, int p_y);


	/**
	 *	@fn void getInputWidgetUnderCursor(WEWidget * p_below, WEWidget * * p_toBeTouched, int p_x, int p_y)
	 *
	 *	@brief Recursively find the highest widget below p_below which should get a touch or mouse event.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_below - Widget where searching should terminate.
	 *	@param[out] p_toBeTouched - Pointer to be filled with the found widget.
	 *	@param[in] p_x - Horizontal screen coordinate of the point to be checked.
	 *	@param[in] p_y - Vertical screen coordinate of the point to be checked.
	 *
	 *	@returns Pointer to the highest input widget below p_below whose bounding rect covers p_x, p_y.
	 *	@returns nullptr if there is no such input object
	 *
	 *	This method finds the touched object e.g. below the pointer we create for every pointer DDO.
	 *	It's typically triggered by WEWidget * getBetterWidgetUnderCursor(int p_x, int p_y).
	 *	Hidden containers and their children are ignored in the search.
	 *
	 *	@see WEWidget * getBetterWidgetUnderCursor(int p_x, int p_y)
	 */
	void getInputWidgetUnderCursor(WEWidget * p_below, WEWidget * * p_toBeTouched, int p_x, int p_y);

	/**
	 *	@fn getChildWithController()
	 *
	 *	@brief Retrieve the first child widget containing the controller.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns The child widget containing the controller.
	 *	@returns nullptr if there is not child with a controller.
	 */
	WEWidget * getChildWithController() const;


	// Navigation related functions

	/**
	 *	@fn isInputTypeWidget()
	 *
	 *	@brief Check whether the widget is an input widget or not.
	 *
	 *
	 *	@returns TRUE if the widget itself can change a value based on operator input, FALSE otherwise.
	 *
	 *	It does NOT require that the widget is currently visible to the operator.
	 *
	 *	@see bool isInputWidget()
	 */
	virtual bool isInputTypeWidget() const;


	/**
	 *	@fn isInputWidget()
	 *
	 *	@brief Check whether the widget is an input widget or not.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns TRUE if the widget is an input widget, FALSE otherwise.
	 *
	 *	Only an input widget can get the focus during navigation.
	 *	So this method checks if this widget itself is an enabled and visible input or if
	 *	it has children that are.
	 *	This method is used for navigation - so it requires that the widget is currently visible.
	 *
	 *	@see bool isInputTypeWidget()
	 */
	virtual bool isInputWidget() const;


	virtual int setInputWidget(bool p_inputState);

	bool isSetAsInput(void) const;


	/**
	 *	@fn getBorderState()
	 *
	 *	@brief Retreive the current state of the widget.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns The current state of the widget
	 *	@returns WE::DISABLED if the widget is disabled and cannot get the focus.
	 *	@returns WE::OPEN_FOR_INPUT if the widget value can be edited.
	 *	@returns WE::ENABLED_WITH_FOCUS if the widget is an input widget and has the focus.
	 *	@returns WE::ENABLED_WITHOUT_FOCUS if the widget is an input widget but does not have the focus right now.
	 *
	 */
	WEBorderState getBorderState() const;

	/**
	 *	@fn setBorderState(int state)
	 *
	 *	@brief Change the state of the widget.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@param[in] state The expected new state of the widget.
	 *		By default following states are supported:
	 *		WE::DISABLED if the widget is to be disabled and cannot get the focus.
	 *		WE::OPEN_FOR_INPUT if the widget value can be edited.
	 *		WE::ENABLED_WITH_FOCUS if the widget is an input widget and has the focus.
	 *		WE::ENABLED_WITHOUT_FOCUS if the widget is an input widget but does not have the focus right now.
	 *
	 *	@returns WE_SUCCESS if the state is changed sucessfully.
	 *	@returns WE_STATE_NOT_SUPPORTED_ERROR if the requested state is not supported by the widget.
	 */
	virtual int setBorderState(int p_state);


	/**
	 *	@fn getParent()
	 *
	 *	@brief Retreive the parent widget containing this widget.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns The parent widget containing the widget.
	 *	@returns nullptr if it is top level widget.
	 */
	WEWidget* getParent() const;


	void setParent (WEWidget * p_newParentWidget);


	/**
	 *	@fn hasChildren()
	 *
	 *	@brief Check whether the widget contains child widget under it or not.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns TRUE if the widget contains child widgets under it, FALSE otherwise.
	 *
	 */
	virtual bool hasChildren() const;


	/**
	 *	@fn hasInputChildren()
	 *
	 *	@brief Check whether the widget contains at least one editable child widget under it or not.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@returns TRUE if the widget contains editable child widgets under it, FALSE otherwise.
	 *
	 */
	virtual bool hasInputChildren() const;


	/**
	 *	@fn void removeChild(WEWidget * p_child)
	 *
	 *	@brief Remove a child pointer from the list of children.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param[in] p_child - pointer to widget to be removed from the list
	 *
	 *	@returns nothing - but [s|c]hould return an error code
	 *
	 *	@warning This will only remove the pointer to the widget from the list - not the widget itself.
	 *
	 *	This method can help cleaning up the tree of widgets if is not destroyed completely.
	 *	It does not remove the widget itself but only the reference this widget keeps in its
	 *	list of children.
	 */
	virtual void removeChild(WEWidget * p_child);


	/**
	 *	@fn hasSiblings()
	 *
	 *	@brief Check whether the widget has siblings i.e other children of its parent.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns TRUE if the widget has siblings, FALSE otherwise.
	 *	@returns FALSE if the parent is nullptr.
	 */
	bool hasSiblings() const;


	/**
	 *	@fn getFirstChild()
	 *
	 *	@brief Retrieve the first child widget of this widget.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns The first child of the widget
	 *	@returns nullptr if the widget has no child widgets under it.
	 */
	virtual WEWidget * getFirstChild() const;
	virtual WEWidget * getLastChild() const;
	virtual WEWidget * getFirstInputChild() const;
	virtual WEWidget * getLastInputChild() const;


	/**
	 *	@fn std::vector<WEWidget *> * getChildren(void)
	 *
	 *	@brief Return vector of child widgets of this widget
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *
	 *	@param none
	 *
	 *	@returns A pointer to a vector of pointers to child widgets of this widget
	 *
	 *	@warning This returns a pointer to an internal structure - it is intended to only READ its content.
	 *
	 */
	virtual childrenBiMap * getChildren(void) const
	{
		return this->children;
	}


	/**
	 *	@fn getNextSibling()
	 *
	 *	@brief Retrieve the sibling next in order to this widget.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns The next sibling of the widget.
	 *	@returns nullptr if the widget has no next sibling or it is the last child of its parent.
	 */
	WEWidget* getNextSibling();
	WEWidget* getNextInputSibling() const;


	/**
	 *	@fn getPreviousSibling()
	 *
	 *	@brief Retrieve the sibling previous in order to this widget.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns The previous sibling of the widget.
	 *	@returns nullptr if the widget has no previous sibling or it is the first child of its parent.
	 */
	WEWidget* getPreviousSibling();
	WEWidget* getPreviousInputSibling() const;


	/**
	 *	@fn getFirstInputSibling()
	 *
	 *	@brief Retrieve the first input sibling.
	 *
	 *	@author hendrik.mueller@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns The first input sibling.
	 *	@returns nullptr if no input sibling was found.
	 */
	WEWidget* getFirstInputSibling() const;


	/**
	 *	@fn getLastInputSibling()
	 *
	 *	@brief Retrieve the last input sibling.
	 *
	 *	@author hendrik.mueller@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns The last input sibling.
	 *	@returns nullptr if no input sibling was found.
	 */
	WEWidget* getLastInputSibling() const;


	// Calculation of the Position of the Object inside of the Border

	/**
	 *	@fn getXPositionInsideBorder()
	 *
	 *	@brief Calculate the position of the object inside of the frame
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *
	 *	@returns integer value for the object position
	 *
	 *	This method returns the position of the object inside of the border
	 *  Calculates also if a border if drawn or not.
	 *
	 */
	int getXPositionInsideBorder() const;

	/**
	 *  @copydoc int getXPositionInsideBorder()
	 */
	int getYPositionInsideBorder() const;


	int getWidthInsideBorder() const;


	int getHeightInsideBorder() const;


	/**
	 *	@fn void drawBorderProfile(WEPainter * p_painter, bool drawEllipse = false)
	 *
	 *	@brief draws border profile
	 *
	 *	@author kw@wachendorff.de
	 *	@author Katsiaryna Walesch
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] p_painter pointer to painter object
	 *  @param[in] p_drawEllipse if the border is an ellipse
	 *
	 *	@returns void
	 *
	 *	This method returns the position of the object inside of the border
	 *
	 */
	void drawBorderProfile(WEPainter * p_painter, bool p_drawEllipse = false);


	void updateBorderProfileNodes(WEControllerNode *p_controller, bool p_drawEllipse);


	void drawBorderImage(int p_x, int p_y, WEPainter * p_painter);


	void updateBorderImageNode(WEControllerNode *p_controller);


	void drawBorder(WEPainter * p_painter, bool p_ellipse = false);

	// TODO: Nowhere this is called with 'true'?! Instead arched bargraph has a similar method with only one parameter?! And none of them seems 'virtual'.
	void updateBorderNodes(WEControllerNode *p_controller, bool p_ellipse = false);


	// Note: Before 9. Nov. 2011 "this->m_isTransparent" switched drawing the whole background
	// after that we decided to draw the background image whenever it is present and by "this->m_isTransparent" only switch background color
	void drawBackground(WEPainter* p_painter, bool p_ellipse = false);

	void updateBackgroundNodes(WEControllerNode *p_controller, bool p_ellipse = false);


	void drawBackgroundImage(int p_x, int p_y, WEPainter * p_painter);
	void drawBackgroundImage(WEPainter * p_painter);

	void updateBackgroundImageNode(WEControllerNode *p_controller, int p_x=0, int p_y=0, const WEColor& p_svgBGColor=WEColor::transparentColor);


	/**
	 *	@fn createBackgroundImage(std::string p_strFile)
	 *
	 *	@brief creates the background image
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param p_strFile path and name of the image file to be used as background
	 *
	 *	@returns WE_SUCCESS or failure
	 *
	 *	creates background image and stores pointer
	 *
	 *	@see createImage(std::string p_strFile)
	 */
	virtual int createBackgroundImage(const std::string & p_strFile);


	/**
	 *	@fn createBorderImage(std::string p_strFile)
	 *
	 *	@brief creates border image
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns WE_SUCCESS or failure
	 *
	 *	creates a border image and stores the pointer in the widget
	 *
	 */
	int createBorderImage(const std::string & p_strFile);


	/**
	 *	@fn getBorderProfile()
	 *
	 *	@brief returns pointer to border profile iterface
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns pointer
	 *
	 * 	@warning pointer can be nullptr!
	 *
	 */
	WEBorderProfile* getBorderProfile() const
	{
		return borderProfile;
	}


	/**
	 *	@fn WEImage * getBackgroundImage()
	 *
	 *	@brief Gets the image of the background
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@returns The background image pointer
	 *
	 */
	WEImage * getBackgroundImage() const
	{
		return this->backgroundImage;
	}


	/**
	 *	@fn WEImage * getBorderImage()
	 *
	 *	@brief Gets the image of the border
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@returns The border image pointer
	 *
	 */
	WEImage * getBorderImage() const
	{
		return this->borderImage;
	}


	/**
	 *	@fn int setBackgroundImage(WEImage * backgroundImage)
	 *
	 *	@brief Set the image of the background
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@returns WE_ERROR code
	 *
	 */
	virtual int setBackgroundImage(WEImage * p_backgroundImage);


	/**
	 *	@fn int setBorderImage(WEImage * borderImage)
	 *
	 *	@brief Set the image of the border
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@returns WE_ERROR code
	 *
	 */
	virtual int setBorderImage(WEImage * p_borderImage);


	/**
	 *	@fn setBorderProfile(WEBorderProfile* p_borderProfile)
	 *
	 *	@brief Sets a border profile as a pointer - i.e. we share somebody else's border profile.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 * 	@param[in] p_borderProfile pointer to border profile
	 *
	 *	@returns WE_ERROR code
	 *
	 *	@warning Do not confuse with the similar method taking a reference.
	 */
	virtual int setBorderProfile(WEBorderProfile * p_borderProfile);


	/**
	 *	@fn setBorderProfile(WEBorderProfile* p_borderProfile)
	 *
	 *	@brief Sets a border profile and creates an internal copy
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 * 	@param[in] p_borderProfile object
	 *
	 *	@returns WE_ERROR code
	 *
	 */
	int setBorderProfile(const WEBorderProfile & p_borderProfile);


	/**
	 *	@fn createBorderProfile()
	 *
	 *	@brief Creates an internal border profile if there is none yet, is called as default.
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@returns WE_ERROr code
	 *
	 */
	int createBorderProfile();


	virtual void setVisible(bool p_visible);


	/**
	 *	@brief Forward information about visibility change to (grand*) parent
	 */
	virtual void childVisibilityChanged(WEWidget * p_child, bool p_visible);


	/**
	 *	@fn bool shouldBeVisible()
	 *
	 *	@brief returns if a widget is visible or not
	 *
	 *	@author ml@wachendorff.de
	 *
	 *	@returns true or false
	 *
	 *	@see bool isVisible()
	 */
	bool shouldBeVisible() const
	{
		return this->weVisible;
	}


	/**
	 *	@fn int setActive(bool p_visible)
	 *
	 *	@brief sets a flag inside the widget;
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @param[in] p_active true or false
	 *
	 *  @see setVisible
	 *
	 *  sets a internal flag; if true the widget will draw if false the widget draws nothing.
	 *
	 *  Why? During implementation of WEList and adding string widgets instead of strings
	 *  I faced a problem of very frequent redrawing when using setVisible. If you use it
	 *  the widget is visible but will not going to draw something on the surface.
	 *
	 *  That frequent redrawing causes some infinite loops.
	 *
	 *  To achive that goal, the widgets have to add that flag in the paint routine.
	 */
	int setActive(bool p_active);


	/**
	 *	@fn bool isActive()
	 *
	 *	@brief sets a flag inside the widget;
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *  @see setActive
	 *
	 *  @returns if the widget draws or not;
	 */
	bool isActive() const
	{
		return this->active;
	}


	/**
	 *	@fn int takeScreenshot(std::string p_fileName, std::string p_folderName = "")
	 *
	 *	@brief Take a screenshot of the current widget and save it
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_fileName Name of the screenshot file
	 *	@param p_folderName Folder where screenshot should be stored ("" for current working directory)
	 *
	 *	@returns WE_SUCCESS
	 *	@returns WE_FILE_ERROR
	 *
	 *	p_fileName has to end in a suffix indicating the desired image format, e.g. "myScreenshot.png"
	 *	will create "portable network graphic" format.
	 *
	 *	The screenshot will be stored in the folder indicated by p_folderName.
	 *	If p_folderName is "" or p_folderName does not point to a valid folder
	 *	the screenshot will be saved in "<APP_DATA_PARTITION>/screenshots" and if this does not exist
	 *	the current working directory is used.
	 *
	 */
	int takeScreenshot(const std::string & p_fileName, const std::string & p_folderName = "");


	/**
	 *	@fn void setValueToValid(bool p_valid)
	 *
	 *	@brief This triggers to show the disabled pattern or not.
	 *
	 *	@returns nothing
	 *
	 *	@param p_valid Set to false if disabled pattern should be shown.
	 *
	 *	This method makes sure the disabled pattern is shown on top of this widget.
	 *	Some widgets overwrite this method to propagate the property to their children.
	 *
	 */
	virtual void setValueToValid(bool p_valid);
	bool isValueValid(void) const;


	/**
	 *	@fn setOpaque(bool opaque)
	 *
	 *	@brief This method is used to set the transparency for the canvas(surface) of the
	 *	WEWidget. This method is independent of the method setTransparent() which is
	 *	a setter for a property and has an effect according to the user. For
	 *	e.g. For the Background Color the user sets the brush and draws a rectangle
	 *	and to achieve transparency he can set the property for the transparency using
	 *	setTransparent and with in the paint() he may/may not clip the inner area of the drawn
	 *	rectangle according to the value of the property.
	 *
	 *	@author maneesh.arora@aricent.com
	 *	@author Maneesh Arora
	 *
	 *	@param opaque bool Sets the transparency of the WEWidget.
	 *
	 *	@see setTransparent
	 */
	void setOpaque(bool p_opaque);


	void resetTransparentRegion();


	void drawTransparentRegion(WEPainter * p_painter);


	void setPenStyle(WEPenStyle p_style, unsigned short p_pattern = 0xffff);


	void processMouseEvents(bool p_yes);


	void setLineWidth(unsigned int p_width)
	{
		this->pen.width = p_width;
	}


	/**
	 *	@fn void setBackgroundTexture(WEImage * p_img)
	 *
	 *	@brief This sets the image to be used as background pattern and default display coordinates of its origin.
	 *
	 *	@returns nothing
	 *
	 *	@param p_img Image to be used as background pattern. Set to nullptr for "no pattern".
	 *
	 *	This method sets the image for the background pattern. If p_img == nullptr the pattern is switched off.
	 *	The pattern is created by copying p_img repeatedly to the left and bottom, starting from position
	 *	0, 0 of the screen.
	 *
	 *	@see setBackgroundTexture(WEImage *, WEPoint &)
	 */
	void setBackgroundTexture(WEImage * p_img);


	/**
	 *	@fn void setBackgroundTexture(WEImage * p_img, WEPoint & p_origin)
	 *
	 *	@brief This sets the image to be used as background pattern and the display coordinates of its origin.
	 *
	 *	@returns nothing
	 *
	 *	@param p_img Image to be used as background pattern. Set to nullptr for "no pattern".
	 *	@param p_origin position of the pattern origin on display
	 *
	 *	This method sets the image for the background pattern. If p_img == nullptr the pattern is switched off.
	 *	The pattern is created by copying p_img repeatedly to the left and bottom, starting from position
	 *	p_origin.
	 *	For ISO-VT p_origin has to be set to the top left corner of the data mask area.
	 *
	 */
	void setBackgroundTexture(WEImage * p_img, const WEPoint & p_origin);


	/**
	 *	@fn void setIsInputRoot(bool p_isInputRoot);
	 *
	 *	@brief  Use to set if this widget is a "Input Root Widget".
	 *		That means that we don't want to search after other input widgets behind this one.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@returns nothing
	 *
	 *	@param p_isInputRoot "True" this is a "Input Root Widget"
	 *
	 *
	 *	NOTE:
	 *	This function is used by "Virtual Keyboard" of PClient and ISO-VT.
	 *	If a area is touched where is no input button, we don't want to
	 *	search after other inputs behind the virtual keyboard.
	 *
	 */
	void setIsInputRoot(bool p_isInputRoot);

	/**
	 *	@fn bool isInputRoot();
	 *
	 *	@brief Returns if this widget is a "Input Root Widget"
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@returns "True" this is a "Input Root Widget"
	 *		 "False" this is not a "Input Root Widget"
	 */
	virtual bool isInputRoot() const;

	static void setScrollBarWidth(int p_newWidth)
	{
		// std::cout << "WEWidget::setScrollBarWidth: From " << scrollBarWidth << " to  " << newWidth << "." << std::endl;
		scrollBarWidth = p_newWidth;
	}


	static int getScrollBarWidth( void )
	{
		return scrollBarWidth;
	}


	void addBelow(WEWidget * p_toAdd, WEWidget * p_belowThisOne);


	virtual const std::string & getDisplayedValue(void) const
	{
		static const std::string empty = "";
		return empty;
	}

	/**
	 *	Print out the whole subtree of this widget recursively.
	 *
	 *	@param p_whitespace is used in recursion - call with ""
	 */
	void printSubtree(const std::string & p_whitespace ) const;

	void printChildrenNames(void) const;


	/**
	 *	@fn WEWidget * getReparentedChild( void )
	 *
	 *	@brief Return current value of pointer to reparented child.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns A pointer to the reparented child
	 *
	 *
	 *	Reparenting is used e.g. when a list is edited. Even if the edited list is larger
	 *	because of showing multiple items and a scroll bar it will not be hidden behind other
	 *	widgets if it's a direct child of the window.
	 *
	 *	Deleting this widget while the child is reparented will make it an orphan which will
	 *	be hard to remove.
	 *
	 *	With this method one can figure out that there is something to take care of BEFORE
	 *	deleting this widget.
	 *
	 *	@see setReparentedChild( WEWidget * )
	 */
	WEWidget * getReparentedChild( void ) const;


	/**
	 *	@fn void setReparentedChild( WEWidget * p_child )
	 *
	 *	@brief Set current value of pointer to reparented child.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param A pointer to the reparented child
	 *
	 *	Reparenting is used e.g. when a list is edited. Even if the edited list is larger
	 *	because of showing multiple items and a scroll bar it will not be hidden behind other
	 *	widgets if it's a direct child of the window.
	 *
	 *	Deleting this widget while the child is reparented will make it an orphan which will
	 *	be hard to remove.
	 *
	 *	With this method one can figure out that there is something to take care of BEFORE
	 *	deleting this widget.
	 *
	 *	@see WEWidget * getReparentedChild( )
	 */
	void setReparentedChild( WEWidget * p_child );

	/*
		 * @DEPRECATED
		 *
		 * @see bool sendToggleMessageToWindowManager( void )
		 * @see bool sendAlarmPriorityMessageToWindowManager(int priority)
		 */
		bool sendMessageToWindowManager(const std::string & p_message);


		/**
		 *	@fn bool sendToggleMessageToWindowManager(void)
		 *
		 *	@brief Send message to window manager in order to switch to next window
		 *
		 *	@author mlenhart@topcon.com
		 *	@author Michael Lenhart
		 *
		 *	@return true in case of success.
		 *	@return false otherwise
		 *
	´	 *	This method can only succeed if this instance represents a window.
		 *
		 *	@see int QTWEApplication::sendToggleMassage(void);
		 */
		bool sendToggleMessageToWindowManager( void );


		/**
		 *	@fn bool sendAlarmPriorityMessageToWindowManager(int priority)
		 *
		 *	@brief Send message to window manager in order to set priority of this window.
		 *
		 *	@author mlenhart@topcon.com
		 *	@author Michael Lenhart
		 *
		 *	@return true in case of success.
		 *	@return false otherwise
		 *
		 *	This method can only succeed if this instance represents a window.
		 *
		 *	@see int QTWEApplication::setAlarmPriority(unsigned int p_winID, int p_priority);
		 */
		bool sendAlarmPriorityMessageToWindowManager(int p_priority);


		/**
		 * @fn bool eventFilter(QObject *obj,  QEvent *event)
		 *
		 * @brief Filters events if this object has been installed as an event filter.
		 *
		 * @author mohd.faisal@aricent.com
		 * @author Mohd Faisal
		 *
		 * @param obj	the monitored object (main window) where the filter was installed
		 * @param event The event data
		 *
		 * @returns false we do not want to filter the event out i.e. stop it being handled further
		 *
		 * @see QTWEWidget() constructor
		 * @see QObject::installEventFilter
		 */
		bool eventFilter(QObject *p_obj,  QEvent *p_event) override;

		/**
		 * @fn void windowActiveEvent(QEvent * event)
		 *
		 * @brief Qt Window Active Event when the main window gets the input focus.
		 *	Triggers the listeners if registered for WE::TOOLKIT_WINDOW_ACTIVE_EVENT event.
		 *
		 * @author mohd.faisal@aricent.com
		 * @author Mohd Faisal
		 *
		 * @param event The event data
		 *
		 * @see eventFilter()
		 */
		void windowActiveEvent(QEvent * p_event);

		/**
		 * @fn void windowDeactiveEvent(QEvent * event)
		 *
		 * @brief Qt Window Deactive Event when the main window looses input focus.
		 *	Triggers the listeners if registered for WE::TOOLKIT_WINDOW_DEACTIVE_EVENT event.
		 *
		 * @author mohd.faisal@aricent.com
		 * @author Mohd Faisal
		 *
		 * @param event The event data
		 *
		 * @see eventFilter()
		 */
		void windowDeactiveEvent(QEvent * p_event);

		/**
		 *	@fn void hideEvent ( QHideEvent * p_qtHideEvent )
		 *
		 *	@brief Forward hide event e.g. to release pressed buttons.
		 *
		 *	@author ml@wachendorff.de
		 *	@author Michael Lenhart
		 *
		 *	@param p_qtHideEvent - QEvent without much additional information.
		 *
		 *	@returns nothing.
		 *
		 *	Like the parameter suggests this is a reimplementation of a QWidget method.
		 */
		void hideEvent ( QHideEvent * p_qtHideEvent );


		/**
		 * @fn void wheelEvent(QWheelEvent* p_qtWheelEvent)
		 *
		 * @brief Qt Wheel Event
		 *
		 * @author hm@wachendorff.de
		 * @author Hendrik Mueller
		 *
		 * @param p_qtWheelEvent The event data
		 */
		void wheelEvent(QWheelEvent* p_qtWheelEvent) override;


		int getStartAngle(void) const;
		int getEndAngle(void) const;
		int getSpanAngle(void) const
		{
			return this->spanAngle;
		}
		bool getDrawBorderImage(void) const;
		bool getDrawBackgroundImage(void) const;

		/**
		 * Function used to determine if a mouse event really hit the painting area of a widget.
		 * Default implementation returns true.
		 * Overwrite for non-rectangular widgets (see arched bargraph, meter, gauge...)
		 */
		virtual bool amIHit(int p_globalX, int p_globalY) const;

		QQuickItem *childAt(qreal p_x, qreal p_y) const;

		void move(int p_x, int p_y);
		virtual void setPosition(float p_x, float p_y);
		virtual void resize(int p_w, int p_h);
		void setGeometry(int p_x, int p_y, int p_w, int p_h);

		void show();
		void hide();

		void bringToFront();
		void sendToBack();

		void raise();
		void lower();

		void stackUnder(WEWidget *p_w);

		virtual void repaint();

		WId winId() const;
		bool isActiveWindow() const;

		bool event(QEvent *p_event) override;
		void handleTouchBeginEvent(const QTouchEvent::TouchPoint & p_tp, WEWidget * p_forwarder = nullptr);
		void handleTouchEndEvent(const QTouchEvent::TouchPoint & p_tp, WEWidget * p_forwarder = nullptr);
		void handleTouchUpdateEvent(QEvent *p_event, WEWidget * p_forwarder = nullptr);


	/**
	 * Enable or disable multi touch
	 */
	static void enableMultiTouch(bool p_enable);
	static bool isMultiTouchEnabled(void)
	{
		return WEWidget::multiTouchEnabled;
	}

protected:

		friend class QTWEScrollArea;	//!< Allow QTWEScrollArea to call keyPressEvent and KeyReleaseEvent functions

	// ==============================================================================================


	/**
	 *	@fn int drawDisabledPattern(WEWidget * p_widget, WEPainter * p_painter, int p_color, bool p_elipse)
	 *
	 *	@brief Draw a brush pattern above a widget to mark it as DISABLED
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_painter Pointer to WEPainter currently used
	 *	@param p_color An int from Color enum representing the color in which the pattern shall be drawn
	 *	@param p_ellipse flag if we are painting the pattern to an ellipsoid or rectangle widget
	 *	@param p_isInvalidValue If true, only overlay is created, otherwise a St. Andrews cross is drawn on top additionally
	 *
	 *	@returns WE_SUCCESS
	 *
	 */
	int drawDisabledPattern(WEPainter * p_painter, Color p_color, bool p_ellipse, bool p_isInvalidValue = false);


	/**
	 *
	 *	@param p_controller
	 *	@param p_color An int from Color enum representing the color to be taken if there's no color defined for the 'disabled' border state.
	 *	@param p_ellipse 'true' for widgets such as meter, gauge, arched bargraph, 'false' for rectangular ones
	 *	@param p_isInvalidValue	If 'false' an St. Andrews cross is added to the disabled pattern (to clearly distinguish from background).
	 *	@param p_ellipseLineWidth Keep as default (0) to get a filled pattern.
	 */
	void updateDisabledPatternNodes(WEControllerNode *p_controller, Color p_color, bool p_ellipse, bool p_isInvalidValue, int p_ellipseLineWidth=0);


	void calculateSpanAngle();


	/**
	 *	@fn void repaintBackgroundPixmap()
	 *
	 *	@brief this method causes that the pixmap background will be updated
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@returns nothing
	 *
	 *	Remember to do something in 'createBackgroundPixmap()' because something otherwise
	 *	considered 'static' has changed, e.g. border state, background color ...
	 *
	 *	Additionally this calls 'repaint()' so that the flag gets evaluated (if the
	 *	widget implements the 'static content pixmap') and the whole widget gets refreshed.
	 *
	 *	Most of the times refresh worked even without 'repaint()' here - but there were exceptions.
	 */
	void repaintBackgroundPixmap();


	/**
	 *	@brief Create a pixmap of the static content of this widget or update the existing pixmap.
	 *
	 *	@returns nothing
	 *
	 *	The idea here is to create a pixmap of the more or less static content of
	 *	the widget, assuming it will be more efficient to use that when repainting
	 *	instead of re-constructing the whole widget at every paint event.
	 *
	 *	@note: Even though this most likely includes background image and background color, it's NOT the same.
	 *
	 *	This is a generic dummy doing nothing. Widgets which want to use the pixmap
	 *	have to override it.
	 *
	 */
	virtual void createBackgroundPixmap();


	/**
	 *	@brief Create object instances needed for the pixmap
	 *
	 *	@returns nothing
	 *
	 *	The idea here is to create a pixmap of the more or less static content of
	 *	the widget, assuming it will be more efficient to use that when repainting
	 *	instead of re-constructing the whole widget at every paint event.
	 *
	 *	@note: Even though this most likely includes background image and background color, it's NOT the same.
	 *
	 *	This method creates generic infrastructure i.e. object instances and calls
	 *	'paintBackgroundGraphics()' to fill it.
	 *
	 *	@see paintBackgroundGraphics(WEPainter *) for what really is considered 'static' for a specific widget
	 */
	void createAndFillBackgroundPixmap();


	/**
	 *	@fn WEWidgetTextureCacheKey getBackgroundTextureCacheKey(bool useBorderProperties=true) const
	 *
	 *	@brief Create a texture cache key from the current state of the widget
	 *
	 *	@author preths@topcon.com
	 *	@author Patrick Reths
	 *
	 *	@param useBorderProperties The border specific properties will be set if this is true.
	 *	                           (Set this to false if the widget's background pixmap doesn't contain the border.)
	 *	                           (Only WEArchedbargraphWidget does that currently.)
	 *
	 *	@returns the created WEWidgetTextureCacheKey
	 */
	WEWidgetTextureCacheKey getBackgroundTextureCacheKey(bool p_useBorderProperties=true) const;


	/**
	 *	Seems this is currently only called from
	 *	'WEMeterWidget::createBackgroundPixmap()' and
	 *	'WEGaugeWidget::createBackgroundPixmap()'
	 *	'WEArchedbargraphWidget::updatePaintNode()' has an option to call it but doesn't
	 *
	 *	I.e. all widgets actually implementing the 'static content pixmap' call this method
	 *	from their createBackgroundPixmap() to trigger WEWidget::createAndFillBackgroundPixmap().
	 *
	 *	Implemented in header because of template
	 */
	template <class Cache, class T>
	void createCachedBackgroundTexture(Cache & p_cache, const T & p_key)
	{
		if (this->updateBackgroundPixmap)
		{
			this->updateBackgroundPixmap = false;	// make sure we repaint again if set while painting below

			WETextureCache::Data *cacheData = p_cache.getCacheData(p_key);

			if (nullptr != cacheData)
			{
				//std::cout << "WEWidget::createCachedBackgroundTexture: Use texture from cache" << std::endl;
				if (nullptr == backgroundPixmapImage)
				{
					this->backgroundPixmapImage = new WEImage();
				}
				this->backgroundPixmapImage->createRawImage(cacheData->texture, cacheData);
			}
			else
			{
				//std::cout << "WEWidget::createCachedBackgroundTexture: Create new texture" << std::endl;
				// This method is (only) called from 'WEMeterWidget::createBackgroundPixmap()' and
				// 'WEGaugeWidget::createBackgroundPixmap()'
				this->createAndFillBackgroundPixmap();

				if (nullptr != this->backgroundPixmapImage)
				{
					std::shared_ptr<ToolkitWEImage> toolkitImage = this->backgroundPixmapImage->getToolkitInterface();

					QSGTexture *texture = nullptr;

					if (toolkitImage)
					{
						texture = toolkitImage->getTexture();
					}

					if (nullptr != texture)
					{
						cacheData = p_cache.addTexture(p_key, texture);

						if (toolkitImage)
						{
							toolkitImage->setCacheData(cacheData);
						}
					}
				}
			}
		}
	}


	/**
	 *	@fn void clearBackgroundPixmap()
	 *
	 *	@brief clears the current background pixmap
	 *
	 *	@author afaust@topcon.com
	 *	@author Alexander Faust
	 *
	 *	@returns nothing
	 *
	 *	Clear the pixmap containing the 'mostly static' parts of this widget
	 */
	void clearBackgroundPixmap();


	/**
	 *	@brief Empty dummy implementation to be overwritten.
	 *
	 *	How should I know that I want to overwrite if its purpose is undocumented?
	 *	Seems this method is supposed to fill the 'pixmap of static content' with
	 *	e.g. border, ticks, background color ... of this widget
	 */
	virtual void paintBackgroundGraphics(WEPainter* p_painter);


	void updateParentIsTransparentForMouseEvents(bool p_parentIsTransparent, bool p_parentParentIsTransparent);

	/**
	 * Add a widget that just received a press event to the mousePressWidgets set
	 */
	static void addPressWidget(WEWidget * p_widget);

	/**
	 * Remove a widget that is no longer pressed from the mousePressWidgets set
	 */
	static void removePressWidget(WEWidget * p_widget);

	/**
	 * Check if a widget is currently pressed (is stored in mousePressWidgets set)
	 */
	static bool isPressed(WEWidget * p_widget);



	void * controller{nullptr};			//!< some kind of Controller of this widget - to be returned on request and with events created here
	WEWidget* parent{nullptr};			//!< Parent of this widget
	childrenBiMap * children{nullptr};	//!< List of children of this widget
	WEWidget * reparentedChild{nullptr};	//!< Child that was reparented e.g. for editing. This help cleaning up in this case.
	WE::WEInplaceEditingListener * inplaceEditingListener{nullptr};	//!< Listener interested in receiving the Inplace Editing Events of this Inplace editing widget.
	WEPainterPath * transparentRegion{nullptr};
	QTWEPainter * painter{nullptr};
	QTWEScrollArea * scrollArea{nullptr};

	int typeID{WIDGET_TYPE_ID_UNDEFINED};		//!< type of this widget
	bool active{true};	// seems if (true != this->active) { don't draw this widget; } (see setActive())
					// this might be a bad workaround addressing another basic problem ;-)
					// could be highly "list object" specific since only the list modifies visibility
					// of its children by default

	bool hasValidValue{true};
	bool inputRootFlag{false};		//<! Flag, to avoid searching after other input widgets, behind this widget
					//<! This flag is used for e.g. Virtual Keyboards.

	WEBrush brush{};	//!<Brush object to fill the background with the defined color
	WEPen pen{};		//!<Pen object to define the options for the drawing text



	std::string name{"UndefinedWidgetName"};	//!< Name of the Widget

	QTWEPainter *backgroundPainter{nullptr};
	QPixmap *backgroundPixmap{nullptr};
	WEImage *backgroundPixmapImage{nullptr};
	bool updateBackgroundPixmap{true};

	std::unordered_map< int, WEToolkitEventListener * > eventListenersMap{};

	WERect repaintRect{};

	int touchBeginID{-1}; //!< The ID of the touchBegin touch point

	std::map<int, WEWidget*> updateTouchPointsForwardedToWidgets{}; //<! If a touch update event was forwarded to a widget by another widget, we need to keep track of this to be able to cancel an input (the widget to which the event was forwarded will not receive update events anymore). The key is the ID of the touch point.

	qreal topZ{0};		//!<QWidget compatibility: top of the widget's stack
	qreal bottomZ{0};	//!<QWidget compatibility: bottom of the widget's stack

	bool transparentForMouseEvents{false};
	bool parentIsTransparentForMouseEvents{false};

	WEControllerNode *controllerNode{nullptr};

	static inline QTWEApplication * app{nullptr}; //!< Pointer to the application we are running on. Used to check if touch is enabled and for window managing messages.
	static inline std::set< WEWidget * > mousePressWidgets{};	//!< We want to provide the release event to the same widget that received the press
	static inline WEPen noPen{NO_PEN};	//empty pen that can be used by all widgets
	static inline WEBrush noBrush{NO_BRUSH};	//empty pen that can be used by all widgets
	static inline int scrollBarWidth{0};	//default scrollbar width is 0, meaning off
	static int debugSection;	//initialized in cpp file

	static const Color DEFAULT_DISABLED_COLOR{WE::grey};
	static const std::string SCREENSHOT_FOLDER;	//!< place where screenshot images will be stored, initialized in cpp file

private:

	WEBorderProfile * borderProfile{nullptr};//!<Object with the border profile properties
	WEImage * backgroundImage{nullptr}; //!< pointer to background image
	WEImage * borderImage{nullptr}; //!< pointer to border image

	int borderWidth{0};		//!<Width of the Border of the Widget
	int startAngle{0}; //!<needed for the border profile
	int endAngle{0}; //!<needed for the border profile
	int spanAngle{0};
	bool drawBorderFlag{false};		//!<If the Border should be drawn
	int cornerRadius{0};
	bool transparentBorderFlag{false};	//!<Shold the Border be transparent
	bool transparentFlag{false};	//!<Widget transparency setting
	bool setAsInputFlag{false};	//!<should the Widget be correspondent to the output signal (from Encoder for Example)
	bool weVisible{true};	//!< Intended visible status! Might not correspond to actual Qt visibility status. Had to rename it from "visible" since this is the name of the qt property.
	bool drawBorderImageFlag{false};
	bool drawBackgroundImageFlag{false};
	bool usesForeignBorderProfile{false};	//!< MKA's undocumented magic ... creating segfaults if one just wants to set a border profile and chooses the method that promises to create a copy :-(
				// it seems true == usesForeignBorderProfile should indicate that we are sharing somebody else's border profile
				// whereas false == usesForeignBorderProfile means we are maintaining our own copy
	WEBorderState borderState{WE::BORDER_STATE_UNDEFINED};	//!< Current State of the widget (DISABLED, Enabled with Focus, Enabled without Focus, Open for Input)

	WEColor transparentBorderColor{WE::black};
	WEColor foregroundColor{WE::black};	//!<Foreground Color of the Widget
	WEColor backgroundColor{WE::white};	//!<Background Color of the Widget

	WEEllipseNode * backgroundEllipseNode{nullptr};				//!<Node of the Widget which is used to draw the background
	WERectangleNode * backgroundRectangleNode{nullptr};			//!<Node of the Widget which is used to draw the background
	WEImageNode * backgroundImageNode{nullptr};				//!<Node of the Widget which is used to draw the background
	WEEllipseNode * disabledPatternEllipseNode{nullptr};			//!<Node of the Widget which is used to draw the Disabled Pattern
	WEEllipseNode * disabledPatternEllipseStrokeNode{nullptr};		//!<Node of the Widget which is used to draw the Disabled Pattern
	WERectangleNode * disabledPatternRectangleNode{nullptr};			//!<Node of the Widget which is used to draw the Disabled Pattern
	WERectangleNode * disabledPatternRectangleStrokeNode{nullptr};		//!<Node of the Widget which is used to draw the Disabled Pattern
	WEGeometryNode * disabledPatternGeometryNode{nullptr};			//!<Node of the Widget which is used to draw the Disabled Pattern
	WERectangleNode * borderImageRectangleNode{nullptr};			//!<Node of the Widget which is used to draw the border image
	std::vector<WERectangleNode*> borderProfileRectangleNodes{};	//!<Border Profile Nodes of the Widget
	std::vector<WEEllipseNode*> borderProfileEllipseNodes{};		//!<Border Profile Nodes of the Widget

	static bool multiTouchEnabled; //!< flag to indicate if multi touch (e.g. pressinf multiple at the same time) is enabled or not
};
} //End of namespace WE
#endif /* WEWIDGET_H_ */
