/**
 *	@file we_can.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief WEAPI interface functions for setting hardware related options
 *		to the CAN controller
 *
 *	@see we_can.cpp
 *
 *	This header file provides the WEAPI interfaces for setting
 *	options to the CAN hardware.
 */

#ifndef WE_CAN_H_
#define WE_CAN_H_

#include <string>
#include <memory> //unique_ptr

namespace WE
{

enum CAN_CONTROLLER_STATES
{
	WE_CAN_STATE_OFF		= 0x00,
	WE_CAN_STATE_ACTIVE		= 0x01,
	WE_CAN_STATE_ERROR_ACTIVE	= 0x02,
	WE_CAN_STATE_ERROR_PASSIVE	= 0x03,
	WE_CAN_STATE_BUS_OFF		= 0x04,
	WE_CAN_STATE_DEVICE_ERROR	= 0xff
};

enum WINDOWS_CAN_HARDWARE
{
	WINDOWS_CAN_NO_CAN		= -1,
	WINDOWS_CAN_INTERFACE_PEAK 	= 0,
	WINDOWS_CAN_INTERFACE_KVASER 	= 1
};

/**
 *	@class WE_CAN
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief WEAPI interface to the CAN hardware.
 *
 *	@namespace WE
 *
 *	@see peak_linux.h
 *	@see flexCAN.h
 *	@see we_can.cpp
 *
 *	This class provides an interface to the underlying CAN
 *	hardware. The interface can be used independend from
 *	the hardware.
 */
class WECAN
{
public:
	/**
	 *	@fn ~WE_CAN()
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Destructor of WE_CAN class
	 */
	virtual ~WECAN() = default;

	/**
	 *	@fn setBaudrate(unsigned int port, unsigned int baudrate)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief "Wrapper" interface for setting the baudrate of
	 *		a CAN port.
	 *
	 *	@param[in] port Number of the CAN port
	 *	@param[in] baudrate CAN baudrate to set. The baudrate can be passed either as
	 *		kbit/s (e.g. 250 => 250kbit/s => 250000bit/s) or
	 *		bit/s  (e.g. 250000 => 250kbit/s => 250000bit/s)
	 *
	 *	@returns WE_SUCCESS if setting the baudrate was successful.
	 *		WE_INIT_ERROR if initializing the CAN port with selected
	 *			baudrate was not successful.
	 *		WE_DEVICE_ERROR if opening the device was not successful
	 *			or the selected bitrate could not be retrieved.
	 *		WE_GENERAL_ERROR if something went completely wrong.
	 *			Should never happen.
	 *
	 *	@see flexCAN.h peak_linux.h
	 *
	 *	This function calls the setCANBaudrate of the WECANPort
	 *	hardware interface to set the baudrate of a CAN port.
	 *	This function can be called independent from the hardware.
	 *	I is decided on compile which hardware interface shall be used.
	 */
	virtual int setBaudrate(unsigned int p_port, unsigned int p_baudrate) = 0;

	/**
	 * Retreive baudrate of selected CAN port
	 * @param port Port number starting with 1
	 * @return baudrate in bit/s, -1 on error
	 */
	virtual int getBaudrate(unsigned int p_port) const = 0;

	/**
	 *	@fn static WE_CAN * create(int p_hardwareType = WINDOWS_CAN_INTERFACE_PEAK)
	 *
	 *	@brief Creates the hardware interface to the CAN hardware.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_hardwareType When using this class on Windows, the type of CAN dongle can be specified.
	 *		See enum WINDOWS_CAN_HARDWARE. For other platforms this parameter is not processed.
	 *
	 *	@returns A pointer to the CAN hardware interface.
	 *
	 *	This function needs to be called to create an instance of the CAN hardware
	 *	interface.
	 */
	static std::unique_ptr<WECAN> create(int p_hardwareType = WINDOWS_CAN_INTERFACE_PEAK);

	/**
	 * Retreive state of CAN port
	 * @param port CAN port starting with 1
	 * @return the state of the CAN port as in enum CAN_CONTROLLER_STATES
	 */
	static int getState(unsigned int p_port);

	/**
	 * Set the state of the CAN controlelr
	 * @param port Port number starting with 1
	 * @param p_state only WE_CAN_STATE_OFF or WE_CAN_STATE_ACTIVE are accepted
	 * @return WE_SUCCESS if setting state was successful, WE_GENERAL_ERROR on failure or WE_OUT_OF_BOUNDS_ERROR of port is != 1 or 2
	 */
	static int setState(unsigned int p_port, int p_state);

protected:
	std::string devicename;	//!< Name of the hardware interface (e.g. /dev/pcanusb0 on a PC)
	static int debugSection;
};
}//end of namespace WE
#endif /* WE_CAN_H_ */
