/**
 *	@file socket_can_interface.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Definition of an interface to access the socketCAN interface.
 *
 *	@see socket_can_interface.cpp
 *	@see can_interface.h
 *	@see we_can_raw.h
 *
 *	The class defined in this file allows access to the socketCAN interface
 */


#ifndef SOCKET_CAN_INTERFACE_H_
#define SOCKET_CAN_INTERFACE_H_

#include <net/if.h>
#include <linux/can.h>
#include <vector>

#include "../include/we_can_raw.h"

//Only create variables if there are no defines in Linux' standard header files (newer versions of Linux defines them directly)
#ifndef PF_CAN
// migrate "domain" setting for socket from #define in Makefile to constants:
static const int PF_CAN = 29;		//!< domain setting (?)
static const int AF_CAN = PF_CAN;	//!< address family (?)
#endif


namespace WE
{

/**
 *	@class CANSocket
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Interface for accessing socketCAN interface.
 *
 *	@namespace WE
 *
 *	@see CANInterface
 *
 *	This class is derived from WECANRaw and provides functions
 *	for accessing the socketCAN interface (initialize, read, write, etc.)
 *	A factory pattern is realized here. The functions of WECANRaw are overwritten
 *	and the create functions are implemented in the cpp file.
 */
class CANSocket : public WECANRaw
{
private:
	struct sockaddr_can address;//!< Address definition(AF_CAN), used for bind call
	struct ifreq ifr; 	//!< Interface request structure used for socket ioctl's
	socklen_t len;		//!< The length of a CAN socket

	fd_set FD;		//!< File descriptor for receiving CAN messages
	//fd_set writeFD;		//!< File descriptor for sending CAN messages

	bool initialized;	//!< indicates if the interface has been correctly initialized
	bool pollingRead;	//!< indicates if polling is enabled or not
	unsigned int pollIntervall;	//!< time in milliseconds in which shall be polled

	int endPointID;
	bool connected;

	std::vector<struct can_filter> filter;

	int numberOfRetries;
	int retryTimeout;

	int errorRv;		//!< Return code of last error (replacing a 'static' variable in 'CANSocket::send()')
	int oldErrno;		//!< Errno of last error (replacing a 'static' variable in 'CANSocket::send()')

public:
	/**
	 *	@fn CANSocket()
	 *
	 *	@brief Standard constructor of CANSocket class
	 *
	 *	The constructor initializes the classes attributes
	 *	with sane values.
	 */
	CANSocket();

	/**
	 *	@fn CANSocket(const char *  interfaceName)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Constructor for CANSocket interface that also
	 *		initializes the interface.
	 *
	 *	This constructor also does the work of the interfaceInit()
	 *	function. User can choose to create an instance with
	 *	standard constructor and then initialize or only use this
	 *	constructor without the need of initialize function.
	 *
	 *	@param[in] interfaceName Name of the interface that
	 *			shall be used
	 */
	explicit CANSocket(const char * p_interfaceName);

	/**
	 *	@fn ~CANSocket()
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Destructor of CANSocket class
	 *
	 *	The destructor closes the open sockets.
	 *
	 */
	virtual ~CANSocket();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	CANSocket(const CANSocket &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	CANSocket & operator=(const CANSocket &) = delete;

	/**
	 *	@fn initialize(unsigned int portNo, unsigned int baudrateToSet, const char * interfaceName)
	 *
	 *	@brief Function to initialize the Socket CAN stack
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param[in] portNo Number of the CAN port
	 *	@param[in] baudrateToSet The baudrate to be set in KBit/s
	 *	@param[in] interfaceName Name of the interface
	 *		that shall be initialized.
	 *
	 *	@returns 	WE_SUCCESS if message was sent out successfully
	 *	@returns	WE_INIT_ERROR if instance of interface is faulty
	 *	@returns	else: return code of interface function
	 *
	 *	With this function the SocketCAN stack can be initialized.
	 */
	int initialize(unsigned int p_portNo, unsigned int p_baudrateToSet, const char * p_interfaceName) override;

	/**
	 *	@fn int initialize(unsigned int p_portNo)
	 *
	 *	@brief Initialize a CAN port without setting the baudrate
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_portNo Number of the CAN port that shall be used
	 *
	 *	@returns WE_SUCCESS if the CAN port was successfully initialized
	 *	@returns The return value of int CANSocket::interfaceInit(const char *  interfaceName)
	 *
	 *	@see int CANSocket::interfaceInit(const char *  interfaceName)
	 */
	int initialize(unsigned int p_portNo) override;

	/**
	 *	@fn int interfaceInit(const char * interfaceName)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Initializes socketCAN interface
	 *
	 *	interfaceInit initializes the socket interface with
	 *	the provided interface name.
	 *
	 *	@param interfaceName const char* that represents the
	 *		interface name (like can0)
	 *
	 *	@returns 	WE_SUCCES if everything was OK
	 *			WE_
	 */
	int interfaceInit(const char * p_interfaceName);

	/**
	 *	@fn virtual int receive(WECANMessage * p_messageToStoreReceivedDataTo) = 0
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Receives a CAN message and stores it provided
	 *		WECANMessage for further processing. This call is the
	 *		prefered one (using endPoints) instead of the above blockign one.
	 *
	 *	@param p_messageToStoreReceivedDataTo A pointer to a struct
	 *		in which the received CAN message gets stored.
	 *
	 *	@returns WE_SUCCESS if a new CAN message was successfully received
	 *	@returns WE_TIMEOUT_ERROR if select timed out before a message was received
	 *	@returns WE_INTERFACE_ERROR if select returned an error (other than timeout)
	 *
	 *	@see int interfaceReceive(void)
	 *
	 *	@note This is the prefered function for reading messages. Used within onEndPointReadble()
	 *	      it will never block.
	 *
	 *	The function makes reads a socket. When beeing called from onEndPointReadable()
	 *	function, this call will never block. Otherwise, if no data are readable, this
	 *	function will block until data become available.
	 */
	int receive(WECANMessage * p_messageToStoreReceivedDataTo) override;

	/**
	 *	@fn int send(WECANMessage * messageToSend)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Sends a CAN message provided as parameter
	 *
	 *	The interfaceSend function sends out the CAN message
	 *	provided as parameter.
	 *
	 *	@param messageToSend CAN message that shall be sent
	 *
	 *	@returns WE_SUCCESS if message was sent out successfully
	 */
	int send(WECANMessage * p_messageToSend) override;

	/**
	 *	@fn inline void setPolling(bool polling, unsigned int timeout)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@brief Sets the reading of the socket to blocked or polling
	 *
	 *	The reading behavior gets set to polling or blocking.
	 *	Timeout can be set for polling.
	 *
	 *	@param[in] polling Enable or disables polling read
	 *	@param[in] timeout polling timeout in milliseconds
	 *
	 *	@returns void
	 *
	 *	@see interfaceReceive() interfaceReceive(WECANMessage * messageToStoreDataIn)
	 */
	void setPolling(bool p_polling, unsigned int p_timeout) override;

	//WEEndpoint functions:
	/**
	 *	@fn getReadEndPointID()
	 *
	 *	@brief Returns the Identifier used to read from the end point.
	 *
	 *	@returns End point Identifier.
	 *
	 *	This could be same as the identifier returned from the getWriteEndPointID()
	 */
	virtual int getReadEndPointID() const override;

	/**
	 *	@fn getWriteEndPointID()
	 *
	 *	@brief Returns the Identifier used to write to the end point.
	 *
	 *	@returns End point Identifier.
	 *
	 *	This could be same as the identifier returned from the getReadEndPointID()
	 */
	virtual int getWriteEndPointID() const override;

	/**
	 *	@fn isConnected()
	 *
	 *	@brief Check whether the end point is still connected to the peer entity
	 *			according to the last access result not the current status.
	 *
	 *	@returns true If last access was successful.
	 *	@returns false If last access produced error.
	 */
	virtual bool isConnected() const override;

	/**
	 *	@fn reconnect()
	 *
	 *	@brief reconnect the endpoint to the same peer with which it was connected earlier
	 *
	 *	@returns WE_SUCCESS If connect with server is re-established successfully.
	 *	@returns WE_INIT_ERROR If end point could not be reinitialised successfully.
	 *	@returns WE_TIMEOUT_ERROR If connection could not be established before connectTimeout expires.
	 *	@returns WE_CONNECT_ERROR If connection could not be established for any other reason.
	 *
	 *
	 *	It depends on the type of the end point if it is really useful because there may be cases
	 *	where endpoint seems always connected e.g; WEFileEndpoint.
	 */
	virtual int reconnect() override;

	int addIDFilter(unsigned int p_ID, unsigned int p_mask) override;

	void setRetry(int p_numberOfRetries, int p_retryTimeout);
};
}//end of namespace WE

#endif /* SOCKET_CAN_INTERFACE_H_ */
