/**
 *	@file we_dimmable.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Interface for accessing dimmable hardware (backlight, beeper, etc.)
 *
 *	@warning Dimmables are not supported for PC
 *	@warning Before including this header somewhere, we_datatypes.h has to be included.
 *
 *	This file defines a class for accessing dimmable hardware
 *	like display backlight, keyboard backlight, multi color LED and beeper.
 *	This hardware is not supported on a PC!
 */


#ifndef WE_DIMMABLE_H_
#define WE_DIMMABLE_H_

#include <fstream>
#include <mutex>

namespace WE
{

class WEColor;

class DimmableHardwareInterface;

/*
 * enum defining the dimmable hardware devices
 */
typedef enum
{
	DIMMABLE_DISPLAY,
	DIMMABLE_KEYBOARD,
	DIMMABLE_ENCODER,
	DIMMABLE_MULTI_COLOR_LED,
	DIMMABLE_BEEPER
}WEDimmableDevice;

/**
 *	@class WEDimmable
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Abstract class providing interface for accessing dimmable hardware
 *
 *	@namespace WE
 *
 *	@warning Dimmable hardware not supported on PC!
 *
 *	@see we_dimmable.h
 *	@see dimmable_interface.h
 *	@see dimmable_interface_pc.h
 *	@see dimmable_interface_a3.h
 *
 *	This class provides hardware independent access to the display backlight,
 *	keyboard backlight, multi color LED and beeper.
 *	This hardware is not supported on a PC.
 *	Before using the hardware the create function has to be called to provide
 *	the appropriate interface.
 */
class WEDimmable
{
public:

	/*--------------------------------------------------------*/
	/**
	 *	@fn ~WEDimmable()
	 *
	 *	@brief Standard destructor of WEDimmable class
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	The destructor closes all open file handles
	 */
	/*--------------------------------------------------------*/
	virtual ~WEDimmable();

	/*
	 * getIntensity returns the currently set intensity of a device identified by
	 * p_device. The intensity in % will be written into storeIntensity.
	 * @returns WE_SUCCESS if getting the intensity was successful. Else another WE error
	 * code is returned and a debug log entry is made.
	 */
	virtual int getIntensity(unsigned int * p_storeIntensity, WEDimmableDevice p_device);

	/**
	 * setIntensity is used to set the intensity of a device
	 * @param intensity intensity in % that shall be set
	 * @param p_device device identifier identifying the device
	 * @return WE_SUCCESS if setting was successful, else another WE error code
	 * is returned and a apropriate debug log entry is made.
	 */
	virtual int setIntensity(unsigned int p_intensity, WEDimmableDevice p_device);

	/**
	 * getcolor is used to retreive a set color (e.g. from multi color LED)
	 * @param storeColor The returned color as rgb value is stored in here
	 * @param p_device Identifier from which the color shall be retreived
	 * @return WE_SUCCESS if getting the color was successful, else another WE error code
	 * is returned and a apropriate debug log entry is made.
	 */
	virtual int getColor(unsigned int * p_storeColor, WEDimmableDevice p_device);

	/**
	 * setColor is used to change the current color of a device (e.g. multi color LED)
	 * @param color The color as rgb value that shall be set
	 * @param p_device Identifier for the device
	 * @return WE_SUCCESS if setting was successful, else another WE error code
	 * is returned and a apropriate debug log entry is made.
	 */
	virtual int setColor(unsigned int p_color, WEDimmableDevice p_device);

	/*--------------------------------------------------------*/
	/**
	 *	@fn setMinDisplayBacklightIntensity(unsigned int minBrightness) = 0
	 *
	 *	@brief With this function the minimal display backlight intensity can be set.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param[in] minBrightness The value of the min brightness that shall be set
	 *
	 *	@returns WE_SUCCESS if everything went OK
	 *	@returns WE_WRITE_ERROR if it was not possible to write the brightness
	 *
	 *	This function calls the hardware interface's function
	 *	to set the min display backlight intensity.
	 */
	/*--------------------------------------------------------*/
	virtual int setMinDisplayBacklightIntensity(unsigned int p_minBrightness);

	/*--------------------------------------------------------*/
	/**
	 *	@fn getMinDisplayBacklightIntensity(unsigned int * storeMinBrightness) = 0
	 *
	 *	@brief With this function the min backlight intensity can be
	 *		retrieved
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param[out] storeMinBrightness Pointer to the address to where the
	 *		retrieved backlight value shall be written to.
	 *
	 *	@returns WE_SUCCESS if everything went OK
	 *	@returns WE_READ_ERROR if it was not possible to read the
	 *		brightness value from backlight device
	 *
	 *	This function calls the hardware interface's function
	 *	to retrieve the min display backlight intensity. The read intensity
	 *	will be stored to the address to which the parameter points to.
	 */
	/*--------------------------------------------------------*/
	virtual int getMinDisplayBacklightIntensity(unsigned int * p_storeMinBrightness);

	virtual bool isColorSupported(WEDimmableDevice p_device) const;

	/**
	 *	@fn static WEDimmable * getInstance(void)
	 *
	 *	@brief Returns an instance of WEDimmable.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Pointer to the instance of this class.
	 *
	 *	Since this class is a singleton, this function has to be used to get
	 *	a pointer to the instance of this class. If no instance exists, yet one will be created.
	 */
	static WEDimmable * getInstance(void);


private:
	/*
	 * Private constructor that opens all device files for later access.
	 * Private because this class is a singleton
	 */
	WEDimmable();

	unsigned int displayBacklightIntensity;	//!< last retrieved displ. backlight intensity
	unsigned int keyboardBacklightIntensity; //!< last retrieved keyboard. backlight intensity
	unsigned int encoderBacklightIntensity; //!< last retrieved keyboard. backlight intensity

	static const unsigned int maxBeeperVolume = 100; //!< max beeper volume that can be set
	static const unsigned int maxDisplayBacklightBrightness = 100; //!< max display brightness that can be set
	static const unsigned int maxKeyboardBacklightBrightness = 100; //!< max keyboard brightness that can be set
	static const unsigned int maxKeyboardBacklightBrightnessIfColorIsSupported = 0xffffff; //!< max keyboard brightness that can be set
	static const unsigned int maxColorValue = 255; //!< max value of one color component that can be set

	static const unsigned int minBeeperVolume = 0; //!< min beeper volume that can be set
	unsigned int minDisplayBacklightBrightness; //!< min display brightness that can be set
	static const unsigned int minKeyboardBacklightBrightness = 0; //!< min keyboard brightness that can be set
	static const unsigned int minColorValue = 0; //!< min value of one color component that can be set

	static int debugSection;

	static std::mutex LEDMutex;
	static std::mutex beeperMutex;
	static std::mutex displayMutex;

	std::string deviceNameBacklight;	//!< Filename of display backlight device
	std::string deviceNameKeyBacklight;	//!< Filename of keyboard backlight device
	std::string deviceNameEncoderBacklight;	//!< Filename of keyboard backlight device
	std::string deviceNameBeeper;	//!< Filename of beeper device
	std::string deviceNameLED;	//!< Filename of LED device

	//the sys/device files:
	std::fstream displayBacklightFile;
	std::fstream keyboardBacklightFile;
	std::fstream encoderBacklightFile;
	std::fstream LEDFile;
	std::fstream beeperFile;

	bool keyboardColorSupported;
	bool encoderColorSupported;

	unsigned int keyboardColor;
	unsigned int encoderColor;
};
}//end of namespace WE
#endif /* WE_DIMMABLE_H_ */
