/**
*	@file hardware_interface.h
*
*	@author dsl@wachendorff.de
*	@author Dirk Sterzel
*
*	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
*
*	@brief Abstract interface to retrive and values for internal hardware variables
*
*	@see WEHardwareDaemon
*	@see WEDEI
*	@see WEInternalVariable
*/

#ifndef HARDWARE_INTERFACE_H_
#define HARDWARE_INTERFACE_H_

#include <string>
#include <map>
#include <mutex>
#include <condition_variable>
#include <QObject>

#include "../../input_output/include/we_io.h"
#include "../../can/include/wecan.h"
#include "../../../event_handling/include/we_signal.h"
#include "../../../os_abstraction/thread/include/we_thread.h"
#include "../../../os_abstraction/network/include/we_network_config.h"

namespace WE
{

class WEDimmable;
class WEInputOutput;
class WEEventQueue;
class WEThread;
class WENetworkConfigFactory;
class WEPowerManagement;
class WEEndPointManager;
class WEDEI;
class WECAN;
class HardwareInterface;

/* The following enum defines the supported beeper types. The order of the types
 * reflects their priority. Smaller numbers have higher priority.
 */
enum beeperRequestType{
	BEEPER_REQ_ALARM_PATTERN,	// is handled with same priority as BEEPER_REQ_ALARM_ON
	BEEPER_REQ_ALARM_ON,		// is handled with same priority as BEEPER_REQ_ALARM_PATTERN
	BEEPER_REQ_REPETITION,
	BEEPER_REQ_EVENT_PATTERN,	// is handled with same priority as BEEPER_REQ_EVENT_ON
	BEEPER_REQ_EVENT_ON,		// is handled with same priority as BEEPER_REQ_EVENT_PATTERN
	BEEPER_REQ_CLICK_ON,
	BEEPER_REQ_NONE
};

struct beeperRequestProperties{
	HardwareInterface * interface;
	int value;
	int maxVolume;
	int clickVolume;
	int alarmVolume;
	int eventVolume;
	int variableVolume;
	int minEmergencyAlarmVolume;
	int onTime;
	int offTime;
	int repetitions;
	int alarmRepetitions;
	int eventRepetitions;
	int alarmPattern;
	int eventPattern;
	int clickOn;
	int alarmOn;
	int eventOn;
};

class HardwareInterface : public QObject, public WESignalListener
{
	Q_OBJECT
public:
	explicit HardwareInterface(const std::string & p_ownerName, QObject * p_parent = nullptr);
	virtual ~HardwareInterface();
	//don't allow copy or assignment
	HardwareInterface(const HardwareInterface & toCopy) = delete;
	HardwareInterface& operator=(const HardwareInterface& toCopy) = delete;

	void initPowerManagementInterface(WEEndPointManager * mgr);

	/* Index 0x2000 - Terminal Hardware */
	int getTerminalHardware_NoOfEntries(int * valueToReturn) const;
	int getTerminal(std::string * valueToReturn) const;
	int getWE_PartNumber(std::string * valueToReturn) const;
	int setWE_PartNumber(const std::string & valueToSet);
	int getBrandOwnerPartNumber(std::string * valueToReturn) const;
	int setBrandOwnerPartNumber(const std::string & valueToSet);
	int getSerialNumber(std::string * valueToReturn) const;
	int setSerialNumber(const std::string & valueToSet);
	int getOperatingTime(int * valueToReturn) const;

	/* Index 0x2010 - SD + USB */
	int getSDUSB_NoOfEntries(int * valueToReturn) const;
	int getUSBMemStatus(int * valueToReturn); //non-const, manipulates this->usbUmountPID
	int setUSBMemStatus(int valueToSet);
	int getUSBMemSize(int * valueToReturn) const;
	int getUSBMemUsed(int * valueToReturn) const;
	int getUSBMountPath(std::string * valueToReturn) const;
	// no setters since the hardware of SD USB is read only

	//	/* Index 0x2020 - Terminal Software */
	int getTerminalSoftware_NoOfEntries(int * valueToReturn) const;
	int getSystemUptime(int * valueToReturn) const;
	int getOSKernelVersion(std::string * valueToReturn) const;
	int getEnableServiceMode(int * valueToReturn) const;
	int setEnableServiceMode(int valueToSet);

	//	/* Index 0x2030 - Power Modes */
	int getPowerModes_NoOfEntries(int * valueToReturn) const;
	int getPWR_TimeToSleepMode(int * valueToReturn) const;
	int getPWR_TimeToPowerOff(int * valueToReturn) const;
	int getPWR_TimeFromSilentOnToLowPower(int * valueToReturn) const;
	int getPWR_LowPowerDisplayState(int * valueToReturn) const;
	int getPWR_LowPowerDisplayBacklightState(int * valueToReturn) const;
	int getPWR_LowPowerEncoderBacklightState(int * valueToReturn) const;
	int getPWR_LowPower_RGB_LED_State(int * valueToReturn) const;
	int getPWR_LowPowerBeeperState(int * valueToReturn) const;
	int getPWR_LowPowerKeypadState(int * valueToReturn) const;
	int getPWR_LowPowerEncoderState(int * valueToReturn) const;
	int getPWR_LowPowerTouchState(int * valueToReturn) const;
	int getPWR_LowPowerConsoleState(int * valueToReturn) const;
	int getPWR_LowPowerOutput0State(int * valueToReturn) const;
	int getPWR_LowPowerOutput1State(int * valueToReturn) const;
	int getPWR_LowPowerOutput2State(int * valueToReturn) const;
	int getPWR_LowPowerCamOutputState(int * valueToReturn) const;
	int getPWR_LowPowerVideoState(int * valueToReturn) const;
	int getPWR_CurrentPowerState(int * valueToReturn) const;
	int getPWR_SleepModePropagationTime(int * valueToReturn) const;
	int getPWR_LowPowerDigitalInput1State(int * valueToReturn) const;
	int getPWR_LowPowerDigitalInput2State(int * valueToReturn) const;
	int getPWR_LowPowerDigitalInput3State(int * valueToReturn) const;
	int getPWR_LowPowerDigitalInput4State(int * valueToReturn) const;
	int getPWR_LowPower1stStatusLEDState(int * valueToReturn) const;
	int getPWR_LowPower2ndStatusLEDState(int * valueToReturn) const;

	int setPWR_TimeToSleepMode(int valueToSet);
	int setPWR_TimeToPowerOff(int valueToSet);
	int setPWR_TimeFromSilentOnToLowPower(int valueToSet);
	int setPWR_LowPowerDisplayState(int valueToSet);
	int setPWR_LowPowerDisplayBacklightState(int valueToSet);
	int setPWR_LowPowerKeypadBacklightState(int valueToSet);
	int setPWR_LowPowerEncoderBacklightState(int valueToSet);
	int setPWR_LowPower_RGB_LED_State(int valueToSet);
	int setPWR_LowPowerBeeperState(int valueToSet);
	int setPWR_LowPowerKeypadState(int valueToSet);
	int setPWR_LowPowerEncoderState(int valueToSet);
	int setPWR_LowPowerTouchState(int valueToSet);
	int setPWR_LowPowerConsoleState(int valueToSet);
	int setPWR_LowPowerOutput0State(int valueToSet);
	int setPWR_LowPowerOutput1State(int valueToSet);
	int setPWR_LowPowerOutput2State(int valueToSet);
	int setPWR_LowPowerCamOutputState(int valueToSet);
	int setPWR_LowPowerVideoState(int valueToSet);
	int setPWR_CurrentPowerState(int valueToSet);
	int setPWR_SleepModePropagationTime(int valueToSet);
	int setPWR_LowPowerDigitalInput1State(int valueToSet);
	int setPWR_LowPowerDigitalInput2State(int valueToSet);
	int setPWR_LowPowerDigitalInput3State(int valueToSet);
	int setPWR_LowPowerDigitalInput4State(int valueToSet);
	int setPWR_LowPower1stStatusLEDState(int valueToSet);
	int setPWR_LowPower2ndStatusLEDState(int valueToSet);

	/* Index 0x2040 - LEDs */
	int getLEDNoOfEntries(int * valueToReturn) const;
	int getLED_MultiColor(int * valueToReturn) const;
	int setLED_MultiColor(int valueToSet);
	int getLEDMultiColorRepetitions(int * valueToReturn) const;
	int setLEDMultiColorRepetitions(int valueToSet);
	int getLEDMultiColorPattern(int * valueToReturn) const;
	int setLEDMultiColorPattern(int valueToSet);

	/* Index 0x2050 - Sensors */
	int getSensors_NoOfEntries(int * valueToReturn) const;
	int getDigitalInputs_NoOfEntries(int * valueToReturn) const;
	int getFrequencyInputs_NoOfEntries(int * valueToReturn) const;

	int setInputConfiguration(int port, int valueToSet);

	/* Index 0x20a0 - Digital Outputs */
	int getDigitalOutputs_NoOfEntries(int * valueToReturn) const;
	int getDigitalOutput01(int * valueToReturn) const;
	int setDigitalOutput01(int valueToSet);
	int getDigitalOutput02(int * valueToReturn) const;
	int setDigitalOutput02(int valueToSet);
	int getDigitalOutput03(int * valueToReturn) const;
	int setDigitalOutput03(int valueToSet);
	int getDigitalOutput04(int * valueToReturn) const;
	int setDigitalOutput04(int valueToSet);
	int getDigitalOutputLoopback(int * valueToReturn, int portNumber) const;
	int getDigitalOutputFailure(int * valueToReturn, int portNumber) const;

	/* Index 0x20b0 - Display Backlight */
	int getDisplayBacklight_NoOfEntries(int * valueToReturn) const;
	int getDispBacklightIntensity(int * valueToReturn) const;
	int getMinDispBacklightIntensity(int * valueToReturn) const;
	int getDispBacklightAmbientLowPass(int * valueToReturn) const;
	int getDispBacklightTimeConstant(int * valueToReturn) const;
	int getDispBacklightFilterFactor(int * valueToReturn) const;
	int getDispBacklightIntensity0(int * valueToReturn) const;
	int getDispBacklightIntensity10(int * valueToReturn) const;
	int getDispBacklightIntensity20(int * valueToReturn) const;
	int getDispBacklightIntensity30(int * valueToReturn) const;
	int getDispBacklightIntensity40(int * valueToReturn) const;
	int getDispBacklightIntensity50(int * valueToReturn) const;
	int getDispBacklightIntensity60(int * valueToReturn) const;
	int getDispBacklightIntensity70(int * valueToReturn) const;
	int getDispBacklightIntensity80(int * valueToReturn) const;
	int getDispBacklightIntensity90(int * valueToReturn) const;
	int getDispBacklightIntensity100(int * valueToReturn) const;
	int getDispBacklightOffset(int * valueToReturn) const;
	int getDispBacklightStartupIntensity(int * valueToReturn) const;

	int setDispBacklightIntensity(int valueToSet);
	int setMinDispBacklightIntensity(int valueToSet);
	int setDispBacklightAmbientLowPass(int valueToSet);
	int setDispBacklightTimeConstant(int valueToSet);
	int setDispBacklightFilterFactor(int valueToSet);
	int getAutoDispBacklightIntensity(int * valueToReturn, int p_index) const;
	int setAutoDispBacklightIntensity(int valueToSet, int p_index);
	int setDispBacklightOffset(int valueToSet);
	int setDispBacklightStartupIntensity(int valueToSet);

	/* Index 0x20c0 - Keyboard Backlight */
	int getKeyblayBacklight_NoOfEntries(int * valueToReturn) const;
	int getKeybBacklightIntensity(int * valueToReturn); //non-const because it also sets intensity
	int getKeybBacklightColor(int * valueToReturn); //non-const because it also sets intensity
	int getKeybBacklightStartupIntensity(int * valueToReturn) const;
	int getKeybBacklightStartupColor(int * valueToReturn) const;
	int getEncoderbBacklightIntensity(int * valueToReturn); //non-const because it also sets intensity
	int getEncoderBacklightColor(int * valueToReturn); //non-const because it also sets intensity
	int setKeybBacklightIntensity(int valueToSet);
	int setEncoderBacklightIntensity(int valueToSet);
	int setKeybBacklightColor(int valueToSet);
	int setEncoderBacklightColor(int valueToSet);
	int setKeyBacklightStartupIntensity(int valueToSet);
	int setKeyBacklightStartupColor(int valueToSet);

	/* Index 0x20d0 - Beeper */
	int getBeeper_NoOfEntries( int * valueToReturn) const;
	int getBeeperMaxVolume(int * valueToReturn) const;
	int getBeeperClickVolume(int * valueToReturn) const;
	int getBeeperAlarmVolume(int * valueToReturn) const;
	int getBeeperEventVolume(int * valueToReturn) const;
	int getBeeperMinEmergencyAlarmVolume(int * valueToReturn) const;
	int getBeeperFrequency(int * valueToReturn) const;
	int getBeeperOnTime(int * valueToReturn) const;
	int getBeeperOffTime(int * valueToReturn) const;
	int getBeeperRepetitions(int * valueToReturn) const;
	int getBeeperAlarmRepetitions(int * valueToReturn) const;
	int getBeeperEventRepetitions(int * valueToReturn) const;
	int getBeeperClickOn(int * valueToReturn) const;
	int getBeeperAlarmOn(int * valueToReturn) const;
	int getBeeperEventOn(int * valueToReturn) const;
	int getBeeperVariableVolume(int * valueToReturn) const;
	int getBeeperAlarmPattern(int * valueToReturn) const;
	int getBeeperEventPattern(int * valueToReturn) const;
	int getBeeperIsActive(int * valueToReturn) const;
	int getVolumeLevelSpeaker(int * valueToReturn) const;
	int getVolumeLevelLineOut(int * valueToReturn) const;
	int setBeeperMaxVolume(int valueToSet);
	int setBeeperClickVolume(int valueToSet);
	int setBeeperAlarmVolume(int valueToSet);
	int setBeeperEventVolume(int valueToSet);
	int setBeeperMinEmergencyAlarmVolume(int valueToSet);
	int setBeeperFrequency(int valueToSet);
	int setBeeperOnTime(int valueToSet);
	int setBeeperOffTime(int valueToSet);
	int setBeeperRepetitions(int valueToSet, bool startBeeperThread = true);
	int setBeeperAlarmRepetitions(int valueToSet, bool startBeeperThread = true);
	int setBeeperEventRepetitions(int valueToSet, bool startBeeperThread = true);
	int setBeeperClickOn(int valueToSet, bool startBeeperThread = true);
	int setBeeperAlarmOn(int valueToSet, bool startBeeperThread = true);
	int setBeeperEventOn(int valueToSet, bool startBeeperThread = true);
	int setBeeperVariableVolume(int valueToSet);
	int setBeeperAlarmPattern(int valueToSet);
	int setBeeperEventPattern(int valueToSet);
	int setBeeperIsActive(int valueToSet);
	int setVolumeLevelSpeaker(int valueToSet);
	int setVolumeLevelLineOut(int valueToSet);

	int getRTCNoOfEntries(int * valueToReturn) const;
	int getRTCYear(int * valueToReturn) const;
	int getRTCMonth(int * valueToReturn) const;
	int getRTCDay(int * valueToReturn) const;
	int getRTCHours(int * valueToReturn) const;
	int getRTCMinutes(int * valueToReturn) const;
	int getRTCSeconds(int * valueToReturn) const;
	int getRTCDayTime(int * valueToReturn) const;
	int getRTCHours12h(int * valueToReturn) const;


	int setRTCYear(int valueToSet);
	int setRTCMonth(int valueToSet);
	int setRTCDay(int valueToSet);
	int setRTCHours(int valueToSet);
	int setRTCMinutes(int valueToSet);
	int setRTCSeconds(int valueToSet);
	int setRTCDayTime(int valueToSet);
	int setRTCHours12h(int valueToSet);

	//Index 0x2120, 0x2121 CAN state and statistics
	int getCANState(unsigned int port, int * valueToReturn) const;
	int getCANRxPackets(unsigned int port, int * valueToReturn); //non-const, stores packets in member
	int getCANTxPackets(unsigned int port, int * valueToReturn); //non-const, stores packets in member
	int getCANBitrate(unsigned int port, int * valueToReturn) const;

	int setCANState(unsigned int port, int valueToSet);
	int setCANBitrate(unsigned int port, int valueToSet);

	/* Index 0x2130 - FileSystem */
	int getFileSystem_NoOfEntries	( int * valueToReturn) const;
	int getCurrentDirectory		( std::string * valueToReturn) const;
	int getCurrentFile		( std::string * valueToReturn) const;
	int getDestinationDirectory	( std::string * valueToReturn) const;
	int getCopyFileOrFolder		( int * valueToReturn) const;
	int getFileSystemPath		( std::string * valueToReturn) const;
	int getUsedMemoryPartition	( int * valueToReturn) const;
	int getFreeMemoryPartition	( int * valueToReturn) const;
	int getCurrentFileSize		( int * valueToReturn) const;

	int setCurrentDirectory		( const std::string & p_valueToSet);
	int setCurrentFile		( const std::string & p_valueToSet);
	int setDestinationDirectory	( const std::string & p_valueToSet);
	int setCopyFileOrFolder		( int p_valueToSet);
	int setFileSystemPath		( const std::string & p_valueToSet);

	/*index 0x2140 Ethernet*/
	int getEthernet_NoOfEntries	(int * valueToReturn) const;

	int setIPaddress(unsigned short p_port, const std::string & p_valueToSet, bool p_apply);
	int setNetmaskIP(unsigned short p_port, const std::string & p_valueToSet, bool p_apply);
	int setGatewayIP(unsigned short p_port, const std::string & p_valueToSet, bool p_apply);

	int applyEthernetSettings(unsigned short p_port);

	int getEthernet0MasterConfig(int * valueToReturn);
	int setEthernet0MasterConfig(int valueToSet);

	/* Index 0x2150 - RS232 */
	int getRS232_NoOfEntries(int * valueToReturn) const;
	int getConsoleMode(int * valueToReturn) const;
	int setConsoleMode(int valueToSet);

	int configureIOEvents(WEEventQueue * p_eventQueue, WEEndPointManager * p_endpointManager);

	int enableAutoDisplayBacklight(void);
	int disableAutoDisplayBacklight(void);
	static bool isAutoBacklightrunning(void)
	{
		return autoBacklightRunning;
	}
	static int getScaledBacklightValueBasedOnAmbientLight(int p_currentLightPerMille, WEDEI * p_dei);
	static void autoBacklightThread(void * obj);

	void setInitialized(void)
	{
		this->initialized = true;
	}
	bool isInitialized(void) const
	{
		return this->initialized;
	}



//	bool beeperAlarmPatternThreadRunning;

	int signalOccurred(WESignal & p_signal) override;

	int processIOEvents(WEEventQueue * eventQueueIO);

	int getIOIndicesByEventInfo(unsigned int p_info, unsigned short * p_index, unsigned short * p_subindex) const;





	WEDimmable * getDimmableInterface(void) const
	{
		return this->dimmableInterface;
	}

	void setDEI(WEDEI * p_dei)
	{
		this->dei = p_dei;
	}

	void createSpeakerSettings(void);
	void setSpeakerVolume(int p_volume, bool p_forceRepeat = false);

	/*
	 * Function only used by OPUS devices to set the beeper volume.
	 * On egspro devices it sets the speaker volume, on G1 devices it sets the volume to the beeper device file.
	 */
	void setBeeperVolume(int p_volume);

	static void beeperThreadFunction(void *);

protected:
	void startBeeperThread(void);

private:
	static std::string getRemanentValueOrSetDefault(const std::string & p_fileName, const std::string & p_defaultValue, bool (functionPointer(const std::string &)));

	int usbUmountPID;	//!< PID while 'umount' command is running, 0 otherwise

	WEDimmable* dimmableInterface; //!< pointer to singelton. Do not delete!
	WEInputOutput ioInterface{};
	WENetworkConfigFactory ethconfig{};
	std::unique_ptr<WEPowerManagement> powerDaemon;

	//serials
	std::string serialNumber;
	std::string partNumber;
	std::string brandOwnerPartNumber;

	//LED pattern and repetitions:
	int ledRepetitions;
	int ledPattern;

	//hardware daemon manages beeper settings:
	int beeperMaxVolume;
	int beeperClickVolume;
	int beeperAlarmVolume;
	int beeperEventVolume;
	int beeperMinEmergencyAlarmVolume;
	int beeperOnTime;
	int beeperOffTime;
	int beeperRepetitions;
	int beeperAlarmRepetitions;
	int beeperEventRepetitions;
	int beeperClickOn;
	int beeperAlarmOn;
	int beeperEventOn;
	int beeperVariableVolume;
	int beeperAlarmPattern;
	int beeperEventPattern;
	int beeperIsActive;
	int volumeLevelSpeaker;
	int volumeLevelLineOut;

	std::string packagePath; //!< Used for variable "@PackagePath" 0x2020 0x02

	int ambientLowPassValue;
	int backlightTimeConstant;
	int backlightFilterFactor;
	int autoBacklightCurve[11];
	int backlightOffset;

	int ambientLightValue;

	int timeToSleep;
	int timeToOff;
	int timeFromSilentOnToLowPower;
	int lowPowerDisplayState;
	int lowPowerDisplayBacklightState;
	int lowPowerKeypadBacklightState;
	int lowPowerEncoderBacklightState;
	int lowPower_RGB_LED_State;
	int lowPowerBeeperState;
	int lowPowerKeypadState;
	int lowPowerEncoderState;
	int lowPowerTouchState;
	int lowPowerConsoleState;
	int lowPowerOutput0State;
	int lowPowerOutput1State;
	int lowPowerOutput2State;
	int lowPowerCamOutputState;
	int lowPowerVideoState;
	int lowPowerInput1State;
	int lowPowerInput2State;
	int lowPowerInput3State;
	int lowPowerInput4State;
	int lowPower1stStausLEDState;
	int lowPower2ndStausLEDState;
	int currentPowerState;
	int sleepModePropagationTime;

	int lastCANRxPackages[4];
	int lastCANTxPackages[4];

	WEThread backlightThread{};

	bool initialized;

	WEDEI * dei;

	std::unique_ptr<WE::WECAN> canInterface{};

	std::string ownerName;

	static struct beeperRequestProperties beeperThreadData;


	static bool autoBacklightRunning;
	static bool beeperAvailable;
	static bool speakerAvailable;
	// Data needed for beeper thread handling
	WEThread beeperThread{"Beeper Thread"};
	static std::mutex beeperMutex;	// limit ALL access to variables below (even priority comparison of new request
	// (i.e. read access) is only done after locking)
	static std::condition_variable beeperCondition;
	static WEDimmable * beeperInterface;	// only used in thread and hence not protected by mutex?!
	// Access to the following variables is only allowed when having locked the beeperMutex
	static bool beeperDataReady;	// true: 'beeperThreadData' below only consists of values the thread can work with
	static bool terminateBeeperThread;
	static bool beeperThreadRunning;
	static beeperRequestType  nextBeeperRequest;			// directly create a static variable of the above enum type
	static const int minAutoBacklightTimeConstant = 100;
	static const std::string digitalOutPathPrefix;


	static int startNextBeeperRequest(void * myInterface, enum beeperRequestType req, int valueToSet);
	static void playBeeperPattern(struct beeperRequestProperties &properties);
	static int processBeeperRepetition(struct beeperRequestProperties &properties);
	static int processBeeperAlarmPattern(struct beeperRequestProperties &properties);
	static int processBeeperAlarmOn(struct beeperRequestProperties &properties);
	static int processBeeperEventPattern(struct beeperRequestProperties &properties);
	static int processBeeperEventOn(struct beeperRequestProperties &properties);
	static int processBeeperClickOn(struct beeperRequestProperties &properties);

	signals:
	//the following signal is only used when using QSoundEffect as beeper emulation
	void setSpeakerVolumeSignal(int p_volume);
};

} // namespace WE

#endif /* HARDWARE_INTERFACE_H_ */
