/**
*	@file we_hardware_daemon.h
*
*	@author dsl@wachendorff.de
*	@author Dirk Sterzel
*
*	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
*
*	@brief Definition of WEHardwareDaemon class
*
*	@see HardwareInterface
*	@see WEDEI
*	@see WEVariable
*
*	This file defines the WEHardwareDaemon class. It's purpose is to
*	manage all internal hardware variables so that other processes
*	can connect to this daemon and there is only one process that accesses
*	the hardware directly.
*/

#ifndef WE_HARDWARE_DAEMON_H_
#define WE_HARDWARE_DAEMON_H_

#include <QObject>

#include "../../../dei/variables/include/we_variable.h"
#include "../../../dei/include/we_variable_owner.h"

#include "../../../event_handling/include/we_event_queue_base.h"

#include "../../../os_abstraction/watchdog_daemon/include/watchdog_daemon_interface.h"

#include <vector>

namespace WE
{

class HardwareInterface;
class WEEndPointManager;
class WEDEI;

typedef enum
{
	VARIABLE_PRIORITY_AUTO 		= 0,
	VARIABLE_PRIORITY_LOW 		= 1,
	VARIABLE_PRIORITY_MEDIUM 	= 2,
	VARIABLE_PRIORITY_HIGH 		= 3,
	VARIABLE_PRIORITY_SECOND 	= 4,
	VARIABLE_PRIORITY_MINUTE	= 5,
	VARIABLE_PRIORITY_NONE 		= 6
}WEHardwareVariablePriority;

class WEHardwareDaemon : public QObject, public WEVariableOwner
{
	Q_OBJECT
public:
	/**
	*	@fn WEHardwareDaemon(const std::string& p_IPCServerAddress, WE::WEEndPointManager * p_endPointManager)
	*
	*	@brief Constructor for WEHardwareDaemon
	*
	*	@author dsl@wachendorff.de
	*	@author Dirk Sterzel
	*
	*	@param[in] p_IPCServerAddress The IPC address of with which the hardware daemon can be accessed over IPC
	*	@param[in] p_endPointManager The end point manager that has to be used by the DEI IPC handling.
	*
	*	Constructor of WEHardwareDaemon that takes a string as parameter to define the IPC address
	*	of this process.
	*/
	WEHardwareDaemon(const std::string& p_IPCServerAddress, WE::WEEndPointManager * p_endPointManager);
	//don't allow copy or assignment
	WEHardwareDaemon(const WEHardwareDaemon & p_copy) = delete;
	WEHardwareDaemon & operator=(const WEHardwareDaemon & p_copy) = delete;

	/**
	*	@fn ~WEHardwareDaemon()
	*
	*	@brief Destructor of WEHardwareDaemon. Frees all memory and quits all connections.
	*
	*	@author dsl@wachendorff.de
	*	@author Dirk Sterzel
	*
	*/
	virtual ~WEHardwareDaemon();

	int initialize(void);

	/**
	*	@fn unsigned int getTimeperiodHighPriority()
	*
	*	@brief Returns the time period in which high priority varibales shall be updated
	*
	*	@author dsl@wachendorff.de
	*	@author Dirk Sterzel
	*
	*	@returns RETURN DESCRIPTION
	*
	*	@see REFERENCE TO SOMETHING INTERESTING
	*
	*	@warning DANGEROUS STUFF
	*
	*	DETAILED DESCRIPTION
	*/
	static unsigned int getTimeperiodHighPriority()
	{
		return TIME_PERIOD_HIGH_PRIORITY;
	}
	static unsigned int getTimeperiodMediumPriority()
	{
		return TIME_PERIOD_MEDIUM_PRIORITY;
	}
	static unsigned int getTimeperiodLowPriority()
	{
		return TIME_PERIOD_LOW_PRIORITY;
	}

	//int update(unsigned short p_index, unsigned short p_subindex, std::string p_owner, int p_intValue);
	//int update(unsigned short p_index, unsigned short p_subindex, std::string p_owner, std::string p_stringValue);

	int updateRequest(unsigned short p_index, unsigned short p_subindex, int p_intValue) override;
	int updateRequest(unsigned short p_index, unsigned short p_subindex, const std::string & p_stringValue) override;

	// JU
	/**
	*	@fn getValueRequest(unsigned short p_index, unsigned short p_subindex)
	*
	*	@brief NOT SUPPORTED for hardware daemon
	*
	*	@author ju@wachendorff.de
	*	@author Joachim Urbach
	*
	*	@returns WE_NOT_SUPPORTED_ERROR
	*
	*	@warning Function not supported for hardware daemon
	*/
	int getValueRequest(unsigned short p_index, unsigned short p_subindex) override;

	/**
	 *	@fn static void setExitState(void *)
	 *
	 *	@brief Callback function to be called if PClient is killed from outside
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns void
	 *
	 *	On a Linux system, this function gets called if one of these signals
	 *	are received:
	 *	SIGHUP, SIGINT, SIGQUIT, SIGTERM, SIGKILL.
	 *	It just sets the pclientState to EXIT so the application is exited in a controlled
	 *	way and the onProjectExit Action can be executed.
	 */
	static void setExitState(void *);

	/**
	 *	@fn int setTerminationListenerFunction(void (*functionPointer)(void*))
	 *
	 *	@brief Set a callback function that gets kalled when the process gets killed from outside
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *	original implemented by:
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param functionPointer pointer to the function that shall be called
	 *
	 *	@returns WE_SUCCESS if the callback function was correctly set
	 *	@returns WE_GENERAL_ERROR else
	 *
	 *	The implementation of this function is OS dependent.
	 *	On Linux system signal handlers are installed which call the
	 *	function pointed to by functionPointer.
	 */
	int setTerminationListenerFunction(void (*functionPointer)(void*));

	static int debugSectionKey;

public slots:
	//the run slot is called during QCoreApplications idle processing and represents the main loop of hardware daemon
	void run(void);

private:

	int createVariableAndAddToDEI(unsigned short p_index,
			unsigned short p_subindex,
			std::string p_name,
			WEVariableStorageType p_storageType,
			WEHardwareVariablePriority p_priority,
			WEVariableAccessType p_accessType,
			bool remanent,
			bool p_forceSetting = false);
	int createAndInitializeVariables(void);
	//following function is generated by json to cpp converter
	int createPredefinedVariables(void);

	int retrieveVariableValue(unsigned short p_index, unsigned short p_subindex, int * p_intToReturn);
	int retrieveVariableValue(unsigned short p_index, unsigned short p_subindex, std::string * p_stringToReturn);

	int setHardwareValue(unsigned short p_index, unsigned short p_subindex, int p_value);
	int setHardwareValue(unsigned short p_index, unsigned short p_subindex, const std::string & p_value);

	int retrieveVariableValueTerminalHardware(unsigned short p_subindex, int * p_intToReturn);
	int retrieveVariableValueTerminalHardware(unsigned short p_subindex, std::string * p_stringToReturn);
	int setHardwareValueTerminalHardware(unsigned short p_subindex, const std::string & p_value);

	int retrieveVariableValueUSB(unsigned short p_subindex, int * p_intToReturn);
	int retrieveVariableValueUSB(unsigned short p_subindex, std::string * p_stringToReturn);
	int setHardwareValueUSB(unsigned short p_subindex, int p_value);

	int retrieveVariableValueTerminalSoftware(unsigned short p_subindex, int * p_intToReturn);
	int retrieveVariableValueTerminalSoftware(unsigned short p_subindex, std::string * p_stringToReturn);
	int setHardwareValueTerminalSoftware(unsigned short p_subindex, int p_value);

	int retrieveVariableValuePowerModes(unsigned short p_subindex, int * p_intToReturn);
	int setHardwareValuePowerModes(unsigned short p_subindex, int p_value);

	int retrieveVariableValueLED(unsigned short p_subindex, int * p_intToReturn);
	int setHardwareValueLED(unsigned short p_subindex, int p_value);

	int retrieveVariableValueSensors(unsigned short p_index, unsigned short p_subindex, int * p_intToReturn);

	int retrieveVariableValueDigitalInputs(unsigned short p_index, unsigned short p_subindex, int * p_intToReturn);

	int retrieveVariableValueAnalogInputs(unsigned short p_index, unsigned short p_subindex, int * p_intToReturn);

	int retrieveVariableValueFrequencyInputs(unsigned short p_index, unsigned short p_subindex, int * p_intToReturn);

	int retrieveVariableValueInputConfiguration(unsigned short p_index, unsigned short p_subindex, int * p_intToReturn);
	int setHardwareValueInputConfiguration(unsigned short p_index, unsigned short p_subindex, int p_value);

	int retrieveVariableValueDigitalOutputs(unsigned short p_subindex, int * p_intToReturn);
	int setHardwareValueDigitalOutputs(unsigned short p_subindex, int p_value);

	int retrieveVariableValueDisplayBacklight(unsigned short p_subindex, int * p_intToReturn);
	int setHardwareValueDisplayBacklight(unsigned short p_subindex, int p_value);

	int retrieveVariableValueKeyboardBacklight(unsigned short p_subindex, int * p_intToReturn);
	int setHardwareValueKeyboardBacklight(unsigned short p_subindex, int p_value);

	int retrieveVariableValueBeeper(unsigned short p_subindex, int * p_intToReturn);
	int setHardwareValueBeeper(unsigned short p_subindex, int p_value);

	int retrieveVariableValueRTC(unsigned short p_subindex, int * p_intToReturn);
	int setHardwareValueRTC(unsigned short p_subindex, int p_value);

	int retrieveVariableValueCAN(unsigned int p_port, unsigned short p_subindex, int * p_intToReturn);
	int setHardwareValueCAN(unsigned int p_port, unsigned short p_subindex, int p_value);

	int retrieveVariableValueFileSystem(unsigned short p_subindex, int * p_intToReturn);
	int retrieveVariableValueFileSystem(unsigned short p_subindex, std::string * p_stringToReturn);
	int setHardwareValueFileSystem(unsigned short p_subindex, int p_value);
	int setHardwareValueFileSystem(unsigned short p_subindex, const std::string & p_value);

	int retrieveVariableValueEthernet(unsigned short p_index, unsigned short p_subindex, int * p_intToReturn);
	int retrieveVariableValueEthernet(unsigned short p_index, unsigned short p_subindex, std::string * p_stringToReturn);
	int setHardwareValueEthernet(unsigned short p_subindex, int p_value);
	int setHardwareValueEthernet(unsigned short p_subindex, const std::string & p_value);

	int retrieveVariableValueRS232(unsigned short p_subindex, int * p_intToReturn);
	int setHardwareValueRS232(unsigned short p_subindex, int p_value);

	int retrieveValuesWithPriority(WEHardwareVariablePriority p_priority);


	static unsigned int calculateCombinedKey(unsigned short p_varIndex, unsigned short p_varSubIndex);
	int addVariableWithPriority(unsigned short p_index, unsigned short p_subindex, WEHardwareVariablePriority p_priority);

	const std::string serverAddress;

	std::unique_ptr<HardwareInterface> hardwareInterface;

	std::unique_ptr<WEEventQueue> eventQueueIO;

	WEDEI * dei;
	WEEndPointManager * endPointManager;

	std::vector<unsigned int> highPriorityVariables;
	std::vector<unsigned int> mediumPriorityVariables;
	std::vector<unsigned int> lowPriorityVariables;
	std::vector<unsigned int> secondPriorityVariables;
	std::vector<unsigned int> minutePriorityVariables;
	std::vector<unsigned int> autoPriorityVariables;

	static const unsigned int TIME_PERIOD_HIGH_PRIORITY = 77;
	static const unsigned int TIME_PERIOD_MEDIUM_PRIORITY = 444;
	static const unsigned int TIME_PERIOD_LOW_PRIORITY = 1111;

	static const unsigned int TIME_PERIOD_SECOND = 1000;
	static const unsigned int TIME_PERIOD_MINUTE = 60000;

	bool initializationFinished;

	static std::string getVarPath(unsigned short p_index, unsigned short p_sub_index);
	static std::string getRemanentVariableString(const std::string & p_fileName);
	static void setDefaultForRemanentVariable(const std::string & p_fileName, const std::string & p_defaultValue, bool (functionPointer(const std::string &)));

	unsigned int timeSpentInCreateVariableAndAddToDEI;
	unsigned int timeSpentInRetreive;
	unsigned int timeSpentInStringRetreive;
	unsigned int timeSpentInIntSet;
	unsigned int timeSpentInStringSet;

	//configure watchdog interface to restart us when we hang for more than 10 seconds
	WE::WatchdogDaemonInterface wdd{"hardwareDaemon", "/opt/etc/init.d/01_hwdaemon restart", 10};
};

} // namespace WE

#endif /* WE_HARDWARE_DAEMON_H_ */
