/**
 *	@file we_hardware_interface.h
 *
 *	@author dsterzel@topcon.com
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2016
 *
 *	@brief Definition of class WEHardwareInterface that implements the sunrise abstract THardwareInterface
 */

#ifndef SRC_HARDWARE_SUNRISE_HARDWARE_LIB_SRC_INCLUDE_WE_HARDWARE_INTERFACE_H_
#define SRC_HARDWARE_SUNRISE_HARDWARE_LIB_SRC_INCLUDE_WE_HARDWARE_INTERFACE_H_

#include "THardwareInterface.h"	//!< The abstract interface for sunrise hardware access

#include "../../../../dei/variables/include/we_variable_definitions.h"

#include <QString>


/**
 * The WEHardwareInterface class implements the abstract THardwareInterface class.
 * All functions in this class are copied from the original abstract interface except of
 * constructor, destructor, the slots and everything in the private section.
 * You will not find any WE specific classes in this file.
 */
class WEHardwareInterface : public THardwareInterface
{
	Q_OBJECT

public:
	/**
	 * Constructor for creating the instance of the hardware interface.
	 * All initialization is done in the constructor and the connection to the
	 * WE hardware daemon is established. If hardware daemon is not running
	 * when the instance of this class is created, it will automatically
	 * try to connect to it later.
	 * A Qtimer is started inside the constructor to cyclic tasks.
	 */
	WEHardwareInterface();

	/**
	 * Free memory
	 */
	virtual ~WEHardwareInterface();

	/**
	 * \brief Get the consoles serial number.
	 * \return The serial number is a string of 20 characters
	 */
	virtual QString GetSerialNumber() override;

	// Ambient Light
	/**
	 * \brief Get the ambient light value.
	 * \return The ambient light as a percentage between 0 and 100.
	 */
	virtual double GetAmbientLight() override;

	/**
	 * \brief Set the hysterisis of the ambient light input.
	 * \param[in] hysterisis The hysterisis value.
	 * \return True if the set succeeded, false if there was an error.
	 */
	virtual bool SetAmbientLightHysterisis(double hysterisis) override;

	/**
	 * \brief Set the minimum threshold the ambient light can be before emitting AmbientLightMinThresholdReached().
	 * \param[in] threshold The ambient light minimum threshold.
	 * \return True if the set succeeded, false if there was an error.
	 */
	virtual bool SetAmbientLightMinThreshold(double threshold) override;

	/**
	 * \brief Set the maximum threshold the ambient light can be before emitting AmbientLightMaxThresholdReached().
	 * \param[in] threshold The ambient light maximum threshold.
	 * \return True if the set succeeded, false if there was an error.
	 */
	virtual bool SetAmbientLightMaxThreshold(double threshold) override;


	// Brightness
	/**
	 * \brief Get the screen brightness.
	 * \return The screen brightness as a percentage between 0 and 100.
	 */
	virtual double GetBrightness() override;

	/**
	 * \brief Set the screen brightness.
	 * \param[in] brightness The screen brightness as a percentage between 0 and 100.
	 * \return True if the set succeeded, false if there was an error.
	 */
	virtual bool SetBrightness(double brightness) override;


	// Power
	/**
	 * \brief Get the state of the screen backlight power
	 * \return True if the power is suspended, false if it isn't.
	 */
	virtual bool IsPowerSuspended() override;

	/**
	 * \brief Resume the backlight power.
	 * \return True if the resume succeeded, false if there was an error.
	 */
	virtual bool ResumePower() override;

	/**
	 * \brief Suspend the backlight power.
	 * \return True if the resume succeeded, false if there was an error.
	 */
	virtual bool SuspendPower() override;

	// Digital I/O
	/**
	 * \brief Disable digital output for a pin.
	 * \param[in] pin The pin to disable digital output on.
	 * \return True if the disable succeeded, false if there was an error.
	 */
	virtual bool DisableDigitalOutput(int pin) override;

	/**
	 * \brief Enable digital output for a pin.
	 * \param[in] pin The pin to enable digital output on.
	 * \return True if the enable succeeded, false if there was an error.
	 */

	virtual bool EnableDigitalOutput(int pin) override;
	/**
	 * \brief Get the status of a digital input pin.
	 * \param[in] pin The pin to get the status of.
	 * \return True if the pin is powered/on, false if it is unpowered/off.
	 */
	virtual bool IsDigitalInputOn(int pin) override;

	// Radar
	/**
	 * \brief Set the frequency of the radar output.
	 * \param[in] frequency The frequency in Hertz.
	 * \return True if the set succeeded, false if there was an error.
	 */
	virtual bool SetRadarFrequency(uint frequency) override;

	// Temperature
	/**
	 * \brief Get the console temperature.
	 * \return The console temperature in degrees celcius.
	 */
	virtual double GetTemperature() override;

	// Voltage
	/**
	 * \brief Get the supply voltage of the console.
	 * \return The console supply voltage in Volts.
	 */
	virtual double GetSupplyVoltage() override;

	/**
	 * \brief Get the battery voltage of the console.
	 * \return The console battery voltage in Volts.
	 */
	virtual double GetBatteryVoltage() override;

	/**
	 * \brief Get the presence of physical controls for the Virtual Terminal.
	 * \return True if the console has hardware buttons to control the Virtual Terminal.
	 */
	virtual bool HasPhysicalVTControls() override;

	public slots:
	;	// just stop Eclipse showing yellow lines any further

	/**
	 * Slot called by a cyclic timer to handle incomming variable change messages from hardware daemon
	 * and forward them to a listener (see variableValueChanged() below) which then will emit the appropriate signals defined above.
	 */
	void doCyclicTasks(void);

	/**
	 * Slot called after a variable change was received in the cyclic task.
	 * Emits the appropriate signals defined by the hardware interface.
	 * @param p_key Unique idetifier of the hardware variable that has changed (see definitions.h)
	 * @param value The new value of the hardware variable
	 */
	void variableValueChanged(unsigned int p_key, int p_value);

	private:

	/**
	 * Helper function to create a variable and add it to DEI (which then subscribes to hardware daemon)
	 * and create a connection with WEVariableEventDispatcher so we get noticed when a value changed in hardware daemon.
	 * This function has less parameters than normally needed to create a variable because we e.g.
	 * only create int variables in this interface and no strings.
	 * @param p_index	Index of the variable
	 * @param p_subindex	Subindex of the variable
	 * @param p_name	The name of the variable
	 * @param p_lengthInBit	Length of the variable in bit
	 * @param p_min		Minimum value of the variable
	 * @param p_max		Maximum value of the variable
	 * @param p_default	The default value of the variable
	 * @param p_initAsInvalid	Flag to indicate that the variable shall be marked as invalid until we received a valid value from the owner (hardware daemon)
	 * @param p_remanent	Flag to indicate that the latest value of the variable shall be stored and be reloaded after a cold start
	 * @param p_signed	Flag to tell if we have a signed or unsigned value
	 * @param p_forceWriting	Flag to indicate that a value change request is sent to the owner even if the "new" value is the same as the current value
	 * @param p_groupName	Name of the variable group
	 * @param p_storageType defaults to STORAGE_TYPE_INTEGER i.e. only has to be set for STORAGE_TYPE_STRING
	 */
	void createVariableAndSubscribe(
			unsigned short p_index,
			unsigned short p_subindex,
			std::string p_name,
			int p_lengthInBit,
			int p_min,
			int p_max,
			int p_default,
			bool p_initAsInvalid,
			bool p_remanent,
			bool p_signed,
			bool p_forceWriting,
			std::string p_groupName,
			WE::WEVariableStorageType p_storageType = WE::STORAGE_TYPE_INTEGER
	);

	/**
	 * Helper function in which all variables the interface is interested in are created
	 * via createVariableAndSubscribe() (see above)
	 */
	void subscribeToHardwareDaemon(void);
};



#endif /* SRC_HARDWARE_SUNRISE_HARDWARE_LIB_SRC_INCLUDE_WE_HARDWARE_INTERFACE_H_ */
