/**
 *	@file we_endpoint.h
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd Faisal
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Interface for defining an endpoint that can be managed by the WEEndPointManager
 *
 *	@see WEEndPointManager, WEIPCEndPoint, WECANRaw, WEPipe, WEFileEndPoint
 *
 */


#ifndef WE_ENDPOINT_H_
#define WE_ENDPOINT_H_

namespace WE
{
/**
 *	@class WEEndPoint
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd. Faisal
 *
 *	@brief Interface for defining an endpoint that can be managed by the WEEndPointManager
 *
 *	@namespace WE
 *
 *	@see WEEndPointManager, WEIPCEndPoint, WECANRaw, WEPipe, WEFileEndPoint
 *
 *	This is interface is implemented by the IPC socket, Can, pipe and file so that
 *	they all can be waited for readability in a single call endPointManager->waitForEndPoints()
 */
class WEEndPoint
{
public:
	/**
	 *	@fn WEEndPoint(bool p_autoReconnect)
	 *
	 *	@brief End point constructor.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_autoReconnect Set to true if automatic reconnect by the endPointManager is allowed,
	 *				   otherwise set p_autoReconnect to false.
	 *
	 *	This constructor will open the connection to the controlled endpoint.
	 *
	 */
	explicit WEEndPoint(bool p_autoReconnect)
	 :isAutoReconnectAllowed(p_autoReconnect)
	{
		//nothing more to do here
	}

	/**
	 *	@fn virtual ~WEEndPoint()
	 *
	 *	@brief End point destructor.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	The destructor will disconnect the connection to the controlled endpoint.
	 *
	 */
	virtual ~WEEndPoint() = default;

	/**
	 *	@fn getReadEndPointID()
	 *
	 *	@brief Returns the Identifier used to read from the end point.
	 *
	 *	@returns End point Identifier.
	 *
	 *	This could be same as the identifier returned from the getWriteEndPointID()
	 */
	virtual int getReadEndPointID() const = 0;

	/**
	 *	@fn getWriteEndPointID()
	 *
	 *	@brief Returns the Identifier used to write to the end point.
	 *
	 *	@returns End point Identifier.
	 *
	 *	This could be same as the identifier returned from the getReadEndPointID()
	 */
	virtual int getWriteEndPointID() const = 0;

	/**
	 *	@fn bool getAutoReconnectAllowed()
	 *
	 *	@brief Returns if auto reconnection by the endpoint manager is allowed.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return true if auto reconnection is allowed, otherwise false
	 *
	 */
	bool getAutoReconnectAllowed() const
	{
		return this->isAutoReconnectAllowed;
	}

	/**
	 *	@fn isConnected()
	 *
	 *	@brief Check whether the end point is still connected to the peer entity
	 *			according to the last access result not the current status.
	 *
	 *	@returns true If last access was successful.
	 *	@returns false If last access produced error.
	 */
	virtual bool isConnected() const = 0;

	/**
	 *	@fn reconnect()
	 *
	 *	@brief reconnect the endpoint to the same peer with which it was connected earlier
	 *
	 *	@returns WE_SUCCESS If connect with server is re-established successfully.
	 *	@returns WE_INIT_ERROR If end point could not be reinitialized successfully.
	 *	@returns WE_TIMEOUT_ERROR If connection could not be established before connectTimeout expires.
	 *	@returns WE_CONNECT_ERROR If connection could not be established for any other reason.
	 *
	 *
	 *	It depends on the type of the end point if it is really useful because there may be cases
	 *	where endpoint seems always connected e.g; WEFileEndpoint.
	 */
	virtual int reconnect() = 0;

private:
	bool isAutoReconnectAllowed;	//!< true if the endpoint manager is allowed to reconnect the endpoint if needed.
};
}	//namespace WE

#endif /* WE_ENDPOINT_H_ */
