/**
 *	@file we_health_management.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Class to send messages to health management daemon
 *
 */


#ifndef WE_HEALTH_MANAGEMENT_H_
#define WE_HEALTH_MANAGEMENT_H_

namespace WE {

/**
 *	@enum WEHealthManagementCommand
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief enum with commands for the health management daemon
 *
 *	@namespace WE
 */
typedef enum
{
	WE_HEALTH_CMD_GET_OTC = 0,
	WE_HEALTH_CMD_OTC,
	WE_HEALTH_CMD_GET_EE_PRODUCT_TYPE,
	WE_HEALTH_CMD_EE_PRODUCT_TYPE,
	WE_HEALTH_CMD_GET_EE_PRODUCT_VARIANT,
	WE_HEALTH_CMD_EE_PRODUCT_VARIANT,
	WE_HEALTH_CMD_GET_EE_BOARD_REVISION,
	WE_HEALTH_CMD_EE_BOARD_REVISION,
	WE_HEALTH_CMD_GET_EE_BOARD_NAME,
	WE_HEALTH_CMD_EE_BOARD_NAME,
	WE_HEALTH_CMD_GET_EE_PART_NUMBER,
	WE_HEALTH_CMD_EE_PART_NUMBER,
	WE_HEALTH_CMD_GET_EE_SERIAL_NUMBER,
	WE_HEALTH_CMD_EE_SERIAL_NUMBER,
	WE_HEALTH_CMD_GET_EE_BOOT_MODE,
	WE_HEALTH_CMD_EE_BOOT_MODE,
	WE_HEALTH_CMD_GET_EE_UNIQUE_RELEASE_ID,
	WE_HEALTH_CMD_EE_UNIQUE_RELEASE_ID,
	WE_HEALTH_CMD_GET_EE_DELETE_DIAG,
	WE_HEALTH_CMD_EE_DELETE_DIAG,
	WE_HEALTH_CMD_GET_EE_CONSOLE_MODE,
	WE_HEALTH_CMD_EE_CONSOLE_MODE,
	WE_HEALTH_CMD_GET_API_VERSION,
	WE_HEALTH_CMD_API_VERSION,
	WE_HEALTH_CMD_ERROR,
	WE_HEALTH_CMD_SET_EE_CONSOLE_MODE,
	WE_HEALTH_CMD_GET_EE_DISPLAY_BRIGHTNESS,	//p_data is char* (0..255)
	WE_HEALTH_CMD_SET_EE_DISPLAY_BRIGHTNESS,	//p_data is char* (0..255)
	WE_HEALTH_CMD_EE_DISPLAY_BRIGHTNESS,
	WE_HEALTH_CMD_GET_EE_KEYPAD_BRIGHTNESS,		//p_data is char* (0..255)
	WE_HEALTH_CMD_SET_EE_KEYPAD_BRIGHTNESS,		//p_data is char* (0..255)
	WE_HEALTH_CMD_EE_KEYPAD_BRIGHTNESS,
	WE_HEALTH_CMD_GET_EE_KEYPAD_COLOR,		//p_data is char* (0..0xffffff)
	WE_HEALTH_CMD_SET_EE_KEYPAD_COLOR,		//p_data is char* (0..0xffffff)
	WE_HEALTH_CMD_EE_KEYPAD_COLOR,
	WE_HEALTH_CMD_SET_EE_BOOT_MODE
} WEHealthManagementCommand;


/**
 *	@class WEHealthManagement
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Utility class to send a message to the health management daemon
 *
 *	@namespace WE
 *
 *	@warning Currently only compatible with A3 BSP 1.1.0. On other either no implementation is present
 *	or WE_NOT_SUPPORTED_ERROR will be returned for all functions.
 *
 *	This class only has a static function to send a command to the health management daemon.
 *	It can be waited for the answer or not. See member function documentation
 */
class WEHealthManagement
{
public:
	/**
	 *	@fn ~WEHealthManagement()
	 *
	 *	@brief Destructor that does nothing
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WEHealthManagement() = delete;
	WEHealthManagement(const WEHealthManagement & p_copy) = delete;
	WEHealthManagement& operator=(const WEHealthManagement& p_copy) = delete;

	/**
	 *	@fn static int sendMessage(WEHealthManagementCommand p_command, void * p_data, bool waitForAnswer = false);
	 *
	 *	@brief static function to send a command to the health management daemon
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_command Command to send to the health management daemon
	 *	@param p_data Pointer to the data that fits to the command (platform and command dependent)
	 *		if waitForAnswer is true, the result will also be copied into the p_data pointer
	 *	@param waitForAnswer true: The function returns after an answer was received from the daemon
	 *		false: The function returns immediately after the command was sent
	 *
	 *	@returns WE_SUCCESS if the command was successfully sent and the answer correctly received (if requested by "waitForAnswer")
	 *	@returns WE_NOT_SUPPORTED_ERROR for any command that is not supported
	 *	@returns WE_SEND_ERROR if the message could not be sent to the health management daemon
	 *	@returns WE_RECEIVE_ERROR if the requested answer could not be received
	 *	@returns WE_GENERAL_ERROR if the received answer does not fit the sent command
	 *
	 *	@warning Only the following commands are currently supported on OPUS A3 BSP >= 1.1.0:
	 *		WE_HEALTH_CMD_GET_EE_DISPLAY_BRIGHTNESS, WE_HEALTH_CMD_SET_EE_DISPLAY_BRIGHTNESS,
	 *		WE_HEALTH_CMD_GET_EE_KEYPAD_BRIGHTNESS,	WE_HEALTH_CMD_SET_EE_KEYPAD_BRIGHTNESS
	 *
	 *	Static function to send a command to the health management daemon and eventually wait for an answer.
	 */
	static int sendMessage(WEHealthManagementCommand p_command, void * p_data, bool p_waitForAnswer = false);

	/**
	 *	@fn WEHealthManagement()
	 *
	 *	@brief private constructor. Only static functions and no member.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WEHealthManagement() = delete;
};
}  // namespace WE


#endif /* WE_HEALTH_MANAGEMENT_H_ */
