/**
 *	@file we_network_config.h
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief factory and wrapper to get and set network interface information
 */

#ifndef WE_NETWORK_CONFIG_H_
#define WE_NETWORK_CONFIG_H_

namespace WE
{
#define WE_MAX_ETHERNET_ADAPTER 4 //!< max. ethernet adapters to be handled

class WENetwork; //!< forward declaration
class WENetworkConfig; //!< forward declaration

class WENetworkConfigFactory
{
public:
	WENetworkConfigFactory();
	virtual ~WENetworkConfigFactory();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WENetworkConfigFactory(const WENetworkConfigFactory &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WENetworkConfigFactory & operator=(const WENetworkConfigFactory &) = delete;

	/**
	 *	@fn int create(int p_adapter_index)
	 *
	 *	@brief creates instance of the adapter
	 *
	 *	@param p_adapter_index > 0 and < WE_MAX_ETHERNET_ADAPTER
	 *
	 *	@returns -1 when device is not available
	 *	@returns 0 when ok
	 */
	int create(int p_adapterIndex);

	/**
	 *	@fn void release(int p_adapter_index)
	 *
	 *	@brief kills the instance
	 *
	 *	@param p_adapter_index > 0 and < WE_MAX_ETHERNET_ADAPTER
	 *
	 *	@returns n/a
	 */
	void release(int p_adapterIndex);

	/**
	 *	@fn WENetworkConfig* getConfig(int p_adapter_index);
	 *
	 *	@brief returns pointer to WENetworkConfig
	 *
	 *	@param p_adapter_index > 0 and < WE_MAX_ETHERNET_ADAPTER
	 *
	 *	@returns pointer to WENetworkConfig
	 *	@returns nullptr when no instance
	 *	@see WENetworkConfig
	 */
	const WENetworkConfig* getConfig(int p_adapterIndex) const;

	/**
	 *	@fn bool setNetworkSettings(int p_adapter_index, const std::string & p_ip_address, const std::string & p_netmask, const std::string & p_gateway);
	 *
	 *	@brief set the network configuration
	 *
	 *	@param p_adapter_index > 0 and < WE_MAX_ETHERNET_ADAPTER
	 *	@param p_ip_address valid address, syntax will be checked
	 *	@param p_netmask valid address, syntax will be checked
	 *	@param p_gateway valid address, syntax will be checked
	 *
	 *	@returns true settings could be set
	 *	@returns false when settings could not be applied
	 */
	bool setNetworkSettings(int p_adapterIndex, const std::string & p_ipAddress, const std::string & p_netmask, const std::string & p_gateway);

	/**
	 *	@fn bool readNetworkSettings(int p_adapter_index)
	 *
	 *	@brief reads the settings from an adapter
	 *
	 *	@param p_adapter_index > 0 and < WE_MAX_ETHERNET_ADAPTER
	 *
	 *	@returns true when read was ok
	 *	@returns false when read failed
	 */
	bool readNetworkSettings(int p_adapterIndex);

	/**
	 *	@fn void releaseAll()
	 *
	 *	@brief kill all created instances
	 *
	 *	@returns n/a
	 */
	void releaseAll();

	WENetwork* getEthInstance(int p_adapterIndex);

	const std::string & getDefaultIPAdress(int p_adapterIndex) const;
	const std::string & getDefaultNetmaskIP(int p_adapterIndex) const;
	const std::string & getDefaultGatewayIP(int p_adapterIndex) const;

private:

	/**
	 *	@fn void init()
	 *
	 *	@brief more or less do a reset
	 *
	 *	@returns n/a
	 */
	void init();

	/**
	 *	@fn void reset(int p_adapter_index)
	 *
	 *	@brief resets all settings
	 *
	 *	@param p_adapter_index > 0 and < WE_MAX_ETHERNET_ADAPTER
	 *
	 *	@returns n/a
	 */
	void reset(int p_adapterIndex);

	WENetwork* eth_instance[WE_MAX_ETHERNET_ADAPTER]{};

	static const std::string emptyString;
};
}
#endif /* WE_NETWORK_CONFIG_H_ */
