/**
 *	@file we_pipe.h
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd Faisal
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief This file defines the class WEPipe which can be used for thread synchronization
 *	as long as there is only one write end and may be multiple read ends.
 *
 *	DESCRIPTION
*/


#ifndef WE_PIPE_H_
#define WE_PIPE_H_

#include "platform_specific_settings.h"
#include "../../endpoint_manager/include/we_endpoint.h"

namespace WE
{



/**
 *	@class WEPipe
 *
 *	@author mohd.faisal@aricent.com
 *	@author Mohd Faisal
 *
 *	@brief This class is an abstraction for the pipe which reads data from one end and writes them to buffer at the other end
 *		This pipe is supposed to be used for EventQueue and ActionQueue
 *		It is a WEEndPoint and can be used in WEEndPointManager
 */
class WEPipe : public WEEndPoint
{
private:
	bool isInitialized{false};	//!< true if pipe is initialized else it is false
	int readEndpointID{-1};	//!< read end of the pipe
	int writeEndpointID{-1};	//!< write end of the pipe
	PlatformSpecificPipeSettings * settings{nullptr};
	int myDebugKey{-1};
public:

	/**
	 *	@fn WEPipe()
	 *
	 *	@brief default constructor for the class WEPipe. init() needs to be called
	 *	on the pipe before it can be used for reading and writing.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *
	 */
	WEPipe();
	WEPipe(const WEPipe & p_copy) = delete;
	WEPipe& operator=(const WEPipe & p_copy) = delete;

	/**
	 *	@fn WEPipe
	 *
	 *	@brief the destructor for the class WEpipe which closes both read write ends of pipe.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 */
	virtual ~WEPipe();

	/**
	 *	@fn init
	 *
	 *	@brief this function creates read write ends of the pipe. after successful return
	 *	pipe can be used for reading and writing.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *  @returns WE_SUCCESS if the pipe is created successfully
	 *  @returns WE_INIT_ERROR if the pipe creation failed
	 *                         or is already initialized
	 */
	int init();

	/**
	 *	@fn deinit
	 *
	 *	@brief deallocates the system resources by closing both read write ends of pipe.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *  @returns WE_SUCCESS
	 */
	int deinit();

	/**
	 *	@fn readFromPipe(void* bufferToReadInto , int bytesToRead)
	 *
	 *	@brief reads specified number of bytes from the pipe into the buffer.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@param[out] bufferToReadInto pointer to the buffer where the data read from the pipe is copied.
	 *	@param[in] bytesToRead size of the data to be read.
	 *
	 *	@returns WE_SUCCESS if the reading process is successful
	 *	@returns WE_READ_ERROR if requested number of bytes could not be read OR
	 *							end of file is reached (although not likely) OR
	 *							reading fails for unknown reasons
	 *	@returns WE_INVALID_PARAMETER_ERROR if bufferToReadInto is nullptr or bytesToRead <= 0 or pipe is uninitialized.
	 *
	 */
	int readFromPipe(void* p_bufferToReadInto , int p_bytesToRead);

	/**
	 *	@fn writeIntoPipe(void* bufferToWrite , int bytesToWrite)
	 *
	 *	@brief this function writes the data on the buffer into the pipe
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@param[in] bufferToBeWritten buffer to be written to the pipe.
	 *	@param[in] bytesToWrite number of bytes of buffer to write to the pipe.
	 *
	 *	@returns WE_SUCCESS is successful
	 *	@returns WE_WRITE_ERROR if requested number of bytes could not be written OR
	 *							writing fails for unknown reasons
	 *	@returns WE_INVALID_PARAMETER_ERROR if bufferToWrite is nullptr or bytesToWrite <= 0 or pipe is uninitialized.
	 */
	int writeIntoPipe(void* p_bufferToBeWritten , int p_bytesToWrite) const;

	/*
	 *
	 *	@fn getReadEndPointID()
	 *
	 *	@brief Returns the Identifier used to read from the end point.
	 *
	 *	@returns Read End point Identifier.
	 *
	 *	This is a different the identifier as returned from the getWriteEndPointID()
	 */
	int getReadEndPointID() const override;

	/*
	 *	@fn getWriteEndPointID()
	 *
	 *	@brief Returns the Identifier used to write to the end point.
	 *
	 *	@returns Write End point Identifier.
	 *
	 *	This is a different the identifier as returned from the getReadEndPointID()
	 */
	int getWriteEndPointID() const override;

	/*
	 *	@fn isConnected()
	 *
	 *	@brief This should be connected unless it is uninitialized / deinitialized.
	 *
	 *	@returns true If last access was successful.
	 *	@returns false If uninitialized / deinitialized.
	 */
	bool isConnected() const override;

	/**
	 *	@fn reconnect()
	 *
	 *	@brief reconnect the endpoint to the same peer with which it was connected earlier
	 *
	 *	@returns WE_SUCCESS If connect with server is re-established successfully.
	 *	@returns WE_INIT_ERROR If end point could not be reinitialised successfully.
	 *	@returns WE_CONNECT_ERROR If connection could not be established for any other reason.
	 *
	 *
	 *	It depends on the type of the end point if it is really useful because there may be cases
	 *	where endpoint seems always connected e.g; WEFileEndpoint.
	 */
	int reconnect() override;

	/**
	 *	@fn isReadable()
	 *
	 *	@brief checks whether data is available in the Pipe to be read or not
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@returns true if data is available to be read
	 *	@returns false if the pipe is empty
	 *
	 */
	bool isReadable() const;

	/**
	 *	@fn getNoOfBytesAvailable()
	 *
	 *	@brief finds the number of bytes available in the pipe
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd Faisal
	 *
	 *	@returns number of bytes available in the pipe
	 *
	 */
	int getNoOfBytesAvailable() const;

	/**
	 *	@fn clear()
	 *
	 *	@brief remove all the events still there in the pipe.
	 *
	 *	@author mohd.faisal@aricent.com
	 *	@author Mohd. Faisal
	 *
	 *	@returns WE::WE_SUCCESS
	*/
//	int clear();
};
}
#endif /* WE_PIPE_H_ */
