/**
 *	@file we_power_management.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of WEAPI interface functions for accessing Power Management Daemon
 *
 *	The interface defined in this file provides functions for accessing
 *	power management API on an OPUS platform.
 */

#ifndef WE_POWER_MANAGEMENT_H_
#define WE_POWER_MANAGEMENT_H_

#include <memory> //unique_ptr

#include "../../endpoint_manager/include/we_endpoint_handler.h"

namespace WE
{

class WESignalListener;

typedef enum
{
	WE_POWER_STATE_ON,
	WE_POWER_STATE_LOW_POWER,
	WE_POWER_STATE_SLEEP,
	WE_POWER_STATE_OFF,
	WE_POWER_STATE_SILENT_ON,
	WE_POWER_STATE_LOW_POWER_TIMER_ON,
	WE_POWER_STATE_REBOOT
}wePowerStates;

typedef enum
{
	WE_POWER_STATE_TRANSITIONS_ON_TO_LOW_POWER,
	WE_POWER_STATE_TRANSITIONS_LOW_POWER_TO_ON,
	WE_POWER_STATE_TRANSITIONS_LOW_POWER_TO_SLEEP,
	WE_POWER_STATE_TRANSITIONS_SLEEP_TO_ON,
	WE_POWER_STATE_TRANSITIONS_ON_TO_OFF,
	WE_POWER_STATE_TRANSITIONS_LOW_POWER_TO_OFF
}wePowerStateTransitions;

typedef enum
{
	WE_POWER_MESSAGE_TIMEOUT,
	WE_POWER_MESSAGE_DEVICE,
	WE_POWER_MESSAGE_DEVICELIST,
	WE_POWER_MESSAGE_MODE,
	WE_POWER_MESSAGE_MODE_CHANGE
}wePowerManagementValues;

enum weDeviceStates
{
	DEVICE_STATE_OFF,
	DEVICE_STATE_ON
};

enum PowerDevices
{
	WE_POWER_DEVICE_DISPLAY,
	WE_POWER_DEVICE_DISPLAY_BACKLIGHT,
	WE_POWER_DEVICE_KEYPAD_BACKLIGHT,
	WE_POWER_DEVICE_ENCODER_BACKLIGHT,
	WE_POWER_DEVICE_RGB_LED,
	WE_POWER_DEVICE_BEEPER,
	WE_POWER_DEVICE_KEYPAD,
	WE_POWER_DEVICE_ENCODER,
	WE_POWER_DEVICE_TOUCH,
	WE_POWER_DEVICE_CONSOLE,
	WE_POWER_DEVICE_OUTPUT_0,
	WE_POWER_DEVICE_OUTPUT_1,
	WE_POWER_DEVICE_OUTPUT_2,
	WE_POWER_DEVICE_CAM_OUTPUT,
	WE_POWER_DEVICE_VIDEO,
	WE_POWER_DEVICE_INPUT_0,
	WE_POWER_DEVICE_INPUT_1,
	WE_POWER_DEVICE_INPUT_2,
	WE_POWER_DEVICE_INPUT_3,
	WE_POWER_DEVICE_STATUS_LED_1,
	WE_POWER_DEVICE_STATUS_LED_2
};

/**
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief Interface for power management
 *
 *	@namespace WE
 *
 *	This class provides an platform independent interface for accessing
 *	the power management of an OPUS platform.
 *	Use the init method to connect the class to the power management of the
 *	operating system.
 */
class WEPowerManagement : public WEEndPointHandler
{	// We don't check return values of all class member functions since 2011.
public:

	/**
	 *	@brief Standard destructor
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	virtual ~WEPowerManagement() = default;

	/**
	 *	@brief Creates an instance of WEPowerManagement
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns unique_ptr to the appropriate object.
	 *
	 *	@see WEPowerManagementInterfaceA3
	 *
	 *	This function creates an instance of WEPowerMangement for the
	 *	appropriate platform. It is implemented by the derived classes
	 *	who will create a new object and return a pointer to it.
	 *	With this function the factory pattern is realized.
	 */
	static std::unique_ptr<WEPowerManagement> create(void);

	/**
	 *	@brief Open connection to Power Management Daemon and reads out the available hardware devices
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS if establishing connection and reading was successful
	 *	@returns WE_SOCKET_ERROR if socket can not be created
	 *	@returns WE_CONNECT_ERROR if opening connection failed
	 *	@returns WE_NOT_SUPPORTED_ERROR on a platform that does not support power management
	 */
	virtual int init(void) = 0;

	/**
	 *	@brief Close connection to Power Management Daemon
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS when socket was closed successful
	 *	@returns WE_FILE_ERROR when closing socket failed
	 *	@returns WE_NOT_SUPPORTED_ERROR on a platform that does not support power management
	 */
	virtual int release(void) = 0;

	/**
	 *	@brief Set the event listener for power management events
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] eventListener eventListener that has to be invoked on occurrence of an event.
	 *
	 *	@returns WE_SUCCESS when the eventListener was registered successful
	 *	@returns WE_INVALID_PARAMETER_ERROR eventListener was invalid
	 *	@returns WE_NOT_SUPPORTED_ERROR on a platform that does not support power management
	 *
	 *	Data read from power management daemon are evented to the eventListener.
	 */
	virtual int setEventListener(WESignalListener * p_eventListener) = 0;

	/**
	 *	@brief Set timeout limit
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_state Holds the corresponding power mode. Allowed values are
	 *	POWER_STATE_LOW_POWER and POWER_STATE_SLEEP
	 *	@param[in] p_timeout Timeout limit in seconds to set
	 *
	 *	@returns WE_SUCCESS if setting timeout was successful
	 *	@returns WE_WRITE_ERROR if setting timeout value fails
	 *	@returns WE_INVALID_PARAMETER_ERROR if timeout is not supported by given p_state
	 *	@returns WE_NOT_SUPPORTED_ERROR on a platform that does not support power management
	 */
	virtual int setTimeoutForState(wePowerStates p_state, unsigned int p_timeout) = 0;

	virtual int setPropagationTimeForState(wePowerStates p_state, unsigned int p_time) = 0;

	/**
	 *	@brief Get timeout limit
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_state Holds the corresponding power mode. Allowed values are
	 *	POWER_STATE_LOW and POWER_STATE_SLEEP
	 *
	 *	@returns WE_SUCCESS if reading timeout values was successful
	 *	@returns WE_READ_ERROR if reading timeout value failed
	 *	@returns WE_NOT_SUPPORTED_ERROR on a platform that does not support power management
	 */
	virtual int getTimeoutForState(wePowerStates p_state) = 0;

	virtual int getPropagationTimeForState(wePowerStates p_state) = 0;


	/**
	 *	@brief Set device state for low-power mode
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_device Holds the corresponding device id
	 *	@param[in] p_state Holds the low-power mode state (on (1) / off (0))
	 *
	 *	@returns WE_SUCCES if the set request was successfully forwarded to the hardware (e.g. power daemon)
	 *	@returns Something else on error
	 *
	 */
	virtual int setDeviceStateForLowPower(int p_device, int p_state) = 0;

	/**
	 *	@brief Get device state for low-power mode
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_device Holds the corresponding device id
	 *
	 *	@returns WE_SUCCESS if the get rquest was forwarded to power daemon.
	 *
	 *	@warning The answer to the get request will come via an event!
	 */
	virtual int getDeviceStateInLowPower(int p_device) = 0;

	/**
	 *	@brief Set power mode
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_state The power state to set
	 *
	 *	@returns WE_SUCCESS if setting power state was successful
	 *	@returns WE_WRITE_ERROR if setting power state failed
	 *	@returns WE_OUT_OF_BOUNDS_ERROR if power state is unknown
	 *	@returns WE_INVALID_PARAMETER_ERROR if it is not possible to switch into given power state
	 *	@returns WE_NOT_SUPPORTED_ERROR on a platform that does not support power management
	 */
	virtual int setPowerState(wePowerStates p_state) = 0;

	/**
	 *	@brief Get current power state
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS if reading timeout values was successful
	 *	@returns WE_READ_ERROR if reading timeout value failed
	 *	@returns WE_NOT_SUPPORTED_ERROR on a platform that does not support power management
	 */
	virtual int getPowerState(void) = 0;

	/**
	 *	@fn virtual int subscribeToEndpointManager(WEEndPointManager * p_mgr) = 0
	 *
	 *	@brief Add power management as an endpoint to the end point manager
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_mgr Pointer to the endpoint manager to which to subscribe
	 *
	 *	@returns WE_SUCCESS if adding the power management as an endpoint to the endpoint manager was successful
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_mgr is nullptr
	 *
	 *	@see int WEEndPointManager::addEndPoint(WEEndPoint* endPointToAdd, WEEndPointHandler* handler)
	 *
	 *	This function is used to add this power management  as an endpoint to the endpoint manager.
	 *	As soon as the power management is readable (a message is received) the endpoint manager calls
	 *	the "onEndPointReadable" function of the power management.
	 */
	virtual int subscribeToEndpointManager(WEEndPointManager * p_mgr) = 0;

	/**
	 * Disable the transition from on to lowPower via short press of power button.
	 * Reason: this transition is application dependent and shall be configured in the project
	 */
	virtual int disablePowerbuttonOnToLowPower(void) = 0;

	static int debugSection;
};
} /* namespace WE */

#endif /* WE_POWER_MANAGEMENT_H_ */
