/**
 *	@file we_power_management.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of WEAPI interface functions for accessing Power Management on Linux PC.
 *
 */

#ifndef WE_POWER_MANAGEMENT_INTERFACE_WINDOWS_PC_H_
#define WE_POWER_MANAGEMENT_INTERFACE_WINDOWS_PC_H_

#include "../include/we_power_management.h"

namespace WE
{


class WEPowerManagementInterfaceWindowsPC : public WEPowerManagement
{	// Works without 'check return values of all class member functions' since 2011
public:

	/**
	 *	@brief Standard destructor
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	virtual ~WEPowerManagementInterfaceWindowsPC() {};

	/**
	 *	@brief Open connection to Power Management Daemon and reads out the available hardware devices
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR
	 */
	int init(void) override;

	/**
	 *	@brief Close connection to Power Management Daemon
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_NOT_SUPPORTED
	 */
	int release(void) override;

	/**
	 *	@brief Set the event listener for power management events
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] eventListener eventListener that has to be invoked on occurrence of an event.
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR on a platform that does not support power management
	 */
	int setEventListener(WESignalListener * p_eventListener) override;

	/**
	 *	@brief Set timeout limit
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_state Holds the corresponding power mode. Allowed values are
	 *	POWER_STATE_LOW_POWER and POWER_STATE_SLEEP
	 *	@param[in] p_timeout Timeout limit in seconds to set
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR
	 */
	int setTimeoutForState(wePowerStates p_state, unsigned int p_timeout) override;

	int setPropagationTimeForState(wePowerStates p_state, unsigned int p_time) override;

	/*
	 *	@brief Get timeout limit
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_state Holds the corresponding power mode. Allowed values are
	 *	POWER_STATE_LOW and POWER_STATE_SLEEP
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR
	 */
	int getTimeoutForState(wePowerStates p_state) override;

	int getPropagationTimeForState(wePowerStates p_state) override;


	/**
	 *	@brief Set device state for low-power mode
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_device Holds the corresponding device id
	 *	@param[in] p_state Holds the low-power mode state (on/off)
	 *
	 *	@returns value -1 because platform does not support power management.
	 */
	int setDeviceStateForLowPower(int p_device, int p_state) override;

	/**
	 *	@brief Get device state for low-power mode
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_device Holds the corresponding device id
	 *	@param[in] p_state The read low-power mode state (on/off)
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR
	 */
	int getDeviceStateInLowPower(int p_device) override;

	/**
	 *	@brief Set power mode
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_state The power state to set
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR
	 */
	int setPowerState(wePowerStates p_state) override;

	/**
	 *	@brief Get current power state
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR
	 */
	int getPowerState(void) override;

	/**
	 *	@fn onEndPointReadable(WEEndPoint* endPoint, WEEndPointManager* endPointMgr)
	 *
	 *	@brief Invoked everytime from the waitForEndPoints() if the corresponding endpoint
	 *		becomes readable.
	 *
	 **	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endPoint The End point that is found to be readable.
	 *	@param p_endPointMgr The End Point manager managing the endpoint.
	 *
	 *	Generally it is invoked when a connection request is received on a server endpoint
	 *	or when a message is received on a client endpoint. For a server endpoint
	 *	WEIPCServer::acceptConnection() inside this function should return successfully.
	 *	And for a client endpoint WEIPCEndPoint::receiveMessage() inside this function
	 *	should return successfully.
	 *	It might also be invoked when connection is broken but only once. This is becuase
	 *	the connection status is not the immediate status but based on the last access results.
	 *	In such a case WEIPCEndPoint::receiveMessage() will return WE_RECEIVE_ERROR, therefore
	 *	for a failing WEIPCEndPoint::receiveMessage() it should be checked if the endpoint is
	 *	still connected using endPoint->isConnected(). And if it is not connected the endPoint
	 *	could be removed from the endPointManager using WEEndPointManager::removeEndPoint().
	 */
	void onEndPointReadable(std::shared_ptr<WEEndPoint> p_endPoint, WEEndPointManager* p_endPointMgr) override;

	/**
	 *	@fn void onEndPointReconnected(WEEndPoint* endpoint)
	 *
	 *	@brief This function gets called every time a endpoint gets re-connected
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endpoint The endpoint which was reconnected
	 */
	void onEndPointReconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;

	/**
	 *	@fn void onEndPointDisconnected(WEEndPoint* p_endpoint)
	 *
	 *	@brief Call this function if a IPC connection was lost
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endpoint Endpoint of the connection that was disconnected
	 *
	 */
	void onEndPointDisconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;

	/**
	 *	@fn int subscribeToEndpointManager(WEEndPointManager * p_mgr)
	 *
	 *	@brief Add power management as an endpoint to the end point manager
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_mgr Pointer to the endpoint manager to which to subscribe
	 *
	 *	@returns WE_SUCCESS if adding the power management as an endpoint to the endpoint manager was successful
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_mgr is nullptr
	 *	@returns WE_ALREADY_EXISTS_ERROR if the power management was already registered.
	 *
	 *	@see int WEEndPointManager::addEndPoint(WEEndPoint* endPointToAdd, WEEndPointHandler* handler)
	 *
	 *	This function is used to add this power management  as an endpoint to the endpoint manager.
	 *	As soon as the power management is readable (a message is received) the endpoint manager calls
	 *	the "onEndPointReadable" function of the power management.
	 */
	int subscribeToEndpointManager(WEEndPointManager * p_mgr) override;

	virtual int disablePowerbuttonOnToLowPower(void) override;

};

} /* namespace WE */

#endif /* WE_POWER_MANAGEMENT_INTERFACE_WINDOWS_PC_H_ */
