/**
 *	@file we_thread.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Interface for using threads within WEAPI applications
 *
 *	The class defined in this file can be used for building multi threaded
 *	applications. It is also used internally in WEAPI.
 */

#ifndef WE_THREAD_H_
#define WE_THREAD_H_

#include <string>
#include <thread>

namespace WE
{

enum scheduler
{
	WE_SCHEDULER_NON_REALTIME,	// default non realtime scheduler
	WE_SCHEDULER_ROUND_ROBIN,	// real time scheduler using round robin
	WE_SCHEDULER_FIFO		// real time scheduler using FIFO
};

/**
 *	@class WEThread
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Basic interface for using threads inside WEAPI applications
 *
 *	@namespace WE
 *
 *	This class provides a simple interface for using threads in applications.
 *	Usage: Create a thread: WEThread * thread = WEThread::create();
 *	After creating start a thread: thread->startThread(functionPointer, objPtr).
 *	The function pointer should point to a static function, "anyObject" is a void pointer
 *	to any object that you want to use inside the static function.
 *	The join() function should be called to synchronize the threads when finishing them.
 *	Please also have a look at thread synchronization, e.g. via std::mutex.
 */
class WEThread
{
public:
	/**
	 * Default constructor. Members initialized via nsdmi.
	 * Use startThread() to create a std::thread and start it
	 */
	WEThread() = default;

	/**
	 * Constructor taking the name of the thread as parameter.
	 * Use startThread() to create a std::thread and start it
	 */
	explicit WEThread(const std::string & p_threadName);

	/**
	 * Destructor. Calls join()
	 */
	~WEThread();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEThread(const WEThread &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEThread & operator=(const WEThread &) = delete;

	/**
	 *	@fn virtual void startThread(void (*functionPointer)(void*), void * objPtr) = 0
	 *
	 *	@brief With this function a thread is started. A function that runs inside the thread has to be provided.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param functionPointer A pointer to a (static) function that represent the thread functionality
	 *	@param objPtr A pointer to any object can be provided that can be used inside the threads function.
	 *
	 *	@returns WE_SUCCESS if thread could be started.
	 *	@returns WE_INTERFACE_ERROR if boost::thread could not be created
	 *
	 *	@see Example implementation
	 *
	 *	With this function the created thread can be started. A function pointer
	 *	to a static function has to be provided. This function represents the functionality
	 *	of the thread. As second parameter a pointer to any existing object may be provided
	 *	so that it can be used inside the static function.
	 */
	int startThread(void (*p_functionPointer)(void*), void * p_objPtr);

	/**
	 *	@fn virtual int join(void) = 0
	 *
	 *	@brief When a thread finishes it has to be joined with the main thread.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS if the thread exists, is joinable and was successfully joined,
	 *		WE_GENERAL_ERROR if the thread was not started or is not joinable (e.g. after detach() or if the thread was not started)
	 *
	 *	@warning Always join threads to avoid zombies!
	 *	@warning Make sure that the thread exits its thread loop before calling join! join() will block until the thread has finished!
	 *
	 *	After starting a thread, this function should be called to join
	 *	a finished thread with the main thread.
	 */
	int join(void);

	/**
	 * Calls std::thread's detach:
	 * Separates the thread of execution from the thread object, allowing execution to continue independently.
	 * Any allocated resources will be freed once the thread exits.
	 * After calling detach *this no longer owns any thread.
	 */
	void detach(void);

	/**
	 *	@fn virtual int getPriority(int &p_scheduler, int &p_prio)
	 *
	 *	@brief Reads the scheduler and the priority of this thread.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_scheduler The scheduler queue that is used to schedule the thread. Values are defined by 'enum scheduler'.
	 *	@param[in] p_prio The priority of the thread. Use command 'chrt -m' to check on your target which values are valid.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	REads the real time scheduler and priority of the thread.
	 *
	 *	@see we_thread.h
	 */
	int getPriority(int &p_scheduler, int &p_prio);

	/**
	 *	@fn virtual int setPriority(int p_scheduler, int p_prio)
	 *
	 *	@brief Changes the scheduler and the priority of this thread.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param[in] p_scheduler The scheduler queue to be used and defined by enum scheduler.
	 *	@param[in] p_prio The priority to be set for the thread. Use command 'chrt -m' to check on your target which values are valid.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	Sets the rea ltime scheduler and priority of the thread.
	 *
	 *	@WARNING: The process needs to have proper rights to change it's thread priorities (e.g. executed as root).
	 *
	 *	@see we_thread.h
	 */
	int setPriority(int p_scheduler, int p_prio);

	/**
	 * On Linux this function sends a signal with p_signalNo to the thread via pthread_kill.
	 * This can be used to wake up a thread that is blocked (e.g. blocking read)
	 */
	void sendSignal(int p_signalNo);

	/**
	 *	@fn static int sleep(unsigned int timeToSleep)
	 *
	 *	@brief Puts the currently running thread to sleep
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param timeToSleep This parameter represents the time that the thread
	 *		shall sleep in milliseconds. If 0 is provided, the thread will slepp
	 *		forever.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	The sleep functions suspends the currently running thread for the given time
	 *	in milliseconds.
	 *	One should consider to use blocking functions inside the thread
	 *	instead of the sleep() call.
	 */
	static int sleep(unsigned int p_timeToSleep);

	/**
	 * Set name of the thread. If the thread is already running, it is tried to set the new name
	 * to the running thread. If it is not running the name will be used when startThread() is called.
	 * Will only work on Linux.
	 */
	void setThreadName(const std::string & p_name);

private:

	std::unique_ptr<std::thread> myThread{}; //!< unique_ptr to a std::thread object (don't try to change to a non-pointer member => the class needs the thread function in constructor!).
	std::string threadName{};
};
}//end of namespace WE

#endif /* WE_THREAD_H_ */
