/*
 *	@file we_can_freestyle.h
 *
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Header file for WECANFreestyle class (version 2)
 *
 */

#ifndef WE_CAN_FREESTYLE_H_
#define WE_CAN_FREESTYLE_H_

#include "../../common/include/we_can_protocol.h"	// the base class will be required in any case
#include "../../../common/include/we_error.h"



namespace WE
{


class WECANMappingObject;
class WECANTransmitSequence;
class WECANFreestyleResources;
class WECANFreestyleStateMachine;


class WECANFreestyle : public WECANProtocol
{



public:
	// ************************************************************************************************
	// ************************ (de)initialization related ********************************************
	// ************************************************************************************************


	/**
	 *	@fn WECANFreestyle(const std::string & p_instanceName )
	 *
	 *	@brief Default constructor. Initializes pointers to nullptr and other members to sane
	 *		start values. Also calls Constructor of WECANProtocol.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_instanceName name this instance should use as a WEVariableOwner
	 *
	 */
	explicit WECANFreestyle(const std::string & p_instanceName );



	/**
	 *	@fn virtual ~WECANFreestyle()
	 *
	 *	@brief Virtual destructor. Frees all occupied memory.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 */
	virtual ~WECANFreestyle();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WECANFreestyle(const WECANFreestyle &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WECANFreestyle & operator=(const WECANFreestyle &) = delete;

	/**
	 *	@fn virtual int create(WEDEI * const p_dei,
				WEVariableEventDispatcher * const p_varEventDispatcher,
				unsigned int p_portNumber,
				WECANFreestyleStateMachine * p_machine)
	 *
	 *	@brief Initialize the CANFreestyle stack and allocates needed memory
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_dei Pointer to the DEI instance of this process
	 *	@param p_portNumber The CAN port number which CANFreestyle shall use (starting with 1, not 0)
	 *	@param p_machine State machine to be used to run this stack instance.
	 *
	 *	@returns WE_SUCCESS if a valid WETick, DEI pointer and a CAN port number are present
	 *	@returns WE_INIT_ERROR if setting the CAN port number or the DEI pointer failed
	 *
	 *	@see int init(void)
	 *
	 *	You can use this create function if you don't want to set the port number and DEI before.
	 *	This function implicitly calls the init() function without parameter.
	 *
	 */
	virtual int create(WEDEI * const p_dei,
				WEVariableEventDispatcher * const p_varEventDispatcher,
				unsigned int p_portNumber,
				WECANFreestyleStateMachine * p_machine);



	/**
	 *	@fn virtual int init();
	 *
	 *	@brief Initialize all class attributes with meaningfull default values. Does not allocate memory.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int init() override;



	/**
	 *	@fn virtual int reset();
	 *
	 *	@brief Reset all class attributes with default values. Does not free allocated memory.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int reset() override;



	/**
	 *	@fn virtual int createResources(WEDEI * p_dei, WEVariableEventDispatcher * p_varEventDispatcher, unsigned int p_portNo,	WECANFreestyleResources ** p_resources)
	 *
	 *	@brief Set resources for stack (and possibly state machine) in case you don't want to use the init method.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_dei pointer to DEI instance to be used
	 *	@param p_varEventDispatcher pointer to dispatcher providing events from p_dei
	 *	@param p_portNo no of CAN port to be used (starting with 1)
	 *	@param p_resources Returns the pointer of the created resources
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if any of the pointers provided is nullptr or p_portNo is 0
	 *
	 *	This method checks the content and creates a new instance of WECANProtocolResources for the stack to use.
	 *
	 *	@warning The stack will delete both of the instances provided when it is destroyed itself.
	 */
	virtual int createResources(
				WEDEI * p_dei,
				WEVariableEventDispatcher * p_varEventDispatcher,
				unsigned int p_portNo,
				WECANFreestyleResources ** p_resources
			);



	/**
	 *	@fn virtual int release();
	 *
	 *	@brief Frees allocated memory and sets all class attributes to default values.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@return WE_SUCCESS
	*/
	virtual int release() override;

	/**
	 *	@fn int addMappingMessage( WECANMessageObject * p_mapMsg)
	 *
	 *	@brief Method to add a receive or transmit mapping to the stack.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_mapMsg Pointer to the WECANMessageObject to be added.
	 *
	 *	@returns WE_SUCCESS if the received message object could be stored
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_mapMsg or p_mapMsg->getDataMapping()is nullptr
	 *
	 *	This method adds the supplied transmit or receive mapping message to the stack. It does NOT create a copy
	 *	of the object but stores the object supplied.
	 *
	 *	The object supplied is expected to NOT be modified from outside the stack any more.
	 *
	 *	When the stack is destroyed or whenever releaseMappingMessages() is called the object supplied will be
	 *	erased,
	 *
	 *
	 *	@see releaseMappingMessages()
	 *
	 */
	int addMappingMessage( WECANMessageObject * p_mapMsg);

	/**
	 *	@fn int addMappingDataToID( unsigned int p_ID, WECANMappingData * p_data, bool p_isTriggerOnly = false );
	 *
	 *	@brief Add (a copy of) p_data to this stack's mapping message identified by p_ID.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_ID	ID of the mapping message object which should get the new mapping data.
	 *	@param p_data	Pointer to mapping data which will be copied into the mapping object of p_ID
	 *	@param p_isTriggerOnly Don't map into CAN message but consider as trigger to send the mapping
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INIT_ERROR if p_data is nullptr
	 *	@returns WE_NOT_FOUND_ERROR if a mapping message object with ID p_ID could not be found
	 *	@returns WE_INVALID_PARAMETER_ERROR if the mapping message object with ID p_ID did not contain a mapping object
	 *
	 *	This method adds the mapping data described by p_data to the mapping object attached to
	 *	the mapping message identified by p_ID.
	 *	To do so it creates a copy of p_data.
	 *
	 *	The mapping message has to be added to the stack before by using addMappingMessage().
	 *
	 *	@see addMappingMessage(WECANMessageObject * p_object)
	 *
	 *	This method could also be part of the common stack (?!) Although it would have to be overwritten to maintain protocol specific data.
	 *
	 */
	int addMappingDataToID( unsigned int p_ID, WECANMappingData * p_data, bool p_isTriggerOnly = false );

	/**
	 *	@fn int addTransmitSequence(WECANTransmitSequence * p_sequence)
	 *
	 *	@brief Add a new CANFreestyle transmit sequence to this stack
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_sequence Pointer to the data of a WECANFreestyleTransmitSequence that shall be added (can be deleted afterwards)
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	@see WECANFreestyleTransmitSequence
	 *
	 *	When adding a transmit sequence, a new class will be instantiated and the memory will be
	 *	managed by this stack. The instance p_sequence can be deleted after calling this function.
	 */
	int addTransmitSequence(WECANTransmitSequence * p_sequence);

	/**
	 *	@fn int createMappingMessage( WECANMessageObject * p_mapMsg)
	 *
	 *	@brief Method to add a receive or transmit mapping to the stack.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_mapMsg Pointer to the WECANMessageObject to be added.
	 *
	 *	@returns WE_SUCCESS if the received message object could be stored
	 *
	 *	This method adds the supplied transmit or receive mapping message to the stack. It creates a copy
	 *	of the object. The object supplied has to be deleted by the caller.
	 *
	 *	When the stack is destroyed or whenever release() is called the object supplied will be
	 *	erased,
	 *
	 *
	 *	@see addMappingMessage()
	 *
	 */
	int createMappingMessage( WECANMessageObject * p_mapMsg);

	/**
	 *	@brief Method to add a receive or transmit mapping to the stack.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_objectID,
	 *	@param p_canID,
	 *	@param p_canIDMask,
	 *	@param p_timeoutPeriod,
	 *	@param p_timeoutListener,
	 *	@param p_mapping,
	 *	@param p_requestData);
	 *
	 *	@returns WE_SUCCESS if the received message object could be stored
	 *
	 *	This method adds the supplied transmit or receive mapping message to the stack. It creates a new object
	 *	and also copies the supplied p_mapping and p_requestData.
	 *
	 *
	 *	When the stack is destroyed or whenever release() is called the object created here will be
	 *	erased,
	 *
	 *
	 *	@see releaseMappingMessages()
	 *
	 */
	int createMappingMessage(
			unsigned int p_objectID,
			unsigned int p_canID,
			unsigned int p_canIDMask,
			unsigned int p_timeoutPeriod,
			WESignalListener * p_timeoutListener,
			WECANMappingObject * p_mapping,
			WECANRequestData * p_requestData);

	/**
	 *	@fn int createTransmitSequence(int p_ID, unsigned int p_transmitPeriod, std::vector< unsigned int > * p_msgIDs)
	 *
	 *	@brief Method to add a receive or transmit mapping to the stack.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_ID ID of this sequence
	 *	@param p_transmitPeriod gap in ms between messages of the sequence
	 *	@param p_msgIDs vector of IDs of WECANMessageObjects to be sent as the sequence
	 *
	 *	@returns WE_SUCCESS if sequence could be added to the stack
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_transmitPeriod is 0 or p_msgIDs is nullptr
	 *	@returns WE_ if p_ID already exists as a sequence
	 *
	 *
	 *
	 *	@see addMappingMessage()
	 *	@see createMappingMessage()
	 *
	 */
	int createTransmitSequence(int p_ID, unsigned int p_transmitPeriod, std::vector< unsigned int > * p_msgIDs);


	/**
	 *	@fn WECANMessageObject* getMessageObjectByObjectID(int p_id);
	 *
	 *	@brief Return a WECANMessageObject identified by p_ID of the map ID2MappingMsg
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_id ID of the WECANMessageObject
	 *
	 *	@returns WECANMessageObject 	if the p_id was found in the map
	 *		 nullptr 			if p_id was not found
	 *
	 */
	WECANMessageObject* getMessageObjectByObjectID(int p_id);



	// ************************************************************************************************
	// ************************ periodic handling related *********************************************
	// ************************************************************************************************

	/**
	 *	@fn int sendSequences( unsigned int p_tick, unsigned int p_timeoutTick )
	 *
	 *	@brief Send out transmit mappings with repeat rate.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_time - current time tick in ms
	 *	@param p_timeoutTick	NOT USED
	 *
	 */
	int sendPeriodicTransmitMappings( unsigned int p_tick, unsigned int p_timeoutTick );


	/**
	 *	@fn int sendPeriodicRequests(unsigned int p_time)
	 *
	 *	@brief Send out requests with repeat rate.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_time - current time tick in ms
	 *	@param p_timeoutTick	NOT USED
	 *
	 *	@return WE_SUCCESS in case of success
	 *		WE_GENERAL_ERROR if no we found no request data
	 *
	 */
	int sendPeriodicRequests(unsigned int p_time);


	/**
	 *	@fn int sendSequences( unsigned int p_tick, unsigned int p_timeoutTick )
	 *
	 *	@brief	Checks if a transmit sequence is attached in the transmitsequence map. If
	 *		there are sequences then check if it is time to send a transmit mapping.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_tick		current time
	 *	@param p_timeoutTick	NOT USED
	 *
	 *	@returns WE_SUCCESS 	in any case
	 *
	 *	Checks if a transmit sequence is attached in the transmitsequence map. If
	 *	there are sequences then check if it is time to send a transmit mapping and
	 *	choose the mapping which we have to send next.
	 *
	 *	If the we have to send a message we create a sendMessage and fill it with
	 *	the data of the transmit mapping. After that the message is send with the
	 *	method sendMessage();
	 *
	 */
	int sendSequences( unsigned int p_tick, unsigned int p_timeoutTick );



	/**
	 *	@fn int handleReceiveTimeouts( unsigned int p_tick, unsigned int p_timeoutTick )
	 *
	 *	@brief Check receive messages timestamps and trigger handlers if they timed out.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_timeNow	Current time to use for timeout calculation
	 *	@parame p_timeoutTick	NOT USED
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	This method checks all the receive mappings with a timeout period defined if they have timed out.
	 *	For all those that have time out the mapped variables are set to "invalid" in DEI and handlers
	 *	for the event are triggered.
	 *
	 */
	int handleReceiveTimeouts( unsigned int p_tick, unsigned int p_timeoutTick );



	// ************************************************************************************************
	// ************************ variable owner related ************************************************
	// ************************************************************************************************

	/**
	 *	@fn int updateRequest(unsigned int p_varKey, int p_intValue)
	 *
	 *	@brief Implementation of WEVariableOwner's updateRequest function
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_varKey Index and subindex of the variable for which an update (write) request shall be sent.
	 *	@param p_intValue New integer value that shall be written to the variable.
	 *
	 *	@returns WE_SUCCESS if the CAN write request was sent out successfully
	 *	@returns WE_NOT_FOUND_ERROR if no variable was found that matches p_index and p_subindex
	 *	@returns return value of WE::WECANRaw::send(WECANMessage * messageToSend)
	 *
	 *	@see WECANRaw
	 *
	 *	The updateRequest function gets called by WEDEI when an application wants to change the value
	 *	of an external variable.
	 *	With this method the DEI asks the owner of the variable (here the CAN stack) to change the
	 *	variable value. Since the CAN stack can typically not decide whether the new value is reasonable
	 *	or not it will send a "write request" to the ECU which actually maintains the variable.
	 *	A "write request" is any CAN mapping that contains the respective variable and has the
	 *	"WE_CAN_TRANSMIT_TYPE_ON_WRITE_REQUEST" as well as the "WE_CAN_TRANSMIT_TYPE_ANY_VALUE_CHANGED" or
	 *	"WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_CHANGED" bits set ("WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_CHANGED" only works
	 *	if the "certain value" indicates the variable that should be modified).
	 *	The actual value of the variable will only be changed in DEI if a read response for that variable
	 *	arrives on the bus.
	 */
	virtual int updateRequest(unsigned int p_varKey, int p_intValue) override;



	/**
	 *	@fn int updateRequest(unsigned int p_varKey, const std::string & p_stringValue)
	 *
	 *	@brief Implementation of WEVariableOwner's updateRequest function
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_index Index of the variable for which an update (write) request shall be sent.
	 *	@param p_subindex Subindex of the variable for which an update (write) request shall be sent.
	 *	@param p_stringValue New string value that shall be written to the variable.
	 *
	 *	@returns WE_NOT_SUPPORTED
	 *
	 *	@see WECANRaw
	 *
	 *	The updateRequest function gets called by WEDEI when an application wants to change the value
	 *	of an external variable.
	 *	With this method the DEI asks the owner of the variable (here the CAN stack) to change the
	 *	variable value. Since the CAN stack can typically not decide whether the new value is reasonable
	 *	or not it will send a "write request" to the ECU which actually maintains the variable.
	 *	A "write request" is any CAN mapping that contains the respective variable and has the
	 *	"WE_CAN_TRANSMIT_TYPE_ON_WRITE_REQUEST" as well as the "WE_CAN_TRANSMIT_TYPE_ANY_VALUE_CHANGED" or
	 *	"WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_CHANGED" bits set ("WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_CHANGED" only works
	 *	if the "certain value" indicates the variable that should be modified).
	 *	The actual value of the variable will only be changed in DEI if a read response for that variable
	 *	arrives on the bus.
	 *	Please note that the actual stack implementation using this method may not support sending out string values.
	 */
	virtual int updateRequest(unsigned int p_varKey, const std::string & p_stringValue) override;

	/**
	 *	@fn virtual int getValueRequest(unsigned int p_varKey)
	 *
	 *	@brief 	UNUSED by CANFreestyle
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_varkey
	 *
	 *	@returns WE::WE_NOT_SUPPORTED_ERROR
	 *
	 *	This Method will only be used by CANopen. So don't care for CANFreestyle.
	 *
	 *	@warning: Will not be used by CANFreestyle
	 *
	 */
	virtual int getValueRequest(unsigned int /*p_varKey*/) override
	{
		return WE::WE_NOT_SUPPORTED_ERROR;
	}

	// ************************************************************************************************
	// ************************ WESignalListener related ***********************************************
	// ************************************************************************************************

	/**
	 *	@brief Handle variable change events
	 *
	 *	@param p_signal as provided by the event system
	 *
	 *	@returns WE_SUCCESS if the event was handled successfully
	 *	@returns WE error code
	 *
	 *	From there other methods of the protocol could be called to e.g. send out a message with the updated
	 *	value.
	 */
	virtual int signalOccurred(WESignal & p_signal) override;



	// ************************************************************************************************
	// ************************ CAN port related ******************************************************
	// ************************************************************************************************



	/**
	 *	@fn int handleReceiveMapping(WECANMessage * p_msg, unsigned int p_timeNow)
	 *
	 *	@brief  Check message against mappings and transfer values to DEI if one or more mapping(s) fit(s).
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_msg 		Message to be checked.
	 *	@param p_timeNow	Time to use for timestamp in case message fits a mapping
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INIT_ERROR if p_data is nullptr
	 *	@returns WE_NOT_FOUND_ERROR if a mapping message object could not be found for the message provided
	 *
	 *	This method checks if the message provided fits any mapping stored and if it does it forwards the
	 *	value using WECANMapping::extractMappedValues().
	 *	In this case the timestamp of the respective mapping message object is set to p_timeNow.
	 *
	 *	The mapping message has to be added to the stack before by calling addMappingMessage().
	 *
	 *	@see WECANMapping::extractMappedValues()
	 */
	int handleReceiveMapping(WECANMessage * p_msg, unsigned int p_timeNow);


	/**
	 *	@fn int handleRequestsReceived(WECANMessage * p_msg, unsigned int p_timeNow)
	 *
	 *	@brief Check if we have a transmit mapping to fit the request and send it out.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_msg		Message to be checked if it contains a request.
	 *	@param p_timeNow	Time to use for timestamp in case request fits a mapping
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_NOT_FOUND_ERROR if a mapping message object could not be found for the message provided
	 *
	 *	This method checks if the message provided is a request for one of our
	 *	transmit mappings. If one of the transmit mapping messages fits it is sent out.
	 *
	 *	The mapping message has to be added to the stack before by calling addMappingMessage().
	 *
	 */
	int handleRequestsReceived(WECANMessage * p_msg, unsigned int p_timeNow);

	virtual int enableMessage(unsigned int p_objectID, bool p_enable) override;



private:
	// Create data structures as suggested in WECANProtocol:

	WECANFreestyleResources * resourcesFreestyle{nullptr};	//!< pointer to data that will not change while stack is running and can be shared with state machine.

	// one map keeping all mapping messages accessible by their unique Object IDs
	std::unordered_map<unsigned int, WECANMessageObject *>		ID2MappingMsg{};

	// (1) (masked) CAN ID -> receive message description (one list/map per ID mask; to forward new values)
	// When populating this map we could also set ID filters for the CAN socket - assuming the socket is exclusively used for this stack instance.
	// Using std::unordered_multimap to support multiple Mappings with same CAN ID and same MASK ID
	std::unordered_multimap <unsigned int, WECANMessageObject *>	maskedCANID2ReceiveMappingMsg{};

	// (2) Variable key -> transmit message description (to satisfy "updateRequest" and variable change event)
	std::unordered_multimap <unsigned int, WECANMessageObject *> 	varKey2TransmitMappingMsg{};

	// (3) list of receive message descriptions with timeout (to set variables to "invalid") if message expires
	std::set < WECANMessageObject * >				receiveMappingMsgsWithTimeout{};

	// (4) list of transmit messages with repeat time (to send them out in time)
	std::set < WECANMessageObject * >				transmitMappingsWithPeriod{};

	// (4a) list of transmit messages with proprietary requests (to check them quickly)
	std::set < WECANMessageObject * >				transmitMappingsWithRequests{};

	// (4b) list of transmit messages with repeat time (to send them out in time)
	std::set < WECANMessageObject * >				receiveMappingsWithPeriodicRequest{};

	// (5) list of transmit sequences (to send out their messages in time)
	std::unordered_map<unsigned int, WECANTransmitSequence * >		ID2transmitSequence{};

	std::unordered_multimap <unsigned int, WECANMessageObject * >	varKey2ReceiveMappingMsg{};

	int errorObjectID{-1};	// Replace 'static' variable in 'triggerOccurred()'
	int errorReturnValue{WE_GENERAL_ERROR};	// Replace 'static' variable in 'triggerOccurred()'

	int errorObjectID2{-1};	// Replace 'static' variable in 'triggerOccurred()'
	int errorReturnValue2{WE_GENERAL_ERROR};	// Replace 'static' variable in 'triggerOccurred()'

	static const int debugSection;
};
}	// namespace

#endif /* WE_CAN_FREESTYLE_H_ */
