/*
 *
 *	@file we_can_freestyle_state_machine.h
 *
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Header file for WECANFreestyleStateMachine class (version 2)
 *
 */

#ifndef WE_CAN_FREESTYLE_STATE_MACHINE_H_
#define WE_CAN_FREESTYLE_STATE_MACHINE_H_



#include "../../common/include/we_can_protocol_state_machine.h"		// base class will be needed in any case



namespace WE
{

class WECANFreestyleResources;


class WECANFreestyleStateMachine : public WECANProtocolStateMachine
{

public:


	WECANFreestyleStateMachine();
	virtual ~WECANFreestyleStateMachine();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WECANFreestyleStateMachine(const WECANFreestyleStateMachine &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WECANFreestyleStateMachine & operator=(const WECANFreestyleStateMachine &) = delete;

	int init() override;
	int reset() override;


	// ************************************************************************************************
	// ************************ (de)initialization related ********************************************
	// ************************************************************************************************


	/**
	 *	@fn virtual int create(WECANProtocolResources * p_res)
	 *
	 *	@brief Create everything that requires memory allocation for this class.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_res
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	This methods creates instances of everything the state machine needs to work.
	 *
	 *	@see release()
	 *
	 */
	virtual int create(WECANProtocolResources * p_res) override;



	/**
	 *	@fn virtual int reset(void)
	 *
	 *	@brief Delete everything this instance of state machine allocated before.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	This method deletes all instances of classes created by this state machine instance
	 *	- especially those in method create().
	 *
	 *
	 *	@see create()
	 *
	 */
	virtual int release(void) override;




	// ************************************************************************************************
	// ************************ state machine related *************************************************
	// ************************************************************************************************


	/**
	 *	@fn virtual int run(unsigned int p_time)
	 *
	 *	@brief Switch state machine until p_time is reached or current work is done.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_time WETick where this method should be finished or interrupt itself.
	 *
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_TIMEOUT_ERROR in case execution was terminated prematurely because p_time expired
	 *
	 *	This method executes the periodically driven state machine of the CAN Freestyle stack.
	 *	It will be called frequently by the stack's "process()" method in order to perform all the stack activities
	 *	NOT driven by (external) events.
	 *	This includes (but is not limited to) e.g.
	 *	- intialization: populate lists
	 *	- normal operation: sending messages, checking timeouts,
	 *	- termination: send log off, cleaning up memory
	 *
	 *	This method should NOT be used for activities which could directly be driven by external events
	 *	such as
	 *	- receiving a CAN message
	 *	- receiving an event
	 *
	 *	When implementing this method one should be aware that it blocks the event handling methods.
	 *	So similar to them it should do things quickly and not perform any waiting by itself.
	 *	If longer lasting tasks are implemented they should be interrupted when p_time is reached.
	 *	This requires a design keeping the current state when time is up and continuing when this method
	 *	is called next time.
	 *
	 *	For longer lasting tasks it could be considered starting threads to execute them although this
	 *	requires careful synchronization of shared resources.
	 *
	 *	When this method is not running it is expected that the process/thread handling the stack is either
	 *	waiting for events/endpoints or executing handlers for these.
	 *
	 *
	 */
	virtual int run(unsigned int p_time) override;



	/**
	 *	@fn int handleInitState(unsigned int p_time)
	 *
	 *	@brief Switch state machine until p_time is reached or current work is done.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_time WETick where this method should be finished or interrupt itself.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_TIMEOUT_ERROR in case execution was terminated prematurely because p_time expired
	 *
	 *	This method will be called by "run()" in order to cover everything that has to be done in "init" state.
	 *	This includes
	 *
	 *	When implementing this method one should be aware that it blocks the event handling methods.
	 *	So similar to them it should do things quickly and not perform any waiting by itself.
	 *	If longer lasting tasks are implemented they should be interrupted when p_time is reached.
	 *	This requires a design keeping the current state when time is up and continuing when this method
	 *	is called next time.
	 *
	 *	For longer lasting tasks it could be considered starting threads to execute them although this
	 *	requires careful synchronization of shared resources.
	 *
	 *	When this method is not running it is expected that the process/thread handling the stack is either
	 *	waiting for events/endpoints or executing handlers for these.
	 *
	 *
	 */
	int handleInitState(unsigned int p_time);



	/**
	 *	@fn int handleOperationalState(unsigned int p_time)
	 *
	 *	@brief Switch state machine until p_time is reached or current work is done.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_time WETick where this method should be finished or interrupt itself.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_TIMEOUT_ERROR in case execution was terminated prematurely because p_time expired
	 *
	 *	This method will be called by "run()" in order to cover everything that has to be done in "init" state.
	 *	This includes
	 *
	 *	When implementing this method one should be aware that it blocks the event handling methods.
	 *	So similar to them it should do things quickly and not perform any waiting by itself.
	 *	If longer lasting tasks are implemented they should be interrupted when p_time is reached.
	 *	This requires a design keeping the current state when time is up and continuing when this method
	 *	is called next time.
	 *
	 *	For longer lasting tasks it could be considered starting threads to execute them although this
	 *	requires careful synchronization of shared resources.
	 *
	 *	When this method is not running it is expected that the process/thread handling the stack is either
	 *	waiting for events/endpoints or executing handlers for these.
	 *
	 *
	 */
	int handleOperationalState(unsigned int p_time);

	// ************************************************************************************************
	// ************************ event related *********************************************************
	// ************************************************************************************************


	/**
	 *	@fn virtual int handleCANmsg(WECANMessage * p_msg)
	 *
	 *	@brief Method to handle and distribute all incoming CAN messages.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg Pointer to the message received which should be evaluated/handled here.
	 *
	 *	@returns WE_SUCCESS if the received message could be handled (or the state machine is paused)
	 *	@returns WE_NOT_SUPPORTED_ERROR if the received message was not intended for this CAN protocol instance
	 *
	 *	This function get's called whenever a CAN message is read from the CAN bus by WECANProtocol::onEndPointReadable().
	 *
	 *
	 *	@warning "p_msg" is only valid in this method - if (parts of) the message should be stored one has to copy them/it
	 *
	 *	@see WECANProtocol::onEndPointReadable()
	 *
	 */
	virtual int handleCANmsg(WECANMessage * p_msg) override;

private:
	WECANFreestyleResources * resourcesFreestyle{nullptr};
};
}	// namespace

#endif /* WE_CAN_FREESTYLE_STATE_MACHINE_H_ */
