/**
 *	@file we_can_open_message_handler.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Interface class for CANopen (CAN) message handlers
 *
 */


#ifndef WE_CAN_OPEN_MESSAGE_HANDLER_H_
#define WE_CAN_OPEN_MESSAGE_HANDLER_H_

#include "../../../hardware/can_raw/include/we_can_message.h"

namespace WE
{

//forward declarations
class WECANopenObjectDictionary;
class WECANRaw;
class WEEmergencyHandler;
class WETick;
class WECANopen;

/**
 *	@class WECANopenMessageHandler
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Base class for specialized (CAN) message handler classes for CANopen
 *
 *	@namespace WE
 *
 *	@see WEPDOHandler
 *	@see WESDOHandler
 *	@see WESyncHandler
 *	@see WENMTHandler
 *	@see WEEmergencyHandler
 *
 *	This base class provides some common members and getter/setter functions for them.
 *	Every inheriting class needs to implement the handleMessage function.
 */
class WECANopenMessageHandler
{
public:
	/**
	 *	@fn WECANopenMessageHandler()
	 *
	 *	@brief Default constructor
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	This default constructor sets the members od, canPort and emcy to nullptr but creates an
	 *	instance of WECANMessage (msg).
	 *	Implementing classes should call this constructor inside their own.
	 */
	WECANopenMessageHandler() = default;

	/**
	 *	@fn virtual ~WECANopenMessageHandler()
	 *
	 *	@brief Virtual destructor, deletes msg
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WECANopenMessageHandler() = default;

	/**
	 *	@fn void setObjectDictionary(WECANopenObjectDictionary * p_od)
	 *
	 *	@brief Provide a pointer to the CANopen Object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_od Pointer to the CANopen Object Dictionary
	 */
	void setObjectDictionary(WECANopenObjectDictionary * p_od);

	/**
	 *	@fn void setEmergency(WEEmergencyHandler * p_emcy)
	 *
	 *	@brief Provide a pointer for the CANopen emergency message handler
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_emcy Pointer to an existing WEEmergencyHandler instance
	 */
	void setEmergency(WEEmergencyHandler * p_emcy);

	/**
	 *	@fn virtual int handleMessage(WECANMessage * p_msg) = 0
	 *
	 *	@brief Abstract method to be implemented by every inheriting class
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_msg Pointer to a WECANMessage with data that shall be handled
	 *
	 *	@returns WE_SUCCESS if message was handled successfully
	 *	@returns Implementation specific error code
	 *
	 *	@see WEPDOHandler
	 *	@see WESDOHandler
	 *	@see WESyncHandler
	 *	@see WENMTHandler
	 *	@see WEEmergencyHandler
	 */
	virtual int handleMessage(WECANMessage * p_msg) = 0;

	void setCANopenInstance(WECANopen * p_co)
	{
		this->co = p_co;
	}

protected:
	WECANMessage msg{};	//!< The CAN message that can be used to construct CAN messages for sending
	WECANopenObjectDictionary * od{nullptr};	//!< Pointer to the CANopen Object Dictionary
	WEEmergencyHandler * emcy{nullptr};	//!< Pointer to the CANopen Emergency message handler
	WECANopen * co{nullptr};	//!< Pointer to CANopen protocol instance that this handler belongs to

	int sendCANMessage(void);
private:
	/**
	 * Private copy constructor! Don't copy!
	 */
	WECANopenMessageHandler(const WECANopenMessageHandler & p_handler) = delete;
	WECANopenMessageHandler & operator=(const WECANopenMessageHandler &p_handler) = delete;
};
}  // namespace WE

#endif /* WE_CAN_OPEN_MESSAGE_HANDLER_H_ */
