/**
 *	@file we_can_open_state_machine.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Header file for WECANopenStateMachine class
 *
 */

#ifndef WE_CAN_OPEN_STATE_MACHINE_H_
#define WE_CAN_OPEN_STATE_MACHINE_H_

#include "../../../hardware/can_raw/include/we_can_message.h"
#include "we_can_open_definitions.h"

namespace WE
{

//forward declarations
class WECANopen;
class WECANopenObjectDictionary;
class WECANRaw;
class WETick;
class WEEmergencyHandler;
class WEDEI;
class WESyncHandler;
class WENMTHandler;
class WESDOHandler;
class WEPDOHandler;
class WESignalListener;

/**
 *	@class WECANopenStateMachine
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Class for a CANopen State machine
 *
 *	@namespace WE
 *
 *	This class represents a CANopen state machine.
 *	Before starting the machine, be sure to set all the pointers (setMessageHandlers, setObjectDictionary,
 *	setDEI, setCANinterface. Don't use this class from outside. It is intended to be used by CANopen class only.
 */
class WECANopenStateMachine
{
public:
	/**
	 *	@fn WECANopenStateMachine(WECANopen * p_instance)
	 *
	 *	@brief Constructor
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_instance Pointer to the CANopen instance this state machine belongs to
	 */
	explicit WECANopenStateMachine(WECANopen * p_instance);

	WECANopenStateMachine() = delete;

	/**
	 *	@fn ~WECANopenStateMachine()
	 *
	 *	@brief Destructor, frees memory of this->recMsg, all other references need to be freed by other classes (=> WECANopen)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WECANopenStateMachine();

	/**
	 *	@fn int run(unsigned int p_time)
	 *
	 *	@brief The run function needs to be called frequently so that all tasks can be executed
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_time The tick count until which this function is allowed to run
	 *
	 *	@returns If everything went OK, WE_SUCCESS is returned. Everything else means error but is hard to define at the moment
	 *
	 *	This functions needs to be called frequently so that all SENDING tasks of the CANopen stack
	 *	can be triggered at the right time.
	 *	The receiving part is done event based by the EndPoint system.
	 */
	int run(unsigned int p_time);

	/**
	 *	@fn void setDEI(WEDEI * p_dei)
	 *
	 *	@brief Set pointer this processes WEDEI instance
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_dei Pointer to WEDEI instance
	 */
	void setDEI(WEDEI * p_dei);

	/**
	 *	@fn void setMessageHandlers(WESDOHandler * p_sdo, WEPDOHandler * p_pdo, WESyncHandler * p_sync, WENMTHandler * p_nmt, WEEmergencyHandler * p_emcy)
	 *
	 *	@brief Set all WECANOpenMessageHandlers that are needed by the state machine
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_sdo Pointer to this CANopen instance's SDO handler instance
	 *	@param p_pdo Pointer to this CANopen instance's PDO handler instance
	 *	@param p_sync Pointer to this CANopen instance's Sync handler instance
	 *	@param p_nmt Pointer to this CANopen instance's NMT handler instance
	 *	@param p_emcy Pointer to this CANopen instance's Emergency handler instance
	 *
	 *	@see WESDOHandler, WEPDOHandler, WESyncHandler, WENMTHandler, WEEmergencyHandler
	 *
	 *	@warning Don't set nullptr :/
	 */
	void setMessageHandlers(WESDOHandler * p_sdo,
			WEPDOHandler * p_pdo,
			WESyncHandler * p_sync,
			WENMTHandler * p_nmt,
			WEEmergencyHandler * p_emcy);

	/**
	 *	@fn void setObjectDictionary(WECANopenObjectDictionary * p_od)
	 *
	 *	@brief Set pointer to the CANopen object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_od Pointer to the CANopen object dictionary associated with this CANopen instance
	 */
	void setObjectDictionary(WECANopenObjectDictionary * p_od);

	/**
	 *	@fn void pauseMachine(bool p_pause)
	 *
	 *	@brief Pause or un-pause state machine.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_pause true for pause, false for continue
	 *
	 *	If p_puase == true, all incoming CAN messages are ignored and no messages will be sent out.
	 */
	void pauseMachine(bool p_pause);

	bool isPaused(void) const;

	/**
	 *	@fn int receiveCANMessages(unsigned int p_time)
	 *
	 *	@brief Function to handle and distribute all incoming CAN messages
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_time The tick count until this function is allowed to run
	 *
	 *	@returns WE_SUCCESS if the received message could be handled (or the state machine is paused)
	 *	@returns WE_NOT_SUPPORTED_ERROR if the received message was not intended for this CANopen instance
	 *	@returns WE_INTERFACE_ERROR if either the canPort member or one of the message handlers are nullptr
	 *	@returns WE_READ_ERROR if an error occurred while reading from canPort (can happen and may be normal because no select is used)
	 *
	 *	@see run()
	 *
	 *	This function get's called whenever a CAN message can be read from the CAN bus. It passes the received
	 *	CAN message to every Message handlers that then checks if it is intended for the specific handler.
	 *	If one handler says that it does not support the message it is passed to the next.
	 *	Sequence is NMT->SYNC->PDO->SDO->EMCY.
	 */
	int receiveCANMessages(unsigned int p_time);

	/**
	 *	@fn int getCurrentState(void)
	 *
	 *	@brief Return the current state of the state machine
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns the current state of the state machine
	 *
	 *	@see enum CANopenNMTStates for available states
	 */
	int getCurrentState(void) const
	{
		return this->currentState;
	}

	/**
	 *	@fn int initHandlers(void)
	 *
	 *	@brief This function is called in initialization state to initialize the single message handlers
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS if successful
	 *	@returns WE_INTERFACE_ERROR if one of the handlers is nullptr
	 */
	int initHandlers(void);

	void addStateListener(WESignalListener * p_listener);

private:
	/**
	 * Private copy constructor! Don't copy!
	 */
	WECANopenStateMachine(const WECANopenStateMachine & p_machine) = delete;
	WECANopenStateMachine & operator=(const WECANopenStateMachine &p_machine) = delete;

	friend class WENMTHandler; //only NMT handler may set the state
	friend class WECANopen;

	/**
	 *	@fn void setState(int p_state)
	 *
	 *	@brief Set the current state of the state machine
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_state State to set (see CANopenNMTStates)
	 *
	 *	This function should only be called by WENMTHandler. To pause the machine or use
	 *	the auto start feature use the appropriate methods.
	 */
	void setState(int p_state);

	/**
	 *	@fn int handleInitState(unsigned int p_time)
	 *
	 *	@brief Initialize anything needed in initializiation phase of the state machine
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_time Tick count until this function may run
	 *
	 *	@returns Currently the returnValue of initHandlers() function is returned
	 *
	 *	@see initHandlers()
	 */
	int handleInitState(unsigned int p_time);

	/**
	 *	@fn int sendCANMessages(unsigned int p_time)
	 *
	 *	@brief Sends all cyclic CANMessages
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_time Tick count until this function is allowed to run
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This function is called in the states Pre-Operational, Operational and Stopped to send
	 *	messages on the bus that are not triggered by external events (e.g. heartbeat, cyclic PDOs) and it
	 *	also checks if timeouts are expired (e.g. SDO Download responses)
	 */
	int sendCANMessages(unsigned int p_time);

	WESyncHandler * sync{nullptr};	//!< Pointer to Sync Handler
	WENMTHandler * nmt{nullptr};	//!< Pointer to NMT Handler
	WESDOHandler * sdo{nullptr};	//!< Pointer to SDO Handler
	WEPDOHandler * pdo{nullptr};	//!< Pointer to PDO Handler
	WEEmergencyHandler * emcy{nullptr};	//!< Pointer to Emergency Handler
	WECANopenObjectDictionary * od{nullptr};	//!< Pointer to Object Dictionary
	WECANopen * canOpenInstance{nullptr};	//!< Pointer to the CANopen instance of this state machine

	WEDEI * dei{nullptr};		//!< Pointer to the DEI instance
	WECANMessage recMsg{};	//!< An instance of a CAN message in which data of received CAN messages is stored

	WESignalListener * stateListener{nullptr};

	unsigned int timeLeftToNextSend{5};
	int currentState{CANOPEN_STATE_INITIALISATION};	//!< The current state of the state machine
	int resetCommand{CANOPEN_STATE_RESET_APPLICATION}; 	//!< Reset Node or Reset Communication
	bool paused{false};		//!< flag that indicates if state machine is currently paused
};
} //end of namespace WE

#endif /* WE_CAN_OPEN_STATE_MACHINE_H_ */
