/**
 *	@file we_nmt_handler.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Class definition of WENMTHandler
 */

#ifndef WE_NMT_HANDLER_H_
#define WE_NMT_HANDLER_H_

#include <map>
#include <set>

#include "we_can_open_message_handler.h"

namespace WE
{

class WECANRaw;
struct WECANMessage;
class WECANopenStateMachine;
class WECANopenObjectDictionary;
class WESignalListener;

/**
 *	@class WENMTHandler
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Handle CANopen NMT specific CAN messages and process other NMT data
 *
 *	@namespace WE
 *
 *	@see WECANopenMessageHandler
 *
 *	This class is used to determine and handle incoming CANopen NMT messages.
 *	It is also responsible for handling heartbeat consumer/producer behaviour
 */
class WENMTHandler : public WECANopenMessageHandler
{
public:
	/**
	 *	@fn WENMTHandler()
	 *
	 *	@brief Initialize member with sane values and call constructor of WECANopenMessageHandler
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WENMTHandler();

	/**
	 *	@fn virtual ~WENMTHandler()
	 *
	 *	@brief Destructor. Frees memory (no explicit freeing necessary)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WENMTHandler();

	/**
	 *	@fn void setNodeID(unsigned int p_nodeID)
	 *
	 *	@brief Set the CANopen Node ID (needed for determining if a NMT message is addressed at this stack)
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_nodeID the node ID of this CANopen stack
	 */
	void setNodeID(unsigned int p_nodeID);

	/**
	 *	@fn void setStateMachine(WECANopenStateMachine * p_machine)
	 *
	 *	@brief Set the pointr to the CANopen state machine
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_machine Pointer to the CANopen state machine
	 */
	void setStateMachine(WECANopenStateMachine * p_machine);

	/**
	 *	@fn void setHeartbeatProducerTime(WECANopenObjectDictionary * p_od)
	 *
	 *	@brief Get a pointer to the heartbeat producer time from object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_od Pointer to the object dictionary
	 */
	void setHeartbeatProducerTime(WECANopenObjectDictionary * p_od);

	/**
	 *	@fn void sendHeartbeatMessage(int p_state)
	 *
	 *	@brief Send the heartbeat message on the CAN bus
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_state Current state of the CANopen state machine
	 *
	 *	This function gets frequently called by the CANopen state machine so that the
	 *	heartbeat message can be sent.
	 *	The NMT handler sends the heartbeat message if heartbeat producer time is > 0.
	 */
	void sendHeartbeatMessage(int p_state);

	/**
	 *	@fn int handleMessage(WECANMessage * p_msg)
	 *
	 *	@brief Handle a received CAN message
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_msg The received CAN message that shall be handled
	 *
	 *	@returns WE_SUCCESS if the message could be handled and appropriate action was executed (like
	 *		state change, etc.). Note that NMT messages that are not broadcast and not addressed to this
	 *		node get ignored and the return value will also be WE_SUCCESS
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_msg is nullptr
	 *	@returns WE_NOT_SUPPORTED_ERROR if the message could not be handled (and should be checked by another message handler)
	 *
	 *	This functions checks if a CAN message is a valid NMT message and shall be processed by this
	 *	NMT message handler. Appropriate actions get executed if the NMT messages is addressed to this stack
	 *	(like telling the state machine to change the state)
	 */
	int handleMessage(WECANMessage * p_msg) override;

	void setListenToNodeIDOnly(bool p_onlyNodeID)
	{
		this->listenToNodeIDOnly = p_onlyNodeID;
	}

	void updateHeartbeatConsumerSettings(void);
	void checkHeartbeatTimeouts(void);
	void addHeartbeatListener(WESignalListener * p_listener);
	void addHeartbeatListner(WESignalListener * p_listener, int p_nodeID);

private:
	int handleNMTCommand(WECANMessage * p_msg);
	void handleNMTHeartbeat(WECANMessage * p_msg);
	WECANopenStateMachine * stateMachine;	//!< Pointer to the CANopen state machine
	unsigned int nodeID;			//!< The node ID of this CANopen stack
	bool listenToNodeIDOnly;		//!< If set to true, NMT broadcasts are ignored
	unsigned int lastHeartbeatTick;	//!< The tick count of when the last heartbeat tick was received

	const int * heartbeatProducerTime;	//!< Pointer to the value of the heartbeat producer time (0x1017/0x00)
	std::map<unsigned char, unsigned int> heartbeatConsumerMapping;
	std::map<unsigned char, unsigned int> heartbeatConsumerTimeouts;
	std::map<unsigned char, WE::WESignalListener*> hertbeatListenersByNodeID;
	std::map<unsigned char, bool> timedOutECUs;
	std::set<unsigned int> heartbeatConsumerCOBIDs;
	WESignalListener * heartbeatConsumerListener;
};
}

#endif /* WE_NMT_HANDLER_H_ */
