/**
 *	@file we_pdo.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of a CANopen PDO
 *
 */


#ifndef WE_PDO_H_
#define WE_PDO_H_

#include <set>
#include "../../common/include/we_can_mapping_object.h"

namespace WE
{

//forward declarations
class WECANMappingObject;
class WEDEI;
class WEVariableOwner;
class WECANopenObjectDictionary;
struct WECANMessage;

/**
 *	@class WEPDO
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief WE interpretation of a CANopen PDO
 *
 *	@namespace WE
 *
 *	@see WECANMappingObject
 *
 *	The WEPDO is a specialized version of the WECANMappingObject. It is specialized in that way that it does
 *	overwrite the handleExtractedValue functions so that that received PDOs can pass the extracted values
 *	to the object dictionary and not to the DEI directly.
 *	The fillMessage function also uses values from Object Dictionary to fill a Tx CAN Message.
 *	Furthermore the WEPDO has some additional members to store some specific information.
 */
class WEPDO : public WECANMappingObject
{
public:
	/**
	 *	@fn WEPDO(unsigned int p_messageLength, WETransmitType p_transmitType, unsigned short p_inhibitTime, unsigned int p_transmitPeriod, WECANopenObjectDictionary * p_od, WECANopenPDOTransmissionMode p_txMode)
	 *
	 *	@brief Constructor. See detailed description and parameter description.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_messageLength The length of the PDO in bytes
	 *	@param p_transmitType The transmit type of the PDO
	 *	@param p_inhibitTime The inhibit time of the PDO
	 *	@param p_tranmsitPeriod The transmit period
	 *	@param p_od A pointer to the CANopen object dictionary
	 *	@param p_txMode the CANopen PDO transmission mode
	 *
	 *	This constructor implicitly calls the constructor of WECANMappingObject with the following parameters:
	 *	p_messageLength = p_messageLength, p_transmitType = p_trnsmitType, p_transmitTriggerKey = 0,
	 *	p_transmitPeriod = p_transmitPeriod, p_sequenceID = 0 and p_littleEndian = true.
	 */
	WEPDO(unsigned int p_messageLength,
			WETransmitType p_transmitType,
			unsigned short p_inhibitTime,
			unsigned int p_transmitPeriod,
			WECANopenObjectDictionary * p_od,
			WECANopenPDOTransmissionMode p_txMode);

	/**
	 *	@fn virtual ~WEPDO()
	 *
	 *	@brief Free memory, only set pointer to object dictionary to nullptr.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WEPDO();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEPDO(const WEPDO &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEPDO & operator=(const WEPDO &) = delete;

	/**
	 *	@fn int handleExtractedValue(unsigned int p_key, int p_value, WEDEI * p_dei, WEVariableOwner * p_owner, int p_valueValid)
	 *
	 *	@brief Process values extracted from a CAN message matching this PDO
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_key Index and subindex of the object dictionary sub entry that's value has been extracted from the CAN message
	 *	@param p_value The extracted integer value
	 *	@param p_dei In this case nullptr because we don't use DEI directly but the object dictionary
	 *	@param p_owner Pointer to the owner of the variable (ignored)
	 *
	 *	@returns WE_INTERFACE_ERROR the internal pointer to the object dictionary is nullptr (od)
	 *	@returns WE_SUCCESS if p_index is <= 0x1000 (that means dummy mapping)
	 *	@returns the return value of WECANopenObjectDictionary::setValue()
	 *
	 *	This overwrites the default behavior of the WECANMappingObject. It does not pass the extracted value
	 *	to DEI but to the CANopen object dictionary. The object dictionary then takes care of handling the value correctly.
	 */
	int handleExtractedValue(unsigned int p_key, int p_value, WEDEI * p_dei, WEVariableOwner * p_owner, int p_valueValid) const override;

	/**
	 *	@fn int handleExtractedValue(unsigned int p_key, std::string & p_value, WEDEI * p_dei, WEVariableOwner * p_owner, int p_valueValid)
	 *
	 *	@brief Process values extracted from a CAN message matching this PDO
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_key Index and subindex of the object dictionary sub entry that's value has been extracted from the CAN message
	 *	@param p_value The extracted string value
	 *	@param p_dei In this case nullptr because we don't use DEI directly but the object dictionary
	 *	@param p_owner Pointer to the owner of the variable (ignored)
	 *
	 *	@returns WE_INTERFACE_ERROR the internal pointer to the object dictionary is nullptr (od)
	 *	@returns WE_SUCCESS if p_index is <= 0x1000 (that means dummy mapping)
	 *	@returns the return value of WECANopenObjectDictionary::setValue()
	 *
	 *	This overwrites the default behavior of the WECANMappingObject. It does not pass the extracted value
	 *	to DEI but to the CANopen object dictionary. The object dictionary then takes care of handling the value correctly.
	 */
	int handleExtractedValue(unsigned int p_key, const std::string & p_value, WEDEI * p_dei, WEVariableOwner * p_owner, int p_valueValid) const override;

	/**
	 *	@fn std::set<unsigned int> getAllMappedVariableKeys(void)
	 *
	 *	@brief Retrieve the keys of all mapped variables in this PDO
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns A set containing all variable keys that are mapped to this PDO
	 */
	std::set<unsigned int> getAllMappedVariableKeys(void) const;

	/**
	 *	@fn int fillMessage(WECANMessage * p_msg)
	 *
	 *	@brief Fill a CAN message with current values of mapped variables / object dictionary subentries
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_msg pointer to a CAN message that shall be filled with values
	 *
	 *	@returns WE_INVALID_PARAMETER_ERRORif p_msg is nullptr
	 *	@returns WE_SUCCESS if p_msg was successfully filled with current values
	 */
	int fillMessage(WECANMessage * p_msg);

	inline unsigned short getODIndex(void) const
	{
		return this->odIndex;
	}

	inline void setVariablesInvalid(bool p_invalid)
	{
		this->variablesInvalid = p_invalid;
	}

	inline bool areVariablesInvalid(void) const
	{
		return this->variablesInvalid;
	}
private:
	friend class WEPDOHandler;	//!< The PDO handler may access private members
	friend class WESyncHandler;	//!< The Sync handler may access private members

	WECANopenObjectDictionary * od;	//!< Pointer to the object dictionary
	WECANopenPDOTransmissionMode txMode;	//!< The PDO transmission mode
	unsigned short odIndex;	//!< The object dictionary index of this PDO
	unsigned char numberOfBytes;	//!< The number of bytes used by this PDO
	unsigned int cobID;	//!< The COB-ID of this PDO (as from object dictionary)
	unsigned short inhibitTime;	//!< The inhibit time of the PDO (as from object dictionary)
	unsigned int tickOfLastSend;	//The tick of the last sending of this PDO. To be checked if inhibit time is != 0
	int action;
	unsigned int actionVariable;
	bool variablesInvalid;
};
}  // namespace WE

#endif /* WE_PDO_H_ */
