/**
 *	@file we_pdo_handler.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Class definition of WEPDOHandler
 *
 */

#ifndef WE_PDO_HANDLER_H_
#define WE_PDO_HANDLER_H_

#include <map>
#include <set>
#include <mutex>

#include "we_can_open_definitions.h"
#include "we_can_open_message_handler.h"

namespace WE
{

//forward declarations
class WECANRaw;
class WECANopenObjectDictionary;
class WEPDO;
class WESyncHandler;
class WECANopenStateMachine;
class WETick;
class WECANopen;
class WEDEI;

struct PDOActionStruct
{
	int number{0};
	WECANopenPDOType type{PDO_TX};
	int action{PDO_NO_ACTION};
	unsigned int key{0};
	PDOActionStruct() = default;
};

/**
 *	@class WEPDOHandler
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief WEPDOHandler handles receiving and sending of PDO messages
 *
 *	@namespace WE
 *
 *	@see WEPDO
 *
 *	@warning RTR PDOs ARE NOT SUPPORTED!
 *
 *	This class provides everything necessary to process Rx- and send TxPDOs.
 */
class WEPDOHandler : public WECANopenMessageHandler
{
public:
	/**
	 *	@fn WEPDOHandler()
	 *
	 *	@brief Constructor. Init with defaults.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	explicit WEPDOHandler(WECANopen * p_canopenInstance);

	/**
	 *	@fn virtual ~WEPDOHandler()
	 *
	 *	@brief Destructor. Clears all stored PDOs and container.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WEPDOHandler();

	/**
	 *	@fn int init()
	 *
	 *	@brief Clear and delete existing PDOs and create new ones from Object Dictionary entries.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_INTERFACE_ERROR if no object dictionary exists
	 *	@returns WE_SUCCESS if initialization was successful
	 *
	 *	This function clears all existing PDO information and reads the PDO data from object dictionary
	 *	to create new PDOs.
	 */
	int init();

	/**
	 *	@fn int handleMessage(WECANMessage * p_msg)
	 *
	 *	@brief Handle a CAN message and process data if it is a configured PDO.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_msg Pointer to a CAN message that may contain PDO data
	 *
	 *	@returns WE_NOT_SUPPORTED_ERROR if p_msg is either nullptr OR the current CANopen state is not OPERATIONAL
	 *		OR there is no state machine available OR the message does not contain PDO data for this stack.
	 *	@returns EMCY_PDO_LENGTH_EXCEEDED if p_msg is longer than the the configured PDO length
	 *	@returns WE_INTERFACE_ERROR if the configured PDO is a synchronous one and no sync handler is available
	 *	@returns WE_SUCCESS if the p_msg contains correct PDO data for this stack and could be processed
	 *
	 *	This function is called by the CANopen state machine whenever a CAN message is recieived and the
	 *	message was not handled by a higher priority message handler.
	 *	It checks the message if it fits any of the configured RxPDO information from object dictionary.
	 *	If the message is a correct PDO message, the data is extracted and the mapped values will be extracted
	 *	and passed to the object dictionary (with respect of the configuration and transmission mode of the PDO)
	 */
	int handleMessage(WECANMessage * p_msg) override;

	/**
	 *	@fn void signalPDOChange(unsigned short p_index)
	 *
	 *	@brief Signal the PDO handler that PDO data in object dictionary changed
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Object dictionary index that contains the data of the changed PDO
	 *
	 *	If the object dictionary gets changed at runtime, this function is called to signal the
	 *	PDO handler that data has changed and that it needs to update it's objects.
	 */
	void signalPDOChange(unsigned short p_index);

	/**
	 *	@fn void setSync(WESyncHandler * p_sync)
	 *
	 *	@brief Set a pointer to the sync handler
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_sync Pointer to this stack's sync handler
	 */
	void setSync(WESyncHandler * p_sync)
	{
		this->sync = p_sync;
	}

	/**
	 *	@fn void setStateMachine(WECANopenStateMachine * p_sm)
	 *
	 *	@brief Set pointer to the CANopen state machine
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_sm Pointer to the CANopen state machine
	 */
	void setStateMachine(WECANopenStateMachine * p_sm)
	{
		this->sm = p_sm;
	}

	/**
	 *	@fn int handleVariableChange(unsigned int p_variableKey, int p_value)
	 *
	 *	@brief Check if a txPDO needs to be sent because of variable value change
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_variableKey Combination of index and subindex of the changed variable
	 *	@param p_value The changed value
	 *
	 *	@returns WE_INTERFACE_ERROR if no state machine is available OR the state is not OPERATIONAL
	 *		OR a synchronous PDO could not be handed over to the sync handler (because it is nullptr)
	 *	@returns WE_SUCCESS if the variable is mapped to a TxPDO and the PDO was sent (or queued)
	 *	@returns WE_NOT_FOUND_ERROR if the variable is NOT mapped to a PDO
	 *
	 *	This function gets called whenever the CANopen stack gets informed about a variable change.
	 *	It will check if the variable is mapped to a TxPDO and then sends the PDO with the new data
	 *	(or queues it for sending)
	 */
	int handleVariableChange(unsigned int p_variableKey, int p_value);

	/**
	 *	@fn int sendPDOs(unsigned int p_timeout, unsigned int *p_timeleft = nullptr)
	 *
	 *	@brief Send cyclic asynchronous PDOs or PDOs with inhibit time
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_timeout Tick count until this function is allowed to run
	 *
	 *	@returns WE_INTERFACE_ERROR if the state machine can't be accessed or the current state is not OPERATIONAL
	 *	@returns WE_TIMEOUT_ERROR if timeout expired
	 *	@returns WE_SUCCESS if all queued PDOs where sent out
	 *
	 *	This function gets frequently called by the CANopen state machine to send out queued
	 *	cyclic asynchronous PDOs or PDOs that where queued because of a set inhibit time.
	 */
	int sendPDOs(unsigned int p_timeout, unsigned int *p_timeleft = nullptr);

	int setPDOAction(int p_PDONumber, WECANopenPDOType p_type, int p_action, unsigned int p_key);

	void checkRxTimeouts(void);

	void addInvalidPDOIndex(unsigned short p_index);

private:
	friend class WESyncHandler;	//!< The sync handler may access some functions

	std::map<unsigned int, WEPDO*> rxIDtoMappingObject;	//!< Map containing all RxPDOs ordered by their configured COB-ID
	std::set<WEPDO*> txPDOs;	//!< Set with all TxPDOs
	std::set<WEPDO*> queuedTXPDOs;	//!< All TxPDOs that are queued for sending
	std::multimap<unsigned int, WEPDO*> varsMappedToTxPDOs;	//!< Multimap: Variable keys mapped to TxPDOs
	std::multimap<unsigned int, WEPDO*> timeToTxPDOMapping;	//!< Multimap with a tick count mapped to a TxPDO (the PDOs shall be sent at that tick count)
	std::set<WEPDO*> rxPDOsWithTimeout;	//!< All RxPDOs that have an event time set
	std::map<unsigned short, PDOActionStruct> pdoActions;
	std::set<unsigned short> invalidPDOs;	//!< If an invalid PDO mapping was detected, the OD index of this PDO will be added here. PDOs in this set might never be enabled!

	WESyncHandler * sync;	//!< Pointer to the sync handler
	WECANopenStateMachine * sm;	//!< Pointer to the object dictionary
	WECANopen * canopenInstance;
	WEDEI * dei;
	std::mutex sendMutex;	//!< sendPDODirectly may be called from different threads => need a mutex here for locking to one thread at a time

	/**
	 *	@fn WEPDO * getPDODataFromObjectDictionary(unsigned short p_index)
	 *
	 *	@brief Create a PDO with data from object dictionary
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Object Dictionary index that contains data of the PDO to create
	 *
	 *	@returns A PDO configured with object dictionary data found at p_index or nullptr if p_index is not present
	 *		or if the PDO was marked as invalid due to an invalid mapping configuration
	 *
	 *	This function is used to create a PDO based on data from the object dictionary.
	 */
	WEPDO * getPDODataFromObjectDictionary(unsigned short p_index);

	/**
	 *	@fn int sendPDODirectly(WEPDO * p_pdo)
	 *
	 *	@brief Try to send a PDO out to the CAN bus.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_pdo Pointer to the PDO that shall be sent
	 *
	 *	@returns WE_SUCCESS if the PDO was sent successfully
	 *	@returns WE_INTERFACE_ERROR If the state machine is not present or the state is not OPERATIONAL
	 *	@returns WE_TIMEOUT_ERROR If the inhibit time of the PDO prevents direct sending (caller has to queue it into queuedTXPDOs if this is returned)
	 *	@returns WE_SUCCESS if sending the PDO directly was successful
	 *
	 *	This function is used to send PDOs directly on the CAN bus. If the inhibit time prevents direct
	 *	sending of the PDO, WE_TIMEOUT_ERROR is returned; caller has to insert the PDO into queuedTXPDOs so
	 *	the sendPDOs function can send it later.
	 */
	int sendPDODirectly(WEPDO * p_pdo);

	/**
	 *	@fn unsigned char getNumberOfUsedBytes(WEPDO * p_pdo)
	 *
	 *	@brief Determine how many bytes are used for the mapped variables of a PDO.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_pdo Pointer to a PDO that's length shall be checked
	 *
	 *	@returns Number of bytes that a PDO occupies in a CAN message
	 *
	 *	This method is used to check if an incoming PDO is too long.
	 */
	unsigned char getNumberOfUsedBytes(WEPDO * p_pdo);
};
}

#endif /* WE_PDO_HANDLER_H_ */
