/**
 *	@file we_can_message_object.h
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Common implementation to be used by all or most of the CAN protocols.
 *
 *	The content of this file is collected to cover a common interface and implementation
 *	of features likely to be needed by all CAN protocols.
 *
 *	This file provides a generic description of a CAN message which e.g. would be extended to a PGN definition for J1939.
 *	This is NOT implementing the raw CAN message to be sent on the bus (see WECANMessage for that) but
 *	providing some common interfaces for a CAN stack to maintain such messages.
 *
 *	The "common" part of CAN messages is considered the resulting CAN ID and the data mapping.
 *	For the implementation of an actual protocol this is expected to be extended.
 *
 *	@see WECANMessage
 *	@see WECANMappingObject
 *
*/

#ifndef WE_CAN_MESSAGE_OBJECT_H_
#define WE_CAN_MESSAGE_OBJECT_H_

#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include "we_can_mapping_object.h"	// WETransmitType

namespace WE
{

class WECANRaw;
struct WECANMessage;
class WECANMappingObject;	// don't know why the #include above does not work
class WETick;

/**
 *	@class WECANRequestData
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Struct that stores information of how a request looks like for a WECANMappingObject
 *
 *	@warning Here we assume that no request needs more than one CAN message.
 *
 *
 *	@namespace WE
 */
class WECANRequestData
{
public:
	/**
	 *	@fn WECANRequestData()
	 *
	 *	@brief Standard constructor that initializes all member variables.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WECANRequestData()
	{
		this->CANID		= 0;
		this->IDMask		= 0;
		this->longID		= false;
		this->messageLength	= 8;
		this->dataBytes[0]	= 0;
		this->dataBytes[1]	= 0;
		this->dataBytes[2]	= 0;
		this->dataBytes[3]	= 0;
		this->dataBytes[4]	= 0;
		this->dataBytes[5]	= 0;
		this->dataBytes[6]	= 0;
		this->dataBytes[7]	= 0;

		this->dataMasks[0]	= 0;
		this->dataMasks[1]	= 0;
		this->dataMasks[2]	= 0;
		this->dataMasks[3]	= 0;
		this->dataMasks[4]	= 0;
		this->dataMasks[5]	= 0;
		this->dataMasks[6]	= 0;
		this->dataMasks[7]	= 0;
	}

	virtual ~WECANRequestData()
	{
		// this is only here to be virtual :-)
	}

	void setIDType(int p_idType)
	{
		this->longID = p_idType == 0 ? false : true;
		if (this->longID)
		{
			this->CANID = this->CANID | 0x80000000;
		}
	}

	void setCANID(unsigned int p_CANID)
	{
		if (longID)
		{
			this->CANID = p_CANID | 0x80000000;
		}
		else
		{
			this->CANID = p_CANID;
		}
	}

	unsigned int	CANID;		//!< The CAN ID
	unsigned int	IDMask;		//!< Corresponding CAN ID Mask
	bool		longID; 	//!< Long ID	True = 29 bit	false = 11 bit
	unsigned char	messageLength;	//!< The message length
	unsigned char	dataBytes[8];	//!< Array containing the data bytes
	unsigned char	dataMasks[8];	//!< Array containing the mask bytes that will be applied to the data bytes

	// SERIALIZATION
	friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & p_ar, const unsigned int )
	{
		// no parent class here: ar & boost::serialization::base_object< WECANRequestData > (*this);
		p_ar & CANID;		//!< The CAN ID
		p_ar & IDMask;		//!< Corresponding CAN ID Mask
		p_ar & messageLength;	//!< The message length
		p_ar & dataBytes;		//!< Array containing the data bytes
		p_ar & dataMasks;		//!< Array containing the mask bytes that will be applied to the data bytes
		// cppcheck-suppress bitwiseOnBoolean
		p_ar & longID; 		//!< Long ID	True = 29 bit	false = 11 bit
	}
};


class WECANMessageObject
{
public:
	/**
	 *	@fn WECANMessageObject(void)
	 *
	 *	@brief Default constructor initializing members with default values.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	Member 'timeStamp' is initialized with WETick::getTickCount() to allow one full timeout period
	 *	before first timeout calculation may fail.
	 */
	WECANMessageObject();


	/**
	 *	@fn WECANMessageObject(unsigned int p_objectID)
	 *
	 *	@brief Constructor initializing members with default values - except object ID.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_objectID - ID of this object (NOT the CAN ID of this object)
	 *
	 *	Member 'timeStamp' is initialized with WETick::getTickCount() to allow one full timeout period
	 *	before first timeout calculation may fail.
	 */
	explicit WECANMessageObject(unsigned int p_objectID);


	/**
	 *	@fn WECANMessageObject(const WECANMessageObject & p_object)
	 *
	 *	@brief Copy constructor.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_object - object to be copied
	 *
	 */
	WECANMessageObject(const WECANMessageObject & p_object);
	WECANMessageObject & operator=(const WECANMessageObject & p_object);

	/**
	 *	@brief Constructor initializing members with values provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_objectID - ID of this object (NOT the CAN ID of this object)
	 *	@param p_canID - CAN ID to be used for this message
	 *	@param p_canIDMask - ID mask to be applied (only relevant for receive messages)
	 *	@param p_timeoutPeriod - period to be applied for timeout of receive message or as sending period of transmit message
	 *	@param p_timeoutListener - this will get events if receive message times out or becomes valid again
	 *	@param p_mapping	- mapping describing the data part of this message (will be copied)
	 *	@param p_requestData	- request to be sent out to get receive message or to be answered with transmit message (will be copied)
	 *
	 *	Member 'timeStamp' is initialized with WETick::getTickCount() to allow one full timeout period
	 *	before first timeout calculation may fail.
	 */
	WECANMessageObject(
			unsigned int p_objectID,
			unsigned int p_canID,
			unsigned int p_canIDMask,
			unsigned int p_timeoutPeriod,
			WESignalListener * p_timeoutListener,
			WECANMappingObject * p_mapping,
			WECANRequestData * p_requestData
	);

	/**
	 *	@fn virtual ~WECANMessageObject()
	 *
	 *	@brief Default destructor destroying members and freeing allocated memory.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 */
	virtual ~WECANMessageObject();


	/**
	 *	@fn WECANRequestData * getRequest(void)
	 *
	 *	@brief Get pointer to request attached to this mapping.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns A pointer to the request of this mapping
	 *	@returns nullptr if there is no request defined
	 *
	 *	@warning This does not copy the request.
	 */
	WECANRequestData * getRequest(void) const;


	/**
	 *	@fn int setRequest(WECANRequestData * p_request)
	 *
	 *	@brief Use request attached for this object.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_request - pointer to request to be used in this message.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	@warning This will in any case delete the existing request of this message.
	 *	@warning The request provided will be deleted when this message gets destroyed.
	 */
	int setRequest(WECANRequestData * p_request);


	/**
	 *	@fn int createRequest(WECANRequestData * p_request)
	 *
	 *	@brief Copy request provided and use it for this object.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_request - pointer to request to be copied into this message.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	@warning This will in any case delete the existing request of this message.
	 */
	int createRequest(WECANRequestData * p_request);


	/**
	 *	@fn unsigned int getobjectID( void )
	 *
	 *	@brief Get object ID as provided to constructor.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns The object ID as provided to constructor
	 *
	 */
	unsigned int getobjectID( void ) const;


	/**
	 *	@fn unsigned int getEffectiveCANID( void )
	 *
	 *	@brief Get CAN ID resulting from CAN protocol definitions.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns The CAN ID as it should appear on CAN bus with mask applied.
	 *
	 */
	unsigned int getEffectiveCANID( void ) const;


	/**
	 *	@fn unsigned int getCANID( void )
	 *
	 *	@brief Get CAN ID resulting from CAN protocol definitions.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns The CAN ID as it should appear on CAN bus
	 *
	 */
	unsigned int getCANID( void ) const;


	/**
	 *	@fn unsigned int getCANIDMask( void )
	 *
	 *	@brief Get CAN ID mask to be applied.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns The CAN ID mask to apply in order to recognize CAN ID on the bus.
	 *
	 */
	unsigned int getCANIDMask( void ) const;


	/**
	 *	@fn WECANMappingObject * getDataMapping( void )
	 *
	 *	@brief Get data mapping object attached to this message.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns pointer to this CAN message's data mapping object.
	 *
	 *	@see WECANMappingObject
	 *
	 *	@warning The object returns will be deleted when this message is destroyed.
	 */
	WECANMappingObject * getDataMapping( void ) const;


	/**
	 *	@fn bool isReceiveMessage( void )
	 *
	 *	@brief Return true if the mapping attached identifies this as a receive message.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns true if thie message is intended to be received.
	 *	@returns false otherwise
	 *
	 *	@see WECANMappingObject
	 *
	 */
	virtual bool isReceiveMessage( void ) const;


	/**
	 *	@fn unsigned int getTransmitType( void )
	 *
	 *	@brief Get transmit type as stored in the data mapping object belonging to this message.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns Transmit type as stored in the data mapping object belonging to this message.
	 *	@returns WE_CAN_TRANSMIT_TYPE_UNDEFINED if this should not have a data mapping.
	 *
	 *	@see WECANMappingObject
	 *
	 */
	unsigned int getTransmitType( void ) const;


	/**
	 *	@fn int setTransmitType(WETransmitType p_transmitType );
	 *
	 *	@brief Set transmit type for data mapping object.
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_transmitType transmit type to set.
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	@see WETransmitType in WECANMappingObject
	 *
	 */
	int setTransmitType(WETransmitType p_transmitType);


	/**
	 *	@fn bool matchesMsg(WECANMessage * p_msg, unsigned int p_timeNow = 0)
	 *
	 *	@brief Check if this message description fits the message provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg - message to be compared with this object
	 *	@param p_timeNow - value for timestamp in case return value is "true"
	 *
	 *	@returns true if this is a receive object and the CAN message provided fits to it
	 *	@returns false if this belongs to a transmit object
	 *	@returns false if definition doesn't fit
	 *	@returns false if this has no mapping attached
	 *
	 *	Time stamp of this object is set to "p_timeNow" if this method returns true.
	 *	If "0 == p_timeNow" timestamp is set to WETick::getInstance()->getTickCount() instead.
	 *
	 *	@see WECANMappingObject
	 *	@see WETick
	 */
	virtual bool matchesMsg(WECANMessage * p_msg, unsigned int p_timeNow = 0);


	/**
	 *	@fn bool matchesMsg(unsigned int p_ID, unsigned char * p_data, unsigned int p_size, unsigned int p_timeNow = 0)
	 *
	 *	@brief Check if this message description fits the (extended) message provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_ID - CAN ID of message to be compared with this object
	 *	@param p_data - data buffer of message to be compared with this object
	 *	@param p_size - size of p_data in bytes
	 *
	 *	@returns true if this is a receive object and the CAN message provided fits to it
	 *	@returns false if this belongs to a transmit object
	 *	@returns false if definition doesn't fit
	 *	@returns false if this has no mapping attached
	 *
	 *	This could be used e.g. to check if this object matches a sequence of more than one CAN message
	 *	- if the protocol implementation supports something like transport protocol.
	 *
	 *	Time stamp of this object is set to "p_timeNow" if this method returns true.
	 *	If "0 == p_timeNow" timestamp is set to WETick::getInstance()->getTickCount() instead.
	 *
	 *
	 *	@see WECANMappingObject
	 */
	virtual bool matchesMsg(unsigned int p_ID, unsigned char * p_data, unsigned int p_size, unsigned int p_timeNow = 0);


	/**
	 *	@fn bool matchesRequest(WECANMessage * p_msg)
	 *
	 *	@brief Check if the request of this message description fits the message provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg - message to be compared with this object
	 *
	 *	@returns true if this is a transmit object and the CAN message provided fits to it
	 *	@returns false if this belongs to a receive object or definition doesn't fit.
	 *
	 *	This method also requires the size to match.
	 *
	 *	@see fitsConstRequestData()
	 */
	virtual bool matchesRequest(WECANMessage * p_msg) const;


	/**
	 *	@fn bool matchesRequest(unsigned int p_ID, unsigned char * p_data, unsigned int p_size)
	 *
	 *	@brief Check if the request of this message description fits the message provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_ID - CAN ID of message to be compared with this object
	 *	@param p_data - data buffer of message to be compared with this object
	 *	@param p_size - size of p_data in bytes
	 *
	 *	@returns true if this is a transmit object and the CAN message provided fits to it
	 *	@returns false if this belongs to a receive object or definition doesn't fit.
	 *
	 *	This does NOT consider p_size as a condition to match but only as a buffer size.
	 *
	 *	@see WECANMappingObject
	 */
	virtual bool matchesRequest(unsigned int p_ID, unsigned char * p_data, unsigned int p_size) const;


	/**
	 *	@fn bool fitsConstRequestData(unsigned char * p_data, unsigned int p_size)
	 *
	 *	@brief Check if the data provided fits this mapping's request.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_data CAN data to be checked
	 *	@param p_size length of p_data in bytes
	 *
	 *	@returns true if this mapping fits p_data and p_length
	 *	@returns false otherwise
	 *
	 *	This method checks the const data of the data provided and succeeds when the the shorter
	 *	one of both ends and patterns matched.
	 *	It is assumed that the CAN ID was checked before.
	 *
	 *	@warning At the moment checking requests only works for single messages because WECANRequestData is defined that way.
	 *
	 *	@see fitsRequest(WECANMessage * p_msg)
	 *
	 */
	virtual bool fitsConstRequestData(unsigned char * p_data, unsigned int p_size) const;


	/**
	 *	@fn int fillMsgWithRequest(WECANMessage * p_msg, unsigned int p_timeNow)
	 *
	 *	@brief Fill p_msg with data defined in this->requestData
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg - message which should get the ID
	 *	@param p_timeNow - Time to apply to this->requestTimeStamp if filling p_msg succeeds
	 *
	 *	@returns WE_SUCCESS if message could be written to send buffer
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_msg is nullptr
	 *	@returns WE_INVALID_PARAMETER_ERROR if this->requestData is nullptr
	 *
	 *	Just set the CAN ID, DLC and data of this->requestData in p_msg.
	 *	In the base implementation this method does not consider if this instance is a send or receive message.
	 *	It also does not consider any mask in any way (since those are intended for receiving anyways).
	 *
	 *	@see createRequest()
	 */
	virtual int fillMsgWithRequest(WECANMessage * p_msg, unsigned int p_timeNow);


	/**
	 *	@fn int fillID( WECANMessage * p_msg )
	 *
	 *	@brief Add ID of this object to p_msg.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg - message which should get the ID
	 *
	 *	@returns WE_SUCCESS if message could be written to send buffer
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_msg is nullptr
	 *
	 *	Just set the CAN ID of this object in p_msg.
	 *	In the base implementation this method does not consider if this instance is a send or receive message.
	 *	It also does not consider the ID mask in any way (since it's intended for receiving anyways).
	 *
	 *	@see WECANRaw
	 */
	virtual int fillID( WECANMessage * p_msg ) const;


	/**
	 *	@fn virtual int fillSendMessage(
			WECANMessage * p_msg,
			WEDEI * p_dei,
			unsigned int p_timeNow = 0,
			unsigned int p_varKey = 0,
			int p_varValue = 0,
			bool p_setTimeStamp = true)
	 *
	 *	@brief If this is a transmit message populate p_msg with its data.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg - CAN message to be filled with ID and data
	 *	@param p_dei - DEI which should provide variable values
	 *	@param p_timeNow - Tick in milliseconds to be set to the timestamp of this object
	 *	@param p_varKey - key of variable which should NOT be sent with DEI value but with p_varValue.
	 *	@param p_varValue - value to be used for variable identified by "p_varKey"
	 *	@param p_setTimeStamp - true if timestamp should be set in case of success - false otherwise
	 *
	 *	@returns WE_SUCCESS if message could be filled
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_msg == nullptr
	 *	@returns WE_READ_ERROR if this is a read and not a write message
	 *	@returns any value WECANMappingObject::fillMessageFromDEI() returns
	 *
	 *	If this is a send message is determined by the mapping object attached.
	 *	If message could be filled successfully with DEI data only, timestamp of this object is set to p_timeNow.
	 *	If p_varKey = 0 AND p_varValue = 0 timestamp is set (otherwise this is considered a write request)
	 *	If p_timeNow == 0 and timestamp should be set timestamp is set to WETick::getInstance()->getTickCount();
	 *	p_setTimeStamp can be used if message is filled to answer a request but periodic sending should continue
	 *	ignoring this.
	 *
	 *	@see WECANMappingObject
	 *	@see WECANMappingObject::fillMessageFromDEI()
	 *	@see WETick
	 */
	virtual int fillSendMessage(
			WECANMessage * p_msg,
			WEDEI * p_dei,
			unsigned int p_timeNow = 0,
			unsigned int p_varKey = 0,
			int p_varValue = 0,
			bool p_setTimeStamp = true);


	/**
	 *	@fn unsigned int getTimeoutPeriod( void )
	 *
	 *	@brief Get timeout period set for this object instance.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns this->timeoutPeriod
	 *
	 *	Timeout period only applies to receive messages.
	 *	Transmit type should have WE_CAN_TRANSMIT_TYPE_RECEIVE_PERIODIC bit set.
	 *
	 *	@see WETick
	 *	@see WETransmitType
	 *	@see WECANMappingObject
	 */
	unsigned int getTimeoutPeriod( void ) const
	{
		return this->timeoutPeriod;
	}


	/**
	 *	@fn bool hasTimedOut ( unsigned int p_timeNow = 0, long * p_timeLeft = nullptr )
	 *
	 *	@brief Check if this object has timed out.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_timeNow - timestamp in milliseconds for now
	 *	@param p_timeLeft - if != nullptr this is filled with the time until timeout (negative value if true is returned)
	 *
	 *	@returns true if this has timed out already
	 *	@returns false otherwise
	 *	@returns true if timeout period is set to 0; p_timeLeft is set to p_timeNow in this case
	 *
	 *	This method checks if timeout period has elapsed i.e. if message was sent/received within this period.
	 *	If p_timeNow == 0 current time is taken from WETick::instance()->getTickCount();
	 *
	 *	@see WETick
	 */
	bool hasTimedOut ( unsigned int p_timeNow = 0, long * p_timeLeft = nullptr ) const;


	/**
	 *	@fn int handleTimedOut( bool p_timedOut, WEVariableOwner * p_owner, WEDEI * p_dei )
	 *
	 *	@brief Send out events if the state of p_timedOut has changed.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_timedOut - true if this has timed out, false otherwise
	 *	@param p_owner - owner of variables who should be set to "invalid" if p_timedOut is true
	 *	@param p_dei - dei instance containing p_owner's variables
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns any error code signalOccurred() may return from one of the listeners
	 *
	 *
	 *	This method sends out all events required to timeout listeners for the message and
	 *	the attached mappings' variables.
	 *	Parameter p_timedOut could be the result of hasTimedOut().
	 *	p_owner and p_dei can be set to "nullptr" if no variables should be modified.
	 *
	 *	@warning This will only inform about timeout if the state p_timedOut has changed compared to earlier calls.
	 *	@warning I.e. variable will not be set invalid again if set valid elsewhere after it was set invalid here.
	 *
	 *	@see WECANMessageObject::hasTimedOut()
	 *	@see WECANMessageObject::sendTimeoutEvent()
	 */
	int handleTimedOut( bool p_timedOut, WEVariableOwner * p_owner = nullptr, WEDEI * p_dei = nullptr );


	/**
	 *	@fn int sendTimeoutEvent(bool p_engaged)
	 *
	 *	@brief This function is called whenever a receive timeout occurs. The event will be forwarded to timoutListener of this message.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_engaged true: timeout occurred, false: timeout resolved
	 *
	 *	@returns The return value of WESignalListner::signalOccurred()
	 *
	 *	@see WESignalListener
	 *
	 *	When a receive timeout happens, timeoutListener of this message will receive a TIMEOUT_EVENT event.
	 *	If the appropriate message is received again, another TIMEOUT_EVENT is sent to timeoutListener.
	 *	Both events differ in parameter "ep3.isValid" which is set to p_engaged.
	 *	The event parameter 1 is set to "0" to avoid mixup with a variable key.
	 *	The event parameter 2 contains a pointer to "this" as the origin of the event.
	 *
	 *	For send mappings this method does nothing.
	 *
	 *	@see WECANMessageObject::handleTimedOut()
	 *	@see WECANMappingObject::sendTimeoutEvent()
	 */
	int sendTimeoutEvent(bool p_engaged);

	int sendReceivedEvent(void);


	/**
	 *	@fn bool isDueToBeSent ( unsigned int p_timeNow, long * p_timeLeft )
	 *
	 *	@brief Check if this object has to be sent or how much time is left.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_timeNow - timestamp in milliseconds for now
	 *	@param p_timeLeft - if != nullptr this is filled with the time until sending time (negative value if true is returned)
	 *
	 *	@returns true if this has to be sent out already
	 *	@returns false otherwise
	 *	@returns true if timeout period is set to 0; p_timeLeft is set to p_timeNow in this case
	 *
	 *	This method checks if transmit period (in data mapping) has elapsed i.e. if message has to be sent.
	 *	If p_timeNow == 0 current time is taken from WETick::instance()->getTickCount();
	 *
	 *	@see WETick
	 */
	bool isDueToBeSent ( unsigned int p_timeNow = 0, long * p_timeLeft = nullptr ) const;


	/**
	 *	@fn bool isDueToBeRequested ( unsigned int p_timeNow, long * p_timeLeft )
	 *
	 *	@brief Check if this object has to be requested or how much time is left.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_timeNow - timestamp in milliseconds for now
	 *	@param p_timeLeft - if != nullptr this is filled with the time until sending time (negative value if true is returned)
	 *
	 *	@returns true if this has to be requested already
	 *	@returns false otherwise
	 *	@returns true if timeout period is set to 0; p_timeLeft is set to p_timeNow in this case
	 *
	 *	This method checks if transmit period (in data mapping; interpreted as request period here) has elapsed i.e.
	 *	if message has to be requested.
	 *	If p_timeNow == 0 current time is taken from WETick::instance()->getTickCount();
	 *
	 *	@see WETick
	 */
	bool isDueToBeRequested ( unsigned int p_timeNow = 0, long * p_timeLeft = nullptr ) const;


	/**
	 *	@fn void setCANID( unsigned int p_CANID )
	 *
	 *	@brief Set CAN ID resulting from CAN protocol definitions.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_CANID - CAN ID as it will/should appear on the bus
	 *
	 *	The effective CAN ID for this object will be (p_CAN_ID & this->effectiveCANIDMask).
	 *	"p_CANID" will anyways be remembered in case the mask should be changed.
	 *
	 *	@see setEffectiveCANIDMask()
	 *
	 */
	void setCANID( unsigned int p_CANID );


	/**
	 *	@fn void setCANIDMask( unsigned int p_CANID )
	 *
	 *	@brief Set CAN ID mask resulting from CAN protocol definitions.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_CANIDMask - Mask to be applied to CAN IDs from the bus.
	 *
	 */
	void setCANIDMask( unsigned int p_CANIDMask );


	/**
	 *	@fn void setDataMapping( WECANMappingObject * p_mapping )
	 *
	 *	@brief Set data mapping for this CAN message.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_mapping - pointer to mapping
	 *
	 *	@warning The object this points to will be deleted when this message is destroyed. Use "createDataMapping()" to avoid that.
	 */
	void setDataMapping( WECANMappingObject * p_mapping );


	/**
	 *	@fn void createDataMapping( WECANMappingObject * p_mapping )
	 *
	 *	@brief Create a copy of p_mapping and set it as the data mapping for this CAN message.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_mapping - pointer to mapping which will be copied
	 *
	 */
	virtual void createDataMapping( WECANMappingObject * p_mapping );

	virtual void setEnabled(bool p_enabled);
	virtual bool isEnabled(void) const;

	int setEventAction(int p_eventType, std::shared_ptr<WE::WEAction> p_varAction);

private:
	/**
	 *	@fn void init();
	 *
	 *	@brief Initializes all class members
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	Member 'timeStamp' is initialized with WETick::getTickCount() to allow one full timeout period
	 *	before first timeout calculation may fail.
	 */
	void init();


protected:
	unsigned int		objectID;	//!< The ID of this object
	unsigned int		CANID;		//!< CAN ID as it will/should appear on the bus
	unsigned int		CANIDMask;	//!< mask to be applied to CAN ID from the bus to recognize a receive message
	unsigned int		effectiveCANID;	//!< CAN ID as it will/should appear on the bus - with mask applied

	WECANMappingObject *	dataMapping;	//!< mapping defining the data part

	WECANRequestData *	requestData;	//!< Description of a request to be sent out or received
	unsigned int		requestTimeStamp;	//!< Timestamp to check if periodic request is due

	unsigned int		timeStamp;	//!< The applications tick count when the last matching CAN message was received or sent.
	unsigned int		timeoutPeriod;	//!< Timeout period in milliseconds. To be used to check periodic reception.
						// For messages to be sent periodically use "transmitPeriod" in WECANMappingObject
	WESignalListener * 	eventListener; //!< Pointer to a WESignalListener. To be called if message could not be received (or sent) in time
	bool			timedOut;	//!< Flag that indicates that this object is currently in timed out state

	std::map<int, std::shared_ptr<WE::WEAction>> eventActionMap;

	bool enabled;
};
}	// namespace

#endif /* WE_CAN_MESSAGE_OBJECT_H_ */
