/**
 *	@file we_can_protocol.h
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Common implementation to be used by all or most of the CAN protocols.
 *
 *	The content of this file is collected to cover a common interface and implementation
 *	of features likely to be needed by all CAN protocols.
 *
 *	These include (but are not necessarily limited to):
 *	- data mapping:
 *		Recognize, evaluate and assemble the data part based on constant and variable data
 *		supporting up to 8 bytes of a single CAN message as well as data buffers of arbitrary length.
 *	- message description:
 *		A generic description of a CAN message which e.g. would be extended to a PGN definition for J1939.
 *	- request description:
 *		A generic description of CAN message to trigger sending a transmit mapping or to be sent out to get
 *		receive mapping data.
 *	- sequence description:
 *		A set of messages to be sent or received as a sequence with constant time gap between messages.
 *	- basic implementation of DEI communication:
 *		Implementation of a CAN stack as variable owner (in place of external ECU) and subscriber.
 *	- maintenance of all the above in extendible data structures
 *	- serialization of all the above
 *
 *	Most of the interfaces provided in this stack are a generalization of the original CAN Freestyle implementation
 *	which were considered to be useful for other protocols as well.
 *
 *
 *	@see WECANFreestyle
*/

#ifndef WE_CAN_PROTOCOL_H_
#define WE_CAN_PROTOCOL_H_

#include "../../../common/include/we_error.h"
#include "../../../dei/variables/include/we_variable.h"
#include "../../../dei/include/we_variable_owner.h"
#include "../../../event_handling/include/we_signal.h"
#include "../../../dei/include/we_variable_event_dispatcher.h"
#include "../../../os_abstraction/endpoint_manager/include/we_endpoint_handler.h"
#include "../../../hardware/can_raw/include/we_can_raw.h"
#include "we_can_protocol_base.h"
#include "we_can_message_object.h"

namespace WE
{

//class WECANMessageObject;

typedef std::unordered_map< unsigned int, WECANMessageObject * > long2MessageObjType; //!< Map to be used for storing the message objects of this stack (by their ID)
typedef std::unordered_map< unsigned int, WEVariable * > long2VariableType; //!< Map to be used for storing the Variables owned by this stack

class WEDEI;
struct WECANMessage;
class WECANProtocolStateMachine;


/**
 *	@class WECANProtocolResources
 *
 *	@brief Keep resources public in this class instead of providing a lot of get/set methods.
 *
 *	This class is supposed to be inherited if other protocols need additional parameters.
 *	The intention is to provide all the data needed to initialize a stack in one shot rather than
 *	maintaining large parameter lists or a lot of get/set methods.
 *
 *	Additionally this information can be shared with the state machine easily.
 *
 *	We expect them to be constant for a stack instance - so the members are also "const".
 *
 */
class WECANProtocolResources
{
public:
	WEDEI * const		dei;		//!< pointer to DEI instance to be used
	WEVariableEventDispatcher * const varEventDispatcher; //!< pointer to variable event source
	unsigned int		portNo;		//!< no of CAN port to be used (starting with 1)

	WECANProtocolResources(WEDEI * p_dei, WEVariableEventDispatcher * p_varEventDispatcher, unsigned int p_portNo)
	 :dei(p_dei),
	 varEventDispatcher(p_varEventDispatcher),
	 portNo(p_portNo)
	{
		//nothing more to do here
	}

	virtual ~WECANProtocolResources() = default;
};




class WECANProtocol : public WEVariableOwner, public WESignalListener, public WEEndPointHandler, public WECANProtocolBase
{
public:




	// ************************************************************************************************
	// ************************ (de)initialization related ********************************************
	// ************************************************************************************************


	/**
	 *	@fn ~WECANProtocol(void)
	 *
	 *	@brief Default constructor doing nothing - see init()
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@see init(WECANProtocolResources * p_res, WECANProtocolStateMachine * p_machine)
	 *
	 *
	 */
	WECANProtocol(const std::string & p_instanceName, WEProtocolType p_type);

	/**
	 *	@fn ~WECANProtocol(void)
	 *
	 *	@brief Default destructor doing nothing
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@warning Since this interface can't know how the members are used,
	 *	the implementing classes are responsible for freeing memory correctly
	 *
	 *	Making this destructor virtual ensures that one can even delete an instance of a derived
	 *	class completely by deleting a pointer to WECANProtocol.
	 *
	 *
	 */
	virtual ~WECANProtocol();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WECANProtocol(const WECANProtocol &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WECANProtocol & operator=(const WECANProtocol &) = delete;

	bool startProtocol() override
	{
		return true;
	}

	bool stopProtocol() override
	{
		return true;
	}

	bool resetProtocol() override
	{
		return true;
	}

	bool runProtocol(unsigned int p_time) override;

	virtual int enableMessage(unsigned int p_objectID, bool p_enable) = 0;

	static int commonDebugSection;

	//if the executeMappingActionsDirectly switch is set to true, all mapping actions
	//(timeout, timeout resolved and mapping received) will be directly execute.
	//If it is set to false, an event will be send to an event listener (PClient: mapping DDO).
	//In case of PClient it will post the action to an action queue and later be executed in the main thread.
	static const bool executeMappingActionsDirectly = true;

protected:

	/**
	 *	@fn 	int create(
	 *		WEDEI * const p_dei,
	 *		WEVariableEventDispatcher * const p_varEventDispatcher,
	 *		unsigned int p_portNumber,
	 *		WECANProtocolStateMachine * p_machine,
	 *		WECANProtocolResources * p_res )
	 *
	 *	@brief Initialize members of the stack to defaults and values provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_dei WEDEI instance to be used by this stack and its state machine
	 *	@param p_varEventDispatcher WEVariableEventDispatcher instance providing events from p_dei
	 *	@param p_portNumber CAN port number to be used by this stack
	 *	@param p_machine pointer of an instance of the state machine to be used for this stack (nullptr for CANopen)
	 *	@param p_res pointer to an instance of (a class derived from) WECANProtocolResources to be used in this stack
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_dei is nullptr
	 *	@returns WE_INTERFACE_ERROR if no CAN (socket) interface could be obtained with portNo provided
	 *
	 *	This method initializes the stack and assigns the state machine to be used.
	 *
	 *	The state machine parameter can be used to modify the default behaviour by using a derived state machine.
	 *
	 *	p_res is mainly intended for protocol stacks derived from this one which want to keep more information in their
	 *	protocol resources and therefore need an instance of a class derived from WECANProtocolResources.
	 *	If p_res is nullptr an instance of WECANProtocolResources will be created and filled by this method.
	 *	Otherwise this method will check that the resources instance contains all the data provided.
	 *
	 *
	 *	@warning If p_res is provided here be aware that it will be deleted when this stack instance is deleted.
	 *
	 */
	int create(	WEDEI * const p_dei,
			WEVariableEventDispatcher * const p_varEventDispatcher,
			unsigned int p_portNumber,
			WECANProtocolStateMachine * p_machine,
			WECANProtocolResources * p_res);


public:

	/**
	 *	@fn 	virtual int create(
	 *		WEDEI * const p_dei,
	 *		WEVariableEventDispatcher * const p_varEventDispatcher,
	 *		unsigned int p_portNumber,
	 *		WECANProtocolStateMachine * p_machine)
	 *
	 *	@brief Initialize members of the stack to defaults and values provided.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_dei WEDEI instance to be used by this stack and its state machine
	 *	@param p_varEventDispatcher WEVariableEventDispatcher instance providing events from p_dei
	 *	@param p_portNumber CAN port number to be used by this stack
	 *	@param p_machine pointer of an instance of the state machine to be used for this stack (nullptr for CANopen)
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_dei is nullptr
	 *	@returns WE_INTERFACE_ERROR if no CAN (socket) interface could be obtained with portNo provided
	 *
	 *	This method initializes the stack and assigns the state machine to be used.
	 *
	 *	The state machine parameter can be used to modify the default behaviour by using a derived state machine.
	 *
	 */
	virtual int create(WEDEI * const p_dei,
			WEVariableEventDispatcher * const p_varEventDispatcher,
			unsigned int p_portNumber,
			WECANProtocolStateMachine * p_machine);

	/**
	 *	@fn virtual int init( void )
	 *
	 *	@brief Initialize all class attributes with meaningfull default values. Does not allocate memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_instanceName name this instance should use as a WEVariableOwner
	 *
	 *	@return WE_SUCCESS
	 *
	 *	@see WEVariableOwner
	 *
	 *
	*/
	virtual int init( void );


	/**
	 *	@fn virtual int release();
	 *
	 *	@brief Frees allocated memory and sets all class attributes to default values.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik MuellerStateMachine
	 *
	 *	@return WE_SUCCESS
	*/
	virtual int release();


	/**
	 *	@fn int setResources(WECANProtocolResources * p_res)
	 *
	 *	@brief Set resources for stack (and possibly state machine) in case you don't want to use the init method.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_res pointer to filled resources instance to be used by this stack and its state machine
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if any of the pointers provided in p_res is nullptr
	 *
	 *	This method only checks the content and sets pointers.
	 *
	 *	@warning The stack will delete p_res when it is destroyed - so it should be an instance created specifically.
	 */
	int setResources(WECANProtocolResources * p_res);


	/**
	 *	@fn int createResources(WEDEI * p_dei,
			WEVariableEventDispatcher * p_varEventDispatcher,
			unsigned int p_portNo)
	 *
	 *
	 *	@brief Set resources for stack (and possibly state machine) in case you don't want to use the init method.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_dei pointer to DEI instance to be used
	 *	@param p_varEventDispatcher pointer to event dispatcher providing events of p_dei
	 *	@param p_portNo no of CAN port to be used (starting with 1)
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if any of the pointers provided is nullptr or p_portNo is 0
	 *
	 *	This method checks the content and creates a new instance of WECANProtocolResources for the stack to use.
	 *
	 *	@warning The stack will delete both of the instances provided when it is destroyed itself.
	 */
	int createResources(WEDEI * p_dei,		//!< pointer to DEI instance to be used
			WEVariableEventDispatcher * p_varEventDispatcher,	//!< Get variable events from here
			unsigned int p_portNo	//!< no of CAN port to be used (starting with 1)
			);


	/**
	 *	@fn virtual int reset(void)
	 *
	 *	@brief Reset all members of the stack to "equivalent to new" state.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	This method resets all all the members of the protocol stack into "equivalent to new" state.
	 *	Especially all the pointers are initialized to "nullptr" here.
	 *
	 *	This method does not free any memory and does not (and actually can not) check if that should be done.
	 *
	 *	@warning Only call this method when there is no memory allocated for the members.
	 */
	virtual int reset(void);




	// ************************************************************************************************
	// ************************ state machine related *************************************************
	// ************************************************************************************************


	/**
	 *	@fn int process(unsigned int p_time)
	 *
	 *	@brief The process function needs to be called frequently so that all tasks can be executed
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_time The tick count until which this function is allowed to run
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_TIMEOUT_ERROR in case execution was terminated prematurely because p_time expired
	 *
	 *	This method executes the periodically driven state machine of a stack i.e. it mainly calls the run()
	 *	method of the state machine.
	 *
	 *	@see WECANProtocolStateMachine::run()
	 *
	 */
	virtual int process(unsigned int p_time);

	/**
	 * Create the WECANRaw instenace that is used to send and receive CAN messages.
	 * This function is automatically called when subscribeToEndpointManager() is called.
	 * Using an endpoint manager is the preffered way to use the WE CAN protocol stacks.
	 * If no endpoint manager shall be used for whatever reason, this function can be clled manually.
	 * @return WE_SUCCESS if a WECANRaw instance was successfully created.
	 */
	virtual int createCANPort(void);

	inline WE::WECANRaw * getCANPort(void) const
	{
		return this->canPort.get();
	}

	// ************************************************************************************************
	// ************************ variable mappping related *********************************************
	// ************************************************************************************************

protected:


	/**
	 *	@fn int addMappingMessage(WECANMessageObject * p_object)
	 *
	 *	@brief Add a mapping message object to the CAN stack
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_object Pointer to an existing WECANMessageObject (should be created by a child stack).
	 *
	 *	@returns WE_INVALID_PARAMETER_ERROR If p_object is nullptr
	 *	@returns WE_ALREADY_EXISTS_ERROR If p_object with the same ID was added to the stack before
	 *	@returns WE_WRITE_ERROR If inserting p_var into the variable map failed
	 *	@returns WE_SUCCESS if adding the variable was successful
	 *	@returns anything handleNewVariable(WEVariable * p_var) may return
	 *
	 *	This adds a WECANMessageObject to this stack. Its main purpose is to add the message to the
	 *	internal map of message objects.
	 *
	 *	@warning p_object should point to an object created by a stack inheriting from this one (because it will be deleted in "release()")
	 *	@warning This method does not create copies - mainly in order to also handle types derived from WECANMessageObject.
	 *
	 *	@see getMappingMessageByID(unsigned int p_ID)
	 *
	 */
	int addMappingMessage(WECANMessageObject * p_object);


	/**
	 *	@fn WECANMessageObject * getMappingMessageByID(unsigned int p_ID)
	 *
	 *	@brief Get a mapping message object of the CAN stack by its ID
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_ID ID of the object to be found.
	 *
	 *	@returns nullptr if object with the ID provided is not stored in the map
	 *	@returns A pointer to the mapping message with ID p_ID otherwise
	 *
	 *	The WECANMessageObject returned here has to be stored in the map by calling addMappingMessage() before
	 *
	 *	@warning Object returned will be deleted in "release()" i.e. when this stack instance is deleted.
	 *
	 *	@see addMappingMessage(WECANMessageObject * p_object)
	 *
	 */
	WECANMessageObject * getMappingMessageByID(unsigned int p_ID) const;




	// ************************************************************************************************
	// ************************ variable owner related ************************************************
	// ************************************************************************************************

public:

	/**
	 *	@fn int addVariable(WEVariable * p_var)
	 *
	 *	@brief Add a variable owned by the CAN stack to the CAN stack
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_var Pointer to an existing WEVariable.
	 *
	 *	@returns WE_INVALID_PARAMETER_ERROR If p_var is nullptr
	 *	@returns WE_ALREADY_EXISTS_ERROR If p_var was added to the stack before
	 *	@returns WE_WRITE_ERROR If inserting p_var into the variable map failed
	 *	@returns WE_SUCCESS if adding the variable was successful
	 *	@returns anything handleNewVariable(WEVariable * p_var) may return
	 *
	 *	@warning Currently only 32 bit signed integers can be added. WEVariable has to be enhanced to use more data types.
	 *
	 *	This adds a WEVariable that is owned by this CAN stack to the stack. The variable type will
	 *	automatically set to the correct type for the stack (assuming this will be overwritten) and the owner of the variable
	 *	is set to this instance.
	 *	The variable will also automatically be added as external variable to the DEI.
	 *
	 *	After adding the variable to the stack "handleNewVariable(WEVariable * p_var)" will be called.
	 *	Implementation of this method should be stack specific (e.g. CANopen stack will add it to its object dictionary).
	 *
	 *	@see addVariable(WEVariable * p_var)
	 *	@see handleNewVariable(WEVariable * p_var, WEVariableType p_type)
	 *
	 */
	virtual int addVariable(WEVariable * p_var);


	/**
	 *	@fn int updateRequest(unsigned short p_index, unsigned short p_subindex, int p_intValue)
	 *	@fn int updateRequest(unsigned int p_varKey, int p_intValue)
	 *
	 *	@brief Implementation of WEVariableOwner's updateRequest function
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the variable for which an update (write) request shall be sent.
	 *	@param p_subindex Subindex of the variable for which an update (write) request shall be sent.
	 *	@param p_intValue New integer value that shall be written to the variable.
	 *
	 *	@returns WE_SUCCESS if the CAN write request was sent out successfully
	 *	@returns WE_NOT_FOUND_ERROR if no variable was found that matches p_index and p_subindex
	 *	@returns return value of WE::WECANRaw::send(WECANMessage * messageToSend)
	 *
	 *	@see WECANRaw
	 *
	 *	The updateRequest function gets called by WEDEI when an application wants to change the value
	 *	of an external variable.
	 *	With this method the DEI asks the owner of the variable (here the CAN stack) to change the
	 *	variable value. Since the CAN stack can typically not decide whether the new value is reasonable
	 *	or not it will send a "write request" to the ECU which actually maintains the variable.
	 *	A "write request" is any CAN mapping that contains the respective variable and has the
	 *	"WE_CAN_TRANSMIT_TYPE_ON_WRITE_REQUEST" as well as the "WE_CAN_TRANSMIT_TYPE_ANY_VALUE_CHANGED" or
	 *	"WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_CHANGED" bits set ("WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_CHANGED" only works
	 *	if the "certain value" indicates the variable that should be modified).
	 *	The actual value of the variable will only be changed in DEI if a read response for that variable
	 *	arrives on the bus.
	 */
	int updateRequest(unsigned short p_index, unsigned short p_subindex, int p_intValue) override;

	virtual int updateRequest(unsigned int p_varKey, int p_intValue) = 0;



	/**
	 *	@fn int updateRequest(unsigned short p_index, unsigned short p_subindex, std::string p_stringValue)
	 *	@fn int updateRequest(unsigned int p_varKey, std::string & p_stringValue)
	 *
	 *	@brief Implementation of WEVariableOwner's updateRequest function
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_index Index of the variable for which an update (write) request shall be sent.
	 *	@param p_subindex Subindex of the variable for which an update (write) request shall be sent.
	 *	@param p_stringValue New string value that shall be written to the variable.
	 *
	 *	@returns WE_SUCCESS if the CAN write request was sent out successfully
	 *	@returns WE_NOT_FOUND_ERROR if no variable was found that matches p_index and p_subindex
	 *	@returns return value of WE::WECANRaw::send(WECANMessage * messageToSend)
	 *
	 *	@see WECANRaw
	 *
	 *	The updateRequest function gets called by WEDEI when an application wants to change the value
	 *	of an external variable.
	 *	With this method the DEI asks the owner of the variable (here the CAN stack) to change the
	 *	variable value. Since the CAN stack can typically not decide whether the new value is reasonable
	 *	or not it will send a "write request" to the ECU which actually maintains the variable.
	 *	A "write request" is any CAN mapping that contains the respective variable and has the
	 *	"WE_CAN_TRANSMIT_TYPE_ON_WRITE_REQUEST" as well as the "WE_CAN_TRANSMIT_TYPE_ANY_VALUE_CHANGED" or
	 *	"WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_CHANGED" bits set ("WE_CAN_TRANSMIT_TYPE_CERTAIN_VALUE_CHANGED" only works
	 *	if the "certain value" indicates the variable that should be modified).
	 *	The actual value of the variable will only be changed in DEI if a read response for that variable
	 *	arrives on the bus.
	 *	Please note that the actual stack implementation using this method may not support sending out string values.
	 */
	int updateRequest(unsigned short p_index, unsigned short p_subindex, const std::string & p_stringValue) override;

	virtual int updateRequest(unsigned int p_varKey, const std::string & p_stringValue) = 0;

	/**
	 *	@fn int getValueRequest(unsigned short p_index, unsigned short p_subindex)
	 *	@fn getValueRequest(unsigned int p_varKey)
	 *
	 *	@brief Implementation of WEVariableOwner's updateRequest function
	 *
	 *	@author ju@wachendorff.de
	 *	@author Joachim Urbach
	 *
	 *	@param p_index Index of the variable where the value which we want to read shall be stored.
	 *	@param p_subindex Subindex of the variable where the value which want to we read shall be stored.
	 *		   The Variable needs an external identifier. The read request will be send to them.
	 *
	 *	@returns WE_SUCCESS if the CAN read request was sent out successfully
	 *	@returns WE_NOT_FOUND_ERROR if no variable was found that matches p_index and p_subindex
	 *	@returns return value of WE::WECANRaw::send(WECANMessage * messageToSend)
	 *
	 *	@see WECANRaw
	 *
	 *  The getValueRequest function is called by WEDEI when an application wants to read the value
	 *  of an external variable. With this method the DEI asks the owner of the variable (here a foreign stack)
	 *  to send us the variable value.
	 *  We will send a "read request" to the ECU which actually maintains the variable.
	 *  The index of the Object Dictionary Entry is stored in the external identifier of the variable.
	 */
	int getValueRequest(unsigned short p_index, unsigned short p_subindex) override;
	virtual int getValueRequest(unsigned int p_varKey) = 0;


	// ************************************************************************************************
	// ************************ CAN port related ******************************************************
	// ************************************************************************************************


	/**
	 *	@fn void setMaxMsgAtOnce(unsigned int p_msgs)
	 *
	 *	@brief Set max number of messages to be read if queue isn't empty before.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param max number of messages to be read if queue isn't empty before
	 *
	 *	@returns nothing
	 *
	 */
	void setMaxMsgAtOnce(unsigned int p_msgs);

	/**
	 *	@fn unsigned int getMaxMsgAtOnce(void)
	 *
	 *	@brief Get max number of messages to be read if queue isn't empty before.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *	@returns max number of messages to be read if queue isn't empty before
	 *
	 */
	unsigned int getMaxMsgAtOnce(void) const;



	//WEEndPointHandler:
	/**
	 *	@fn virtual void onEndPointReadable(WEEndPoint* p_endPoint, WEEndPointManager* p_endPointMgr)
	 *
	 *	@brief Signal that a message on the CAN port of this stack arrived
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_endPoint the endpoint (in this case CAN port) that got readable
	 *	@param p_endPointManager the corresponding endpoint manager (not used)
	 *
	 *	This function will be called by WEEndPointManager as soon as the CAN port of this stack gets
	 *	readable (a CAN message arrived). The CANopen stack then reads the message from CAN and starts
	 *	to process it. The message will be passed to the single message handlers until one of them identifies it
	 *	to belong to it and then processes it.
	 */
	virtual void onEndPointReadable(std::shared_ptr<WEEndPoint> p_endPoint, WEEndPointManager* p_endPointMgr) override;


	/**
	 *	@fn virtual void onEndPointReconnected(WEEndPoint* p_endpoint)
	 *
	 *	@brief Signal that a disconnected endpoint got connected again
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_endpoint the re-connected endpoint
	 *
	 *	This function is not used.
	 */
	virtual void onEndPointReconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;

	/**
	 *	@fn void onEndPointDisconnected(WEEndPoint* p_endpoint)
	 *
	 *	@brief Call this function if a IPC connection was lost
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_endpoint Endpoint of the connection that was disconnected
	 *
	 */
	virtual void onEndPointDisconnected(std::shared_ptr<WEEndPoint> p_endpoint) override;

	/**
	 *	@fn int subscribeToEndpointManager(WEEndPointManager * p_mgr)
	 *
	 *	@brief Tell the endpoint manager provided that we are interested in messages from our CAN port.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_mgr Endpoint manager to connect to.
	 *
	 *	@returns anything WEEndPointManager::addEndPoint() could return
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_mgr is nullptr
	 *
	 *	@see WEEndPointManager::addEndPoint()
	 *
	 */
	virtual int subscribeToEndpointManager(WEEndPointManager * p_mgr) override;


	/**
	 *	@fn int unsubscribeFromEndpointManager(WEEndPointManager * p_mgr)
	 *
	 *	@brief Tell the endpoint manager provided that we are NOT interested in messages from our CAN port any more.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_mgr Endpoint manager to disconnect from.
	 *
	 *	@returns anything WEEndPointManager::removeEndPoint() could return
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_mgr is nullptr
	 *
	 *	@see WEEndPointManager::removeEndPoint()
	 *
	 */
	int unsubscribeFromEndpointManager(WEEndPointManager * p_mgr);


	/**
	 *	@fn int sendMessage(WECANMessage * p_msg)
	 *
	 *	@brief Send message provided via this->canPort
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_msg Pointer to message to be sent out
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_msg == nullptr
	 *	@returns WE_INTERFACE_ERROR if this->canPort == nullptr
	 *	@returns any return value of WECANRaw::send()
	 *
	 *	This method sends out the message provided via the port assigned to this protocol stack.
	 *
	 *	@see WECANRaw::send()
	 *
	 */
	int sendMessage(WECANMessage * p_msg);




	// ************************************************************************************************
	// ************************ variable owner related ************************************************
	// ************************************************************************************************

protected:

	/**
	 *	@fn int addVariable(WEVariable * p_var, WEVariableType p_type)
	 *
	 *	@brief Add a variable owned by the CAN stack to the CAN stack
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_var Pointer to an existing WEVariable.
	 *	@param p_type Type to be set for the variable before adding it to DEI
	 *
	 *	@returns WE_INVALID_PARAMETER_ERROR If p_var is nullptr
	 *	@returns WE_ALREADY_EXISTS_ERROR If p_var was added to the stack before
	 *	@returns WE_WRITE_ERROR If inserting p_var into the variable map failed
	 *	@returns WE_SUCCESS if adding the variable was successful
	 *	@returns anything handleNewVariable(WEVariable * p_var) may return
	 *
	 *	@warning Currently only 32 bit signed integers can be added. WEVariable has to be enhanced to use more data types.
	 *
	 *	This adds a WEVariable that is owned by this CAN stack to the stack. The variable type will
	 *	automatically set to p_type and the owner of the variable is set to this instance.
	 *	The variable will also automatically be added as external variable to the DEI.
	 *
	 *	After adding the variable to the stack "handleNewVariable(WEVariable * p_var)" will be called.
	 *	Implementation of this method should be stack specific (e.g. CANopen stack will add it to its object dictionary).
	 *
	 *	@see addVariable(WEVariable * p_var)
	 *	@see handleNewVariable(WEVariable * p_var)
	 *
	 */
	int addVariable(WEVariable * p_var, WEVariableType p_type);


	/**
	 *	@fn handleNewVariable(WEVariable * p_var)
	 *
	 *	@brief Do stack specific handling when adding a variable owned by the CAN stack to the CAN stack
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_var Pointer to an existing WEVariable.
	 *
	 *	@returns WE_SUCCESS if adding the variable was successful
	 *
	 *
	 *	This is the place to implement anything protocol specific e.g. CANopen stack will add
	 *	the new variable to its object dictionary.
	 *
	 *	@see handleNewVariable(WEVariable * p_var)
	 *
	 */
	virtual int handleNewVariable(WEVariable * p_var);


	WECANProtocolResources	* resources{nullptr};	//!< pointer to data that will not change while stack is running and can be shared with state machine.

	std::shared_ptr<WECANRaw> canPort{};		//!< instance of CAN port to be used

	unsigned int maxMsgAtOnce{1};	//!< When CAN port becomes readable empty it but don't read more than maxMsgAtOnce messages in one pass

	WECANProtocolStateMachine * stateMachine{nullptr};	//!< Pointer to this CAN stack's state machine - separate class, so that it could be chosen at runtime

	// WECANMessage * lastReceivedMessage;	//!< The last received message of the stack i.e. the one to be handled currently


	// In order to react to CAN messages and DEI requests we generally need some maps:
	long2VariableType key2varOwnedByStack{};	//!< variables with values coming from distant ECUs --> the var is owned by the stack

	long2MessageObjType id2MsgObj{};	// all the message objects - accessible by their ID

	// (1) (masked) CAN ID -> receive message description (one list/map per ID mask; to forward new values)
	// (2) Variable key -> transmit message description (to satisfy "updateRequest" and variable change event)
	// (3) list of receive message descriptions with timeout (to set variables to "invalid") if message expires
	// (4) list of transmit messages with repeat time (to send them out in time)
	// (5) list of transmit sequences (to send out their messages in time)

	// Depending on the protocol these maps will have different keys and different data content - so it seems we can not provide them here.


	// Another task for a stack could be to maintain the state of different ECUs on the bus
	// Basic support for this will probably be implemented in a separate class.
};
}	// namespace

#endif /* WE_CAN_PROTOCOL_H_ */
