/**
 *	@file we_can_protocol_base.h
 *
 *	@author mka@wachendorff.de
 *	@author Markus Kalin
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2012
 *
 *	@brief abstract base class for all CAN protocols
 *
 *	If CANfreestyle is inherting the CANprotocol sub class this class is not needed.
 */

#ifndef WE_CAN_PROTOCOL_BASE_H_
#define WE_CAN_PROTOCOL_BASE_H_

namespace WE
{
class WEEndPointManager;

class WECANProtocolBase
{
public:
	WECANProtocolBase() = default;

	 virtual ~WECANProtocolBase() = default;
	 /**
	  *	@fn bool startProtocol()
	  *
	  *	@brief called inside thread to start the state machine
	  *	@brief
	  *	@brief Currently not implemented by the stacks
	  *
	  *	@author mka@wachendorff.de
	  *	@author Markus Kalin
	  *
	  *	@returns true if the stack could be started and false if there is any error
	  *
	  */
	 virtual bool startProtocol() = 0;

	 /**
	  *	@fn bool stopProtocol()
	  *
	  *	@brief called inside thread to stop the state machine
	  *	@brief
	  *	@brief Currently not implemented by the stacks
	  *
	  *	@author mka@wachendorff.de
	  *	@author Markus Kalin
	  *
	  *	@returns true if the stack could be stopped and false if there is any error
	  *
	  */
	 virtual bool stopProtocol() = 0;

	 /**
	  *	@fn bool resetProtocol()
	  *
	  *	@brief called inside thread to stop the state machine
	  *	@brief
	  *	@brief Currently not implemented by the stacks
	  *
	  *	@author mka@wachendorff.de
	  *	@author Markus Kalin
	  *
	  *	@returns true if the stack could be reset and false if there is any error
	  *
	  */
	 virtual bool resetProtocol() = 0;

	 /**
	  *	@fn virtual bool runProtocol(unsigned int p_time) = 0
	  *
	  *	@brief called inside thread frequently, here you have to code your own machine
	  *
	  *	@author mka@wachendorff.de
	  *	@author Markus Kalin
	  *
	  *	@returns true if the machine returns without problem and false if there is any error
	  *
	  */
	 virtual bool runProtocol(unsigned int p_time) = 0;

	 /**
	  *	@fn void setInstanceName(const std::string & p_instanceName)
	  *
	  *	@brief you can set a name here to identify the stack instance. Helps debugging
	  *
	  *	@author mka@wachendorff.de
	  *	@author Markus Kalin
	  *
	  *	@returns nothing
	  *
	  */
	 void setInstanceName(const std::string & p_instanceName)
	 {
		 this->myName = p_instanceName;
	 }

	 const std::string & getInstanceName() const
	 {
		 return this->myName;
	 }

	 void setWaitForEndpointsTime(unsigned int p_time)
	 {
		 this->waitForEntpointsTime = p_time;
	 }

	 unsigned int getWaitForEndpointsTime() const
	 {
		 return this->waitForEntpointsTime;
	 }

	 /*
	  * max wait for entpoints time:
	  *
	  * It is used to set the maximum time to wait. Currently it is set to 5ms. That means that the CAN thread has
	  * has a maximum wait time of 5ms. Currently it is hard coded. for a better handling we can think about setting it
	  * during XML parsing. If you have only transmit mappings which have to be send for example 100ms we can think about
	  * to wait 10 or 20 ms.
	  */
	 void setMaxWaitTime(unsigned int p_time)
	 {
		 this->maxWaitForEntpointsTime = p_time;
	 }

	 unsigned int getMaxWaitTime() const
	 {
		 return this->maxWaitForEntpointsTime;
	 }

	 virtual int subscribeToEndpointManager(WEEndPointManager* p_mgr) = 0;
private:

	 std::string myName{};
	 unsigned int waitForEntpointsTime{5};
	 unsigned int maxWaitForEntpointsTime{5};
};
}//end of namespace WE


#endif /* WE_CAN_PROTOCOL_BASE_H_ */
