/**
 *	@file we_can_transmit_sequence.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2009
 *
 *	@brief Class that stores data of a transmit sequence
 *
 *	This file was originally designed for CAN Freestyle implementation but we figured out
 *	that it is useful for other protocols such as J1939 also.
 *
 *
 *	@see WECANMappingObject
*/

#ifndef WE_CAN_TRANSMIT_SEQUENCE_H_
#define WE_CAN_TRANSMIT_SEQUENCE_H_


#include <vector>

namespace WE
{

class WECANMappingObject; //!< Forward declaration of WECANMappingObject
class WECANMessageObject;

/**
 *	@class WECANTransmitSequence
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief A CANFreestyle transmit sequence can be used to send out different
 *		CANFreestyleMappingObjects with a defined period.
 *
 *	@namespace WE
 *
 *	@see WECANMappingObject
 *
 *	This class is mainly for data storage. Most of the work is done by the CANFreestyle that
 *	to which this sequence belongs to.
 */
class WECANTransmitSequence
{
public:
	/**
	 *	@fn WECANTransmitSequence(int p_ID, unsigned int p_period)
	 *
	 *	@brief Constructor
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_ID The ID of this transmit sequence
	 *	@param p_period Every p_period milliseconds a transmit object will be sent
	 */
	WECANTransmitSequence(int p_ID, unsigned int p_period);



	/**
	 *	@fn WECANTransmitSequence(int p_ID, unsigned int p_period)
	 *
	 *	@brief Constructor
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 */
	virtual ~WECANTransmitSequence();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WECANTransmitSequence(const WECANTransmitSequence &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WECANTransmitSequence & operator=(const WECANTransmitSequence &) = delete;

	/**
	 *	@fn int release( void )
	 *
	 *	@brief Clear this sequence and delete all message objects it contains.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *	This method reduces this sequence to 0 messages and deletes every pointer to message object stored in it.
	 *	The objects themselves have to be deleted elsewhere.
	 *
	 *	As a fallback this is also called in the destructor.
	 *
	 *	@see addTransmitObject()
	 *
	 */
	int release( void );



	/**
	 *	@fn int release( void )
	 *
	 *	@brief Set all values to new and unused state.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none.
	 *
	 *	@returns WE_SUCCESS in case of success
	 *
	 *
	 *	@warning This method even resets pointers which may point to allocated memory and could thus cause memory leaks.
	 */
	int reset( void );



	/**
	 *	@fn int addTransmitObject(WECANMessageObject * p_object)
	 *
	 *	@brief Add a transmit object to this transmit sequence
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_object Pointer to a WECANMappingObject.
	 *
	 *	@returns WE_SUCCESS in case pointer is stored
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_object or the data mapping it contains is nullptr
	 *	@returns WE_RECEIVE_ERROR if p_object's transmit type does not include WE_CAN_TRANSMIT_TYPE_SEQUENCE
	 *
	 *	With this function a transmit object can be added to this transmit sequence.
	 *	The object must be created BEFORE this function is called i.e. it is NOT copied here.
	 *
	 *	@see release()
	 *
	 */
	int addTransmitObject(WECANMessageObject * p_object);



	/**
	 *	@fn bool messageToBeSentNow( unsigned int p_tickNow, WECANMessageObject * * p_msg)
	 *
	 *	@brief Return if and which message has to be sent now.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_tickNow Current time as tick.
	 *	@param p_msg address of pointer to a message object - will be set to message object to send
	 *
	 *	@returns true if a message has to be sent now; in this case *p_msg will point to this message
	 *	@returns false if there is no message to be sent now - *p_msg will be nullptr
	 *
	 */
	bool messageToBeSentNow( unsigned int p_tickNow, WECANMessageObject * * p_msg);

	int getID(void) const
	{
		return this->ID;
	}

	unsigned int getPeriod(void) const
	{
		return this->period;
	}

	const std::vector< WECANMessageObject * > & getSequence(void) const
	{
		return this->sequence;
	}

private:
	int ID{0};						//!< The sequence ID
	std::vector< WECANMessageObject * > sequence{};	//!< Vector containing pointers to the transmit objects of this sequence
	unsigned int lastSentIndex{0};			//!< The last (vector) index of the last sent transmit object
	unsigned int tickOfLastSending{0};		//!< The application's tick count at the time when the last transmit object was sent out
	unsigned int period{100};				//!< The period in milliseconds when a transmit object is sent
};
} // namespace WE



#endif /* WE_CAN_TRANSMIT_SEQUENCE_H_ */
