/**
 *	@file we_j1939_definitions.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief This file contains definitions of types, error codes, constants, etc.
 *
 */

#ifndef WE_J1939_DEFINITIONS_H_
#define WE_J1939_DEFINITIONS_H_

namespace WE
{

/**
 *	@enum J1939AddressClaimStates
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief Possible J1939 address states
 *
 *	@namespace WE
 */
enum J1939AddressClaimStates
{
	J1939_ADDRESS_CLAIM_STATE_NO_ADDRESS,
	J1939_ADDRESS_CLAIM_STATE_ADDRESS_CLAIMED,
	J1939_ADDRESS_CLAIM_STATE_ADDRESS_CLAIM_SENT,
	J1939_ADDRESS_CLAIM_STATE_CLAIM_ADDRESS,
	J1939_ADDRESS_CLAIM_STATE_DISCOVER
};


enum J1939AckControlByteValues
{
	ACK_CONTROL_BYTE_ACK				= 0,
	ACK_CONTROL_BYTE_NACK				= 1,
	ACK_CONTROL_BYTE_ACCESS_DENIED			= 2,
	ACK_CONTROL_BYTE_CANNOT_RESPOND			= 4
};


enum J1939AckGroupFunctionByteValues
{
	ACK_GROUP_FUNCTION_BYTE_NOT_APPLICABLE		= 0xff
};


/**
 * Definition of special addresses
 */
static const unsigned char J1939_NULL_ADDRESS = 254;		//!< The null address defined by J1939
static const unsigned char J1939_BROADCAST_ADDRESS = 255;	//!< The broadcast address defined by J1939
static const unsigned int J1939_NUM_ADDRESSES = 256;		//!< Number of available J1939 addresses

/**
 * Definition of special PGNs
 */
static const unsigned int J1939_PGN_REQUEST_FOR_PGN = 59904;		// 0x00EA00
static const unsigned int J1939_PGN_CLAIMED_ADDRESS = 60928;		// 0x00EE00
// AXIVION Next Line Generic-LocalScope CWE-Bad-Coding-Practices-1126 : Next variable could be declared more locally but this is the more readable place for it
static const unsigned int J1939_PGN_COMMANDED_ADDRESS = 65240;		// 0x00FED8
static const unsigned int J1939_PGN_TP_CONNECTION_MANAGEMENT = 60416;	// 0x00EC00
static const unsigned int J1939_PGN_TP_DATA_TRANSFER = 60160;		// 0x00EB00
static const unsigned int J1939_PGN_ETP_CONNECTION_MANAGEMENT = 51200;	// 0x00C800
static const unsigned int J1939_PGN_ETP_DATA_TRANSFER = 50944;		// 0x00C700
static const unsigned int J1939_PGN_ACK = 0x00E800;			//
// AXIVION Next Line Generic-LocalScope CWE-Bad-Coding-Practices-1126 : Next variable could be declared more locally but this is the more readable place for it
static const unsigned int J1939_PGN_DM1 = 0x00FECA;			//

/**
 * Definitions of J1939 Transport Protocol
 */
static const unsigned char J1939_TP_ABORT_REASON_HAVE_OTHER_CONNECTION = 1;
// AXIVION Next Line Generic-LocalScope CWE-Bad-Coding-Practices-1126 : Next variable could be declared more locally but this is the more readable place for it
static const unsigned char J1939_TP_ABORT_REASON_NO_RESOURCES = 2;
static const unsigned char J1939_TP_ABORT_REASON_TIMEOUT = 3;
static const unsigned char J1939_TP_ABORT_REASON_ANY_OTHER = 0xff;

/**
 * Definition of default priorities
 */
static const unsigned int J1939_PRIO_CONTROL_MSG = 6;
static const unsigned int J1939_PRIO_INFO_MSG = 6;
static const unsigned int J1939_PRIO_PROPRIETARY_MSG = 6;
static const unsigned int J1939_PRIO_REQUEST_MSG = 6;
// AXIVION Next Line Generic-LocalScope CWE-Bad-Coding-Practices-1126 : Next variable could be declared more locally but this is the more readable place for it
static const unsigned int J1939_PRIO_ACK_MSG = 6;
static const unsigned int J1939_PRIO_TP_MSG = 7;

/**
 * Definition for timeout values
 */
static const unsigned int J1939_TIME_REQUEST_FOR_ADDRESS_CLAIMED_RESPONSE	= 1250;	//!< [ms]
// AXIVION Next Line Generic-LocalScope CWE-Bad-Coding-Practices-1126 : Next variable could be declared more locally but this is the more readable place for it
static const unsigned int J1939_TIME_ADDRESS_CLAIMED_RESPONSE			=  250;	//!< [ms]

// timeouts if we transfer data:
static const unsigned int J1939_TIMEOUT_CLEAR_TO_SEND				= 1250;	//!< [ms] Max time we wait for CTS after our RTS is sent

// timeouts if we receive data:
static const unsigned int J1939_TIMEOUT_DPO_OR_FIRST_DATA			= 1250;	//!< [ms] Max time we wait for DPO (ETP) or first DT (TP) after sending CTS
static const unsigned int J1939_TIMEOUT_TP_DATA_RECEIVE			=  750;	//!< [ms] Max time between DPO and DT or DT and DT
static const unsigned int J1939_TIMEOUT_ETP_DATA_RECEIVE			=  750;	//!< [ms] Max time between DPO and DT or DT and DT

/**
 * Definition of J1939 DEI variable indexes
 */
static const int J1939_DEI_INITIAL_ADDRESS_CAN1 = 0x23030002;
static const int J1939_DEI_INITIAL_ADDRESS_CAN2 = 0x23130002;
static const int J1939_DEI_INITIAL_ADDRESS_CAN3 = 0x23230002;
static const int J1939_DEI_INITIAL_ADDRESS_CAN4 = 0x23330002;
}	// namespace

#endif /* WE_J1939_DEFINITIONS_H_ */
