/**
 *	@file we_j1939_mapping_object.h
 *
 *	@author ml@wachendorff.de
 *	@author Michael Lenhart
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definitions of data structures used for J1939 protocol variable mapping
 *
 *	This file only contains deviations from the default implementation in we_can_mapping_object.h
 *
 *
 *	@see WECANMappingObject
 *	@see WECANMessageObject
 *	@see WECANFreestyle
*/

#ifndef WE_J1939_MAPPING_OBJECT_H_
#define WE_J1939_MAPPING_OBJECT_H_


#include "../../common/include/we_can_mapping_object.h"		// base class definition is required in any case



namespace WE
{
struct WECANMappingData;


class WEJ1939MappingObject : public WECANMappingObject
{
public:
	//base object has two checkRawValue virtual functions but we override only one
	using WECANMappingObject::checkRawValue;

	/**
	 *	@fn WEJ1939MappingObject(
			unsigned int p_messageLength,
			unsigned int p_transmitType,
			unsigned int p_transmitTriggerKey,
			unsigned int p_transmitPeriod,
			int p_mappingType,
			bool p_littleEndian,
			bool p_useInvalidTransmitPattern = false,
			unsigned int p_invalidTransmitPattern = 0
			)
	 *
	 *	@brief Contructor providing all necessary data for creating an instance of this class
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_messageLength Number of bytes of the CAN message (DLC)
	 *	@param p_transmitType trigger definition for a transmit object or WE_CAN_TRANSMIT_TYPE_RECEIVE
	 *	@param p_transmitTriggerKey if p_transmitType includes XXX_CERTAIN_VALUE_XXX, this contains the key of the variable
	 *	@param p_transmitPeriod if p_transmitType is XXX_PERIODIC_XXX, this contains the transmit period in milliseconds
	 *	@param p_mappingType type of this mapping (i.e. length, delimiter or mixed mapping type)
	 *	@param p_littleEndian If true, received and sent CAN message data bytes will be treated as little endian values
	 *	@param p_useInvalidTransmitPattern If false the value from DEI will be used for transmit even if marked as "invalid"
	 *	@param p_invalidTransmitPattern pattern to replace value if it is reported "invalid" by DEI
	 *
	 *	This constructor only calls the similar one of the parent class.
	 *
	 */
	WEJ1939MappingObject(
			unsigned int p_messageLength,
			unsigned int p_transmitType,
			unsigned int p_transmitTriggerKey,
			unsigned int p_transmitPeriod,
			int p_mappingType,
			bool p_littleEndian,
			bool p_useInvalidTransmitPattern = false,
			unsigned int p_invalidTransmitPattern = 0
			);


	/**
	 *	@fn ~WEJ1939MappingObject()
	 *
	 *	@brief Detructor of this class
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param none
	 *
	 *
	 */
	virtual ~WEJ1939MappingObject( ) = default;

	/**
	 * Deleted assignment operator. If we need to copy assign this class please implement this.
	 */
	WEJ1939MappingObject & operator=(const WEJ1939MappingObject &) = delete;


	/**
	 *	@fn WEJ1939MappingObject(const WEJ1939MappingObject & p_mapping)
	 *
	 *	@brief Copy contructor providing copying all data to a new instance of this class.
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_mapping - object to be copied
	 *
	 *	@warning This does NOT copy the vector of pointers to mapping data to avoid trouble with memory allocation and inheritance.
	 *
	 *	This method currently only calls the copy constructor of the parent class.
	 *
	 *	@see WECANMappingObject::WECANMappingObject(const WECANMappingObject & p_mapping)
	 *
	 */
	WEJ1939MappingObject(const WEJ1939MappingObject & p_mapping);


	/**
	 *	@fn int checkRawValue(unsigned int p_rawValue, WECANMappingData * p_mapping)
	 *
	 *	@brief Check if the raw value extracted from CAN message may have a dedicated meaning (such as "don't care)
	 *
	 *	@author ml@wachendorff.de
	 *	@author Michael Lenhart
	 *
	 *	@param p_rawValue Value extracted from CAN message (assembled according to endianess but no scale, offset ... applied)
	 *	@param p_mapping The WECANMappingData used to fetch p_rawValue from CAN message
	 *
	 *	@returns WE_SUCCESS if p_rawValue is a normal, valid value
	 *	@returns WE_INVALID_MESSAGE_RECEIVED_ERROR if the value should be set to "invalid" (e.g. in DEI)
	 *	@returns WE_NOT_SUPPORTED_ERROR if the value should just be ignored (i.e. indicates "don't care")
	 *
	 *	This function gets called by the extractMappedValues function for each mapped variable.
	 *	This implementation checks if all bits of a mapped value are set to 1.
	 *	If they are it returns WE_NOT_SUPPORTED_ERROR so that the value is ignored.
	 *
	 *	p_mapping can be used to get more information about p_rawValue e.g. the number of bits actually extracted from the CAn message.
	 *
	 *
	 *	@see int extractMappedValues(WECANMessage * p_msg, WEDEI * p_dei, WEVariableOwner * p_owner)
	 *	@see int handleExtractedValue(unsigned int p_key, int p_value, WEDEI * p_dei, WEVariableOwner * p_owner, int p_valueValid)
	 *	@see WEPDO
	 */
	int checkRawValue(unsigned int p_rawValue, WECANMappingData * p_mapping) const override;
};	// class WEJ1939MappingObject
}	// namespace
#endif /* WE_J1939_MAPPING_OBJECT_H_ */
