/**
 *	@file we_can_open_state_machine.h
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Header file for WEJ1939StateMachine class
 *
 */

#ifndef WE_J1939_STATE_MACHINE_H_
#define WE_J1939_STATE_MACHINE_H_

#include "../../common/include/we_can_protocol_state_machine.h"

namespace WE
{

class WEJ1939Resources;

/**
 *	@class WEJ1939StateMachine
 *
 *	@author hm@wachendorff.de
 *	@author Hendrik Mueller
 *
 *	@brief Class for a J1939 State machine
 *
 *	@namespace WE
 *
 *	This class represents a J1939 state machine.
 *	Before starting the machine, be sure to set all the pointers (setMessageHandlers,
 *	setDEI, setCANinterface. Don't use this class from outside. It is intended to be
 *	used by J1939 class only.
 */
class WEJ1939StateMachine : public WECANProtocolStateMachine
{
public:
	/**
	 *	@fn WEJ1939StateMachine()
	 *
	 *	@brief Constructor
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 */
	WEJ1939StateMachine();

	/**
	 *	@fn ~WEJ1939StateMachine()
	 *
	 *	@brief Destructor, frees memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 */
	virtual ~WEJ1939StateMachine();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEJ1939StateMachine(const WEJ1939StateMachine &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEJ1939StateMachine & operator=(const WEJ1939StateMachine &) = delete;

	/**
	 *	@fn virtual int create(WECANProtocolResources * p_res)
	 *
	 *	@brief Initialize the J1939 state machine and allocates needed memory
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_res pointer to filled resources instance to be used by the stack and this state machine
	 *
	 *	@returns WE_SUCCESS in case of success
	 *	@returns WE_INVALID_PARAMETER_ERROR if any of the pointers provided in p_res is nullptr
	 *
	 *	You can use this init function if you don't want to set the port number and DEI before.
	 *	This function implicitly calls the init() function without parameter.
	 */
	virtual int create(WECANProtocolResources * p_res) override;

	/**
	 *	@fn virtual init()
	 *
	 *	@brief Initialize members of the state machine to defaults and values provided.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	This method initializes the J1939 state machine to defaults and the values provided.
	 *	Default for the current state is CAN_PROTOCOL_STATE_INITIALISATION.
	 *
	 *
	 *	@warning The state machine will only delete memory allocated by itself i.e. care has to be taken to free the pointers provided here.
	 */
	virtual int init() override;

	/**
	 *	@fn virtual int reset();
	 *
	 *	@brief Reset all class attributes with default values. Does not free allocated memory.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int reset() override;

	/**
	 *	@fn virtual int release();
	 *
	 *	@brief Frees allocated memory and sets all class attributes to default values.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@return WE_SUCCESS
	 */
	virtual int release() override;

	/**
	 *	@fn virtual int run(unsigned int p_time)
	 *
	 *	@brief The run function needs to be called frequently so that all tasks can be executed
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_time The tick count until which this function is allowed to run
	 *
	 *	@returns If everything went OK, WE_SUCCESS is returned. Everything else means error but is hard to define at the moment
	 *
	 *	This functions needs to be called frequently so that all SENDING tasks of the J1939 stack
	 *	can be triggered at the right time.
	 *	The receiving part is done event based by the EndPoint system.
	 */
	virtual int run(unsigned int p_time) override;


	/**
	 *	@fn virtual int handleCANmsg(WECANMessage * p_msg)
	 *
	 *	@brief Method to handle and distribute all incoming CAN messages.
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@param p_msg Pointer to the message received which should be evaluated/handled here.
	 *
	 *	@returns WE_SUCCESS if the received message could be handled (or the state machine is paused)
	 *	@returns WE_NOT_FOUND_ERROR if this message was not treated by any handler yet
	 *	@returns WE_NOT_SUPPORTED_ERROR if the received message isn't a valid J1939 message
	 *	@returns WE_NOT_SUPPORTED_ERROR if the J1939 instance this state machine belongs to does not have a valid address yet
	 *	@returns WE_NOT_SUPPORTED_ERROR was not intended for the J1939 instance this state machine belongs to
	 *
	 *	This function get's called whenever a CAN message is read from the CAN bus by WECANProtocol::onEndPointReadable().
	 *
	 *
	 *	@warning "p_msg" is only valid in this method - if (parts of) the message should be stored one has to copy them/it
	 *
	 *	@see WECANProtocol::onEndPointReadable()
	 *
	 */
	virtual int handleCANmsg(WECANMessage * p_msg) override;

	/**
	 *	@fn int getCurrentState(void)
	 *
	 *	@brief Return the current state of the state machine
	 *
	 *	@author hm@wachendorff.de
	 *	@author Hendrik Mueller
	 *
	 *	@returns the current state of the state machine
	 *
	 *	@see enum J1939States for available states
	 */
	int getCurrentState(void) const
	{
		return this->currentState;
	}


private:



protected:
	WEJ1939Resources	* resourcesJ1939{nullptr};	//!< pointer to data that will not change while stack is running and can be shared with state machine.
};
}

#endif /* WE_J1939_STATE_MACHINE_H_ */
