/*
 * we_eeprom_information.h
 *
 *  Created on: 02.03.2011
 *      Author: mka
 */

#ifndef WE_DEBUG_CONFIG_H_
#define WE_DEBUG_CONFIG_H_

#include <string>
#include <map>
#include <vector>
#include "../../ini_reader/include/ini_reader.h"

namespace WE
{

class WEDebugConfig
{
public:

	WEDebugConfig() = delete;
	virtual ~WEDebugConfig() = delete;

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WEDebugConfig(const WEDebugConfig &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WEDebugConfig & operator=(const WEDebugConfig &) = delete;

	static const std::string & getConfigFileName()
	{
		return WEDebugConfig::configFile;
	}
	//set configfile with full path


	/**
	 *	@fn static bool setConfigFileName(std::string p_name)
	 *
	 *	@brief Set config file now.
	 *
	 *	@param Path and name of config file
	 *
	 *	@returns true: file was found, false otherwise
	 *
	 *	If this is called after WEAppication has been created "bool readConfig()" has to be
	 *	called afterwards in order for the settings to be used.
	 *
	 *	It is suggested to either accept the default ini file or to set the file
	 *	before WEApplication is created.
	 *
	 *	@see bool readConfig()
	 */
	static bool setConfigFileName(const std::string & p_name);


	/**
	 *	@fn static bool readConfig()
	 *
	 *	@brief Read the settings for debug sections from config file now.
	 *
	 *	@param None
	 *
	 *	@returns true: success, false otherwise
	 *
	 *	This method is called by constructor of WEApplication().
	 *	It has to be called whenever file is modified by "bool setConfigFileName(std::string p_name)".
	 *
	 *	It is suggested to either accept the default ini file or to set the file
	 *	before WEApplication is created.
	 *
	 *	@see bool setConfigFileName(std::string p_name)
	 */
	static bool readConfig();


	/**
	 *	@fn unsigned int getIniVal(int p_key)
	 *
	 *	@brief get value for section
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param p_indexKey key value (index) for the according section
	 *
	 *	@returns value according to key, if error it returns 0 so nothing is done.
	 */
	static unsigned int getIniVal(int p_indexKey);

	/**
	 *	@fn int getLogLevel()
	 *
	 *	@brief get current log level
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param n/a
	 *
	 *	@returns log level or 0
	 */
	static int getLogLevel();

	static int translateConfigStringToInt(const std::string & p_name);

	/**
	 *	@fn int insertNewDebugSection(std::string p_section)
	 *
	 *	@brief inserts a new section to the map
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param p_section string which has to be the same like inside the ini file.
	 *
	 *	@returns current key value. If the p_section already exists it returns the already found key if error it returns -1
	 */
	static int insertNewDebugSection(const std::string & p_section);

	/**
	 *	@fn int getDebugSectionKey(std::string p_section)
	 *
	 *	@brief finds the according key to the section
	 *
	 *	@author mka@wachendorff.de
	 *	@author Markus Kalin
	 *
	 *	@param p_section string which has to be the same like inside the ini file.
	 *
	 *	@returns the key value or -1 if not found
	 */
	static int getDebugSectionKey(const std::string & p_section);

	static int logToConsole();
	static int logToFile();
	static int getMaxFileSize();
	static int getLogContinue();

	inline static const bool & isConfigFileOK(void)
	{
		return WEDebugConfig::iniFileOk;
	}

private:
	static inline IniReader theIniFile{};
	static inline bool iniFileOk{false};

	static inline std::string configFile{"./debug.ini"};

	static inline std::multimap<std::string, int> theConfigMap{};	//!< map name of debug section to its internal ID

	static inline std::vector<int> theValueVector{};		//!< index: internal ID of section, value: value defined in debug.ini for this section
	static inline int sectionIndex{0};				//!< number of debug sections inserted so far
	static inline int generalLogLevel{-1};
	static inline int toConsole{-1};
	static inline int toFile{-1};
	static inline int maxFileSize{24599};
	static inline int continousLogging{-1};

	static inline const std::string WE_DEBUG_INI_SECTION{"debug"};
};
} // namespace WE

#endif /* WE_DEBUG_CONFIG_H_ */
