/*
 * ini_reader.h
 *
 *  Created on: Sep 18, 2020
 *      Author: dsterzel
 */

#ifndef SRC_UTILITY_INI_READER_INCLUDE_INI_READER_H_
#define SRC_UTILITY_INI_READER_INCLUDE_INI_READER_H_

#include <map>

namespace WE {

/**
 * Simple class to read a .ini file and store its content in memory. Values can
 * be accessed via getValue() after loadFile() was called (and returned true).
 * If the used ini file contains sections (e.g. [SECTION_NAME]), use getValue()
 * with p_sectionName parameter. If absolutely no sections are included in th ini file
 * use the getValue without section identifier;
 */
class IniReader
{
public:
	/**
	 * Constructor
	 */
	IniReader() = default;

	/**
	 * Destructor
	 */
	~IniReader();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	IniReader(const IniReader &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	IniReader & operator=(const IniReader &) = delete;

	/**
	 * Load and parse the ini file. This needs to be called before getValue() can be used.
	 * @param p_filePath path to the file to load
	 * @return true if file was loaded (even if file is empty), false if file could not be found or opened for reading
	 */
	bool loadFile(const std::string & p_filePath);

	/**
	 * Clear loaded content. After this was called, the ini file needs to be laoded again with loadFile()
	 */
	void clearLoadedContent(void);

	/**
	 * @return true if ini file was loaded with loadFile(), false if no file was loaded
	 */
	bool isFileLoaded(void) const;

	/**
	 * Get a value from loaded ini file. Use this functions if sections are used in the ini file.
	 * If e.g. a section [MY_SECTION] is contained in the ini file, provide "MY_SECTION" as first parameter.
	 *
	 * @param p_sectionName Name of the section
	 * @param p_id The identifier of the value to get (e.g. HEIGHT=100 => Provide "HEIGHT" as parameter)
	 * @param p_defaultValue If value with p_id was not found in section p_sectionName, this value will be returned. If omitted, an empty string will be returned.
	 *
	 * @return String representation of the value identified by p_id inside p_sectionName. If p_id was not found, p_defaultValue will be returned.
	 */
	std::string getValue(const std::string & p_sectionName, const std::string & p_id, const std::string & p_defaultValue = "") const;

	/**
	 * Get a value from loaded ini file. Use this functions if NO sections are used in the ini file.
	 *
	 * @param p_id The identifier of the value to get (e.g. HEIGHT=100 => Provide "HEIGHT" as parameter)
	 * @param p_defaultValue If value with p_id was not found, this value will be returned. If omitted, an empty string will be returned.
	 *
	 * @return String representation of the value identified by p_id. If p_id was not found, p_defaultValue will be returned.
	 */
	std::string getValue(const std::string & p_id, const std::string & p_defaultValue = "") const;
private:
	std::map<std::string, std::string> noSectionValues{};	//!< If ini file contains no sections, all values will be stored here
	std::map<std::string, std::map<std::string, std::string>* > sectionValues{}; //!< If ini file contains sections, every section has its own map
	bool fileLoaded{false};	//!< flag to tell if ini file was loaded
};
}  // namespace WE



#endif /* SRC_UTILITY_INI_READER_INCLUDE_INI_READER_H_ */
