/**
 *	@file we_string_tokenizer.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2010
 *
 *	@brief A simple string tokenizer class
 *
 */

#ifndef WE_STRING_TOKENIZER_H_
#define WE_STRING_TOKENIZER_H_

#include <string>
#include <vector>

namespace WE
{

/**
 *	@class WEStringTokenizer
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Small class that provides functionality to split a string into several single tokens
 *
 *	@namespace WE
 */
class WEStringTokenizer
{
public:
	WEStringTokenizer() = delete;
	~WEStringTokenizer() = delete;

	/**
	 *	@fn static std::vector<std::string> tokenize(const std::string& p_str,const std::string& p_delimiters)
	 *
	 *	@brief Split a string into tokens delimited by one or more delimiter characters
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_str String that shall be tokenized
	 *	@param p_delimiters A String that's single characters will be used as delimiters for the string
	 *
	 *	@returns A vector containing strings. The strings in the vector are the tokens taken from p_str
	 *
	 *	To tokenize a string, just pass the string that shall be tokenized and another string that
	 *	contain at least one character that shall be used as delimiter.
	 *	The single tokens extracted in of the string will be stored in the returned vector.
	 *	If the delimiter is not found in the string, the returned vector will contain the original string.
	 *	If there is no string between two delimiters, an empty string will be placed in the returned vector at that position
	 *	(e.g.: string: "This;is;;a;test", delimiter: ";", result: vector[0] = "This", vector[1] = "is", vector[2] = "", vector[3] = "a", vector[5] = "test").
	 */
	static std::vector<std::string> tokenize(const std::string& p_str,const std::string& p_delimiters);

private:
};
} // namespace WE

#endif /* WE_STRING_TOKENIZER_H_ */
