/**
 *	@file we_timer.h
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@copyright ( c ) Topcon Electronics GmbH & Co. KG 2011
 *
 *	@brief Definition of WETimer class
 *
 */


#ifndef WE_TIMER_H_
#define WE_TIMER_H_

#include <map>

namespace WE {

class WETick;
class WESignalListener;

typedef enum
{
	WE_TIMER_SINGLE_SHOT,
	WE_TIMER_REPEATING
}WETimerType;

/**
 *	@struct RepeatTimeout
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Small struct to store timeout and interval of repeating timeouts
 *
 *	@namespace WE
 */
struct RepeatTimeout
{
	RepeatTimeout(unsigned int p_timeout, unsigned int p_interval)
	{
		this->timeout = p_timeout;
		this->interval = p_interval;
	}
	RepeatTimeout()
	{
		this->timeout = 0;
		this->interval = 0;
	}
	unsigned int timeout;
	unsigned int interval;
};

/**
 *	@class WETimer
 *
 *	@author dsl@wachendorff.de
 *	@author Dirk Sterzel
 *
 *	@brief Singleton class for timer usage
 *
 *	@namespace WE
 *
 *	WEtimer can be used to install either single shot or repeating timers.
 *	A TIMEOUT_EVENT will be sent to the event listener when timeout occurred.
 *	It is only allowe for each event listener to install one timeout of each type.
 *	The checkTimeouts method needs to be called frequently to check the timeouts.
 *	WETimer is a singleton.
 */
class WETimer
{
public:
	/**
	 *	@fn ~WETimer()
	 *
	 *	@brief Destructor. Clear maps and delete all timeouts.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	virtual ~WETimer();

	/**
	 * Deleted copy constructor. If we need to copy this class please implement this.
	 */
	WETimer(const WETimer &) = delete;

	/**
	 * Deleted assignment operator. If we need to copy this class please implement this.
	 */
	WETimer & operator=(const WETimer &) = delete;

	/**
	 *	@fn static WETimer * instance()
	 *
	 *	@brief Retreive the one and only instance of this class
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns Pointer to the one and only instance of this class.
	 */
	static WETimer * instance();

	/**
	 *	@brief Install a timeout
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_interval Timeout in milliseconds (e.g. 100 if the timeout shall occur in 100ms from now on)
	 *	@param p_type. Either single shot timeout (WE_TIMER_SINGLE_SHOT) or a repeating one (WE_TIMER_REPEATING)
	 *	@param p_listener. Pointer to a WESignalListener set gets a TIMEOUT_EVENT as soon as the timeout occurred.
	 *
	 *	@returns WE_SUCCESS if the timeout was installed successfully
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_listener is nullptr
	 *	@returns WE_TYPE_ERROR if a wrong type was provided
	 *
	 *	@warning It can only be one timeout type per listener be installed. Existing ones get overwritten.
	 */
	int installTimer(unsigned int p_interval, WETimerType p_type, WESignalListener * p_listener);

	/**
	 *	@brief Remove p_type timeout belongign to p_listener
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@param p_type Type of timeout to remove
	 *	@param p_listener. The listener of the timeout that shall be removed
	 *
	 *	@returns WE_SUCCESS If the timer was removed (or not found)
	 *	@returns WE_INVALID_PARAMETER_ERROR if p_listener is nullptr
	 *	@returns WE_TYPE_ERROR if a wrong type was provided
	 */
	int removeTimer(WETimerType p_type, WESignalListener * p_listener);

	/**
	 *	@fn int checkTimeouts(void)
	 *
	 *	@brief check all stored timeouts
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 *
	 *	@returns WE_SUCCESS
	 *
	 *	Checks all stored timeouts if they are expired. If yes, call the listener's
	 *	signalOccurred function with a WE::TIMEOUT_EVENT event.
	 */
	int checkTimeouts(void);

	/**
	 * Function to remove ALL timers
	 */
	void removeAllTimers(void);
private:
	/**
	 *	@fn WETimer()
	 *
	 *	@brief Private constructor. Only retrieves WETick pointer.
	 *
	 *	@author dsl@wachendorff.de
	 *	@author Dirk Sterzel
	 */
	WETimer() = default;

	std::map< WESignalListener *, unsigned int > singleShotTimers;	//!< All stored single shot timer
	std::map< WESignalListener *, RepeatTimeout > repeatingTimers;	//!< All stored repeating timer
};
}  // namespace WE


#endif /* WE_TIMER_H_ */
