/*
 * hardware_daemon_ipc_example.cpp
 *
 *  Created on: Jul 5, 2019
 *      Author: dsl
 */

#include <unistd.h> //unlink

//socket includes
#include <sys/socket.h>
#include <sys/un.h>

#include <errno.h>
#include <string.h> //various stuff like strlen, strerror

#include <stdio.h>
#include <stdlib.h>

#include <iostream> //cout

//make sure that API_PATH is correctly defined in Makefile
#include <dei_ipc_functions.h>


void handleMessage(IPCMessage * msg)
{
	if (nullptr != msg)
	{
		IPCData data;
		dataFromBuffer(msg->data, msg->header.dataLength, &data);
		switch(msg->header.messageType)
		{
			case VARIABLE_UPDATE_EVENT:
				//FALLTHROUGH
			case VARIABLE_GET_RESPONSE:
				//FALLTHROUGH
			case EVENT_SUBSCRIPTION_RESPONSE:
				switch(data.index)
				{
					case 0x2020:
					{
						switch(data.subIndex)
						{
							case 0x08:
							{
								std::cout << "Running on OS " << data.data.stringValue.value << std::endl;
							}
							break;
						}
					}
					break;
					case 0x2100:
					{
						switch(data.subIndex)
						{
							case 0x07:
								std::cout << "@RTCSeconds: " << data.data.intValue << std::endl;
								break;

						}
					}
					break;
					case 0x2130:
						switch(data.subIndex)
						{
							case 0x02:
								std::cout << "@CurrentFile: " << data.data.stringValue.value << std::endl;
								break;
						}
						break;
				}
				break;

			default:
				break;
		}
	}
	//freeMessageData(msg);
}

int main(void)
{
	//NOTE:
	//We are trying to stay compatible with older C++ versions in this example (e.g. OPUS A3/A6 G1).
	//We also try to use "standard" Linux c/c++ functions with the least possible dependencies to other libraries.


	//define the address of this process
	std::string myAddress = "/tmp/hwdTest";
	//append process-ID to the address so it changes at each start (if this was c++11, std::to_string() can be used instead of snprintf)
	char pid[12];
	sprintf(pid, "%d", getpid());
	myAddress += pid;


	//address information for hwd
	//will be filled by createSocket()
	struct sockaddr_un socketAddressHWD;
	//hardware daemon listens on socket /tmp/hardwareDaemon
	int hwdSocket = createSocket(myAddress, "/tmp/hardwareDaemon", &socketAddressHWD);

	//now we can try to connect
	//this will block
	//if it shall not block, you have to set the socket to non-blocking (see man pages)...
	int ret = connect(hwdSocket, (struct sockaddr*) &socketAddressHWD, sizeof(socketAddressHWD));
	if (0 == ret)
	{
		std::cout << "CONNECT SUCCESSFULL" << std::endl;
		//(Note that this is only an example and error handling can be made a lot better!)
		//Now that we are connected, we can send out requests...
		//Let's just subscribe for value changes of @RTCSeconds.
		ret = subscribeToVariable(hwdSocket, 0x2100, 0x07, DATA_TYPE_INT, 100);
		//Now the same but for the string variable @CurrentFile
		ret = subscribeToVariable(hwdSocket, 0x2130, 0x02, DATA_TYPE_STRING, 100);

		//request the value for @OSKernelVersion just once (without subscription for value changes)
		ret = requestValue(hwdSocket, 0x2020, 0x08, DATA_TYPE_STRING, 100);

		//if last sending was OK
		if (ret > 0)
		{
			//Create a message used for reception
			struct IPCMessage msg;

			//sending was successful
			std::cout << "Sending request was successful!" << std::endl;

			int i = 0;
			//now we can wait for responses
			//this loop can be part of an application's main loop or can be
			//moved to a thread (make sure to synchronize properly)
			while(1)
			{
				//wait max.  1000ms for a message from hardware daemon to be received...
				ret = receiveMessage(hwdSocket, &msg, 1000);
				if (ret >= 0)
				{
					//std::cout << "Received message type " << msg.messageType << " with data length " << msg.dataLength << std::endl;
					//a message was received, pass it to handleMessage
					//(modify the handling inside handleMessage as it fits for the application)
					handleMessage(&msg);
					//free the memory used by the message
					//freeMessageData(&msg);
				}
				else if (ret == -7)
				{
					//only timeout
					//=> no messages received within the given timeout
				}
				else
				{
					//check receiveMessage() return codes!
					std::cout << "RECEIVING FAILED! EXIT " << ret << std::endl;
					break;
				}


				//lets test setting a value every 10 loops (this does not mean every 10 seconds!)
				//(we are trying to change @RTCSeconds, so we immediately get an updated value from hardware daemon since we subscribed to that variable)
				if (i % 10 == 0)
				{
					//std::cout << i << ": Request change to 0" << std::endl;
					requestValueChange(hwdSocket, 0x2100, 0x07, 0, 50);
				}
				++i;
			}
		}
		else
		{
			std::cout << "Sending message to hardware daemon failed! EXIT " << ret << std::endl;
		}
	}
	else
	{
		std::cout << "FAILED TO CONNECT!" << std::endl;
	}

	std::cout << "CLEAN UP" << std::endl;

	close(hwdSocket);



	return ret;
}
