/*
 * dei_ipc_data.h
 *
 * This header file contains useful functions to make it easier to communicate with
 * a WEapi DEI application via IPC (such as hardware daemon or PClient).
 * Simply include the dei_ipc_functions.h in your application code and add this file to the
 * source files of your application.
 * General function documentation see header file.
 *
 * Copyright 2019 Topcon Electronics GmbH & Co. KG
 *
 * This API and it's functions are provided as is in source code without any guarantee.
 * You are allowed to use this code to transfer data from any custom application to a WEapi
 * DEI application such as PClient or hardware daemon. The sources are provided as source code
 * to make custom application connections easier and to reduce dependencies to other libraries
 * (you don't need to link to anything, just use the headers and source files in your application).
 * Although most of the implementation of this API uses C code (because of Linux system calls), you need to use a
 * C++ compiler to use this API.
 * Please do not modify the data parts of the API since this might break compatibility with existing applications.
 * You are free to modify the function code to your liking (e.g. memory management, return values, etc.).
 * You are however NOT free to spread this code anywhere. Only for direct Topcon Electronics customers.
 * Future releases of WEapi and PClient may use an updated version of this IPC API. Make sure to check
 * for changes and update your application accordingly.
 */

#ifndef _WIN32
#include <sys/socket.h>
#include <sys/un.h>
#endif
#include <unistd.h>	//unlink

#include "dei_ipc_functions.h"


int createSocket(const std::string & p_myAddress, const std::string & p_serverAddress, struct sockaddr_un * p_serverAddressStruct)
{
#ifndef _WIN32
	//create a socket
	int sock = socket(AF_UNIX, SOCK_STREAM, 0);

	//we need to provide a name for our side of the socket so that the "server" socket can distinguish between different connections
	struct sockaddr_un socketAddress{};
	memset(&socketAddress, 0, sizeof(socketAddress));
	socketAddress.sun_family = AF_LOCAL;
	strncpy(socketAddress.sun_path, p_myAddress.c_str(), p_myAddress.length());

	//give our socket an address via bind
	if (0 == bind(sock,  (struct sockaddr*) &socketAddress, sizeof(socketAddress)))
	{
		//since we can consider this as the private part of the connection, we can remove (unlink) it (won't show on file system)
		unlink(p_myAddress.c_str());

		//address information for hwd
		memset(p_serverAddressStruct, 0, sizeof(*p_serverAddressStruct));
		p_serverAddressStruct->sun_family = AF_LOCAL;
		strncpy(p_serverAddressStruct->sun_path, p_serverAddress.c_str(), p_serverAddress.length());

		return sock;
	}
	else
	{
		//bind to myAddress failed!
		close(sock);
		return -1;
	}
#else
	static_cast<void>(p_myAddress);
	static_cast<void>(p_serverAddress);
	static_cast<void>(p_serverAddressStruct);
	return -1;
#endif
}

int createSocketAndConnect(const std::string & p_myAddress, const std::string & p_serverAddress)
{
#ifndef _WIN32
	struct sockaddr_un serverSocketAddress{};
	int sock = createSocket(p_myAddress, p_serverAddress, &serverSocketAddress);
	if (sock > 0)
	{
		if (0 == connect(sock, (struct sockaddr*) &serverSocketAddress, sizeof(serverSocketAddress)))
		{
			return sock;
		}
		else
		{
			return -1;
		}
	}
	else
	{
		return -2;
	}
#else
	static_cast<void>(p_myAddress);
	static_cast<void>(p_serverAddress);
	return -1;
#endif
}

int sendMessage(int p_sock, IPCMessage* p_message, int p_timeoutMS)
{
	int returnStatus = 0;
#ifndef _WIN32
	if (p_sock >= 0 &&
		nullptr != p_message &&
		0 != p_message->header.dataLength &&
		nullptr != p_message->data)
	{
		struct timeval sendTimeout{};
		fd_set sendFDset{};

		// initialize the sendTimeout to be used with select
		sendTimeout.tv_sec = p_timeoutMS / 1000;
		sendTimeout.tv_usec = (p_timeoutMS - (sendTimeout.tv_sec * 1000)) * 1000;

		// initialize the sendFDset to be used with select
		FD_ZERO(&sendFDset);
		FD_SET(p_sock, &sendFDset);

		int messageLength = messageHeaderLength + p_message->header.dataLength;

		// set the "Send low water mark" as the message length, so that select waits
		// for the specified number of bytes to be available in the socket send buffer before it can return
		setsockopt(p_sock, SOL_SOCKET, SO_SNDLOWAT, &messageLength, sizeof(messageLength));

		if (p_timeoutMS < 0)
		{
			// wait forever till socket becomes writable
			returnStatus = select(p_sock + 1, nullptr, &sendFDset, nullptr, nullptr);
		}
		else
		{
			// wait till the socket becomes writable or the sendTimeout expires
			returnStatus = select(p_sock + 1, nullptr, &sendFDset, nullptr, &sendTimeout);
		}
		if (returnStatus > 0)
		{
			returnStatus = 0;
			// socket is writable
			if (FD_ISSET(p_sock, &sendFDset))
			{
				// Construct the message payload so that the complete message can be
				// sent at one go rather than sending header bytes and data bytes
				// separately in two send() calls.
				char* messagePayload = new char[messageLength];
				memset(messagePayload, '\0', messageLength);
				memcpy(messagePayload, p_message, messageHeaderLength);
				memcpy(messagePayload + messageHeaderLength, p_message->data, p_message->header.dataLength);

				int bytesSent = 0;
				// Try to send the complete message at one go. It should be successful
				// after the select returned successfully.
				// Also disable SIGPIPE on the socket to avoid abnormal termination
				// of the application in case of other end being closed before this.
				// Handle this situation by checking errno for EPIPE.
				bytesSent = send(p_sock, messagePayload, messageLength, MSG_NOSIGNAL);

				//Release the message payload
				delete [] messagePayload;
				messagePayload = nullptr;

				// Even then if complete message could not be sent, return error.
				if (bytesSent != messageLength)
				{
					if (bytesSent < 0)
					{
						if (errno == EPIPE)
						{
							//the connection is broken now so close the socket to free the system resources
							//after this the endpoint has to initialized and connected again before sending/receiving
							//more messages.
							returnStatus = -1;
						}
						else
						{
							returnStatus = -2;
						}
					}
					else
					{
						if (0 == bytesSent)
						{
							//the connection is broken now so close the socket to free the system resources
							//after this the endpoint has to initialized and connected again before sending/receiving
							//more messages.
							returnStatus = -3;
						}
						else
						{
							returnStatus = -4;
						}
					}
				}
				else
				{
					//full success
					returnStatus = bytesSent;
				}
			}
		}
		else
		{
			if (0 == returnStatus)
			{
				returnStatus = -5;
			}
			else
			{
				returnStatus = -6;
			}
		}
	}
	else
	{
		if (p_sock < 0)
		{
			returnStatus = -7;
		}
		else if (nullptr == p_message)
		{
			returnStatus = -8;
		}
		else if (0 == p_message->header.dataLength)
		{
			returnStatus = -9;
		}
		else if (nullptr == p_message->data)
		{
			returnStatus = -10;
		}
	}
#else
	static_cast<void>(p_sock);
	static_cast<void>(p_message);
	static_cast<void>(p_timeoutMS);
#endif
	return returnStatus;
}

int receiveMessage(int p_sock, IPCMessage *p_message, long p_timeoutMS)
{
	int returnStatus = 0;
#ifndef _WIN32
	if (p_sock >= 0 &&
		nullptr != p_message)
	{
		struct timeval receiveTimeout{};
		fd_set receiveFDset{};

		// initialize the receiveTimeout to be used with select
		receiveTimeout.tv_sec = p_timeoutMS / 1000;
		receiveTimeout.tv_usec = (p_timeoutMS - (receiveTimeout.tv_sec * 1000)) * 1000;

		// initialize the receiveFDset to be used with select
		FD_ZERO(&receiveFDset);
		FD_SET(p_sock, &receiveFDset);

		int messageLength = messageHeaderLength + p_message->header.dataLength;

		// set the "Receive low water mark" as the message length, so that select waits
		// for the specified number of bytes to be available in the socket receive
		// buffer before it can return
		setsockopt(p_sock, SOL_SOCKET, SO_RCVLOWAT, &messageLength, sizeof(messageLength));



		if (p_timeoutMS < 0)
		{
			// wait forever till socket becomes readable
			returnStatus = select(p_sock + 1, &receiveFDset, nullptr, nullptr, nullptr);
		}
		else
		{
			// wait till the socket becomes readable or the receiveTimeout expires
			returnStatus = select(p_sock + 1, &receiveFDset, nullptr, nullptr, &receiveTimeout);
		}

		if (returnStatus > 0)
		{
			returnStatus = 0;
			// socket is readable
			if (FD_ISSET(p_sock, &receiveFDset))
			{
				int headerBytesReceived = 0;
				// Try to receive the message header first. Should be successful after
				// the select returns successfully.
				headerBytesReceived = recv(p_sock, p_message, messageHeaderLength, 0);
				if (headerBytesReceived == messageHeaderLength)
				{
					int dataBytesReceived = 0;
					//Calculate the originally supplied data length
					int originalDataLength = messageLength - messageHeaderLength;
					// If the data length in the received message header is less
					// than equal to the original data length then use the same
					// message->data memory.
					if ((int)p_message->header.dataLength <= originalDataLength)
					{
						memset(p_message->data, '\0', originalDataLength);
					}
					// Otherwise deallocate the original message->data pointer
					// and allocate new memory as the original memory was not
					// sufficient enough.
					else
					{
						if (nullptr != p_message->data)
						{
							delete (char*)p_message->data;
							p_message->data = nullptr;
						}
						//make sure that caller will delete this memory!
						p_message->data = new unsigned char[p_message->header.dataLength];
						memset(p_message->data, '\0', p_message->header.dataLength);
					}

					dataBytesReceived = recv(p_sock, p_message->data, p_message->header.dataLength,0);
					if (dataBytesReceived != (int)p_message->header.dataLength)
					{
						if (dataBytesReceived < 0)
						{
							returnStatus = -1;
						}
						else if (0 == dataBytesReceived)
						{
							//the connection is broken now so close the socket to free the system resources
							//after this the endpoint has to initialized and connected again before sending/receiving
							//more messages.
							returnStatus = -2;
						}
						else
						{
							returnStatus = -3;
						}
					}
					else
					{
						//full success
						returnStatus = dataBytesReceived + messageHeaderLength;
					}
				}
				// if complete message header could not be received, return error.
				else
				{
					if (headerBytesReceived < 0)
					{
						returnStatus = -4;
					}
					else if (0 == headerBytesReceived)
					{
						//the connection is broken now so close the socket to free the system resources
						//after this the endpoint has to initialized and connected again before sending/receiving
						//more messages.
						returnStatus = -5;
					}
					else
					{
						returnStatus = -6;
					}
				}
			}
		}
		else
		{
			if (0 == returnStatus)
			{
				returnStatus = -7;
			}
			else
			{
				returnStatus = -8;
			}
		}
	}
	else
	{
		if (p_sock < 0)
		{
			returnStatus = -9;
		}
		else if (nullptr == p_message)
		{
			returnStatus = -10;
		}
		else if (0 == p_message->header.dataLength)
		{
			returnStatus = -11;
		}
		else if (nullptr == p_message->data)
		{
			returnStatus = -12;
		}
	}
#else
	static_cast<void>(p_sock);
	static_cast<void>(p_message);
	static_cast<void>(p_timeoutMS);
#endif
	return returnStatus;
}

int subscribeToVariable(int p_sock, unsigned short p_index, unsigned short p_subindex, data_type p_dataType, int p_timeoutMS)
{
	int ret = 0;
	if (p_sock > 0)
	{
		//and the data part (which is of type IPCEventData for subscription requests)
		struct IPCData data;
		//The IPC message to send
		struct IPCMessage msg;

		//now set the options (for index and subindex see variable table in Projektor and the user manual variable reference)
		data.index = p_index;				//!< the index of the variable group
		data.subIndex = p_subindex;				//!< the subindex of the variable
		data.type = p_dataType;
		//now the rest of the message
		msg.header.messageType = EVENT_SUBSCRIPTION_REQUEST;	//!< subscribe
		msg.header.dataLength = getBufferSize(&data);			//!< size of the data pointer in bytes
		msg.data = new unsigned char[msg.header.dataLength];		//!< reserve space for data
		//and assign the data
		convertToBuffer(msg.data, &data);
		//send
		ret = sendMessage(p_sock, &msg, p_timeoutMS);
	}

	return ret;
}

int requestValueChange(int p_sock, unsigned short p_index, unsigned short p_subindex, int p_newValue, int p_timeoutMS)
{
	int ret = 0;
	if (p_sock > 0)
	{
		IPCData dat;
		dat.index = p_index;
		dat.subIndex = p_subindex;
		dat.type = DATA_TYPE_INT;
		dat.data.intValue = p_newValue;
		IPCMessage msg;
		msg.header.messageType = VARIABLE_SET_REQUEST;
		msg.header.messageStatus = true;	// we don't want to make the value 'invalid'
		msg.header.dataLength = getBufferSize(&dat);
		msg.data = new unsigned char[msg.header.dataLength];
		if (convertToBuffer(msg.data, &dat) > 0)
		{
			ret = sendMessage(p_sock, &msg, p_timeoutMS);
		}
	}
	return ret;
}

int requestValueChange(int p_sock, unsigned short p_index, unsigned short p_subindex, const std::string & p_newValue, int p_timeoutMS)
{
	int ret = 0;
	if (p_sock > 0)
	{
		IPCData dat;
		dat.index = p_index;
		dat.subIndex = p_subindex;
		dat.type = DATA_TYPE_STRING;
		dat.data.stringValue.length = p_newValue.length();
		dat.data.stringValue.value = new unsigned char[dat.data.stringValue.length];
		strncpy((char*)dat.data.stringValue.value, p_newValue.c_str(), dat.data.stringValue.length);
		IPCMessage msg;
		msg.header.messageType = VARIABLE_SET_REQUEST;
		msg.header.messageStatus = true;	// we don't want to make the value 'invalid'
		msg.header.dataLength = getBufferSize(&dat);
		msg.data = new unsigned char[msg.header.dataLength];
		if (convertToBuffer(msg.data, &dat) > 0)
		{
			ret = sendMessage(p_sock, &msg, p_timeoutMS);
		}
	}
	return ret;
}

int requestValue(int p_sock, unsigned short p_index, unsigned short p_subindex, data_type p_dataType, int p_timeoutMS)
{
	int ret = 0;
	if (p_sock > 0)
	{
		IPCData dat;
		dat.index = p_index;
		dat.subIndex = p_subindex;
		dat.type = p_dataType;
		dat.data.stringValue.length = 0;

		IPCMessage msg;
		msg.header.messageType = VARIABLE_GET_REQUEST;
		msg.header.dataLength = getBufferSize(&dat);
		msg.data = new unsigned char[msg.header.dataLength];
		if (convertToBuffer(msg.data, &dat) > 0)
		{
			ret = sendMessage(p_sock, &msg, p_timeoutMS);
		}
	}
	return ret;
}

unsigned int getBufferSize(IPCData * p_data)
{
	if (nullptr != p_data)
	{
		unsigned int size = 2 + 2 + 1; //index + subindex + type
		size += 4; //either int value or string length
		if (p_data->type == DATA_TYPE_STRING)
		{
			size += p_data->data.stringValue.length;
			size += 1; //for terminating 0
		}

		return size;
	}
	else
	{
		return 0;
	}
}

unsigned int convertToBuffer(unsigned char * p_buffer, IPCData * p_ipcData)
{
	unsigned int i = 0;
	if (p_buffer != nullptr && p_ipcData != nullptr && p_ipcData->type != DATA_TYPE_INVALID)
	{
		p_buffer[i++] = p_ipcData->index & 0xff;
		p_buffer[i++] = (p_ipcData->index >> 8) & 0xff;
		p_buffer[i++] = p_ipcData->subIndex & 0xff;
		p_buffer[i++] = (p_ipcData->subIndex >> 8) & 0xff;
		p_buffer[i++] = p_ipcData->type & 0xff;
		if (p_ipcData->type == DATA_TYPE_INT)
		{
			p_buffer[i++] = (p_ipcData->data.intValue >>  0) & 0xff;
			p_buffer[i++] = (p_ipcData->data.intValue >>  8) & 0xff;
			p_buffer[i++] = (p_ipcData->data.intValue >> 16) & 0xff;
			p_buffer[i++] = (p_ipcData->data.intValue >> 24) & 0xff;
		}
		else
		{
			unsigned int length = p_ipcData->data.stringValue.length + 1; //+1 for terminating 0
			p_buffer[i++] = (length >>  0) & 0xff;
			p_buffer[i++] = (length >>  8) & 0xff;
			p_buffer[i++] = (length >> 16) & 0xff;
			p_buffer[i++] = (length >> 24) & 0xff;
			for (size_t j = 0; j < length - 1; ++j)
			{
				p_buffer[i++] = p_ipcData->data.stringValue.value[j];
			}
			p_buffer[i++] = 0;
		}
	}
	return i;
}

unsigned int dataFromBuffer(unsigned char * p_buffer, unsigned int p_bufferLength, IPCData * p_data)
{
	unsigned int ret = 0;
	if (nullptr != p_buffer && nullptr != p_data && p_bufferLength >= POSITION_STRING_VALUE_START)
	{
		p_data->index = p_buffer[POSITION_INDEX_MSB];
		p_data->index = p_data->index << 8;
		p_data->index |= (p_buffer[POSITION_INDEX_LSB] & 0xff);
		p_data->subIndex = p_buffer[POSITION_SUBINDEX_MSB];
		p_data->subIndex = p_data->subIndex << 8;
		p_data->subIndex |= p_buffer[POSITION_SUBINDEX_LSB];
		p_data->type = (data_type)p_buffer[POSITION_TYPE];
		if (p_data->type == DATA_TYPE_INT)
		{
			p_data->data.intValue = p_buffer[POSITION_INT_OR_STRING_LENGTH_MMSB];
			p_data->data.intValue = p_data->data.intValue << 8;
			p_data->data.intValue |= p_buffer[POSITION_INT_OR_STRING_LENGTH_MSB];
			p_data->data.intValue = p_data->data.intValue << 8;
			p_data->data.intValue |= p_buffer[POSITION_INT_OR_STRING_LENGTH_LSB];
			p_data->data.intValue = p_data->data.intValue << 8;
			p_data->data.intValue |= p_buffer[POSITION_INT_OR_STRING_LENGTH_LLSB];
			ret = POSITION_STRING_VALUE_START;
		}
		else
		{
			p_data->data.stringValue.length = p_buffer[POSITION_INT_OR_STRING_LENGTH_MMSB];
			p_data->data.stringValue.length = p_data->data.stringValue.length << 8;
			p_data->data.stringValue.length |= p_buffer[POSITION_INT_OR_STRING_LENGTH_MSB];
			p_data->data.stringValue.length = p_data->data.stringValue.length << 8;
			p_data->data.stringValue.length |= p_buffer[POSITION_INT_OR_STRING_LENGTH_LSB];
			p_data->data.stringValue.length = p_data->data.stringValue.length << 8;
			p_data->data.stringValue.length |= p_buffer[POSITION_INT_OR_STRING_LENGTH_LLSB];

			if (p_data->data.stringValue.length > 0
				&& p_bufferLength >= (p_data->data.stringValue.length + POSITION_STRING_VALUE_START))
			{
				if (p_data->data.stringValue.value != nullptr)
				{
					delete [] p_data->data.stringValue.value;
				}
				p_data->data.stringValue.value = new unsigned char[p_data->data.stringValue.length];

				for (unsigned int i = 0; i < p_data->data.stringValue.length; ++i)
				{
					p_data->data.stringValue.value[i] = p_buffer[POSITION_STRING_VALUE_START + i];
				}
				ret = p_data->data.stringValue.length + POSITION_STRING_VALUE_START;
			}
		}
	}
	return ret;
}


