/*
 * dei_ipc_functions.h
 * This header file contains useful functions to make it easier to communicate with
 * a WEapi DEI application via IPC (such as hardware daemon or PClient).
 *
 * Copyright 2019 Topcon Electronics GmbH & Co. KG
 *
 * This API and it's functions are provided as is in source code without any guarantee.
 * You are allowed to use this code to transfer data from any custom application to a WEapi
 * DEI application such as PClient or hardware daemon. The sources are provided as source code
 * to make custom application connections easier and to reduce dependencies to other libraries
 * (you don't need to link to anything, just use the headers and source files in your application).
  * Although most of the implementation of this API uses C code (because of Linux system calls), you need to use a
 * C++ compiler to use this API.
 * Please do not modify the data parts of the API since this might break compatibility with existing applications.
 * You are free to modify the function code to your liking (e.g. memory management, return values, etc.).
 * You are however NOT free to spread this code anywhere. Only for direct Topcon Electronics customers.
 * Future releases of WEapi and PClient may use an updated version of WEapi and the IPC API. Make sure to check
 * for changes and update your application accordingly.
 */

#ifndef SRC_DEI_DEI_IPC_API_DEI_IPC_FUNCTIONS_H_
#define SRC_DEI_DEI_IPC_API_DEI_IPC_FUNCTIONS_H_

#include <string>

//Make sure that the -I include path to the directory where these header are is set correctly when compiling
#include "dei_ipc_data.h"

/**
 * Create a Linux domain socket that can be later used to connect to another application.
 *
 * @param p_myAddress - The address of your application. This is used by the other application to determine where the
 * 	requests are coming from. Use a more or less random address at each application start (e.g. with added process ID => see examples)
 * @param p_serverAddress - Address of the other application (e.g. "/tmp/hardwareDaemon" or "/tmp/pclient_<PROJECT_NAME>")
 * @param p_serverAdressStruct - Pointer to a sockaddr_un struct needed later for the connect() call
 *
 * @returns -1 on failure (e.g. creation of socket failed or bind to myAddress failed) or the socket (value > 0) if everything worked.
 *
 * See also createSocketAndConnect() function that will also do the connect for you.
 */
int createSocket(const std::string & p_myAddress, const std::string & p_serverAddress, struct sockaddr_un * p_serverAddressStruct);

/**
 * Same as createSocket() above but also includes the connect() call.
 * When this function returns with a value > 0, the socket can be used to send and receive
 * IPCMessages from the server application.
 *
 * @returns -1 if internal call to createSocket() failed
 * @returns -2 if connect failed
 * @returns the socked file descriptor that can be used for sending and receiving (value > 0)
 */
int createSocketAndConnect(const std::string & p_myAddress, const std::string & p_serverAddress);

/**
 * Send a IPC message to the provided socket
 *
 * @param p_sock - Socket to the IPC counterpart
 * @param p_message - Pointer to the IPCMessage to send. Make sure that all data in the message is already filled.
 * @param p_timeoutMS - A timeout in milliseconds. The function will call select() on the socket and waits until it
 * 	is writable for the specified amount of milliseconds.
 *
 * @returns The number of bytes sent (positive number if the complete message was sent)
 * @returns -1 if the connection is broken. The socket needs to be closed and re-created and re-connected
 * @returns -2 if an unknown error happened during sending (check errno)
 * @returns -3 if the connection is broken. The socket needs to be closed and re-created and re-connected
 * @returns -4 if an unknown error happened during sending but 1 or more byte were already sent (checkk errno)
 * @returns -5 the select call timed out (e.g. socket was not ready for write within the specified timeout)
 * @returns -6 the select call returned with an error (check errno)
 * @returns -7 if an invalid socket was provided (socket < 0)
 * @returns -8 if a nullptr pointer was provided as message
 * @returns -9 if the dataLength of the provided message is == 0
 * @returns - 10 if a nullptr pointer was provided as data array of message
 */
int sendMessage(int p_sock, IPCMessage * p_message, int p_timeoutMS);

/**
 * Receive messages from a socket
 *
 * @param p_sock - The socket to receive messages from
 * @param p_message - Pointer to a IPCMessage struct that will be filled if a message was received
 * @param p_timoutMS - A timeout in milliseconds. The function will call select() on the socket and waits until it
 * 	is readable for the specified amount of milliseconds.
 *
 * @returns The number of bytes received (positive number on success => message parameter is filled with data)
 * @returns -1 If receiving the data part of the message failed (check errno)
 * @returns -2 If the socket connection is broken. The socket needs to be closed and re-created and re-connected
 * @returns -3 If the received data does not match the data size of the to-be-received IPC message
 * @returns -4 If receiving the header part of the message failed (check errno)
 * @returns -5 If the socket connection is broken. The socket needs to be closed and re-created and re-connected
 * @returns -6 If the first received bytes do not match the size of an IPC message header
 * @returns -7 If the select call timed out (e.g. socket was not ready for read within the specified timeout)
 * @returns -8 If the select call returned with an error (check errno)
 * @returns -9 If an invalid socket was provided (socket < 0)
 * @returns -10 If a nullptr pointer was provided as message
 * @returns -11 If the dataLength of the provided message is == 0
 * @returns -12 If a nullptr pointer was provided as data array of message
 */
int receiveMessage(int p_sock, IPCMessage * p_message, long p_timeoutMS);

/**
 * Subscribe to variable value updates of another application.
 * This is a helper function that creates a IPCMessage and IPCData for the subscription request
 * and sends it via sendMessage()
 *
 * @param p_sock - The socket connected to the other application
 * @param p_index - Index of the variable to subscribe to
 * @param p_subindex - Subindex of the variable to subscribe to
 * @param p_dataType - The data type of the variable
 * @param p_timeoutMS - A timeout in milliseconds. The function will call select() on the socket and waits until it
 * 	is writable for the specified amount of milliseconds.
 *
 * @returns 0 if a bad socket (<0) was provided or the return value of sendMessage() function (see above, value > 0 means possible success)
 */
int subscribeToVariable(int p_sock, unsigned short p_index, unsigned short p_subindex, data_type p_dataType, int p_timeoutMS);

/**
 * Send a request to change the value of an integer variable
 * This is a helper function that creates a IPCMessage and IPCData for the value change request
 * and sends it via sendMessage()
 *
 * @param p_sock - The socket connected to the other application
 * @param p_index - Index of the variable to subscribe to
 * @param p_subindex - Subindex of the variable to subscribe to
 * @param p_newValue - The new integer value that shall be set
 * @param p_timeoutMS - A timeout in milliseconds. The function will call select() on the socket and waits until it
 * 	is writable for the specified amount of milliseconds.
 *
 * @returns 0 if a bad socket (<0) was provided or the return value of sendMessage() function (see above, value > 0 means possible success)
 */
int requestValueChange(int p_sock, unsigned short p_index, unsigned short p_subindex, int p_newValue, int p_timeoutMS);

/**
 * Send a request to change the value of a string variable
 * This is a helper function that creates a IPCMessage and IPCData for the value change request
 * and sends it via sendMessage()
 *
 * @param p_sock - The socket connected to the other application
 * @param p_index - Index of the variable to subscribe to
 * @param p_subindex - Subindex of the variable to subscribe to
 * @param p_newValue - The new string value that shall be set
 * @param p_timeoutMS - A timeout in milliseconds. The function will call select() on the socket and waits until it
 * 	is writable for the specified amount of milliseconds.
 *
 * @returns 0 if a bad socket (<0) was provided or the return value of sendMessage() function (see above, value > 0 means possible success)
 */
int requestValueChange(int p_sock, unsigned short p_index, unsigned short p_subindex, const std::string & p_newValue, int p_timeoutMS);

/**
 * Send a request to get the value of a variable
 * This is a helper function that creates a IPCMessage and IPCData for the value get request
 * and sends it via sendMessage()
 *
 * @param p_sock - The socket connected to the other application
 * @param p_index - Index of the variable to subscribe to
 * @param p_subindex - Subindex of the variable to subscribe to
 * @param p_dataType - The data type of the variable
 * @param p_timeoutMS - A timeout in milliseconds. The function will call select() on the socket and waits until it
 * 	is writable for the specified amount of milliseconds.
 *
 * @returns 0 if a bad socket (<0) was provided or the return value of sendMessage() function (see above, value > 0 means possible success)
 */
int requestValue(int p_sock, unsigned short p_index, unsigned short p_subindex, data_type p_dataType, int p_timeoutMS);

/**
 * Use the getBufferSize function to check how long a serialized character buffer needs to be to fit
 * the data of the provided IPCData.
 * Use this function to know how much memory needs to be allocated for the IPCMessage::data
 * character array:
 * 	struct IPCData data;
 *	struct IPCMessage msg;
 *	data.index = 0x2010;
 *	data.subIndex = 0x02;
 *	data.type = DATA_TYPE_INT;
 *	msg.messageType = EVENT_SUBSCRIPTION_REQUEST;
 *	msg.dataLength = getBufferSize(&data);
 *	msg.data = new unsigned char[msg.dataLength];
 *	convertToBuffer(msg.data, &data);
 *	//message can be sent now
 *
 * @param p_data - Pointer to the filled IPCdata struct for which to check the needed size
 *
 * @returns - The needed number of bytes to serialize data (or 0 if data is a nullptr pointer)
 */
unsigned int getBufferSize(IPCData * p_data);

/**
 * Use convertToBuffer to serialize an IPCData struct into a character array
 *
 * 	struct IPCData data;
 *	struct IPCMessage msg;
 *	data.index = 0x2010;
 *	data.subIndex = 0x02;
 *	data.type = DATA_TYPE_INT;
 *	msg.messageType = EVENT_SUBSCRIPTION_REQUEST;
 *	msg.dataLength = getBufferSize(&data);
 *	msg.data = new unsigned char[msg.dataLength];
 *	convertToBuffer(msg.data, &data);
 *	//message can be sent now
 *
 * @param p_buffer - Pointer to an already allocated character array
 * 	with enough space to fit the serialized data (see example above)
 * @param p_data - Pointer to the IPCData struct that shall be serialized
 *
 * @returns The size in bytes that were serialized or 0 on error (e.g. nullptr parameter or invalid data type in IPCData)
 */
unsigned int convertToBuffer(unsigned char * p_buffer, IPCData * p_data);

/**
 * Use dataFromBuffer to convert the IPCMessage::data buffer to a IPCData struct after receiving an IPCMessage
 *
 * IPCMessage msg;
 * receiveMessage(sock, &msg, 100);
 * IPCData dat;
 * dataFromBuffer(msg.data, msg.dataLength, &dat);
 *
 * @param p_buffer - Pointer to IPCMessage::data
 * @param p_bufferLength - Size of the buffer (IPCMessage::dataLength)
 * @param p_data - Pointer to an IPCData struct were the data from the buffer will be extracted into
 *
 * @returns a value > 0 if data was extracted or 0 on error (e.g. nullptr parameter or
 * 	too small bufferLength to extract the IPCData from the buffer)
 *
 */
unsigned int dataFromBuffer(unsigned char * p_buffer, unsigned int p_bufferLength, IPCData * p_data);


#endif /* SRC_DEI_DEI_IPC_API_DEI_IPC_FUNCTIONS_H_ */
